

/**
 * Class BlockOutputStream is used to implement the block access the
 * data on disk. The instances of this class are constructed with
 * two arguments, disk file name and buffer size.
 *
 * From the IO test results, the buffer size should be between 16K
 * and 64K, which gives a good tradeoff between performance and space.
 *
 * @author Biao Hao (bhao@cs.mcgill.ca)
 * @version 0.5, 07/04/1997
 */

import java.io.*;

public class BlockOutputStream extends DataOutputStream
implements TypeConstants
{


  /* to deal with DK and DC for string/integer/long/double
   * if DK, then writebyte(2) -- ^B in the file
   * if DC, then writebyte(5) -- ^E in the file
   * -- Weizhong
   */

   public BlockOutputStream(String filename, int buffersize)
   throws IOException
   {
	  super(new BufferedOutputStream(
		 new FileOutputStream(filename), buffersize));
   }      

   // Our string is stored as : short + byte array
   // Note there is shift of short value for the length of
   // string, so that the maximum length of string is
   // 2*Short.MAX_VALUE - 1 = 64*1024 - 3
   public void writeString(String s)
   throws IOException
   {
	   if (s.equals(STRING_DK))
		super.writeByte(2);		// Weizhong
	   else if (s.equals(STRING_DC))
		super.writeByte(5);		// Weizhong
	  else {
		 super.writeBytes(s);
	  }
   }      
   // Our text is stored as : int + byte array
   // Note there is no value shift for the length of test, so that
   // the maximum length of text is Integer.MAX_VALUE.
   public void writeText(String s)
   throws IOException
   {
	   if (s.equals(STRING_DK))
		super.writeByte(2);		// Weizhong
	   else if (s.equals(STRING_DC))
		super.writeByte(5);		// Weizhong
	  else {
		 super.writeBytes(s);
	  }
   }      

   // Weizhong, for writing Integer/long/double... DC
   public void writeMyBytes(int Type, String s)
   throws IOException
   {
	String sDK = new String("");
	String sDC = new String("");

	switch(Type) {
	case INTEGER :
	  sDK = sDK.valueOf(INT_DK);
	  sDC = sDC.valueOf(INT_DC);
	  break;
	case STRING :	 
	case STMT :
	case EXPR :
	  // have been dealed with in writeString function
	  sDK = sDK.valueOf(STRING_DK);
	  sDC = sDC.valueOf(STRING_DC); 
	  break;
	case BOOLEAN :
	  sDK = sDK.valueOf(BOOL_DK);
	  sDC = sDC.valueOf(BOOL_DC);
	  break;
	case SHORT :
	  sDK = sDK.valueOf(SHORT_DK);
	  sDC = sDC.valueOf(SHORT_DC);
	  break;
	case LONG :
	case IDLIST :
	  sDK = sDK.valueOf(LONG_DK);
	  sDC = sDC.valueOf(LONG_DC);
	  break;
	case TEXT :
	  // have been dealed with in writeText function
	  sDK = sDK.valueOf(STRING_DK);
	  sDC = sDC.valueOf(STRING_DC);
	  break;
	case FLOAT :
	  sDK = sDK.valueOf(FLOAT_DK);
	  sDC = sDC.valueOf(FLOAT_DC);
	  break;
	case DOUBLE :
	  sDK = sDK.valueOf(DOUBLE_DK);
	  sDC = sDC.valueOf(DOUBLE_DC);
	  break;
	}

	   if (s.equals(sDK))
		super.writeByte(2);		// Weizhong
	   else if (s.equals(sDC))
		super.writeByte(5);		// Weizhong
	  else {
		 super.writeBytes(s);
	  }
   }      
}
