/**
 * Static class with methods that return a byte value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */


import java.util.*;


public class CastToBoolByte implements Constants, TypeConstants {  
    //static private final byte RESULT_TYPE_DC = BOOL_DC;
    //static private final byte RESULT_TYPE_DK = BOOL_DK;
    public static byte assertValidValue(byte castFromBoolByte) 
        throws InterpretError {
        return value(castFromBoolByte);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     *
     * @return  a boolean true if the RHS type can be cast to LHS type.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
            // We must provide a value() method for the compatible types.
        case DK:	//added by Yi Zheng Sept 28, 2002
        case DC:	//added by Yi Zheng Sept 28, 2002
            return true;

        case STRING:
        case TEXT:
            return true;

        case STMT:
        case EXPR:
            return false;

        case IDLIST:
            // handled in Actualizer.
        case UNIVERSAL:  // moved here by Fan Guo September 29, 2004
            return true;

        case COMP:
    //    case DC: //Yi Zheng Sept 28, 2002
    //    case DK: //Yi Zheng Sept 28, 2002
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
     // case UNIVERSAL:   // moved up, return true by Fan
        default:
            return false;
        }
    }

    public static byte value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return (byte) BOOL_DK;
        else if (castFromStrn.equals(STRING_DC))
            return (byte) BOOL_DC;
        else if (Boolean.valueOf(castFromStrn).booleanValue() == true) 
            return BOOL_TRUE;
        else
            return BOOL_FALSE;
    }

    public static byte value(byte castFromBoolByte) 
        throws InterpretError {
        if ((castFromBoolByte == (byte) BOOL_TRUE) 
            || (castFromBoolByte == (byte) BOOL_FALSE) 
            || (castFromBoolByte == (byte) BOOL_DK)
            || (castFromBoolByte == (byte) BOOL_DC))
            return castFromBoolByte;
        else 
            throw new InterpretError(
                    "CastToBoolByte.value: error: Can't cast from unknown BOOL.");
    }

    public static byte value(int castFromInt) {
        if (castFromInt == INT_DK)
            return (byte) BOOL_DK;
        else if (castFromInt == INT_DC)
            return (byte) BOOL_DC;
        else if (castFromInt == 0)
            return BOOL_FALSE;
        else
            return BOOL_TRUE;
    }

    public static byte value(short castFromShort) {
        if (castFromShort == SHORT_DK)
            return (byte) BOOL_DK;
        else if (castFromShort == SHORT_DC)
            return (byte) BOOL_DC;
        else if (castFromShort == 0)
            return BOOL_FALSE;
        else
            return BOOL_TRUE;
    }

    public static byte value(long castFromLong) {
        if (castFromLong == LONG_DK)
            return (byte) BOOL_DK;
        else if (castFromLong == LONG_DC)
            return (byte) BOOL_DC;
        else if (castFromLong == 0)
            return BOOL_FALSE;
        else
            return BOOL_TRUE;
    }

    public static byte value(double castFromDouble) 
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return (byte) BOOL_DC;
            else if (castFromDouble == DOUBLE_DK)
                return (byte) BOOL_DK;
            else if (Double.isInfinite(castFromDouble))
                throw new InterpretError(
                        "CastToBoolByte.value: error: Can't cast +/- Infinity.");
            else if (Double.isNaN(castFromDouble))
                throw new InterpretError(
                        "CastToBoolByte.value: error: Can't cast NaN.");
            else if (castFromDouble == 0.0)
                return BOOL_FALSE;
            else
                return BOOL_TRUE;
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToBoolByte.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast with error: " + e);
        }
    }

    public static byte value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return (byte) BOOL_DC;
            else if (castFromFloat == FLOAT_DK)
                return (byte) BOOL_DK;
            else if (Float.isInfinite(castFromFloat))
                throw new InterpretError(
                        "CastToBoolByte.value: error: Can't cast +/- Infinity.");
            else if (Float.isNaN(castFromFloat))
                throw new InterpretError(
                        "CastToBoolByte.value: error: Can't cast NaN.");
            else if (castFromFloat == 0.0)
                return BOOL_FALSE;
            else
                return BOOL_TRUE;
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToBoolByte.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast with error: " + e);
        }
    }

    public static byte value(boolean castFromBoolean) {
        if (castFromBoolean == true) 
            return (byte) BOOL_TRUE; 
        else //if (castFromBoolean == false) 
            return (byte) BOOL_FALSE;
    }

    public static byte value(number castFromNumber) 
        throws InterpretError {
        if (castFromNumber.iszero)
            return (byte) BOOL_FALSE;
        else
            return (byte) BOOL_TRUE; 
        // Does not handle Dc/Dk yet.
        //throw new InterpretError(
                //"CastToBoolByte.value(): error: Can't cast from castFromNumber: Not yet implemented.");
    }
}
