

/**
 * Static class with methods that return a double value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */
//yiyi modified Sept'03
import java.util.*;


public class CastToDouble implements Constants, TypeConstants {  
    static private final double RESULT_TYPE_DC = DOUBLE_DC;
    static private final double RESULT_TYPE_DK = DOUBLE_DK;
    public static double assertValidValue(double castFromDouble) 
        throws InterpretError {
        return value(castFromDouble);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     * We must provide a value() method for all the castable types.
     *
     * @return  a boolean true if the RHS type can be cast double type.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DK:	//added by Yi Zheng Sept 28, 2002
        case DC:	//added by Yi Zheng Sept 28, 2002
        case UNIVERSAL:  // moved here by Fan Guo September 29, 2004
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
 //       case DC:	//Yi Zheng Sept 28, 2002
 //       case DK: //Yi Zheng Sept 28, 2002
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
     //   case UNIVERSAL:   // moved up, return true by Fan
        default:
            return false;
        }
    }

    public static double value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return RESULT_TYPE_DK;
        else if (castFromStrn.equals(STRING_DC))
            return RESULT_TYPE_DC;
        double retVal = value(CastToNum.value(castFromStrn));
       /* if (retVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (retVal == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");
        */
        if(retVal == RESULT_TYPE_DC || retVal == RESULT_TYPE_DK)
           return RESULT_TYPE_DK;
        else 
            return retVal;
    }

    public static double value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return RESULT_TYPE_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return RESULT_TYPE_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return 0;
        else return 1;
    }

    public static double value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DC)
            return RESULT_TYPE_DC;
        else if (castFromLong == LONG_DK)
            return RESULT_TYPE_DK;
        /*else if ((double) castFromLong == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((double) castFromLong == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");
         */
        else if((double) castFromLong == RESULT_TYPE_DC || (double) castFromLong == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((double) castFromLong);
    }

    public static double value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DC)
            return RESULT_TYPE_DC;
        else if (castFromInt == INT_DK)
            return RESULT_TYPE_DK;
       /* else if ((double) castFromInt == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((double) castFromInt == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");
        */
        else if((double) castFromInt == RESULT_TYPE_DC || (double) castFromInt == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((double) castFromInt);
    }

    public static double value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DC)
            return RESULT_TYPE_DC;
        else if (castFromShort == SHORT_DK)
            return RESULT_TYPE_DK;
        /*else if ((double) castFromShort == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((double) castFromShort == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");
         */
        else if((double) castFromShort == RESULT_TYPE_DC || (double) castFromShort == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((double) castFromShort);
    }

    public static double value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return RESULT_TYPE_DC;
            else if (castFromFloat == FLOAT_DK)
                return RESULT_TYPE_DK;
            /*else if ((double) castFromFloat == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((double) castFromFloat == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");
             */
            else if((double) castFromFloat == RESULT_TYPE_DC || (double) castFromFloat == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
            else if (Double.isInfinite((double) castFromFloat))
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast to +/- Infinity.");
            else if (Double.isNaN((double) castFromFloat))
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast NaN.");
            else {
                return ((double) castFromFloat);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToDouble.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast to:" + (new Double((double) castFromFloat)).toString() + " with error: " + e);
        }
    }

    public static double value(double castFromDouble)
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return RESULT_TYPE_DC;
            else if (castFromDouble == DOUBLE_DK)
                return RESULT_TYPE_DK;
            /*else if ((double) castFromDouble == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((double) castFromDouble == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");
             */
            else if((double) castFromDouble == RESULT_TYPE_DC || (double) castFromDouble == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
            else if (Double.isInfinite((double) castFromDouble))
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast to +/- Infinity.");
            else if (Double.isNaN((double) castFromDouble))
                throw new InterpretError(
                        "CastToDouble.value: error: Can't cast NaN.");
            else {
                return ((double) castFromDouble);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToDouble.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast to:" + (new Double((double) castFromDouble)).toString() + " with error: " + e);
        }
    }

    //yiyi revisited sept'03
    public static double value(number castFromNumber) 
        throws InterpretError {
        //yiyi added            
        //if (castFromNumber.iszero)
            
           // System.out.println("original = " + castFromNumber.toString());
            
        if(number.iszero(castFromNumber))
            return 0;
        
        else if(castFromNumber.compareTo(NUMERIC_DK) == 0)
            return RESULT_TYPE_DK;
            
        else if(castFromNumber.compareTo(NUMERIC_DC) == 0)
            return RESULT_TYPE_DC;
        
        else if((number.abs(castFromNumber)).compareTo(CastToNum.value(Double.MIN_VALUE)) < 0 ){
           // System.out.println(Double.MIN_VALUE);
           // System.out.println(castFromNumber.compareTo(CastToNum.value(Double.MIN_VALUE)));
            return RESULT_TYPE_DK;
        }
        else{
            double doubleVal = Double.parseDouble(CastToStrn.value(castFromNumber));
            
           // System.out.println("casted = " + doubleVal);

            if(doubleVal == Double.NEGATIVE_INFINITY || doubleVal == Double.POSITIVE_INFINITY)
                return RESULT_TYPE_DK;
        
        
        /*if (doubleVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (doubleVal == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToDouble.value: error: Can't cast to RESULT_TYPE_DK.");*/
            if(doubleVal == RESULT_TYPE_DC || doubleVal == RESULT_TYPE_DK)
                return RESULT_TYPE_DK;
        //end
            return doubleVal;
        }
    }
}
