

/**
 * Static class with methods that return a float value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */


import java.util.*;


public class CastToFloat implements Constants, TypeConstants {  
    static private final float RESULT_TYPE_DC = FLOAT_DC;
    static private final float RESULT_TYPE_DK = FLOAT_DK;
    public static float assertValidValue(float castFromFloat) 
        throws InterpretError {
        return value(castFromFloat);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     * We must provide a value() method for each castable type.
     *
     * @return  a boolean true if the RHS type is castable.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DC:	//added by Yi Zheng Sept 28, 2002
        case DK:	//added by Yi Zheng Sept 28, 2002
        case UNIVERSAL:  // moved here by Fan Guo September 29, 2004
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
 //       case DC:	//Yi Zheng Sept 28, 2002
 //       case DK:  //Yi Zheng Sept 28, 2002
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
     // case UNIVERSAL:   // moved up, return true by Fan
        default:
            return false;
        }
    }

    /**
     * Returns a float value from a string.
     *
     * @return  a float corresponding to the first numeric
     *          string in the entire castFromStrn.
     *          Uses the standard Java methods to convert the
     *          numeric string to a number.
     * @since jRelCaNu
     */
    public static float value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return RESULT_TYPE_DK;
        else if (castFromStrn.equals(STRING_DC))
            return RESULT_TYPE_DC;
        float retVal = value(CastToDouble.value(castFromStrn));
        /*if (retVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (retVal == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DK.");
         */
        if(retVal == RESULT_TYPE_DC || retVal == RESULT_TYPE_DK)
           return RESULT_TYPE_DK;
        else 
            return retVal;
    }

    public static float value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return RESULT_TYPE_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return RESULT_TYPE_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return 0;
        else return 1;
    }

    public static float value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DC)
            return RESULT_TYPE_DC;
        else if (castFromLong == LONG_DK)
            return RESULT_TYPE_DK;
        /*else if ((float) castFromLong == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((float) castFromLong == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DK.");
         */
        else if(Math.abs(castFromLong) < Float.MIN_VALUE || (float)castFromLong == Float.POSITIVE_INFINITY
                ||(float)castFromLong == Float.NEGATIVE_INFINITY )
            return RESULT_TYPE_DK;
        else if((float) castFromLong == RESULT_TYPE_DC || (float) castFromLong == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((float) castFromLong);
    }

    public static float value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DC)
            return RESULT_TYPE_DC;
        else if (castFromInt == INT_DK)
            return RESULT_TYPE_DK;
       /* else if ((float) castFromInt == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((float) castFromInt == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DK.");
        */
        else if(Math.abs(castFromInt) < Float.MIN_VALUE || (float)castFromInt == Float.POSITIVE_INFINITY
                ||(float)castFromInt == Float.NEGATIVE_INFINITY)
            return RESULT_TYPE_DK;
        else if((float) castFromInt == RESULT_TYPE_DC || (float) castFromInt == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((float) castFromInt);
    }

    public static float value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DC)
            return RESULT_TYPE_DC;
        else if (castFromShort == SHORT_DK)
            return RESULT_TYPE_DK;
        /*else if ((float) castFromShort == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((float) castFromShort == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToFloat.value: error: Can't cast to RESULT_TYPE_DK.");*/
        else if((float) castFromShort == RESULT_TYPE_DC || (float) castFromShort == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((float) castFromShort);
    }

    public static float value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return RESULT_TYPE_DC;
            else if (castFromFloat == FLOAT_DK)
                return RESULT_TYPE_DK;
            /*else if ((float) castFromFloat == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((float) castFromFloat == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast to RESULT_TYPE_DK.");
             */
        else if((float) castFromFloat == RESULT_TYPE_DC || (float) castFromFloat == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
            else if (Float.isInfinite((float) castFromFloat))
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast to +/- Infinity.");
            else if (Float.isNaN((float) castFromFloat))
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast NaN.");
            else {
                return ((float) castFromFloat);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToFloat.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast to:" + (new Float((float) castFromFloat)).toString() + " with error: " + e);
        }
    }

    public static float value(double castFromDouble)
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return RESULT_TYPE_DC;
            else if (castFromDouble == DOUBLE_DK)
                return RESULT_TYPE_DK;
            /*else if ((float) castFromDouble == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((float) castFromDouble == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast to RESULT_TYPE_DK.");
             */
        else if(Math.abs(castFromDouble) < Float.MIN_VALUE || (float)castFromDouble == Float.POSITIVE_INFINITY
                ||(float)castFromDouble == Float.NEGATIVE_INFINITY )
            return RESULT_TYPE_DK;
        else if((float) castFromDouble == RESULT_TYPE_DC || (float) castFromDouble == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
            else if (Float.isInfinite((float) castFromDouble))
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast to +/- Infinity.");
            else if (Float.isNaN((float) castFromDouble))
                throw new InterpretError(
                        "CastToFloat.value: error: Can't cast NaN.");
            else {
                return ((float) castFromDouble);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToFloat.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast to:" + (new Float((float) castFromDouble)).toString() + " with error: " + e);
        }
    }

    public static float value(number castFromNumber) 
        throws InterpretError {
            return value(CastToDouble.value(castFromNumber));
    }
}
