

/**
 * Static class with methods that return a long value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */

//yiyi modified Sept'03

import java.util.*;


public class CastToLong implements Constants, TypeConstants {  
    static private final long RESULT_TYPE_DC = LONG_DC;
    static private final long RESULT_TYPE_DK = LONG_DK;
    public static long assertValidValue(long castFromLong) 
        throws InterpretError {
        return value(castFromLong);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     * We must provide a value() method for each compatible type.
     *
     * @return  a boolean true if the RHS type can be cast to LHS type.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DK:	//added by Yi Zheng Sept 28, 2002
        case DC:	//added by Yi Zheng Sept 28, 2002
        case UNIVERSAL:  // moved here by Fan Guo September 29, 2004
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
//        case DC:
 //       case DK:
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
     // case UNIVERSAL:   // moved up, return true by Fan
        default:
            return false;
        }
    }

    /**
     * Returns a long value from a string.
     *
     * @return  an long corresponding to the first numeric
     *          string in the entire castFromStrn.
     *          Uses the standard Java methods to convert the
     *          numeric string to a number.
     * @since jRelCaNu
     */
    public static long value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return RESULT_TYPE_DK;
        else if (castFromStrn.equals(STRING_DC))
            return RESULT_TYPE_DC;
        long retVal = value(CastToNum.value(castFromStrn));
     /*   if (retVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (retVal == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");
      */
        if(retVal == RESULT_TYPE_DC || retVal == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return retVal;
    }

    public static long value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return RESULT_TYPE_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return RESULT_TYPE_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return 0;
        else return 1;
    }

    public static long value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DC)
            return RESULT_TYPE_DC;
        else if (castFromLong == LONG_DK)
            return RESULT_TYPE_DK;
       /* else if ((long) castFromLong == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((long) castFromLong == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");
        */
        else if ((long) castFromLong == RESULT_TYPE_DC || (long) castFromLong == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((long) castFromLong);
    }

    public static long value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DC)
            return RESULT_TYPE_DC;
        else if (castFromInt == INT_DK)
            return RESULT_TYPE_DK;
        /*else if ((long) castFromInt == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((long) castFromInt == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");*/
        else if((long) castFromInt == RESULT_TYPE_DC || (long) castFromInt == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((long) castFromInt);
    }

    public static long value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DC)
            return RESULT_TYPE_DC;
        else if (castFromShort == SHORT_DK)
            return RESULT_TYPE_DK;
       /* else if ((long) castFromShort == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
        else if ((long) castFromShort == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");
        */
        else if((long) castFromShort == RESULT_TYPE_DC || (long) castFromShort == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        else 
            return ((long) castFromShort);
    }

    public static long value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return RESULT_TYPE_DC;
            else if (castFromFloat == FLOAT_DK)
                return RESULT_TYPE_DK;
            /*else if ((long) castFromFloat == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((long) castFromFloat == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");
             */
            else if((long) castFromFloat == RESULT_TYPE_DC || (long) castFromFloat == RESULT_TYPE_DK)
                return RESULT_TYPE_DK;
            else if (Float.isInfinite(castFromFloat))
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast to +/- Infinity.");
            else if (Float.isNaN(castFromFloat))
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast NaN.");
            else {
                return ((long) castFromFloat);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToLong.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast to:" + (new Long((long) castFromFloat)).toString() + " with error: " + e);
        }
    }

    public static long value(double castFromDouble)
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return RESULT_TYPE_DC;
            else if (castFromDouble == DOUBLE_DK)
                return RESULT_TYPE_DK;
           /* else if ((long) castFromDouble == RESULT_TYPE_DC)
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
            else if ((long) castFromDouble == RESULT_TYPE_DK)
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");
            */
            else if((long) castFromDouble == RESULT_TYPE_DC || (long) castFromDouble == RESULT_TYPE_DK)
                return RESULT_TYPE_DK;
            else if (Double.isInfinite(castFromDouble))
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast to +/- Infinity.");
            else if (Double.isNaN(castFromDouble))
                throw new InterpretError(
                        "CastToLong.value: error: Can't cast NaN.");
            else {
                return ((long) castFromDouble);
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToLong.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast to:" + (new Long((long) castFromDouble)).toString() + " with error: " + e);
        }
    }

    //yiyi revisited Sept'03
    public static long value(number castFromNumber) 
        throws InterpretError {

            //yiyi added
//        if (castFromNumber.iszero) yiyi revisited
        if(number.iszero(castFromNumber))
            return 0;

        else if(castFromNumber.compareTo(NUMERIC_DK) == 0)
            return RESULT_TYPE_DK;
            
        else if(castFromNumber.compareTo(NUMERIC_DC) == 0)
            return RESULT_TYPE_DC;
                
        else if((number.abs(castFromNumber)).compareTo(CastToNum.value(Long.MAX_VALUE)) > 0)
            return RESULT_TYPE_DK;
        else{
            long longVal = 0;

            for (int i = 0; i < castFromNumber.expo; i++) {
               longVal *= 10;
                if (i < castFromNumber.fractionLength)
                  if (castFromNumber.sign == '-')
                       longVal -= (long) castFromNumber.fraction[i];
                   else
                       longVal += (long) castFromNumber.fraction[i];
        }
        //yiyi added, OUT OF RANGE NUMBER, e.g. 0.0098
          if(longVal == 0)
             return RESULT_TYPE_DK;
            
        if(longVal == RESULT_TYPE_DC || longVal == RESULT_TYPE_DK)
            return RESULT_TYPE_DK;
        
       /* if (longVal == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DC.");
        else if (longVal == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToLong.value: error: Can't cast to RESULT_TYPE_DK.");*/
        return longVal;
        }
    }
}
