import java.util.*;

/**
 * Class encapsulating a single entry in .dom system table.
 * These entries are described in the documentation for the class
 * DomTable.
 */

public class Domain implements TypeConstants, Constants
{
	public String name;    //name of this domain
   	public int type;       //type of this domain
   	public int numref;     //a reference count of the number of relations
						  //defined on this domain,
   	public SimpleNode tree;//sytax tree if the domain is virtual, else null
	public int isStateDom = 0;//1 if this domain is declared as a state var of a computation

	//Yi Zheng April 23, 2002
	public int isRedTemp = 0; //1 if this is a temporary representing a reduction used for level lifting, Yi Zheng, June 28, 2002
	public int isRedTempTop = 0; //1 if this is a temporary domain as in the following case: A <- [red ujoin of U] in R;
//	public long surr = 0;	//surrogate value for comp domain (assumption: no polymorphism) Yi Zheng May 10, 2002

        // Yu
	public SimpleNode[] union;		//store union types		June 28,2004 by Yu
  	public static final String temp_dname_prefix = "_temp_d_";
  	private static final int MAX_TEMP_FILE = Integer.MAX_VALUE;
  	private static int next_number = 0;
  	private static final int MAX_INITIALIZE_SIZE = 10000;

// constructor -- added by Zongyan 	Aug, 2002
	public Domain ()
	{
		name = null;
	    type = 0;
	}
	/**
	 * duplicate this domain (not IDLIST or COMP)
	 * @author Yi Zheng May 17, 2002
	 */
	public Domain duplicate(Environment env) throws InterpretError
	{
		Domain nd = null;
/*		if(this.type == IDLIST || this.type == COMP)
		{
			nd = new Domain(this.name, this.type, this.numref, null);
			Relation rel = env.lookupRel("." + this.name, true);
			if (rel == null)
				throw new InterpretError("Relation ." + this.name + " not found");
			Relation drel = rel.duplicate(true);
			env.put(drel);
			if(drel.data == null) drel.load();
			drel.save();
		}
		else*/
//		{
			nd = new Domain(this.name, this.type, this.numref,
				(SimpleNode)(((SimpleNode) this.tree).jjtDuplicate()));
//		}
		return nd;
	}
   /**
	* This constructor is used for domains of non-primitive type.
	*   ex1: domain X(A,B);
	*   ex2: domain C comp(I,O);
	*/
  public Domain(String name, int type, Domain[] domlist, Environment env)
   throws InterpretError
   {  this(name, type, 0, null);

	  // Certain names are reserved by the system and should not be
	  // used by the user.
	  if(name.equals("rel") || name.equals("dom") ||
		 name.equals("rd") || name.equals("comp"))
	  {  throw new InterpretError("'"+name+"' is a reserved word.");
	  }

	  // fill a Domain[] array for creating a new relation in reltable.
	  // Note: we can't increment the numref of domain in domlist
	  // here, since addRel() will take care of that.

		//if switch added by Yi Zheng April 23, 2002
	Domain [] doms = null;
	if(type == IDLIST) //Yi Zheng June 19, 2002
	{
		if(domlist != null)
		  doms = new Domain[domlist.length+1];
	    else
		  doms = new Domain[1];
	  	doms[0] = env.lookupDom(".id", true);
	  	if(domlist != null)
	  	{
	  		for(int i = 0; i < domlist.length; i++)
	  		{  doms[i+1] = domlist[i];

	  		}
	  	}
	}
	else if(type == COMP) //Yi Zheng June 19, 2002
	{
		if(domlist != null)
			doms = new Domain[domlist.length];
		else
			doms = new Domain[0];
		if(domlist != null)
		{
			for(int j = 0; j < domlist.length; j++)
				doms[j] = domlist[j];
		}
	}
	else
		throw new InterpretError("unknown domain type in domain declaration");

	  // Add it to relTable with idlist info.
	  Relation r = new Relation("."+name, doms, RELATION, 0,
								 doms.length, 0, null, env);

	  env.put(r);

   }

   /**
	* This constructor is used for domains of union type.
	*   ex: domain X(A,B)|integer|Y;
	*/
  public Domain(String name, int type, int numr, SimpleNode[] complist, Environment env)
   throws InterpretError
   {  this(name, type, numr, null);

	  // Certain names are reserved by the system and should not be
	  // used by the user.
	  if(name.equals("rel") || name.equals("dom") ||
		 name.equals("rd") || name.equals("comp"))
	  {  throw new InterpretError("'"+name+"' is a reserved word.");
	  }
	  union = complist;
   }

   /**
	* This constructor is used for domains of primitive type.
	*   e.g. domain a intg;
	*/
   public Domain(String nm, int tp, int ref, SimpleNode nd)
   {  name = nm;
	  type = tp;
	  numref = ref;
	  tree = nd;
   }

   //zhan /** recursive **/
   public Domain (String nm, int tp)
   {
   	  name = nm;
   	  type = tp;
   }
   //zhan

   /**
	* Get the syntax tree for virtual domain.
	*/
   public SimpleNode getTree()
   {  // if there is such domain and it is a virtual domain..
	  if(tree != null)
	  {  if(tree.jjtGetNumChildren() == 1)
		 {  return (SimpleNode) tree.jjtGetChild(0);

		 }
		 else if(tree.jjtGetNumChildren() == 2)
		 {  return (SimpleNode) tree.jjtGetChild(1);
		 }
		 else
		 {  // Should never happen, otherwise system error.
			Global.println("getTree(): system error, wrong number"
							  +" of children.");
			return null;
		 }
	  }
	  else
	  {  return null;
	  }
   }
   public String getType()
   {
   	String typeName="";
   	switch (type) 
   	{
   		case INTEGER:
   		typeName="integer";
   		break;
   		case BOOLEAN:
   		typeName="boolean";
   		break;
   		case SHORT:
   		typeName="short";
   		break;
   		case LONG:
   		typeName="long";
   		break;
   		case FLOAT:
   		typeName="float";
   		break;
   		case DOUBLE:
   		typeName="double";
   		break;
   		case NUMERIC:
   		typeName="numeric";
   		break;
   		case STRING:
   		typeName="string";
   		break;
   		case TEXT:
   		typeName="text";
   		break;
   		case ATTRIBUTE:
   		typeName="attribute";
   		break;
   		case UNIVERSAL:
   		typeName="universal";
   		break;
   		case TYPE:
   		typeName="type";
   		break;
   		case UNION:
   		typeName="union";
   		break;
   	}
   	 return typeName;	
   }

  public static String nextTempDomainName() //July 01, 2004 by Yu
  {
    int n = (next_number++)%MAX_TEMP_FILE;
    return temp_dname_prefix + Integer.toString(n);
  }

  public static void setNextNumber(int n)      //July 02, 2004 by Yu
  {
    next_number = n;
  }

   /*
    * Check if the domain type is scalar;
    */
   public boolean IsTypeScalar() // Fan Guo June 3 2004
   {
                boolean isScalar;

        if (type!=IDLIST) isScalar = true;
        else isScalar = false;

                return isScalar;
        }

}
