import java.util.*;
/**
 *  Class implementing a frame of an environment.
 *  The environment consists of entries of five different "kinds"
 *     PARAM
 *     LOCAL
 *     STATE
 *     REL
 *     DOM
 * These are subclasses of IDInfo which provide convenient storage for
 * each of these, respectively.
 */
class Environment implements Constants, TypeConstants
//class Environment extends Environment76
{
//debugging Yi Zheng
private boolean DEBUG=false;
/**
	*  Constructor called at boot time.
	*  It does disk I/O to load the RelTable and DomTable from disk.
	*/
   private Environment parent;   //Pointer to parent environment
   private NREnvironment NRenv;  //For use with domain algebra
   //Table holding information about params, locals and states
   private Hashtable table;
   //For holding local domains and relations
   public RelTable reltable;
   public DomTable domtable;
   public TextTable textTable; // Jiantao 2004
   public Object[] data; //used in computation
   public int row;//used in computation
   public boolean constraint_mode = false;//added by Yi Zheng March 31, 2002
   public TriggerNode triggerroot;
   // START Since jRelixNS: optionally turn on Nested-Sigma join code.
   //final public boolean     isSigmaJoinImplemented = true;
   // END Since jRelixNS.

// private boolean isExported; //limits parent pointer info to state info..
//	public boolean isNested = false;//Yi Zheng June 10, 2002, true if lookupRel returns a nested rel from NREnvironment
//	public String nameNested = null; //Yi Zheng June 10, 2002

   public Environment()
   {  table = new Hashtable();
	  parent = null;
	  NRenv = null;

	  //This is the top-level environment which is
	  //only created when the system boots
	  reltable = new RelTable(this, true);      //true: load from .rel
	  domtable = new DomTable(this, true);      //true: load from .dom
          textTable = new TextTable(this, true);        //true: load from .text by Jian tao

   }

   /**
    * set the parent environment
    * @author Yi Zheng, May 14, 2002
    */
   public void setParent(Environment pEnv)
  { parent = pEnv;
   }

   /**
    * get the parent environment
    * @author Yi Zheng, May 31, 2002
    */
   public Environment getParent()
  { return parent;
   }

   /**
	* General purpose constructor.
	* @parentEnv    Pointer to the parent frame of the environment
	*/
   public Environment(Environment parentEnv)
   {  table = new Hashtable();
	  parent = parentEnv;
	  NRenv = null;

	  reltable = new RelTable(this, false);     //false: no disk I/O
	  domtable = new DomTable(this, false);     //ditto
          textTable = new TextTable(this, true);        //true: load from .text

   }

   // Weizhong, for update command, to initialize the trigger root
   public Environment(RelTable prel, DomTable pdom)
   {
	table = new Hashtable();
	parent = null;
	NRenv = null;
	reltable = prel;
	domtable = pdom;
	triggerroot = new TriggerNode(this);
   }

	//Yi Zheng May 20, 2002, for use by updates in Computation block
   public Environment(Environment parent, Hashtable tbl, RelTable prel, DomTable pdom)
   { this(prel, pdom);
     this.table = tbl;
     this.parent = parent;
   }

   /****************************************************************/
   /*           METHODS FOR MANIPULATING THE NRENVIRONMENT         */
   /****************************************************************/

   public NREnvironment addNRFrame(Relation r)
   {  NREnvironment e = new NREnvironment(r, NRenv, this);
	  NRenv = e;
	  return e;
   }
   public void clear()
   {  //reltable.clear();
	  //domtable.clear();
	  //REVISIT: Anything else to do
   }
   /**
	* Provided for debugging purposes.
	*/
   public void dump(boolean dumpParent)
   throws InterpretError
   {Global.println("The environment contains "+table.size()+" entries:");
	  for(Enumeration e=table.keys(); e.hasMoreElements(); )
	  {  String name = (String) e.nextElement();
		 IDInfo id = (IDInfo) table.get(name);
		 if (id == null)
		 {  throw new InterpretError("table lookup failed.");
		 }
		 if (id.kind == PARAMETER)
		 {  ((ParamInfo) id).dump(name);
		 }
		 else if (id.kind == LOCAL)
		 {  ((LocalInfo) id).dump(name);
		 }
		 else if (id.kind == STATE)
		 {  ((StateInfo) id).dump(name);
		 }
	  }
	  for(Enumeration e=reltable.keys(); e.hasMoreElements(); )
	  {  String name = (String) e.nextElement();
		 Relation r = reltable.getRel(name);
		 Global.println("Rel "+r.name);
	  }
	  for(Enumeration e=domtable.keys(); e.hasMoreElements(); )
	  {  String name = (String) e.nextElement();
		 Domain d = domtable.getDom(name);
		 Global.println("Dom "+d.name);
	  }

	  if (dumpParent && parent!=null)
	  {  Global.println("----------------------------------------");
		 parent.dump(true);
	  }
   }
   /****************************************************************/
   /*                 METHODS NEEDED BY COMPUTATIONS               */
   /****************************************************************/

   public String[] getStateNamesLocal()
   {  int curr = 0;
	  String[] states = new String[table.size()];
	  for(Enumeration e=table.keys(); e.hasMoreElements(); )
	  {  String idName = (String) e.nextElement();
		 IDInfo id = (IDInfo) table.get(idName);
		 if(id.kind == STATE)
		 {  states[curr++] = idName;
		 }
	  }
	  String[] result = new String[curr];
	  for(int i=0; i<curr; i++)
	  {  result[i] = states[i];
	  }
	  return result;
   }
   /****************************************************************/
   /*                        LOOKUP ROUTINES                       */
   /****************************************************************/

   /**
	* This is the generic lookup method. It searches the whole
	* environment and returns what it finds as a generic IDInfo object.
	*/
   public IDInfo lookup(String name, boolean searchParent)
   {  IDInfo id;
	  Relation r;
	  Domain d;
	  //First search the NREnvironment
	  if (NRenv != null)
	  {  if ( (id = NRenv.lookup(name, true)) != null )
		 {  return id;
		 }
	  }
	  //next search in the table
	  if ( (id = (IDInfo) table.get(name)) != null )
	  {  return id;
	  }
	  //If not found, check reltable
	  if ( (r = reltable.getRel(name)) != null)
	  {  return new RelInfo(r);
	  }
	  //If not found, check domtable
	  if ( (d = domtable.getDom(name)) != null)
	  {  return new DomInfo(d);
	  }
	  if(searchParent && parent!=null)  //search parent env
	  {  return parent.lookup(name, true);
	  }
	  return null; //name is not in the env
   }

   /***start from zhan ***/
   public Domain lookupTryDom(String name, String z, boolean searchParent)
   {  Domain d;
      //System.out.println("name=" + name +"  z="+ z);//zhan
	if (name == null) return null; //added by Yi Zheng May 27, 2002 as a bug fix
	  //Search domtable
	  if ( (d = domtable.getDom(name)) != null)
	  {  return d;
	  }
	  if(searchParent && parent!=null)  //search parent env
	  {  return parent.lookupDom(name, true);
	  }
	  if(z.equals(name))
	  {
		Domain recur = new Domain(name,IDLIST);
        //System.out.println("return recursive");
        return recur;
      }
      //name is not in the env (or it is domain which is an error)
	  return null;
   }
   /*** end by zhan ***/


   /**
	* This routine searches the whole environment for a domain.
	*/
   public Domain lookupDom(String name, boolean searchParent)
   {  Domain d;
	if (name == null) return null; //added by Yi Zheng May 27, 2002 as a bug fix
	  //Search domtable
	  if ( (d = domtable.getDom(name)) != null)
	  {  return d;
	  }
	  if(searchParent && parent!=null)  //search parent env
	  {  return parent.lookupDom(name, true);
	  }

	  //name is not in the env (or it is domain which is an error)
	  return null;
   }

   //Yi Zheng June 20, 2002
   public ParamInfo lookupParam(String name)
   {
   	 IDInfo id = (IDInfo) table.get(name);
   	 if (id == null) return null;
   	 if (id instanceof ParamInfo)
   	 	return (ParamInfo) id;
   	 else return null;
  }
   /**
	* This routine only searches 'NRenv' for a binding. It does not
	* look in 'env'.
	*/
   public NRInfo lookupNR(String name, boolean searchParent)
   {  NRInfo nri;
	  if (NRenv != null)
	  {  if ( (nri = NRenv.lookup(name, searchParent)) != null )
		 {  return nri;
		 }
	  }
	  return null;
   }
   /**
	* This routine searches the environment for a parameter,
	*/
   public ParamInfo lookupParam(String name, boolean searchParent)
   {  IDInfo id = ((IDInfo) table.get(name));
	  if (id == null)
	  {  return null;
	  }
	  if (id.kind == PARAMETER)
	  {  return (ParamInfo) id;
	  }
	  if(searchParent && parent!=null)  //search parent env
	  {  return parent.lookupParam(name, true);
	  }
	  return null; // id isn't a parameter
   }
   /**
	* This routine searches the whole environment for a relation
	* (or computation or view).
	*/
   public Relation lookupRel(String name, boolean searchParent)
   throws InterpretError
   {  IDInfo id;
	  Relation r;
   //System.out.println("in lookupRel name = "+name);//zhan
/*	  if(DEBUG)
	  {
	  	System.out.println("reltable entries when entering lookupRel()");
		Hashtable relt = reltable.getRelTable();
		for(Enumeration enum = relt.keys();
			enum.hasMoreElements(); )
			{
				String reln = (String) enum.nextElement();
				System.out.println(reln);
//				Relation rele = reltable.getRel(reln);
//				rele.print();

			}
		}
		*/
	  //First search the NREnvironment
	  if (NRenv != null)
	  {
//		if(DEBUG)
//			System.out.println("in lookupRel, SEARCH NRenv");
	     if ( (r = NRenv.lookupRel(name))!= null )
		 {

		  return r;
		 }
	  }

	  //Next search in the table
	  if ( (id = (IDInfo) table.get(name)) != null )
	  {
	//	 if(DEBUG)
	//		System.out.println("in lookupRel, SEARCH table for name= " + name);

		//addition by Yi Zheng May 17, 2002
		//if(id.kind == LOCAL)
		//	return ((LocalInfo)id).relVal;

	     if(id.kind == PARAMETER)
		 {  ParamInfo pi = (ParamInfo) id;
			if(pi.type != IDLIST && pi.type != COMP) //REVISIT: Could be COMP not just IDLIST; Yi Zheng May 17, 2002
			{  return null;
			}
			//addition by Yi Zheng May 17, 2002
			else if(pi.type == COMP)
			{
				long sur = -1;

			   if(((ParamInfo)id).rel instanceof Computation)
				return ((ParamInfo)id).rel;
			   else if ((sur = ((long[]) pi.env.data[pi.index])[pi.env.row]) > 0)
			   {//ijoin
			   	Computation c = Global.comptable.lookup(sur);
			   	if (c != null)
			//   		return c;
			     return new Computation(c.tree, this); //Yi Zheng June 17, 2002, each comp passed through parameter list is considered a new one
			   	else throw new InterpretError("Computation parameter: " + name + " not found");

			   }
			   else
			   throw new InterpretError("Computation parameter not found");
			}
			//end of addition May 17, 2002

			//Need to extract the nested relation
			Environment e = pi.env;
			long surrNo  = ((long[]) e.data[pi.index])[e.row];
//			if(DEBUG)
//				System.out.println("in lookupRel, surrNo = " + surrNo);
			Relation dom = lookupRel("."+name, true);
/*
			if(DEBUG)
			{
				if (dom != null)
				{
					System.out.println("dom is not null");
					dom.print();
				}
				else System.out.println("dom is null!!!");

				if(pi.rel == null)
					System.out.println("about to extract, pi.rel is null!!!");
				else
					System.out.println("about to extract, pi.rel is not null");
			}
*/
		//	if(pi.rel == null)//added by Yi Zheng April 25, 2002 //commented out by Yi Zheng June 17, 2002
			{
//				System.out.println("****pi.rel is null*****");
				pi.rel = dom.duplicate(false);
				pi.rel.name = "";		//REVISIT
				pi.rel.numattrs--;
				pi.rel.domains = new Domain[pi.rel.numattrs];
				int jj=0;
				for (int kk = 0; kk < dom.numattrs; kk++)
				{
					if(dom.domains[kk].name.compareTo(".id") != 0)
						pi.rel.domains[jj++] = dom.domains[kk];
				}
				pi.rel.createDataArray(pi.rel.numtuples);
			}

			//DEBUGGING
			//System.out.println("about to extract, dom is: ");
			//dom.print();
			//System.out.println("surrNo = " + surrNo);


			dom.extract(surrNo, pi.rel);
			return pi.rel;
		 }
		 if( id.kind == STATE)
		 { // return null;   //REVISIT:Need extractRel() method
			return ((StateInfo)id).relVal; //Yi Zheng May 17, 2002

		 }
		 //"name" must be a local. Locals are
		 //supposed to be of primitive type. We return null
		 //Yi Zheng March 25, 2002, see Computation.processCallParams.OP_IDENTIFIER

//		 return null; commented out by Yi Zheng
//		if(id.kind == LOCAL)
//			return null; //REVISIT Yi Zheng April 25, 2002

		//addition by Yi Zheng May 17, 2002
		if(id.kind == LOCAL)
			return ((LocalInfo)id).relVal;
	  }
	  //If not found, check reltable

	  if ( (r = reltable.getRel(name)) != null)
	  {
//		if(DEBUG)
//			System.out.println("in lookupRel, SEARCH reltable");
	     return r;
	  }


//	  if(DEBUG)
//	  	System.out.println("No luck here, going for parent");

	  if(searchParent && parent!=null)  //search parent env
	  {
//System.out.println("in lookupRel, SEARCH parent");
		return parent.lookupRel(name, true);
	  }
	  //name is not a relation in the env (it could have been a domain
	  //but that is not what was being searched for)
	  return null;
   }
   /**
	* This routine searches the whole environment for a nested relation
	* (or computation or view).
	*  -- Weizhong Sun
	*/
   public Relation lookupNestedRel(String name, boolean searchParent)
   throws InterpretError
   {  IDInfo id;
	  Relation r = null; //Yi Zheng June 3, 2002
//	name = "."+name; //commented out by Yi Zheng June 3, 2002
//System.out.println("In env::lookupNestedRel for " + name);
	  //First search the NREnvironment
	  if (NRenv != null)
	  {
//System.out.println("before call NRenv.lookupNest..., NRenv.row: " + NRenv.row);
	     if ((r = NRenv.lookupNestedRel(name, NRenv.row))!= null ) //modified by Yi Zheng June 3, 2002
		 {
		//	System.out.println("found " + name + " in NR");
			return r;
		 }
	  }
	  //Next search in the table
	  if ( (id = (IDInfo) table.get(name)) != null )
	  {
//		System.out.println("in lookupNestedRel, SEARCH table");
	     if( id.kind == PARAMETER)
		 {

//		 System.out.println("found PARAMETER");
		  ParamInfo pi = (ParamInfo) id;
			if(pi.type != IDLIST && pi.type != COMP) //REVISIT: Could be COMP not just IDLIST
			{  return null;
			}
			//Need to extract the nested relation
			Environment e = pi.env;
//			long surrNo  = ((long[]) e.data[pi.index])[e.row];
			Relation dom = lookupRel("." + name, true); //modified by Yi Zheng June 3, 2002
//			dom.extract(surrNo, pi.rel);
			return pi.rel;
		 }
		 if((id.kind == STATE) || ((id.kind == DOM) && (((DomInfo)id).dom.isStateDom == 1))) //Yi Zheng June 3, 2002
		 {
//		 	System.out.println("found STATE");
		 	 // return null;   //REVISIT:Need extractRel() method
			//addition by Yi Zheng June 3, 2002
			StateInfo si = (StateInfo) id;
			if(si.type != IDLIST)
				return null;
			Environment e = si.env;
			Relation result = e.lookupRel("." + name, true);
			return result;
		 }
		 if(id.kind == LOCAL)
		 {
		   LocalInfo li = (LocalInfo) id;
		   if(li.type != IDLIST)
		   	return null;
		   Relation result = lookupRel("." + name, true);
		   return result;

		 //end of addition by Yi Zheng June 3, 2002

		 //"name" must be a local. Locals are
		 //supposed to be of primitive type. We return null
		 }
		 return null;
	  }
	  //If not found, check reltable
	  if ( (r = reltable.getRel("." + name)) != null)
	  {
//System.out.println("in lookupNestedRel, SEARCH domtable");
	     return r;
	  }
	  if(searchParent && parent!=null)  //search parent env
	  {
//System.out.println("in lookupNestedRel, SEARCH parent");
		return parent.lookupNestedRel(name, true);
	  }
	  //name is not a relation in the env (it could have been a domain
	  //but that is not what was being searched for)
	  return null;
   }
   public Relation lookupNestedRel(String name, boolean searchParent, int rrow)
   throws InterpretError
   {  IDInfo id;
	  Relation r;
	  //First search the NREnvironment
	  if (NRenv != null)
	  {
//System.out.println("before call NRenv.lookupNest..., rrow: " + rrow);
	     if ( (r = NRenv.lookupNestedRel(name, rrow))!= null )
		 {
//System.out.println("after call NRenv.lookupNest...,rrow: " + rrow);
			return r;
		 }
	  }
	  return null;
   }
   /**
	* This routine to get the relation with the surrogate
	* (or computation or view).
	*  -- Weizhong Sun
	*/
   public Relation lookupRelSurr(String name, boolean searchParent)
   throws InterpretError
   {  IDInfo id;
	  Relation r;
	  //First search the NREnvironment
	  if (NRenv != null)
	  {
	     if ( (r = NRenv.lookupRelSurr(name))!= null )
		 {  return r;
		 }
	  }
	  return null;
   }
   // Weizhong Sun
   public long getSurrogate(String name)
   throws InterpretError
   {
	  long i;
	  //search the NREnvironment
	  if (NRenv != null)
	  {
	     i = NRenv.getSurrogate(name);
	     return i;
	  }
	  return -1;
   }

   //Yi Zheng June 3, 2002
   public void setSurrogate(String name, long sur)
   throws InterpretError
   {
	  //search the NREnvironment
	  if (NRenv != null)
	  {
	     NRenv.setSurrogate(name, sur);
	  }
   }
    /**
        * This routine searches the whole environment for a text.
        */
        public Text lookupText(String name, boolean searchParent) // Jiantao 2004
        {       Text t;
                if (name == null) return null;

                //Search textTable
                if ( (t = textTable.getText(name)) != null)
                        return t;

                if(searchParent && parent!=null)  //search parent env
                        return parent.lookupText(name, true);

                //name is not in the env (or it is text which is an error)
                return null;
        }

        // Jiantao 2004
        public void removeText(String name, boolean searchParent)
        {       if(textTable.getText(name) == null)
                        if (searchParent && parent != null)
                                parent.removeText(name, true);
                else
                        textTable.rmText(name);
        }


        // Jiantao 2004
        public void put(Text t)
                throws InterpretError
        {       textTable.addText(t);
        }




   //REVISIT: is this method necessary?
   public void purge()
   {  //Purge reltable
/*	  for(Enumeration e = reltable.keys(); e.hasMoreElements(); )
	  {  String name = (String) e.nextElement();
		 reltable.rmRel(name);
	  }
	  //Purge domtable
	  for(Enumeration e = reltable.keys(); e.hasMoreElements(); )
	  {  String name = (String) e.nextElement();
		 reltable.rmRel(name);
	  }*/
	  //the above are commented out by Yi Zheng April 23, 2002
	  reltable = null;
	  domtable = null;
	  table = null;
   }
   public void put(Domain d)
   {  //Must check to see if the domain already exists
	  //in this frame of the environment
	  Domain dom = lookupDom(d.name, false);
	  if (dom != null)
	  {  if (domtable.rmDom(d.name))
		 {  domtable.addDom(d);
		 }
		 //In this case...
	  }
	  else
	  {  domtable.addDom(d);
	  }
   }

	//zhan starting
	//to handle the recursive nesting case: modify the type of recursive
	//domain during relational initialization
	public void putRecurDom(Domain d)
	{
	  //Must check to see if the domain already exists
	  Domain dom = lookupDom(d.name, false);
	  if (dom != null)
	  {
		 if(domtable.rmRecurDom(d.name))
		   domtable.addDom(d);
	  }
	  else
		  domtable.addDom(d);
	}
	 //zhan ending





   /****************************************************************/
   /*                       INSERTION ROUTINES                     */
   /****************************************************************/

   public void put(String name, IDInfo id)
   {  if(id.kind == REL)
	  {  reltable.addRel(((RelInfo) id).rel, true);
	  }
	  else if (id.kind == DOM) //REVISIT:
	  { // domtable.addDom(((DomInfo) id).dom);
	  //added by Yi Zheng March 24, 2002
	  Global.println("Environment: put (DOM) not implemented yet");
	  }
	  else
	  {  if (id.kind == PARAMETER)
		 {  ((ParamInfo) id).env = this;
		 }
		 else if (id.kind == STATE)
		 {  ((StateInfo) id).env = this;
		 }
		 table.put(name, id);
	  }
   }


   //method used to export inner computations to parent environment(s)
   public void put(String name, IDInfo id, boolean searchParent)
   {  if(id.kind == REL)
	  {  reltable.addRel(((RelInfo) id).rel, true);
	  }
	  else if (id.kind == DOM) //REVISIT:
	  { // domtable.addDom(((DomInfo) id).dom);
	  //added by Yi Zheng March 24, 2002
	  Global.println("Environment: put (DOM) not implemented yet");
	  }
	  else
	  {  if (id.kind == PARAMETER)
		 {  ((ParamInfo) id).env = this;
		 }
		 else if (id.kind == STATE)
		 {  ((StateInfo) id).env = this;
		 }
		 table.put(name, id);
	  }
	  if((searchParent) && (parent != null))
	  	parent.put(name, id, searchParent);
   }


   public void put(Relation r)
   throws InterpretError
   {
 //    Relation rel = lookupRel(r.name, true);//we don't need this Yi Zheng March 24, 2002
     reltable.addRel(r, true);
   }

   /**
    * put the relation in the designated environment
    * @author Yi Zheng June 14, 2002
    */
   public void put2(Relation r, Environment envFound)
   throws InterpretError
    {
    	if(envFound == null)
    		return;
    	if(this == envFound)
    		reltable.addRel(r, true);
    	else if (parent != null)
    		parent.put2(r, envFound);
    }
   public void removeNRFrame()
   {  NRenv = NRenv.getParent();
   }
   /****************************************************************/
   /*                        REMOVAL ROUTINES                      */
   /****************************************************************/

   public void removeRel(String name, boolean searchParent)
   {  if(reltable.getRel(name) == null)
	  {
	     if ((searchParent) && (parent != null)) //modified by Yi Zheng June 14, 2002
		 {  parent.removeRel(name, true);
		 }
	  }
	  else
	  {  //Relation name is in this frame
		 reltable.rmRel(name);
	  }
   }

   //Yi Zheng June 28, 2002
   public void removeDom(String name, boolean searchParent)
   {
   	 if(domtable.getDom(name) == null)
	  {
	     if ((searchParent) && (parent != null)) //modified by Yi Zheng June 14, 2002
		 {  parent.removeDom(name, true);
		 }
	  }
	  else
	  {  //Relation name is in this frame
		 domtable.rmDom(name); //this will also remove dot relations if dom type is IDLIST or COMP
	  }

   	}

   /**
    * remove a named paramInfo
    * @author Yi Zheng June 20, 2002
    */
    public void removeParam(String name)
    {
    	ParamInfo pi = (ParamInfo) table.get(name);
    	if(pi != null)
    		table.remove(name);
	}
   /**
    * remove a relation with name name and record in which environment this relation is found
    * @author Yi Zheng June 14, 2002
    */
   public Environment removeRel2(String name, boolean searchParent)
   {
   	 if(reltable.getRel(name) == null)
	  {
	     if ((searchParent) && (parent != null))
		 {  return parent.removeRel2(name, searchParent);
		 }
		 return null;
	  }
	  else
	  {  //Relation name is in this frame
		 reltable.rmRel(name);
		 return this;
	  }
  }


// convert opcode to string to represent the prefix/action name in event handle
// Weizhong
   public String opcodeToString(int opcode)
   throws InterpretError
  {
//System.out.println("in opcodeToString of Environment.java, opcode: " + opcode);
	switch (opcode) {
	case OP_PRE:
		return new String("pre");
	case OP_POST:
		return new String("post");
	case OP_ADD:
		return new String("add");
	case OP_DELETE:
		return new String("delete");
	case OP_CHANGE:
		return new String("change");
	case OP_CONTAINS:
	        return new String("contains");
	case OP_CMPCONTAINS:
	        return new String("cmpcontains");
	case OP_CMPWITHIN:
	        return new String("cmpwithin");
	case OP_INTERSECT:
	        return new String("intersect");
	case OP_WITHIN:
		return new String("within");
	case OP_WITHINDIST:
		return new String("withindist");
	default:
		throw new InterpretError("wrong opcode to String.");
	}
   }


// convert string that represents the prefix/action name in event handle to opcode
// YuLing
   public int stringToOpcode(String str)
   throws InterpretError
  {
        if(str.compareTo("contains")==0)
	   return OP_CONTAINS;
        else if(str.compareTo("within")==0)
	   return OP_WITHIN;
	else if(str.compareTo("cmpcontains")==0)
	   return OP_CMPCONTAINS;
	else if(str.compareTo("cmpwithin")==0)
	   return OP_CMPWITHIN;
	else if(str.compareTo("intersect")==0)
	   return OP_INTERSECT;
	else if(str.compareTo("withindist")==0)
	   return OP_WITHINDIST;

		throw new InterpretError("wrong string to opcode.");
   }
/*
 * to get an actual event name from a tree.
 */
  public String getEventName(SimpleNode node)
  throws InterpretError
  {
	int numchild = node.jjtGetNumChildren();
// for debug
//System.out.println("in getEventName, numchild: " + numchild);
//System.out.println("in getEventName, node.name; node.type; node.opcode: " + node.name + ";" + node.type + ";" + node.opcode);
/*SimpleNode child;
for ( int i = 0; i < numchild; i++ ) {
  child = (SimpleNode) node.jjtGetChild(i);
  System.out.println("child.name; child.type; child.opcode: "  + child.name + ";" + child.type + ";" + child.opcode);
}
*/
// --- end of debug

	SimpleNode prefixnode;
	SimpleNode actionnode;
	SimpleNode relnode;
	String name = new String("");
//	if (DEBUG)
//		System.out.println("numchild ="+numchild);
	if (numchild == 2) {
		actionnode = (SimpleNode) node.jjtGetChild(0);
		relnode = (SimpleNode) node.jjtGetChild(1);
		name = opcodeToString(OP_POST) + ":" + opcodeToString(actionnode.opcode) + ":" + relnode.name;
	} else if(numchild==1){ //added by YuLing for GIS eventnames
				//eg: comp containts:() is...
                actionnode=(SimpleNode) node.jjtGetChild(0);
	//	System.out.println("opcode="+actionnode.opcode);
		name = opcodeToString(actionnode.opcode)+":";
	//	System.out.println("name="+name);
	}
	     else {
		prefixnode = (SimpleNode) node.jjtGetChild(0);
		actionnode = (SimpleNode) node.jjtGetChild(1);
		relnode = (SimpleNode) node.jjtGetChild(2);
		name = opcodeToString(prefixnode.opcode) + ":" + opcodeToString(actionnode.opcode) + ":" + relnode.name;
	}

//System.out.println("in getEventName, name: " + name);
	return name;
  }
	//Yi Zheng May 20, 2002
	public Hashtable getTable()
	{ return table;}

} /* class Environment */
