/**
 * This class contains evaluators for expressions that should return 
 * one of the follwoing types:
 *     boolean
 *     short
 *     integer
 *     long
 *     float
 *     double
 *     string
 */
import java.util.*;
import java.io.*;

class EvalExpr implements Constants, TypeConstants
{  
//debugging Yi Zheng
private static boolean DEBUG = false;

//added to provide wrapper for evalIntExpr by Yi Zheng
public static boolean WEIZHONG = false;

//In order to use evaluateTLExpression() in Interpreter class
   public static Interpreter interpreter;

   //Empty hashtable needed as hack since evaluateTLExpression()
   //REVISIT: method evaluateTLExpression() should check if ht is null.
   static Hashtable emptyht = new Hashtable();


   public static boolean evalBooleanExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  { 
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalBooleanExpr(thenNode, env);
			} else {
				return evalBooleanExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case BOOLEAN :
					 return ((byte[]) r.data[0])[0] == 1;
				default://added by Yi Zheng April 13, 2002
					throw new InterpretError("Expression does not return a boolean");
			   }
			}
			throw new InterpretError("Expression does not return a boolean");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
			if (node.jjtGetNumChildren() != 2)
			{   throw new 
				InterpretError("Bad binary operator");
			}
			SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
			SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
//Object lo = evalObjectExpr(leftNode, env);
//Object ro = evalObjectExpr(rightNode, env);
//commented out and added to each of the following relevant cases by
//Yi Zheng, March 31, 2002
			switch (node.opcode)
			{  case OP_AND :
				  return evalBooleanExpr(leftNode, env) &&
						 evalBooleanExpr(rightNode, env);
			   case OP_OR :
				  return evalBooleanExpr(leftNode, env) ||
						 evalBooleanExpr(rightNode, env);
			   case OP_EQ :{ //REVISIT: equality of relations??
//				  return evalLongExpr(leftNode, env) ==  evalLongExpr(rightNode, env);
Object lo = evalObjectExpr(leftNode, env); //added by Yi Zheng March 31, 2002
Object ro = evalObjectExpr(rightNode, env); //added by Yi Zheng March 31, 2002

				if (lo.getClass().getName() == ro.getClass().getName()) {
					int ri = lo.toString().compareTo(ro.toString());
					if (ri == 0) return true;
					else return false;
				} else return false;
			}
			   case OP_GE :{
//				  return evalLongExpr(leftNode, env) >=  evalLongExpr(rightNode, env);
Object lo = evalObjectExpr(leftNode, env);//added by Yi Zheng March 31, 2002
Object ro = evalObjectExpr(rightNode, env);//added by Yi Zheng March 31, 2002

				if (lo.getClass().getName() == ro.getClass().getName()) {
					int ri = lo.toString().compareTo(ro.toString());
					if (ri >= 0) return true;
					else return false;
				} else return false;
			}	
			   case OP_GT :{
//				  return evalLongExpr(leftNode, env) >   evalLongExpr(rightNode, env);
Object lo = evalObjectExpr(leftNode, env);//added by Yi Zheng March 31, 2002
Object ro = evalObjectExpr(rightNode, env);//added by Yi Zheng March 31, 2002

				if (lo.getClass().getName() == ro.getClass().getName()) {
					int ri = lo.toString().compareTo(ro.toString());
					if (ri > 0) return true;
					else return false;
				} else return false;
			}
			   case OP_LE :{
//				  return evalLongExpr(leftNode, env) <=  evalLongExpr(rightNode, env);
Object lo = evalObjectExpr(leftNode, env);//added by Yi Zheng March 31, 2002
Object ro = evalObjectExpr(rightNode, env);//added by Yi Zheng March 31, 2002

				if (lo.getClass().getName() == ro.getClass().getName()) {
					int ri = lo.toString().compareTo(ro.toString());
					if (ri <= 0) return true;
					else return false;
				} else return false;
			}
			   case OP_LT :{
//				  return evalLongExpr(leftNode, env) <   evalLongExpr(rightNode, env);
Object lo = evalObjectExpr(leftNode, env);//added by Yi Zheng March 31, 2002
Object ro = evalObjectExpr(rightNode, env);//added by Yi Zheng March 31, 2002

				if (lo.getClass().getName() == ro.getClass().getName()) {
					int ri = lo.toString().compareTo(ro.toString());
					if (ri < 0) return true;
					else return false;
				} else return false;
			}
			   case OP_NEQ :{
//				  return evalLongExpr(leftNode, env) !=  evalLongExpr(rightNode, env);
Object lo = evalObjectExpr(leftNode, env);//added by Yi Zheng March 31, 2002
Object ro = evalObjectExpr(rightNode, env);//added by Yi Zheng March 31, 2002

				if (lo.getClass().getName() == ro.getClass().getName()) {
					int ri = lo.toString().compareTo(ro.toString());
					if (ri != 0) return true;
					else return false;
				} else return false;
			}
			   case OP_SUBSTR :
				  String leftResult = evalStringExpr(leftNode, env);
				  String rightResult = evalStringExpr(rightNode, env);
				  return rightResult.indexOf(leftResult) != -1;
			   case OP_CAST :
				  throw new 
				  InterpretError("Cast operator not implemented");
			   default :
				  throw new         
				  InterpretError("Invalid binary operation");
			}
		 case OP_UOPERATOR : //REVISIT: assuming a unary tree
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			boolean childResult = evalBooleanExpr(childNode, env);
			switch (node.opcode)
			{  case OP_UNOT :
				  return !childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL :
			switch (node.opcode) 
			{  case OP_BOOLEAN :
				  Byte b = (Byte) node.getObject();
				  if (b == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return b.intValue() == 1;
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case BOOLEAN :
					 return ((byte[]) nri.NRenv.data[col])[row] == 1;
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case BOOLEAN :
					 return ((byte[]) pi.env.data[col])[row] == 1;
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if(id.kind == LOCAL)
			{  switch (((LocalInfo) id).type)
			   {  case BOOLEAN :
					 return (((LocalInfo) id).intVal == 1) ? true : false;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{ //added by Yi Zheng April 18, 2002
				if(!WEIZHONG)
				{
					switch (((StateInfo) id).type)
					{
						case BOOLEAN:
							return (((StateInfo) id).intVal == 1)?true:false;
						default:
							throw new InterpretError("Incompatible identifier");
					}
				}else
				{
					StateInfo si = (StateInfo) id;
					col = si.index;
					int row = si.env.row;
					switch(si.type)
					{
						case BOOLEAN:
							return ((byte[]) si.env.data[col])[row] == 1;
				  		default :
					 		throw new 
					 			InterpretError("Incompatible identifier");
			   		}	
				}
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
if (firstchild.opcode == OP_IDENTIFIER) {
	SimpleNode virnode = firstchild.jjtDeepDuplicate();
	virnode.name = iddom.name;
	boolean virret = evalBooleanExpr(virnode, env);
	return virret;
} else {
 				boolean ret =  evalBooleanExpr(iddom.getTree(), env);
				return ret;
}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
			InterpretError("This is not implemented yet.");
		 default :
			throw new
			InterpretError("Invalid boolean statement");
	  }
   }   
   public static double evalDoubleExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  { 
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalDoubleExpr(thenNode, env);
			} else {
				return evalDoubleExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case DOUBLE :
					 return ((double[]) r.data[0])[0];
				  case FLOAT :
					 return ((float[]) r.data[0])[0];
				  case LONG :
					 return ((long[]) r.data[0])[0];
				  case INTEGER :
					 return ((int[]) r.data[0])[0];
				  case SHORT :
					 return ((short[]) r.data[0])[0];
			   }
			}
			throw new InterpretError("Expression does not return a double");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
		 if (node.jjtGetNumChildren() != 2)
		 {   throw new 
			 InterpretError("Bad binary operator");
		 }
		 SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
		 double leftResult = evalDoubleExpr(leftNode, env);
		 double rightResult = evalDoubleExpr(rightNode, env);
		 switch (node.opcode)
		 {  case OP_DIV :
			   return leftResult / rightResult;
			case OP_MIN :
			   return Math.min(leftResult, rightResult);
			case OP_MINUS :
			   return leftResult - rightResult;
			case OP_MAX :
			   return Math.max(leftResult, rightResult);
			case OP_MUL :
			   return leftResult * rightResult;
			case OP_PLUS :
			   return leftResult + rightResult;
			case OP_POW :
			   return Math.pow(leftResult, rightResult);
			case OP_MOD :
			   throw new 
			   InterpretError("Modulo operator used incorrectly");
			case OP_EQ :
			case OP_GE :
			case OP_GT :
			case OP_LE :
			case OP_LT :
			case OP_NEQ :
			   throw new 
			   InterpretError("Comparison operator used incorrectly");
			case OP_CAST :
			   throw new 
			   InterpretError("Cast operator not implemented");
			default :
			   throw new 
			   InterpretError("Uknown binary operator");
		 }
		 case OP_UOPERATOR :
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			double childResult = evalDoubleExpr(childNode, env);
			switch (node.opcode) //REVISIT: assuming a binary tree
			{  case OP_UMINUS :
				  return -childResult;
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL : //REVISIT:
			switch (node.opcode)
			{  case OP_DOUBLE :
				  Double d = (Double) node.getObject();
				  if (d == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return d.doubleValue();
			   case OP_FLOAT :
				  Float f = (Float) node.getObject();
				  if (f == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return f.floatValue();
			   case OP_LONG :
				  Long l = (Long) node.getObject();
				  if (l == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return l.longValue();
			   case OP_INTEGER :
				  Integer i = (Integer) node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return i.intValue();
			   case OP_SHORT :
				  Short s = (Short) node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return s.shortValue();
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case DOUBLE :
					 return ((double[]) nri.NRenv.data[col])[row];
				  case FLOAT :
					 return ((float[]) nri.NRenv.data[col])[row];
				  case LONG :
					 return ((long[]) nri.NRenv.data[col])[row];
				  case INTEGER :
					 return ((int[]) nri.NRenv.data[col])[row];
				  case SHORT :
					 return ((short[]) nri.NRenv.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case DOUBLE :
					 return ((double[]) pi.env.data[col])[row];
				  case FLOAT :
					 return ((float[]) pi.env.data[col])[row];
				  case LONG :
					 return ((long[]) pi.env.data[col])[row];
				  case INTEGER :
					 return ((int[]) pi.env.data[col])[row];
				  case SHORT :
					 return ((short[]) pi.env.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if(id.kind == LOCAL)
			{  switch (((LocalInfo) id).type)
			   {  case DOUBLE :
					 return ((LocalInfo) id).doubleVal;
				  case FLOAT :
					 return ((LocalInfo) id).floatVal;
				  case LONG :
					 return ((LocalInfo) id).longVal;
				  case INTEGER :
					 return ((LocalInfo) id).intVal;
				  case SHORT :
					 return ((LocalInfo) id).intVal;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{ //added by Yi Zheng April 18, 2002
				if(!WEIZHONG)
				{
					switch (((StateInfo) id).type)
				   {  case DOUBLE :
						 return ((StateInfo) id).doubleVal;
					  case FLOAT :
						 return ((StateInfo) id).floatVal;
					  case LONG :
						 return ((StateInfo) id).longVal;
					  case INTEGER :
						 return ((StateInfo) id).intVal;
					  case SHORT :
						 return ((StateInfo) id).intVal;
					  default :
						 throw new 
						 InterpretError("Incompatible identifier");
				   }
				 }else
				 {
				 	StateInfo si = (StateInfo) id;
			   		col = si.index;
			   		int row = si.env.row;
				   switch ( si.type )
				   {  case DOUBLE :
						 return ((double[]) si.env.data[col])[row];
					  case FLOAT :
						 return ((float[]) si.env.data[col])[row];
					  case LONG :
						 return ((long[]) si.env.data[col])[row];
					  case INTEGER :
						 return ((int[]) si.env.data[col])[row];
					  case SHORT :
						 return ((short[]) si.env.data[col])[row];
					  default :
						 throw new 
						 InterpretError("Incompatible identifier");
				   }	
				 	
				 	
				 }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
if (firstchild.opcode == OP_IDENTIFIER) {
	SimpleNode virnode = firstchild.jjtDeepDuplicate();
	virnode.name = iddom.name;
	double virret = evalDoubleExpr(virnode, env);
	return virret;
} else {
 				double ret =  evalDoubleExpr(iddom.getTree(), env);
				return ret;
}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
			InterpretError("This is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}
			
			double ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.sin(arg);
					return ret;
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.cos(arg);
					return ret;	
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (double) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					double denom = (double) Math.log(10);
					ret = (double) Math.log(arg) / denom;
					return ret;
				}
				case OP_ROUND://assume argument is of type double
				{//round() returns integer, not double
					throw new InterpretError("Bad use of round()");
				}
				case OP_CEIL://assume argument is of type double
				{
                                        double arg = evalDoubleExpr(argument, env);
                                        ret = (double) Math.ceil(arg);
                                        return ret;
// Dec 9 2004 Zongyan

					//throw new InterpretError("Bad use of ceil()");
				}
				case OP_FLOOR://assume argument is of type double
				{
                                        double arg = evalDoubleExpr(argument, env);
                                        ret = (double) Math.floor(arg);
                                        return ret;
// Dec 9 2004 Zongyan
					//throw new InterpretError("Bad use of floor()");
				}
				case OP_CHR:
					throw new InterpretError("Bad use of chr()");					
				case OP_ORD:
				{
					String s = evalStringExpr(argument, env);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (double) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}	
		//	throw new InterpretError("Functions not implemented yet");

		 default :
			throw new 
			InterpretError("Invalid double statement");
	  }
   }   
   public static float evalFloatExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  { 
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalFloatExpr(thenNode, env);
			} else {
				return evalFloatExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case FLOAT :
					 return ((float[]) r.data[0])[0];
				  case LONG :
					 return ((long[]) r.data[0])[0];
				  case INTEGER :
					 return ((int[]) r.data[0])[0];
				  case SHORT :
					 return ((short[]) r.data[0])[0];
			   }
			}
			throw new InterpretError("Expression does not return a float");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
		 if (node.jjtGetNumChildren() != 2)
		 {   throw new 
			 InterpretError("Bad binary operator");
		 }
		 SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
		 float leftResult = evalFloatExpr(leftNode, env);
		 float rightResult = evalFloatExpr(rightNode, env);
		 switch (node.opcode)
		 {  case OP_DIV :
			   return leftResult / rightResult;
			case OP_MIN :
			   return Math.min(leftResult, rightResult);
			case OP_MINUS :
			   return leftResult - rightResult;
			case OP_MAX :
			   return Math.max(leftResult, rightResult);
			case OP_MUL :
			   return leftResult * rightResult;
			case OP_PLUS :
			   return leftResult + rightResult;
			case OP_POW :
			   return (float) Math.pow(leftResult, rightResult);
			case OP_MOD :
			   throw new 
			   InterpretError("Modulo operator used incorrectly");
			case OP_EQ :
			case OP_GE :
			case OP_GT :
			case OP_LE :
			case OP_LT :
			case OP_NEQ :
			   throw new 
			   InterpretError("Comparison operator used incorrectly");
			case OP_CAST :
			   throw new 
			   InterpretError("Cast operator not implemented");
			default :
			   throw new 
			   InterpretError("Uknown binary operator");
		 }
		 case OP_UOPERATOR :
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			float childResult = evalFloatExpr(childNode, env);
			switch (node.opcode) //REVISIT: assuming a binary tree
			{  case OP_UMINUS :
				  return -childResult;
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL : //REVISIT:
			switch (node.opcode)
			{  case OP_FLOAT :
				  Float f = (Float) node.getObject();
				  if (f == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return f.floatValue();
			   case OP_LONG :
				  Long l = (Long) node.getObject();
				  if (l == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return l.longValue(); //why don't we return floatValue()? Yi Zheng April 10, 2002
			   case OP_INTEGER :
				  Integer i = (Integer) node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return i.intValue();
			   case OP_SHORT :
				  Short s = (Short) node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return s.shortValue();
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case FLOAT :
					 return ((float[]) nri.NRenv.data[col])[row];
				  case LONG :
					 return ((long[]) nri.NRenv.data[col])[row];
				  case INTEGER :
					 return ((int[]) nri.NRenv.data[col])[row];
				  case SHORT :
					 return ((short[]) nri.NRenv.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case FLOAT :
					 return ((float[]) pi.env.data[col])[row];
				  case LONG :
					 return ((long[]) pi.env.data[col])[row];
				  case INTEGER :
					 return ((int[]) pi.env.data[col])[row];
				  case SHORT :
					 return ((short[]) pi.env.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal"+((ParamInfo) id).type);
			   }
			}
			else if(id.kind == LOCAL)
			{  switch (((LocalInfo) id).type)
			   {  case FLOAT :
					 return ((LocalInfo) id).floatVal;
				  case LONG :
					 return ((LocalInfo) id).longVal;
				  case INTEGER :
					 return ((LocalInfo) id).intVal;
				  case SHORT :
					 return ((LocalInfo) id).intVal;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{//Yi Zheng April 18, 2002
				if(!WEIZHONG)
				{
					switch (((StateInfo) id).type)
				   {  case FLOAT :
						 return ((StateInfo) id).floatVal;
					  case LONG :
						 return ((StateInfo) id).longVal;
					  case INTEGER :
						 return ((StateInfo) id).intVal;
					  case SHORT :
						 return ((StateInfo) id).intVal;
					  default :
						 throw new 
						 InterpretError("Incompatible identifier");
				   }
				 }else
				 {
				 	StateInfo si = (StateInfo) id;
				 	col = si.index;
				 	int row = si.env.row;	
				 	switch ( si.type )
			   		{  case FLOAT :
					 		return ((float[]) si.env.data[col])[row];
				  		case LONG :
					 		return ((long[]) si.env.data[col])[row];
				  		case INTEGER :
					 		return ((int[]) si.env.data[col])[row];
				  		case SHORT :
					 		return ((short[]) si.env.data[col])[row];
				  		default :
					 		throw new 
					 		InterpretError("Incompatible identifier");
			   		}

				 	
				 }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
if (firstchild.opcode == OP_IDENTIFIER) {
	SimpleNode virnode = firstchild.jjtDeepDuplicate();
	virnode.name = iddom.name;
	float virret = evalFloatExpr(virnode, env);
	return virret;
} else {
 				float ret =  evalFloatExpr(iddom.getTree(), env);
				return ret;
}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
			InterpretError("This is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}
			
			float ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type float
				{
					float arg = evalFloatExpr(argument, env);
					ret = (float) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.sin(arg);
					return ret;
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.cos(arg);
					return ret;	
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (float) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					float denom = (float) Math.log(10);
					ret = (float) Math.log(arg) / denom;
					return ret;
				}
				case OP_ROUND://assume argument is of type float
				{
					throw new InterpretError("Bad use of round()");
				}
				case OP_CEIL://assume argument is of type double
				{
                                        double arg = evalDoubleExpr(argument, env);
                                        ret = (float) Math.ceil(arg);
                                        return ret;
// Dec 9 2004 Zongyan

					//throw new InterpretError("Bad use of ceil()");
				}
				case OP_FLOOR://assume argument is of type double
				{
                                        double arg = evalDoubleExpr(argument, env);
                                        ret = (float) Math.floor(arg);
                                        return ret;
// Dec 9 2004 Zongyan
					//throw new InterpretError("Bad use of floor()");
				}
				case OP_CHR:
					throw new InterpretError("Bad use of chr()");					
				case OP_ORD:
				{
					String s = evalStringExpr(argument, env);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (float) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}	
		//	throw new InterpretError("Functions not implemented yet");
		 default :
			throw new 
			InterpretError("Invalid float statement");
	  }
   } 
   
   //method added by Yi Zheng April 18, 2002 to fix state bug
   //Weizhong's modification to the case of STATE causes 
   //ArrayIndexOutOfBoundsException in the following case:
   //
   //domain cur intg;
   //comp cnt(cur) is
   //state cur_ intg;
   //{cur_ <- cur;}
   //alt
   //{cur_ <- cur_ + 1; cur <- cur_;};
   //
   //A <- cnt[1];
   //A <- cnt[]; << exception occurs here
   //
   //the fix is to enhance StateInfo object with value holders
   //and check for these values when evalIntExpr
   //to leave the original code intact, the wrapper function is provided
   // -- Yi Zheng April 18, 2002
   public static int evalIntExpr(SimpleNode node, Environment env)
   throws InterpretError
   {
   		if(node.type == OP_IDENTIFIER)
   		{ 
			IDInfo id = env.lookup(node.name, true);
			if(id != null && id.kind == STATE && (!WEIZHONG))
			{
		//	System.out.println("in evalIntExpr for " + node.name);
				switch(((StateInfo)id).type)
				{
					case SHORT:
						return ((StateInfo)id).intVal;
					case INTEGER:
						return ((StateInfo)id).intVal;
					default:
						throw new InterpretError("Incompatible type");
				}
			}
			else return evalIntExpr_old(node,env);
		}
		else return evalIntExpr_old(node,env);
	}
	

   	public static int evalIntExpr_old(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
//System.out.println("node.type:"+node.type+node.opcode);
	  switch(node.type)
	  { 
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalIntExpr(thenNode, env);
			} else {
				return evalIntExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_ARRAY :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			String tempFileName = interpreter.nextTempName();
			Relation r = interpreter.evaluateTLExpression(
									 node,tempFileName,emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   interpreter.deleteFile(tempFileName);
			   switch (r.domains[0].type)
			   {  case SHORT :
					 return ((short[]) r.data[0])[0];
				  case INTEGER :
					 return ((int[]) r.data[0])[0];
				  default:
				  throw new InterpretError("Data type not implemented yet");
			   }
			}
			throw new InterpretError("Expression does not return an integer");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree

			
			if (node.jjtGetNumChildren() != 2)
			{   throw new 
				InterpretError("Bad binary operator");
			}
			SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
			SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
			int leftResult = evalIntExpr(leftNode, env);


			int rightResult = evalIntExpr(rightNode, env);	


			switch (node.opcode)
			{  case OP_DIV :
				  return leftResult / rightResult;
			   case OP_MIN :
				  return Math.min(leftResult, rightResult);
			   case OP_MINUS :
				  return leftResult - rightResult;
			   case OP_MAX :
				  return Math.max(leftResult, rightResult);
			   case OP_MOD :
				  return leftResult % rightResult;
			   case OP_MUL :
				  return leftResult * rightResult;
			   case OP_PLUS :
				  return leftResult + rightResult;
			   case OP_POW :
				  return (int) Math.pow(leftResult, rightResult);
			   case OP_EQ :
			   case OP_GE :
			   case OP_GT :
			   case OP_LE :
			   case OP_LT :
			   case OP_NEQ :
				  throw new 
				  InterpretError("Comparison operator used incorrectly");
			   case OP_CAST :
				  throw new 
				  InterpretError("Cast operator not implemented");
			   default :
				  throw new 
				  InterpretError("Invalid binary operator");
			}
		 case OP_UOPERATOR : //REVISIT: assuming a unary tree
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			int childResult = evalIntExpr(childNode, env);
			switch (node.opcode)
			{  case OP_UMINUS :
				  return -childResult;
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL :
			switch (node.opcode) 
			{  case OP_INTEGER :
				  Integer i = (Integer) node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  
				  return i.intValue();
			   case OP_SHORT :
				  Short s = (Short) node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return s.shortValue();
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
//System.out.println("	OP_IDENTIFIER:"+id.kind);	
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{ 
			   NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case INTEGER :
					 return ((int[]) nri.NRenv.data[col])[row];
				  case SHORT :
					 return ((short[]) nri.NRenv.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal" + Utility.type2String(nri.type));
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case INTEGER :
					 return ((int[]) pi.env.data[col])[row];
				  case SHORT :
					 return ((short[]) pi.env.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if(id.kind == LOCAL)//weizhong
			{  
			   switch (((LocalInfo) id).type)
			   {  case INTEGER :
					 return ((LocalInfo) id).intVal;
				  case SHORT :
					 return ((LocalInfo) id).intVal;
				  case IDLIST:
				  	 String nr_name = ((LocalInfo) id).nrVal;		

				 	 IDInfo NRid = env.lookup(nr_name, true);
					 if (NRid == null)
					 {
					   throw new InterpretError("Unknown variable " + nr_name);
					 }  

					 if (NRid.kind == NR)
					 {
						NRInfo nri = (NRInfo) NRid;
						col = nri.index;
						int row = nri.NRenv.row;
						switch (nri.type)
						{ case INTEGER :
;					
							   return ((int[]) nri.NRenv.data[col])[row];
						  case SHORT :
						  	   return ((short[]) nri.NRenv.data[col])[row];

						  default :
						           throw new
							   InterpretError("Invalid literal");
						}
					  } 
					  else 
					  {
					        throw new
						InterpretError("Wrong type of input parameter.");
					  }		   
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{
			// weizhong: add for update..computation..
			   StateInfo si = (StateInfo ) id;
			   col = si.index;
			   int row = si.env.row;
			   switch ( si.type )
			   {  case INTEGER :
			           int i = ((int[]) si.env.data[col])[row];
			           return ((int[]) si.env.data[col])[row];
			      case SHORT :
			           return ((short[]) si.env.data[col])[row];
			      default :
	        	           throw new
			           InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
					SimpleNode idtree = (SimpleNode) iddom.tree;
					SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
					if (firstchild.opcode == OP_IDENTIFIER) {
						SimpleNode virnode = firstchild.jjtDeepDuplicate();
						virnode.name = iddom.name;
						int virret = evalIntExpr(virnode, env);
						return virret;
					} else {
					 				int ret =  evalIntExpr(iddom.getTree(), env);
									return ret;
					}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
				InterpretError("This type is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}
			
			int ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type int
				{
					int arg = evalIntExpr(argument, env);
					ret = (int) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					throw new InterpretError("Bad use of sin()");
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					throw new InterpretError("Bad use of cos()");
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					double denom = (double) Math.log(10);
					ret = (int) (Math.log(arg) / denom);
					return ret;
				}
				case OP_ROUND://assume argument is of type float
				{
					float arg = evalFloatExpr(argument, env);
					ret = (int) Math.round(arg);
					return ret;
				}
				case OP_CEIL://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.ceil(arg);
					return ret;
				}
				case OP_FLOOR://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (int) Math.floor(arg);
					return ret;
				}
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					int code = evalIntExpr(argument, env);
					if(code < 0 || code > 255)
						throw new InterpretError("Invalid argument to chr()");
					else
						return code;
				}					
				case OP_ORD:
				{
					String s = evalStringExpr(argument, env);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (int) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}	
		//	throw new InterpretError("Functions not implemented yet");

		 default :
			throw new
			InterpretError("Invalid integer statement");
	  }
   }   
   
   // Yiyi added
   public static number evalNumExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  {
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode;
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalNumExpr(thenNode, env);
			} else {
				return evalNumExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar.
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {
            	case NUMERIC:
               	return ((number[])r.data[0])[0];
            	case LONG :
					 return number.longToNum(((long[])r.data[0])[0]);
				  case INTEGER :
					 return number.intToNum(((int[])r.data[0])[0]);
				  case SHORT :
					 return number.shortToNum(((short[])r.data[0])[0]);
			   }
			}
			throw new InterpretError("Expression does not return a number");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
		 if (node.jjtGetNumChildren() != 2)
		 {   throw new
			 InterpretError("Bad binary operator");
		 }
		 SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
		 number leftResult = evalNumExpr(leftNode, env);
		 number rightResult = evalNumExpr(rightNode, env);
		 switch (node.opcode)
		 {  case OP_DIV :
			   return leftResult.divide(rightResult);
			case OP_MIN :
			   return number.min(leftResult, rightResult);
			case OP_MINUS :
			   return leftResult.subtract(rightResult);
			case OP_MAX :
			   return number.max(leftResult, rightResult);
			case OP_MUL :
			   return leftResult.multiply(rightResult);
			case OP_PLUS :
			   return leftResult.add(rightResult);
			case OP_POW :
			   return number.pow(leftResult, rightResult);
			case OP_MOD :
			   throw new
			   InterpretError("Modulo operator used incorrectly");
			case OP_EQ :
			case OP_GE :
			case OP_GT :
			case OP_LE :
			case OP_LT :
			case OP_NEQ :
			   throw new
			   InterpretError("Comparison operator used incorrectly");
			case OP_CAST :
			   throw new
			   InterpretError("Cast operator not implemented");
			default :
			   throw new
			   InterpretError("Uknown binary operator");
		 }
		 case OP_UOPERATOR :
			if (node.jjtGetNumChildren() != 1)
			{   throw new
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			number childResult = evalNumExpr(childNode, env);
			switch (node.opcode) //REVISIT: assuming a binary tree
			{  case OP_UMINUS :
				  return number.negate(childResult);
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL : //REVISIT:
			switch (node.opcode){
         	case OP_NUMERIC:
            	number n = (number)node.getObject();
               if(n == null){
               	throw new InterpretError("Literal has no value");
               }
               return n;
         	case OP_LONG :
				  Long l = (Long)node.getObject();
				  if (l == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return number.longToNum(l.longValue());
			   case OP_INTEGER :
				  Integer i = (Integer)node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return number.intToNum(i.intValue());
			   case OP_SHORT :
				  Short s = (Short)node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return number.shortToNum(s.shortValue());
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER :
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type ){
            	case NUMERIC:
               	return ((number[])nri.NRenv.data[col])[row];
            	case LONG :
					 return number.longToNum(((long[])nri.NRenv.data[col])[row]);
				  case INTEGER :
					 return number.intToNum(((int[]) nri.NRenv.data[col])[row]);
				  case SHORT :
					 return number.shortToNum(((short[])nri.NRenv.data[col])[row]);
				  default :
					 throw new
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {
					case NUMERIC:
               	return ((number[])pi.env.data[col])[row];
            	case LONG :
					 return number.longToNum(((long[])pi.env.data[col])[row]);
				  case INTEGER :
					 return number.intToNum(((int[])pi.env.data[col])[row]);
				  case SHORT :
					 return number.shortToNum(((short[])pi.env.data[col])[row]);
				  default :
					 throw new
					 InterpretError("Invalid literal"+((ParamInfo) id).type);
			   }
			}
			else if(id.kind == LOCAL)
			{  switch (((LocalInfo) id).type)
			   {
            	case NUMERIC:
						return ((LocalInfo) id).numberVal;
            	case LONG :
					 return ((LocalInfo) id).numberVal;
				  case INTEGER :
					 return ((LocalInfo) id).numberVal;
				  case SHORT :
					 return ((LocalInfo) id).numberVal;
				  default :
					 throw new
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{//added by Yi Zheng April 18, 2002
				if(!WEIZHONG)
				{
					switch (((StateInfo) id).type)
				   {
               	case NUMERIC:
                  	return ((StateInfo)id).numberVal;
               	case LONG :
						 return ((StateInfo) id).numberVal;
					  case INTEGER :
						 return ((StateInfo) id).numberVal;
					  case SHORT :
						 return ((StateInfo) id).numberVal;
					  default :
						 throw new
						 	InterpretError("Incompatible identifier");
				   }
			    }else
			    {
			    	StateInfo si = (StateInfo) id;
			   		col = si.index;
			   		int row = si.env.row;
			   		switch ( si.type )
			   		{
                  	case NUMERIC:
								return ((number[])si.env.data[col])[row];
                  	case LONG:
			   				return number.longToNum(((long[])si.env.data[col])[row]);
			   			case INTEGER :
			           		return number.intToNum(((int[])si.env.data[col])[row]);
			      		case SHORT :
			           		return number.shortToNum(((short[])si.env.data[col])[row]);
			      		default :
	        	           throw new
			           		InterpretError("Incompatible identifier");
			   		}

			    }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
				SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
				if (firstchild.opcode == OP_IDENTIFIER) {
					SimpleNode virnode = firstchild.jjtDeepDuplicate();
					virnode.name = iddom.name;
					number virret = evalNumExpr(virnode, env);
					return virret;
				} else {
 					number ret =  evalNumExpr(iddom.getTree(), env);
					return ret;
					}
			 } else
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new
			InterpretError("This is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);

			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}

			number ret;
         number arg = evalNumExpr(argument, env);
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type long
				{
					ret = number.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					ret = number.sqrt(arg);
					return ret;
				}
				case OP_SIN://assume argument is of type double
				{
					throw new InterpretError("Bad use of sin()");
				}
				case OP_ASIN://assume argument is of type double
				{
					ret = number.arcsin(arg);
					return ret;
				}
				case OP_COS://assume argument is of type double
				{
					throw new InterpretError("Bad use of cos()");
				}
				case OP_ACOS://assume argument is of type double
				{
					ret = number.arccos(arg);
					return ret;
				}
				case OP_TAN://assume argument is of type double
				{
					ret = number.tan(arg);
					return ret;
				}
				case OP_ATAN://assume argument is of type double
				{
					ret = number.arctan(arg);
					return ret;
				}
				case OP_LOG://assume argument is of type double
				{
					ret = number.ln(arg);
					return ret;
				}
				case OP_LOG10://assume argument is of type double
				{
					ret = number.log(arg);
					return ret;
				}
				case OP_ROUND://assume argument is of type double
				{
					ret = number.round(arg);
					return ret;
				}
				case OP_CEIL://assume argument is of type double
				{
					ret = number.ceil(arg);
					return ret;
				}
				case OP_FLOOR://assume argument is of type double
				{
					ret = number.floor(arg);
					return ret;
				}
				case OP_CHR:
				case OP_ORD:
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}
		//	throw new InterpretError("Functions not implemented yet");

		 default :
			throw new
			InterpretError("Invalid long statement");
	  }
   }
   public static long evalLongExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  {  
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalLongExpr(thenNode, env);
			} else {
				return evalLongExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case LONG :
					 return ((long[]) r.data[0])[0];
				  case INTEGER :
					 return ((int[]) r.data[0])[0];
				  case SHORT :
					 return ((short[]) r.data[0])[0];
			   }
			}
			throw new InterpretError("Expression does not return a long");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
		 if (node.jjtGetNumChildren() != 2)
		 {   throw new 
			 InterpretError("Bad binary operator");
		 }
		 SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
		 long leftResult = evalLongExpr(leftNode, env);
		 long rightResult = evalLongExpr(rightNode, env);
		 switch (node.opcode)
		 {  case OP_DIV :
			   return leftResult / rightResult;
			case OP_MIN :
			   return Math.min(leftResult, rightResult);
			case OP_MINUS :
			   return leftResult - rightResult;
			case OP_MAX :
			   return Math.max(leftResult, rightResult);
			case OP_MUL :
			   return leftResult * rightResult;
			case OP_PLUS :
			   return leftResult + rightResult;
			case OP_POW :
			   return (long) Math.pow(leftResult, rightResult);
			case OP_MOD :
			   throw new 
			   InterpretError("Modulo operator used incorrectly");
			case OP_EQ :
			case OP_GE :
			case OP_GT :
			case OP_LE :
			case OP_LT :
			case OP_NEQ :
			   throw new 
			   InterpretError("Comparison operator used incorrectly");
			case OP_CAST :
			   throw new 
			   InterpretError("Cast operator not implemented");
			default :
			   throw new 
			   InterpretError("Uknown binary operator");
		 }
		 case OP_UOPERATOR :
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			long childResult = evalLongExpr(childNode, env);
			switch (node.opcode) //REVISIT: assuming a binary tree
			{  case OP_UMINUS :
				  return -childResult;
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL : //REVISIT:
			switch (node.opcode)
			{  case OP_LONG :
				  Long l = (Long) node.getObject();
				  if (l == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return l.longValue();
			   case OP_INTEGER :
				  Integer i = (Integer) node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return i.intValue();
			   case OP_SHORT :
				  Short s = (Short) node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return s.shortValue();
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case LONG :
					 return ((long[]) nri.NRenv.data[col])[row];
				  case INTEGER :
					 return ((int[]) nri.NRenv.data[col])[row];
				  case SHORT :
					 return ((short[]) nri.NRenv.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case LONG :
					 return ((long[]) pi.env.data[col])[row];
				  case INTEGER :
					 return ((int[]) pi.env.data[col])[row];
				  case SHORT :
					 return ((short[]) pi.env.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal"+((ParamInfo) id).type);
			   }
			}
			else if(id.kind == LOCAL)
			{  switch (((LocalInfo) id).type)
			   {  case LONG :
					 return ((LocalInfo) id).longVal;
				  case INTEGER :
					 return ((LocalInfo) id).intVal;
				  case SHORT :
					 return ((LocalInfo) id).intVal;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{//added by Yi Zheng April 18, 2002
				if(!WEIZHONG)
				{
					switch (((StateInfo) id).type)
				   {  case LONG :
						 return ((StateInfo) id).longVal;
					  case INTEGER :
						 return ((StateInfo) id).intVal;
					  case SHORT :
						 return ((StateInfo) id).intVal;
					  default :
						 throw new 
						 	InterpretError("Incompatible identifier");
				   }
			    }else
			    {
			    	StateInfo si = (StateInfo) id;
			   		col = si.index;
			   		int row = si.env.row;
			   		switch ( si.type )
			   		{	case LONG:
			   				return ((long[]) si.env.data[col])[row];
			   			case INTEGER :
			           		return ((int[]) si.env.data[col])[row];
			      		case SHORT :
			           		return ((short[]) si.env.data[col])[row];
			      		default :
	        	           throw new
			           		InterpretError("Incompatible identifier");
			   		}
			    	
			    }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
if (firstchild.opcode == OP_IDENTIFIER) {
	SimpleNode virnode = firstchild.jjtDeepDuplicate();
	virnode.name = iddom.name;
	long virret = evalLongExpr(virnode, env);
	return virret;
} else {
 				long ret =  evalLongExpr(iddom.getTree(), env);
				return ret;
}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
			InterpretError("This is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}
			
			long ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type long
				{
					long arg = evalLongExpr(argument, env);
					ret = (long) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					throw new InterpretError("Bad use of sin()");
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					throw new InterpretError("Bad use of cos()");
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					double denom = (double) Math.log(10);
					ret = (long) (Math.log(arg) / denom);
					return ret;
				}
				case OP_ROUND://assume argument is of type double
				{
					double arg = evalFloatExpr(argument, env);
					ret = (long) Math.round(arg);
					return ret;
				}
				case OP_CEIL://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.ceil(arg);
					return ret;
				}
				case OP_FLOOR://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (long) Math.floor(arg);
					return ret;
				}
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					int code = evalIntExpr(argument, env);
					if(code < 0 || code > 255)
						throw new InterpretError("Invalid argument to chr()");
					else
						return (long) code;
				}					
				case OP_ORD:
				{
					String s = evalStringExpr(argument, env);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (long) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}	
		//	throw new InterpretError("Functions not implemented yet");

		 default :
			throw new 
			InterpretError("Invalid long statement");
	  }
   }   
   public static short evalShortExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  { 
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalShortExpr(thenNode, env);
			} else {
				return evalShortExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case SHORT :
					 return ((short[]) r.data[0])[0];
			   }
			}
			throw new InterpretError("Expression does not return a short");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
			if (node.jjtGetNumChildren() != 2)
			{   throw new 
				InterpretError("Bad binary operator");
			}
			SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
			SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
			short leftResult = evalShortExpr(leftNode, env);
			short rightResult = evalShortExpr(rightNode, env);
			switch (node.opcode)
			{  case OP_DIV :
				  return (short) (leftResult / rightResult);
			   case OP_MIN :
				  return (short) Math.min(leftResult, rightResult);
			   case OP_MINUS :
				  return (short) (leftResult - rightResult);
			   case OP_MAX :
				  return (short) Math.max(leftResult, rightResult);
			   case OP_MOD :
				  return (short) (leftResult % rightResult);
			   case OP_MUL :
				  return (short) (leftResult * rightResult);
			   case OP_PLUS :
				  return (short) (leftResult + rightResult);
			   case OP_POW :
				  return (short) Math.pow(leftResult, rightResult);
			   case OP_EQ :
			   case OP_GE :
			   case OP_GT :
			   case OP_LE :
			   case OP_LT :
			   case OP_NEQ :
				  throw new 
				  InterpretError("Comparison operator used incorrectly");
			   case OP_CAST :
				  throw new 
				  InterpretError("Cast operator not implemented");
			   default :
				  throw new 
				  InterpretError("Invalid binary operator");
			}
		 case OP_UOPERATOR : //REVISIT: assuming a unary tree
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			short childResult = evalShortExpr(childNode, env);
			switch (node.opcode)
			{  case OP_UMINUS :
				  return (short) -childResult;
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL :
			switch (node.opcode) 
			{  case OP_SHORT :
				  Short s = (Short) node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return s.shortValue();
			   case OP_INTEGER:
				  Integer i = (Integer) node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return i.shortValue();
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case SHORT :
					 return ((short[]) nri.NRenv.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER) //used in ijoin
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case SHORT :
					 return ((short[]) pi.env.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if(id.kind == LOCAL) //used in in/out
			{  switch (((LocalInfo) id).type)
			   {  case SHORT :
					 return (short) ((LocalInfo) id).intVal;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{//Yi Zheng April 18, 2002
			   if(!WEIZHONG)  //invoked by top level comp call
			   {
					switch (((StateInfo) id).type)
			   		{  case SHORT :
					 	return (short) ((StateInfo) id).intVal;
				  		default :
					 		throw new 
					 		InterpretError("Incompatible identifier");
			   		}	
			   } else //invoked by update comp call
			   {
			   	   StateInfo si = (StateInfo) id;
			   	   col = si.index;
			   	   int row = si.env.row;
			   		switch ( si.type )
			   		{
			      		case SHORT :
			           		return ((short[]) si.env.data[col])[row];
			      		default :
	        	           throw new
			           			InterpretError("Incompatible identifier");
			   		}
			   }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
if (firstchild.opcode == OP_IDENTIFIER) {
	SimpleNode virnode = firstchild.jjtDeepDuplicate();
	virnode.name = iddom.name;
	short virret = evalShortExpr(virnode, env);
	return virret;
} else {
 				short ret =  evalShortExpr(iddom.getTree(), env);
				return ret;
}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
			InterpretError("This type is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}
			
			short ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type int
				{
					int arg = evalIntExpr(argument, env);
					ret = (short) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					throw new InterpretError("Bad use of sin()");
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					throw new InterpretError("Bad use of cos()");
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					double denom = (double) Math.log(10);
					ret = (short) (Math.log(arg) / denom);
					return ret;
				}
				case OP_ROUND://assume argument is of type float
				{
					float arg = evalFloatExpr(argument, env);
					ret = (short) Math.round(arg);
					return ret;
				}
				case OP_CEIL://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.ceil(arg);
					return ret;
				}
				case OP_FLOOR://assume argument is of type double
				{
					double arg = evalDoubleExpr(argument, env);
					ret = (short) Math.floor(arg);
					return ret;
				}
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					int code = evalIntExpr(argument, env);
					if(code < 0 || code > 255)
						throw new InterpretError("Invalid argument to chr()");
					else
						return (short) code;
				}					
				case OP_ORD:
				{
					String s = evalStringExpr(argument, env);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (short) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}	
		//	throw new InterpretError("Functions not implemented yet");

		 default :
			throw new
			InterpretError("Invalid short expression");
	  }
   }   
   public static String evalStringExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	  switch(node.type)
	  { 
		case OP_IFTHENELSE:
		 	SimpleNode ifNode = (SimpleNode)node.jjtGetChild(0);
		 	SimpleNode thenNode = (SimpleNode)node.jjtGetChild(1);
			SimpleNode elseNode; 
			if (node.jjtGetNumChildren() == 3)
				elseNode = (SimpleNode) node.jjtGetChild(2);
			else
				throw new InterpretError("There must be ELSE clause");

			if (evalBooleanExpr(ifNode, env)) {
				return evalStringExpr(thenNode, env);
			} else {
				return evalStringExpr(elseNode, env);
			}
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case STRING :
//				case BOOLEAN:
					 return ((String[]) r.data[0])[0];
			   }
			}
			throw new InterpretError("Expression does not return a string");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
		 if (node.jjtGetNumChildren() != 2)
		 {   throw new 
			 InterpretError("Bad binary operator");
		 }
		 SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
		 String leftResult = evalStringExpr(leftNode, env);
		 String rightResult = evalStringExpr(rightNode, env);
		 switch (node.opcode)
		 {  case OP_CAT :
			   return leftResult.concat(rightResult);
			case OP_CAST :
			   throw new 
			   InterpretError("Cast operator not implemented");
			default :
			   throw new 
			   InterpretError("Uknown binary operator");
		 }
		 case OP_UOPERATOR ://302 -- by Yu
//System.out.println("in EvalExpr:"+node.opcode+","+node.type);
			if (node.opcode==OP_SUBSTR)		//added by Yu, Oct 19,2004
			{
		        SimpleNode id = (SimpleNode)node.jjtGetChild(0);
				String val = evalStringExpr(id,env);
        		SimpleNode start = (SimpleNode)node.jjtGetChild(1);
        		int from=0;
        		if (start.getInfo()!=null)
         			from = ((Integer)start.getInfo()).intValue();
         		else if (start.type!=OP_LITERAL)
		 		{
		 			from = evalIntExpr(start,env);
		 		}
        		int to;
        		String substr;
				if (node.jjtGetNumChildren()==3)
				{
         		 SimpleNode end = (SimpleNode)node.jjtGetChild(2);
//System.out.println("!!!	actSubstr:"+end.getInfo()+"name:"+end.name+"opcode:"+end.opcode+"type:"+end.type+"id:"+end.identifier);		
				 if (end.getInfo()!=null)
				 {
					to = ((Integer)end.getInfo()).intValue();
					substr = val.substring(from,to);
		 		 }
		 		 else
		 		 {
		 			if (end.type!=OP_LITERAL)
		 			{
		 				to = evalIntExpr(end,env);
		 				substr = val.substring(from,to);
		 			}
		 			else
					substr = val.substring(from);
		 		 }
				}
				else
		 	 		substr = val.substring(from);
			 return substr;
			}
			else
				  throw new 
				  InterpretError("Invalid unary operation");
		 case OP_LITERAL : //REVISIT:
			switch (node.opcode)
			{  case OP_STRING :
				  String str = (String) node.getObject();
				  if (str == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return str;
//			   case OP_BOOLEAN:
//				 byte btr = (byte) node.getObject();
//				 String str = new String("");
//				 return str.ValueOf(btr);
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case STRING :
//				case BOOLEAN:
					 return ((String[]) nri.NRenv.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case STRING :
//				case BOOLEAN:
					 return ((String[]) pi.env.data[col])[row];
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if(id.kind == LOCAL)//REVISIT:
			{  
			   switch (((LocalInfo) id).type)
			   {  case STRING :
//				case BOOLEAN:
					 return ((LocalInfo) id).stringVal;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{//Yi Zheng April 19, 2002
				if(!WEIZHONG)
				{
					switch (((StateInfo) id).type)
				   {  case STRING :
	//				case BOOLEAN:
						 return ((StateInfo) id).stringVal;
					  default :
						 throw new 
						 InterpretError("Incompatible identifier");
				   }
				 }else
				 {
				 	StateInfo si = (StateInfo) id;
				 	col = si.index;
				 	int row = si.env.row;
				 	switch (si.type)
				 	{
				 		case STRING:
				 			return 	((String[]) si.env.data[col])[row];
				 		default:
				 			throw new InterpretError("Incompatible identifier");
				 	}
				 }
			}
			else if (id.kind == DOM)
			{  DomInfo domi = (DomInfo) id;
			   Domain iddom = domi.dom;
			   if (iddom.tree != null) {
				SimpleNode idtree = (SimpleNode) iddom.tree;
SimpleNode firstchild = (SimpleNode)iddom.tree.jjtGetChild(0);
if (firstchild.opcode == OP_IDENTIFIER) {
	SimpleNode virnode = firstchild.jjtDeepDuplicate();
	virnode.name = iddom.name;
	String virret = evalStringExpr(virnode, env);
	return virret;
} else {
 				String ret =  evalStringExpr(iddom.getTree(), env);
				return ret;
}
			   } else 
				throw new
				InterpretError("Virtual domain set error.");
			}
			throw new 
			InterpretError("This is not implemented yet.");
		case OP_FUNCTION: //added by Yi Zheng April 10, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			if(DEBUG)
			{
				Utility.dump(argument, "argument ");
			}
			
			String ret;
			switch(node.opcode)
			{
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					int code = evalIntExpr(argument, env);
					if(code < 0 || code > 255)
						throw new InterpretError("Invalid argument to chr()");
					else
					{
						char startC = '\u0000';
						char[] newC = new char[1];
						newC[0] = (char)(startC + ((char) code));
						ret = new String(newC);
						return ret;
					}
				}					
				default:
					throw new InterpretError("Unknown function operator");
			}	
		//	throw new InterpretError("Functions not implemented yet");

		 default :
			throw new 
			InterpretError("Invalid string statement");
	  }
   }   

// for in If case, 
   public static Object evalObjectExpr(SimpleNode node, Environment env)
   throws InterpretError
   {  int col;
	Object o = null;	// for return value
	long longl;
	int intl;
	short shortl;
	double doubled;
	float doublef;
	byte byteb;
	  switch(node.type)
	  {  
		case OP_JOINOPERATOR :
		 case OP_PROJECT :
		 case OP_TSELECT :
		 case OP_SELECT :
		 case OP_QSELECT :
		 case OP_PICK :
			//REVISIT: need to give filename. Will need to delete it.
			Relation r = interpreter.evaluateTLExpression(node,"gogo",emptyht,env);
			if(r.numtuples==1 && r.numattrs==1)
			{  //Relation r is a singleton scalar. 
			   //Return the value it contains as an integer.
			   if (r.data == null)
			   {  r.load();
			   }
			   switch (r.domains[0].type)
			   {  case LONG :
			   	  case IDLIST:
			   	  case COMP:
					 longl = ((long[]) r.data[0])[0];
					o = new Double(longl);
					return o;
				  case INTEGER :
					 intl = ((int[]) r.data[0])[0];
					o = new Double(intl);
					return o;
				  case SHORT :
					 shortl = ((short[]) r.data[0])[0];
					o = new Double(shortl);
					return o;
				case STRING:
			   	  case TEXT:
			   	  case EXPR:
			   	  case STMT:
					return ((String[]) r.data[0])[0];
				  case DOUBLE :
					doubled = ((double[]) r.data[0])[0];
					o = new Double(doubled);
					return o;
				  case FLOAT :
					doublef = ((float[]) r.data[0])[0];
					o = new Double(doublef);
					return o;
				  case BOOLEAN :
					byteb = ((byte[]) r.data[0])[0];
					o = new Byte(byteb);
					return o;
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			throw new InterpretError("Expression does not return a long");
		 case OP_BIOPERATOR : //REVISIT: assuming a binary tree
		 if (node.jjtGetNumChildren() != 2)
		 {   throw new 
			 InterpretError("Bad binary operator");
		 }
		 SimpleNode leftNode = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode rightNode = (SimpleNode) node.jjtGetChild(1);
		 Object leftResult = evalObjectExpr(leftNode, env);
		 Object rightResult = evalObjectExpr(rightNode, env);
		// assume no string type should be here
		if(leftResult.getClass().getName().compareTo("java.lang.String") == 0)
			throw new InterpretError("Comparison not implemented for String type");
		double ro;
		double lResult = ((Double)leftResult).doubleValue();
		double rResult = ((Double)rightResult).doubleValue();
		 switch (node.opcode)
		 {  case OP_DIV :
			   ro = lResult / rResult;
			   break;
			case OP_MIN :
			   ro = Math.min(lResult, rResult);
			   break;
			case OP_MINUS :
			   ro = lResult - rResult;
			   break;
			case OP_MAX :
			   ro = Math.max(lResult, rResult);
			   break;
			case OP_MUL :
			   ro = lResult * rResult;
			   break;
			case OP_PLUS :
			   ro = lResult + rResult;
			   break;
			case OP_POW :
			   ro = Math.pow(lResult, rResult);
			   break;
			case OP_MOD :
			   throw new 
			   InterpretError("Modulo operator used incorrectly");
			case OP_EQ :
			case OP_GE :
			case OP_GT :
			case OP_LE :
			case OP_LT :
			case OP_NEQ :
			   throw new 
			   InterpretError("Comparison operator used incorrectly");
			case OP_CAST :
			   throw new 
			   InterpretError("Cast operator not implemented");
			default :
			   throw new 
			   InterpretError("Unknown binary operator");
		 }
		o = new Double(ro);
		return o;
		 case OP_UOPERATOR :
// String shouldn't be here
			if (node.jjtGetNumChildren() != 1)
			{   throw new 
				InterpretError("Bad unary operator");
			}
			SimpleNode childNode = (SimpleNode) node.jjtGetChild(0);
			Object childResult = evalObjectExpr(childNode, env);
			switch (node.opcode) //REVISIT: assuming a binary tree
			{  case OP_UMINUS :
				double ol = ((Double)childResult).doubleValue();
				ol = -ol;
				o = new Double(ol);
				return o;
			   case OP_UPLUS :
				  return childResult;
			   default :
				  throw new 
				  InterpretError("Invalid unary operator");
			}
		 case OP_LITERAL : //REVISIT:
			switch (node.opcode)
			{
			  case OP_LONG :
			   case OP_COMP:
				  Long l = (Long) node.getObject();
				  if (l == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  o = new Double(l.doubleValue());
			//	  return l; modified by Yi Zheng March 23, 2002
				return o;
			   case OP_INTEGER :
				  Integer i = (Integer) node.getObject();
				  if (i == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  o = new Double(i.doubleValue());
				  return o;
			   case OP_SHORT :
				  Short s = (Short) node.getObject();
				  if (s == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  o = new Double(s.doubleValue());
				  return o;
			   case OP_STRING:
				  String str = (String) node.getObject();
				  if (str == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return str;
			   case OP_DOUBLE :
				  Double doub = (Double) node.getObject();
				  if (doub == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return doub;
			   case OP_FLOAT :
				  Float f = (Float) node.getObject();
				  if (f == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  o = new Double(f.doubleValue());
				  return o;
			   case OP_BOOLEAN :
				  Byte b = (Byte) node.getObject();
				  if (b == null)
				  {  throw new  //should never happen
					 InterpretError("Literal has no value");
				  }
				  return b;
			   case OP_DC :
			   case OP_DK :
			   case OP_NULL :
				  throw new 
				  InterpretError("Null values not implemented in comps");
			   case OP_TEXT:
			   case OP_STMT:
			   case OP_EXPR:
			   case OP_IDLIST:
			   default :
				  throw new 
				  InterpretError("Invalid literal");
			}
		 case OP_IDENTIFIER : 
			IDInfo id = env.lookup(node.name, true);
			if(id ==null)
			{  throw new 
			   InterpretError("Unknown variable "+ node.name);
			}
			if (id.kind==REL)//added by Yu Nov 15,2004
			{
				Relation rels = env.lookupRel(node.name, true);
				if (rels.rvc==SCALAR)
				{
				 id.kind=DOM;
				 Domain d = env.lookupDom(node.name, true);
				 id = new DomInfo(d);
				}
			}	//end of addition by Yu
			if (id.kind == NR)
			{  NRInfo nri = (NRInfo) id;
			   col = nri.index;
			   int row = nri.NRenv.row;
			   switch ( nri.type )
			   {  case LONG :
			   	  case COMP:
					longl = ((long[]) nri.NRenv.data[col])[row];
					o = new Double(longl);
					return o;
				  case INTEGER :
					intl = ((int[]) nri.NRenv.data[col])[row];
					o = new Double(intl);
					return o;
				  case SHORT :
					shortl = ((short[]) nri.NRenv.data[col])[row];
					o = new Double(shortl);
					return o;
				  case STRING:
					 return ((String[]) nri.NRenv.data[col])[row];
				  case DOUBLE :
					doubled = ((double[]) nri.NRenv.data[col])[row];
					o = new Double(doubled);
					return o;
				  case FLOAT :
					doublef = ((float[]) nri.NRenv.data[col])[row];
					o = new Double(doublef);
					return o;
				  case BOOLEAN :
					byteb = ((byte[]) nri.NRenv.data[col])[row];
					o = new Byte(byteb);
					return o;
			   	  case TEXT:
			   	  case STMT:
			   	  case EXPR:
			   	  case IDLIST:
				  default :
					 throw new 
					 InterpretError("Invalid literal");
			   }
			}
			else if (id.kind == PARAMETER)
			{  ParamInfo pi = (ParamInfo ) id;
			   col = pi.index;
//               data= pi.env.data;
			   int row = pi.env.row;
			   switch ( pi.type )
			   {  case LONG :
			   	  case COMP:
					longl = ((long[]) pi.env.data[col])[row];
					o = new Double(longl);
					return o;
				  case INTEGER :
					intl = ((int[]) pi.env.data[col])[row];
					o = new Double(intl);
					return o;
				  case SHORT :
					shortl = ((short[]) pi.env.data[col])[row];
					o = new Double(shortl);
					return o;
				  case STRING :
					 return ((String[]) pi.env.data[col])[row];
				  case DOUBLE :
					doubled = ((double[]) pi.env.data[col])[row];
					o = new Double(doubled);
					return o;
				  case FLOAT :
					doublef = ((float[]) pi.env.data[col])[row];
					o = new Double(doublef);
					return o;
				  case BOOLEAN :
					byteb = ((byte[]) pi.env.data[col])[row];
					o = new Byte(byteb);
					return o;
			   	  case TEXT:
			   	  case STMT:
			   	  case EXPR:
			   	  case IDLIST:
				  default :
					 throw new 
					 InterpretError("Invalid literal"+((ParamInfo) id).type);
			   }
			}
			else if(id.kind == LOCAL)
			{  switch (((LocalInfo) id).type)
			   {  case LONG :
			   	  case COMP:
					longl = ((LocalInfo) id).longVal;
					o = new Double(longl);
					return o;
				  case INTEGER :
					intl = ((LocalInfo) id).intVal;
					o = new Double(intl);
					return o;
				  case SHORT :
					shortl = (short)((LocalInfo) id).intVal;
					o = new Double(shortl);
					return o;
				  case STRING :
					String so = ((LocalInfo) id).stringVal;
					return so;
				  case DOUBLE :
					doubled = ((LocalInfo) id).doubleVal;
					o = new Double(doubled);
					return o;
				  case FLOAT :
					doublef = ((LocalInfo) id).floatVal;
					o = new Double(doublef);
					return o;
				  case BOOLEAN :
					byteb = (byte)((LocalInfo) id).intVal;
					o = new Byte(byteb);
					return o;
			   	  case TEXT:
			   	  case STMT:
			   	  case EXPR:
			   	  case IDLIST:
				  default :
					 throw new 
					 InterpretError("Incompatible identifier");
			   }
			}
			else if(id.kind == STATE)
			{//Yi Zheng April 18, 2002
				if(!WEIZHONG)
				{
				switch (((StateInfo) id).type)
				   {  case LONG :
				   	  case COMP:
						longl = ((StateInfo) id).longVal;
						o = new Double(longl);
						return o;
					  case INTEGER :
						intl = ((StateInfo) id).intVal;
						o = new Double(intl);
						return o;
					  case SHORT :
						shortl = (short)((StateInfo) id).intVal;
						o = new Double(shortl);
						return o;
					  case STRING :
						String so = ((StateInfo) id).stringVal;
						return so;
					  case DOUBLE :
						doubled = ((StateInfo) id).doubleVal;
						o = new Double(doubled);
						return o;
					  case FLOAT :
						doublef = ((StateInfo) id).floatVal;
						o = new Double(doublef);
						return o;
					  case BOOLEAN :
						byteb = (byte)((StateInfo) id).intVal;
						o = new Byte(byteb);
						return o;
				   	  case TEXT:
				   	  case STMT:
				   	  case EXPR:
				   	  case IDLIST:
					  default :
						 throw new 
						 InterpretError("Incompatible identifier");
				   }
				 }else
				 {
					 StateInfo si = (StateInfo) id;
				   col = si.index;
				   int row = si.env.row;
				   switch ( si.type )
				   {  case LONG :
				   	  case COMP:
						longl = ((long[]) si.env.data[col])[row];
						o = new Double(longl);
						return o;
					  case INTEGER :
						intl = ((int[]) si.env.data[col])[row];
						o = new Double(intl);
						return o;
					  case SHORT :
						shortl = ((short[]) si.env.data[col])[row];
						o = new Double(shortl);
						return o;
					  case STRING :
						 return ((String[]) si.env.data[col])[row];
					  case DOUBLE :
						doubled = ((double[]) si.env.data[col])[row];
						o = new Double(doubled);
						return o;
					  case FLOAT :
						doublef = ((float[]) si.env.data[col])[row];
						o = new Double(doublef);
						return o;
					  case BOOLEAN :
						byteb = ((byte[]) si.env.data[col])[row];
						o = new Byte(byteb);
						return o;
				   	  case TEXT:
				   	  case STMT:
				   	  case EXPR:
				   	  case IDLIST:
					  default :
						 throw new 
						 InterpretError("Incompatible identifier");
				   }	
				 	
				 }
			}
			throw new 
			InterpretError("This is not implemented yet.");
		 default :
			throw new 
			InterpretError("Invalid long statement");
	  }
   }   

}
