import java.util.*;
import java.io.*;

public class ExprTable implements Constants, TypeConstants
{  private static final String EXPR_FILE_NAME = ".expr";
   private Environment env;
   private RelTable reltable;
   private DomTable domtable;
   private CompTable comptable;
   private Hashtable exprtable;    // expression table.
  

   public ExprTable(Environment env, CompTable comp)
   throws InterpretError
   {  exprtable = new Hashtable();
	  this.env = env;
	  reltable = env.reltable;
	  domtable = env.domtable;
	  comptable = comp;
	  load();
   }      
   // Dump the expression information to disk file .expr.
   public void dump()
   {  // clear the hash table first.
	  exprtable.clear();

	  domtable.fillExprTable(exprtable);
	  reltable.fillExprTable(exprtable);
	  comptable.fillExprTable(exprtable);

	  try
	  {  FileOutputStream out = new FileOutputStream(EXPR_FILE_NAME);
		 ObjectOutput s = new ObjectOutputStream(out);
		 s.writeObject(exprtable);
		 s.flush();
	  }
	  catch(NotSerializableException e)
	  {  Global.println("ExprTable.dump(): "+e); 
	  }
	  catch(IOException e)
	  {  Global.println("ExprTable.dump(): "+e); 
	  }
   }         

/*
   // Dump the expression information to disk file .expr.
   //	-- modified by Weizhong Sun, June 14th 1999
   public void dump()
   {  // clear the hash table first.
	  exprtable.clear();

	  domtable.fillExprTable(exprtable);
	  reltable.fillExprTable(exprtable);
	  comptable.fillExprTable(exprtable);

	  try
	  {  
	  	DataOutputStream out = new DataOutputStream(new FileOutputStream(EXPR_FILE_NAME, true));
		 out.writeBytes(exprtable);
		 out.flush();
	  }
	  catch(NotSerializableException e)
	  {  Global.println("ExprTable.dump(): "+e); 
	  }
	  catch(IOException e)
	  {  Global.println("ExprTable.dump(): "+e); 
	  }
   }         
*/

   // Load the expression information from .expr and .view.
   public void load()
   throws InterpretError
   {  // load expression table from disk.
	  try
	  {  // read .expr to exprTable.
		 FileInputStream in = new FileInputStream(EXPR_FILE_NAME);
		 ObjectInputStream s = new ObjectInputStream(in);
		 exprtable = (Hashtable) s.readObject();
	  }
	  catch(ClassNotFoundException e)
	  {}
	  catch(IOException e)
	  {}

	  // Display the table contents.
//	   showTable();

	  // insert tree root into domTable and relTable.
	  for(Enumeration e = exprtable.keys(); e.hasMoreElements();)
	  {  ExprEntry exprEnt = (ExprEntry)exprtable.get(e.nextElement());
		 String name = exprEnt.name;
//System.out.println("in ExprTable, load(), name: " + name);

		 if(name.startsWith("."))
		 {  // virtual domain name
			Domain d = env.lookupDom(name.substring(1), true);
			if(d != null)
			{ d.tree = (SimpleNode) exprEnt.root;
			}
			else
			{  Global.println(
					   "Warning: Attempt to load a tree failed since "+
					   "virtual domain "+name+" can not be found. "+
					   "The system will skip this entry."  );
			}
//            domtable.insertRoot(name.substring(1),(SimpleNode) exprEnt.root);
		 }
		 else if(name.startsWith("#")) //Modified by Yi Zheng May 14, 2002
		 {  // This is for a computation that is stored in a relation
			// It belongs in the comptable
			// name consists of two parts: (1) #surr, where surr is the String rep. of surrogate value
			// (2) $fullName, where fullName is the fully qualified name of a computation,
			// which is formed by parentName!selfName.
			// e.g. #12$COLLECTIVE!FARMER!CNT denotes a surrogate of 12, a computation
			// CNT defined inside the body of FARMER, which in turn is defined inside
			// COLLECTIVE; thus parentName is COLLECTIVE!FARMER, selfName is CNT
			try
			{  SimpleNode root = (SimpleNode) exprEnt.root;
			   Computation c=new Computation(root, env);
			 //  comptable.addEntry(name.substring(1), c);
	
			   //step 1, parse the surrogate
			   //note that we are dealing with nested computations with surrogates,
			   //so it's got to have non-empty parentName
			   int index1 = name.indexOf((int) '$');
			   String surrogate = name.substring(1, index1);
			   comptable.addEntry(surrogate, c);
			   //step 2, parse the fully qualified computation name
			   int index2 = name.lastIndexOf((int) '!');
			   String fullName = name.substring(index1 + 1);
			   comptable.addEntry(fullName, new Long(surrogate));
			   //step 3, update the parentName attribute of c
			   String parentName = fullName.substring(0, index2);
			   c.parentName = parentName;
			}
			catch (InterpretError err)
			{  Global.println(
				 "Warning: Attempt create a computation for "
				 +name+" failed. The system will skip this entry"  );
			}
			catch (Exception ee)
			{	Global.processThrowable(ee);}
		 }
/*		 else if(name.startsWith("#")) //Commented out by Yi Zheng May 14, 2002, see above modification
		 {  // This is for a computation that is stored in a relation
			// It belongs in the comptable
			try
			{  SimpleNode root = (SimpleNode) exprEnt.root;
			   Computation c=new Computation(root, env);
			   comptable.addEntry(name.substring(1), c);
			}
			catch (InterpretError err)
			{  Global.println(
				 "Warning: Attempt create a computation for "
				 +name+" failed. The system will skip this entry"  );
			}
		 }*/
		 else
		 {  // view  or computation name.
			// If its a computation, it is a top level one.
			SimpleNode root = (SimpleNode) exprEnt.root;
//System.out.println("in ExprTable, load(), root.opcode: " + root.opcode);
//printNodes(root);
			Relation  r = env.lookupRel(name, true);
			if (r == null)
			{  //root must be the syntax tree of a computation
			   //since the computations aren't in the reltable yet.
			   if (root.opcode == OP_COMPUTATION)
			   {  //The Computation object must be created from the 
				  //sytax tree and then put into the reltable.
				  try
				  {  Computation c=new Computation(root, env);
					 reltable.addRel(c, false);
					 //false: Don't increment reference counts of domains
				  }
				  catch (InterpretError err)
				  {  Global.println(
					   "Warning: Attempt create a computation for "
					   +name+" failed. The system will skip this entry"  );
				  }
			   }
			   else
			   {  //should never happen
				  Global.println(
					   "Warning: Attempt to load a tree for "
					   +name+" failed since it isn't a computation" +
					   "The system will skip this entry." );
			   }
			}
			else // r != null
			{  //root should be for a view
			   if (r.rvc == VIEW)
			   {  //The Relation object was already put into the 
				  //reltable, we only need to add the tree for the view
				  r.tree = root;
			   }
			   else
			   {  Global.println(
					   "Warning: Attempt to load a tree for "
					   +name+" failed since it isn't a view" +
					   "The system will skip this entry."  );
			   }
			}
		 }
	  }
   }            
   // Print table information.
   public void showTable()
   {Global.println("+---------------- Expression Table ---------------+");
	for(Enumeration e = exprtable.keys(); e.hasMoreElements();)
	{  ExprEntry exprEnt = (ExprEntry)exprtable.get(e.nextElement());
	   Global.println("name: "+ exprEnt.name);
	   if(exprEnt.name.startsWith("#")) //it is a computation
	   {  //((Computation) exprEnt.root).tree.dump(" ");
	   }
	   else //it is either a virtual domain or a view
	   {  //((SimpleNode) exprEnt.root).dump(" ");
	   }
	   //outstream.println("---");
	}
	Global.println("+-------------------------------------------------+");
   }         
   // Print table information.
   //Orignial version:
   public void showTable2(){
	  Global.println("+---------------- Expression Table ---------------+");
	  for(Enumeration e = exprtable.keys(); e.hasMoreElements();){
	ExprEntry exprEnt = (ExprEntry)exprtable.get(e.nextElement());
	Global.println("name: "+ exprEnt.name);
	((SimpleNode) exprEnt.root).dump(" ");
	Global.println("---");
	  }
	  Global.println("+-------------------------------------------------+");
   }         

   public Hashtable getExprtable() {
	return exprtable;
   }

/*
public void printNodes(SimpleNode node) {
int childnum = node.jjtGetNumChildren();
System.out.println("node.name: " + node.name + " childrennum: " + childnum);
SimpleNode child;
for (int i = 0; i < childnum; i++) {
	child = (SimpleNode)node.jjtGetChild(i);
	System.out.println("i: " + i + " child.name: " + child.name);
	System.out.println("child.opcode: " + child.opcode);
	if (i == childnum-1)
		printNodes(child);
}
}
*/

}
