
/**
 *  The purpose of this class is to collect methods used for 
 *  checking the semantics of user code. These methods should all
 *  be public and static.
 */
class SemanticChecker implements Constants
{


   private static void SC_CompCommand(SimpleNode stmt, Environment env)
   throws SemanticCheckError
   {  int opcode = stmt.opcode;

	  switch(opcode) 
	  {  case OP_HELP : 
	    	case OP_QUIT : 
	    	case OP_TRACE : 
	    	case OP_INPUT : 
	    	case OP_SSD : 
	    	case OP_SSR : 
	    	case OP_DEBUG : 
	    	case OP_BATCH : 
	    	case OP_EXPERT : 
	    	case OP_TIME : 
	    	case OP_DELD : 
	    	case OP_DELR : 
	    	case OP_PR : 
	    	case OP_SC : 
	    	case OP_SD : 
	    	case OP_SR : 
	    	case OP_SRD : 
	    	case OP_PRINT : 
	    	case OP_UNDO : 
			break;
		 default : // Parser should prevent flow from reaching here
			throw new SemanticCheckError("Invalid statement");
	  }
   }      

   private static void SC_CompAssignment(SimpleNode stmt, Environment env)
   throws SemanticCheckError
   {  int opcode = stmt.opcode;

	  switch(opcode) 
	  {  case OP_ASSIGN : 
			break;
		 case OP_ASSIGNADD :
			break;
		 default : // Parser should prevent flow from reaching here
			throw new SemanticCheckError("Invalid statement");
	  }
   }      
   private static void SC_CompDeclaration(SimpleNode stmt, Environment env)
   throws SemanticCheckError
   {  int opcode = stmt.opcode;

	  switch(opcode) 
	  {  case OP_COMPUTATION :
			break;
		 case OP_DOMAIN :
			//Make sure no new domains have the same name as a parameter
			SimpleNode newDomains = (SimpleNode) stmt.jjtGetChild(0);
			int numChildren = newDomains.jjtGetNumChildren();
			for(int i=0; i<numChildren; i++)
			{  String idname = ((SimpleNode) newDomains.jjtGetChild(i)).name;
			   if (env.lookupParam(idname, true) != null)
			   {  throw new SemanticCheckError
				  ("Declaration of domain '"+idname+"' failed. "+
				   "There is a parameter of the same name");
			   }
			}
			break;
		 case OP_LET :
			break;
		 case OP_RELATION :
			break;
		 case OP_VIEW :
			break;
		 default : // Parser should prevent flow from reaching here
			throw new SemanticCheckError("Invalid statement");
	  }
   }      
   private static void SC_CompStatement(SimpleNode stmt, Environment env)
   throws SemanticCheckError
   {  int opcode = stmt.opcode;

	  switch(opcode) 
	  {  case OP_SEQUENCE :
		 case OP_PARALLEL :
		 case OP_CHOICE :
		 case OP_COMPCALL :
		 case OP_CONDITIONAL :
		 case OP_FORLOOP :
		 case OP_WHILELOOP :
		 case OP_EXIT :
		 case OP_DEADLOCK :
		 case OP_EXEC :
		 case OP_STMTBLOCK :
			//These have not been implemented, so we do not have
			//code to do semantic checking for it. So do nothing
			break;
		 default : // Parser should prevent flow from reaching here
			throw new SemanticCheckError("Invalid statement");
	  }
   }      
   private static void SC_CompUpdate(SimpleNode stmt, Environment env)
   throws SemanticCheckError
   {  int opcode = stmt.opcode;

	  switch(opcode) 
	  {  case OP_ADD :
		 case OP_DELETE :
		 case OP_CHANGE : 
			break;
		 default : // Parser should prevent flow from reaching here
			throw new SemanticCheckError("Invalid statement");
	  }
   }      
   /***************************************************************/
   /*               METHODS FOR CHECKING COMPUTATIONS             */
   /***************************************************************/
   public static void SC_Computation(SimpleNode stmt, Environment env)
   throws SemanticCheckError
   {  int type = stmt.type;

	  switch(type) 
	  {  case OP_ASSIGNMENT :
			SC_CompAssignment(stmt, env);
			break;
		 case OP_COMMAND :
//System.out.println("OP_COMMAND, semanticcheck error.");
			SC_CompCommand(stmt, env);
			break;
		 case OP_DECLARATION :
			SC_CompDeclaration(stmt, env);
			break;
		 case OP_STATEMENT :
			SC_CompStatement(stmt, env);
			break;
		 case OP_UPDATE :
			SC_CompUpdate(stmt, env);
			break;
		 default : // Parser should prevent flow from reaching here
			throw new SemanticCheckError("Invalid statement");
	  }
   }      
}
