import java.io.*;
import java.util.*;
import java.util.regex.*;
/**
 * Class Text: for text type elements
 */
class Text implements TypeConstants, Constants
{
	public String name;
	public long length;
	public String abstr;	// abstract of the content
	protected Environment env; 	//Pointer to the frame of the environment
	private static final String DOM_VAL = "_val";
	private static final String DOM_POS = "_pos";
	private static final String DOM_SEQ = "_seq";	
	private static final int NUM_DOM = 3;
	
	/**
	* Constructor
	* @param	name	New text name
	* @param	length	New text length
	* @param	abstr	New text abstract
	* @param	env		Environment
	*/
   	public Text(String name, long length, String abstr, Environment Env)
   	{	this.name = name;
   		this.length = length;
   		this.abstr = abstr;
   		env = Env;
   	}

	/**
	* Init a text from relation, text, disk file or string.
	* @param	name		The name of the text.
	* @param	input		Initial value, could be element names, file path or string value
	* @param	inputType	Input type, could be "Identifier" / "FilePath" / "String"
	* @param	Env			Environment
	*/
   	public Text(String name, String input, String inputType, Environment Env) throws InterpretError
   	{	
   		this.name = name;
   		env = Env;
		char buf[] = new char[BUF_SIZE];
		int cnt;
   		StringBuffer readin = new StringBuffer();
   		String cp;
   		BufferedReader in;
   		BufferedWriter out;
   		PrintStream ps;
   		File outFile;
   		this.length = 0;
   		String tmpStr;
   		
		try
   		{	outFile = new File(name + ".text");
   			if(inputType.compareTo("Identifier") == 0)
   			{	Text t = Env.lookupText(input, true);
   				if(t == null) // init from relation
   				{	Relation r = Env.lookupRel(input, true);
   					if(r == null)
   					{	throw new InterpretError("Text " + input + " not found in text initialization!"); }
   					
   					ps = new PrintStream(new FileOutputStream(outFile),true);
					Global.outstream = ps;
   					r.printDat();
					ps.close();
					Global.outstream = System.out;
   					
   					this.length = (new File(name + ".text")).length();
   					tmpStr = this.toString().replaceAll("\n", "").replaceAll("\t", " ");
   					this.abstr = tmpStr.length() > 40 ? tmpStr.substring(0, 40) : tmpStr;
   					return;
   				}
   				else // init from text
				{	in = new BufferedReader(new FileReader(input + ".text"));
 					while ((cnt = in.read(buf, 0, BUF_SIZE)) != -1)
					{	readin.append(String.copyValueOf(buf, 0, cnt)); }
					in.close();
				}
			}
			else if(inputType.compareTo("FilePath") == 0) // init from disk file
			{	in = new BufferedReader(new FileReader(input));
				while ((cnt = in.read(buf, 0, BUF_SIZE)) != -1)
				{	readin.append(String.copyValueOf(buf, 0, cnt)); }
				in.close();
			}					
			else if(inputType.compareTo("String") == 0) // init from string in parameter "input"
			{	readin.append(input);
			}
			else
			{	throw new InterpretError("Incorrect input type: " + inputType);
   			}
			cp = readin.toString().replaceAll("\r\n", "\n");
			out = new BufferedWriter(new FileWriter(outFile));
			out.write(cp);
			out.close();
			
			tmpStr = cp.replaceAll("\n", "").replaceAll("\t", " ");
			this.abstr = tmpStr.length() > 40 ? tmpStr.substring(0, 40) : tmpStr;
			
			this.length = (new File(name + ".text")).length();
   		}
   		catch(IOException e){ throw new InterpretError("IOException in TextInitialization: " + e.getMessage()); }
   	}

	/**
	* Grep matches into a relation
	* @param	tmpname		The name of returned relation
	* @param	patn		Match pattern
	* @param	val			Given domain for value attr
	* @param	pos			Given domain for position attr
	* @param	seq			Given domain for sequence attr
	*/
	public Relation grep(String tmpname, String patn, Domain val, Domain pos, Domain seq) throws InterpretError
   	{
   		if(val == null && (val = env.lookupDom(DOM_VAL, true)) == null)
   		{	val = new Domain(DOM_VAL, STRING);
   			env.put(val);
   		}
   		
   		if(pos == null && (pos = env.lookupDom(DOM_POS, true)) == null)
   		{	pos = new Domain(DOM_POS, INTEGER);
   			env.put(pos);
   		}

   		if(seq == null && (seq = env.lookupDom(DOM_SEQ, true)) == null)
   		{	seq = new Domain(DOM_SEQ, INTEGER);
   			env.put(seq);
   		}
   		
   		Domain [] domArray = new Domain[3];
		domArray[0] = val;
		domArray[1] = pos;
		domArray[2] = seq;
   			
		Relation r = new Relation(tmpname, domArray, 0, NUM_DOM, 0, env);
   		
   		String readin = this.toString();

		Pattern pattern = Pattern.compile(patn, Pattern.DOTALL|Pattern.MULTILINE); // to grep across lines
		Matcher matcher = pattern.matcher(readin);
		
		Vector vals = new Vector();
		Vector poss = new Vector();
		for(int i=0;matcher.find();i++)
		{	vals.add(i, matcher.group());
			poss.add(i, new Integer(matcher.start()));
		}

		Object[] data = Relation.createDataArray(domArray, vals.size());
		
		for(int i=0;i<vals.size();i++)
		{	((String[])data[0])[i] = (String)vals.get(i);
			((int[])   data[1])[i] = ((Integer)poss.get(i)).intValue();
			((int[])   data[2])[i] = i;
		}
		
		r.append(data,vals.size());
		r.load();
   		return r;
	}
	
	/**
	* Convert text to string
	*/
	public String toString()
	{	BufferedReader in;
		char buf[] = new char[BUF_SIZE];
		StringBuffer readin = new StringBuffer();
		int cnt;
		
   		try
   		{	in = new BufferedReader(new FileReader(name + ".text")); }
		catch (FileNotFoundException fnfe) 
		{	return "";
		}
   		
		try
   		{	while ((cnt = in.read(buf, 0, BUF_SIZE)) != -1)
			{	readin.append(String.copyValueOf(buf, 0, cnt)); }
			in.close();
		}
		catch(IOException e)
		{	Global.println("IO error: " + e.getMessage()); 
			return null;
		}
		
		return readin.toString();
	}
	
	/**
	* Return text file
	*/
	public File getFile()
	{	return new File(name + ".text").getAbsoluteFile();
	}
	
	/**
	* Print text
	*/
	public void print() throws InterpretError
	{
		String readin = this.toString();
		PrettyPrint out = new PrettyPrint();
		
		out.pprintRule(80);
		out.println();
		out.pprint(" ", 35);
		out.pprint("Text " + name, 45);
		out.println();
		out.pprintRule(80);
		out.println();
		out.println(readin);
		out.pprintRule(80);
		out.println();
		out.println("Text " + name + " has " + length + (length > 1 ? " characters" : " character"));
	}
}
