import java.io.*;
import java.util.*;

public class TextTable implements TypeConstants,Constants
{	
	private static final String TEXT_FILE  = ".text";
	
	private Hashtable table;

	/**
	* Constructor
	* @param	env		Evironment
	* @param	load	To load flag
	*/
	public TextTable(Environment env, boolean load)
	{	table = new Hashtable();
		if(load)
		{	this.load(env);
		}
	}
	
	/**
	* Create new text and add to the text table
	* @param	name	New text name
	* @param	length	New text length
	* @param	abstr	New text abstract
	* @param	env		New text environment
	*/
	private void addText(String name, long length, String abstr, Environment env)
	{	if(table.containsKey(name))
			table.remove(name);
		table.put(name, new Text(name, length, abstr, env));
	}
		
	/**
	* Add text to the text table
	* @param	text	Text instance
	*/
	public void addText(Text text)
	{	if(table.containsKey(text.name))
			table.remove(text.name);
		table.put(text.name, text);
	}
	
	/**
	* Return text by given name
	* @param	name	Text name
	*/
	public Text getText(String name){
		return((Text) table.get(name));
   }
	
	/**
	* Load text table
	* @param	env		Environment
	*/
	public void load(Environment env)
	{	
		 // load relation information from .rel.
		try
		{	BufferedReader txtis  = new BufferedReader(new FileReader(TEXT_FILE));
			String name;
			long length;
			String abstr;
			
			// load data to hash table.
			try
			{	while(true)
				{	name = txtis.readLine();
					if (name == null)
						break;
					length = Long.parseLong(txtis.readLine());
					abstr = txtis.readLine();
					
					addText(name, length, abstr, env);
				}
				txtis.close();
			}
			catch(IOException e) { Global.println("error: TextTable: load IOException"); }
		}
		catch(FileNotFoundException e)
		{	addText(".text", 0, "", env);
		}
	}
	
	/**
	* Show text
	* @param	t	Text instance
	*/
	public void showText(Text t) 
	{
		PrettyPrint out = new PrettyPrint();

		// print the title.
		out.pprintRule(34);	
		out.pprint(" Text Entry ", 12);
		out.pprintRule(34);    
		out.println();
		out.pprint("Name                Length              Content", 80);
		out.println();
		out.pprintRule(80);
		out.println();

		out.pprint(t.name, 20);
		out.pprint(t.length, 20);
		out.pprint(t.abstr, 40);
		out.println();

		out.pprintRule(80);
		out.println();
	}
	
	/**
	* Show text table
	* @param	flag	To show system variable flag
	*/
	public void showTable(boolean flag) 
	{
		Text t;
		PrettyPrint out = new PrettyPrint();

		// print the title.
		out.pprintRule(34);    
		out.pprint(" Text Table ", 12);
		out.pprintRule(34);    
		out.println();
		out.pprint("Name                Length              Content", 80);
		out.println();
		out.pprintRule(80);
		out.println();

		for(Enumeration e = table.keys(); e.hasMoreElements(); )
		{	t = (Text) table.get(e.nextElement());
			if(flag || !t.name.startsWith("."))
			{	out.pprint(t.name, 20);
				out.pprint(t.length, 20);
				out.pprint(t.abstr, 40);
				out.println();
			}
		}
		
		out.pprintRule(80);
		out.println();
	}		
	
	/**
	* Dump text table to disk
	*/
	public void dump()
	{	
		Text t;
		try
		{	PrintWriter txtos = new PrintWriter(new BufferedWriter(new FileWriter(TEXT_FILE)));
			
			for(Enumeration e = table.keys(); e.hasMoreElements();)
			{	t = (Text)table.get(e.nextElement());
				txtos.println(t.name);
				txtos.println(t.length);
				txtos.println(t.abstr);
			}
			
			txtos.close();
		}
		catch(IOException e){ Global.println("error: TextTable: dump IOException");}
	}
	
	/**
	* Delete a text from the TextTable.
	* @param   name   The name of the text.
	*/
	public void rmText(String name)
	{	Text t = (Text) table.get(name);
		
		if(t != null)
		{	table.remove(name);
			
			File tf = new File(name + ".txt");
			if(tf.exists())
				tf.delete();
		}
		else
			Global.println("warning: text '"+name+"' is not found in " + "the TextTable.");
	}	
}
