/**
 * Structure that stores information about XfigObj
 * Author: Lili Zhu
 * May, 2005
**/
import java.util.*;
import java.io.*;
import java.text.*;

class XfigObj
{
	public int obj_type;
	public int sub_type;
	public int line_style;
	public int thickness;
	public int line_color;
	public int text_color;
	public int fill_color;
	public int depth;
	public int pen_style;
	public int fill_pattern;
	public float style_val;
	public int join_style;
	public int cap_style;
	public int radius;
	public int forward_arrow;
	public int backward_arrow;
	public int num_points;
	public int [][] points;
	//public boolean newfile; 728
	public long centroid_x;
	public long centroid_y;
	
	//for text
	public int font;
	public double font_size;
	public double angle;
	public int font_flags;
	public int height;
	public int length;
	public long pos_x;
	public long pos_y;
	public String s;
	//end for text
	
	//public PrintStream MyOutput;
	// static int tmp=0;
	//String filename=".fig";
      
public XfigObj (int obj_type) //728   , boolean newfile)
{
	//this.newfile = newfile;   728
	this.obj_type = obj_type;
	sub_type = 1;
	line_style = 0;
	thickness = 1;
	line_color = 0;
	text_color = 0;
	fill_color = 7;
	depth = 50;
	pen_style = -1;
	fill_pattern = -1;
	style_val = 0;
	join_style = 0;
	cap_style = 0;
	radius = -1;
	forward_arrow = 0;
	backward_arrow = 0;
	font= 0;
	font_size = 12.00;
	angle = 0.000;
	font_flags = 4;
	height = 135;
	length = 90;
}

public XfigObj ()
{
	sub_type = 1;
	line_style = 0;
	thickness = 1;
	line_color = 0;
	text_color = 0;
	fill_color = 7;
	depth = 50;
	pen_style = -1;
	fill_pattern = -1;
	style_val = 0;
	join_style = 0;
	cap_style = 0;
	radius = -1;
	forward_arrow = 0;
	backward_arrow = 0;
	font = 0;
	font_size = 12.00;
	angle = 0.000;
	font_flags = 4; 
	height = 135;
	length = 90;
}

/*public String outFigFile_nonData() //not going to use this anymore !!!!
{	
	String s = "#FIG 3.2\nLandscape\nCenter\nMetric\nLetter\n100.00\nSingle\n-2\n1200 2\n";
	return s;
}*/

public String outFigFile_Objs ()
{
	DecimalFormat formatter = new DecimalFormat("#0.000");
	String  style_val_temp= formatter.format(style_val);
	
	String s = obj_type +" "+sub_type +" "+line_style +" "+thickness +" "+ line_color +" "+fill_color +" "+depth +" "+pen_style +" "+fill_pattern +" "+style_val_temp +" "+join_style +" "+cap_style +" "+radius +" "+forward_arrow+" "+backward_arrow+" "+num_points+"\n\t ";//"\n         ";
	
	for (int i = 0; i < num_points; i++)
	{
		if(i != num_points - 1)
			s = s + points[i][0]+" "+points[i][1]+" ";
		else
			s = s + points[i][0]+" "+points[i][1];
	}
	//s = s + "\n";
	return s;
}

public String outFigFile_Text ()
{
	/*String str = "\n6 "+(pos_x-180)+" "+(pos_y-360)+" "+(pos_x+length+180)+" "+(pos_y+180)+" \n";
	str = str + (pos_x-180)+" "+(pos_y-360)+" "+(pos_x+length+180)+" "+(pos_y-360)+" "+(pos_x+length+180)+" "+(pos_y+180)+" "+
	(pos_x-180)+" "+(pos_y+180)+" "+(pos_x-180)+" "+(pos_y-360)+" \n";*/
	
	String str = "6 "+(int)(pos_x-270*Math.sin(angle))+" "+(int)(pos_y-270*Math.cos(angle))+" "
	+(int)(pos_x+length*Math.cos(angle))+" "+(int)(pos_y-length*Math.sin(angle))+" \n";
	
	str = str + "2 2 0 1 0 7 "+depth+" -1 20 0.000 0 0 -1 0 0 5\n\t"; 
	//we set the default depth for text+its box = 49
	//so it will shown on top of other shapes (with depth = 50

	//white box outside the text
	str = str +
	(int)(pos_x-270*Math.sin(angle)) + " " + (int)(pos_y-270*Math.cos(angle))+" "+
	(int)(pos_x-270*Math.sin(angle)+length*Math.cos(angle))+" "+(int)(pos_y-270*Math.cos(angle)-length*Math.sin(angle))+ " "
	+(int)(pos_x+length*Math.cos(angle))+" "+(int)(pos_y-length*Math.sin(angle))+" "
	+(pos_x)+" "+(pos_y)+" "
	+(int)(pos_x-270*Math.sin(angle))+" "+(int)(pos_y-270*Math.cos(angle))+" \n";
	
	str = str + obj_type +" "+sub_type +" "+ text_color +" "+depth+" "+pen_style +" "+font +" "+font_size +" "+angle +" "+font_flags+" "+height+" "+length+" "+pos_x+" "+pos_y+" "+s+"\\001"+"\n"+"-6";
	
	return str;
}

public  int findLength(double font_size)
{
	int lg = 0; 
	if(font_size<=6)                   lg=30;
	if(font_size>=7 && font_size<=9)   lg=45;
	if(font_size==10 || font_size==11) lg=60;
	if(font_size==12 || font_size==13) lg=90;
	
	if(font_size>=14 && font_size<=16) lg=105;
	if(font_size>=17 && font_size<=19) lg=120;
	if(font_size==20 || font_size==21) lg=135;
	if(font_size==22 || font_size==23) lg=150;
	if(font_size==24 || font_size==25) lg=165;
	if(font_size==26 || font_size==27) lg=180;
	if(font_size>=28 && font_size<=30) lg=195;
	if(font_size==31 || font_size==32) lg=210;
	if(font_size==33 || font_size==34) lg=225;
	if(font_size==35 || font_size==36) lg=240;
	if(font_size>=37 && font_size<=39) lg=255;
	if(font_size >39 && font_size <=500)
	{
		int i=4; int y=0;
		
		if(((int)font_size-31)%9==0) y=0;
		
		if(((int)font_size-31)%9==1) y=0;
		
		if(((int)font_size-31)%9==2) y=1;
		
		if(((int)font_size-31)%9==3) y=1;
		
		if(((int)font_size-31)%9==4) y=2;
		
		if(((int)font_size-31)%9==5) y=2;
		
		if(((int)font_size-31)%9==6) y=3; 
		
		if(((int)font_size-31)%9==7) y=3;
		
		if(((int)font_size-31)%9==8) y=3;
		
		i=i*(((int)font_size-31)/9)+y;
		lg= i*15+210;
	}
	return lg;
}

}