import java.io.*;
import java.util.*;
import java.util.regex.*;
/**
 * Class xMLParser: for parsing xML files, i.e. XML & HTML.
 * Example: Please refer to the tester
 */
class xMLParser implements TypeConstants, Constants
{	private static String xstr;
	private static int currpos; // pointer of current position
	private static int prev; // preceding text characters, tags not included
	
	/**
	* Class tester
	* @param	av	To specify input text file
	*/
	public static void main(String av[]) // tester
	{	char buf[] = new char[BUF_SIZE];
		int cnt;
   		StringBuffer readin = new StringBuffer();
   		String xstr;
   		BufferedReader in;
   		
   		try
   		{	in = new BufferedReader(new FileReader(av[0]));
 			while ((cnt = in.read(buf, 0, BUF_SIZE)) != -1)
				readin.append(String.copyValueOf(buf, 0, cnt));
			in.close();
			xstr = readin.toString();
			xMLParser parser = new xMLParser(xstr);
			SimpleNode node = parser.parse(null);
			Utility.dumpxML(node, "");
		}
		catch(IOException e){ System.out.println(e.getMessage());}
	}
	
	/**
	* Constructor
	* @param	xstr	String to be parsed
	*/
	public xMLParser(String xstr)
	{	this.xstr = xstr;
		this.currpos = 0;
		this.prev = 0;
	}

	/**
	* Parse process
	* @param	parent	Parent node, for back-off use
	*/
	public SimpleNode parse(SimpleNode parent)
	{	Pattern tagpttn, endtagpttn;
		Matcher matcher, tagmatcher, endtagmatcher;
		String tagstr, tagname;
		int startpos, blankpos;
		
		startpos = currpos;
		tagpttn = Pattern.compile("<.*?>"); // to match any tag
		endtagpttn = Pattern.compile("<\\s*/.*?>"); // to match only end tag
		
		// get the tag string from the first tag
		matcher = tagpttn.matcher(xstr.substring(currpos));
		matcher.find();
		tagstr = matcher.group().substring(1, matcher.end()-1).trim();
		// get the tag name
		blankpos = tagstr.indexOf(" ");
		if(blankpos >= 0)
		{	tagname = tagstr.substring(0, blankpos); }
		else
		{	tagname = tagstr; }
		SimpleNode node = new SimpleNode(tagname);
		node.addAttr(RELATIVE_POS, new Integer(prev));
		currpos += matcher.end();
		
		// parse xML attributes into the tree
		if(blankpos >= 0)
		{	String attrstr = tagstr.substring(blankpos).trim().replaceAll("\\s*=\\s*", "="); // remove blanks around "="
			String attrlist[] = attrstr.split("\\s+"); // split into attribute pairs
			for(int i=0;i<attrlist.length;i++)
			{	int espos = attrlist[i].indexOf("="); // equal sign position
				node.addAttr(attrlist[i].substring(0, espos), attrlist[i].substring(espos + 1));
			}
		}		
		
		// generate children
		while(true)
		{	tagmatcher = tagpttn.matcher(xstr.substring(currpos));
			tagmatcher.find();
			endtagmatcher = endtagpttn.matcher(xstr.substring(currpos));
			endtagmatcher.find();
			if(tagmatcher.start() != endtagmatcher.start()) // encounter another open tag
			{	currpos += tagmatcher.start();
				prev += tagmatcher.start();
				SimpleNode childNode = parse(node); // recursive parsing
				if(childNode != null)
					node.jjtAddChild(childNode);
			}
			else // next tag is an end tag
			{	if(!endtagmatcher.group().substring(2,endtagmatcher.group().length() - 1).equals(tagname)) // not its own end tag, curr tag is like <br>
				{	for(int i=0;i<node.jjtGetNumChildren();i++) // transfer children to parent
						parent.jjtAddChild(node.jjtGetChild(i));
					node = null;
				}
				else
				{	currpos += endtagmatcher.end();
					prev += endtagmatcher.start();
				}
				break;
			}
		}
		
		if (node != null)
			node.setInfo(xstr.substring(startpos, currpos));
		
		return node;
	}
}