/*
 * Database interpreter (with semantic check)  
 *
 * @author Biao Hao (bhao@cs.mcgill.ca)  
 *
 * May 31, 1997 ---    
 *  
 * TODO :
 * 1. mutual recursive view - semantics
 * 2. select, project, assignment, mujoin with virtual attributes
 * 3. initialization using another relation
 *    relation R(A,B) <- S; << S exists and defined on A and B
 * 4. update
 * 5. RAM version of evaluateExpression
 *
 * DONE :
 * 1. project with virtual attributes in the project list
 * 2. select with virtual attribute (October 20, 97)
 * 3. assignment with virtual attribute (October 21, 97)
 * 4. mujoin with virtual attribute (October 21, 97)
  *
 * Changes Since jRelCaNu
 *  - Special handling for OP_CAST.
 *    explicitely checked children for cast compatibility.
 *  - Also store the CastFromType in bits of the OP_TYPE node.
 *  - Allow select and project in let expressions.

 */

import java.lang.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
//import sql2relix.*;

public class Interpreter implements Constants, TypeConstants
//public class Interpreter extends Interpreter76
{
//debugging Yi Zheng
private boolean DEBUG=false;
private boolean DEBUG_09_22_2002 = false;

  public Environment env;   // the top level environment
  private CompTable comptable; // Relation system table
  private ExprTable exprtable;	// Expression system table
  private Parser parser;	// the parser

  private Stack instack;      //stack for old input streams
  private InputStream in;     //current parser input stream

  // the prefix and current number for temporary file name
//  private static final String temp_name_prefix = "_temp_X9X_"; //changed by Yi Zheng March 27, 2002
  public static final String temp_name_prefix = "_temp_X9X_";
  private static final int MAX_TEMP_FILE = Integer.MAX_VALUE;
  private static int next_number = 0;

  private static final int MAX_INITIALIZE_SIZE = 10000;

//added by Yi Zheng June 24, 2002
public Relation destrel = null;
public int currow = -1;

  //   private int nested; 	// set the nested level for the relations

private int levels =0; //from zhan : record of the level of recursive nesting
private int level = 0; //from zhan : NE level

private String isKStar = null; //from zhan
private String KStarName = null; //from zhan
private Vector allPath = new Vector();//from zhan

/******* start adding by Fan  *******/
public int relLevel = -1; // Fan Guo June 6,2004
public Vector NRelName;   // Fan Guo June 6,2004
private int rcrLoop = 0; //  Fan Guo August 8, 2004
private SimpleNode recurpNode = null;  // August 31, 2004;
private int childOrder = 0;    // August 31, 2004;
private int printLevel=0; // June 8, 2004 Fan Guo. used for method Mytraverse(node) to print out a tree;
/******* end of Fan's addition *******/

  public Interpreter () {
    this.env = Global.environment;
    this.comptable = Global.comptable;
    this.exprtable = Global.exprtable;
    //	nested = 0;	// no nested at first
  }
  private void assignLiteral(Relation rel, int i, int offset, SimpleNode node)
       throws InterpretError
  {
    Domain[] doms = rel.domains;
    switch(doms[i + offset].type)
      {
      case BOOLEAN : {
	if (node.opcode == OP_BOOLEAN)
	  ((byte[]) rel.data[i+offset])[rel.capacity]
	    = ((Byte) node.info).byteValue();
	else if (node.opcode == OP_DK)
	  ((byte[]) rel.data[i+offset])[rel.capacity] = BOOL_DK;
	else if (node.opcode == OP_DC)
	  ((byte[]) rel.data[i+offset])[rel.capacity] = BOOL_DC;
	else
	  throw new InterpretError("invalid boolean literal");
	break;
      }
      case SHORT : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT
	    || node.opcode == OP_DOUBLE)
	  ((short[]) rel.data[i+offset])[rel.capacity]
	    = ((Number) node.info).shortValue();
	else if (node.opcode == OP_DK)
	  ((short[]) rel.data[i+offset])[rel.capacity] = SHORT_DK;
	else if (node.opcode == OP_DC)
	  ((short[]) rel.data[i+offset])[rel.capacity] = SHORT_DC;
	else
	  throw new InterpretError("invalid short literal");
	break;
      }
      case INTEGER : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT
	    || node.opcode == OP_DOUBLE)
	  ((int[]) rel.data[i+offset])[rel.capacity]
	    = ((Number) node.info).intValue();
	else if (node.opcode == OP_DK)
	((int[]) rel.data[i+offset])[rel.capacity] = INT_DK;
	else if (node.opcode == OP_DC)
	  ((int[]) rel.data[i+offset])[rel.capacity] = INT_DC;
	else
	  throw new InterpretError("invalid integer literal");
	break;
      }
      case LONG : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT
	    || node.opcode == OP_DOUBLE)
	  ((long[]) rel.data[i+offset])[rel.capacity]
	    = ((Number) node.info).longValue();
	else if (node.opcode == OP_DK)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DK;
	else if (node.opcode == OP_DC)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DC;
	else
	  throw new InterpretError("invalid long literal");
	break;
      }

      	case NUMERIC:{//case added by Yiyi
	  /*  if(node.opcode == OP_NUMERIC ||node.opcode == OP_STRING){
       		String s = (String) node.info;
		((number[])rel.data[i+offset])[rel.capacity] =
		    new number(s);
	    }*/
       if (node.opcode == OP_INTEGER)      //added by Yu, July 15,2004
       {
            ((number[])rel.data[i+offset])[rel.capacity] = new number((node.info).toString()); //won't work, suspect class number
       }
       else if(node.opcode == OP_NUMERIC)
       	((number[])rel.data[i+offset])[rel.capacity] = (number)node.info;
       else if(node.opcode == OP_STRING){
       	String s = (String) node.info;
			((number[])rel.data[i+offset])[rel.capacity] =
		    new number(s);
       }
	    else if(node.opcode == OP_DK)
		((number[])rel.data[i+offset])[rel.capacity]
		    = NUMERIC_DK;
	    else if(node.opcode == OP_DC)
		((number[])rel.data[i+offset])[rel.capacity]
		    = NUMERIC_DC;
	    else
		throw new InterpretError("invalid numeric literal");
	    break;
	}
      	 case UNIVERSAL:  // this line of code waas added by Fan to support universal type initialization on September 17, 2004
      	 case UNION:{//added by Yu Aug 04,2004
      	  if (node.opcode == OP_INTEGER)
      	  {
      	  	((String[]) rel.data[i+offset])[rel.capacity] = 
      		((Number) node.info).toString();
      	  }
      	else if (node.opcode == OP_STRING)
      {	  
      	  ((String[]) rel.data[i+offset])[rel.capacity] = (String) node.info;
      }	    //from Andrey
         else if ((node.opcode == OP_ATTRIBUTE)||(node.opcode == OP_TEXT))
      	  ((String[]) rel.data[i+offset])[rel.capacity]
      		= (String) node.info;
      	    //end from Andrey

      	else if (node.opcode == OP_DK)
      	  ((String[]) rel.data[i+offset])[rel.capacity] = STRING_DK;
      	else if (node.opcode == OP_DC)
      	  ((String[]) rel.data[i+offset])[rel.capacity] = STRING_DC;
      	else if (node.opcode == OP_NULL)
      	  ((String[]) rel.data[i+offset])[rel.capacity] = "";
      	else
      	  throw new InterpretError("invalid String literal");
      	break;
            }
      	
      //case UNIVERSAL ://from Andrey  // commetned out by Fan on September 17, 2004. 
      case ATTRIBUTE://from Andrey
      case TYPE : //from Yu  June 2004
      case STRING :
      case TEXT :
      case STMT :
      case EXPR : {
	/*if( rel==null)
	  System.out.println( " NUll rel" );
	  else if ( rel.data[i+offset] == null )
	  System.out.println( " Null data" );
	  else if ( node.info == null )
	  System.out.println( " Null node info" );
	  else if ( rel.capacity == 0 )
	  System.out.println( " Null rel capacity" );
	  System.out.println( "rel.data size" +
	  (rel.data).length );
	  System.out.println( "node.info is " +
	  (node.info).toString() );*/


	if (node.opcode == OP_STRING)
	  ((String[]) rel.data[i+offset])[rel.capacity]
	    = (String) node.info;
	    //from Andrey
   else if ((node.opcode == OP_ATTRIBUTE)||(node.opcode == OP_TEXT))
	  ((String[]) rel.data[i+offset])[rel.capacity]
		= (String) node.info;
	    //end from Andrey
//	else if (node.opcode == OP_NUMERIC) //from Yiyi
//	  ((number[]) rel.data[i+offset])[rel.capacity]
//	    = (number) node.info;

	else if (node.opcode == OP_DK)
	  ((String[]) rel.data[i+offset])[rel.capacity] = STRING_DK;
	else if (node.opcode == OP_DC)
	  ((String[]) rel.data[i+offset])[rel.capacity] = STRING_DC;
	else if (node.opcode == OP_NULL)
	  ((String[]) rel.data[i+offset])[rel.capacity] = "";
	else
	  throw new InterpretError("invalid String literal");
	break;
      }
      case FLOAT : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT
	    || node.opcode == OP_DOUBLE)
	  ((float[]) rel.data[i+offset])[rel.capacity]
	    = ((Number) node.info).floatValue();
	else if (node.opcode == OP_DK)
	  ((float[]) rel.data[i+offset])[rel.capacity] = FLOAT_DK;
	else if (node.opcode == OP_DC)
	  ((float[]) rel.data[i+offset])[rel.capacity] = FLOAT_DC;
	else
	  throw new InterpretError("invalid float literal");
	break;
      }
      case DOUBLE : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT
	    || node.opcode == OP_DOUBLE)
	  ((double[]) rel.data[i+offset])[rel.capacity]
	    = ((Number) node.info).doubleValue();
	else if (node.opcode == OP_DK)
	  ((double[]) rel.data[i+offset])[rel.capacity] = DOUBLE_DK;
	else if (node.opcode == OP_DC)
	((double[]) rel.data[i+offset])[rel.capacity] = DOUBLE_DC;
	else
	  throw new InterpretError("invalid double literal");
	break;
      }
      case IDLIST : { // dk and dc for nested relation
	if (node.opcode == OP_DK)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DK;
	else if (node.opcode == OP_DC)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DC;
	else
	throw new InterpretError("invalid surrogate literal");
	break;
      }
      } // end of switch
  }

// by Yu
  private Vector CreateDAGUnionRecursive(String relname, Environment env,String recurDom,int recurLevel)
       throws InterpretError
  {
   boolean recu = false;
   Vector vec = new Vector(64); //stack of names to process

	//System.out.println("top level relation name:"+relname);
   vec.addElement(relname); // add top level relation name

   int size = 0;
   while (size < vec.size()) 
   {
    String s = (String) vec.elementAt(size);
    //System.out.println("relation name in CreatDAG:"+s);
    Domain[] doms = env.lookupRel(s, true).domains;
    for(int i = 0; i < doms.length; i++) 
    {
     //System.out.println("domain name in CreatDAG:"+doms[i].name);
      	//if (doms[i].name.equals("c1"))	//union type
      	//	vec.addElement("." + "n1");
     if (doms[i].type == UNION)
	 {
     //System.out.println("domain type in CreatDAG is union:"+doms[i].name);
       Vector idlistnames = new Vector(); //stack of domain names to process
	   idlistnames.addElement(doms[i].name);
	   while (idlistnames.size()>0)//hasMoreUnion
	   {	
        String domnm = (String) idlistnames.lastElement();
        idlistnames.removeElementAt(idlistnames.size() - 1);
 	    Domain domC = env.lookupDom(domnm, true);
	    //System.out.println("union type in vector:"+domC.name);
	    for (int j = 0; j<domC.union.length; j++) //check for each possible type if it is IDLIST
	    {//System.out.println("the "+j+" "+"union type in vector:"+domC.union[j].name);
//		 if (doms[i].union[j].opcode==OP_IDLIST||doms[i].union[j].opcode==OP_IDENTIFIER)
		 if (domC.union[j].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;
		 {
		  String nameD = domC.union[j].name;
		  if (env.lookupDom(nameD,true)!=null)
		  {
			if ((env.lookupDom(nameD,true)).type==IDLIST)
			{
	   		 //System.out.println("here I am:union types with idlist:"+domC.union[j].name);
	//System.out.println("recursive domain:"+recurDom+" and level is :"+ recurLevel);
			 if (recurDom.equals(nameD))
			 {  if (!recu)
			 	{
			 	 recu = true;	
			 	 for (int k = 0; k<recurLevel; k++)
			 	 {
			  	  vec.addElement("." + nameD+"_"+k);
			 	 }
			 	}
			 }
			 else
			  vec.addElement("." + domC.union[j].name);
			  //hasMoreUnion=false;//--no!!! idlist can have union domain! deal with it at the following
			}
			else if ((env.lookupDom(nameD,true)).type==UNION)
				idlistnames.addElement(nameD);
		  }
		 }
	    }	//for
	   } //while (domnames.size()>0)(hasMoreUnion)
	 }
	 else if(doms[i].type == IDLIST && !doms[i].name.equals(".id"))
	 {
	 	 vec.addElement("." + doms[i].name);
	 }
    }//for
    size++;
    if (size > 1024) // has cycle or nested too deep
		throw new InterpretError("nested relation " + relname
				 + " has cycle or nested too deep");
   }

    Hashtable ht = new Hashtable(64);

    for (int i = size - 1; i >= 0; i--) {
      String name = (String) vec.elementAt(i);
      if (ht.containsKey(name))
	vec.removeElementAt(i);
      else
	ht.put(name, name);
    }

    return vec;
  }

// by Yu
  private Vector CreateDAGUnion(String relname, Environment env)
       throws InterpretError
  {
   Vector vec = new Vector(64); //stack of names to process

	//System.out.println("top level relation name:"+relname);
   vec.addElement(relname); // add top level relation name

   int size = 0;
   while (size < vec.size()) 
   {
    String s = (String) vec.elementAt(size);
    //System.out.println("relation name in CreatDAG:"+s);
    Domain[] doms = env.lookupRel(s, true).domains;
    for(int i = 0; i < doms.length; i++) 
    {
     //System.out.println("domain name in CreatDAG:"+doms[i].name);
      	//if (doms[i].name.equals("c1"))	//union type
      	//	vec.addElement("." + "n1");
     if (doms[i].type == UNION)
	 {
     //System.out.println("domain type in CreatDAG is union:"+doms[i].name);
       Vector idlistnames = new Vector(); //stack of domain names to process
	   idlistnames.addElement(doms[i].name);
	   while (idlistnames.size()>0)//hasMoreUnion
	   {	
        String domnm = (String) idlistnames.lastElement();
        idlistnames.removeElementAt(idlistnames.size() - 1);
 	    Domain domC = env.lookupDom(domnm, true);
	    //System.out.println("union type in vector:"+domC.name);
	    for (int j = 0; j<domC.union.length; j++) //check for each possible type if it is IDLIST
	    {//System.out.println("the "+j+" "+"union type in vector:"+domC.union[j].name);
//		 if (doms[i].union[j].opcode==OP_IDLIST||doms[i].union[j].opcode==OP_IDENTIFIER)
		 if (domC.union[j].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;
		 {
		  String nameD = domC.union[j].name;
		  if (env.lookupDom(nameD,true)!=null)
		  {
			if ((env.lookupDom(nameD,true)).type==IDLIST)
			{
	   		 //System.out.println("here I am:union types with idlist:"+domC.union[j].name);
			 vec.addElement("." + domC.union[j].name);
			  //hasMoreUnion=false;//--no!!! idlist can have union domain! deal with it at the following
			}
			else if ((env.lookupDom(nameD,true)).type==UNION)
				idlistnames.addElement(nameD);
		  }
		 }
	    }	//for
	   } //while (domnames.size()>0)(hasMoreUnion)
	 }
	 else if(doms[i].type == IDLIST && !doms[i].name.equals(".id"))
	 {
	 	 vec.addElement("." + doms[i].name);
	 }
    }//for
    size++;
    if (size > 1024) // has cycle or nested too deep
		throw new InterpretError("nested relation " + relname
				 + " has cycle or nested too deep");
   }

    Hashtable ht = new Hashtable(64);

    for (int i = size - 1; i >= 0; i--) {
      String name = (String) vec.elementAt(i);
      if (ht.containsKey(name))
	vec.removeElementAt(i);
      else
	ht.put(name, name);
    }

    return vec;
  }
  private Vector CreateDAG(String relname, Environment env)
       throws InterpretError
  {
    Vector vec = new Vector(64); //stack of names to process

    vec.addElement(relname); // add top level relation name

    int size = 0;
    while (size < vec.size()) {
      String s = (String) vec.elementAt(size);
      Domain[] doms = env.lookupRel(s, true).domains;
      for(int i = 0; i < doms.length; i++) {
	if(doms[i].type == IDLIST && !doms[i].name.equals(".id"))
	  vec.addElement("." + doms[i].name);
      }
      size++;
      if (size > 1024) // has cycle or nested too deep
	throw new InterpretError("nested relation " + relname
				 + " has cycle or nested too deep");
    }

    Hashtable ht = new Hashtable(64);

    for (int i = size - 1; i >= 0; i--) {
      String name = (String) vec.elementAt(i);
      if (ht.containsKey(name))
	vec.removeElementAt(i);
      else
	ht.put(name, name);
    }

    return vec;
  }
  // create data array for a relation and put the relaton into
  // the hashtable
  private void CreateData(Hashtable ht, String name, Environment env)
       throws InterpretError
  {
    Relation rel = env.lookupRel(name, true);
    rel.createDataArray(MAX_INITIALIZE_SIZE);
    rel.capacity = 0;
    ht.put(name, rel);
  }
  /**
   * Create all relations and their data arrays that need to be updated
   * (including nested relations) for initializing relation 'name'. Each
   * of these relations is stored in the resulting hash table with the
   * name of the relation as the key.
   */
  private Hashtable CreateDataHashTable(String name, Environment env)
       throws InterpretError
  {
    Hashtable rels = new Hashtable();
    Vector relnames = new Vector(); //stack of names to process
    int size;

    // initialize the stack of names
    Domain[] doms = env.lookupRel(name, true).domains;
    for(int i = 0; i < doms.length; i++) {
      if(doms[i].type == IDLIST)
		relnames.addElement("." + doms[i].name);
	  else if(doms[i].type == UNION)		// Added on July 08,2004 by Yu
	  {
	   //boolean hasMoreUnion=true;
       Vector domnames = new Vector(); //stack of domain names to process
	   domnames.addElement(doms[i].name);
	   while (domnames.size()>0)//hasMoreUnion
	   {	
        String dnm = (String) domnames.lastElement();
        domnames.removeElementAt(domnames.size() - 1);
 	    Domain dc = env.lookupDom(dnm, true);
 	    if (dc==null) System.out.println("domain is null,"+dnm);
	    //System.out.println("union type:"+dc.name);
	    for (int j = 0; j<dc.union.length; j++) //check for each possible type if it is IDLIST
	    {
//		 if (doms[i].union[j].opcode==OP_IDLIST||doms[i].union[j].opcode==OP_IDENTIFIER)
		 if (dc.union[j].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;
		 {
		 	String nameD = dc.union[j].name;
		 	//System.out.println("union:"+nameD);
		 	if (env.lookupDom(nameD,true)==null) 
		 		  throw new InterpretError("Domain "+nameD+" has not been declared");
			if ((env.lookupDom(nameD,true)).type==IDLIST)
			{
	   		 //System.out.println("union types with idlist:"+dc.union[j].name);
			 relnames.addElement("." + dc.union[j].name);
			 //hasMoreUnion=false;//--no!!! idlist can have union domain! deal with it at the following
			}
			else if ((env.lookupDom(nameD,true)).type==UNION)
			{//System.out.println(nameD+" is union type of union domain :"+dc.name);
				domnames.addElement(nameD);
			}
		 }
	    }	//for
	   } //while (domnames.size()>0)(hasMoreUnion)
	  }//else if(doms[i].type == UNION)	end of Added on July 08,2004 by Yu
    }

    // create top level relation and its data array, store in hashtable
    CreateData(rels, name, env);

    while (0 < (size = relnames.size())) {	//all nested domain -- comment by Yu
      String nm = (String) relnames.lastElement();
      relnames.removeElementAt(size - 1);
      if (!rels.containsKey(nm)) {	//System.out.println("CreateDataHashTable for nested relation:"+nm);
	CreateData(rels, nm, env);
	doms = env.lookupRel(nm, true).domains;
	for (int i = 0; i < doms.length; i++) 
	{
	 String s = doms[i].name;
	 if (!rels.containsKey(s)) {
	  if(doms[i].type == IDLIST && !s.equals(".id"))//.id of type IDLIST
	  {//System.out.println("domain:"+doms[i].name);
	    	  relnames.addElement("." + doms[i].name);
	  }
	   //added by Yu, July 24,2004		
	  else if (doms[i].type == UNION)
	  {
           Vector idlistnames = new Vector(); //stack of domain names to process
	   idlistnames.addElement(doms[i].name);
	   while (idlistnames.size()>0)//hasMoreUnion
	   {	
            String domnm = (String) idlistnames.lastElement();
            idlistnames.removeElementAt(idlistnames.size() - 1);
 	    Domain domC = env.lookupDom(domnm, true);
//	    System.out.println("union type:"+doms[i].name);

	    for (int j = 0; j<domC.union.length; j++) //check for each possible type if it is IDLIST
	    {
//		 if (doms[i].union[j].opcode==OP_IDLIST||doms[i].union[j].opcode==OP_IDENTIFIER)
		 if (doms[i].union[j].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;
		 {
		  String nameD = doms[i].union[j].name;
		  //System.out.println("right here, domain is "+doms[i].name+" and union type is :"+nameD);
		  Domain d = env.lookupDom(nameD,true);
		  if (d!=null)
		  {
			if ((env.lookupDom(nameD,true)).type==IDLIST)
			{
	   		 //System.out.println("union types with idlist:"+doms[i].union[j].name);
			 relnames.addElement("." + doms[i].union[j].name);
			 //hasMoreUnion=false;//--no!!! idlist can have union domain! deal with it at the following
			}
			else if ((env.lookupDom(nameD,true)).type==UNION)
				idlistnames.addElement(nameD);
		  }//if (d!=null)
		 }
	    }	//for
	   } //while (domnames.size()>0)(hasMoreUnion)
	  }//end of else if (doms[i].type == UNION)
	 }
	   //end of added by Yu, July 24,2004		
		
		
//---------------		commented off by Yu
/*	  String s = doms[i].name;
	  if (!rels.containsKey(s)) {
	    if(doms[i].type == IDLIST && !s.equals(".id"))//.id of type IDLIST
	    {//System.out.println("domain:"+doms[i].name);
	    	  relnames.addElement("." + doms[i].name);
	    }
	  }*///----------------
	}//end of 	for (int i = 0; i < doms.length; i++) 
      }
    }
    return rels;
  }
  // delete a data file on disk
  public static void deleteFile(String filename)
  {
    File f = new File(filename);
    if (f.exists()) f.delete();
  }
  // REVISIT: short cut for equality of two relations
  private boolean equals(Relation rl, Relation rr)
       throws InterpretError
  {
    String tmpname = nextTempName();

    // check if the two relations are defined on same set of attributes
    if (!rl.isDomainsIn(rr.domains) || !rr.isDomainsIn(rl.domains))
      throw new InterpretError("two relations are not comparable");
    if (rl.numtuples != rr.numtuples)
      return false;
    Relation rel = rl.mujoin(OP_SJOIN, null, rr, null, tmpname);
    if (rel.numtuples == 0) {
      deleteFile(tmpname);
      return true;
    } else {
      deleteFile(tmpname);
      return false;
    }
  }

     /**
   * Processes the node for evaluating the join specified in the node's tree.
   *
   * JRelixNS modified/enhanced this method:
   *  Now handles Sigma-Joins in addition to mu-joins.
   */
  private Relation evaluateJoin(SimpleNode node, String filename,
				Hashtable htable, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    Domain[] domsl = null, domsr = null;
    Relation rl = null, rr = null;
    String lname = nextTempName();
    String rname = nextTempName();

    // step 1. get the join attributes and evaluate join expressions
    if (node.jjtGetNumChildren() == 2) { // no join attributes specified
      
	  /*
      // evaluate left relational expressions
      rl = evaluateTLExpression((SimpleNode) node.jjtGetChild(0),
				lname, htable, env);
      // evaluate right relational expressions
      rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1),
				rname, htable, env);
	  */// original code commented out by Fan, use the following code
	  
	  /******* Start modifying by Fan  *******/
      // evaluate left relational expressions
      rl = evaluateTLExpression((SimpleNode) node.jjtGetChild(0),
				lname, htable, env);
      rl = IntermediateLevelLift(rl);// Fan Guo July 18, 2004	
      // evaluate right relational expressions
      rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1),
				rname, htable, env);
      rr = IntermediateLevelLift(rr);  // Fan Guo July 18, 2004
      
      if(rl == null && rr != null)  // Fan August 18, 2004
      	return rr;
      else if (rl != null && rr == null)
      	return rl;      
      else if(rl == null && rr == null)
      	return null;
      
      /****** end of Fan's modification   *******/
       	
    } else { // node.jjtGetNumChildren() == 4
      // get an array of domains for left ExpressionList
      SimpleNode nodel = (SimpleNode) node.jjtGetChild(1);
      domsl = ExpressionListToDomains(nodel);

      // get an array of domains for right ExpressionList
      SimpleNode noder = (SimpleNode) node.jjtGetChild(2);
      domsr = ExpressionListToDomains(noder);
	  /*
	  // evaluate left relational expressions
      rl = evaluateTLExpression((SimpleNode) node.jjtGetChild(0),
				lname, htable, env);
      // evaluate right relational expressions
      rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(3),
				rname, htable, env);
	  */// original code commented out by Fan, use the following code
	   
      /******* Start modifying by Fan  *******/
      // evaluate left relational expressions
      rl = evaluateTLExpression((SimpleNode) node.jjtGetChild(0),
				lname, htable, env);
      rl = IntermediateLevelLift(rl);// Fan Guo July 18, 2004	
      // evaluate right relational expressions
      rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(3),
				rname, htable, env);
      rr = IntermediateLevelLift(rr); // Fan Guo July 18, 2004
      
      if(rl == null && rr != null)  // Fan August 18, 2004
      	return rr;
      else if (rl != null && rr == null)
      	return rl;
      else if(rl == null && rr == null)
      	return null;      
      /****** end of Fan's modification   *******/
      
      // actualize if necessary
      if (rl.rvc == RELATION && !rl.isDomainsIn(domsl)) {
	Relation rltmp = new Actualizer(domsl, rl, env).actualize();
	rltmp.name = lname;
	rltmp.save();
	rltmp.unload();
	rl = rltmp;
      }
      if (rr.rvc == RELATION && !rr.isDomainsIn(domsr)) {
	Relation rrtmp = new Actualizer(domsr, rr, env).actualize();
	rrtmp.name = rname;
	rrtmp.save();
	rrtmp.unload();
	rr = rrtmp;
      }
    }

    // step 2. invoke the join method
    if(rl.rvc==COMPUTATION && rr.rvc==COMPUTATION ){
      throw new
	InterpretError("join of two computations is not implemented.");
    }
    else if(rl.rvc == COMPUTATION){
    	if(opcode == OP_IJOIN) { //if switch added by Yi Zheng July 4, 2002
      Relation rel =
	((Computation) rl).applyIjoin(rr, domsl, domsr, filename, env);
      deleteFile(lname);
      deleteFile(rname);
      return rel;
     }
     else
     	throw new InterpretError("join other than IJOIN between computation and relation not implemented");
    }else if(rr.rvc == COMPUTATION){
    	if (opcode == OP_IJOIN) {
      Relation rel =
	((Computation) rr).applyIjoin(rl, domsr, domsl, filename, env);
      deleteFile(lname);
      deleteFile(rname);
      return rel;
      }
      else
     	throw new InterpretError("join other than IJOIN between computation and relation not implemented");
    }
    else
      {  //this is a regular join of two relations/views
	switch(opcode)
	  {  case OP_IJOIN :
	  case OP_UJOIN :
	  case OP_SJOIN :
	  case OP_LJOIN :
	  case OP_RJOIN :
	  case OP_DLJOIN : // also OP_DJOIN
	  case OP_DRJOIN :
	    Relation rel = rl.mujoin(node.opcode, domsl, rr, domsr, filename);
	  deleteFile(lname);
	  deleteFile(rname);
	  return rel;
	  case OP_ICOMP :
	  case OP_EQJOIN :
	  case OP_GTJOIN :
	  case OP_GEJOIN :
	  case OP_LTJOIN :
	  case OP_LEJOIN :
	  case OP_IEJOIN :
	  case OP_NOTICOMP :
	  case OP_NOTEQJOIN :
	  case OP_NOTGTJOIN :
	  case OP_NOTGEJOIN :
	  case OP_NOTLTJOIN :
	  case OP_NOTLEJOIN :
	  case OP_NOTIEJOIN :
	    // START Since jRelixNS
	    // Process a sigma join only if it the optional flag
	    // Relation.isSigmaJoinImplemented is set.
	    // Otherwise, issue a "TO BE IMPLEMENTED" message.
	  //  if (env.isSigmaJoinImplemented == true) {
	     if (Relation.isSigmaJoinImplemented == true) { //Andy, May 12, 2002
	      rel = rl.sigmaJoin(node.opcode, domsl, rr, domsr, filename);
	      deleteFile(lname);
	      deleteFile(rname);
	      return rel;
	    }
	    else
	      throw new InterpretError("TO BE IMPLEMENTED : sigma-join");
	    // END Since jRelixNS
	  case OP_NOP : {//case added by Yi Zheng July 4, 2002

	  	Relation noprel = null;
	  	if(node.jjtGetNumChildren() == 2)
	  	{
	  		boolean Ran = new Random().nextBoolean();
	  		if (!Ran)
	  		{
	  			noprel = rl.project(rl.domains, filename);
	  			deleteFile(lname);
	  			deleteFile(rname);
	  			return noprel;
	  		}
	  		else
	  		{
	  			noprel = rr.project(rr.domains, filename);
	  			deleteFile(lname);
	  			deleteFile(rname);
	  			return noprel;
	  		}
	  	}
	  	else
	  		throw new InterpretError("invalid relational expression");
	  }
	  default :
	    throw new
	      InterpretError("invalid/not implemented relational expression");
	  } /* switch */
     }
  }


  //YuLing
  private Relation evaluateGedit(SimpleNode node,String filename, Hashtable
				 htable, boolean flag, Environment env, Domain[] doms)
       throws InterpretError
  {
    int opcode=node.opcode;
    //	Domain[] doms = null;
    String tmpname = nextTempName();
    boolean projector2=false;
    SimpleNode noder=null;
    SimpleNode nodel1=null;
    SimpleNode nodel2=null;

    //get the first  porjector
    nodel1=(SimpleNode)node.jjtGetChild(0);

    //get the number of children
    if (node.jjtGetNumChildren() == 2)  // no second projector
      {
	//append Color, symbol, and layer to rel.
	noder=(SimpleNode)node.jjtGetChild(1);
      }
    else
      {
	nodel2=(SimpleNode)node.jjtGetChild(1);
	projector2=true;

	noder=(SimpleNode)node.jjtGetChild(2);
      }
    if (nodel1.isBitSet(0)) { // contains QT-Operator hongyu
      throw new InterpretError("TO BE IMPLEMENTED : QT-Operator");
    }
    // evaluate the relational expression (projection) it is projected on
    Relation rel = evaluateTLExpression(noder, tmpname, htable, env);

    // Jiantao 2004
    if(rel.name.equals(STRING_DC))
                return rel;

    if (nodel1.isBitSet(1)) { // contains ExpressionList in the first projector
      // get node containing ExpressionList in projector
      SimpleNode nd = (SimpleNode) nodel1.jjtGetChild(0);
      if(doms==null)
	doms=new Domain[8];

      doms=ExpressionListToGeditDomains(nd,doms,1);
    } else { // empty projector is not allowed
      throw new
	InterpretError("empty projector is not allowed in gedit");
    }
    if(projector2==true)
      {
	if (nodel2.isBitSet(1)) { // contains ExpressionList in projector
	  // get node containing ExpressionList in projector
	  SimpleNode nd = (SimpleNode) nodel2.jjtGetChild(0);
	  doms = ExpressionListToGeditDomains(nd,doms,2);
	} else { // empty projector not allowed
	  throw new
	    InterpretError("Empty projector is not allowed in gedit");
	}
      }
    Domain[] domains=null;

    if(doms[5]==null)
      {
        executeRelixCommand("dd .C;");
        executeRelixCommand("domain .C intg;");
        executeRelixCommand("relation COLOR(.C)<-{(204055055)};");

        domains=rel.domains;

	String attrs="";
	for(int i=0;i<domains.length;i++)
	  attrs=attrs+domains[i].name+",";
	//System.out.println("attrs in Interpret[doms5]="+attrs);

        executeRelixCommand(noder.name+"<-["+attrs+".C]("+noder.name+" ijoin COLOR);");
	doms[5]=env.lookupDom(".C",true);
	//System.out.println("doms[5].name="+doms[5].name);
	rel = getRelbyName(rel.name);
      }

    if(doms[6]==null)
      {
        executeRelixCommand("dd .Symb;");
        executeRelixCommand("domain .Symb intg;");
        executeRelixCommand("relation SYMBOL(.Symb)<-{(201)};");

        domains=rel.domains;

	String attrs="";
	for(int i=0;i<domains.length;i++)
	  attrs=attrs+domains[i].name+",";
	//System.out.println("attrs in Interpret doms[6]="+attrs);

        executeRelixCommand(noder.name+"<-[" +attrs+".Symb]("+noder.name+" ijoin SYMBOL);");
	doms[6]=env.lookupDom(".Symb",true);
	rel = getRelbyName(rel.name);
      }

    if(doms[7]==null)
      {
        executeRelixCommand("dd .L;");
        executeRelixCommand("domain .L string;");
        executeRelixCommand("relation LAYER(.L)<-{(\"DefautLayer\")};");

	domains=rel.domains;

	String attrs="";
	for(int i=0;i<domains.length;i++)
	  attrs=attrs+domains[i].name+",";
	//System.out.println("attrs in Interpret doms[7]="+attrs);

        executeRelixCommand(noder.name+"<-["+attrs+".L]("+noder.name+" ijoin LAYER);");
	doms[7]=env.lookupDom(".L",true);
	rel=getRelbyName(rel.name);
      }


    if (opcode == OP_WHENFROM) {

    }

    if(flag){

      domains=rel.domains;
      Actualizer actualizer = new Actualizer(domains, rel, env);
      Relation rr = actualizer.actualize();

      actualizer.cleanup();

      deleteFile(tmpname);

      /*Domain[] sortdoms= new Domain[3];
	System.arraycopy(rr.domains,0,sortdoms,0,3);
	rr.sort(sortdoms);*/

      int S=((int[])(rr.data[2]))[0];
      System.out.println("Please wait a few minutes for the Geditor to start...");
      Geditor win=new Geditor(this,rr,doms);
      win.setVisible(true);

      return rr;
    }else{
      new Actualizer(doms, rel, env);
      return null;
    }
  }

  //YuLing's addition ends here

  private Relation evaluateProject(SimpleNode node, String filename,
				   Hashtable htable, boolean flag, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    Domain[] doms = null;
    String tmpname = nextTempName();

	//zhan' addition
	node = sugarInProject(node,env); //zhan
	//end of zhan' addition
	
	/******* start adding by Fan  *******/
	int tp = ((SimpleNode)node.jjtGetChild(1)).type;
	node = WildcardInProject(node, env);  // Fan August 18, 2004
    
	if (node == null ) 
		return null;
    else if(node.opcode!=opcode)  // changed from project to join
    	return evaluateTLExpression(node,filename, htable, env);
    /******* end of Fan's addition *******/
    	
    // get the projector
    SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
    //get the selection
    SimpleNode noder = (SimpleNode) node.jjtGetChild(1);

    //The follwing is added by hongyu. Feb 14, 02
    if(nodel.isBitSet(0)) { // contains QT-Operator
  //    System.out.println("here");
      SimpleNode nodeCopy = (SimpleNode)node.jjtDeepDuplicate();
      SimpleNode nodeCopyNodel = (SimpleNode)nodeCopy.jjtGetChild(0);
      SimpleNode nodeCopyNoder = (SimpleNode)nodeCopy.jjtGetChild(1);

      int num = nodeCopyNodel.jjtGetNumChildren();
      SimpleNode nodeQT=(SimpleNode)nodeCopyNodel.jjtGetChild(0);
 //     System.out.println("num "+num);
 //     System.out.println("node QT");
 //     printNode(nodeQT);
      nodeCopyNodel.jjtRemoveChild(0);
      int tempBits = nodeCopyNodel.bits;
      tempBits --;
      tempBits = tempBits | 0x10;
      if(num==1){
	switch (nodeQT.opcode){
	case OP_NUMBER:{
	  doms = new Domain[1];
	  doms[0] = env.lookupDom(".integer", true);
	  Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	  Object[] datal = r.createDataArray(doms,1);
	  ((int[])datal[0])[0] =1;
	  r.data = datal;
	  return r;
	}
	case OP_DOT:
	case OP_PERCENT:{
	  doms = new Domain[1];
	  doms[0] = env.lookupDom(".double", true);
	  Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	  Object[] datal = r.createDataArray(doms,1);
	  ((double[])datal[0])[0] =1.0;
	  r.data = datal;
	  return r;
	}
	}
      }
      else{
	nodeCopyNodel.bits = tempBits;// clear bits of QT-Operator option

	Relation R = evaluateProject(nodeCopy,tmpname,htable,flag, env);
	int nmuber =R.numtuples;
	if(nodeQT.opcode == OP_NUMBER){
	  deleteFile(tmpname);
	  doms = new Domain[1];
	  doms[0] = env.lookupDom(".integer", true);
	  Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	  Object[] datal = r.createDataArray(doms,1);
	  ((int[])datal[0])[0] =nmuber;
	  r.data = datal;
	  return r;
	}
	else{// case opcode == OP_DOT or opcode == OP_PERCENT
	  if((num=nodeCopyNoder.jjtGetNumChildren())>= 2){
	    nodeCopyNoder =(SimpleNode)nodeCopyNoder.jjtGetChild(num-1);
	  }

	  Relation RR = evaluateTLExpression(nodeCopyNoder,filename, htable, env);// org relation
	  Actualizer actualizer = new Actualizer(R.domains, RR, env);
	  RR = actualizer.actualize();
	  RR.name = filename;
	  RR.save();
	  RR = RR.project(R.domains,tmpname);
	  actualizer.cleanup();

	  int total = RR.numtuples;
	  if(total == 0){
	    throw new
	      InterpretError(" the total number is 0");
	  }
	  else{
	    deleteFile(tmpname);
	    deleteFile(filename);
	    doms = new Domain[1];
	    doms[0] = env.lookupDom(".double", true);
	    Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	    Object[] datal = r.createDataArray(doms,1);
	    ((double[])datal[0])[0] =(double)nmuber/(double)total;
	    r.data = datal;
	    return r;
	  }
	}
      }
    }
    //end of hongyu added. Feb 14,02

    /******* start adding by Fan  *******/

     if(noder.type == OP_IDENTIFIER){// can not get the name of a virtual nested relation; as in case 
       					//let srvir be [srel] in fnameweb; followed by SRVir<-[red ujoin of srvir] in products;
     	addRelName(noder.name);
     	relLevel++;
     }      
     
     if(noder.type == OP_JOINOPERATOR || noder.type == OP_PROJECT) {
     	if (noder.info!=null){ // virtual nested relation name
     		addRelName((String)noder.info);
     		relLevel++;
     	} 
     	 
     	  else if(relLevel == 0) {      	
     		//System.out.println("self can not be actualized on an unnamed relation.");
     		addRelName("null");
     		relLevel++;
     	}
     	
     	/*
     	else if(node.opcode != OP_RED){      	
     		System.out.println("self can not be actualized on an unnamed relation.");
     		addRelName("null");
     		relLevel++;
     	}
     	*/
     }  
     /******* end of Fan's addition *******/


    // evaluate the relational expression (projection) it is projected on
    //evaluateQSelect() is here.Hongyu
    Relation rel = evaluateTLExpression(noder, tmpname, htable, env);


    /******* start adding by Fan  *******/
    if (rel ==null) return null;


     //Jiantao 2004
     if(rel.name.equals(STRING_DC)) return rel;
     
    rel = IntermediateLevelLift(rel);  // August 7, 2004
    /******* end of Fan's addition *******/

//Bug Fix --  commented off by Zongyan Feb, 2005
    //if (rel.data==null) rel.load();		//source relation --- July 05, 2004 by Yu

    if (nodel.isBitSet(1)) { // contains ExpressionList in projector  for example pr [t1,x] in R;--by Yu t1: top level scalar
      // get node containing ExpressionList in projector
      SimpleNode nd = (SimpleNode) nodel.jjtGetChild(0);
  //Modified by Yi Zheng April 30, 2002    doms = ExpressionListToDomains(nd); // REVISIT: generalization
	//doms = ExpressionListToDomains(nd, rel);//modified by Yi Zheng May 26, 2002, to integrate Andrey's code
	doms = ExpressionListToDomains(nd, rel, htable, env);	//destination domains ---comment by Yu
    } else { // empty projector, pass domain ".bool"
      doms = new Domain[1];
      doms[0] = env.lookupDom(".bool", true);
      Relation rtmp = rel.project(doms, filename);
      deleteFile(tmpname);
      return rtmp;
    }


    if (opcode == OP_WHENFROM) {
    	throw new InterpretError("OP_WHENFROM not implemented");
     }

    if(flag){
      Actualizer actualizer = new Actualizer(doms, rel, env);
      Relation rtmp = actualizer.actualize();	//same as source --comment by Yu
      /******* start adding by Fan  *******/      
      if(noder.type == OP_IDENTIFIER){  // August 5, 2004
        delRelName();
        relLevel--;
      }

      if(noder.type == OP_JOINOPERATOR || noder.type == OP_PROJECT) {
        if (noder.info!=null){
                delRelName();
                relLevel--;
        }

         else if(relLevel == 0){
                delRelName();
                        relLevel--;
                }
                /*
        else if(node.opcode != OP_RED){
                addRelName("null");
                relLevel++;
        }
        */
      }
      /******* end of Fan's addition *******/

      rtmp.myEnv = env; //Yi Zheng June 28, 2002
      // Save the result relation to disk file. This is required by
      // other operations e.g. projection, selection etc.
      rtmp.name = tmpname;
      rtmp.save();


      // Do the projection..

 //     Relation rr = rtmp.project(doms, filename); //modified by Yi Zheng June 28, 2002

      //addition by Yi Zheng June 28, 2002
      //deal with red domain used for level lifting
      Relation rr = null;
      String rrTempName = nextTempName();

      Relation rrTemp = rtmp.project(doms, rrTempName);

      if((doms.length == 1) && (doms[0].isRedTempTop == 1))
      {
      	if(doms[0].type == IDLIST)
      	{
      		Relation dotRR = env.lookupRel("." + doms[0].name, true);
      		if(dotRR == null)
      			throw new InterpretError("relation ." + doms[0].name + " not found");

      		Domain[] realDoms = new Domain[dotRR.domains.length - 1];
      		int jj = 0;
      		for(int i = 0; i < dotRR.domains.length; i++)
      		{
      			if(dotRR.domains[i].name.compareTo(".id") == 0)
      				continue;
      			realDoms[jj++] = dotRR.domains[i];
      		}

      		rr = dotRR.project(realDoms, filename);
      		env.removeDom(doms[0].name, true);
      	}
      	else
      		rr = rrTemp.project(doms, filename); //this is only used in Computations: eg. Cnt <- [red + of 1] in customers;
      }
      else
      {
      	rr = rrTemp.project(doms, filename);
      }

      //end of addition June 28, 2002


      // System.out.println("rtmp.data in project ");
      //System.out.println(((int[])rtmp.data[3])[1]);
      // Do the clean-up.  Note, we have to put this procedure
      // *after* all the operations on those relations which are
      // generated by the actualizer. This is too bad as OO scheme
      // seems to be broken here. However, this is a must, at least
      // for now. The reason is that when intermediate nested domains
      // are generated by the actualizer (e.g. in case
      // "let x be S ijoin T ijoin U;", a temporary nested domain '0'
      // is generated.), their data contents are written to corresponding
      // disk files (e.g. '.0'). These intermediate disk files are
      // necessary for the operations (such as the above-written
      // "rtmp.project(doms, filename);" since all relational
      // operations are based on disk files but not memory data. ;-(
      // As actualizer's cleanup() method will delete all the
      // intermediate files, we have to make sure this happens *only*
      // after our major task (projection here) has finished.

      actualizer.cleanup();

      //Relation rtmp = rel.project(doms, filename);
      deleteFile(tmpname);
      deleteFile(rrTempName); //Yi Zheng June 28, 2002

//      System.out.println("rr to be returned is:");
//      rr.print();
      rr.load(true);
      return rr;
    }else{
      new Actualizer(doms, rel, env);
      return null;
    }

  }


  //start of zhan's addition 'SugarInProject' Sept 2003
  //handle syntax sugar in projects
  private SimpleNode sugarInProject(SimpleNode node,Environment env)
          throws InterpretError
  {

   //modify the tree to handle 'street/name in company/address'
   //modify the tree to: company/address/street/name;
    SimpleNode projectLeft = (SimpleNode)node.jjtGetChild(0);
    SimpleNode projectRight = (SimpleNode)node.jjtGetChild(1);

    if(((SimpleNode)node.jjtGetChild(0)).jjtGetNumChildren()==0)
      return node;

    if(projectLeft.identifier.equals("Project"))
    {
	   SimpleNode verticalL = (SimpleNode)projectLeft.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);
       if(verticalL.type == OP_RED)
       {
		boolean isNest=false;
		//get name of top identifier of the left child of the node
        String leftName = ((SimpleNode)projectLeft.jjtGetChild(1)).name;
        Domain Dom = env.lookupDom(leftName,true);
        String nameOfbottomIdentifier = null;
        SimpleNode tmp = null;

        Relation Rel = null;
		//get the name of last identifier of the right child of the node
		if(projectRight.opcode == OP_IDENTIFIER)
		{
			nameOfbottomIdentifier = projectRight.name;
			Rel = env.lookupRel(nameOfbottomIdentifier,true);
            isNest = Rel.isDomainIn(Dom);
            tmp = projectRight;
		}
	   else if(projectRight.identifier.equals("Project"))
	   {
		 SimpleNode tempNode = (SimpleNode)projectRight.jjtGetChild(0);
		 //SimpleNode tmp;
		 while(true)
		 {
		   tmp = (SimpleNode)tempNode.jjtGetChild(0);
		   if(tmp.opcode == OP_IDENTIFIER)
		   {   nameOfbottomIdentifier = tmp.name;
			   break;
		   }
		   else
			  tempNode = tmp;
		 }

		 Rel = env.lookupRel("."+ nameOfbottomIdentifier,true);
	     isNest = Rel.isDomainIn(Dom);
	   }
	    //check recursive nesting case
	    //e.g N/A in R/N;
	    String domName;
	    if(isNest == false)
	    {
		  for(int i=0;i<Rel.domains.length;i++)
		  {
			  domName = Rel.domains[i].name;
			  if(domName.indexOf('_')!=-1)
			  {
				if(domName.substring(0,domName.indexOf('_')).equals(nameOfbottomIdentifier))
				{ isNest = true;
				   break;
			    }
			  }
	      }
		}

		if(isNest == true)
		{
		  //copy the tree of project
		  SimpleNode dupProLower = (SimpleNode)projectLeft.jjtDeepDuplicate();
		  //SimpleNode dupProUpper = (SimpleNode)projectRight.jjtDeepDuplicate();

		  //create the tree like
		  //            Project
		  //             /    \
		  //       Projector  Identifier
		  //           /
		  //  ExpressionList
		  //          /
		  //      vertical

          SimpleNode proNode = SimpleNode.jjtCreateNode("Project",421,427,null);
		  SimpleNode protNode = SimpleNode.jjtCreateNode("Projector",420,420,null);
		  protNode.setBit(1);
		  SimpleNode expNode = SimpleNode.jjtCreateNode("ExpressionList",592,592,null);
		  SimpleNode verNode = SimpleNode.jjtCreateNode("Vertical",306,362,null);
		  SimpleNode ideNode = SimpleNode.jjtCreateNode("Identifier",230,230,nameOfbottomIdentifier);

		  verNode.jjtSetParent(expNode);
		  (expNode).jjtAddChild(verNode);
		  expNode.jjtSetParent(protNode);
		  (protNode).jjtAddChild(expNode);
		  protNode.jjtSetParent(proNode);
		  (proNode).jjtAddChild(protNode);
		  ideNode.jjtSetParent(proNode);
		  (proNode).jjtAddChild(ideNode);

		  //connect 2 trees
          tmp.jjtReplace(proNode);

          //create a big tree
	      dupProLower.jjtSetParent(verNode);
	      (verNode).jjtAddChild(dupProLower);

	      //replace the left tree
	      node.jjtReplace(projectRight);
         }
		 else
           throw new InterpretError("Domain "+ leftName + " is not in Relation " + nameOfbottomIdentifier);

	  }//end of verticalL.type == OP_RED
    } //end of projectLeft.identifier.equals("Project")

     //if it is a recursive nesting case, modify the node
     //change RR/recurR/recurR to RR/recurR/recurR_0 and so on;
     SimpleNode nodeW = node;
     SimpleNode tmpNode,tmpNodeR;
     String nameOfBottomNode;
     String nameOfIdentifier = " ";
     String tmpName = null;
     int numOfRecur = 0;


     // only in the Project sugar case R/N/N and R/aa
     // the modification of the tree should be taken
     // first determine whether the project sugar case if yes modify the tree

     if(node.identifier.equals("Project") ) //&& node.jjtGetNumChildren()>3)
     {
   	   SimpleNode verti =(SimpleNode) node.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);
       SimpleNode id = (SimpleNode) node.jjtGetChild(1);

       //the case B in R/N should be handled later
       if(verti.type == OP_RED && id.opcode==OP_IDENTIFIER)
       {
        nameOfIdentifier = id.name;
        while(true)
        {
		   tmpNode = (SimpleNode)nodeW.jjtGetChild(0);
		   if(tmpNode.identifier.equals("Project"))
		   {

			   tmpNodeR = (SimpleNode)tmpNode.jjtGetChild(1);
			   if(tmpNodeR.opcode == OP_IDENTIFIER)
			   {
				  tmpName = tmpNodeR.name;
				  if(nameOfIdentifier.equals(tmpNodeR.name))
				  {
					 tmpNodeR.name = tmpNodeR.name + "_" + numOfRecur;
					 numOfRecur++;
				  }
				  nameOfIdentifier = tmpName;
			   }

		   }
		   if(tmpNode.opcode == OP_IDENTIFIER)
		   {
			 if(tmpNode.name.equals(nameOfIdentifier))
			   tmpNode.name = tmpNode.name + "_" + numOfRecur;

			 nameOfBottomNode = tmpNode.name;
			 break;
		   }

                   /******* start adding by Fan  *******/
                   else if(tmpNode.opcode == OP_SELF){  // Fan Guo September 28, 2004
                        nameOfBottomNode = "self";
                        break;  // [red ujoin of [self] in device] in products;
                   }else if(tmpNode.type == OP_LITERAL){  // Fan Guo January 7, 2005
                   		tmpNode = LookupValidRelName(0, tmpNode, env); 
                   		nameOfBottomNode = tmpNode.name;
                    break;  // [red ujoin of [2000] in Rel;
                   }
                   /******* end of Fan's addition  *******/

		   nodeW = tmpNode;
	   }

       //System.out.println("nameOfBottomNode = "+nameOfBottomNode);

	   //Note: Need to check whether the last identifier of the tree
	   //is a Domain or Relation (except it is a reddom...)
	   //if the nameOfBottomNode is a domain name,remove the vertical node above it
	   if(nameOfBottomNode.startsWith("reddom")==false)
	   {
		 SimpleNode nodeV = (SimpleNode)tmpNode.jjtGetParent();
		 if(nodeV.type == OP_RED) //Feb2004, exclude not-sugar cases
         {
			Relation nestR = env.lookupRel("."+nameOfBottomNode,true);
            if(nestR == null) //it's not a relation
            {
	          Domain d = env.lookupDom(nameOfBottomNode,true);
              //if it is a domain, its parent vertical node should be removed
              if(d != null)
		      {
		        //SimpleNode nodeV = (SimpleNode)tmpNode.jjtGetParent();
                        // Modified by Zongyan 2005 Nov. Excluding T <- [red + of a] in R;
                        //if(nodeV.type == OP_RED)                           
                        if(nodeV.type == OP_RED && nodeV.opcode != OP_PLUS)
		        {
					nodeV.jjtReplace(tmpNode);
			        nodeV.jjtRemoveChild(0);
		        }
		      }
		      else
		      {
			     //SimpleNode nodeComp = (SimpleNode)tmpNode.jjtGetParent();
		         //if(nodeComp.type == OP_RED)
			     //zhan modify on Feb 2004
		       throw new InterpretError("Error: Domain "+nameOfBottomNode+" is not found.");
		      }
		  }
         }
       }
     }
   }

  return node;
  }
  //end of zhan's addition sugarInProject



  //The follwoing is added by  Hongyu ( Feb 14, 02)
  private Relation evaluateQSelect(SimpleNode orgNode, String filename,
				   Hashtable htable, Environment env)
       throws InterpretError
  {
    Relation rel = null;
    switch(orgNode.opcode){
    case OP_WHEREIN: {

      /*STEP 1:
	This is to get the original relation. (The after "in" part in a query)
	or if the query is without "in", to get the part after the "quant" list.
	*/
      SimpleNode node =  (SimpleNode)orgNode.jjtDeepDuplicate();
      SimpleNode qtCond = new SimpleNode(nextTempName()),orgRel = null;
      int numOfChildren = orgNode.jjtGetNumChildren();
      boolean cond = false,projIsSth = false,flag=true;
      String orgRelName = nextTempName();
      Domain d=null;
      Vector reOrderDomsV = new Vector();

      // case OP_WHEREFROM: the query is without "in" part
      if (numOfChildren == 2){
	orgRel = (SimpleNode)node.jjtGetChild(1);
	cond = true;
      }
      else {
	orgRel = (SimpleNode)node.jjtGetChild(numOfChildren-1);
	qtCond.jjtAddChild((SimpleNode)node.jjtGetChild(1));
	d = new Domain(nextTempName(), BOOLEAN, 0, qtCond);// define a temp boolean domain for "where" clause
      }
      rel = evaluateTLExpression(orgRel,orgRelName,htable,env);// get the org relation
      //rel.load();
      //rel.print();

      /*STEP 2: get the project domain list */
      Vector allDoms = new Vector();
      allDoms = qtEvaluateProj(orgNode,env,allDoms);

      Domain[] projList = new Domain[allDoms.size()];
      allDoms.copyInto(projList);
      int p = projList.length;
      if(p == 1){
	if((projList[0].name.compareTo("0qtProjIsEmpty") == 0) ||
	   (projList[0].name.compareTo("0qtProjIsSth") == 0))
	  {
	    p = 0;
	    allDoms.removeAllElements();
	    if(projList[0].name.compareTo("0qtProjIsSth") == 0)
	      projIsSth = true;
	  }
      }

     // for(int i=0;i<allDoms.size();i++)
//	System.out.println("Porj "+ i+ " "+((Domain)allDoms.elementAt(i)).name);

      /*STEP 3: get the quantifier list */
      allDoms = qtEvaluateQuantDom(node,env,allDoms);

    //  for(int i=0;i<allDoms.size();i++)
//	System.out.println("Porj + Quant "+ i+ " "+((Domain)allDoms.elementAt(i)).name);

      /*STEP 4: if the project domain list is empty, we have to compare the reOrderDomsV with ori relation domain list */
      if (projList[0].name.compareTo("0qtProjIsEmpty") == 0){
	Vector reOrderDomsV1 =  new Vector();
	p=rel.domains.length;
	for (int i=0;i<p;i++)
	  reOrderDomsV1.addElement(rel.domains[i]);

	for (int i=0;i<allDoms.size();i++){
	  if(!(reOrderDomsV1.contains(allDoms.elementAt(i))))
	    reOrderDomsV1.addElement(allDoms.elementAt(i));
	  else{
	    p=0;
	    flag = false;
	    break;
	  }
	}
	if(flag){
	  allDoms= reOrderDomsV1;
	  flag=false;
	}
      }

      int q = allDoms.size() - p;

      /*STEP 5: get the "where" caluse if the query has it*/
      if(!cond)
	allDoms.addElement(d);

      /*STEP 6: get PredList */
      Domain[] predList = qtEvaluatePred(node,env);

      /*STEP 7: get domains inside each (# ) or (. ) or(% ) caluse. */
      for(int i=0;i<q;i++)
	allDoms = qtEvaluateDomInQuant(predList[i].tree,env,allDoms);


      /*STEP 8: get the reorder domain list */
      for(int i=0;i<allDoms.size();i++){
	if(!(reOrderDomsV.contains(allDoms.elementAt(i)))) // delete duplicate domains in allDoms list
	  reOrderDomsV.addElement(allDoms.elementAt(i));
      }


      /*STEP 9: get the domains in rel list */
      if(flag){
	for(int i=0;i<rel.domains.length;i++){
	  if(!(reOrderDomsV.contains(rel.domains[i])))
	    reOrderDomsV.addElement(rel.domains[i]);
	}
      }

      /*STEP 10: get the reorder domain list */
      Domain[] doms = new Domain[reOrderDomsV.size()];
      reOrderDomsV.copyInto(doms);

//	for(int i=0;i<doms.length;i++){
//	System.out.println("allDoms "+ i +" " + doms[i].name );
//	System.out.println("allDoms type "+ i +" " + doms[i].type );
//     }

      /*STEP 11: evaluating */
      Actualizer qtActualizer = new Actualizer(doms,rel,env,true);
      qtActualizer.qtActive = true;
      qtActualizer.p = p;
      qtActualizer.q = q;
      qtActualizer.predList = predList;
      qtActualizer.cond = cond;
      qtActualizer.qtCond = qtCond;
      qtActualizer.projIsSth = projIsSth;
      Relation r = qtActualizer.actualize();
      if(r.data== null)
	r.load();
      if(!flag){
	r.project(rel.domains);
      }
      qtActualizer.cleanup();
      rel=r;
      deleteFile(orgRelName);// might cause problems?

      //System.out.println("End");// End of evaluation.

      break;
    }

    case OP_WHENIN:{
      orgNode.opcode = OP_WHEREIN;
      rel =  evaluateQSelect(orgNode,filename,htable,env);
      orgNode.opcode = OP_WHENIN;
      break;
    }

    // End of Hongyu Added. July 15, 01.

    case OP_WHEREFROM :
    case OP_WHENFROM :

      throw new InterpretError("TO BE IMPLEMENTED : QSelector");
    }
    return rel;
  }

  /* The following methods are for QSelector.
     Added by hongyu (June 25, 01) */
  private Vector qtEvaluateProj(SimpleNode node,Environment env, Vector allDoms)
       throws InterpretError
  {
    Domain[] doms = null;
    Domain d = null;
    SimpleNode nodeParent = (SimpleNode) node.jjtGetParent();
    //System.out.println(" nodeParent");
    //printNode(nodeParent);
    if(nodeParent.identifier.compareTo("Project")== 0){//case: a <- #[ITEM] in qaunt (#>2) COMP in SUPPLY
      d = new Domain("0qtProjIsEmpty",BOOLEAN,0,null);
      doms = new Domain[1];
      doms[0] = d;
      //System.out.println("--- Empty project list"+d.name);
    }
    else{
      SimpleNode orgGetExpresstion = (SimpleNode)nodeParent.jjtGetChild(0);

      //clone the orgGetExpresstion, and keep the orginal tree untouched.(recrusive view)
      SimpleNode getExpresstion =  (SimpleNode)orgGetExpresstion.jjtDeepDuplicate();

      if(getExpresstion.isBitSet(0)){
	//let s be # quant (#>2)COMP in SUPPLY; which is not a valid statement
	//a <- # quant (#>2)COMP in SUPPLY; which is not a valid statement either
	throw new  InterpretError(" QTselector: need to be implemented ");
      }
      else if(getExpresstion.isBitSet(1)){
	SimpleNode expresstionList =
	  (SimpleNode)getExpresstion.jjtGetChild(0);
	doms = ExpressionListToDomains(expresstionList);
      }

      else {// empty project list
	if(getExpresstion.type == OP_IDENTIFIER){//case: a<-  quant (#>2)COMP in SUPPLY;
	  //System.out.println("0qtProjIsEmpty 1  is set ");
	  d = new Domain("0qtProjIsEmpty",BOOLEAN,0,null);
	  doms = new Domain[1];
	  doms[0] = d;
	}
	else{
	  if(getExpresstion.jjtGetNumChildren() == 0){//case: a<- [] quant (#>2)COMP in SUPPLY;
	    //System.out.println(" getExpresstion.jjtGetNumChildren() == 0 ");
	    //System.out.println("0qtProjIsSth  is set ");
	    //System.out.println("type "+ getExpresstion.type);
	    d = new Domain("0qtProjIsSth",BOOLEAN,0,null);
	    doms = new Domain[1];
	    doms[0] = d;
	  }
	  else {
	    //System.out.println("loop ");
	    getExpresstion = (SimpleNode)getExpresstion.jjtGetChild(0);
	    doms = ExpressionListToDomains(getExpresstion);
	  }
	}
      }
    }
    for(int i=0;i<doms.length;i++){
      if(!(allDoms.contains(doms[i])))
	allDoms.addElement(doms[i]);
      else
	throw new InterpretError(" duplicate attributes in project domain list. ");
    }
    return allDoms;
  }

  private Vector qtEvaluateQuantDom(SimpleNode node,Environment env,Vector allDoms)
       throws InterpretError
  {
    SimpleNode quantifierList = (SimpleNode)node.jjtGetChild(0);
    int n = quantifierList.jjtGetNumChildren();
    Vector doms= new Vector();
    Domain d = null;
    for(int i=0; i<n; i++){
      SimpleNode quantifier =
	(SimpleNode)quantifierList.jjtGetChild(i).jjtGetChild(1);
      d = env.lookupDom(quantifier.name,true);

      if (d == null)
	throw new  InterpretError("Domain " + quantifier.name + " is not declared.");
      else if(doms.contains(d))
	throw new InterpretError(" duplicate attributes in quantifier domain list. ");
      else if (allDoms.contains(d))
	throw new InterpretError(" projected attribute appears in quantifier domain list.");
      else{
	doms.addElement(d);
	allDoms.addElement(d);
      }
    }

    return allDoms;
  }

  private Domain[] qtEvaluatePred(SimpleNode node,Environment env)
       throws InterpretError
  {
    SimpleNode quantifierList = (SimpleNode)node.jjtGetChild(0);
    int n = quantifierList.jjtGetNumChildren();
    Domain[] doms = new Domain[n];

    for(int i=0; i<n; i++){
      SimpleNode quantifier =
	(SimpleNode)quantifierList.jjtGetChild(i).jjtGetChild(0);
      SimpleNode sn = new SimpleNode(" ");
      sn.list = new Vector();
      sn.list.addElement( new Vector()); // for #
      sn.list.addElement( new Vector()); // for " ." and " %"
      sn.jjtAddChild(quantifier);
      Domain d = new Domain(" ",BOOLEAN,0,sn);
      doms[i] = d;
      //System.out.println("Pred "+ i+ " "+d.name);
    }
    return doms;
  }

  private Vector qtEvaluateDomInQuant(SimpleNode node, Environment env,Vector alldoms)
       throws InterpretError
  {
    if(node.type == OP_IDENTIFIER){
      //System.out.println("qtEvaluateDomInQuant opcode "+node.opcode);
      Domain d = env.lookupDom(node.name,true);

      /* Here is a defect.
	 for query:
	 a <- [ITEM] quant (# >= #[ITEM] in SUPPLY) COMP in  SUPPLY;
	 the parser will let it in, but actually #[ITEM] in SUPPLY returns A RELATIOPN WITH SINGLE ATTRIBUTE OF
	 TYPE .interger. So the query should be illegal.
	 When executes this, it will throw interpreter buildtree error.
	 */
      if (d == null){
	Relation r = env.lookupRel(node.name,true);
	if(r == null)
	  throw new  InterpretError(node.name + " is not declared.");
	else throw new  InterpretError(node.name + " is a relation. The query is invalid");
      }
      else if(!alldoms.contains(d))
	alldoms.addElement(d);
    }

    for(int i = 0; i < node.jjtGetNumChildren();i++)
      alldoms = qtEvaluateDomInQuant((SimpleNode)node.jjtGetChild(i),env,alldoms);

    return alldoms;
  }


  /*End of hongyu Added. Feb 14, 02 */

     /**
   * Returns relation only with tuples selected by the selector as
   *   specified by the tree of node.
   *
   * JRelixNS modified/enhanced this method:
   *   In jRelixSJ, nullary-Relation/IDLIST and boolean are now compatible.
   *   Now selector may contain IDLIST within and need cleanup
   *   of temporary relation files created for tuple-by-tuple
   *   actualization.
   */

  private Relation evaluateSelect(SimpleNode node, String filename,
				  Hashtable htable, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    //System.out.println("!!!!"+node.opcode+node.type); //selector--YU
	//zhan's addition to handle sugar in selection Sept 2003
	node = sugarInSelection(node);
	//zhan


    String tmpname = nextTempName();
    Relation rel = null;

    // get conditional expression for selection
    SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
    //System.out.println("!!!!l"+node.opcode+node.type);
    // get relational expression for selection
    SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
    //System.out.println("!!!!r"+node.opcode+node.type);
    //REVISIT:
    //This code is not computation aware.
    //Patrick added this next part to handle compuations
    //The whole code should be look over taking into  consideration
    //computations
//System.out.println("evaluateTLExpression in  evaluateSelect with name:"+ tmpname);
    Relation rl = evaluateTLExpression(noder, tmpname, htable, env);
		//System.out.println("relation: "+rl.name);
    if (rl.rvc == COMPUTATION)
      {
       //return ((Computation) rl).applySelect(nodel,filename, env);
       //modified by Yi Zheng June 25, 2002
       if(this.destrel != null) //r1 is a nested attribute of destrel
       {
       	//System.out.println("nested computation");
       	Computation nr = ((Computation)rl).duplicate(); //each row get a new computation with new states
       	this.destrel.load(true); //make sure data is not null
       	return nr.applySelect(this.destrel, this.currow, nodel, filename, env);
       }
       else
       	return ((Computation) rl).applySelect(nodel,filename, env); //r1 is top level

      }
    deleteFile(tmpname);
    //End of code added by Patrick

    // check for special case for optimization
    //         where A in R;


    // here A is in R and of boolean type, no actualization is needed
    if (nodel.type == OP_IDENTIFIER) {
      Domain d = env.lookupDom(nodel.name, true);
      if (d == null)
				throw new   InterpretError("attribute " + nodel.name + " not defined");
      // evaluate the relational expression for selection

	  //Yi Zheng May 13, 2002, following code is added to exclude state domains
	  //from being projected
	  	if (d.isStateDom == 1)
	  		throw new InterpretError(nodel.name + " is a state domain");

	  //end of Yi's Addition May 13, 2002

      rel = evaluateTLExpression(noder, tmpname, htable, env);

 // START Since jRelixNS BUGFIX
// It can be either a boolean or an IDLIST.
// Can no longer assume it is a boolean here.
			if (Relation.isSigmaJoinImplemented != true) {

   	   if (rel.isDomainIn(d)) { // domain d is in rel, no actualization
				Relation rr = rel.select(d, filename);
				deleteFile(tmpname);
				return rr;
       }
      }
// END Since jRelixNS BUGFIX

    }
    // such two cases :
    //    where A = 2 and B = 3 in R;
    //    <OR>
    //    let COND be A = 2 and B = 3;
    //    where COND in R;
//System.out.println("type of nodel:"+nodel.type);
    if (nodel.type != OP_IDENTIFIER)
      rel = evaluateTLExpression(noder, tmpname, htable, env);//This seems redudancy as the one above hongyu
    //System.out.println(" after red getting org rel");
    //rel.print();
    SimpleNode sn = new SimpleNode("virtmp");
    sn.jjtAddChild(nodel);

    //added by Yi Zheng May 13, 2002, to prevent state domains from being used in condition
//System.out.println("traverseNode:"+nodel.identifier+" type:"+nodel.type+" opcode"+nodel.opcode);
		
    traverseNode(nodel,0,env);
    //end of Yi Zheng's addition May 13, 2002

    Domain d = new Domain("0select0", BOOLEAN, 0, sn);
    Domain[] doms = new Domain[1];
    doms[0] = d;

    // the follwing is  added by Andy
    // START Since jRelixNS
    // added support to cleanup
 //   if (env.isSigmaJoinImplemented != true) {
 	 if (Relation.isSigmaJoinImplemented != true) {
      //System.out.println(" isSigmaJoinImplemented != true ");
      Relation rr = new Actualizer(doms, rel, env).actualize();
      rr.name = tmpname;
      rr.save();
      // select only the attributes in relation rel and with true value
      // in domain d.
      Relation rrr = rr.select(d, rel.numattrs, filename);

      deleteFile(tmpname);

      return rrr;
    }
    else{
      // This is added for the boolean-IDLIST compatibility.
      // Now selector may contain IDLIST within and need cleanup
      // of temporary relation files created for tuple-by-tuple
      // actualization.
     // System.out.println(" isSigmaJoinImplemented == true ");
      //rel.print();
      Actualizer actualizer = new Actualizer(doms, rel, env);
      Relation rr = actualizer.actualize();
      rr.name = tmpname;
      rr.save();
      // select only the attributes in relation rel and with true value
      // in domain d.
      Relation rrr = rr.select(d, rel.numattrs, filename);
      actualizer.cleanup();

      deleteFile(tmpname);

      return rrr;
    }
             // END Since jRelixNS

  }


 /* --------------- zhan's addition Sept 2003 ----------- */
private SimpleNode sugarInSelection(SimpleNode node)
{
     //modify tree for query "[B] where N/A="y" in R"; - sugar in the selection
     //to the tree: [B] where ([] where A ="y" in N) in R;

     // The tree of query where N/A = "y" in R
     //               Selector
     //                /   \
     //           compare  identifier (R)
     //              /
     //           project
     //              /
     // Should be modified to

     //               Selector
     //                /   \
     //            Tselect identifier (R)
     //              /   \
     //       projector  Selector
     //                  /    \
     //              compare   identifier (N)

     //Node: following code is used to the tree modification

    //get the left tree of "Selector" node
    SimpleNode selectLeft = (SimpleNode)node.jjtGetChild(0);

    //to check whether the selection contains a sugar(sugars)
    boolean isSugar=false;
    SimpleNode tmpNode = selectLeft;
  	SimpleNode tmpChildNode;
  	String nameOfbottomIdentifier; // get the name of the bottom domain

  	if(selectLeft.identifier.equals("Compare"))
  	{
  	 while(true)
  	 {
  	   tmpChildNode = (SimpleNode)tmpNode.jjtGetChild(0);
  	   if(tmpChildNode.type == OP_RED)
  	   {  isSugar = true;
  	      break;
  	   }

           /*
                if(tmpChildNode.opcode == OP_IDENTIFIER)
                break;
                */ //Commented out on July 7, 2004 use the following modification version, Fan Guo
           /******* Start modifying by Fan *******/
        SimpleNode pNode = (SimpleNode)tmpChildNode.parent;
        boolean isCast = (tmpChildNode.type == OP_TYPE && pNode.opcode == OP_CAST);
        if(tmpChildNode.opcode == OP_IDENTIFIER || isCast)      // Fan Guo July 7, 2004
                                                      // for supporting "OP_CAST"
           break;
        /******* end of Fan's modification *******/
  	   else
  	      tmpNode = tmpChildNode;
  	 }

      if(isSugar==true)   //contains sugar in the selector
      {
      //System.out.println("isSugar == true ");
  	  //create the base duplicator tree like
  	  //              TSelect(tselect)
  	  //              /     \
  	  //       Projector   Selector(selector)
  	  //                    /   \
  	  //                tmpLeft Identifier(idNode)

  	  SimpleNode tselect = SimpleNode.jjtCreateNode("TSelect",422,422,null);
  	  SimpleNode projector = SimpleNode.jjtCreateNode("Projector",420,420,null);
  	  SimpleNode selector = SimpleNode.jjtCreateNode("Selector",423,425,null);
  	  SimpleNode idNode = SimpleNode.jjtCreateNode("Identifier",230,230,null);

  	  SimpleNode tmpLeft = SimpleNode.jjtCreateNode("Identifier",230,230,null);
  	  //tmpLeft is used to save the position of left child of Selector

  	  projector.jjtSetParent(tselect);
  	  (tselect).jjtAddChild(projector);
  	  tmpLeft.jjtSetParent(selector);
  	  (selector).jjtAddChild(tmpLeft);
  	  idNode.jjtSetParent(selector);
  	  (selector).jjtAddChild(idNode);
  	  selector.jjtSetParent(tselect);
  	  (tselect).jjtAddChild(selector);

  	  //get the name of the identifiers in the right of "Project" node
  	  String projectRightId = null;
  	  Vector tselVec = new Vector();
  	  SimpleNode dupTSelect,lastdupTSelect;
  	  tmpNode = selectLeft;

  	  while(true)
  	  {
  	    tmpChildNode = (SimpleNode)tmpNode.jjtGetChild(0);
  	    if(tmpChildNode.identifier.equals("Project"))
  	    {
  		  //copy the tree of TSelect
  		  dupTSelect = (SimpleNode)tselect.jjtDeepDuplicate();
            projectRightId = ((SimpleNode)tmpChildNode.jjtGetChild(1)).name;
            ((SimpleNode)dupTSelect.jjtGetChild(1).jjtGetChild(1)).name = projectRightId;

            tselVec.addElement(dupTSelect);
  		}
  	    else if(tmpChildNode.opcode == OP_IDENTIFIER)
  	    {  nameOfbottomIdentifier = tmpChildNode.name;
  	       break;
          }

          tmpNode = tmpChildNode;

  	 }
      //create the compareNode
      //             Compare
      //              /    \
      //        Identifier Literal

      SimpleNode compare =(SimpleNode)node.jjtGetChild(0);
      SimpleNode compareLeft = (SimpleNode)compare.jjtGetChild(0);
      compareLeft.jjtReplace(tmpChildNode);
      compareLeft.jjtRemoveChild(0);
      compareLeft.jjtRemoveChild(0);

      //connect
      SimpleNode lastTselect = (SimpleNode)(tselVec.lastElement());
      SimpleNode lastSelect = (SimpleNode)lastTselect.jjtGetChild(1).jjtGetChild(0);
      lastSelect.jjtReplace(compare);

      SimpleNode ttNode,tSelector;
      SimpleNode tNode = null;

      if(tselVec.size()==1)
      	tNode = lastTselect;
  	  else
  	  {
  		for(int i=tselVec.size()-1;i>0;i--)
  		{
  			ttNode = (SimpleNode)tselVec.elementAt(i);
  			tNode = (SimpleNode)tselVec.elementAt(i-1);
  			tSelector = (SimpleNode)tNode.jjtGetChild(1).jjtGetChild(0);
  			tSelector.jjtReplace(ttNode);
  		 }
  	  }

      SimpleNode selectorRight = ((SimpleNode)node.jjtGetChild(1)).jjtDeepDuplicate();
      node.jjtRemoveChild(0);
      node.jjtRemoveChild(0);

      tNode.jjtSetParent(node);
  	  (node).jjtAddChild(tNode);

  	  selectorRight.jjtSetParent(node);
  	  (node).jjtAddChild(selectorRight);


     }//end if sugar ==true
    }//end of is SelectorLeft = compare


  node = RecursiveSelection(node,0,null);//zhan

  return node;
}

 private SimpleNode RecursiveSelection(SimpleNode node,int L,String recurName)
 {
      //System.out.println("recursiveSeletion -------------");
      int j = L;
      String recursiveDomain = recurName;
      //System.out.println("node:"+node.identifier+"type:"+node.type+"opcode:"+node.opcode);
      if(node.jjtGetNumChildren()!=0)
      {
      //System.out.println("node:"+node.identifier+"has children:"+node.jjtGetNumChildren());
		SimpleNode nd1 = (SimpleNode) node.jjtGetChild(1);
		//case1: where N/A=10 in R;
		//          selector
		//           /      \
	    //       Tselect IDENTIFIER
		if(nd1.type == OP_IDENTIFIER)
		{
			if(recursiveDomain != null)
			 {
				if(recursiveDomain.equals(nd1.name))
				{
					 nd1.name = recursiveDomain + "_"+ j;
					 j = j +1;
				 }
				 else
					 recursiveDomain = nd1.name;
			 }
			else
				recursiveDomain = nd1.name;
	    }

	    //END OP_IDENTIFIER
         //case2: where N/A=10 in R/N;
         // selector
	     //  /      \
	     // Tselect project
        if(nd1.identifier.equals("Project"))
        {  SimpleNode nodeW = nd1;
           SimpleNode tmpNode,tmpNodeR;
		   SimpleNode verti =(SimpleNode) nd1.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);
           SimpleNode id = (SimpleNode) nd1.jjtGetChild(1);
           if(verti.type == OP_RED && id.opcode==OP_IDENTIFIER)
           {
             recursiveDomain = id.name;
             while(true)
             {
		       tmpNode = (SimpleNode)nodeW.jjtGetChild(0);
		       if(tmpNode.identifier.equals("Project"))
		       {
			      tmpNodeR = (SimpleNode)tmpNode.jjtGetChild(1);
			      if(tmpNodeR.opcode == OP_IDENTIFIER)
			      {
					  if(recursiveDomain.equals(tmpNodeR.name))
					  {
						 tmpNodeR.name = tmpNodeR.name + "_" + j;
						 j++;
					  }
					  recursiveDomain = tmpNodeR.name;
			      }
		       }

		      if(tmpNode.opcode == OP_IDENTIFIER)
		      {
			     if(tmpNode.name.equals(recursiveDomain))
    			  tmpNode.name = tmpNode.name + "_" + j;

			     recursiveDomain = tmpNode.name;
                 break;
		      }
  		      nodeW = tmpNode;
 		     }//end while
 		    }
	   }

        // end of case2

        SimpleNode nd0 = (SimpleNode) node.jjtGetChild(0); //TSELECT
		if(nd0.type == OP_TSELECT && nd0.opcode== OP_TSELECT)
        {
	       SimpleNode selectNode = (SimpleNode)nd0.jjtGetChild(1);
		   RecursiveSelection(selectNode,j,recursiveDomain);
    	}
		 //    selector
		 //     /       \
		 // Tselect   identifier
		 //   /  \
		 //project selector
		else if(nd0.type == 300) //AND OR
		{
		   SimpleNode nr = (SimpleNode)nd0.jjtGetChild(1);
		   if(nr.type == OP_TSELECT)
		   {
			  SimpleNode nrChild1 = (SimpleNode)nr.jjtGetChild(1);
			  RecursiveSelection(nrChild1,j,recursiveDomain);
		   }

		}
		//    selector
		//    /       \
		//  And     identifier
		//   / \
	 // compare Tselect
	   //         /  \
	     //   project selector

       }//end if
       return node;
   }

 /* --------------- end of zhan's addition Sept 2003 ----------- */
 
 
/**
@author Yu Gu Fall Sep 19, 2004
@param String filename: name of the returned relation
@return relation
@exception InterpretError
Grep a pattern from a relation, not use unix grep, return a relation
*/
public Relation evalGrepRel(Relation grepR,Relation rel,String filename,Hashtable htable, Environment env) 
       throws InterpretError
{
  String[] whereC =extractWhere(grepR);
  int numOfTuple = rel.numtuples;
  int numOfAttr = rel.numattrs;
  boolean [] flag = new boolean[numOfTuple];
  int asci = 6;
  String aChar = new Character((char)asci).toString();
  String tmpname = nextTempName();
//  String tmpname = "result";
//  Relation result = new Relation(filename, rel.domains, rel.rvc, 0,rel.numattrs, 0,  null, env);
  Relation result = new Relation(tmpname, rel.domains, rel.rvc, 0,rel.numattrs, 0,  null, env);
  for (int i =0; i<whereC.length; i++)
  {
  // System.out.println(i+": "+whereC[i]);
  }
  try {
  DataInputStream relFile = new DataInputStream(new BufferedInputStream(new  FileInputStream(rel.name)));
  String line=relFile.readLine();
//  PrintWriter resultFile = new PrintWriter(new BufferedWriter(new FileWriter((result.name))));
  DataOutputStream resultFile = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(result.name)));
  for (int i =0; i<numOfTuple; i++)
  {
   flag[i] = false;
  }
  int lineCounter = 0;
  int numTuple = 0;
  while (line!=null)
  {
   StringTokenizer st = new StringTokenizer(line,aChar);
       
   //while (col!=null) 
   for (int j =0; j<numOfAttr; j++)
   {
    String col = st.nextToken();   
    //System.out.println("column:"+col);
    for (int i =0; i<whereC.length; i++)
    {
     int find = col.indexOf(whereC[i]);
     //System.out.println("find:"+find);
     if (find>=0)
     {
      if (!flag[lineCounter])
      {
        flag[lineCounter] = true;
//       resultFile.print(line);
	resultFile.writeBytes(line);
	resultFile.writeBytes("\n");
	numTuple++;
      }
      break;
     }
    // System.out.println(i+": "+whereC[i]);
    }
    if (flag[lineCounter])
    break;//col = st.nextToken();
    //System.out.println("col:"+col);
   }
   lineCounter++;
   line=relFile.readLine();
  }//while
  relFile.close();
  resultFile.close();
  result.numtuples=numTuple;
  }
  catch (IOException e)
  { System.out.print("Error: " + e);
      System.exit(1);
  }
  return result;
   //return relR;
}

/**
@author Yu Gu Fall Oct 14, 2004
@param String filename: name of the returned relation
@return relation
@exception InterpretError
Grep a pattern from a top level scalar, not use unix grep, return a relation
*/
public Relation evalGrepScalar(String pattern,Relation rel,String filename,Hashtable htable, Environment env) 
       throws InterpretError
{
  //String[] whereC =extractWhere(grepR);
  String whereC = pattern;
  int numOfTuple = rel.numtuples;
  int numOfAttr = rel.numattrs;
  boolean [] flag = new boolean[numOfTuple];
  int asci = 6;
  String aChar = new Character((char)asci).toString();
  String tmpname = nextTempName();
//  String tmpname = "result";
//  Relation result = new Relation(filename, rel.domains, rel.rvc, 0,rel.numattrs, 0,  null, env);
  Relation result = new Relation(tmpname, rel.domains, rel.rvc, 0,rel.numattrs, 0,  null, env);
  try {
  DataInputStream relFile = new DataInputStream(new BufferedInputStream(new  FileInputStream(rel.name)));
  String line=relFile.readLine();
//  PrintWriter resultFile = new PrintWriter(new BufferedWriter(new FileWriter((result.name))));
  DataOutputStream resultFile = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(result.name)));
  for (int i=0;i<numOfTuple;i++)
   flag[i] = false;
  int lineCounter = 0;
  int numTuple = 0;
  while (line!=null)
  {
   StringTokenizer st = new StringTokenizer(line,aChar);
       
   //while (col!=null) 
   for (int j =0; j<numOfAttr; j++)
   {
    String col = st.nextToken();   
    //System.out.println("column:"+col);
    //for (int i =0; i<whereC.length; i++)
    //{
     int find = col.indexOf(whereC);
     //System.out.println("find:"+find);
     if (find>=0)
     {
      if (!flag[lineCounter])
      {
        flag[lineCounter] = true;
//       resultFile.print(line);
	resultFile.writeBytes(line);
	resultFile.writeBytes("\n");
	numTuple++;
      }
      break;
     }
    // System.out.println(i+": "+whereC[i]);
    //}
    //col = st.nextToken();
    //System.out.println("col:"+col);
   }
   lineCounter++;
   line=relFile.readLine();
  }//while
  relFile.close();
  resultFile.close();
  result.numtuples=numTuple;
  }
  catch (IOException e)
  { System.out.print("Error: " + e);
      System.exit(1);
  }
  return result;
   //return relR;
}

/**
@author Yu Gu Fall Oct 14, 2004
@param String filename: name of the returned relation
@return relation
@exception InterpretError
Grep a pattern from a scalar, not use unix grep, return a relation with parameter
*/
//	   Relation relG = evalGrepScalarPV(patt,rel,tmpname,htable,env);//Oct 14, 2004
public Relation evalGrepScalarPV(Domain[] domarray, String pattern,Relation rel,String filename,Hashtable htable, Environment env) 
       throws InterpretError
{
  String whereC =pattern;
  int numOfTuple = rel.numtuples;
  int numOfAttr = rel.numattrs;
  int asci = 6;
  String aChar = new Character((char)asci).toString();
  String tmpname = nextTempName();
  //String tmpname = "result";
  int ndom = rel.numattrs+domarray.length;       //destination domain length
  Domain [] typeArray = new Domain[ndom];
  for (int i = 0; i<rel.numattrs; i++)
 	typeArray[i]=rel.domains[i];
  for (int i = rel.numattrs; i<ndom; i++)
	typeArray[i]=domarray[i-rel.numattrs];
//	 Relation result = new Relation(tmpnameR, typeArray, RELATION, 0, typeArray.length, 0, null, env);
  Relation result = new Relation(tmpname, typeArray, RELATION, 0, typeArray.length, 0, null, env);
  int[] types = result.toTypes(typeArray);  // domain types
//  Relation result = new Relation(tmpname, rel.domains, rel.rvc, 0, rel.numattrs, 0,  null, env);
  try {
  DataInputStream relFile = new DataInputStream(new BufferedInputStream(new  FileInputStream(rel.name)));
  String line=relFile.readLine();
//  PrintWriter resultFile = new PrintWriter(new BufferedWriter(new FileWriter((result.name))));
  DataOutputStream resultFile = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(result.name)));
  int numTuple = 0;
  boolean hasPos = false;
  for (int k = rel.numattrs; k<typeArray.length; k++)
  {
	 if(types[k]==INTEGER) 
		hasPos = true;
  }
  while (line!=null)
  {
   StringTokenizer st = new StringTokenizer(line,aChar);
       
   //while (col!=null) 
   for (int j =0; j<numOfAttr; j++)//for each column
   {
    String col = st.nextToken();   
    //System.out.println("column:"+col);
    //for (int i =0; i<whereC.length; i++)
    //{
     int find = col.indexOf(whereC);
     //System.out.println("find:"+find);
     if (find>=0)
     {
//       resultFile.print(line);
	Integer pos[] = findInLine(whereC,col);
	String val = whereC;
	String attr = rel.domains[j].name;
	String type = rel.domains[j].getType();
	if (hasPos)
	for (int p = 0; p<pos.length;p++)
	{
 	 resultFile.writeBytes(line);
//	 resultFile.writeBytes(aChar);
	 for (int k = rel.numattrs; k<typeArray.length; k++)
	 {
		switch(types[k]) {
		case INTEGER :
			resultFile.writeBytes(pos[p].toString());
			resultFile.writeBytes(aChar);
		     break;
		case ATTRIBUTE :
			resultFile.writeBytes(attr);
			resultFile.writeBytes(aChar);
		     break;
 		case UNIVERSAL :
			resultFile.writeBytes(type+":"+val);
			resultFile.writeBytes(aChar);
		     break;
		case TYPE :
			resultFile.writeBytes(type);
			resultFile.writeBytes(aChar);
		     break;   }//switch
         }
	 resultFile.writeBytes("\n");
	 numTuple++;
	}//for (int p = 0; p<pos.length;p++)
	else//not hasPos
	{
 	 resultFile.writeBytes(line);
	 for (int k = rel.numattrs; k<typeArray.length; k++)
	 {
		switch(types[k]) {
		case ATTRIBUTE :
			resultFile.writeBytes(attr);
			resultFile.writeBytes(aChar);
		     break;
 		case UNIVERSAL :
			resultFile.writeBytes(type+":"+val);
			resultFile.writeBytes(aChar);
		     break;
		case TYPE :
			resultFile.writeBytes(type);
			resultFile.writeBytes(aChar);
		     break;   }//switch
         }
	 resultFile.writeBytes("\n");
	 numTuple++;
	}
//	resultFile.writeBytes("\n");
     }//if find
    // System.out.println(i+": "+whereC[i]);
    //}//for (int i =0; i<whereC.length; i++)
    //col = st.nextToken();
    //System.out.println("col:"+col);
   }//for (int j =0; j<numOfAttr; j++)
//   lineCounter++;
   line=relFile.readLine();
  }//while
  relFile.close();
  resultFile.close();
  result.numtuples=numTuple;
  //System.out.println("result has tuples:"+result.numtuples);
  }
  catch (IOException e)
  { System.out.print("Error: " + e);
      System.exit(1);
  }
  return result;
   //return relR;
}

/**
@author Yu Gu Fall Sep 20, 2004
@param String filename: name of the returned relation
@return relation
@exception InterpretError
Grep a pattern from a relation, not use unix grep, return a relation
*/
	   
public Relation evalGrepRelPV(Domain[] domarray, Relation grepR,Relation rel,String filename,Hashtable htable, Environment env) 
       throws InterpretError
{
  String[] whereC =extractWhere(grepR);
  int numOfTuple = rel.numtuples;
  int numOfAttr = rel.numattrs;
  int asci = 6;
  String aChar = new Character((char)asci).toString();
  String tmpname = nextTempName();
  //String tmpname = "result";
  int ndom = rel.numattrs+domarray.length;       //destination domain length
  Domain [] typeArray = new Domain[ndom];
  for (int i = 0; i<rel.numattrs; i++)
 	typeArray[i]=rel.domains[i];
  for (int i = rel.numattrs; i<ndom; i++)
	typeArray[i]=domarray[i-rel.numattrs];
//	 Relation result = new Relation(tmpnameR, typeArray, RELATION, 0, typeArray.length, 0, null, env);
  Relation result = new Relation(tmpname, typeArray, RELATION, 0, typeArray.length, 0, null, env);
  int[] types = result.toTypes(typeArray);  // domain types
//  Relation result = new Relation(tmpname, rel.domains, rel.rvc, 0, rel.numattrs, 0,  null, env);
  try {
  DataInputStream relFile = new DataInputStream(new BufferedInputStream(new  FileInputStream(rel.name)));
  String line=relFile.readLine();
//  PrintWriter resultFile = new PrintWriter(new BufferedWriter(new FileWriter((result.name))));
  DataOutputStream resultFile = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(result.name)));
  int numTuple = 0;
  boolean hasPos = false;
  for (int k = rel.numattrs; k<typeArray.length; k++)
  {
	 if(types[k]==INTEGER) 
		hasPos = true;
  }
  while (line!=null)
  {
   StringTokenizer st = new StringTokenizer(line,aChar);
       
   //while (col!=null) 
   for (int j =0; j<numOfAttr; j++)//for each column
   {
    String col = st.nextToken();   
    //System.out.println("column:"+col);
    for (int i =0; i<whereC.length; i++)
    {
     int find = col.indexOf(whereC[i]);
     //System.out.println("find:"+find);
     if (find>=0)
     {
//       resultFile.print(line);
	Integer pos[] = findInLine(whereC[i],col);
	String val = whereC[i];
	String attr = rel.domains[j].name;
	String type = rel.domains[j].getType();
	if (hasPos)
	for (int p = 0; p<pos.length;p++)
	{
 	 resultFile.writeBytes(line);
//	 resultFile.writeBytes(aChar);
	 for (int k = rel.numattrs; k<typeArray.length; k++)
	 {
		switch(types[k]) {
		case INTEGER :
			resultFile.writeBytes(pos[p].toString());
			resultFile.writeBytes(aChar);
		     break;
		case ATTRIBUTE :
			resultFile.writeBytes(attr);
			resultFile.writeBytes(aChar);
		     break;
 		case UNIVERSAL :
			resultFile.writeBytes(type+":"+val);
			resultFile.writeBytes(aChar);
		     break;
		case TYPE :
			resultFile.writeBytes(type);
			resultFile.writeBytes(aChar);
		     break;   }//switch
         }
	 resultFile.writeBytes("\n");
	 numTuple++;
	}//for (int p = 0; p<pos.length;p++)
	else//not hasPos
	{
 	 resultFile.writeBytes(line);
	 for (int k = rel.numattrs; k<typeArray.length; k++)
	 {
		switch(types[k]) {
		case ATTRIBUTE :
			resultFile.writeBytes(attr);
			resultFile.writeBytes(aChar);
		     break;
 		case UNIVERSAL :
			resultFile.writeBytes(type+":"+val);
			resultFile.writeBytes(aChar);
		     break;
		case TYPE :
			resultFile.writeBytes(type);
			resultFile.writeBytes(aChar);
		     break;   }//switch
         }
	 resultFile.writeBytes("\n");
	 numTuple++;
	}
//	resultFile.writeBytes("\n");
     }//if find
    // System.out.println(i+": "+whereC[i]);
    }//for (int i =0; i<whereC.length; i++)
    //col = st.nextToken();
    //System.out.println("col:"+col);
   }//for (int j =0; j<numOfAttr; j++)
//   lineCounter++;
   line=relFile.readLine();
  }//while
  relFile.close();
  resultFile.close();
  result.numtuples=numTuple;
  //System.out.println("result has tuples:"+result.numtuples);
  }
  catch (IOException e)
  { System.out.print("Error: " + e);
      System.exit(1);
  }
  return result;
   //return relR;
}

/**
@author Yu Gu Sep 20, 2004
@param String word: word that want to find in line
@param String line: find in line
@return position int array
*/
public Integer [] findInLine(String word,String line)
{
 Vector posList = new Vector();
 int position = line.indexOf(word);
 while (position>=0)
 {
 	Integer p = new Integer(position);
 	posList.addElement(p);
	int start = position+word.length(); 
	position = line.indexOf(word,start);
 }
 Integer [] pos = new Integer[posList.size()];
 posList.copyInto(pos);                   
 
 return pos;
}

/**
@author Yu Gu Oct 14, 2004
@param String name: name of the top level scalar
@return String
Get pattern from a top level scalar
*/
public String getPatternFromScalar(String name)
{
  Domain d = env.lookupDom(name, true);
  SimpleNode nd = (SimpleNode) d.tree;
  //declare t1 integer<-40; opcode=453,type=470
  //System.out.println("tree:"+nd.opcode+" child:"+((SimpleNode) nd.jjtGetChild(0)).opcode+":"+((SimpleNode) nd.jjtGetChild(0)).type);
  SimpleNode pnd = (SimpleNode) nd.jjtGetChild(0);
  String pattern = (pnd.getInfo()).toString();
  return pattern;
}
/**
@author Yu Gu Summer 2004
@param SimpleNode node: syntax node from parser
@param String filename: name of the returned relation
@param Hashtable htable: hashtable
@param Environment env: 
@return relation
@exception InterpretError
Grep a paticular pattern from a relation, return a relation
*/
  public Relation evaluateGrep(SimpleNode node,
				       String filename, Hashtable htable, Environment env)
       throws InterpretError
  {	
   SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
   SimpleNode pnd = nd;
   String pattern = "";
   SimpleNode rnd =(SimpleNode) node.jjtGetChild(1); //relation
   boolean needAttr = false;
   boolean needType = false;
   boolean needVal = false;
   boolean needPosition = false;
   boolean needParameter = false;
   String tmpname = nextTempName();
   if (!(nd.toString()).equals("GIDList"))
   {
//System.out.println("start to  evaluateTLExpression of the right child node(second grep)"+tmpname);
    Relation rel = evaluateTLExpression(rnd, tmpname, htable, env);     //grep in rel
		 //env.put(rel);
		 //rel.save();
//     rel.load();
    Relation[] relA = new Relation[rel.numattrs];
    if (!(pnd.type==OP_IDENTIFIER))          //July 15,2004 by Yu
    {
		pattern = (pnd.getInfo()).toString();		                                 	//Grep
    }//end of if (!pnd.type==OP_IDENTIFIER)          //July 15,2004 by Yu
    else          //need to deal with the situation like: grep fix in BoM; (fix is a relation or top level variable) July 15,2004 by Yu
    {
	 Relation grepR = env.lookupRel(pnd.name, true);
     if(grepR == null)
      			throw new InterpretError("relation " + grepR.name + " not found");
     if (grepR.rvc==SCALAR)
	 {
	 String patt = getPatternFromScalar(grepR.name);
	 tmpname = nextTempName();
	 Relation relG = evalGrepScalar(patt,rel,tmpname,htable,env);//Oct 14, 2004
         deleteFile(tmpname);
	 return relG;
	 }
	 else 
	 if (grepR.numattrs*grepR.numtuples==0)
      			throw new InterpretError("pattern relation" + grepR.name + " is empty");
	//pattern = extractPattern(grepR);
	 tmpname = nextTempName();
	 Relation relG = evalGrepRel(grepR,rel,tmpname,htable,env);
     deleteFile(tmpname);
	 return relG;
    }
    String where = "where ";
    SimpleNode condiNode = null;
    boolean find = false;
    for (int j =0; j < rel.numattrs; j++)				//search each column
    {
     Domain[] doms = new Domain[1];
     doms[0]=rel.domains[j];
     int t = rel.domains[j].type;
     int opcodeOfDom = getOpcode(t);
//	     System.out.println(doms[0].name);
     String domname=doms[0].name;
//System.out.println("domain:"+ domname);
     relA[j] = rel.project(doms,"rel"+j);
	 //need to convert the unary relation to text file. June 08,2004 otherwise grep "x$" cannot be found,
								//because there is a ^F at end of each attribute.
     try {
		relA[j].saveToFile();
            	}
     catch (IOException e) {
                System.err.println(e);
            	}
     String fname=relA[j].name+".txt";
     String[] cmd={"egrep",pattern,fname};
     //System.out.println(cmd);
     //delete rel[j].txt --Yu Nov 19, 2004
     try {
	ExecComm command=new ExecComm();
	String outlist[] = command.runCommand(cmd);
	File filef = new File(fname);
	filef.delete(); //--Yu Nov 19, 2004
		                // display its output
        if (outlist.length>0)                         //catch the result of egrep
	{
		        /*for (int i = 0; i < outlist.length; i++)
			{
         		 System.out.println("outlist"+i+":"+outlist[i]);
			}*/	
			int start = 0;
			if (!find)	//construct the first condiNode. June 22,2004
			{
			 find = true;
			 //possible situation: outlist[0] is rel0.txt:10
			 //System.out.println("where:"+where+" outlist[0] is "+outlist[0]+" !!!");
			 condiNode=SimpleNode.jjtCreateNode("Compare",300, 323, null);
			 SimpleNode idNode0 = SimpleNode.jjtCreateNode("Identifier",230, 230,domname);
			 idNode0.setInfo(domname);
					//note: need to change to corresponding type -- by Yu
			 SimpleNode litNode0 = SimpleNode.jjtCreateNode("Literal",470, opcodeOfDom, null);
			 int colon = outlist[0].indexOf(":");
			 //System.out.println("here:outlist[0]is "+outlist[0]);
			 //System.out.println("and filename is "+fname);
			 if (colon>=0)
			  if (outlist[0].substring(0,colon).equals(fname))
			   outlist[0]=outlist[0].substring(colon+1);
			 where=where+domname+"="+"\""+outlist[0]+"\"";
			 //System.out.println("here:outlist[0]"+outlist[0]);
           		 Object o = getInfoFromString(outlist[0],t);
			 litNode0.setInfo(o);
			 idNode0.jjtSetParent(condiNode);
			 condiNode.jjtAddChild(idNode0);
			 litNode0.jjtSetParent(condiNode);
			 condiNode.jjtAddChild(litNode0);
			 start = 1;
			}
		        for (int i = start; i < outlist.length; i++)
			{
			 int colon = outlist[i].indexOf(":");
			 if (colon>=0)
			  if (outlist[i].substring(0,colon).equals(fname))
			   outlist[i]=outlist[i].substring(colon+1);
         		 //System.out.println("here outlist"+i+" is "+outlist[i]);
		    		//construct the where clause	//June 15,2004 by Yu
			 where = where+" or "+domname+"="+"\""+outlist[i]+"\"";
			 SimpleNode orNode=SimpleNode.jjtCreateNode("Or",300, 320,  null);
			 condiNode.jjtSetParent(orNode);
			 orNode.jjtAddChild(condiNode);
			 SimpleNode compNode=SimpleNode.jjtCreateNode("Compare",300, 323, null);
			 SimpleNode idNode = SimpleNode.jjtCreateNode("Identifier",230, 230,domname);
			 idNode.setInfo(domname);
			 SimpleNode litNode = SimpleNode.jjtCreateNode("Literal",470, opcodeOfDom, null);
				//note: need to change to corresponding type -- by Yu
             		 Object o = getInfoFromString(outlist[i],t);
			 litNode.setInfo(o);
			 idNode.jjtSetParent(compNode);
			 compNode.jjtAddChild(idNode);
			 litNode.jjtSetParent(compNode);
			 compNode.jjtAddChild(litNode);
			 compNode.jjtSetParent(orNode);
			 orNode.jjtAddChild(compNode);
			 condiNode=orNode;
		        }
	}           //end of if(outlist.length>0)
     }         //try
     catch (IOException e) {
                System.err.println(e);
            	}
    }//end of for (int j =0; j < rel.numattrs; j++)
	//construct a node to represent the where clause like: where false or y="x3 hello" or z="y3 hi,hello there"
								//selector 425/423
							//				\
						//or 320/300				R 230/230   (rel)
						//		\
				//	compare		 	compare 323/300
				//	323/300                	/  	\
			// 		 /      \ 	    	Id     Literal
			//    	Id   		Literal		z      458/470
			//     	230/230		458/470   	y3 hi,hello there
			//	  y           x3 hello
    if (find)
    {
      //System.out.println("where:"+where);		
      SimpleNode seleNode=SimpleNode.jjtCreateNode("Selector",423,425, null);
      condiNode.jjtSetParent(seleNode);
      seleNode.jjtAddChild(condiNode);
      SimpleNode relNode=SimpleNode.jjtCreateNode("Identifier",230, 230, rel.name);
      relNode.setInfo(rel.name);
      relNode.jjtSetParent(seleNode);
      seleNode.jjtAddChild(relNode);
//			Relation relR = evaluateTLExpression(seleNode, tmpname, htable, env);
      Relation relR = evaluateTLExpression(seleNode, filename, htable, env);
//System.out.println("return from grep:"+relR.name);
      //env.put(relR);
      deleteFile(tmpname);
//      deleteFile(tmpnameR);
      return relR;
    }
    else
    {
//			System.out.println("grep return nothing"); //should return empty relation
//			Relation r = new Relation(tmpname, rel.domains, RELATION, 0, rel.numattrs, 0, null, env);
      Relation r = new Relation(filename, rel.domains, RELATION, 0, rel.numattrs, 0, null, env);
      deleteFile(tmpname);
//      deleteFile(tmpnameR);
      return r;
    } //end of if (find) June 16, 2004 by Yu
   }// end of  if (!(nd.toString()).equals("GIDList"))

   else
   {							//deal with the GIDList  
   	//GIDList
//IDList1   ;	//IDList2
//System.out.println("GIDList");	//June 07,2004 by Yu        example: grep(attr,type;x,y) "\xpattern\y" in R;
    pnd = (SimpleNode) node.jjtGetChild(1);		//contain the pattern
    rnd = (SimpleNode) node.jjtGetChild(2);
    Relation rel = evaluateTLExpression(rnd, tmpname, htable, env);     //grep in rel
    Relation[] relA = new Relation[rel.numattrs];
    SimpleNode nd1 = (SimpleNode) nd.jjtGetChild(0);	               //IDList1

    Vector domlist = new Vector();
    Domain[] domarray = null;

    for(int i = 0; i < nd1.jjtGetNumChildren(); i++)	//check domain type of each of the IDList1
    {
	String dname = ((SimpleNode) nd1.jjtGetChild(i)).name;
	        //Domain oldD = env.lookupTryDom(dname,z,true);//from zhan May 2003
    	Domain oldD = env.lookupDom(dname,true); //Original //zhan
	    	// if(env.lookupDom(dname, true) == null) Yi Zheng May 9, 2002
    	if(oldD == null)
        {  throw new InterpretError("domain "+ dname + " has not been declared");
        }
    	else if (oldD.isStateDom == 1) //Yi Zheng May 9, 2002
      		throw new InterpretError("domain " + dname + " is a state domain");
        else
        {  //domlist.addElement(env.lookupDom(dname,true)); Yi Zheng May 9, 2002
        	//June 07,2004 by Yu;	check if the type is one of (attribute,integer,type,universal)
		if ((oldD.type==INTEGER) ||(oldD.type==ATTRIBUTE)||(oldD.type==TYPE)||(oldD.type==UNIVERSAL))
		{
		  domlist.addElement(oldD);           //keep domains of IDList1 to domlist
		  switch(oldD.type)
		  {
			case  INTEGER: //need to check if this domain is used by the relation that grep in
			for (int p = 0; p<rel.numattrs; p++)
			{
                            if (oldD.name == rel.domains[p].name)
	      			throw new InterpretError("domain " + oldD.name + " has been used in the grep relation");
                        }
   			needPosition = true;
			break;
			case  ATTRIBUTE:
			needAttr = true;
			break;
                	case  TYPE:
	 		needType = true;
			break;
	            	case  UNIVERSAL:
		 	needVal = true;
			break;
		  }
		}
		else throw new InterpretError("Type of domain " + dname + " is invalid");
        }
    }       //end of for(int i = 0; i < nd1.jjtGetNumChildren(); i++)

    domarray= new Domain[domlist.size()];
    domlist.copyInto(domarray);                     //hold the domains from IDList1

//	   pattern = (pnd.getInfo()).toString();
    if (!(pnd.type==OP_IDENTIFIER))          //July 15,2004 by Yu
    {
	pattern = (pnd.getInfo()).toString();		                                 	//Grep
    }//end of if (!pnd.type==OP_IDENTIFIER)          //July 15,2004 by Yu
    else   //need to deal with the situation like: grep fix in BoM; (fix is a relation or top level variable) July 15,2004 by Yu
    {
      Relation grepR = env.lookupRel(pnd.name, true);
      if(grepR == null)
      			throw new InterpretError("relation " + pnd.name + " not found");
      if (grepR.rvc==SCALAR)
	  {
	   String patt = getPatternFromScalar(grepR.name);
	   tmpname = nextTempName();
	   Relation relG = evalGrepScalarPV(domarray,patt,rel,tmpname,htable,env);//Oct 14, 2004
       deleteFile(tmpname);
	   return relG;
	  }
	  else 
      if (grepR.numattrs*grepR.numtuples==0)
      			throw new InterpretError("pattern " + grepR.name + " is empty");
//      pattern = extractPattern(grepR);
	tmpname = nextTempName();
	Relation relG = evalGrepRelPV(domarray,grepR,rel,tmpname,htable,env);
        deleteFile(tmpname);
	return relG;
    }
    String oldPattern = pattern;		//remember the pattern in case it has parameters in it
    Domain[] domarry2 = null;
    boolean hasBegin = false; //has parameter at the beginning of the pattern	Sep 23, 2004
    boolean hasEnd = false;//has parameter at the end of the pattern
    String begin = null;
    String end = null;
    String patternA=pattern;
    Vector domV = new Vector();
    String [] domN = null;
    if (nd.jjtGetNumChildren()==2)                 //has IDList2
    {
	    //System.out.println("has parameters");
	    needParameter = true;
	    Vector domlist2 = new Vector();
	    SimpleNode nd2 =(SimpleNode) nd.jjtGetChild(1); 			//(...;x,y)
	    for(int i = 0; i < nd2.jjtGetNumChildren(); i++)
	    {  String dname = ((SimpleNode) nd2.jjtGetChild(i)).name;

	    	 //Domain oldD = env.lookupTryDom(dname,z,true);//from zhan May 2003
	       Domain oldD = env.lookupDom(dname,true); //Original //zhan

	    	// if(env.lookupDom(dname, true) == null) Yi Zheng May 9, 2002
	       if(oldD == null)
	       {  throw new InterpretError("domain "+ dname + " has not been declared");
	       }
	       else if (oldD.isStateDom == 1) //Yi Zheng May 9, 2002
	    	  			throw new InterpretError("domain " + dname + " is a state domain");
	            else
	    	    {  //domlist2.addElement(env.lookupDom(dname,true)); Yi Zheng May 9, 2002
			if (oldD.type==STRING) 
			{
			  domlist2.addElement(oldD);           //keep domains of IDList2 to domlist2
  			  for (int p = 0; p<rel.numattrs; p++)
			  {
                             if (oldD.name == rel.domains[p].name)
	      			throw new InterpretError("domain " + oldD.name + " has been used in the grep relation");
                          }
			}
			else throw new InterpretError("Type of domain " + dname + " should be STRING");
	     	    }
	    } //for

	    domarry2= new Domain[domlist2.size()];
	    domlist2.copyInto(domarry2);
//	    String domName[] = new String[domarry2.length]		//parameters  from IDList2
  	    //pick up the real pattern
	    //System.out.println("The original pattern is :"+pattern+" and its length is "+pattern.length());
	    //int numPara = domarray2.length;
	    for (int i=0; i<domarry2.length; i++)
	    {
	     String dName = domarry2[i].name;
	     //System.out.println(dName);
	     String par = "\\"+dName;
	     //System.out.println(par);
	     int x = pattern.indexOf(par);
	     if (x<0)
	     	throw new InterpretError("There is no parameter " + dName + " in the pattern");
	     //System.out.println("indexof "+par+" is "+x);
	     int l = dName.length();
	     if (x==0)
	     {
	     	hasBegin = true;
		begin = new String(dName);
	     	pattern = pattern.substring(x+l+1);
	     }
	     else
	     if (x+l+1==pattern.length()) //hasEnd
	     {
	        hasEnd = true;
		end = new String(dName);
		pattern = pattern.substring(0,x);
	     }
	     else
	     {
	     	pattern = pattern.substring(0,x)+".*"+pattern.substring(x+l+1);
		domV.addElement(dName);
	     }
				//check \x\y illegal situation June 15, 2004 by Yu
       	     String ill=".*.*";
	     int y =pattern.indexOf(ill);
	     if (y>=0)
	     	throw new InterpretError("It is illegal to put parameter or .* consecutively" );

	     //System.out.println("length of pattern is "+pattern.length());
	    }
	    if (hasBegin)
	    {
	    	patternA = patternA.substring(begin.length()+1);
		//numPara--;
	    }
	    if (hasEnd)
	    {
	    	patternA = patternA.substring(0,patternA.length()-end.length()-1);
		//numPara--;
	    }
	    domN= new String[domV.size()];
	    domV.copyInto(domN);
	
	    //System.out.println("patternA:"+patternA+","+"pattern:"+pattern);
    } //end of if GIDlist has children=2

    String tmpnameR = nextTempName();
    Domain [] typeArray = null;
    Relation result = null;
    if (!needParameter)     //deal with IDList1, no IDList2
    {
	 int ndom = rel.numattrs+domarray.length;       //destination domain length
     	 typeArray = new Domain[ndom];
	 for (int i = 0; i<rel.numattrs; i++)
	 	typeArray[i]=rel.domains[i];
     	 for (int i = rel.numattrs; i<ndom; i++)
		typeArray[i]=domarray[i-rel.numattrs];
//	 Relation result = new Relation(tmpnameR, typeArray, RELATION, 0, typeArray.length, 0, null, env);
	 result = new Relation(filename, typeArray, RELATION, 0, typeArray.length, 0, null, env);
    }
    else
    {
//	System.out.println("i am here! and pattern is "+pattern+"old pattern:"+oldPattern);
//	pattern = oldPattern;
	int ndom = rel.numattrs+domarray.length+domarry2.length;       //destination domain length
     	typeArray = new Domain[ndom];
	for (int i = 0; i<rel.numattrs; i++)
	 	typeArray[i]=rel.domains[i];
     	for (int i = rel.numattrs; i<rel.numattrs+domarray.length; i++)
		typeArray[i]=domarray[i-rel.numattrs];
     	for (int i = rel.numattrs+domarray.length; i<ndom; i++)
		typeArray[i]=domarry2[i-rel.numattrs-domarray.length];
	//String tmpnameR = nextTempName();;
//	 Relation result = new Relation(tmpnameR, typeArray, RELATION, 0, typeArray.length, 0, null, env);
	result = new Relation(filename, typeArray, RELATION, 0, typeArray.length, 0, null, env);    
    }
    env.put(result);
    //System.out.println("put this:"+result.name);
    result.save();
    for (int j =0; j < rel.numattrs; j++)		//search each column need to construct a where clause for each line(row)
    {
	  SimpleNode condiNode = null;
	  boolean find = false;
  	  Domain[] doms = new Domain[1];
	  doms[0]=rel.domains[j];
 	  int t = rel.domains[j].type;
          int opcodeOfDom = getOpcode(t);
//	    System.out.println(doms[0].name);
	  String domname=doms[0].name;
	  relA[j] = rel.project(doms,"rel"+j);
					//need to convert the unary relation to text file. June 08,2004 otherwise grep "x$" cannot be found,
								//because there is a ^F at end of each attribute.
	  try {
			relA[j].saveToFile();
            	}
      	  catch (IOException e) {
                	System.err.println(e);
            	}
	  String fname=relA[j].name+".txt";
//	  if (!needParameter)
//	  {
//	   String cmd="egrep "+pattern+" "+fname;
	   String[] cmd={"egrep",pattern,fname};
	   //System.out.println(cmd);
//	  }
//	  else 
      	  try {
	    ExecComm command=new ExecComm();
	    String outlist[] = command.runCommand(cmd);
  	    if (outlist.length>0)                         //catch the result of egrep
	    {
  	      String where = "where ";
              for (int i = 0; i < outlist.length; i++)
              {
		//if (star>=0)
		//Grep[] posVal = positionValueS(outlist[i],pattern);  *****Sep 9,2004
		//else
		int colon = outlist[i].indexOf(":");
		if (colon>=0) 
		  if (outlist[i].substring(0,colon).equals(fname))
		   outlist[i]=outlist[i].substring(colon+1);
		//System.out.println("outlist[i]:"+outlist[i]+" and pattern:"+pattern);
		Grep[] posVal = null;
		if (!needParameter)		
		 posVal = positionValue(outlist[i],pattern);	
		 //to remember the position and value of each occurence of pattern. Yu, June 25,2004
		else
		{		
		 posVal = positionValueP(outlist[i],pattern,patternA,domN,domarry2,hasBegin,hasEnd,begin,end);
/*		 for (int grepi = 0;grepi<posVal.length;grepi++)
		 {
//		   System.out.print("posVal:"+grepi+" is "+posVal[grepi].value+","+posVal[grepi].position);
		   for (int grepj = 0; grepj<posVal[grepi].paraValue.length;grepj++)
//		    System.out.println(","+posVal[grepi].paraValue[grepj]);
		 }*/
		}
		find = true;
		where=where+domname+"="+"\""+outlist[i]+"\"";
//System.out.println(where);
		condiNode=SimpleNode.jjtCreateNode("Compare",300, 323, null);
      		SimpleNode idNode0 = SimpleNode.jjtCreateNode("Identifier",230, 230,domname);
      		idNode0.setInfo(domname);
		SimpleNode litNode0 = SimpleNode.jjtCreateNode("Literal",470, opcodeOfDom, null);
           	Object o = getInfoFromString(outlist[i],t);
		litNode0.setInfo(o);
//      		 SimpleNode litNode0 = SimpleNode.jjtCreateNode("Literal",470, 458, null);
//      		 litNode0.setInfo(outlist[0]);
      		idNode0.jjtSetParent(condiNode);
      		condiNode.jjtAddChild(idNode0);
      		litNode0.jjtSetParent(condiNode);
      		condiNode.jjtAddChild(litNode0);
      	 	//System.out.println(where);
           	SimpleNode seleNode=SimpleNode.jjtCreateNode("Selector",423,425, null);
        	condiNode.jjtSetParent(seleNode);
        	seleNode.jjtAddChild(condiNode);
        	SimpleNode relNode=SimpleNode.jjtCreateNode("Identifier",230, 230, rel.name);
        	relNode.setInfo(rel.name);
        	relNode.jjtSetParent(seleNode);
        	seleNode.jjtAddChild(relNode);
		tmpnameR = nextTempName();
		Relation r = evaluateTLExpression(seleNode, tmpnameR, htable, env);    //where domainName=outlist[0] in rel.name
          		 	 //return relR;   ********** attr, type, val, pos                 //possible more than one tuples with same value in this field
		tmpnameR = nextTempName();
		Relation rell = new Relation(tmpnameR, typeArray, RELATION, 0, typeArray.length, 0, null, env);
		int[] types = rell.toTypes(typeArray);  // domain types
           	int numTuples = r.numtuples*posVal.length;
//System.out.println("num of tuples of destination:"+numTuples);
		rell.data = rell.createDataArray(rell.domains, numTuples);
		if (r.data==null)
		 	r.load();
           	for (int ii = 0; ii <  r.numtuples; ii++) //for each tuple in r
           	{
              	 for (int jj = 0; jj < posVal.length; jj++)	//for each position
		 {
		    for (int ij = 0; ij < r.numattrs ; ij++)
		    {
							//System.arraycopy(r.data[i], 0, rell.data[i], 0, r.numtuples);
						  //(rell.data[ij])[ii]=(r.data[ij])[ii];
			 System.arraycopy(r.data[ij], ii, rell.data[ij], ii*posVal.length+jj, 1);
		    }
		    rell.numtuples = r.numtuples*posVal.length;
		    for (int k = r.numattrs; k<typeArray.length; k++)
		    {
			switch(types[k]) {
			case INTEGER :
			     ((int[])(rell.data[k]))[ii*posVal.length+jj]=posVal[jj].position;	
			     //need to compute the position
//System.out.println("position of "+jj+" is "+ posVal[jj].position);
			     break;
			case ATTRIBUTE :
			     ((String[])(rell.data[k]))[ii*posVal.length+jj]=(rell.domains[j]).name;
			     break;
 			case UNIVERSAL :
//			  ((String[])(rell.data[k]))[i]=(rell.domains[j]).getType() + ":" + ((String[])(rell.data[j]))[i];
			     ((String[])(rell.data[k]))[ii*posVal.length+jj]=(rell.domains[j]).getType() + ":" + posVal[jj].value;
			     break;
			case TYPE :
			     ((String[])(rell.data[k]))[ii*posVal.length+jj]=(rell.domains[j]).getType();
			     break;   }//switch
         	    }
		    if (needParameter)
		    {
		     for (int k = r.numattrs+domarray.length; k<typeArray.length; k++)
		     ((String[])(rell.data[k]))[ii*posVal.length+jj]= posVal[jj].paraValue[k-r.numattrs-domarray.length];	
		    }
                 }//for each postion
		} //for each tuple in r
		rell.save();
		env.put(rell);
		//System.out.println("put this:"+rell.name);
		SimpleNode joinNode=SimpleNode.jjtCreateNode("Join",301,362, null);
		SimpleNode lNode=SimpleNode.jjtCreateNode("Identifier",230, 230, result.name);
		SimpleNode rNode=SimpleNode.jjtCreateNode("Identifier",230, 230, rell.name);
		lNode.setInfo(result.name);
		rNode.setInfo(rell.name);
		lNode.jjtSetParent(joinNode);
		joinNode.jjtAddChild(lNode);
		rNode.jjtSetParent(joinNode);
		joinNode.jjtAddChild(rNode);
		result = evaluateJoin(joinNode, result.name, htable, env);
//result.print();
		if (result.data==null) result.load();
//				 if (result.data==null) System.out.println("result's data is still null");
		result.save();
		env.put(result);//important Sep 16,2004 by Yu
		//System.out.println("put this:"+result.name);
   		env.removeRel(rell.name, true);
	      }//         for (int i = 1; i < outlist.length; i++)
	    }    //end of  if (outlist.length>0)
          }	//end of try
          catch (IOException e) {
                       System.err.println(e);
                  	}

       	 }//end of for (int j =0; j < rel.numattrs; j++)
      deleteFile(tmpname);
      deleteFile(tmpnameR);
      env.removeRel(result.name, true);
      //System.out.println("^^^^^here:result:"+result);
      //result.print();
      if (result.numtuples!=0)
      result.project(null); //duplicate eliminated  By Yu, Sep 20,2004
      return result;
//    }    //end of if (!needParameter)
  //  else	//July 21, 2004 by Yu	take care the case: grep(attr1,pos;x,y) "\xhello\y"
   } //end of else of if ((nd.toString()).equals("GIDList"))
  }

  // evaluate Top Level expression, which results in a relation
  // note that the hashtable is used to provide the initial value for
  // evaluating recursive view, and filename is the data file name for
  // the result of this expression
  public Relation evaluateTLExpression(SimpleNode node,
				       String filename, Hashtable htable, Environment env)
       throws InterpretError
  {//System.out.println("!!!evaluateTLExpression with filename:"+filename);
    int type = node.type, opcode = node.opcode;
    String name = node.name;
   	//start zongyan's addition, July 2002
	if (node.url != null ) {
	aldatpClient.request(node,env);  // need to check response and return it ^_^
	  if (aldatpClient.rel ==null)
		throw new InterpretError( aldatpClient.rspCode);

	  if (aldatpClient.rel.rvc == 17) { // comp
		 Computation c = new Computation(aldatpClient.rel.tree,env);
		 c.name =aldatpClient.rel.name;
		 return c;
	  }
	  else 	// regular relation
		  return aldatpClient.rel;
	}

	else
	// end Zongyan's addition, July 2002

    switch(type) {

    /******* start adding by Fan  *******/
    case OP_RELATION: // added on June 29, 2004 by Fan Guo
        return evaluateRelation(node, filename, htable, env);
    case OP_TRANSPOSENEW: // added on July 30, 2004 by Fan Guo
        return evaluateTranspose(node, filename, htable, env);
    case OP_LITERAL: // added by Fan   // for pr ".2000"; Nov. 19, 2004
        name = (String)node.info;
        /******* end of Fan's addition  *******/

     //Added by Lili Zhu in 2005
    case OP_DISP2D:
    {
       //Global.println("==case OP_DISP2D==");
       return evaluateDisplay2D(node, filename, htable, env);
    }
    //End of case OP_DISP2D

    case OP_EXPSQL : { //KHAYA 2007
		String s = name.substring(1, name.length() - 1);
		try {
			Relation rel = env.lookupRel(".rd", true);
			rel.numtuples = env.reltable.getRDTableSize();
			env.reltable.dump();
			rel.load();
		  String translated = Main.translate(false, s, filename, rel.data);
		  //System.out.println(translated);
		  executeRelixCommand(translated);
		} 
		catch (Exception e) {
			//System.out.print(e.);
		  throw new InterpretError("A problem happened during the SQL translation");
		}
		Relation rl = env.lookupRel(filename, true).duplicate(true);
		executeRelixCommand("dr " + filename + ";");
		rl.name = filename;
		return rl;
	      }
    case OP_EXPSQLFILE : {//KHAYA 2007
  	  //String tmp = nextTempName();
		String s = name.substring(1, name.length() - 1);
		InputStream fin = null;
		try {
			Relation rel = env.lookupRel(".rd", true);
			rel.numtuples = env.reltable.getRDTableSize();
			env.reltable.dump();
			rel.load();
  		  String translated = Main.translate(true, s, filename, rel.data);
		  executeRelixCommand(translated);
		}
		catch (Exception e) {
	  		  throw new InterpretError("A problem happened during the SQL translation");
	  		}
		Relation rl = env.lookupRel(filename, true).duplicate(true);
		executeRelixCommand("dr " + filename + ";");
		rl.name = filename;
		return rl;
	      }
    case OP_IDENTIFIER :
    case OP_EVENTNAME :
      {
	Relation rel = null;
	rel = (Relation) htable.get(name);
	if (rel != null) // there is initial for substitution
	  {
	  return rel;
	  }
	rel = env.lookupRel(name, true);
	//REVISIT:This env lookup should be the first one
	//NREnvironment e2 = (NREnvironment) htable.get("#env2");
	//if (e2 != null){
	//   rel = e2.lookup(name);
	//}

	if (rel == null) {
	 // throw new InterpretError("no such relation : " + name);
		//bug fix by Yi Zheng Sept 28, 2002
		rel = env.lookupRel("." + name, true);
		if (rel == null)
	 		throw new InterpretError("no such relation : " + name);


		//end of bug fix Sept 28, 2002

	}
	// REVISIT : computation

	if (rel.rvc == COMPUTATION) {
	  return rel;
	}else if (rel.rvc == VIEW) {
	  return evaluateView(rel.tree, filename, htable, env);
	}else if (rel.rvc == SCALAR) {            // by Yu
	 String relname = rel.name;
	 Domain vDom = env.lookupDom(rel.name, true);
	 SimpleNode vnode = vDom.tree;
	 SimpleNode fnode = (SimpleNode)vnode.jjtGetChild(0);
	 //SimpleNode snode = (SimpleNode)vnode.jjtGetChild(1);
	 Domain dom[] = new Domain[1];
	 dom[0] = vDom; 
      	 Relation r = new Relation(relname, dom, RELATION,
				1,1,0, null, env);
	 r.data = r.createDataArray(dom, 1);
	 //r.load();
	 switch (vDom.type){
	  case INTEGER:
	  //System.out.println("node:"+vnode.type+";"+vnode.opcode);
	  //System.out.println("fnode:"+fnode.type+";"+fnode.opcode);
	  int si = EvalExpr.evalIntExpr(fnode,env);
	  //System.out.println("si:"+si);
	  ((int[]) r.data[0])[0]=si;
	  break;
	  case SHORT:
	  short ss = EvalExpr.evalShortExpr(fnode,env);
	  ((short[]) r.data[0])[0]=ss;
	  break;
	  case LONG:
	  long sl = EvalExpr.evalLongExpr(fnode,env);
	  ((long[]) r.data[0])[0]=sl;
	  break;
	  case FLOAT:
	  float sf = EvalExpr.evalFloatExpr(fnode,env);
	  ((float[]) r.data[0])[0]=sf;
	  break;
	  case DOUBLE:
	  double sd = EvalExpr.evalDoubleExpr(fnode,env);
	  ((double[]) r.data[0])[0]=sd;
	  break;
	  case BOOLEAN:
	  boolean sb = EvalExpr.evalBooleanExpr(fnode,env);
	  ((boolean[]) r.data[0])[0]=sb;
	  break;
	  case STRING:
	  //System.out.println("node:"+vnode.type+";"+vnode.opcode);
	  //System.out.println("fnode:"+fnode.type+";"+fnode.opcode);
	  //System.out.println("snode:"+snode.type+";"+snode.opcode);
	  String s = EvalExpr.evalStringExpr(fnode,env);
	  ((String[]) r.data[0])[0]=s;
	  //System.out.println("evaluate scalar:"+s);
	  break;
	 }
	 return r;
	//return a relation Nov 15, 2004
	}else // ordinary relation
	  return rel;
      }
    case OP_JOINOPERATOR : {
      return evaluateJoin(node, filename, htable, env);
    }
    case OP_BIGREP : {      // Jiantao 2004
      return evaluateBigrep(node, filename, env);
    }

    case OP_PROJECT : {
      return evaluateProject(node, filename, htable, true, env);
    }

    //YuLing added
    case OP_GEDIT   : {
      return evaluateGedit(node,filename,htable,true,env,null);
    }
    //end of YuLing added

    case OP_TSELECT : { // REVISIT: optomize TSelector in one pass
      // map this node to : projection + selection
      node.type = OP_PROJECT;
      node.opcode = OP_WHENIN;

      return evaluateTLExpression(node, filename, htable, env);
    }
    case OP_SELECT : {
//System.out.println("in OP_SELECT of evaluateTLExpression:" + filename+ " by evaluateSelect");
      return evaluateSelect(node, filename, htable, env);
    }
    case OP_QSELECT : {
//System.out.println("return from evaluateTLExpression:"+filename+" by evaluateQSelect");
      return evaluateQSelect(node, filename, htable, env);
    }
    case OP_PICK : {
      SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
      String tmpname = nextTempName();
      Relation rel = evaluateTLExpression(nd, tmpname, htable, env);
      Relation rr = rel.pick(filename);
      deleteFile(tmpname);
      return rr;
    }
     case OP_ATTRIBSOF : { //case from Andrey
      SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	  String tmpname = nextTempName();
	  Relation rel = evaluateTLExpression(nd, tmpname, htable, env);
	  Relation rr = rel.AttribsOf(filename);
	  deleteFile(tmpname);
	  return rr;

    }
    case OP_GREP: {//May 19,2004 by Yu
//System.out.println("start evaluate Grep:"+filename);
		  return evaluateGrep(node,filename,htable,env);
		         }
    case OP_TEXT2ATTR : { // Jiantao 2004
        return evaluateText2Attr(node, filename, env);
        }
    case OP_MU2NEST: { // Jiantao 2004  
        String parmname = ((SimpleNode)node.jjtGetChild(1)).name;  
        Text t = env.textTable.getText(parmname);
        String tempname;
        if(t == null) // parm is not a text 
        {       
            if(((SimpleNode)node.jjtGetChild(1)).info.getClass().getName().equals("java.lang.String")) // info is a String      
            {       
                 String textStr = (String)((SimpleNode)node.jjtGetChild(1)).info;
                 if(textStr == STRING_DC) // to terminate recursion
                 {       return new Relation(STRING_DC, null, 0, 0, 0, env); // return a null relation
                 } 
                 else // parm is a String used as a text 
                 {       tempname = nextTempName();
                         t = new Text(tempname, textStr, "String", env);
                         return evaluateMu2Nest(node, filename, t, env);    
                 }   
             } 
             else if(((SimpleNode)node.jjtGetChild(1)).info.getClass().getName().equals("java.lang.Long")) // info is a surrogate
             {       
                 long s = ((Long)((SimpleNode)node.jjtGetChild(1)).info).longValue();
                 if(s == 0)
                 {       return new Relation(STRING_DC, null, 0, 0, 0, env); } 
                 Relation r, tempr, rel = null; 
                 if((r = env.lookupRel("." + parmname, true)) == null)
                 {       throw new InterpretError("Relation ."+ parmname + " does not exist.");
                 } 
                 else // assume that .id is the 1st attr, content or .unanalyzed is the 2nd attr               
                 {    
                         if(r.data == null)
                                {       r.load(); 
                                } 
                         for(int i=0;i<r.numtuples;i++) 
                         {       if(((long[])r.data[0])[i] == s)
                                 {       tempname = nextTempName();
                                         t = new Text(tempname,((String[])r.data[1])[i], "String", env); 
                                         tempr = evaluateMu2Nest(node, filename, t, env);

                                         if(rel == null) // the 1st to be assigned
                                         {       rel = tempr; }       
                                         else  
                                         {       rel = rel.mujoin(OP_UJOIN, null, tempr, null, filename); }    
                                  } 
                           } 
                   }
                   return rel;     
              }
              else // not a String nor a surrogate, throw
              {       throw new InterpretError("Text "+ parmname + " does not exist."); }
        }   
        else // parm is a text  
        {       return evaluateMu2Nest(node, filename, t, env);
        }
   }

    case OP_ARRAY :
      //BUGS:
      //  Suppose Array has three attrs A,B,C.
      //  1) R <- Array[1,2,3,4];
      //     This will run, the 4 is of no consequence. This could be
      //     disallowed by looping through all 64 bits of the 'bits'
      //     field in the Simplenode ArrayList below. This loop is
      //     necessary since the bits field is protected.
      //  2) R <- Array[1,,,,,,,,];
      //     This will run. Furthermore, there is no way to fix this
      //     without changing the 'jjt' file and probably the also the
      //     SimpleNode class. Currently, the parser soed not provide me
      //     with the information that there are so many commas in the call.
      //------------------------------------------------------------

      // Array syntax is syntactic sugar for a TSelector.
      // So convert syntax tree of Array to a syntax tree for a TSelector
      {  //Each domain contributes either to the projection list (ProjList)
	//or to the selection clause (compareNodes)
	SimpleNode ProjList = new SimpleNode("ExpressionList");
	ProjList.set(OP_EXPRLIST, OP_EXPRLIST);
	Vector compareNodes = new Vector();
	//Dissect parts of the actual array syntax tree structure
	//Identifier is the relation that array operator is applied to
	SimpleNode Identifier =
	  ((SimpleNode) node.jjtGetChild(0)).jjtDeepDuplicate();
	SimpleNode ArrayList = (SimpleNode) node.jjtGetChild(1);
	Relation r = env.lookupRel(Identifier.name, true);

	if (r==null)
	  {  throw new InterpretError("No such relation "+Identifier.name);
	  }
	//REVISIT: If r is a view, it must be evaluated first
	//--done March 19, 2002 Yi Zheng added the following
	 if (r.rvc == VIEW)
	 {
	   String tmpname = nextTempName();
	   r = evaluateView(r.tree, tmpname, htable, env);
	   deleteFile(tmpname);
	 }


	int ALindex=0; //index into the children of ArrayList
	for(int i=0; i<r.numattrs; i++)
	  {  if (ArrayList.isBitSet(i))
	    {  //becomes selection attribute
	      //Create this tree structure:
	      //    Compare:300:323:null:0
	      //     Identifier:230:230:I1:0
	      //     Literal:470:453:null:0    <- could be an expression etc.
	      SimpleNode Compare = new SimpleNode("Compare");
	      Compare.set(OP_BIOPERATOR, OP_EQ);
	      SimpleNode id = new SimpleNode("Identifier");
	      id.set(OP_IDENTIFIER, OP_IDENTIFIER,r.domains[i].name);
	      id.jjtSetParent(Compare); //link up Compare and id
	      Compare.jjtAddChild(id);  //link up Compare and id
	      //Graft the node from ArrayList onto new tree structure
	      SimpleNode nd = ((SimpleNode) ArrayList.jjtGetChild(ALindex))
		.jjtDeepDuplicate();
	      nd.jjtSetParent(Compare); //link up Compare and nd
	      Compare.jjtAddChild(nd);  //link up Compare and nd
	      ALindex++;
	      //Store this new 'compare' tree
	      compareNodes.addElement(Compare);
	    }
	  else
	    {  //becomes projection attribute
	      //Add the corresponding domain to the projection list
	      SimpleNode id = new SimpleNode("Identifier");
	      id.set(OP_IDENTIFIER, OP_IDENTIFIER,r.domains[i].name);
	      id.jjtSetParent(ProjList); //link up ProjList and id
	      ProjList.jjtAddChild(id);  //link up ProjList and id
	    }
	  }
	//Now take all compare nodes and 'AND' them together in pairs
	int numElements = compareNodes.size();
	//top points to the top of the synax tree as it is constructed
	SimpleNode top = null;
	if (numElements > 0)
	  {  top = (SimpleNode) compareNodes.lastElement();
	  compareNodes.removeElementAt(numElements-1);
	  numElements--;
	  for(; numElements>0; numElements--)
	    {  SimpleNode And = new SimpleNode("And");
	    And.set(OP_BIOPERATOR, OP_AND);
	    top.jjtSetParent(And); //link up And and top
	    And.jjtAddChild(top);  //link up And and top
	    SimpleNode Compare = (SimpleNode) compareNodes.lastElement();
	    compareNodes.removeElementAt(numElements-1);
	    Compare.jjtSetParent(And); //link up And and Compare
	    And.jjtAddChild(Compare);  //link up And and Compare
	    top = And;
	    }
	  }
	//Finally put everything together into a syntax tree
	if (ProjList.jjtGetNumChildren() == 0 && top == null)
	  {  //REVISIT: Impossible or view
	  }
	//else if(top == null) modified by Yi Zheng April 8, 2002 to fix COMPUTATION bug
	else if(top == null && r.rvc != COMPUTATION)
	  {  //Array is like a simple Project clause, not a TSelect
	    //Create following structure:
	    // Project:421:427:null:0
	    //  Projector:420:420:null:2
	    //   ExpressionList:592:592:null:0
	    //    Identifier:230:230:I1:0
	    //    Identifier:230:230:I2:0
	    //  Identifier:230:230:Array:0
	    SimpleNode Project = new SimpleNode("Project");
	    Project.set(OP_PROJECT, OP_WHENIN);
	    SimpleNode Projector = new SimpleNode("Projector");
	    Projector.set(OP_PROJECTOR, OP_PROJECTOR);
	    Projector.setBit(1);
	    Projector.jjtSetParent(Project); //link up Project and Projector
	    Project.jjtAddChild(Projector);  //link up Project and Projector
	    ProjList.jjtSetParent(Projector);//link up Projector and ProjList
	    Projector.jjtAddChild(ProjList); //link up Projector and ProjList
	    Identifier.jjtSetParent(Project); //link up Project and Identifier
	    Project.jjtAddChild(Identifier);  //link up Project and IDentifier
	    return evaluateTLExpression(Project, filename, htable, env);
	  }
	  //this else case added by Yi Zheng for cases like aComp[];
	  else if(top == null && r.rvc == COMPUTATION)
	  {//Array class is like a TSelect where Selector contains the literal TRUE
	  	SimpleNode TSelect = new SimpleNode("TSelect");
	  	TSelect.set(OP_TSELECT, OP_TSELECT);
	  	SimpleNode Projector = new SimpleNode("Projector");
	  	Projector.set(OP_PROJECTOR, OP_PROJECTOR);
	  	Projector.jjtSetParent(TSelect); //link up TSelect and Projector
	    TSelect.jjtAddChild(Projector);  //link up TSelect and Projector
	    if(ProjList.jjtGetNumChildren() > 0)

	      { Projector.setBit(1);
	      ProjList.jjtSetParent(Projector);//link up Projector and ProjList
	      Projector.jjtAddChild(ProjList); //link up Projector and ProjList
	      }
	    SimpleNode Selector = new SimpleNode("Selector");
	    Selector.set(OP_SELECT, OP_WHEREIN);
	    Selector.jjtSetParent(TSelect); //link up TSelect and Selector
	    TSelect.jjtAddChild(Selector);  //link up TSelect and Selector
	    SimpleNode ctop = new SimpleNode("Literal");
	    ctop.set(OP_LITERAL, OP_BOOLEAN);
	    ctop.setObject(new Byte((byte) 1));
	    ctop.jjtSetParent(Selector);
	    Selector.jjtAddChild(ctop);
	    Identifier.jjtSetParent(Selector);//link up Selector and Identifier
	    Selector.jjtAddChild(Identifier); //link up Selector and IDentifier

	    return evaluateTLExpression(TSelect, filename, htable, env);
	  }
	else
	  {  //Array call is like a TSelect
	    //Create following structure:
	    // TSelect:422:422:null:0
	    //  Projector:420:420:null:2
	    //   ExpressionList:592:592:null:0 <- (This is stored in ProjList)
	    //    Identifier:230:230:I3:0
	    //  Selector:423:425:null:0
	    //   Compare:300:323:null:0        -|
	    //     Identifier:230:230:I1:0      | (This portion is in 'top'
	    //     Literal:470:453:null:0      -|
	    //   Identifier:230:230:Array:0
	    SimpleNode TSelect = new SimpleNode("TSelect");
	    TSelect.set(OP_TSELECT, OP_TSELECT);
	    SimpleNode Projector = new SimpleNode("Projector");
	    Projector.set(OP_PROJECTOR, OP_PROJECTOR);
	    Projector.jjtSetParent(TSelect); //link up TSelect and Projector
	    TSelect.jjtAddChild(Projector);  //link up TSelect and Projector
	    if(ProjList.jjtGetNumChildren() > 0)
	      { Projector.setBit(1);
	      ProjList.jjtSetParent(Projector);//link up Projector and ProjList
	      Projector.jjtAddChild(ProjList); //link up Projector and ProjList
	      }
	    SimpleNode Selector = new SimpleNode("Selector");
	    Selector.set(OP_SELECT, OP_WHEREIN);
	    Selector.jjtSetParent(TSelect); //link up TSelect and Selector
	    TSelect.jjtAddChild(Selector);  //link up TSelect and Selector
	    top.jjtSetParent(Selector); //link up Selector and top
	    Selector.jjtAddChild(top);  //link up Selector and top
	    Identifier.jjtSetParent(Selector);//link up Selector and Identifier
	    Selector.jjtAddChild(Identifier); //link up Selector and IDentifier
	    return evaluateTLExpression(TSelect, filename, htable, env);
	  }
      }
    case OP_EDIT :
    case OP_ZORDER :
    case OP_EVAL : // ???

    case OP_RENAME : { // ???
      throw new InterpretError("TO BE IMPLEMENTED");
    }
    // many more other case ...

    // This case is added by hongyu (June 25, 01)
    // a <- [ ITEM, COMP] quant ( #  = 2) COMP , ( % > 0.1) DEPT where VOL1 > 2 in SUPPLY;
    case OP_BIOPERATOR : {
      // REVISIT: optomize TSelector in one pass
      // map this node to : projection + selection
      node.type = OP_PROJECT;
      node.opcode = OP_WHENIN;
      // System.out.println("here " + node.type + " " + node.opcode);
      return evaluateTLExpression(node, filename, htable, env);
    }
    case OP_FUNCTION:
    	throw new InterpretError("OP_FUNCTION is not implemented yet");
    default : {
      throw new
	InterpretError("invalid/not implemented relational expression"+type);
    }
    }
  }
  // evaluate top level view and return the result relation
  /**
   * Given node is the tree of a view named node.name, evalutate it
   * and return the resulting relation.
   * @param node is the tree of the declared view.
   * @param filename is the name of the result relation.
   * @param htable is the null hash table.
   * @param env is the environment.
   *
   *         node
   *        /  |  \
   *       /   |   \
   *      /    |    \
   *    bit(0)    bit(1)  jjt optional bits
   *   [nodei] nd [nodet]
   * nd is the join expression.
   *
   * jjtGetChild offset #:
   *           0
   *      0    1
   *      0    1     2
   *      0    1     2
   */
  private Relation evaluateView(SimpleNode node, String filename,
	 Hashtable htable, Environment env)
       throws InterpretError
  {
    String name = node.name; // view name
   	String url = node.url;   // Zongyan   Aug 2002
    SimpleNode nodei = null, nodet = null, nd = null;
    Relation rel_init = null; // initial relation for recursive view
    Relation rel_target = null; // target relation for recursive view
    String initname = nextTempName(); // data file for rel_init
    String targetname = nextTempName(); // data file for rel_target

    if (node.isBitSet(0)) { // there is initial part
      nodei = (SimpleNode) node.jjtGetChild(0);
      nd = (SimpleNode) node.jjtGetChild(1); // body of view
      if (node.isBitSet(1)) // there is target part
	nodet = (SimpleNode) node.jjtGetChild(2);
    } else {
      nd = (SimpleNode) node.jjtGetChild(0); // body of view
      if (node.isBitSet(1)) // there is target part
     nodet = (SimpleNode) node.jjtGetChild(1);
    }

    if (isRecursiveView(nd, name, new Hashtable(), env)) {
      // for recursive view


      // find start relation for recursion
      if (nodei != null) { // there is initial part
	// the initial part (if any) should not refer to itself
	if (isRecursiveView(nodei, name, new Hashtable(), env))
		throw new InterpretError("recursive view " + name
					 + " refers to itself in initial expression");
	rel_init = evaluateTLExpression(nodei, initname, htable, env);
      } else { // no initial part, see relational part under this name
	Relation rtmp = env.lookupRel(name, true);
	if (rtmp == null || rtmp.numattrs == 0) { // no relational part
	  throw new InterpretError("nowhere to start in recursive view");
	} else { // use relational part as initial
	  rel_init = rtmp.duplicate(false);
	}
      }

      // find if there is a target
      if (nodet != null) { // there is target part
	// the target part (if any) should not refer to itself
	if (isRecursiveView(nodet, name, new Hashtable(), env))
	  throw new InterpretError("recursive view " + name
				   + " refers to itself in target expression");
	rel_target = evaluateTLExpression(nodet, targetname, htable, env);
      }

      // check if rel_init == rel_target
      if (rel_target != null && equals(rel_init, rel_target)) {
	if (nodei == null) {
	  deleteFile(targetname);
	  return rel_init;
	} else {
	  Relation rel = rel_init;
	  deleteFile(initname);
	  deleteFile(targetname);
	  return rel;
	}
      }

      int i = 0;
      while (true) { // loop for recursion until match target or unchanged
	// add (name, rel_init) pair to hash table to modify environment
	htable.put(name, rel_init);

	Relation rel = evaluateTLExpression(nd, filename, htable, env);
        /******* start adding by Fan  *******/
        // on July 28, 2004
        // for safe sake to sort the rel_init and/or rel_target before comparision to avoid infinitive loop;
        if (rel_init!=null)     rel_init.numsortattrs = 0;
        if (rel_target !=null)  rel_target.numsortattrs =0;
        /******* end of Fan's addition  *******/

	if (rel_target != null && equals(rel, rel_target)) {
	  if (nodei != null) deleteFile(initname);
	  deleteFile(targetname);
	  htable.remove(name);
	  return rel;
	} else if (rel_target == null && equals(rel, rel_init)) {
	  if (nodei != null) deleteFile(initname);
	  htable.remove(name);
	  return rel;
	} else { // loop again
	  rel_init = rel;
		 }
      }
    } else { // for non recursive view, simply evaluate the expression
      //System.out.println("evaluating view : " + name);
      return evaluateTLExpression(nd, filename, htable, env);
    }
  }
    // modified by Zongyan -- Aug 2002

//Start Lili 2005
 private void executeDisplay2D(SimpleNode node, Environment env)
       throws InterpretError
  {

    int opcode = node.opcode;
    int type = node.type;  
    if(opcode==OP_DISP2D) ;
    // Global.println("here, coming soon");
  }
//End Lili 2005

  private void executeAssignment(SimpleNode node, Environment env)
       throws InterpretError
  {
    //start zhan's addition
      node = modifyNodeForKSTAR(node,env);
    //end of zhan's addition Oct 2003


    //printNode(node);
    int opcode = node.opcode;
    int type = node.type;

    switch(opcode)
    {case OP_ASSIGN :
     case OP_ASSIGNADD :
      Domain[] domsr = null, domsl = null;
      Relation rr = null, rl = null;
      String lname = ((SimpleNode) node.jjtGetChild(0)).name;	//x<-10; y<-x;
      //System.out.println( "ExecuteAssignment: "+lname );
      // data file name for right expression
      String tmpname = nextTempName();
      if ( type == OP_ASSIGNMENT)
	 rl = env.lookupRel(lname, true);

      if (node.jjtGetNumChildren() == 2 )
      { 
        //from here added by Yu
	if (rl!=null && rl.rvc != COMPUTATION && rl.rvc !=RELATION && rl.rvc !=VIEW)   //Added by Yu, June 14,2004
	{
	 //check the virtual domain to see if type match
	 Domain vDom = env.lookupDom(lname, true);
	 if (vDom==null) throw new InterpretError(lname+"has not been declared in domain table");
	 if (vDom.tree!=null)
	 {
    		 SimpleNode vn = SimpleNode.jjtCreateNode("Declaration",OP_DECLARATION,OP_LET,null);
	  	 SimpleNode in = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,lname);
		 SimpleNode cn = (SimpleNode)node.jjtGetChild(1);     //new value
	  	 in.jjtSetParent(vn);
	  	 (vn).jjtAddChild(in);
	  	 cn.jjtSetParent(vn);
	  	 (vn).jjtAddChild(cn);
         //need to change the value in domain table Aug 31, 2004 by Yu
		 AddVirtDom2DomTable(vn, env);
	 }
	 else
		  throw new InterpretError("something wrong! no tree found for this domain--should not occur");
	} //if (rl!=null && rl.rvc != COMPUTATION && rl.rvc !=RELATION && rl.rvc !=VIEW) 
	else	//end here added by Yu
		rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1),
				    tmpname, new Hashtable(), env);
      }//if (node.jjtGetNumChildren() == 2 )
      else
      {


	  // node.jjtGetNumChildren() == 4
	  // get the domain array for the left IDList
	  SimpleNode nodel = (SimpleNode) node.jjtGetChild(1);
	  domsl = IDListToDomains(nodel, env);

	  // get the domain array for the right ExpressionList
	  SimpleNode noder = (SimpleNode) node.jjtGetChild(2);
	  domsr = ExpressionListToDomains(noder);

	  rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(3),
				    tmpname, new Hashtable(), env);
	}//end of else if (node.jjtGetNumChildren() == 2 )
      // for case :
      // R <- R; or R <+ R; do nothing
      if (rl != null && rl == rr && domsl == null && domsr == null)
	{  return; // for R <- R; do nothing
	}
      // if domsr is not in relation rr, actualization is needed.
      if (domsr!=null && !rr.isDomainsIn(domsr)) //add domsr!null by Yu, Aug 31, 2004
	{

	//modified by Yi Zheng March 23, 2002
	//have to cleanup actualizer in the end
	Actualizer act = new Actualizer(domsr, rr, env);
	Relation rtmp = act.actualize();
	// can't use rr.name, since this may overwrite the
	// rr if it is ID rather than an expression
	rtmp.name = tmpname;
	rtmp.save(); rtmp.unload();
	rr = rtmp;
	}
      if (rl!=null && rl.rvc!=SCALAR||rl==null)
      {
		  Relation rel = null; //Zongyan's modification
	  String mytmp = nextTempName(); //Zongyan's modification

      if (opcode == OP_ASSIGN)
	{  // REVISIT : better to assign first then delete
	  // if left side relation exists, delete it
	  	  if ( type == OP_ASSIGNMENT) {//added by Zongyan

	  if (rl != null )
	    {  //Only remove rl if it is in the most deeply
	      //nested frame of the environment
	      env.removeRel(lname, false);
	    }
	  env.put(rr.assign(domsr, lname, domsl));
	  	  }
	  else if ( type == ALDATP_ASSIGNMENT) {

		  rel = rr.assign(domsr, mytmp, domsl);
		 env.put(rel);

		 distributedStmt(rel.name,lname,node.url,OP_ASSIGN,env);
		 env.removeRel(rel.name, false);
	  }
	}
      else
	{  // OP_ASSIGNADD

		   if ( type == OP_ASSIGNMENT) {

	  if (rl == null)
	    {  Global.println("Warning: cannot find relation " + lname);
	    env.put(rr.assign(domsr, lname, domsl));
	    }
	  else
	    {  // left relation does not exist
	      rr.assignAdd(domsr, rl, domsl);
	    }
	    	   } else  if ( type == ALDATP_ASSIGNMENT) {
		   rel = rr.assign(domsr, mytmp, domsl);
		   env.put(rel);
		   distributedStmt(rel.name,lname,node.url,OP_ASSIGNADD, env);
		 	 env.removeRel(rel.name, false);


	  }
	}

      //delete the temporary data file

      deleteFile(tmpname);
      }//if (rl.rvc!=SCALAR)
      break;

      default : // impossible to here
	System.out.println("unknown assignment statement!!!");
	break;
      }
  }



/******** searchStar from zhan Sept 2003 ***********/
private void searchSTAR(SimpleNode node)
{
  SimpleNode tmpNode ;
  for(int i = 0; i < node.jjtGetNumChildren(); i++)
  {
	 tmpNode = (SimpleNode)node.jjtGetChild(i);
	 if(tmpNode.opcode == OP_IDENTIFIER && tmpNode.name.indexOf("*") != -1)
	 {
		isKStar = "*";
        //seperate the case R/(N)*/A and R/*/A or R/.*/A
        if(!tmpNode.name.equals("*") && !tmpNode.name.equals(".*"))
       		KStarName = tmpNode.name.substring(1,tmpNode.name.indexOf("*")-1);
       		//R/*/A or R/.*/A
		else if(tmpNode.name.equals("*") ||tmpNode.name.equals(".*"))
		    KStarName = "*";
		break;
	 }
	 else if(tmpNode.opcode == OP_IDENTIFIER && tmpNode.name.indexOf("+") != -1)
	 {
		isKStar = "+";
		//seperate the case R/(N)+/A and R/+/A or R/.+/A
	    if(!tmpNode.name.equals("+") && !tmpNode.name.equals(".+"))
		    KStarName = tmpNode.name.substring(1,tmpNode.name.indexOf("+")-1);
		else if(tmpNode.name.equals("+") ||tmpNode.name.equals(".+"))
		    KStarName = "+";

		break;
	 }
	 else if(tmpNode.opcode==OP_IDENTIFIER && tmpNode.name.indexOf("?")!=-1)
	 {
		isKStar="?";
		KStarName = tmpNode.name.substring(1,tmpNode.name.indexOf("?")-1);
		break;
	 }
	 else if(tmpNode.opcode==OP_IDENTIFIER && tmpNode.name.indexOf("(./")!=-1)
	 {
		isKStar="./";
		break;
	 }

	 else
	   searchSTAR(tmpNode);
   }
}


/******** modify the tree for kleeneStar ********/ //zhan Sept 2003
private SimpleNode modifyNodeForKSTAR(SimpleNode node,Environment env)
  throws InterpretError
{
  //firstly check whether there is kleeneStar or QuestionMark in the tree
   searchSTAR(node);

   String nameOfNode = KStarName;

//System.out.println("isKStar == "+ isKStar);
//System.out.println("KstarName == "+KStarName);

 if (isKStar != null)
 {
  //R/(M/N/)?A  or R/(././)?A
  if(isKStar.equals("?"))
  {
	node = modifyNodeForQuestionMarker(node,env);
  }
  //R/(././)A
  else if(isKStar.equals("./"))
    node = modifyNodeWildCard(node,env);

  //case R/*/cname or R/.*/cname
  else if(KStarName.equals("*") || KStarName.equals("+"))
  {
    //System.out.println("overall case -----------------------");
    Vector allDomsA = new Vector();
	Vector hat = new Vector();
	SimpleNode nodeRight= (SimpleNode) node.jjtGetChild(1);
	SimpleNode newNodeRight = nestDomsA(nodeRight,allDomsA,hat,env,true,KStarName);
	nodeRight.jjtRemoveChild(0);
	nodeRight.jjtRemoveChild(0);
	nodeRight.jjtReplace(newNodeRight);

   //System.out.println("...........................");
   //Global global = new Global();
   //global.traverse(nodeRight);

  }
   //case (N)* -> kleeneStarQueries with recursive relation
   else
  {
   //System.out.println("there is a kstar in the tree");

   Relation recursiveRelation = env.lookupRel("."+KStarName, true);
   RelTable relt = env.reltable;
   Hashtable reltable = relt.getRelTable();
   int recur_level=0;

   //check whether there is the relation and find out the NR level
   if(reltable.containsKey("." + KStarName) == true)
   {
	  while(true)
	  {
		if(reltable.containsKey("."+KStarName+"_"+recur_level)==true)
		  recur_level=recur_level+1;
		else
		  break;
	  }
	}
   else
   {
     KStarName = null;
     isKStar = null;
	 throw new InterpretError("Relation "+ "." + KStarName +" does not exist");
   }

//    System.out.println(" level ======="+recur_level);

    //check whether the relation is recursive
	boolean oneLevelRecur = false; //whether it has only one level

	if(recur_level==0)
	{
	   for(int j=0;j<recursiveRelation.domains.length;j++)
	   {
		  if(recursiveRelation.domains[j].name.equals(KStarName + "_0"))
		  { oneLevelRecur = true;
			break;
		  }
	   }
	   if(oneLevelRecur == false)
	   {
	      KStarName = null;
          isKStar = null;
	      throw new InterpretError(KStarName + " is not recursive, so * should not be used in the query");
       }
    }


    //if it is recursive true with one level or more levels
	if(recur_level>0 || (recur_level==0 && oneLevelRecur==true)) //recursive case
	{
       SimpleNode nodeRight=(SimpleNode)node.jjtGetChild(1); //node is "Assignment"

       //case 1 : R/(N)*  and R/(N)*/A
       // step1
       //         Assignment
       //           /      \
       //    Identifier  project(nodeRight)
       //                  /       \
       //           projector     Identifier(       : R)
       //               /
       //    ExpressionList(nodeExp)
       //            /
       //        Vertical
       //          /
       //      Identifier(nodeID:(N)* )  R/(N)*
       //    [    /
       //      Identifier(         A) ]  R/(N)*/A


       if(nodeRight.type == OP_PROJECT)  //there is star in projection
       {
		 SimpleNode nodeRightL= (SimpleNode)nodeRight.jjtGetChild(0);
		 SimpleNode nodeRightR= (SimpleNode)nodeRight.jjtGetChild(1);
         if(nodeRightR.opcode == OP_IDENTIFIER)
		 {
            if(nodeRightL.type == OP_PROJECTOR)
            {
			  SimpleNode id=(SimpleNode)nodeRightL.jjtGetChild(0).jjtGetChild(0).
			                jjtGetChild(0);

			  if(id.type == OP_IDENTIFIER &&(id.name.indexOf("*")!=-1
			     || id.name.indexOf("+")!=-1) )
			  {
				  //case R/(N)* R/(N)*/A  OR R/(N)+ R/(N)+/A
				  //Node: only R/(N)* and (N)* in R should project
				  // the domains in R, that are not same to N
				  //R(X,Y,N) N(A,B)
				  // R/(N)* and (N)* in R => [X,Y,A,B]
				  // the case R/(N)*/A,(N)*/A in R and [A] in R(N)*
				  // and [B] where A = 1 in R(N)*
				  // do not have to project [X,Y]

				   //use starInEnd to seperate the case (N)* and (N)*/A
				   boolean starInEnd = false;
				   if(id.jjtGetNumChildren()==0)
				      starInEnd = true;

				   SimpleNode newNodeRight = KStarInProjection(nodeRight,recur_level,
				           recursiveRelation,isKStar,env,starInEnd);
                   node.jjtRemoveChild(1);
                   newNodeRight.jjtSetParent(node);
                   node.jjtAddChild(newNodeRight);
			  }
              else if(id.type == OP_PROJECT)
              {
	               //case P/R/(N)*/A: modify to (N)*/A in P/R;
				   KStarName = null;
				   isKStar = null;
		           throw new InterpretError ("To be implemented");

		      }

		    }

		    else if(nodeRightL.type == OP_IDENTIFIER)
		    {
				//case (N)*/A in R
			    //          Assignment
                //           /      \
                //    Identifier  project(nodeRight)
                //                  /           \
                //           Identifier         Identifier(       : R)
                //           (nodeID:(N)* )->(N)* in R
                //           [    /

                //             Identifier(         A) ]  (N)*/A in R

                //use starChildOfProject to handle (N)*/A in R
                nodeRight = starChildOfProject(nodeRight);

                //modify the tree same as R/(N)* or R/(R)*/A
                SimpleNode id=(SimpleNode)nodeRight.jjtGetChild(0).jjtGetChild(0).
                             jjtGetChild(0).jjtGetChild(0);

                boolean starInEnd = false;
				if(id.jjtGetNumChildren()==0)
				    starInEnd = true;


                SimpleNode newNodeRight = KStarInProjection(nodeRight,recur_level,
                     recursiveRelation,isKStar,env,starInEnd);
				node.jjtRemoveChild(1);
				newNodeRight.jjtSetParent(node);
                node.jjtAddChild(newNodeRight);

			 }//end of else if

			 //case  M/(N)*/A in R
			 // modify the tree to the tree of (N)*/A in R/M
			else if(nodeRightL.type == OP_PROJECT)

			{
			   SimpleNode projectDup = (SimpleNode)nodeRightL.jjtDeepDuplicate();
			   String idName = ((SimpleNode)projectDup.jjtGetChild(1)).name;
			   ((SimpleNode)projectDup.jjtGetChild(1)).name=nodeRightR.name;

               //if L/...../M/(N)*/A in R, should do further
			   String tmpName = idName;
			   String newId = null;
			   SimpleNode tmp = (SimpleNode)projectDup.jjtGetChild(0);
			   SimpleNode tmpChild = null;
			   SimpleNode idDup = null;


			   while(true)
			   {
				 tmpChild = (SimpleNode)tmp.jjtGetChild(0);
				 if(tmpChild.type == OP_PROJECT)
				 {
                   newId = ((SimpleNode)tmpChild.jjtGetChild(1)).name;
				   ((SimpleNode)tmpChild.jjtGetChild(1)).name = tmpName;
				    tmpName = newId;
				   }

				  if(tmpChild.type == OP_IDENTIFIER)
				  {
					idDup = (SimpleNode)tmpChild.jjtDeepDuplicate();
					tmpChild.jjtRemoveChild(0);
					tmpChild.name = tmpName;
					break;
				  }

				  tmp = tmpChild;

    		   }//end while

               nodeRightL.jjtRemoveChild(0);
               nodeRightL.jjtRemoveChild(0);

               nodeRightL.jjtReplace(idDup);
               nodeRightR.jjtReplace(projectDup);

               //modify the tree as (N)*/A in R/M
               nodeRight = starChildOfProject(nodeRight);
               SimpleNode id=(SimpleNode)nodeRight.jjtGetChild(0).jjtGetChild(0).
			                                jjtGetChild(0).jjtGetChild(0);

			   boolean starInEnd = false;
			   if(id.jjtGetNumChildren()==0)
				    starInEnd = true;

               SimpleNode newNodeRight = KStarInProjection(nodeRight,recur_level,recursiveRelation,isKStar,env,starInEnd);
			   node.jjtRemoveChild(1);
			   newNodeRight.jjtSetParent(node);
               node.jjtAddChild(newNodeRight);

       		 }//end else if nodeRightL.type == OP_PROJECT
         }

         //case [A] in R/(N)* or [A] in R/(N)+;
         if(nodeRightR.type == OP_PROJECT)
    	 {
           SimpleNode id = (SimpleNode)nodeRightR.jjtGetChild(0).jjtGetChild(0).
                          jjtGetChild(0).jjtGetChild(0);

           //case [A] in R/(N)* or [A] in R/(N)+;
           if(id.name.indexOf("*")!=-1 ||id.name.indexOf("+")!=-1 )
     	   {

			SimpleNode newNodeRight = KStarInProjection(nodeRightR,recur_level,recursiveRelation,isKStar,env,false);
            nodeRight.jjtRemoveChild(1);
            newNodeRight.jjtSetParent(nodeRight);
            nodeRight.jjtAddChild(newNodeRight);
		   }

		   else
		   //Consider : Case (N)*/A in R/M or (N)*/A in P/M/R
		   {
              KStarName = null;
			  isKStar = null;
			  throw new InterpretError ("To be implemented");
	 	   }
		   // else if : the case [A] in R/M/(N)* should be handled
	    }

       } //end if("Project")

 //CASE2: selection has kleeneStar
 else if(nodeRight.opcode==OP_TSELECT )  //"TSelect" there is star in selection
 {

  //System.out.println("selection has star*************************");

  SimpleNode nodeProU = (SimpleNode) nodeRight.jjtGetChild(0);
  SimpleNode nodeIn =
	 (SimpleNode) nodeRight.jjtGetChild(1).jjtGetChild(1);


  if(nodeIn.type == OP_IDENTIFIER)
  {
	//the original tree is
	//           Assignment
	//            /      \
	//       Identifier TSelect(nodeRight)
	//                   /          \
	//Identifier((N)*)|Projector   Selector
	//                 /           /     \
	//    ExpressionList     Compare     Identifier(    :R)
	//        /               /     \
	// Identifier      Identifier  Literal
	//     ( B)        ((N)*)      (value)
	//                 Identifier
	//                 (A)



    //case1: [B] where (N)*/A=value in R;
    SimpleNode starS = (SimpleNode) nodeRight.jjtGetChild(1).jjtGetChild(0).jjtGetChild(0);
    if(nodeProU.type == OP_PROJECTOR && (starS.name.indexOf("*")!=-1||starS.name.indexOf("+")!=-1))
    {
      //System.out.println("there is only one star in here");

      //get name of "Identifier"(B)
      SimpleNode nodePro =
	     (SimpleNode) nodeRight.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);


      //check whether the domain with the name is in the relation(B is in R?)
      Relation proRelation = env.lookupRel(nodeIn.name, true);
      boolean whetherIn = false;
      for(int j=0;j<proRelation.domains.length;j++)
	  {
	    if(proRelation.domains[j].name.equals(nodePro.name))
		{ whetherIn = true;
		   break;
		}
	  }

      if(whetherIn == true)  //the relation has the domain
      {
		SimpleNode topTSelNode = KStarInSelection(nodeRight,recur_level,-1,isKStar,env);

	    //connect to the original tree
        node.jjtRemoveChild(1);
	    topTSelNode.jjtSetParent(node);
	    (node).jjtAddChild(topTSelNode);

	  }
      else
      {
		KStarName = null;
        isKStar = null;
	    throw new InterpretError
	        ("There is no domain "+nodePro.name+" in relation "+nodeIn.name);
	  }

    } //end if (nodePro.name.opcode==... ...)

    //case2: (N)*/B where (N)*/A = value in R;
    else if(nodeProU.type == OP_IDENTIFIER && nodeProU.name.indexOf("*")!=-1)
    {
     //"(N)*/B where (N)*/A = value in R" should be transferred as
     // [B where A=value in R] (if R has A,B)  ujoin
     // N/B where (N)*/A=value in R  ujoin
     // N/N_0/B where (N)*/A=value in R  ujoin ... ...ujoin
     // N/N_0/N_recur_level-1/B where (N)*/A=value in R;

     //handle the case that there is KStar in projection
	 Vector selections = new Vector();
     SimpleNode nodeRightDup = (SimpleNode)nodeRight.jjtDeepDuplicate();

    //change the tree

    //    TSelect(nodeRight)                TSelect(nodeRight)
	//       /           \                   /          \
	//Identifier((N)*)   ...... ======>> Projector   ......
	//    /                                /
	// Identifier                     ExpressionList
	// ( B)                              /
	//                               Identifier
	//                                  ( B)

   	 SimpleNode projectorNode = SimpleNode.jjtCreateNode("Projector",420,420,null);
	 projectorNode.setBit(1);
	 SimpleNode expListNode = SimpleNode.jjtCreateNode("ExpressionList",592,592,null);

     ((SimpleNode)nodeProU.jjtGetChild(0)).jjtSetParent(expListNode);
     expListNode.jjtAddChild((SimpleNode)nodeProU.jjtGetChild(0));
     expListNode.jjtSetParent(projectorNode);
     projectorNode.jjtAddChild(expListNode);

     ((SimpleNode)nodeRightDup.jjtGetChild(0)).jjtReplace(projectorNode);

     SimpleNode allTogether = twoStars(nodeRightDup,recur_level,isKStar,env);

	 node.jjtRemoveChild(1);
	 allTogether.jjtSetParent(node);
	 (node).jjtAddChild(allTogether);

    }//end else if(nodePro... ...)
    else
    { KStarName = null;
	  isKStar = null;
      throw new InterpretError("To be implemented");
    }
  }//end if(nodeIn.type==OP_IDENTIFIER)


  else if(nodeIn.type == OP_PROJECT)
  {
	 //Query:  [B] where A=value in R/(N)*;
	 //or B] where A=value in R/(N)+;
	 SimpleNode id=(SimpleNode)nodeIn.jjtGetChild(0).jjtGetChild(0).
	                       jjtGetChild(0).jjtGetChild(0);
	 //the original tree is
	 //           Assignment
	 //            /      \
	 //       Identifier TSelect(nodeRight)
	 //                   /    \
	 //             Projector   Selector
	 //                 /       /        \
	 //    ExpressionList  Compare         Project(nodeIn)
	 //        /            /     \           /          \
	 //Identifier( B)  Identifier  Literal    Projector   ID
	 //               [((N)*)/(N)+ ]    (value)     /
	 //                  Identifier           ExpList
	 //                  (A)	                   /
     //                                       Vertical
     //                                         /
     //                                        ID(id node) (N)* or (N)+

     if(id.type == OP_IDENTIFIER && (id.name.indexOf("*")!=-1 ||id.name.indexOf("+")!=-1))
     {
        //System.out.println("* and +");
        SimpleNode idStarNode = (SimpleNode)nodeRight.jjtGetChild(1).
                  jjtGetChild(0).jjtGetChild(0);
		if(idStarNode.name.indexOf("*")==-1 && idStarNode.name.indexOf("+")==-1)
		{
	      SimpleNode newNodeIn = KStarInProjection(nodeIn,recur_level,recursiveRelation,isKStar,env,false);
          nodeIn.jjtReplace(newNodeIn);
	    }

	    //[B] where (N)*/A=value in R/(N)*;
	    //[B] where (N)*/A=value in R/(N)+;
	    //[B] where (N)+/A=value in R/(N)*;
	    else
	    {
      //System.out.println("there is 2 star in selection and projection list");

	 //             Projector   Selector    |                   |
	 //                 /       /        \  |                   |
	 //    ExpressionList  Compare          |  Project(nodeIn)  | ID
	 //        /            /     \         |  /          \     |
	 //Identifier( B)  Identifier  Literal  | Projector   ID    |
	 //                  [((N)*) ]   (value)|    /             ==>>
	 //                  Identifier         |  ExpList          |
	 //                  (A)	            |       /           |
     //                                     |  Vertical         |
     //                                     |    /              |
     //                                     | ID(id node) (N)*  |



         //check .... in R/(N)* or ... in R/(N)+;
         SimpleNode starNode = (SimpleNode)nodeIn.jjtGetChild(0).jjtGetChild(0)
                                .jjtGetChild(0).jjtGetChild(0);
         if(starNode.name.indexOf("*")!=-1)
            isKStar = "*";
         else if (starNode.name.indexOf("+")!=-1)
            isKStar = "+";

          //modify the tree
          nodeIn.jjtRemoveChild(0);
          SimpleNode newId = ((SimpleNode)nodeIn.jjtGetChild(0)).jjtDuplicate();
          nodeIn.jjtRemoveChild(0);
          (nodeIn).jjtReplace(newId);

          SimpleNode allTogether = twoStars(nodeRight,recur_level,isKStar,env);

		  node.jjtRemoveChild(1);
		  allTogether.jjtSetParent(node);
		  (node).jjtAddChild(allTogether);
		}

     }
     else
     {
	    KStarName = null;
	    isKStar = null;
	    throw new InterpretError ("To be implemented");
	 }
    }

   }//end if("TSelect")

  }//end if (recur_level>0......)

 }//end else

}//end of isKStar != null

 KStarName = null;
 isKStar = null;
 return node;
}


//handle the case that there is KStar in projection
private SimpleNode KStarInProjection(SimpleNode nodeRight,int recur_level,
                                     Relation recursiveRelation,String isKStar,Environment env,boolean starInEnd)
        throws InterpretError
{
       //System.out.println("in KStarInprojection --------------");
	   //              project(nodeRight)
	   //                  /             \
	   //           projector            [Identifier(       : R)]
	   //               /                OR a "Project" tree
	   //    ExpressionList(nodeExp)
	   //            /
	   //        Verticle
	   //          /
	   //      Identifier(nodeID:(N)* )  R/(N)*
	   //    [    /
	   //      Identifier(         A) ]  R/(N)*/A

     //get copy of nodeRight to be used later to project all domains of R

     SimpleNode nodeRightCopy =(SimpleNode) nodeRight.jjtDeepDuplicate();
	 //get note "ExpressionList"
	 SimpleNode nodeExp =
	 (SimpleNode) nodeRight.jjtGetChild(0).jjtGetChild(0);

     SimpleNode starNode = (SimpleNode) nodeExp.jjtGetChild(0).jjtGetChild(0);
     KStarName = starNode.name.substring(1,starNode.name.indexOf(")"));

	 SimpleNode nodeID = (SimpleNode)(nodeExp.jjtGetChild(0).jjtGetChild(0));

      //R/(N)* or R/(N)+
	 if(nodeID.jjtGetNumChildren()==0)
	 {
		nodeExp.jjtRemoveChild(0);
		String domName;

       	//R/(N)*  should project all domains of N (R/(N)*/A,B,C...)
		//and all domains of R
		//get all domains of N, and connect these nodes to ExpressionList.
		for(int j=0;j<recursiveRelation.domains.length;j++)
		{
			domName = recursiveRelation.domains[j].name;
			if(domName.startsWith(".")!=true && domName.startsWith(KStarName)!=true )
			{
			  SimpleNode nodeD = SimpleNode.jjtCreateNode("Identifier",230,230, domName);
			  nodeD.jjtSetParent(nodeExp);
			  nodeExp.jjtAddChild(nodeD);
			}
		}

	 }
	 //R/(N)*/A or R/(N)+/A:Just remove "Vertical" node
	 else
	 {
		 SimpleNode nodeBottom = (SimpleNode)((nodeExp.jjtGetChild(0)).jjtGetChild(0)).jjtGetChild(0);
		 nodeExp.jjtRemoveChild(0);

		 nodeBottom.jjtSetParent(nodeExp);
		 nodeExp.jjtAddChild(nodeBottom);

	 }

   //if there is a relation in bottom,like R/(N)*/relation, should do further

   //build correct tree corresponding to the queries
   //should be solved as:lift all levels to top level and ujoin them
   //R/(N)*/A ==>(R/N/N_0/.../N_recur_level-1/A) ujoin ...ujoin (R/N/N/A)
   //            ujoin (R/N/A) [ujoin (R/A)] (if R has A)
   // To save the I/O time, just project in each level and ujoin them
   //R/(N)*/A ==>([A] in N_recur_level-1) ujoin ... ujoin ([A] in N_1)
   //            ujoin ([A] in N_0) [ujoin [A] in R] (if R has A)
   //
   // the tree will be
   //               Assignment
   //                /      \

   //          Identifer   Join
   //                      /   \
   //                    Join   Project
   //                     / \   /      \
   //               ... ,,,   Projector Identifier
   //(recur_level "Join" nodes)     /
   //                  Join        ExpressionList
   //                   /  \              /
   //            Project    Project    Identifier
   //           /    \       /      \
   // Projector Identifier Projector Identifier
   //     /                   /
   // ExpressionList     ExpressionList
   //    /                 /
   // Identifier	   Identifier

   // to create the "Join" tree
   Vector JoinNode = new Vector();
   Vector projectHat = new Vector();

   //create the top hat -- nodeRCopy
   //          project
   //      projector id(R)
   //      expList

   //      vertical
    SimpleNode nodeRCopy = (SimpleNode) nodeRightCopy.jjtDeepDuplicate();
    SimpleNode vertical = (SimpleNode) nodeRCopy.jjtGetChild(0).
               jjtGetChild(0).jjtGetChild(0);
    vertical.jjtRemoveChild(0);
    projectHat.addElement(nodeRCopy);

    for(int i=0;i<=recur_level;i++)
    {
	 if(i==0)
	 {
	   SimpleNode projectNode = (SimpleNode) nodeRCopy.jjtDeepDuplicate();
	   SimpleNode verticalNode = (SimpleNode) projectNode.jjtGetChild(0).
	                  jjtGetChild(0).jjtGetChild(0);

	   //i=0 => create tree for red ujoin of N in R
	   //
	   //          project
	   //      projector  id(R)
	   //     ExpList
	   //    Vertical
       //      N

	   SimpleNode IdNode = SimpleNode.jjtCreateNode("Identifier",230,230,KStarName);
       IdNode.jjtSetParent(verticalNode);
       verticalNode.jjtAddChild(IdNode);

       JoinNode.addElement(projectNode);
     }
	 //i>0 create tree for red ujoin of [red ujoin of N_0 in N] in R
	 //               join
	 //    project               project
	 //  projector id(R)   projector id(R)
	 // Exp                Exp
	 // vertical          vertical
	 //  project            .... R/N/N_0/N_1
	 // projector id N
	 //  explist
	 //  vetical N_0

	 if(i>0)
	 {
       SimpleNode lastHatNode = (SimpleNode) projectHat.lastElement();
       SimpleNode lastHatNodeCopy = (SimpleNode) lastHatNode.jjtDeepDuplicate();
       SimpleNode tmpS = lastHatNodeCopy;
	   while(true)

	   {

	     if(tmpS.jjtGetNumChildren()==0)
	    	 break;
	     else
	   		 tmpS = (SimpleNode)tmpS.jjtGetChild(0);
	   }

       SimpleNode Idnode = SimpleNode.jjtCreateNode("Identifier",230,230,KStarName+"_"+(i-1));

       Idnode.jjtSetParent(tmpS);
	   tmpS.jjtAddChild(Idnode);

       SimpleNode Join = SimpleNode.jjtCreateNode("Join",301,362,null);
       SimpleNode lastJoin = (SimpleNode)JoinNode.lastElement();

	   lastJoin.jjtSetParent(Join);
	   (Join).jjtAddChild(lastJoin);
	   lastHatNodeCopy.jjtSetParent(Join);

	   (Join).jjtAddChild(lastHatNodeCopy);

       JoinNode.addElement(Join);

	 }
	 //create new Hat
	 //          ...
	 //           /
     //          project
     //      projector  id(N)
     //     ExpList
     //     Vertical
     //
     SimpleNode newHat = (SimpleNode) nodeRCopy.jjtDeepDuplicate();
     SimpleNode id = (SimpleNode)newHat.jjtGetChild(1);
     if(i==0)
		  id.name = KStarName;
     else
		  id.name = KStarName +"_"+(i-1);

     SimpleNode lastHat = (SimpleNode)projectHat.lastElement();
     SimpleNode HatCopy = (SimpleNode)lastHat.jjtDeepDuplicate();
     SimpleNode tmp = HatCopy;
     while(true)
      {
		 if(tmp.jjtGetNumChildren()==0)
		   break;

		 else
		   tmp = (SimpleNode)tmp.jjtGetChild(0);
      }
     newHat.jjtSetParent(tmp);
     tmp.jjtAddChild(newHat);
	 projectHat.addElement(HatCopy);

  }//end for

   //connect the join tree to the nodeRight("Project")
   SimpleNode topJoin = (SimpleNode)JoinNode.lastElement();

   if(isKStar.equals("*")) //R should be projected
   {
	 //Case: R/(N)*/A; (N)*/A in R;[A] in R(N)*; [B] where A = 1 in R(N)*
     if(starInEnd==false)
     {
      SimpleNode id = (SimpleNode)nodeRight.jjtGetChild(1);
      SimpleNode topID = (SimpleNode)id.jjtDeepDuplicate();
      id.jjtRemoveChild(0);
      id.jjtRemoveChild(0);

      SimpleNode Join = SimpleNode.jjtCreateNode("Join",301,362,null);
      id.jjtReplace(Join);
      SimpleNode newJoin = (SimpleNode)nodeRight.jjtGetChild(1);
      topJoin.jjtSetParent(newJoin);
      newJoin.jjtAddChild(topJoin);
      topID.jjtSetParent(newJoin);
      newJoin.jjtAddChild(topID);
     }
     else //the case: R/(N)* or (N)* in R
     {
	   //get all domain of R and build the projectorList Tree for R;
	   SimpleNode projectorCopy = ((SimpleNode)nodeRightCopy.jjtGetChild(0)).jjtDeepDuplicate();
       SimpleNode idCopy = ((SimpleNode)nodeRightCopy.jjtGetChild(1)).jjtDuplicate();
       SimpleNode nodeExpNode =
	   	          (SimpleNode) projectorCopy.jjtGetChild(0);
       nodeExpNode.jjtRemoveChild(0);

	   Relation topRelation = env.lookupRel(idCopy.name, true);
	   String domName;
       for(int j=0;j<topRelation.domains.length;j++)
	   {
	   	 domName = topRelation.domains[j].name;
	   	 if(domName.startsWith(".")!=true && !domName.equals(KStarName))
	   	 {
	   	    SimpleNode nodeD = SimpleNode.jjtCreateNode("Identifier",230,230, domName);
	   	    nodeD.jjtSetParent(nodeExpNode);
	   	    nodeExpNode.jjtAddChild(nodeD);
	   	 }
	   }

       //join (N)*/N's domain and R/R' domains
       //             join
       //           /     \
       //      project     project
       //      /  \         /      \
       //projector topJoin projecotr R
       //   /    ..           /

       SimpleNode projectNode = SimpleNode.jjtCreateNode("Project",421,427,null);
       projectorCopy.jjtSetParent(projectNode);
       projectNode.jjtAddChild(projectorCopy);
       idCopy.jjtSetParent(projectNode);
       projectNode.jjtAddChild(idCopy);

       SimpleNode id = (SimpleNode)nodeRight.jjtGetChild(1);
       id.jjtReplace(topJoin);

       SimpleNode newnodeRight = (SimpleNode) nodeRight.jjtDeepDuplicate();
       SimpleNode Join = SimpleNode.jjtCreateNode("Join",301,362,null);

       projectNode.jjtSetParent(Join);
       Join.jjtAddChild(projectNode);
       newnodeRight.jjtSetParent(Join);
	   Join.jjtAddChild(newnodeRight);

	   nodeRight.jjtRemoveChild(0);
	   nodeRight.jjtRemoveChild(0);
	   nodeRight.jjtReplace(Join);
     }
   }
   else if(isKStar.equals("+")) //R should not be projected
   {
     SimpleNode id = (SimpleNode)nodeRight.jjtGetChild(1);
     id.jjtReplace(topJoin);
   }

   return nodeRight;

}


private SimpleNode starChildOfProject(SimpleNode nodeRight)
{

    SimpleNode nodeRightL = (SimpleNode)nodeRight.jjtGetChild(0);

	SimpleNode idDupNode =
	  (SimpleNode)nodeRightL.jjtDeepDuplicate();


	//first, transfer the above tree to the tree of R/(N)*[/A]
	//create a new tree like
	//    projector
	//       /
	//    ExpressionList
	//      /
	//    Verticle

	SimpleNode projectorNode = SimpleNode.jjtCreateNode("Projector",420,420,null);
	projectorNode.setBit(1);
	SimpleNode expListNode = SimpleNode.jjtCreateNode("ExpressionList",592,592,null);
	SimpleNode verticalNode = SimpleNode.jjtCreateNode("Vertical",306,362,null);

	verticalNode.jjtSetParent(expListNode);
	expListNode.jjtAddChild(verticalNode);
	expListNode.jjtSetParent(projectorNode);
	projectorNode.jjtAddChild(expListNode);

	//connect to the input tree
	idDupNode.jjtSetParent(verticalNode);
	verticalNode.jjtAddChild(idDupNode);


	if(nodeRightL.jjtGetNumChildren()>0)
		nodeRightL.jjtRemoveChild(0);

	nodeRightL.jjtReplace(projectorNode);

	return nodeRight;
}

//twoStars handles the case that there are 2 stars:
//[B] where (N)*/A=1 (or (N)+/A) in R/(N)*;
private SimpleNode twoStars(SimpleNode nodeRightDup,int recur_level,
          String isKStar,Environment env)
    throws InterpretError

{
     //get the name of R
     SimpleNode tmpRecur =
                (SimpleNode)nodeRightDup.jjtGetChild(1).jjtGetChild(1);
     SimpleNode tmpId =
                (SimpleNode)nodeRightDup.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);

     String tmpRecurName = tmpRecur.name;
     SimpleNode topTSelNode = null;
     Vector TselVector = new Vector();

     //check:[B] where (N)*/A=1 (or (N)+/A) in R/(N)*;
     //last * or + has been checked in modifyNodeForKSTAR

     String firstStar = null;
     SimpleNode IdStarNode =
                (SimpleNode)nodeRightDup.jjtGetChild(1).jjtGetChild(0).jjtGetChild(0);

     if(IdStarNode.name.indexOf("*")!=-1)
         firstStar = "*";

     else if(IdStarNode.name.indexOf("+")!=-1)
      	 firstStar = "+";


	 for(int i=0;i<=recur_level;i++)
	 {
       if(i==0)
	   {
		  topTSelNode = KStarInSelection(nodeRightDup,recur_level,i,firstStar,env);
	      TselVector.addElement(topTSelNode);
       }

       else
       {
		 topTSelNode = KStarInSelection(nodeRightDup,recur_level,i,firstStar,env);

         if(topTSelNode !=null)
         {
      	 //create a "Join" node
	     SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);

	  	 SimpleNode topTS = (SimpleNode)TselVector.lastElement();

	  	 topTS.jjtSetParent(JoinNode);
	  	 JoinNode.jjtAddChild(topTS);

	  	 topTSelNode.jjtSetParent(JoinNode);
	  	 JoinNode.jjtAddChild(topTSelNode);

	  	 TselVector.addElement(JoinNode);

         }

	   }
   	 }//end for

	 //should include the case that if R has B,
	 // "(N)*/B where (N)*/A = value in R;" => "B  where (N)*/A in R;"
	 //even include "B  where A in R;" if R has A

     SimpleNode joinTree = (SimpleNode)TselVector.lastElement();
     SimpleNode allTogether = joinTree;

     //Case: [B] where (N)*/A=1 (or (N)+/A =1) in R/(N)* =>Should project:
     //[B] where (N)*/A=1 (or (N)+/A=1)  in R;
     if(isKStar.equals("*"))
     {
		//first chech B in R?

		Relation recurRel = env.lookupRel(tmpRecurName, true);
		boolean whetherIn = false;
		for(int j=0;j<recurRel.domains.length;j++)
		{
			if(recurRel.domains[j].name.equals(tmpId.name))
			{
			   whetherIn = true;
			   break;

			}
		 }
		//if B in R, ujoin [B] where (N)*/A=1 (or (N)+/A=1 in R;
		if(whetherIn == true)
		{
           topTSelNode = KStarInSelection(nodeRightDup,recur_level,-1,firstStar,env);
           SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);
		   topTSelNode.jjtSetParent(JoinNode);
		   (JoinNode).jjtAddChild(topTSelNode);

           joinTree.jjtSetParent(JoinNode);
		   (JoinNode).jjtAddChild(joinTree);

		   allTogether = JoinNode;
    	 }
	 }

      return allTogether;

 }


//function below handles the case that there is KStar in projection when in R that is indetifier
//[B] where (N)*A = Value in R
private SimpleNode KStarInSelection(SimpleNode nodeRight,int recur_level,
              int currentLevel,String isKStar,Environment env)
        throws InterpretError

{

	//the original tree is
	//              TSelect(nodeRight)
	//                   /    \
	//             Projector   Selector
	//                 /       /     \
	//    ExpressionList  Compare     Identifier(    :R)
	//        /            /     \           rIdNode
	//Identifier( B)  Identifier  Literal
	//                  ((N)*)      (value)
	//                  Identifier
	//                  (A)



  // should be transfered like:
  // query <-  ([B] where ([] where A=value in N) in R) ujoin
  //     ([B] where ([] where ([] where A=value in N_0) in N) in R) ujoin
  //     ([B] where ([] where ([] where ([] where A=value in N_1) in N_0)
  //            in N) in R) ujoin
  //     ([B] where ([] where ([] where ([] where ([] where A=value in N_2)in N_1)
  //            in N_0) in N) in R) ujoin
  //       ... ... ... ...
  //     ([B] where (... ...in N_recur_Level_1) in N_recur_Level-2)...)  ;

  //the correct tree like:
  //            Assignment
  //           /          \

  //   Idnetifier         Join
  //                     /    \
  //                ......     TSelect (from .N_recur_Level-1 to R)
  //(recur_level "Join" nodes)     ... ... ...
  //                 Join
  //                /    \
  //          TSelect     TSelect(from N_0 to R)
  //         /   \            ... ...
  // Projector   Selector
  //   /           /     \
  // ExpList  TSelect   Identi( .N)
  //  /        /   \
  //Identi Projtor Selector
  //               /     \
  //            Compare  Identi( N_0)
  //            /     \
  //          ( A)     value
  //


   // get the left tree unit
   //          TSelect(dupTSel)
   //           /
   //         Projector
   //          /
   //  ExpressionList
   //        /
   //    Identifier
   SimpleNode dupTSel = (SimpleNode)nodeRight.jjtDeepDuplicate();
   dupTSel.jjtRemoveChild(1);

   SimpleNode lnode = (SimpleNode)nodeRight.jjtGetChild(0); //node "Projector"
   SimpleNode dupPro = (SimpleNode)lnode.jjtDeepDuplicate();

   //get the selector node
   SimpleNode rnode = (SimpleNode)nodeRight.jjtGetChild(1); //node "Selector"
   SimpleNode dupSel = (SimpleNode)rnode.jjtDeepDuplicate();

   //get the right identifier node
   SimpleNode rIdNode = (SimpleNode)dupSel.jjtGetChild(1);


   //change (N)* to A in Identifier node;
   SimpleNode idNodeStar =
	 (SimpleNode)dupSel.jjtGetChild(0).jjtGetChild(0);
   SimpleNode idNodeDom=
	 (SimpleNode)dupSel.jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);

   idNodeStar.jjtReplace(idNodeDom);
   idNodeStar.jjtRemoveChild(0);


   //create the "hat" of the selector tree
   //
   //	 selector(selNode)
  //		 /     \
  //	  TSelect Identifier(top-level relation or recursive relation)
  //	  /        (hatIdNode)
  // Projector

   SimpleNode selNode = (SimpleNode)rnode.jjtDeepDuplicate();
   SimpleNode hatIdNode = (SimpleNode)selNode.jjtGetChild(1);
   //  TSelect
   //   /
   // Projector
   SimpleNode tselNode = SimpleNode.jjtCreateNode("TSelect",422,422,null);
   SimpleNode proNode = SimpleNode.jjtCreateNode("Projector",420,420,null);
   proNode.jjtSetParent(tselNode);
   (tselNode).jjtAddChild(proNode);

   SimpleNode CompareNode = (SimpleNode)selNode.jjtGetChild(0);
   CompareNode.jjtReplace(tselNode);


   //create the selector tree(recur_level)
   Vector selectVec = new Vector(); //to save the selector tree;
   //create the "hat"s
   Vector hatVec = new Vector();
   SimpleNode topTSelNode=null;

   if(currentLevel == -1) //the case [B] where (N)*/A=1 in R
   {
    for (int j=0;j<=recur_level;j++)
    {
	 SimpleNode dselNode = (SimpleNode)selNode.jjtDeepDuplicate();
	 SimpleNode TdupTSel = (SimpleNode)dupTSel.jjtDeepDuplicate();

	 SimpleNode dupSelCopy = (SimpleNode)dupSel.jjtDeepDuplicate();
	 SimpleNode rIdNodeCopy = (SimpleNode)dupSelCopy.jjtGetChild(1);

	 SimpleNode tmpJoinLeft =  SimpleNode.jjtCreateNode("Identifier",230,230,null);

	 if(j==0)   // the case [B] where (N)*/A=1 in R
	 {
	   //create the bottom tree
	   //   Selector
	   //    /     \
	   // Compare  Identifier (N)
	   //  /    \ ...

		rIdNodeCopy.name = KStarName;
		//add the "hat"
		dupSelCopy.jjtSetParent(dselNode.jjtGetChild(0));
		(dselNode.jjtGetChild(0)).jjtAddChild(dupSelCopy);
		//add projection tree
		dselNode.jjtSetParent(TdupTSel);
		(TdupTSel).jjtAddChild(dselNode);
		//save the first tree to the vector selectVec
		selectVec.addElement(TdupTSel);
	  }
	  else if (j>0)  // to handle other trees
	  {
		//create a "Join" node
		SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);

		//create the bottom tree
		//         Selector(dupSel)
		//          /    \
		//      compare  Identifier(depends on recursive levels)
		//        /   \   (rIdNode)
		//Identifier  Literal
		rIdNodeCopy.name = KStarName+"_"+(j-1);
		int k;
		SimpleNode   dupHatNode=null;
		SimpleNode   dupHatIdNode=null;
		//add the "hat"
		for(int i=1;i<=j;i++)
		{
		 //create a new "hat"
		  dupHatNode = (SimpleNode)selNode.jjtDeepDuplicate();
		  dupHatIdNode = (SimpleNode)dupHatNode.jjtGetChild(1);

		  //get the hat tree
		  if(j-1-i>=0)
			 dupHatIdNode.name = KStarName+"_"+(j-1-i);
		  else   //j-1=0   the hat of _0
			 dupHatIdNode.name = KStarName;

		  if(i==1)
		  {
			//connect the base node to the hat node
			dupSelCopy.jjtSetParent(dupHatNode.jjtGetChild(0));
			(dupHatNode.jjtGetChild(0)).jjtAddChild(dupSelCopy);
		  }
		  else if(i>1)
		  {
			//add a hat
			SimpleNode lastSel = (SimpleNode)hatVec.lastElement();
			SimpleNode tempTSelNode = (SimpleNode)dupHatNode.jjtGetChild(0);

			lastSel.jjtSetParent(tempTSelNode);
			(tempTSelNode).jjtAddChild(lastSel);

		   }
		   //add the selector tree to the hatVec
		   hatVec.addElement(dupHatNode);
		}//end for
	   //add the top hat
	   dupHatNode.jjtSetParent(dselNode.jjtGetChild(0));
	   (dselNode.jjtGetChild(0)).jjtAddChild(dupHatNode);

	   //add projection tree
	   dselNode.jjtSetParent(TdupTSel);
	   (TdupTSel).jjtAddChild(dselNode);

	   //add tmp node to the left of Join
		tmpJoinLeft.jjtSetParent(JoinNode);
	   (JoinNode).jjtAddChild(tmpJoinLeft);
	   //add TdupTsel Tselect node ot the right of Join
	   TdupTSel.jjtSetParent(JoinNode);
	   (JoinNode).jjtAddChild(TdupTSel);

	   topTSelNode = (SimpleNode)selectVec.lastElement();
	   tmpJoinLeft.jjtReplace(topTSelNode);

	   //save the first tree to the vector selectVec
		selectVec.addElement(JoinNode);
	  }//end else if
	 }//end for

	topTSelNode = (SimpleNode)selectVec.lastElement();

	if(isKStar.equals("*")) //should project [B] where A="m" in R;
	{
	  //System.out.println("isKStar =*");

	  SimpleNode dupNodeRight = (SimpleNode) nodeRight.jjtDeepDuplicate();
	  SimpleNode dName = (SimpleNode) dupNodeRight.jjtGetChild(1).jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);
	  SimpleNode relNode = (SimpleNode) dupNodeRight.jjtGetChild(1).jjtGetChild(1);

	  //check A in R?
	  Relation proRelation = env.lookupRel(relNode.name, true);
	  boolean whetherIn = false;
	  for(int j=0;j<proRelation.domains.length;j++)
	  {
	    if(proRelation.domains[j].name.equals(dName.name))
	    { whetherIn = true;
	  	   break;
	  	}
	  }
	  if(whetherIn == true)
	  {
	    SimpleNode dNameCopy = (SimpleNode) dName.jjtDuplicate();
	    ((SimpleNode) dupNodeRight.jjtGetChild(1).jjtGetChild(0).jjtGetChild(0)).jjtRemoveChild(0);
	    ((SimpleNode) dupNodeRight.jjtGetChild(1).jjtGetChild(0).jjtGetChild(0)).jjtReplace(dNameCopy);

	    SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);
	    topTSelNode.jjtSetParent(JoinNode);
		(JoinNode).jjtAddChild(topTSelNode);

		dupNodeRight.jjtSetParent(JoinNode);
		(JoinNode).jjtAddChild(dupNodeRight);

	    topTSelNode = JoinNode.jjtDeepDuplicate();
	  }
    }
  } //end currentLevel=-1

  //handle the case [B] where (N)*/A=1(or (N)+/A in .N or .N_0...
  if(currentLevel>=0)
  {
	//change the name of the identifier (topLevel) to .N or .N1..
	SimpleNode id = (SimpleNode)selNode.jjtGetChild(1);
    if(currentLevel==0)

      id.name = "."+KStarName;
    else if(currentLevel>0)
      id.name = "." + KStarName+"_"+(currentLevel-1);

    for (int j=currentLevel;j<recur_level;j++)
	{
	  SimpleNode dselNode = (SimpleNode)selNode.jjtDeepDuplicate();
	  SimpleNode TdupTSel = (SimpleNode)dupTSel.jjtDeepDuplicate();
	  SimpleNode dupSelCopy = (SimpleNode)dupSel.jjtDeepDuplicate();
	  SimpleNode rIdNodeCopy = (SimpleNode)dupSelCopy.jjtGetChild(1);

      SimpleNode tmpJoinLeft =  SimpleNode.jjtCreateNode("Identifier",230,230,null);

       if(j==currentLevel)
	   {
	  	   //create the bottom tree
	  	   //   Selector
	  	   //    /     \
	  	   // Compare  Identifier (N)
	  	   //  /    \ ...
	  	rIdNodeCopy.name = KStarName+"_"+j;
	  	//add the "hat"
	  	dupSelCopy.jjtSetParent(dselNode.jjtGetChild(0));
	  	(dselNode.jjtGetChild(0)).jjtAddChild(dupSelCopy);
	  	//add projection tree
	  	dselNode.jjtSetParent(TdupTSel);
	  	(TdupTSel).jjtAddChild(dselNode);

		 //save the first tree to the vector selectVec
	  	 //selectVec.addElement(JoinNode);
	  	 selectVec.addElement(TdupTSel);

	   }
	   else if (j>currentLevel)  // to handle other trees
	   {
		 //create a "Join" node
		 SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);

		 //create the bottom tree
		 //         Selector(dupSel)
		 //          /    \
		 //      compare  Identifier(depends on recursive levels)
		 //        /   \   (rIdNode)
		 //Identifier  Literal

	  	 rIdNodeCopy.name = KStarName+"_"+j;
	  	 SimpleNode   dupHatNode=null;
	  	 SimpleNode   dupHatIdNode=null;

	  	//add the "hat"
		for(int i=0;i<(j-currentLevel);i++)
		{
		  //create a new "hat"
		  dupHatNode = (SimpleNode)selNode.jjtDeepDuplicate();
		  dupHatIdNode = (SimpleNode)dupHatNode.jjtGetChild(1);
		  dupHatIdNode.name = KStarName+"_"+(j-i-1);

   		  if(i==0)
		  {
			//connect the base node to the hat node
			dupSelCopy.jjtSetParent(dupHatNode.jjtGetChild(0));
			(dupHatNode.jjtGetChild(0)).jjtAddChild(dupSelCopy);
		  }
		  else
		  {
			//add a hat
			SimpleNode lastSel = (SimpleNode)hatVec.lastElement();
			SimpleNode tempTSelNode = (SimpleNode)dupHatNode.jjtGetChild(0);

			lastSel.jjtSetParent(tempTSelNode);
			(tempTSelNode).jjtAddChild(lastSel);
		  }
		   //add the selector tree to the hatVec
			 hatVec.addElement(dupHatNode);
		}//end for

	  	   //add the top hat
	  	  dupHatNode.jjtSetParent(dselNode.jjtGetChild(0));
	  	  (dselNode.jjtGetChild(0)).jjtAddChild(dupHatNode);

	  	   //add projection tree
	  	  dselNode.jjtSetParent(TdupTSel);
	  	  (TdupTSel).jjtAddChild(dselNode);

	  	  //add tmp node to the left of Join
	  	  tmpJoinLeft.jjtSetParent(JoinNode);
	  	  (JoinNode).jjtAddChild(tmpJoinLeft);
	  	  //add TdupTsel Tselect node ot the right of Join
	  	  TdupTSel.jjtSetParent(JoinNode);
	  	  (JoinNode).jjtAddChild(TdupTSel);

	  	  topTSelNode = (SimpleNode)selectVec.lastElement();
	  	  tmpJoinLeft.jjtReplace(topTSelNode);

	  	  //save the first tree to the vector selectVec
	  	  selectVec.addElement(JoinNode);

	  	  }//end else if
	   }//end for

    //case [B] where (N)*/A=1 in .N ...
    //should project: [B] where A = 1 in .N...
    if(isKStar.equals("*"))
    {
	  SimpleNode TdupTSel = (SimpleNode)dupTSel.jjtDeepDuplicate();
	  SimpleNode dupSelCopy = (SimpleNode)dupSel.jjtDeepDuplicate();
	  SimpleNode rIdNodeCopy = (SimpleNode)dupSelCopy.jjtGetChild(1);

	  if(currentLevel==0)
	    rIdNodeCopy.name = "."+KStarName;
	  else if(currentLevel>0)
	    rIdNodeCopy.name = "." + KStarName+"_"+(currentLevel-1);

	  //add projection tree
	  dupSelCopy.jjtSetParent(TdupTSel);
	  (TdupTSel).jjtAddChild(dupSelCopy);

      //for example: R has recursive N = > (N,N_0,N_1,N_2) recur_level =3
      //N,N_0,N_1
      if(currentLevel<recur_level)
	  {
		topTSelNode = (SimpleNode)selectVec.lastElement();

	    SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);
	    topTSelNode.jjtSetParent(JoinNode);
	    (JoinNode).jjtAddChild(topTSelNode);

	    TdupTSel.jjtSetParent(JoinNode);
	    (JoinNode).jjtAddChild(TdupTSel);

	    topTSelNode = JoinNode;
	  }
	  //N_2
	  else if(currentLevel == recur_level)
	    topTSelNode = TdupTSel;
    }

    else if(isKStar.equals("+"))
    {
	  if(currentLevel<recur_level)
	  	topTSelNode = (SimpleNode)selectVec.lastElement();
	  else if(currentLevel == recur_level)
	    topTSelNode = null;
	}
  }//end if currentLevel >=0

   //System.out.println(" ********* TopTSelNode *********** ");
   //Global global= new Global();
   //global.traverse(topTSelNode);

   return topTSelNode;

}

//hand the case R/*/A or R/.*/A and R/+/A or R/.+/A
//project all A no matter it is on which level
private SimpleNode nestDomsA(SimpleNode node,Vector allDomsA,Vector hat,
                             Environment env,boolean isTop,String KStarName)
        throws InterpretError
{
    //System.out.println("Relname = "+((SimpleNode)node.jjtGetChild(1)).name);

   SimpleNode nodeRightCopy = (SimpleNode) node.jjtDeepDuplicate();
   if(nodeRightCopy.type == OP_PROJECT)
   {
	 //           project(nodeRight)
	 //          /      \
	 //      projector   topRelNode( R)
	 //         /
	 //      expList
	 //        /
	 //      vertical
	 //        /
	 //      starNode * | .* OR + .+
	 //       /
	 //      domNode A

     SimpleNode topRelNode = (SimpleNode) nodeRightCopy.jjtGetChild(1);
     Relation topRelation = null;
     if(isTop == true)
        topRelation = env.lookupRel(topRelNode.name, true);
     else
        topRelation = env.lookupRel("."+topRelNode.name, true);

     SimpleNode domNode =(SimpleNode) nodeRightCopy.jjtGetChild(0).
       jjtGetChild(0).jjtGetChild(0).jjtGetChild(0).jjtGetChild(0);

     //find all nested domain of R, which also contais A
     for(int i=0;i<topRelation.domains.length;i++)   //R
     {
      //System.out.println("domName = "+ topRelation.domains[i].name);
      if(topRelation.domains[i].type != IDLIST)
      {
		   //contain the dom?
		   if(topRelation.domains[i].name.equals(domNode.name) &&
		      KStarName.equals("*"))
           {
	         //to create a tree like:
			 //           project(nodeRight)
			 //          /      \
			 //      projector   topRelNode( R)
			 //         /
			 //      expList
			 //       /
			 //      domNode A

		  SimpleNode nodeRightC = (SimpleNode) nodeRightCopy.jjtDeepDuplicate();

		  SimpleNode nodeExpList = (SimpleNode)nodeRightC.jjtGetChild(0).jjtGetChild(0);

	      SimpleNode domNodeC = (SimpleNode)nodeExpList.jjtGetChild(0).jjtGetChild(0).
												 jjtGetChild(0);

		  nodeExpList.jjtRemoveChild(0);
		  domNodeC.jjtSetParent(nodeExpList);
		  nodeExpList.jjtAddChild(domNodeC);

          SimpleNode newDomA = null;

		  if(hat.size()==0)  //add it to vector
              newDomA = nodeRightC;

		  else   //hat.size()>0
		  {
			 SimpleNode formerHat=(SimpleNode)hat.lastElement();
             SimpleNode formerHatNew = (SimpleNode)formerHat.jjtDeepDuplicate();
			 //to get the bottom node of the hat tree
			 SimpleNode tmp=formerHatNew;
			 while(true)
             {
			   if(tmp.jjtGetNumChildren()!=0)
			    tmp = (SimpleNode)tmp.jjtGetChild(0);
			   else
			     break;
			 }

			 //connect to the hat(tmp)
			 nodeRightC.jjtSetParent(tmp);
			 tmp.jjtAddChild(nodeRightC);

			 newDomA = formerHatNew;
		   }

           //save it to the allDomsA
		   if(allDomsA.size() ==0)
			 allDomsA.addElement(newDomA);
		   else
		   {
			 SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);
			 SimpleNode formerNode = (SimpleNode)allDomsA.lastElement();

			 formerNode.jjtSetParent(JoinNode);
			 JoinNode.jjtAddChild(formerNode);
			 newDomA.jjtSetParent(JoinNode);
			 JoinNode.jjtAddChild(newDomA);

			 allDomsA.addElement(JoinNode);
			}
		}//end of contains dom :if(topRelation.domain[i]....)

	  }//end if(topRelation.domain[i].type!=IDLIST

  	  else if (topRelation.domains[i].type == IDLIST &&
  	           !topRelation.domains[i].name.equals(".id"))  //nestRelation
      {
         //build hat
         SimpleNode nodeHat = (SimpleNode)node.jjtDeepDuplicate();
         SimpleNode vertical = (SimpleNode)nodeHat.jjtGetChild(0)
                            .jjtGetChild(0).jjtGetChild(0);
         vertical.jjtRemoveChild(0);

         //save it to the hat vector
         if(hat.size()==0)
           hat.addElement(nodeHat);
         else
         {
		    SimpleNode lastHat = (SimpleNode)hat.lastElement();
		    SimpleNode lastHatNew = (SimpleNode)lastHat.jjtDeepDuplicate();

			 //to get the bottom node of the hat tree
			SimpleNode tmp=lastHatNew;
			while(true)
			{
			   if(tmp.jjtGetNumChildren()!=0)
			    tmp = (SimpleNode)tmp.jjtGetChild(0);
			   else
				break;

			 }
			 //connect to the hat(tmp)
			 nodeHat.jjtSetParent(tmp);
			 tmp.jjtAddChild(nodeHat);

             hat.addElement(lastHatNew);
		 }

         //create a node that can pass to the next recursive loop
         SimpleNode nodeNext = (SimpleNode)node.jjtDeepDuplicate();
         SimpleNode nodeNextId = (SimpleNode)nodeNext.jjtGetChild(1);
         nodeNextId.name = topRelation.domains[i].name;

         //call the recursive loop
         SimpleNode tmpNode = nestDomsA(nodeNext,allDomsA,hat,env,false,"*");
         hat.clear();
       }

   }//end For

   SimpleNode finalTree = (SimpleNode)allDomsA.lastElement();
   return finalTree;

  }//end node.type = project
  else
  {   KStarName = null;
  	  isKStar = null;
      throw new InterpretError("To be implemented");
  }


}

//to hand R/(M/N/...)? A or R/(././...)?A
private SimpleNode modifyNodeForQuestionMarker(SimpleNode node,Environment env)
        throws InterpretError
{
	//      assignment
	//        /    \
	//     id    project
	//            /     \
	//        projector  id
	//          /
	//      explist
	//         /
	//      vertical
	//        /
	//    id((M/N/...)?) pathNode (././...)?
	//      /
	// project  or  id( A) pathChild | alterNode
	//
  SimpleNode projectNode = (SimpleNode)node.jjtGetChild(1);
  SimpleNode projectDupNode = (SimpleNode)projectNode.jjtDeepDuplicate();
  SimpleNode tmpNode = projectDupNode;
  SimpleNode tmp = null;
  //boolean containsQuestionMarker =  false;

  //get the node name with question mark
  while(true)
  {
     tmp = (SimpleNode) tmpNode.jjtGetChild(0);
     if(tmp.type == OP_IDENTIFIER && tmp.name.indexOf("?")!=-1)
     {  //containsQuestionMarker = true;
       	break;
	 }
     else
        tmpNode = tmp;
  }

   SimpleNode pathNode = tmp;
   String idName = pathNode.name;
   SimpleNode pathChild = (SimpleNode)pathNode.jjtGetChild(0);

   //R/(M/N/)?A    should be same with R/A ujoin R/M/N/A;
   SimpleNode alterNode = (SimpleNode)pathChild.jjtDeepDuplicate();

   SimpleNode vertical = (SimpleNode)projectDupNode.jjtGetChild(0).jjtGetChild(0).
                          jjtGetChild(0);
   vertical.jjtRemoveChild(0);

   //copy the node for further use
   SimpleNode projectDupCopy = (SimpleNode)projectDupNode.jjtDeepDuplicate();

   if(pathChild.type == OP_IDENTIFIER) //case R/(M/N/)?A ---->if R contains A
   {
	 SimpleNode explist = (SimpleNode) vertical.jjtGetParent();
	 explist.jjtRemoveChild(0);
     alterNode.jjtSetParent(explist);
     explist.jjtAddChild(alterNode);
   }
   else if(pathChild.type == 421) //case R/(M/..)?N.../A ---->if R contains N
   {
	 alterNode.jjtSetParent(vertical);
     vertical.jjtAddChild(alterNode);
   }

      // R/A
      // projectDupNode is:                projectDupCopy is
      //
      //     project                          project
      //      /    \                           /    \
      //  projector id (Company)          projector id
      //    /                                /
      // ExpList                           Explist
      //   /                                /
      // Identifier A                      vertical

   //----------------------------


   String path = idName.substring(1,idName.indexOf(")?"));

   if(idName.indexOf(".") !=-1) //case R/(./)?A or R/(././)?A
	 projectDupCopy = withoutRelName(projectDupCopy,pathChild,path,env);
   else //case R/(M/N/)?A
	 projectDupCopy = hasRelNames(projectDupCopy,pathChild,path,env);


	//should check whether R contains A
	//firstly get the domain name right after the question mark
	String rName = null;
	if(pathChild.type == OP_IDENTIFIER) //R/(M/N/)?A
	   rName = pathChild.name;
	else if(pathChild.type ==421) //R/(M/)? N/...A;
	   rName = ((SimpleNode)pathChild.jjtGetChild(1)).name;

    //check whether R contains domain A
   	SimpleNode topRel = (SimpleNode) projectNode.jjtGetChild(1);
	Relation topR = env.lookupRel(topRel.name,true);
   	boolean containsTheDom = false;

   	for(int i=0;i<topR.domains.length;i++)
   	{
   	 if(topR.domains[i].name.equals(rName))
   	 {
   	   containsTheDom = true;
   	   break;
   	 }
	}

   //join R/A with R/M/N/A
   if(projectDupCopy != null && containsTheDom == true)
   {
    //add it to the join node
    SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);
    projectDupNode.jjtSetParent(JoinNode);
    JoinNode.jjtAddChild(projectDupNode);

    //connect the another tree to the join node
    projectDupCopy.jjtSetParent(JoinNode);
    JoinNode.jjtAddChild(projectDupCopy);

    //add it to the Assignment tree
    projectNode.jjtRemoveChild(0);
    projectNode.jjtRemoveChild(0);
    projectNode.jjtReplace(JoinNode);
   }
   else if(projectDupCopy == null && containsTheDom == true) //N doesn't contain A
   {
     //System.out.println("Warning:"+path+" doesn't contain the domain "+rName);
     projectNode.jjtRemoveChild(0);
	 projectNode.jjtRemoveChild(0);
     projectNode.jjtReplace(projectDupNode);
   }
   else if(projectDupCopy!=null && containsTheDom == false)
   {
     //System.out.println("Warning:"+topRel.name+" doesn't contain the domain "+rName);
     projectNode.jjtRemoveChild(0);
	 projectNode.jjtRemoveChild(0);
     projectNode.jjtReplace(projectDupCopy);
   }
   else
     throw new InterpretError("Error:"+topRel.name+" and "+path+" don't contain the domain "+rName);

 //clear the "global" variable
 isKStar = null;
 KStarName = null;

 return node;
}



//case R/M/N/A
private SimpleNode hasRelNames(SimpleNode projectDupCopy,
            SimpleNode pathChild,String idName, Environment env)
       throws InterpretError
{
   //create a new tree like
   //
   //     project     ->projectCopy
   //      /    \
   //  projector id
   //    /
   // ExpList
   //   /
   //vertical

   SimpleNode projectCopy = (SimpleNode)projectDupCopy.jjtDeepDuplicate();
   SimpleNode verNode = (SimpleNode)projectCopy.jjtGetChild(0).
                          jjtGetChild(0).jjtGetChild(0);
   if(verNode.jjtGetNumChildren() !=0)
         verNode.jjtRemoveChild(0);

    SimpleNode modifiedNode = null;
	String rName = null;

    //get the domain name right after the question mark
	if(pathChild.type == OP_IDENTIFIER) //R/(M/N/)?A
	   rName = pathChild.name;
	else if(pathChild.type ==421) //R/(M/)? N/...A;
	   rName = ((SimpleNode)pathChild.jjtGetChild(1)).name;

	String path = idName.substring(0,idName.lastIndexOf("/"));

	//first check whether N contains A?
	String lastRelationName = null;
	//get relationName: N
	if(path.indexOf("/")== -1)
	  lastRelationName = path;
	else
	  lastRelationName = path.substring(path.indexOf("/")+1,path.length());

    //System.out.println("lastRelationName = " + lastRelationName);

	//check whether .N contains domain A

	Relation bottomRelation = env.lookupRel("."+lastRelationName, true);
	boolean containsTheDom = false;

	for(int i=0;i<bottomRelation.domains.length;i++)
	{
	 if(bottomRelation.domains[i].name.equals(rName))
	 {
	   containsTheDom = true;
	   break;
	 }
	}

   //if .N contains the dom A, should projectR/M/N/A;
   if(containsTheDom == true)
   {
	   String tmpPath = path;
	   String pathName = null;
	   int pos = 0;
	   int x = path.length();
	   Vector pathTree = new Vector();

	   //build "red ujoin of" trees
	   if(path.indexOf("/") == -1)
	   {
		   SimpleNode tmpTree = (SimpleNode)projectCopy.jjtDeepDuplicate();
		   SimpleNode idNode = (SimpleNode)tmpTree.jjtGetChild(1);
		   idNode.name = path;
		   pathTree.addElement(tmpTree);
	   }
	   else
	   {
		 while(pos!= -1)
		 {
		   SimpleNode tmpTree = (SimpleNode)projectCopy.jjtDeepDuplicate();
		   SimpleNode idNode = (SimpleNode)tmpTree.jjtGetChild(1);
		   SimpleNode tmpVerNode = (SimpleNode)tmpTree.jjtGetChild(0).
								   jjtGetChild(0).jjtGetChild(0);
		   pos = tmpPath.lastIndexOf('/');
		   if(pos !=-1)
		   {
			 pathName = tmpPath.substring(pos+1,tmpPath.length());
			 idNode.name = pathName;
			 tmpPath = tmpPath.substring(0,pos);
		   }
		   else
			 idNode.name = tmpPath;

		   //connent these node
		   if(pathTree.size()==0)
			   pathTree.addElement(tmpTree);
		   else
		   {
				SimpleNode lastNode = (SimpleNode)pathTree.lastElement();
				lastNode.jjtSetParent(tmpVerNode);
				tmpVerNode.jjtAddChild(lastNode);
				pathTree.addElement(tmpTree);
		   }

		  }

	   }//end else

	   //connect the node together
	   SimpleNode formerNode = (SimpleNode)pathTree.lastElement();
	   SimpleNode verCopy = (SimpleNode)projectDupCopy.jjtGetChild(0)
							  .jjtGetChild(0).jjtGetChild(0);
	   formerNode.jjtSetParent(verCopy);
	   verCopy.jjtAddChild(formerNode);

	   //to get the bottom node
	   SimpleNode bottomNode = projectDupCopy;
	   while(true)
	   {
		 if(bottomNode.jjtGetNumChildren()!=0)
			bottomNode = (SimpleNode)bottomNode.jjtGetChild(0);
		 else
			break;

	   }

	   pathChild.jjtSetParent(bottomNode);
	   bottomNode.jjtAddChild(pathChild);

	   modifiedNode = projectDupCopy;

   }//end if containsTheDom == true

	return modifiedNode;
}



//case R/./A or R/././A
private SimpleNode withoutRelName(SimpleNode projectDupCopy,
            SimpleNode pathChild,String idName, Environment env)
        throws InterpretError
{

	//           project
	//            /     \
	//        projector  id
	//          /
	//      explist
	//         /
	//      vertical
	//       /
	//    id( ././)
	//      /
	//   id (  A)

	String path = idName;
	int pos;
	int levelNum=0;

	//check how many levels
	while(true)
	{
	  pos = path.indexOf("/");
	  if(pos == -1)
		 break;
	  else
	   {
		 levelNum = levelNum+1;
		 path = path.substring(pos+1,path.length());
	   }
	}
	//System.out.println("levelNum = "+levelNum);

	//check all relation on right level whether contains the domain A
	SimpleNode topRel = (SimpleNode) projectDupCopy.jjtGetChild(1);
	Relation topR = env.lookupRel(topRel.name,true);

	String newPath = "";
	int count=0;

    // ././..../  turn to real paths
	searchAllLevel(topR,env,newPath,count,levelNum,allPath);

	Vector jointTree = new Vector();

    //join all the paths;
	for(int i=allPath.size()-1;i>=0;i--)
	{
	  SimpleNode projectDupC = (SimpleNode) projectDupCopy.jjtDeepDuplicate();
	  SimpleNode pathChildCopy = (SimpleNode) pathChild.jjtDeepDuplicate();
      SimpleNode JoinNode = SimpleNode.jjtCreateNode("Join",301,362,null);
      SimpleNode pathNode = hasRelNames(projectDupC,pathChildCopy,
                       (String)allPath.elementAt(i),env);

      //System.out.println("i="+i+" "+(String)allPath.elementAt(i));
      if(pathNode!=null) //the path (R/M/N or R/P/Q contains dom A)
      {
        if(jointTree.size()==0)
         	jointTree.addElement(pathNode);
        else
        {
		 pathNode.jjtSetParent(JoinNode);
		 JoinNode.jjtAddChild(pathNode);
         SimpleNode tmp = (SimpleNode)jointTree.lastElement();
         tmp.jjtSetParent(JoinNode);
         JoinNode.jjtAddChild(tmp);
		 jointTree.addElement(JoinNode);
	    }
	  }

   }

   allPath.clear();
   SimpleNode finalNode = null;
   if(jointTree.size()>0)
     finalNode= (SimpleNode)jointTree.lastElement();
   return finalNode;
}



//   ././..../   --> real path
private void searchAllLevel(Relation R,Environment env,String newPath,
                            int count,int levelNum,Vector allPath)
        throws InterpretError
{
  if(count<levelNum)   //how many levels(how many ./)
  {
    for(int i=0;i<R.domains.length;i++) //R.domains
    {

      if(R.domains[i].type == IDLIST && !R.domains[i].name.equals(".id"))
      {
		newPath = newPath + R.domains[i].name + "/";

        Relation R1 = env.lookupRel("."+R.domains[i].name,true);

        searchAllLevel(R1,env,newPath,count+1,levelNum,allPath);

        if(count==levelNum-1)
        allPath.addElement(newPath);
        newPath = "";
      }
    }
  }

}

private SimpleNode modifyNodeWildCard(SimpleNode node,Environment env)
        throws InterpretError
{
	//      assignment
	//        /    \
	//     id    project
	//            /     \
	//        projector  id
	//          /
	//      explist
	//         /
	//      vertical
	//        /
	//    id((././...)) pathNode (././...)
	//      /
	// project  or  id( A) pathChild | alterNode
	//
  SimpleNode projectNode = (SimpleNode)node.jjtGetChild(1);
  SimpleNode projectDupNode = (SimpleNode)projectNode.jjtDeepDuplicate();
  SimpleNode tmpNode = projectDupNode;
  SimpleNode tmp = null;


  //get the node name with .
  while(true)
  {
     tmp = (SimpleNode) tmpNode.jjtGetChild(0);
     if(tmp.type == OP_IDENTIFIER && tmp.name.indexOf("./")!=-1)
       	break;
	 else
        tmpNode = tmp;
  }

   SimpleNode pathNode = tmp;
   String idName = pathNode.name;
   SimpleNode pathChild = (SimpleNode)pathNode.jjtGetChild(0);

   //R/(././)A
   SimpleNode alterNode = (SimpleNode)pathChild.jjtDeepDuplicate();

   SimpleNode vertical = (SimpleNode)projectDupNode.jjtGetChild(0).jjtGetChild(0).
                          jjtGetChild(0);
   vertical.jjtRemoveChild(0);

   //copy the node for further use
   SimpleNode projectDupCopy = (SimpleNode)projectDupNode.jjtDeepDuplicate();


      //          projectDupCopy is
      //
      //           project
      //            /    \
      //        projector id
      //           /
      //        Explist
      //          /
      //     vertical

   //----------------------------

   String rName = null;
   //get the domain name right after the question mark
   if(pathChild.type == OP_IDENTIFIER) //R/(././)A
	   rName = pathChild.name;
   else if(pathChild.type ==421) //R/(./)N/...A;
	   rName = ((SimpleNode)pathChild.jjtGetChild(1)).name;

   String path = idName.substring(1,idName.indexOf(")"));
   projectDupCopy = withoutRelName(projectDupCopy,pathChild,path,env);

   if(projectDupCopy!=null)
   {
     projectNode.jjtRemoveChild(0);
	 projectNode.jjtRemoveChild(0);
     projectNode.jjtReplace(projectDupCopy);
   }
   else
     throw new InterpretError("Error:"+path+" don't contain the domain "+rName);

   isKStar = null;
   return node;
}



/*********************end of modifyNodeForStar *********** by zhan




  /**
   * used in computation
   * @ author Yi Zheng
   * added April 8, 2002
   */
  public void executeCompCommand(SimpleNode node, Environment env)
  throws InterpretError
  {
  	executeCommand(node, env);
  }


  private void executeCommand(SimpleNode node, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    String name = node.name;

    RelTable reltable = env.reltable;
    DomTable domtable = env.domtable;

    switch(opcode)
      {
      case OP_HELP : {
		help(name);
		break;
      }
      case OP_QUIT : {
	// REVISIT : may be computation table
		Global.permit_start = false;

	Global.exit(0);
	break;
      }
      case OP_TRACE : {
	Global.trace = Global.trace ? false : true;
	if (Global.trace && Global.traceoutstream == null) {
	  try {
	    Global.traceoutstream = new PrintStream(new FileOutputStream(".trace"), true);
	  } catch (IOException e) {
	    Global.trace = false;
	    Global.println("I/O Exception in creating file '.trace'.");
	  }
	}
	String s = Global.trace ? "on" : "off";
	Global.println("Note: trace is " + s);
	break;
      }

      case OP_INPUT : {
	String s = name.substring(1, name.length() - 1);
	FileInputStream fin = null;
	try {
	  fin = new FileInputStream(s);
	  Global.jin.resetIn(fin);
	  parser.ReInit(Global.jin);
	  instack.push(in);
	  in = fin;
	} catch (FileNotFoundException e) {
	  throw new InterpretError("Input : file " + name + " not found");
	}
	break;
      }
      case OP_INSQL : { //KHAYA 2007
    		String s = name.substring(1, name.length() - 1);
    		InputStream fin = null;
    		try {
    			Relation rel = env.lookupRel(".rd", true);
    			rel.numtuples = env.reltable.getRDTableSize();
    			env.reltable.dump();
    			rel.load();
      		  String translated = Main.translate(false, s, rel.data);
    		  System.out.println(translated);
      		  fin = (InputStream) new ByteArrayInputStream(translated.getBytes());
    		  Global.jin.resetIn(fin);
    		  parser.ReInit(Global.jin);
    		  instack.push(in);
    		  in = fin;
    		} 
    		catch (Exception e) {
    			//System.out.print(e.);
    		  throw new InterpretError("A problem happened during the SQL translation");
    		}
    		break;
    	      }
      case OP_INSQLFILE : {//KHAYA 2007
    		String s = name.substring(1, name.length() - 1);
    		InputStream fin = null;
    		try {
    			Relation rel = env.lookupRel(".rd", true);
    			rel.numtuples = env.reltable.getRDTableSize();
    			env.reltable.dump();
    			rel.load();
      		  String translated = Main.translate(true, s, rel.data);
    		  fin = (InputStream)new ByteArrayInputStream(translated.getBytes());
    		  Global.jin.resetIn(fin);
    		  parser.ReInit(Global.jin);
    		  instack.push(in);
    		  in = fin;
    		}
    		catch (Exception e) {
	  		  throw new InterpretError("A problem happened during the SQL translation");
	  		}
    		break;
    	      }
      case OP_SSD : {
	Global.ssd = Global.ssd ? false : true;
	String s = Global.ssd ? "on" : "off";
	Global.println("Note: show system domain mode is " + s);
	break;
      }
      case OP_SSR : {
	Global.ssr = Global.ssr ? false : true;
	String s = Global.ssr ? "on" : "off";
	Global.println("Note: show system relation mode is " + s);
	break;
      }
      case OP_DEBUG : {
	Global.debug = Global.debug ? false : true;
	String s = Global.debug ? "on" : "off";
	Global.println("Note: debug mode is " + s);
	break;
      }
      case OP_BATCH : {
	Global.batch = Global.batch ? false : true;
	String s = Global.batch ? "on" : "off";
	Global.println("Note: batch mode is " + s);
	break;
      }
      case OP_EXPERT : {
	Global.expert = Global.expert ? false : true;
	String s = Global.expert ? "on" : "off";
	Global.println("Note: expert mode is " + s);
	break;
      }
      case OP_TIME : {
	Global.time = Global.time ? false : true;

	String s = Global.time ? "on" : "off";
	Global.println("Note: timer is " + s);
	break;
      }
      case OP_DELD : {
	SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	for (int i = 0; i < nd.jjtGetNumChildren(); i++) {
	  //Can only remove a domain that is in most deeply
	  //nested scope
	  env.domtable.rmDom(((SimpleNode) nd.jjtGetChild(i)).name);
	}
	break;
      }
      case OP_DELR : {
	SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	String eaname = new String("");
	if (nd.opcode == OP_EVENTNAME) {
	  eaname = getEveAtt(node, env);
	  env.reltable.rmRel(eaname);
	} else {

	  for (int i = 0; i < nd.jjtGetNumChildren(); i++) {
	    //env.reltable.rmRel(((SimpleNode) nd.jjtGetChild(i)).name);
            if(env.lookupText(((SimpleNode) nd.jjtGetChild(i)).name, true) != null) // Jiantao 2004        
               env.textTable.rmText(((SimpleNode) nd.jjtGetChild(i)).name);
            else   
               env.reltable.rmRel(((SimpleNode) nd.jjtGetChild(i)).name);             
	  }
	}
	break;
      }
     case OP_DELT : // Jiantao 2004
        {       SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
                String eaname = new String("");
                for (int i = 0; i < nd.jjtGetNumChildren(); i++)
                {       env.textTable.rmText(((SimpleNode) nd.jjtGetChild(i)).name);
                }
                break;
    }

      case OP_PR : {
	// REVISIT : for COMP -- done. Weizhong.
	// get the temporary file name for the result of expression
	String tmpname = nextTempName();

	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	if (firstchild.opcode == OP_EVENTNAME) {
	  // event name
	  // firstchild.name = env.getEventName(firstchild);
	  firstchild.name = getEveAtt(node, env);
	}
        Text t; // Jiantao 2004 
        if ((t = env.lookupText(firstchild.name, true)) != null)
        {       t.print(); }    
        else    {

	// print trigger in the current environment.
	//first condition showTableadded by Yi Zheng, May 14, 2002
	if ((firstchild.name != null) && (firstchild.name.compareTo("trigger") == 0)) {
	  // handled in the overloaded executeCommand
	} else {

                  /******* start adding by Fan  *******/
                   //firstchild = LookupValidRelNameForPR(firstchild, env); // September 16, 2004
                  firstchild = LookupValidRelName(0, firstchild, env); // Fan January 7, 2005
                  /******* end of Fan's addition *******/

	  // deal with computation as well as relation.

	  Relation rel = evaluateTLExpression( firstchild, tmpname, new Hashtable(), env);

	  if (rel != null) {
	    rel.print(null, rel.name != tmpname);
	  }
          /******* start adding by Fan  *******/
          printRecursiveNestedRelations(rel, env); // September 6, 2004
          /******* end of Fan's addition *******/
        }
	deleteFile(tmpname);
      }
	break;
   }

      case OP_PT : { // Jiantao 2004
                Text t = env.lookupText(name, true);
                if(t == null)
                        throw new InterpretError("text '"+ name + "' does not exist.");
                else
                        t.print();
                break;
        }   

      case OP_SC : {
		 //REVISIT: How about showing not top-level computations
	Relation c = env.lookupRel(node.name, true);
	if (c == null || c.rvc != COMPUTATION)
	  {  throw new InterpretError("No such computation: "+node.name);
	  }
	else
	  {  ((Computation) c).sc();
	  }
	break;
      }
      case OP_SD : {
	if(node.name != null){
	  Domain d = env.lookupDom(node.name, true);
	  if(d == null)
	    throw new InterpretError("domain '"+node.name+
				     "' does not exist.");

	  else
	    env.domtable.showDom(d);
	}else
	  env.domtable.showTable(Global.ssd);

	break;
      }
      case OP_SR : {
        if(node.name != null){
          Text t = env.lookupText(node.name, true); // Jiantao 2004
          if(t == null)
          { 
               Relation r = env.lookupRel(node.name, true);
               if(r == null)         
                   throw new InterpretError("relation or text '"+node.name+ "' does not exist.");   
               else
                   env.reltable.showRel(r);
          }     
          else       
		   env.textTable.showText(t);  
        }
	else
          env.reltable.showTable(Global.ssr);

/*
	if(node.name != null){
	  Relation r = env.lookupRel(node.name, true);
	  if(r == null)
	    throw new InterpretError("relation '"+node.name+
				     "' does not exist.");
	  else

	    env.reltable.showRel(r);
	}else
	  env.reltable.showTable(Global.ssr);
*/
	break;
      }
      case OP_SRD : {//case modified by Yi Zheng Sept. 29, 2002 to allow srd <rel>

	if (((SimpleNode)node).name == null)
	{
		Relation rel = env.lookupRel(".rd", true);
		if (rel != null)
		  rel.print();
		else // shouldn't be here
		  throw new InterpretError("relation .rd is not there");
	}
	else
	{

		Relation rel = env.lookupRel(node.name, true);
		if (rel == null)
			throw new InterpretError("relation " + node.name + " does not exist.");
		else
		{
			Global.environment.reltable.dump();

			executeJRelixCommand("pr (where .rel_name = \"" + node.name +
			"\" in .rd);");


		}

	}
	break;
      }
        case OP_ST : // Jiantao 2004
        {       if(name != null){
                        Text t = env.lookupText(name, true);
                        if(t == null)
                                throw new InterpretError("text '"+ name + "' does not exist.");
                        else
                                env.textTable.showText(t);
                }
                else
                        env.textTable.showTable(Global.sst);

                break;
        }

      case OP_PRINT : {
	Global.println(name.substring(1, name.length() - 1));
	break;
      }
      case OP_EVENTON : {
	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	if (firstchild.opcode == OP_EVENTNAME) {
	  // event name
	  // firstchild.name = env.getEventName(firstchild);
	  firstchild.name = getEveAtt(node, env);
	  Relation temprel = env.lookupRel(firstchild.name, true);
	  if (temprel != null) {
	    temprel.active = 1;
	    temprel.tree.active = 1;
	  } else
	    throw new InterpretError("Event Handler not found!");
	} else {
	  throw new InterpretError ("No such Event Handler!");
	}
	break;
      }
      case OP_EVENTOFF : {
	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	if (firstchild.opcode == OP_EVENTNAME) {
	  // event name
	  // firstchild.name = env.getEventName(firstchild);
	  firstchild.name = getEveAtt(node, env);
	  Relation temprel = env.lookupRel(firstchild.name, true);
	  if (temprel != null) {
	    temprel.active = 0;
	    temprel.tree.active = 0;
	  } else
	    throw new InterpretError ("Event Handler not found!");
	} else {
	  throw new InterpretError ("No such Event Handler!");
	}
	break;
      }
      case OP_UNDO:
	// handler in the overloaded executeCommand function
	break;
      default : {
	throw new InterpretError("No such command");
      }
      }
  }

  // weizhong
  // get the eventname + attrlist
  private String getEveAtt(SimpleNode node, Environment env)
       throws InterpretError
  {
    SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
    String eaname = new String("");
    SimpleNode attrlist;

    eaname = env.getEventName(nd);
    if (node.jjtGetNumChildren() > 1) {
      attrlist = (SimpleNode) node.jjtGetChild(1);
    } else {
      attrlist = null;
    }
    if (attrlist != null) {
      int childnum = attrlist.jjtGetNumChildren();
      SimpleNode childatt;
      if (childnum != 0) {
	childatt = (SimpleNode)attrlist.jjtGetChild(0);
	eaname = eaname + "[" + childatt.name;
	for (int t = 1; t < childnum; t++) {
	  childatt = (SimpleNode)attrlist.jjtGetChild(t);
	  eaname = eaname + "," + childatt.name;
	}

	eaname = eaname + "]";
      }
    }
    return eaname;
  }

  // Weizhong
  // overload the function to handle the different level's undo command
  // in event handlers
  // except OP_UNDO, all commands still be dealed with in original executeCommand
  private void executeCommand(SimpleNode node, Environment env, String rlname, int opc)
       throws InterpretError
  {
    int opcode = node.opcode;
    TriggerNode triggernode = env.triggerroot.getNode(rlname, opc);

    switch(opcode)
      {
      case OP_UNDO:

	doTrigger(env, triggernode, 1);
	break;
      case OP_PR :
	// to handle print trigger command
	String tmpname = nextTempName();

	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);

	if (firstchild.opcode == OP_EVENTNAME) {
			// event name
	  firstchild.name = env.getEventName(firstchild);
	}


	// print trigger in the current environment.
	if (firstchild.name.compareTo("trigger") == 0) {
	  int tid = printAllTrigger(triggernode);
	  if (tid == 0)
	    System.out.println("TRIGGER of " + rlname + " IS EMPTY.");
	} else {
	  // deal with computation as well as relation.
	  Relation rel = evaluateTLExpression( firstchild, tmpname, new Hashtable(), env);
	  if (rel != null) {
	    rel.print(null, rel.name != tmpname);
	  }
	}
	deleteFile(tmpname);
	break;
      default : {

	executeCommand(node, env);
      }
      }
  }


  // Weizhong
  private int printTrigger(CTrigger trigger)
       throws InterpretError
  {
    if (trigger.relname.compareTo("") == 0)
      return 0;
	else {
	  System.out.println("Trigger of "+trigger.relname);

	  String tempname1 = nextTempName();
	  String tempname2 = nextTempName();
	  Relation temp1 = trigger.oldp.mujoin(OP_IJOIN, null, trigger.newp, null, tempname1);
	  Relation temp2 = trigger.oldp.mujoin(OP_DJOIN, null, temp1, null, tempname2);
	  temp2.print(null, false);

	}
    return 1;
  }


  // Weizhong
  private int printAllTrigger(TriggerNode triggernode)
       throws InterpretError
  {
    int tid = printTrigger(triggernode.trigger);
    int childnum = triggernode.jjtGetNumChildren();
    for (int i = 0; i < childnum; i++) {
      TriggerNode childnode = (TriggerNode) triggernode.jjtGetChild(i);
      tid = tid + printAllTrigger(childnode);
    }
    return tid;
  }

  //wrapper used in Computation.java
  //Yi Zheng May 20, 2002
  public void executeCompDeclaration(SimpleNode node, Environment env)
 	throws InterpretError
  {
  	executeDeclaration(node, env);
  }


     /**
   * JRelixNS modified/enhanced this method:
   *   Workaround for IDLIST update.
   */
  private void executeDeclaration(SimpleNode node, Environment env)
       throws InterpretError
   {
     int opcode = node.opcode;

     switch(opcode)
       {
       case OP_RELATION :
	 RelationalDeclaration(node, env);
	 break;

       // start Jiantao modification 2004
       case OP_TEXT :
         TextDeclaration(node, env);
         break;
       // end Jiantao modification 2004

       case OP_DECLARE :		//June 11,2004 by Yu
	 DeclareDeclaration(node, env);
	 break;

       case OP_VIEW :
	 // remove the identifier node representing the name of the view
	 SimpleNode sn = (SimpleNode) node.jjtRemoveChild(0);
	 // optionally set the name of the root to be the name of view
	 node.name = sn.name;
	 // check if a relation of this name already exists, if so
	 // simply append the tree for view and change the type to
	 // view, otherwise, create a new entry
	 //We allow views/relations of the same name to exist in
	 //different frames of the environment so we use false here
	 Relation rel = env.lookupRel(sn.name, false);
	 if (rel != null)
	   {  // add view node and change type
	     rel.rvc = VIEW;
	     rel.tree = node;
	   }
	 else
	   {  // add new view entry
	     Relation r=new Relation(node.name, null, VIEW,
				     0, 0, 0, node, env);
	     env.put(r);
	   }
	 break;

       case OP_DOMAIN :
   Vector domlist = new Vector();
   Domain[] domarry = null;
   SimpleNode idList;
   SimpleNode [] unionType = null;
   SimpleNode type;
   if(node.jjtGetNumChildren() == 1) // null domain type   by Jiantao 2004
   {      type = new SimpleNode("");
          type.set(OP_TYPE, OP_UNDEF);   
   } 
   else   
   {      type = (SimpleNode) node.jjtGetChild(1); }
   if (node.jjtGetNumChildren() >2 || type.opcode ==OP_IDENTIFIER)	//union type        June 28, 2004 by Yu
   {
    unionType = new SimpleNode[node.jjtGetNumChildren()-1];	
	for(int i = 1; i < node.jjtGetNumChildren(); i++)	//deal with each type
	{
	 SimpleNode type1 = (SimpleNode) node.jjtGetChild(i);
	 if(type1.opcode == OP_IDLIST  && type1.jjtGetNumChildren() > 0) //last condition added by Yi Zheng April 23, 2002
	 {
		throw new InterpretError("The nested domain need a name");
	 			//need a temperary name for the nested relation
		/*SimpleNode z1 = (SimpleNode)node.jjtGetChild(0); //from zhan
        SimpleNode z2 = (SimpleNode)z1.jjtGetChild(0); //from zhan
        String z = z2.name; //from zhan

		String tmpnameR = Domain.nextTempDomainName();    //need a temperary domain name
		Vector domlist1 = new Vector();
	 	Domain[] domarry1 = null;
	 	SimpleNode idList1= (SimpleNode) type1.jjtGetChild(0);
		for(int j = 0; j < idList1.jjtGetNumChildren(); j++)
	    {  
	     			String dname = ((SimpleNode) idList1.jjtGetChild(j)).name;
			      Domain oldD = env.lookupTryDom(dname,z,true);//from zhan May 2003
				    if(oldD == null)
			       {  throw new		//need to change, this is possible June 30, 2004  *****
					    InterpretError("domain "+ dname + " has not been declared");
			       }
			      else if (oldD.isStateDom == 1) //Yi Zheng May 9, 2002
			      			throw new InterpretError("domain " + dname + " is a state domain");
			      		 else
			         		domlist1.addElement(oldD);
	      }
	   		Domain [] domarry2= new Domain[domlist1.size()];
	   		domlist1.copyInto(domarry2);
				if(domlist1 != null)
	  			domarry1 = new Domain[domlist1.size()+1];
	    	else
	  			domarry1 = new Domain[1];
	  		domarry1[0] = env.lookupDom(".id", true);
	  		if(domlist1 != null)
  		{
 				for(int j = 0; j < domlist1.size(); j++)
  				{  domarry1[j+1] = domarry2[j];
  				}
  		}
			//need to save this temp domain  && increase its numref
			Domain tmpDom = new Domain(tmpnameR, IDLIST, domarry2, env);
			tmpDom.numref++;
	        	env.put(tmpDom);
	  
		  	Relation r = new Relation("."+tmpnameR, domarry1, RELATION, 0,
					 domarry1.length, 0, null, env);

	  		env.put(r);
	  		//r.print();
	  		idList1.name = tmpnameR;	//record this temp name in the idlist node
	  		unionType[i-1] = idList1;     */            //idlist node
	 	}
	 	else if (type1.opcode == OP_IDENTIFIER) 
		{
			unionType[i-1] = (SimpleNode) type1.jjtGetChild(0); //keep the identifier   //identifier   numref++ July 02,2004 by Yu
            Domain oldD = env.lookupDom(unionType[i-1].name,true);
			if (oldD!=null) oldD.numref++;
        }
		else	//primitive type
		{
			//int t = type1.opcode - OP_BOOLEAN;
			//String ty = getType(t);
			unionType[i-1] = type1;           //other type(not idlist,identifier)
		}
	}// for
   } //   if (node.jjtGetNumChildren() >2)		//union type        June 28, 2004 by Yu

   else          //1
   {


	 // if domain is of type IDLIST, fill the domlist vector
//	 SimpleNode type = (SimpleNode) node.jjtGetChild(1); // by jian tao
	  if((type.opcode == OP_IDLIST || type.opcode == OP_COMP) && type.jjtGetNumChildren() > 0) //last condition added by Yi Zheng April 23, 2002
	  {

       SimpleNode z1 = (SimpleNode)node.jjtGetChild(0); //from zhan
       SimpleNode z2 = (SimpleNode)z1.jjtGetChild(0); //from zhan
       String z = z2.name; //from zhan

   	   idList = (SimpleNode) type.jjtGetChild(0);
	   for(int i = 0; i < idList.jjtGetNumChildren(); i++)
	   {  String dname = ((SimpleNode) idList.jjtGetChild(i)).name;

	     // Check validation of the dom name. But note, we can't
	     // increment the domain's numref at the moment; We have
	     // to complete _all_ validation check. If the check
	     // fails, an exception is thrown and nothing happens.

	    Domain oldD = env.lookupTryDom(dname,z,true);//from zhan May 2003
	     //Domain oldD = env.lookupDom(dname,true); //Original //zhan

	    // if(env.lookupDom(dname, true) == null) Yi Zheng May 9, 2002
	    if(oldD == null)
	    {  
                /******* start adding by Fan  *******/
                        SimpleNode selfNode = ((SimpleNode) idList.jjtGetChild(i));
                        if(selfNode.type == OP_SELF) {
                                Domain selfdom = env.lookupDom("self",true);
                                if(selfdom == null){
                                        selfdom = new Domain("self", ATTRIBUTE, 0, null);
                                        //selfdom= new Domain();
                                        //selfdom.type = ATTRIBUTE;
                                        //selfdom.name = "self";
                                        env.put(selfdom);
                                }
                                oldD = selfdom;
                        } else
                        /******* end of Fan's addition *******/

            throw new
		    InterpretError("domain "+ dname + " has not been declared");
	    }
	    else if (oldD.isStateDom == 1) //Yi Zheng May 9, 2002
	      	throw new InterpretError("domain " + dname + " is a state domain");
	    else
	    {  //domlist.addElement(env.lookupDom(dname,true)); Yi Zheng May 9, 2002
	         domlist.addElement(oldD);
	    }
	   }
	   domarry= new Domain[domlist.size()];
	   domlist.copyInto(domarry);
	  } //end of if((type.opcode == OP_IDLIST || type.opcode == OP_COMP) && type.jjtGetNumChildren() > 0) //last condition added by Yi Zheng April 23, 2002
   }//end of else 1


	//Yi Zheng April 23, 2002
   if (domarry == null)
		domarry = new Domain[0];
	 // get the identifiers.
   idList = (SimpleNode) node.jjtGetChild(0);
   for(int i = 0; i < idList.jjtGetNumChildren(); i++)
   {  SimpleNode id = (SimpleNode) idList.jjtGetChild(i);

	   //If domain is being used, just skip this single declaration.
	   //We allow domains of the same name to exist in different
	   //frames of the environment so we use false here
	   Domain d = env.lookupDom(id.name, false);
	   if(d != null && d.numref != 0)
	  //   {  Global.println("Error: domain '"+ id.name +
		//	       "' is being used.");
	    // }
	   { throw new                // Zongyan Aug 2002
		    InterpretError ("Error: domain '"+ id.name +
			       "' is being used.");
	   }
   	   else if (node.jjtGetNumChildren() >2 || type.opcode ==OP_IDENTIFIER)		//union type        June 30, 2004 by Yu
	   {
//System.out.println("here!!!");
			Domain newDom = new Domain(id.name, UNION, 0, unionType, env);
	        env.put(newDom);
	   }

	   else if (type.opcode == OP_IDLIST)
	     {  Domain newDom = new Domain(id.name, IDLIST, domarry, env);
	        env.put(newDom);

	     }
	   else if (type.opcode == OP_COMP)
	     {  Domain newDom = new Domain(id.name, COMP, domarry, env);
	     env.put(newDom);
	     }

          /****** start adding by Fan  ******/
          else if (type.opcode == OP_TYPE) // May 2004 Fan Guo
             {  Domain newDom = new Domain(id.name, TYPE, 0, null);
                env.put(newDom);
             }
             /******end of Fan's addition       ******/

	   else 
	     {  // Need to domain validation check & type check.
	       Domain newDom = new Domain
		 (id.name, type.opcode - OP_BOOLEAN, 0, null);
//		System.out.println("Type of "+newDom.name+" is: "+newDom.type);
	       env.put(newDom);
	     }
	   } //end of 	 for(int i = 0; i < idList.jjtGetNumChildren(); i++)

	 break;

       case OP_LET :
 //commented out to use Andrey's code, by Yi Zheng May 26, 2002
		SimpleNode tempnode = node.jjtDeepDuplicate();
		tempnode= (SimpleNode)tempnode.jjtGetChild(0);

                /******* start adding by Fan  *******/  // September 14, 2004
                if(tempnode.opcode==OP_CAST){ //  e.g. let (attr) eval year be titles;
                        SimpleNode left = (SimpleNode)tempnode.jjtGetChild(0);
                        SimpleNode right = (SimpleNode)tempnode.jjtGetChild(1);
                        if(left.opcode == OP_ATTRIBUTE && right.type == OP_EVAL){
                                Domain domList[] = ExpressionListToDomains(right,new Hashtable(),env);
                                SimpleNode EvalChild = (SimpleNode)right.jjtGetChild(0);
                                Domain attriDom = env.lookupDom(EvalChild.name, true);
                                if(attriDom.type != IDLIST) { // eval followed by an attribute
                                        String virdName = domList[0].name+"_evaled";
                                        tempnode = (SimpleNode) node.jjtRemoveChild(0);
                                        SimpleNode newNode = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,virdName);
                                        node.jjtAddChildAtHead(newNode);
                                newNode.jjtSetParent(node);
                                // so far it beocmes "let virdName be ABC".
                                // The name, virdName, part of which is "_evaled", indicates the eval operation
                                // will result in a special operation in actualization
                                AddVirtDom2DomTable(node, env);
                                }
                        }
                } else

                /******* end of Fan's addition *******/

		//This is a static evaluation of a "let" statement
		if(tempnode.opcode==OP_EVAL)
		{
			Domain domList[] = ExpressionListToDomains(tempnode,new Hashtable(),env);

                        /******* start adding by Fan  *******/  // September 14, 2004
                        SimpleNode EvalChild = (SimpleNode)tempnode.jjtGetChild(0);
                        //September 24, 2004
                        //new implementation for "let eval (attri)year be title."
                        //no new domain is created at this moment.
                        //need to check the operand of eval is of type ATTRIBUTE            
                        int evalType = traverseType(EvalChild, env);
                        SimpleNode atr = null;
                        if(evalType == ATTRIBUTE){
                                if(EvalChild.type == OP_IDENTIFIER) {
                                        atr = EvalChild;                                    
                                } else if(EvalChild.type == OP_BIOPERATOR && EvalChild.opcode == OP_CAST){
                                        atr = (SimpleNode) EvalChild.jjtGetChild(1);        
                                }
                                Domain attriDom = env.lookupDom(atr.name, true);
                                String virdName = attriDom.name+"_evaled";
                                tempnode = (SimpleNode) node.jjtRemoveChild(0);
                                SimpleNode newNode = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,virdName);
                                node.jjtAddChildAtHead(newNode);
                        newNode.jjtSetParent(node);
                        AddVirtDom2DomTable(node, env);
                                break;
                        /*
                        Domain attriDom = env.lookupDom(EvalChild.name, true);
                        if(attriDom.type != IDLIST) { // eval followed by an attribute


                                String virdName = domList[0].name+"_evaled";
                                tempnode = (SimpleNode) node.jjtRemoveChild(0);
                                SimpleNode newNode = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,virdName);
                                node.jjtAddChildAtHead(newNode);
                        newNode.jjtSetParent(node);
                        // so far it beocmes "let virdName be ABC".
                        // The name, virdName, part of which is "_evaled", indicates the eval operation and
                        // will result in a special operation in actualization
                        AddVirtDom2DomTable(node, env);
                        */// commented out on September 24, 2004.
                                  // and this part was desinged for query d of A7. but I don't like the implementation. try the above code.
                        } else    // In following Andrey's case, Eval must be followed by a relation
                        /******* end of Fan's addition *******/


			for(int i=0;i<domList.length;i++)
			{
			//	tempnode = node.jjtDeepDuplicate();
				tempnode = (SimpleNode) node.jjtRemoveChild(0);
				SimpleNode newNode = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,
											domList[i].name);
				node.jjtAddChildAtHead(newNode);
		        newNode.jjtSetParent(node);
		//		AddVirtDom2DomTable(node);
		AddVirtDom2DomTable(node, env); //modified by Yi Zheng May 26, 2002
			}
		}

		else
		//	AddVirtDom2DomTable(node);
		AddVirtDom2DomTable(node, env); //modified by Yi Zheng May 26, 2002
		break;

       case OP_COMPUTATION :
	 Computation c = new Computation(node, env);

	 Relation temprel = env.lookupRel(c.name, true);
	 if (temprel != null) {
	 	c = null;
	 	//added by Yi Zheng April 6, 2002
	 	//warning already issued in Global.java
	 } else {
	   // env.put(c);

	   if  (Global.saveCompCode(env) == 1){
	     // for a new event handler, default status is active
	     SimpleNode fchild = (SimpleNode) node.jjtGetChild(0);
	     if (fchild.opcode == OP_EVENTNAME) {
	       c.tree.active = 1;
	       c.active = 1;
	     }
	   }
	   else Global.println("Warning: error occurred while saving the computation code");

	   env.put(c);
	 }
	 break;
       }
   }

  //Original code for processing Virtual Domain declaration
  // A. Rozenberg
  // enhanced with Andy's modification, May 26, 2002
  // parameter env added by Yi Zheng May 26, 2002
  private void AddVirtDom2DomTable(SimpleNode node,  Environment env) throws InterpretError
  {
	 // Get the virtual domain name.
	 String domname = ((SimpleNode)node.jjtGetChild(0)).name;

	 // Remove the first child, the rest of the tree will be
	 // stored in the tree field of this domain.
	 //  Weizhong: keep the original node, and operate on the tempnode
	 SimpleNode tempnode = node.jjtDeepDuplicate();
	 tempnode.jjtRemoveChild(0);
	 // node.jjtRemoveChild(0);


	 // START Since
	     // Enhancements to jRelix76 for newVirDomain.
		if (Relation.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
//	 if (env.isSigmaJoinImplemented != true) {
	   // Add it to the domain table.
	   //  Domain newDom = newVirDomain(domname, node, env);
	   Domain newDom = newVirDomain(domname, tempnode, env);
	   env.put(newDom);
	 }else {
                    // Enhancements to jRelix76.
	   // Add it to the domain table.
	   // Domain newDom = newVirDomain(domname, node, env);
	   // Correct problem with IDLIST update.
	   // Old problem: newVirDomain already updated domain's rel structure.
	   // but 'put' of a domain deletes the relation and doesn't
	   // recreate another one.  The 'save' and 'put' combo will
	   // put in the correct relation so no run-time exceptions
	   // will occur after a IDLIST domain update here.

	  // Domain  dom_saved = env.lookupDom(domname,false); modified by Yi Zheng March 20, 2002
	   Domain newDom = newVirDomain(domname, tempnode, env);
	   //System.out.println("virtual domain " + domname + " is added to env");
	   Relation rel_saved = env.lookupRel("."+domname,true);

          /*
                env.put(newDom);
                // if ((rel_saved != null)&&(dom_saved != null)) modified by Yi Zheng March 20, 2002
           if (rel_saved != null)
             env.put(rel_saved);
         */ // this is oroginal code. Modified to the following code on July 31, 2004 by Fan
           /******* start adding by Fan  *******/
            // July 31, 2004
           if (newDom.type != IDLIST ){ // nested virtual domain and it's dot relation have been saved
                        //|| (newDom.type == IDLIST && env.lookupDom(newDom.name,true) == null)) {
                env.put(newDom);
                if (rel_saved != null)
                        env.put(rel_saved);
           }
           /*
           else if(newDom.type == IDLIST && env.lookupDom(newDom.name,true) != null){
                        Relation dotR = env.lookupRel("."+newDom.name,true);
                        if(dotR.numtuples !=0) {
                                Relation Rcopy = dotR.duplicateLoadAll();
                                env.put(newDom);
                                dotR = env.lookupRel("."+newDom.name,true);
                                env.put(Rcopy);  // Sep. 2, 2004 Fan
                                Rcopy.save();

                        }
           }*/  // if it is not the first time to reach the domain, then do nothing
           /******* end of Fan's addition  *******/


	 }
	 // END Since jRelixNS
  }

   //wrapper used in Computation.java
   //Yi Zheng May 20, 2002
   public void executeCompStatement(SimpleNode node, Environment env)
   	throws InterpretError
   	{
   		executeStatement(node, env);
   	}
  private void executeStatement(SimpleNode node, Environment env)
       throws InterpretError
  {  int opcode = node.opcode;

   switch(opcode)
     {
     case OP_CONDITIONAL :
       SimpleNode ifNode, thenNode, elseNode;
       int numchildren = node.jjtGetNumChildren();
       ifNode = (SimpleNode) node.jjtGetChild(0);
       thenNode = (SimpleNode) node.jjtGetChild(1);

       if (EvalExpr.evalBooleanExpr(ifNode, env)) //short-circuit evaluation
	 {
	   interpret(thenNode, env);
	 }
       else if(numchildren == 3)
	 {
	   elseNode = (SimpleNode) node.jjtGetChild(2);
	   interpret(elseNode, env);
	 }
       break;
     case OP_STMTBLOCK :
     case OP_COMPBLOCK :	// Weizhong, for computation statements
       int numChildren = node.jjtGetNumChildren();
       for (int i=0; i<numChildren; i++)
	 {   interpret((SimpleNode) node.jjtGetChild(i), env);
	 }
       break;
     case OP_COMPCALL :
     	 // added by Zongyan Sep 2002
	  Relation r ;
	  SimpleNode compnode = ((SimpleNode) node.jjtGetChild(0));
	  if (compnode.url != null ) {
	  aldatpClient.request(compnode,env);  // need to check response and return it ^_^
	  if (aldatpClient.rel ==null)
		throw new InterpretError( aldatpClient.rspCode);

	  if (aldatpClient.rel.rvc == 17) { // comp
		 Computation c = new Computation(aldatpClient.rel.tree,env);
		 c.name =aldatpClient.rel.name;
		 r = c;
	  }
	  else 	// regular relation
		   throw new InterpretError(aldatpClient.rel.name  + "is not computation");
   	}
	else

	{
	   // end of Zongyan's addition
       String compName = ((SimpleNode) node.jjtGetChild(0)).name;
       r = env.lookupRel(compName, true);
       if (r == null || r.rvc != COMPUTATION)
	 {  throw new InterpretError("'"+compName+"' is not computation");
	 }
	 }
       String tmpname = nextTempName();
       ((Computation) r).applyInOut(node, tmpname, env);
       deleteFile(tmpname);
       break;
     case OP_SEQUENCE :
     case OP_PARALLEL :
     case OP_CHOICE :
     case OP_FORLOOP :
     case OP_WHILELOOP :
     case OP_EXIT :
     case OP_DEADLOCK :
     case OP_EXEC :
       System.out.println("TO BE IMPLEMENTED");
       break;
     }
  }

  // to get the '.id' col position

  // Weizhong
  private int getIDCol(Relation r)
       throws InterpretError
  {
    for(int i = 0; i < r.numattrs; i++) {
      if ((r.domains[i].name).compareTo(".id") == 0)
			return i;
    }

    // shouldn't be here
    return -1;
  }

  // compare the types of the two relations involved in assignment operation
  // if they are of the same type, return 1, else return 0.
  // Because the right relation is a top level relation, so it doesn't constain domain .id
  // and the left relation is a nested attribute, so it contains domain .id
  // Here we just need to make sure the domains except .id are the same.
  public static int CompareRelType(Relation left, Relation right)
       throws InterpretError
  {
    if (left.isDomainsIn(right.domains) && (left.domains.length == right.domains.length + 1))
      return 1;
    else
      return 0;
  }

  //Yi Zheng June 3, 2002
  //used for comparison between two relations:
  // 1. left and right are both top level
  // 2. left and right are both nested
  // 3. left is nested and right is top level
  public static int CompareRelTypeComp(Relation left, Relation right)
       throws InterpretError
  {
    if (left.isDomainsIn(right.domains) && right.isDomainsIn(left.domains))
      return 1; //for the case when both are top level or both are nested
    else
      return CompareRelType(left, right);
  }

  //wrapper used in Computation.java
  //Yi Zheng May 20, 2002
  public int lookCompUpdate(SimpleNode node, Environment env, Relation prtmp, TriggerNode triggernode)
       throws InterpretError
  	{  return lookUpdate(node, env, prtmp, triggernode);
	}

  // to produce the three parts relation of trigger for the certain update
  // command in the certain environment -- each main update command in one
  // environment; execute this command but not make the result effective in
  // the system tables; trigger is only temperary for this environment.
  //  -- implemented by Weizhong, Aug. 1999
  private int lookUpdate(SimpleNode node, Environment env, Relation prtmp, TriggerNode triggernode)
       throws InterpretError
  {
    int opcode = node.opcode;
    String old_name = nextTempName();
    String rest_name = nextTempName();
    String new_name = nextTempName();
    int diff = 0;	// if oldtrigger != newtrigger, diff = 1;
    // else diff = 0, which means it's a void update
    // then Update command won't be activated.
    int assigntime = 0; 	// keep the times of assignments, for loading rtmp once
    int assignidlist = 0;   // if assign a relation to a nested attribute, set it to 1

    triggernode.setName(((SimpleNode) node.jjtGetChild(0)).name);
    triggernode.opcode = opcode;
    switch(opcode)
	  {
	  case OP_ADD :


	    //This can be mapped to a '<+' stmt
	    //update R add expr;  ---> R <+ expr;
	    Relation add_rl = null;	// the relation to be updated
	    Relation add_rr = null;	// the relation to be added
	    Relation rrtemp = null; 	// a temp relation
	    SimpleNode add_idl = (SimpleNode) node.jjtGetChild(0);

	    String rrtemp_name = nextTempName();

	    if ((add_rl = env.lookupRel(add_idl.name, true)) == null)
	    {
	      if ((add_rl = env.lookupNestedRel(add_idl.name, true)) == null)
			{	throw new InterpretError("Relation '"+add_idl.name+"' does not exist");
			}
		}


	    // add_rl ijoin prtmp (project on add_rl -> trigger.oldp
	    if (prtmp != null) {
	      Domain[] iddom = new Domain[1];
	 //     iddom[0] = env.domtable.getDom(".id");
	 	iddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	      Domain[] namedom = new Domain[1];
	 //     namedom[0] = env.domtable.getDom(add_idl.name);
	      namedom[0] = env.lookupDom(add_idl.name, true); //Yi Zheng June 14, 2002

	//      Domain pid = env.domtable.getDom(".id");
			Domain pid = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	      Relation temp_prtmp = prtmp.duplicate(false);

	      int pidj = temp_prtmp.indexOfDomain(pid);

	      if (pidj != ERROR) {
		Domain[] pdoms = new Domain[temp_prtmp.numattrs -1];
		int pdomid = 0;
		for (int pp = 0; pp < temp_prtmp.numattrs; pp++) {
		  if (pp != pidj) {
		    pdoms[pdomid] = temp_prtmp.domains[pp];
		    pdomid++;
		  }
		}
		temp_prtmp.load();
		temp_prtmp.project(pdoms);
	      }

	      triggernode.trigger.oldp = add_rl.mujoin(OP_IJOIN, iddom, temp_prtmp, namedom, old_name);


	      triggernode.trigger.oldp.load();
	      triggernode.trigger.oldp.project(add_rl.domains);
	      triggernode.trigger.oldp.save();
	      triggernode.trigger.relname = "." + add_idl.name;
	    } else {
	      // add_rl -> trigger.oldp
	      //			triggernode.trigger.oldp = add_rl.mujoin(OP_DJOIN, null, add_rl, null, old_name);
	      triggernode.trigger.oldp = add_rl.project(add_rl.domains, old_name);
	      triggernode.trigger.relname = add_idl.name;
	    }


	    // add_rl djoin trigger.oldp -> trigger.restp
	    triggernode.trigger.restp = add_rl.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, rest_name);



	    // trigger.oldp -> rrtemp which to be updated
	    //		  rrtemp = triggernode.trigger.oldp.mujoin(OP_UJOIN, null, triggernode.trigger.oldp, null, rrtemp_name);

	    rrtemp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains, rrtemp_name);

	    // update rrtemp
	    String add_tempname = nextTempName();
	    add_rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), add_tempname, new Hashtable(), env);

	    String tname4 = nextTempName();
	    if (prtmp != null) {
	      // nested!
	      String tname1 = nextTempName();
	      String tname2 = nextTempName();
	      String tname3 = nextTempName();

	      // get the relation to be added with .id domain
	      // to add domain '.id' for add_rr
	      Relation utmp = rrtemp.mujoin(OP_UJOIN, null, add_rr, null, tname1);
	      add_rr = utmp.mujoin(OP_DJOIN, null, add_rl, null, tname2);

	      //Relation addtemp = rrtemp.mujoin(OP_UJOIN, null, rrtemp, null, tname3);
	      Relation addtemp = rrtemp.project(rrtemp.domains, tname3);
	      int idcol = getIDCol(add_rr);
	      if (idcol == -1)
		throw new InterpretError(".id not found in getIDCol");
	      addtemp.load();
	      add_rr.load();
	      rrtemp.load();
	      long old_surr = 0;
	      for (int i = 0; i < rrtemp.numtuples; i++) {
		int rtid = getIDCol(rrtemp);
		if (rtid == -1)
		  throw new InterpretError(".id not found in getIDCol");
		long surr = ((long[])rrtemp.data[rtid])[i];
		if (old_surr != surr) {
		  old_surr = surr;
		  for (int j=0; j<add_rr.numtuples; j++) {
		    ((long[])add_rr.data[idcol])[j] = surr;
		  }
		  add_rr.assignAdd(null, addtemp, null);
		}
	      }
	      //rrtemp = addtemp.mujoin(OP_UJOIN, null, addtemp, null, tname4);
	      rrtemp = addtemp.project(addtemp.domains, tname4);
	      deleteFile(tname1);
	      deleteFile(tname2);
	      deleteFile(tname3);
	    }
	    else {
	      // no nested
	      // rrtemp = add_rr.mujoin(OP_UJOIN, null, add_rr, null, tname4);
	      // rrtemp.assignAdd(null, add_rr, null);
	      // rrtemp = add_rr.project(add_rr.domains, tname4);

	      rrtemp.load(true);
	      add_rr.load(true);

	      add_rr.assignAdd(null, rrtemp, null);

	      rrtemp.load(true);

	    }

	    // rrtemp -> trigger.newp
	    //		  triggernode.trigger.newp = rrtemp.mujoin(OP_UJOIN, null, rrtemp, null, new_name);
	    triggernode.trigger.newp = rrtemp.project(rrtemp.domains, new_name);

	    // compare the old and new trigger
	    Relation comp = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, tname4);
	    if (comp.numtuples != 0)
	      diff = 1;

	    deleteFile(rrtemp_name);
	    deleteFile(add_tempname);
	    deleteFile(tname4);
	    // apply the trigger to get the current result for next statement
	    doTrigger(env, triggernode, 0);
	    break;

	  case OP_DELETE :
	    // Weizhong, July 2 1999
	    //This cn be mapped to a 'djoin' stmt
	    //update R del expr;  ---> R <- R djoin expr;
	    //NOTE: that is not the same as R <- expr djoin expr; !!WRONG

	    Relation del_rl = null;	// the relation to be updated
	    Relation del_rr = null;	// the relation to be deleted
	    Relation del_temp = null; 	// a temp relation
	    SimpleNode del_idl = (SimpleNode) node.jjtGetChild(0);

	    String deltemp_name = nextTempName();

	    if ((del_rl = env.lookupRel(del_idl.name, true)) == null)
	      if ((del_rl = env.lookupNestedRel(del_idl.name, true)) == null)
		{
		  throw new InterpretError("Relation '"+del_idl.name+"' does not exist");
		}


	    // del_rl ijoin prtmp (project on del_rl) -> trigger.oldp
	    if (prtmp != null) {
	      Domain[] iddom = new Domain[1];
	   //   iddom[0] = env.domtable.getDom(".id");
	   iddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14,2002
	      Domain[] namedom = new Domain[1];
	   //   namedom[0] = env.domtable.getDom(del_idl.name);
	      namedom[0] = env.lookupDom(del_idl.name, true); //Yi Zheng June 14, 2002

	  //    Domain dpid = env.domtable.getDom(".id");
	  	Domain dpid = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	      Relation dtemp_prtmp = prtmp.duplicate(false);


	      int dpidj = dtemp_prtmp.indexOfDomain(dpid);

	      if (dpidj != ERROR) {
		Domain[] pdoms = new Domain[dtemp_prtmp.numattrs -1];
		int pdomid = 0;
		for (int pp = 0; pp < dtemp_prtmp.numattrs; pp++) {
		  if (pp != dpidj) {
				pdoms[pdomid] = dtemp_prtmp.domains[pp];
				pdomid++;
		  }
		}
		dtemp_prtmp.load();
		dtemp_prtmp.project(pdoms);
	      }

	      triggernode.trigger.oldp = del_rl.mujoin(OP_IJOIN, iddom, dtemp_prtmp, namedom, old_name);

	      triggernode.trigger.oldp.load();
	      triggernode.trigger.oldp.project(del_rl.domains);
	      triggernode.trigger.oldp.save();
	      triggernode.trigger.relname = "." + del_idl.name;
	    } else {
	      // del_rl -> trigger.oldp
	      triggernode.trigger.oldp = del_rl.project(del_rl.domains, old_name);
	      triggernode.trigger.relname = del_idl.name;
	    }


	    // del_rl djoin trigger.oldp -> trigger.restp
	    triggernode.trigger.restp = del_rl.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, rest_name);



	    // trigger.oldp -> del_temp which to be updated
	    del_temp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains, deltemp_name);

	    // get del_rr which to be deleted
	    String del_tempname = nextTempName();
	    del_rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), del_tempname, new Hashtable(), env);

	    // del_temp djoin del_rr -> trigger.newp
	    String dtname1 = nextTempName();

	    Relation del_rtmp = del_temp.mujoin(OP_DJOIN, null, del_rr, null, dtname1);
	    Domain[] iddoms = new Domain[1];
//	    iddoms[0] = env.domtable.getDom(".id"); //modified by Yi Zheng June 12, 2002
	 	iddoms[0] = env.lookupDom(".id", true);
	    if (iddoms != null) {
	      del_rtmp.load();
	      if (del_rtmp.isDomainIn(iddoms[0]))
		del_rtmp.swapColumns(iddoms);
	    }

	    triggernode.trigger.newp = del_rtmp.project(del_rtmp.domains, new_name);


	    // compare the old and new trigger
	    String dtname = nextTempName();
	    Relation dcomp = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, dtname);
	    if (dcomp.numtuples != 0 || triggernode.trigger.newp.numtuples != triggernode.trigger.oldp.numtuples) {
	      diff = 1;
	    }

	    deleteFile(deltemp_name);
	    deleteFile(del_tempname);
	    deleteFile(dtname);
	    deleteFile(dtname1);

	    // apply the trigger to get the current result for next statement
	    doTrigger(env, triggernode, 0);

	    break;
	  case OP_CHANGE :
	    Relation rel = null; //The relation to be updated
	    Relation real_old = null; // old version of subset of rel
	    Relation temp_real_old = null;
	    Relation rtmp = null;	// the relation for changing
	    Relation rrest = null;	// the rest of source relation
	    Relation rtmp_copy = null;	// a temp copy of rtmp before
	    // discarding the joined domain
	    Relation rtmp_copy2 = null;  // a final copy of rtmp before
	    // discarding the joined domain

	    SimpleNode id = (SimpleNode) node.jjtGetChild(0);
	    SimpleNode stmtList = (SimpleNode) node.jjtGetChild(1);
	    triggernode.trigger.relname = new String(""); // in case, no operate on current level

	    String fname1 = nextTempName();
	    String fname2 = nextTempName();
	    String fname3 = nextTempName();
	    String fname4 = nextTempName();
	    String fname5 = nextTempName();
	    String fname6 = nextTempName();	// for rtmp_copy
	    String fname7 = nextTempName();      // for rtmp_copy2
	    String fname8 = nextTempName();
	    String fname9 = nextTempName();      // for real_old
	    String fname10 = nextTempName();

	    SimpleNode usingExpr; //this is optional..
	    Relation reluser;
	    NREnvironment NRenv;


	    if( (rel = env.lookupRel(id.name, true)) == null){
	      // extract the nested relation
		      if ((rel = env.lookupNestedRel(id.name, true)) == null)
			{	throw new InterpretError("Relation '"+id.name+"' does not exist");
			}
			else if(DEBUG)
			{
			}
	    }else if(DEBUG)
	    {
	    }

	    //check if rel is extracted from a nested relation
	    //if so we can't just replace the current entry of rel in the environment
	    //instead, we need to append the updated result to the dot relation
	    //from which rel is extracted
	    //--Yi Zheng June 10, 2002

	    //Handle using clause first..
	    usingExpr = null; //this is optional..
	    if(node.jjtGetNumChildren() == 3)
	      {
		usingExpr = (SimpleNode) node.jjtGetChild(2);

		SimpleNode rrel;
		SimpleNode ldom;
		SimpleNode rdom;
		Domain[] ldomain;
		Domain[] rdomain;

		if (usingExpr.jjtGetNumChildren() == 3)
		  {

		    ldom = (SimpleNode) usingExpr.jjtGetChild(0);
		    rdom = (SimpleNode) usingExpr.jjtGetChild(1);
		    rrel = (SimpleNode) usingExpr.jjtGetChild(2);
		    ldomain = ExpressionListToDomains(ldom);
		    rdomain = ExpressionListToDomains(rdom);

		  } else {
		    ldomain = null;
		    rdomain = null;
		    rrel = (SimpleNode) usingExpr.jjtGetChild(0);
		  }

		reluser = null;
		reluser = evaluateTLExpression(rrel, fname5, new
					       Hashtable(), env);
		rtmp = rel.mujoin(usingExpr.opcode, ldomain, reluser,rdomain, fname1);
		rtmp.load();


		// make a copy to keep the rtmp with the joined domains
		// for virtual domain uses, e.g. Type <- NewType
		rtmp.swapColumns(rel.domains);
		rtmp.save();
		rtmp.load();



		Domain[] riddoms = new Domain[1];
		riddoms[0] = env.lookupDom(".id", true); //modified by Yi Zheng June 10, 2002
		if (riddoms != null) {
		  rtmp.load();
		  if (rtmp.isDomainIn(riddoms[0]))
		    rtmp.swapColumns(riddoms);
		}
		rtmp.save();
		rtmp.load();

		rtmp_copy = rtmp.project(rtmp.domains, fname6);
		rtmp = rtmp_copy.project(rel.domains, fname8);

		// rtmp may contain new tuples beyond rel (after ujoin)
		// so we need to keep a real_old version of data
		real_old = rtmp.mujoin(OP_IJOIN, null, rel, null,
				       fname9);
		real_old.load();

	      } else {
		// no using option, copy the source relation to rtmp
		rtmp = rel.project(rel.domains, fname1);
		real_old = rel.project(rel.domains, fname9);
	      }

	    // rtmp ijoin prtmp (project on rtmp) -> triggernode.trigger.oldp
	    if (prtmp != null) {
	      Domain[] iddom = new Domain[1];
	      Domain[] namedom = new Domain[1];
	  //    iddom[0] = env.domtable.getDom(".id");
	  		iddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	  //    namedom[0] = env.domtable.getDom(id.name);
	  		namedom[0] = env.lookupDom(id.name, true); //Yi Zheng June 14, 2002

	  //    Domain pcid = env.domtable.getDom(".id");
	  	Domain pcid = env.lookupDom(".id", true);//Yi Zheng June 14, 2002
	      Relation tempc_prtmp = prtmp.duplicate(false);


	      int pcidj = tempc_prtmp.indexOfDomain(pcid);

	      if (pcidj != ERROR) {
		Domain[] pdoms = new Domain[tempc_prtmp.numattrs -1];
		int pdomid = 0;
		for (int pp = 0; pp < tempc_prtmp.numattrs; pp++) {
		  if (pp != pcidj) {
		    pdoms[pdomid] = tempc_prtmp.domains[pp];
		    pdomid++;
		  }
		}
		tempc_prtmp.load();
		tempc_prtmp.project(pdoms);
	      }

	      triggernode.trigger.oldp = rtmp.mujoin(OP_IJOIN, iddom, tempc_prtmp, namedom, old_name);


	      triggernode.trigger.oldp.load();


	      triggernode.trigger.oldp.project(rtmp.domains);

	      if (real_old != null){
		temp_real_old =
		  real_old.mujoin(OP_IJOIN,iddom,tempc_prtmp,namedom,fname10);
		temp_real_old.load();
		real_old =
		  temp_real_old.project(temp_real_old.domains,fname9);
		real_old.load();
	      }
	    } else {
	      // rtmp -> trigger.oldp
	      triggernode.trigger.oldp = rtmp.project(rtmp.domains, old_name);
	    }

	    // triggernode.trigger.oldp -> rtmp
	    rtmp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains, fname1);
	    //		   rtmp.save();
	    rtmp.load();

	    // if (rtmp_copy != null) then rtmp_copy ijoin rtmp -> rtmp_copy2 which includes virtual domains
	    if (rtmp_copy !=null) {
	      rtmp_copy.load();
	      rtmp_copy2 = rtmp_copy.mujoin(OP_IJOIN, null, rtmp, null, fname7);

	    }
	    // rel djoin triggernode.trigger.oldp/rtmp -> triggernode.trigger.restp
	    triggernode.trigger.restp = rel.mujoin(OP_DJOIN, null, rtmp, null, rest_name);

	    //Run each statement in 'stmtList'
	    int numStmts = stmtList.jjtGetNumChildren();

	    for(int i = 0; i < numStmts; i++)
	      {	// for statements

		SimpleNode stmt = (SimpleNode) stmtList.jjtGetChild(i);

		switch(stmt.type)
		  {  // switch stmt type
		  case OP_ASSIGNMENT :
		    // update rtmp
		    // set the current trigger under the current node
		    if (prtmp != null)
		      triggernode.trigger.relname = "." + id.name;
		    else
		      triggernode.trigger.relname = id.name;


		    if (assigntime == 0) {
		      // switch rtmp and rtmp_copy2 if necessary
		      if (rtmp_copy2 != null) {
			Relation temp = null;
			String tempfname = nextTempName();
			// rtmp_copy2 -> temp
			temp = rtmp_copy2.project(rtmp_copy2.domains, tempfname);
			// rtmp -> rtmp_copy2
			rtmp_copy2 = rtmp.project(rtmp.domains, rtmp_copy2.name);
			// temp -> rtmp
			rtmp = temp.project(temp.domains, rtmp.name);

			deleteFile(tempfname);
		      }
		      rtmp.load();
		    }


		    assigntime++;
		    //Initialize the NREnvironment
		    NRenv = env.addNRFrame(rtmp);
		    int nrow = rtmp.numtuples; // avoiding changing rtmp below
		    for(NRenv.row = 0; NRenv.row < nrow; NRenv.row++)
		      {	// for each row

			switch(stmt.opcode)
			  {  // switch stmt opcode
			  case OP_ASSIGN :

			    //REVISIT: I have only handled the case where numchildren == 2
			    // So: R <- S;
			    // But not: R [A: <- :B] S;
			    // Perhaps this last form should not be allowed in an update
			    if(stmt.jjtGetNumChildren() == 2)
			      {  // if 2
				SimpleNode left = (SimpleNode) stmt.jjtGetChild(0);
				SimpleNode expr = (SimpleNode) stmt.jjtGetChild(1);
				NRInfo nri = env.lookupNR(left.name, false);
				if (nri == null)
				  {  throw new InterpretError("'"+left.name+
					 "' is not a domain of "+rtmp.name);
				  }
				int col = nri.index;
				NREnvironment nre = nri.NRenv;

				switch(nri.type)
				  {  // switch 3
				  case BOOLEAN :
				    //REVISIT: implement this case..
				    if(EvalExpr.evalBooleanExpr(expr, env))
				      ((byte[]) nre.data[col])[nre.row] = 1;
				    else
				      ((byte[]) nre.data[col])[nre.row] = 0;
				    break;
				  case SHORT :
				    if (EvalExpr.evalShortExpr(expr,env) != SHORT_DC) {
				      ((short[]) nre.data[col])[nre.row] =
					EvalExpr.evalShortExpr(expr, env);
				    }
				    break;
				  case INTEGER :
				  //Yi Zheng April 18, 2002
				  EvalExpr.WEIZHONG = true;
				  int iv = EvalExpr.evalIntExpr(expr, env);
				  EvalExpr.WEIZHONG = false;
			 if (iv != INT_DC) {
				      ((int[]) nre.data[col])[nre.row] =
				iv;
			//	    if (EvalExpr.evalIntExpr(expr, env) != INT_DC) {
			//	      ((int[]) nre.data[col])[nre.row] =
			//		EvalExpr.evalIntExpr(expr, env);
		//		System.out.println("new value of " + left.name + " is: " + ((int[]) nre.data[col])[nre.row]);
				    }
				    break;

				  case LONG :
				    if (EvalExpr.evalLongExpr(expr, env) != LONG_DC) {
				      ((long[]) nre.data[col])[nre.row] =
					EvalExpr.evalLongExpr(expr, env);
				    }
				    break;
				  case FLOAT :
				    if (EvalExpr.evalFloatExpr(expr, env) != FLOAT_DC) {
				      ((float[]) nre.data[col])[nre.row] =
					EvalExpr.evalFloatExpr(expr, env);
				    }
				    break;
				  case DOUBLE :
				    if (EvalExpr.evalDoubleExpr(expr, env) != DOUBLE_DC) {
				      ((double[]) nre.data[col])[nre.row] =
					EvalExpr.evalDoubleExpr(expr, env);

				    }
				    break;
				    case NUMERIC ://from Yiyi
				      if(EvalExpr.evalNumExpr(expr,env).compareTo(NUMERIC_DC) != 0){
					  ((number[])nre.data[col])[nre.row] =
					      EvalExpr.evalNumExpr(expr,env);
				      }
				      break;

		 case UNIVERSAL ://from Andrey
			case ATTRIBUTE: //from Andrey
				  case STRING :
				    if (EvalExpr.evalStringExpr(expr, env) != STRING_DC) {
				      ((String[]) nre.data[col])[nre.row] =
					EvalExpr.evalStringExpr(expr, env);
				    }
				    break;
				  case IDLIST :
				    // Remember these are nested relations.
				    // assign the new relation to the sub-relation got in 1.

				    Relation nestrtmp = null;
				    String ttmpname1 = nextTempName();
				    String ttmpname2 = nextTempName();
				    String ttmpname3 = nextTempName();
				    String ttmpname4 = nextTempName();
				    long surrogate;
				    Domain[] iiddom = new Domain[1];
                     //              iiddom[0] = env.domtable.getDom(".id");
				   iiddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
				   assignidlist = 1;


				   // extract the nested relation
				   if ((rel = env.lookupNestedRel(left.name, true)) == null)
	//		{	throw new InterpretError("Nested Relation '"+id.name+"' does not exist");}
			{	throw new InterpretError("Nested Relation '"+left.name+"' does not exist");} //Yi Zheng June 2, 2002

				   if ((surrogate = env.getSurrogate(left.name)) == -1) {
				throw new InterpretError("Surrogate not found.");
				   }
				   if( (nestrtmp = env.lookupRelSurr(left.name, true)) == null)
				     {  throw new InterpretError(
						  "Relation '"+left.name+"' does not exist");//changed from id.name by Yi Zheng June 2, 2002
				     }

				   rrest = rel.mujoin(OP_DJOIN, null, nestrtmp, null, ttmpname4);

				   Relation rtmp_tmp1 = evaluateTLExpression(expr, ttmpname1, new Hashtable(), env);

				   // check whether nestrtmp and rtmp_tmp1 is the same type relation (attributes are of same type)
				   // if not, throw an error
				   if (CompareRelType(nestrtmp, rtmp_tmp1) != 1)
				     throw new InterpretError("The assignment is not allowed since the two operand relations are not of the same type!");

                                   // project nestrtmp on .id
				   nestrtmp.project(iiddom);

				   // (add domain .id to rtmp_tmp1) -> rtmp_tmp2
				   Relation rtmp_tmp2  = nestrtmp.mujoin(OP_IJOIN, null, rtmp_tmp1, null, ttmpname2);


				   if (nestrtmp.name == "")
				     nestrtmp.name = "Temp";
				   String nestrtmp_name = nestrtmp.name;
		   		   env.removeRel(nestrtmp_name, false);
		   		   // rename back to nestrtmp relation
		   		   env.put(rtmp_tmp2.assign(null, nestrtmp_name, null));
				   nestrtmp = env.lookupRel(nestrtmp_name, true);
				   nestrtmp.load();
				   if (triggernode.trigger.newp.numtuples == 0)
				     triggernode.trigger.newp = nestrtmp.project(nestrtmp.domains, new_name);
                                   else {
				     triggernode.trigger.newp = triggernode.trigger.newp.mujoin(OP_UJOIN, null, nestrtmp, null, new_name);
				     //				  	 triggernode.trigger.newp.project(triggernode.trigger.newp.domains);
				   };
				   deleteFile(ttmpname1);
				   deleteFile(ttmpname2);
				   deleteFile(ttmpname3);
				   deleteFile(ttmpname4);
				   break;
				  case COMP :
				    // impossible to be here, Weizhong
				    throw new
				      InterpretError("Type 'comp' are not yet supported here!");
				  case TEXT :
				  case STMT :
				  case EXPR :
				    throw new
				      InterpretError("Types 'text', 'stmt', and 'expr' " +
						     "are not yet supported in computations");
				  default :
				    throw new InterpretError("Unknown type");
				  }
				// end of switch 3
			      } // end of if 2
			    break;
			  default :
			    throw new InterpretError ("Statement type not supported in an update");
			  }
			// end of switch stmt opcode
		      } // end of each row

			rtmp.save();
		    rtmp.save(); //Yi Zheng June 10, 2002
		    rtmp.load();
		    // reset NRenv.row for late use
		    env.removeNRFrame();
		    if (assignidlist == 0) {

		      // adjust rtmp domains if necessary
		      if (rtmp_copy2 != null) {
			rtmp = rtmp.project(rtmp_copy2.domains, rtmp.name);
		      } else {
			rtmp.save();
					}

		      // adjust new and old back, to keep the original old piece
		      if (assigntime >1)
			doTrigger(env, triggernode, 0);

		      // rtmp -> newp
		      rtmp.load();
		      triggernode.trigger.newp = rtmp.project(rtmp.domains, new_name);

		    } else {
		      SimpleNode nestid = (SimpleNode) stmt.jjtGetChild(0);
		      Domain[] idom = new Domain[1];
		 //     idom[0] = env.domtable.getDom(".id");
		 		idom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
		      Domain[] nndom = new Domain[1];
		 //     nndom[0] = env.domtable.getDom(nestid.name);
		      nndom[0] = env.lookupDom(nestid.name, true); //Yi Zheng June 14, 2002
		      triggernode.trigger.oldp = triggernode.trigger.oldp.mujoin(OP_IJOIN, nndom , rel, idom, old_name);
		      triggernode.trigger.oldp = triggernode.trigger.oldp.project(rel.domains, old_name);
		      triggernode.trigger.restp = rel.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, rest_name);
		      assignidlist = 0;
		      triggernode.trigger.relname = "." + nestid.name;

		    };


		    // compare the old and new trigger
		    String ctname4 = nextTempName();
		    Relation compc = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, ctname4);
		    if (compc.numtuples != 0) {
		      diff = 1;
		    }

		    deleteFile(ctname4);
		    		    // reset NRenv.row for late use
		//    env.removeNRFrame();

		    // apply the trigger to get the current result for next statement
		    doTrigger(env, triggernode, 0);

		    break;
		  case OP_UPDATE:
		    // nested

		    TriggerNode childnode = new TriggerNode(env, triggernode);

		    triggernode.jjtAddChild(childnode);

		    int stmt_diff = lookUpdate(stmt, env, rtmp, childnode);
		    if (stmt_diff == 1 || diff == 1)
		      diff = 1;
		    else
		      diff = 0;
		    break;

		    /**********************COMPUTATION*************************/
		  case OP_STATEMENT :
		    switch(stmt.opcode)
		      {  // switch 4
		      case OP_COMPCALL :
			triggernode.trigger.relname = id.name;

			String compName = ((SimpleNode) stmt.jjtGetChild(0)).name;

			Domain[] domss = new Domain[1];
			Domain  d = env.lookupDom(compName, true);

			if(d != null)
		 		domss[0] = d;
			else throw new InterpretError("'" + compName + "' is not a valid computation");

			String reltempname = nextTempName();
			Relation temprel = rel.project(domss, reltempname);
			temprel.load();
			long domsur = ((long[]) temprel.data[0])[0];

			Computation  r = comptable.lookup(domsur);
			if (r == null)
			  {  throw new InterpretError("'"+compName+"' is not a valid nested computation for the relation to be changed!");
			  }


			// To get the relation without domains of type COMP
			int j = 0;

			for(int ip = 0; ip < rel.domains.length; ip++)
			  {
			    if (rel.domains[ip].type !=  COMP){
			      j = j + 1;
			    }
			  }

			Domain[] realdomains = new Domain[j];
			Domain[] compdomains = new Domain[rel.domains.length - j];

			int q = 0;
			int z = 0;

			for (int ipp = 0; ipp < rel.domains.length; ipp++)
			  {
			    if (rel.domains[ipp].type != COMP){
			      realdomains[q] = rel.domains[ipp];
			      q = q + 1;
			    } else {
			      compdomains[z] = rel.domains[ipp];
			      z = z + 1;
			    }
			  }

			String reltempname1 = nextTempName();
			Relation temprel1 = rtmp.project(realdomains, reltempname1);
			temprel1.load();

			// Begin to apply this computation statement
			NREnvironment compNRenv;
			compNRenv = env.addNRFrame(temprel1);

			((Computation) r).applyInOut(temprel1, stmt, env);
			rtmp.load();

			String stemp1 = nextTempName();

			Relation rtmp_s1 = rtmp.project(compdomains, stemp1);
			rtmp_s1.load();

			//rtmp = temprel1.mujoin(OP_UJOIN, null, rtmp_s1, null, rtmp.name); Yi Zheng May 1, 2002
			rtmp = temprel1.mujoin(OP_IJOIN, null, rtmp_s1, null, rtmp.name);

			rtmp.load();

			triggernode.trigger.newp = rtmp.project(rtmp.domains, new_name);

			deleteFile(reltempname);
			deleteFile(reltempname1);
			deleteFile(stemp1);

			break;
		      default :
			throw new InterpretError
			  ("Statement type not supported in an update");
		      }
		    // end of switch 4
		    break;
		    /***************************************************************/

		  default :
		    throw new InterpretError("Statement type not supported in an update");
		  }
		// end of switch stmt type
	      } // end of for statements

	    if (assigntime != 0) {
	      // reverse the original old piece
	      doTrigger(env, triggernode, 0);

	      rtmp.save();
	      rtmp.load();

	      // adjust oldp, newp, and rest
	      // in case A<- if B then A else C;

	      // restp = (newp ijoin oldp) ujoin restp
	      triggernode.trigger.oldp.load();
	      String riname = nextTempName();
	      Relation ri = null;
	      ri = triggernode.trigger.newp.mujoin(OP_IJOIN, null, triggernode.trigger.oldp, null, riname);
	      // ri = rtmp.mujoin(OP_IJOIN, null, triggernode.trigger.oldp, null, riname);

	      triggernode.trigger.restp = triggernode.trigger.restp.mujoin(OP_UJOIN, null, ri, null, rest_name);

	      deleteFile(riname);

	      // oldp djoin restp = oldp
	      String tempoldname = nextTempName();
	      String tempoldname1 = nextTempName();
	      Relation temp_old = null;
	      Relation temp_old1 = null;

	      temp_old = triggernode.trigger.oldp.mujoin(OP_DJOIN, null, triggernode.trigger.restp, null, tempoldname);
	      //temp_old.printdata();
	      //rel.printdata();
	      temp_old1 = temp_old.mujoin(OP_IJOIN, null,
					  real_old, null, tempoldname1);
	      triggernode.trigger.oldp = null;
	      triggernode.trigger.oldp = temp_old1.project(temp_old.domains, old_name);

	      deleteFile(tempoldname);
	      deleteFile(tempoldname1);

	      // newp djoin restp = newp
	      String tempnewname = nextTempName();

	      Relation temp_new = null;
	      temp_new = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.restp, null, tempnewname);

	      triggernode.trigger.newp = temp_new.project(temp_new.domains, new_name);
	      deleteFile(tempnewname);

	      // apply the change for this level's assignments
	      doTrigger(env, triggernode, 0);
	    }


	    // clear the temp relation
	    deleteFile(fname1);
	    deleteFile(fname2);
	    deleteFile(fname3);
	    deleteFile(fname4);
	    deleteFile(fname5);
	    deleteFile(fname6);
	    deleteFile(fname7);
	    deleteFile(fname8);
	    deleteFile(fname9);
	    deleteFile(fname10);
	    break;
	    // end of OP_CHANGE
	  default : // impossible to here
	    System.out.println("unknown update statement");
	    break;
	  }
    // end of switch


    // clear up empty trigger file
    if (triggernode.trigger.relname.compareTo("") == 0) {
      deleteFile(old_name);
		deleteFile(new_name);
		deleteFile(rest_name);
    }

    return diff;

  }

  //wrapper used in Computation.java

  //Yi Zheng May 20, 2002
  public int executeCompUpdate(SimpleNode root, SimpleNode node, Environment env)
       throws InterpretError
  { return executeUpdate(root, node, env);
  }
  // implemented by Weizhong Sun, Aug 1999
  private int executeUpdate(SimpleNode root, SimpleNode node, Environment env)
       throws InterpretError
  {  int opcode = node.opcode;
  int assignDone = 0;	// to control the assignments in the same level
  // make sure only do trigger once for all the
  // assignments of one level

  int preevent = executeEventH("pre", node, env);

  switch(opcode)
    {  case OP_ADD :
    case OP_DELETE :
      // Weizhong, July 2 1999
      SimpleNode nid = (SimpleNode) node.jjtGetChild(0);
    String rname = nid.name;
    TriggerNode triggernode = env.triggerroot.getNode(rname, opcode);
    doTrigger(env, triggernode, 0);
    break;
    case OP_CHANGE :
      SimpleNode id = (SimpleNode) node.jjtGetChild(0);
      SimpleNode stmtList = (SimpleNode) node.jjtGetChild(1);

      int numStmts = stmtList.jjtGetNumChildren();
      for(int i = 0; i < numStmts; i++)
	{  // for each statement
	  SimpleNode stmt = (SimpleNode) stmtList.jjtGetChild(i);
	  switch(stmt.type)
	    {  // switch stmt type

	    case OP_ASSIGNMENT :
	      if (assignDone == 0) {
		TriggerNode trignode = env.triggerroot.getNode(id.name, opcode);
		doTrigger(env, trignode, 0);
		assignDone = 1;	// don't do trigger again on other assignments
		// of the same level update
	      }
	      break;
	    case OP_UPDATE:
	      // nested
	      int exeEventH = executeUpdate(root, stmt, env);

	      break;
	    default :
	      throw new InterpretError("Statement type not supported in an update");
	    }
	  // end of switch stmt type
	} // end of each statements
      break;
      // end of OP_CHANGE
    default : // impossible to here
      System.out.println("unknown update statement");
      break;
    }
  // end of switch


  int postevent = executeEventH("post", node, env);

  if (preevent == 1 || postevent == 1)
    return 1;
  return 0;
  }

	//wrapper used in Computation.java
	//Yi Zheng May 20, 2002
  public void doCompTrigger(Environment env, TriggerNode triggernode, int searchchild)
       throws InterpretError
	{ doTrigger(env, triggernode, searchchild);
	}

  // use the env.trigger to produce the new updated relation, and switch
  // old/new relation in env.trigger -- for update/undo command
  // implemented by Weizhong, Aug. 1999
  private void doTrigger(Environment env, TriggerNode triggernode, int searchchild)
       throws InterpretError
  {

    if (triggernode.trigger.relname.compareTo("") != 0) {


      String ctmpname = nextTempName();
	String utmpname1 = nextTempName();
	String utmpname2 = nextTempName();
	String utmpname3 = nextTempName();
	String tmpname1 = nextTempName();
	String tmpname2 = nextTempName();
	String tmpname3 = nextTempName();


	String old_name = triggernode.trigger.oldp.name;
	String new_name = triggernode.trigger.newp.name;
	String rest_name = triggernode.trigger.restp.name;

	String mytempnewname = nextTempName();
	String mytempoldname = nextTempName();
	String mytemprestname = nextTempName();

	// rest U new
	// if the relation contains computation as attribute, we project oldp, newp, restp to realdomains

	int j = 0;
	for (int ip = 0; ip < triggernode.trigger.newp.domains.length; ip++)
	  {
	    if (triggernode.trigger.newp.domains[ip].type != COMP){
	      j = j + 1;
	    }
	  }

	Domain[] realdomains = new Domain[j];
	Domain[] compdomains = new Domain[triggernode.trigger.newp.domains.length - j];
	int q = 0;
	int z = 0;
	for (int ipp = 0; ipp < triggernode.trigger.newp.domains.length; ipp++)
	  {
		if (triggernode.trigger.newp.domains[ipp].type != COMP) {
		  realdomains[q] = triggernode.trigger.newp.domains[ipp];
		  q = q + 1;
		} else {
		  compdomains[z] = triggernode.trigger.newp.domains[ipp];
		  z = z + 1;
		}
	  }

if (z != 0) {
	  Relation mytempnew = triggernode.trigger.newp.project(realdomains, mytempnewname);
	  Relation mytempold = triggernode.trigger.oldp.project(realdomains, mytempoldname);
	  Relation mytemprest = triggernode.trigger.restp.project(realdomains, mytemprestname);

	  //mytempnew.printdata();
	  //mytempold.printdata();
	  //mytemprest.printdata();

	  triggernode.trigger.oldp.load();
	  triggernode.trigger.oldp.swapColumns(compdomains);

  Relation onlycomp = triggernode.trigger.oldp.project(compdomains, ctmpname);
	  onlycomp.load();


	  Relation utmp1 = mytempnew.mujoin(OP_UJOIN, null, mytemprest, null, utmpname1);

	  //utmp1.printdata();

	  utmp1.load();
//Yi Zheng May 1, 2002 Relation utmp2 = utmp1.mujoin(OP_UJOIN, null, onlycomp, null, utmpname2);

	  Relation utmp2 = utmp1.mujoin(OP_IJOIN, null, onlycomp, null, utmpname2);


	 //utmp2.printdata();

	  utmp2.load();
	  utmp2.save();

	  //restore the value of comp domains into utmp2
	  for (int it = 0; it < utmp2.numtuples; it++)
	    {  for (int jt = 0; jt < onlycomp.numattrs; jt++)
	      {
		long tll = ((long[])onlycomp.data[jt])[0];
		int coll = q + jt;
		((long[])utmp2.data[coll])[it] = tll;
	      }
	    }

	 //May 1, 2002
	 utmp2.save();
	 utmp2.load();



	  Relation utmp3 = utmp2.project(utmp2.domains, utmpname3);


	  //utmp3.printdata();



	  // update the relation in system.
	  	String rname = triggernode.trigger.relname;

	  	//addition by Yi Zheng June 11, 2002
	  IDInfo idi = env.lookup(rname, true);
	  if(idi == null)
	  	throw new InterpretError("relation " + rname + " not found"); //Yi Zheng June 14, 2002

	  if(!(idi instanceof NRInfo))
	  {//original code
	  	//env.removeRel(rname, false);
	  	//env.put(utmp3.assign(null, rname, null));

	  		if(idi instanceof RelInfo)            
	  		{                                     
	  			if(rname.startsWith("."))
	  			{
	  				Environment envFound = env.removeRel2(rname, true);
	 				if(envFound != null)
						env.put2(utmp3.assign(null, rname, null), envFound);
					else throw new InterpretError("no environment found for " + rname);
	  			}                             
	  			else        
	  			{
	  				env.removeRel(rname, false);
	  				env.put(utmp3.assign(null, rname, null));
	  			}
	  		}else
	  		{
                        // added 2 lines by Zongyan for computation update , the updated relation is a parameter of the computation  -- Nov 25, 2004
	  			env.removeRel(rname, false);
	  			env.put(utmp3.assign(null, rname, null));


	  		}

	  }else if(((NRInfo)idi).type == IDLIST)
	  {
	  	String relName = "." + rname;
	  	Relation dotRel = env.lookupRel(relName, true);
	  	if(dotRel != null)
	  	{
	  		long oldSur = ((NRInfo)idi).NRenv.getSurrogate(rname);
	  		if(oldSur != -1)
	  		{
	  				dotRel = dotRel.getRelExceptSurr(oldSur, dotRel.name);//CONTINUE HERE


	  			dotRel.insert(oldSur, utmp3);
	  			if(dotRel.data == null)
	  				dotRel.load();
	  			dotRel.project(null);
	  			dotRel.save();

				//System.out.println("removing rel: "  + relName);
	  			env.removeRel(relName, true);
	  			env.put(relName, new RelInfo(dotRel), true);
	  			dotRel.save();
	  		}
	  	}
	  	else throw new InterpretError("relation " + relName + " not found");

	  }//end of addition June 11, 2002



	  // old -> new

//May 1, 2002	  triggernode.trigger.newp = mytempold.mujoin(OP_UJOIN, null, onlycomp, null, new_name);
	  triggernode.trigger.newp = mytempold.mujoin(OP_IJOIN, null, onlycomp, null, new_name);

	  // tempnew -> old

//May 1, 2002	  triggernode.trigger.oldp = mytempnew.mujoin(OP_UJOIN, null, onlycomp, null, old_name);
	  triggernode.trigger.oldp = mytempnew.mujoin(OP_IJOIN, null, onlycomp, null, old_name);

	} else {
	  // no compdomain. use the usual way to doTrigger.

	  Relation utmp1 = triggernode.trigger.newp.mujoin(OP_UJOIN, null, triggernode.trigger.restp, null, utmpname1);

	  if(utmp1.isLoaded()) utmp1.save();
	  utmp1.load(true);

	  String rname = triggernode.trigger.relname;

	  //addition by Yi Zheng June 10, 2002
	  IDInfo idi = env.lookup(rname, true);
	  if(idi == null)
	  	throw new InterpretError("relation " + rname + " not found"); //Yi Zheng June 14, 2002
	  if(!(idi instanceof NRInfo))
	  {			  //original code

	  //		env.removeRel(rname, false);
	  //		env.put(utmp1.assign(null, rname, null));
	  		//end of original code

	  		if(idi instanceof RelInfo) 
	  		{                         

	  			if(rname.startsWith("."))
	  			{
	  				Environment envFound = env.removeRel2(rname, true);
	 				if(envFound != null)
						env.put2(utmp1.assign(null, rname, null), envFound);
					else throw new InterpretError("no environment found for " + rname);
	  			}
	  			else
	  			{
	  				env.removeRel(rname, false);
	  				env.put(utmp1.assign(null, rname, null));
	  			}

	  		}                          
	  		else{                      
                        // added 2 lines by Zongyan for computation update , the updated relation is a parameter of the computation  -- Nov 25, 2004
	  				env.removeRel(rname, false);
	  				env.put(utmp1.assign(null, rname, null));
	  		}                          


	  }else if(((NRInfo)idi).type == IDLIST)
	  {

	  	String relName = "." + rname;
	  	Relation dotRel = env.lookupRel(relName, true);
	  	if(dotRel != null)
	  	{
	  		long oldSur = ((NRInfo)idi).NRenv.getSurrogate(rname);

	  		//System.out.println("oldSur = " + oldSur);

	  		if(oldSur != -1)
	  		{
 			dotRel = dotRel.getRelExceptSurr(oldSur, dotRel.name);//CONTINUE HERE

	  			dotRel.insert(oldSur, utmp1);
	  			if(dotRel.data == null)
	  				dotRel.load();
	  			dotRel.project(null);

	  			dotRel.save();

				//System.out.println("removing rel: " + relName);
	  			env.removeRel(relName, true);
	  			env.put(relName, new RelInfo(dotRel), true);
	  			dotRel.save();
	  		}
	  	}
	  	else throw new InterpretError("relation " + relName + " not found");

	  }//end of addition June 10, 2002
	  //utmp1.printdata();

	  // new -> tempnew
	  Relation tempnew = triggernode.trigger.newp.project(triggernode.trigger.newp.domains, tmpname1);

	  // old -> new
	  triggernode.trigger.newp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains,new_name);

	  // tempnew -> old
	  triggernode.trigger.oldp = tempnew.project(tempnew.domains, old_name);


	}

	deleteFile(ctmpname);
	deleteFile(utmpname1);
	deleteFile(utmpname2);
	deleteFile(utmpname3);
	deleteFile(tmpname1);
	deleteFile(tmpname2);
	deleteFile(tmpname3);
	deleteFile(mytempnewname);
	deleteFile(mytempoldname);
	deleteFile(mytemprestname);
    }

    if (searchchild == 1) {
      // search children's trigger and do it.

      int childnum = triggernode.jjtGetNumChildren();
      for(int i = 0; i < childnum; i++) {
	TriggerNode childnode = (TriggerNode) triggernode.jjtGetChild(i);
	doTrigger(env, childnode, 1);
      }
    }
   }


  // YuLing added the following method
  //convert an ExpressionList node to an arrray of domains for Gedit
  private Domain[] ExpressionListToGeditDomains(SimpleNode node,Domain[] doms,int projectno)
       throws InterpretError
  {
    int n=node.jjtGetNumChildren();
    if(projectno==1)
      {
	if(n!=5) throw new InterpretError("Expecting five attribute in the first projection list of gedit");
	else
	  for(int i=0;i<n;i++)
	    {
	      SimpleNode nd=(SimpleNode)node.jjtGetChild(i);
	      String attrname=nd.name;
	      Domain d=env.lookupDom(attrname,true);
	      if(d==null){throw new InterpretError("domain "+attrname+"not found");}
	      doms[i]=d;
	    }
      }
    else if (projectno==2)
      {
	int Color=5,Symbol=6,Layer=7;
	for(int i=0;i<n;i++){
	  SimpleNode nd=(SimpleNode)node.jjtGetChild(i);
	  if(nd.type!=OP_BIOPERATOR)
	    {throw new
	  InterpretError("Expecting bioperator node instead of "+nd.type);
	    }

	  if (nd.jjtGetNumChildren()!=2)
	    {throw new
	       InterpretError("Expecting two children of comparison node instead of "+nd.jjtGetNumChildren());
	    }

	SimpleNode ndl=(SimpleNode)nd.jjtGetChild(0);
	SimpleNode ndr=(SimpleNode)nd.jjtGetChild(1);
	if(ndl.name.compareTo("Color")!=0&&ndl.name.compareTo("Symbol")!=0&&ndl.name.compareTo("Layer")!=0)
	  {throw new
	     InterpretError("Expecting the following: Color , Symbol, Layer");
	}
	String s=ndr.name;
	Domain d=env.lookupDom(s,true);
	if(d==null){throw new InterpretError("domain "+s+"not found");}
	if(ndl.name.compareTo("Color")==0)
	  doms[Color]=d;
	else if(ndl.name.compareTo("Symbol")==0)
	  doms[Symbol]=d;
	else if(ndl.name.compareTo("Layer")==0)
	  doms[Layer]=d;
	}
      }
    else
      throw new
	InterpretError("Expection 1 or 2 as the projection no. in ExpressionListToGedit()");
    //System.out.println("doms[0].name before return"+doms[0].name);
    return doms;
 }
  //end of YuLing's addition

  // convert an ExpressionList node to an array of domains
  // REVISIT: generalization, may need another argument of type Relation
  private Domain[] ExpressionListToDomains(SimpleNode node)
       throws InterpretError
  {
    int n = node.jjtGetNumChildren();
    Domain[] doms = new Domain[n];
    //System.out.println("children in the Expression List");

    for (int i = 0; i < n; i++) {
      SimpleNode nd = (SimpleNode) node.jjtGetChild(i);
      //printNode(nd);
      if (nd.type != OP_IDENTIFIER) {
	throw new
	  InterpretError("only IDList implemented for ExpressionList");
      }
      String s = nd.name;
      Domain d = env.lookupDom(s, true);
      if (d == null)
	throw new InterpretError("domain : " + s + " not exist");
      else if (d.isStateDom == 1) //Yi Zheng May 6, 2002
      throw new InterpretError("domain: " + s + " is a stae domain");
      else
	doms[i] = d;
    }

    return doms;
  }


   //Andrey Rozenberg
   public Domain[] AttribsRel2DomList(Relation rel)
   throws InterpretError
   {
	   Domain[] domList = null;
		 /*
		//if(rel.domains[0].type==ATTRIBUTE) //modified by Yi Zheng June 24, 2002 to test for unary condition
//		if((rel.domains.length == 1) && (rel.domains[0].type == ATTRIBUTE))
		{
			 rel = rel.duplicateLoadAll();
			 domList = new Domain[rel.numtuples];

			 for(int i=0;i<domList.length;i++)
			 	 domList[i]= env.lookupDom(((String[]) rel.data[0])[i],true);
			 
			  
		}  */  // commented out by Fan on September 28, 2004 using the following part
           /******* start adding by Fan  *******/
            rel = rel.duplicateLoadAll();
                domList = new Domain[rel.numtuples];
                String evaledName = rel.domains[0].name+"_evaled";
                Relation srcrel = null;
                if((rel.domains.length == 1) && (rel.domains[0].type == ATTRIBUTE))
                {

                         Domain evaledDom = env.lookupDom(evaledName,true);
                         Domain dom = null;
                         if(rel.name.startsWith("_temp")){  // eval an attribute of type ATTRIBUTE Nov. 7
                                String srcrelName = getRelName();
                                srcrel = env.lookupRel(srcrelName, true);
                         } else {// eval A, where A is an unary relation.  Nov. 7
                                domList = new Domain[rel.numtuples];

                                 for(int i=0;i<domList.length;i++)
                                         domList[i]= env.lookupDom(((String[]) rel.data[0])[i],true);
                                 return domList;
                         }
                         for(int i=0;i<domList.length;i++){
                                dom = env.lookupDom(((String[]) rel.data[0])[i],true);
                                if (dom!=null && srcrel.isDomainIn(dom))
                                        domList[i] = dom;
                                else if(evaledDom == null){ //
                                        throw new InterpretError("domain "+ ((String[]) rel.data[0])[i] + " does not valid.");
                                }else {
                                        domList[i] = Eval2IFTHENELSE(dom.name, rel.domains[0].type, evaledName, env );
                                }
                         }
                }
                else if (rel.domains[0].type!=ATTRIBUTE && (rel.domains.length == 1)){// October 18, 2004
                        String domname = null;
                        for(int i=0;i<domList.length;i++) {
                                switch(rel.domains[0].type){
                                        case STRING:
                                                domname = ((String[]) rel.data[0])[i];
                                                break;
                                        case INTEGER:
                                                int val = ((int[]) rel.data[0])[i];
                                                domname = CastToStrn.value(val);
                                                break;
                                        default: break;
                                }
                                domList[i] = Eval2IFTHENELSE(domname, rel.domains[0].type, evaledName, env);
                        }
                }
                 /******* end of Fan's addition  *******/
		else
		{
	//		throw new
	//		InterpretError("Can not convert relation " + rel.name + " to domain list");

			//bug fix by Yi Zheng Sept 28, 2002
			domList = new Domain[rel.numattrs];
			for (int j = 0; j < rel.numattrs; j++)
				domList[j] = rel.domains[j];


			//end of bug fix Sept 28, 2002

		}

		return domList;
   }

 	/**
	 * wrapper function to combine the following version (v1) of ExpressionListToDomains
	 * with ExpressionListToDomains(SimpleNode node, Hashtable htable) (v2) from Andrey
	 * @author Yi Zheng May 26, 2002
	 */
  private Domain[] ExpressionListToDomains(SimpleNode node, Relation rel, Hashtable htable, Environment env)
       throws InterpretError
	{ Domain[] domArray = ExpressionListToDomains(node, htable, env); //calls Andrey's code to allow relational expression in projection list

	 //addition by Yi Zheng June 28, 2002
	 //add virtual domains if this is a red node intended for level lifting
	 domArray = handleRedDomain(domArray, node, rel, env);

	 if((domArray.length == 1) && ((domArray[0].isRedTemp == 1) || (domArray[0].isRedTempTop == 1)))
	 {
	 	//System.out.println("return domArray directly");
	 	return domArray;
	 }
	 //end of addition June 28, 2002

	  return ExpressionListToDomains(domArray, rel); //calls Yi's modified code to test for state domain
	 }

	  //this method adds hidden state
  //domains from rel to the domains being returned
  //it also checks to make sure that no state is in the original
  //projection list
  // Yi Zheng April, 2002
  // v1
  private Domain[] ExpressionListToDomains(Domain[] node, Relation rel)
       throws InterpretError
  {
    int n = node.length;
    int numDoms = 0;

    //find the number of state domains in rel
    for(int j = 0; j < rel.numattrs; j++)
    {
    	if (rel.domains[j].isStateDom == 1)
    		numDoms++;
    }
    Domain[] doms = new Domain[n + numDoms];
    //System.out.println("children in the Expression List");

    //check for state doms in the original list
    //this is just superficial check, virtual domains which
    //are renamed from state doms will be caught in virdom declaration code
    int i;
    for (i = 0; i < n; i++) {
 		Domain d = node[i];
 		if (d.isStateDom == 1)
      throw new InterpretError("domain: " + d.name + " is a hidden state domain which cannot be projected upon");
		else
		doms[i] = d;
    }

    //finally, tack on state doms in rel
    	for (int k = 0; k < rel.numattrs; k++)
    	{
    		if (rel.domains[k].isStateDom == 0)
    			continue;
    		else
    			doms[i++] = rel.domains[k];
    	}

    return doms;
  }

  //this method overloads adds hidden state
  //domains from rel to the domains being returned
  //it also checks to make sure that no state is in the original
  //projection list
  // Yi Zheng April, 2002
  //
  private Domain[] ExpressionListToDomains(SimpleNode node, Relation rel)
       throws InterpretError
  {
    int n = node.jjtGetNumChildren();
    int numDoms = 0;

    //find the number of state domains in rel

    for(int j = 0; j < rel.numattrs; j++)
    {
    	if (rel.domains[j].isStateDom == 1)
    		numDoms++;
    }
    Domain[] doms = new Domain[n + numDoms];
    //System.out.println("children in the Expression List");

    //check for state doms in the original list
    //this is just superficial check, virtual domains which
    //are renamed from state doms will be caught in virdom declaration code
    int i;
    for (i = 0; i < n; i++) {
      SimpleNode nd = (SimpleNode) node.jjtGetChild(i);
      //printNode(nd);
      if (nd.type != OP_IDENTIFIER) {
	throw new
	  InterpretError("only IDList implemented for ExpressionList");
      }
      String s = nd.name;
      Domain d = env.lookupDom(s, true);
      if (d == null)
	throw new InterpretError("domain : " + s + " not exist");
      else if (d.isStateDom == 1)
      throw new InterpretError("domain: " + s + " is a hidden state domain which cannot be projected upon");
		else
		doms[i] = d;
    }

    //finally, tack on state doms in rel
    	for (int k = 0; k < rel.numattrs; k++)
    	{
    		if (rel.domains[k].isStateDom == 0)
    			continue;
    		else
    			doms[i++] = rel.domains[k];
    	}

    return doms;
  }

     /*
   This function evaluates an expression and converts the resulting unary relation
   with the attribute of type "attribute" to a list of domains.

   Input: a parsed tree =>   ExpressioList
							/  ...	| ... \

   Output: a list of domains
									A. Rozenberg
   * v2
   */
  // public Domain[] ExpressionListToDomains(SimpleNode node, Hashtable htable) //modified by Yi Zheng May 27, 2002

   public Domain[] ExpressionListToDomains(SimpleNode node, Hashtable htable, Environment env)
   throws InterpretError
   {
//System.out.println("relation or domain????????????");
//Global global = new Global();
//global.traverse(node);
	 int n = node.jjtGetNumChildren();
	 Vector domList = new Vector();

	 for (int i = 0; i < n; i++)
	 {
		SimpleNode nd = (SimpleNode) node.jjtGetChild(i);
		//printNode(nd);
	    switch(nd.type)
		{
                /******* start adding by Fan  *******/
				case OP_LITERAL: // January 7, 2005	      			
	      			nd = LookupValidRelName(i, nd, env); 
	      			/*Domain[] Dom = null;
	      			Dom = ExpressionListToDomains(node, htable, env);
	      			for (i=0;i<Dom.length; i++)
	      				domList.addElement(Dom[i]);
                    */ 
	      			// Now dealing with the node as an IDENTIFIER. The following code is the copy of "case OP_IDENTIFIER:".  
	      			nd = CheckRecursiveNode(nd,env); 
                    nd.jjtSetParent(node);
                    node.jjtReplaceChild(i,nd);
                    String sn = nd.name;
                    Domain dd = env.lookupDom(sn, true);
                    if (dd != null){
                    Domain[] doms = new Domain[1];
                    	doms[0] = dd;
                        domList.addElement(doms);
                    } else{
                    	Relation rd = env.lookupRel(sn, true);
                        if ((rd != null) && (rd.rvc == RELATION)) { //modified by Yi Zheng May 31, 2002
                        	Domain doms[] = AttribsRel2DomList(rd);
                            domList.addElement(doms);
                        } else
                            throw new InterpretError("domain or relation: " + sn + " does not exist");
                    }
	      			break;
                case OP_SELF: // June 8, 2004 Fan Guo  for case SRProd1<-[self] in products;
                        Domain selfdom = env.lookupDom("self",true);
                        if(selfdom == null){
                                selfdom = new Domain ("self", ATTRIBUTE, 0, null);
                                //selfdom = new Domain();  // the following is added on August 5, 2004
                                //selfdom.type = ATTRIBUTE;       // together with code in actualize_old.
                                //selfdom.name = "self";
                                env.put(selfdom);
                        }
                        Domain[] selfdoms = new Domain[1];
                        selfdoms[0] = selfdom;
                        domList.addElement(selfdoms);
                        break;

                case OP_BIOPERATOR: // September, 14, 2004 case: "(arri)A" in "R<-[(attr)([A] in next)] in next;"
                        SimpleNode left = (SimpleNode)nd.jjtGetChild(0);
                SimpleNode right = (SimpleNode)nd.jjtGetChild(1);
                        if(nd.opcode == OP_CAST){// Nov 19, 2004
                                switch(right.type){
                                        case OP_PROJECT:
                                        case OP_SELECT:
                                        case OP_QSELECT:
                                        case OP_TSELECT:
                                        case OP_PICK:
                                        case OP_ATTRIBSOF:
                                        case OP_ARRAY:
                                                                        String tmpname = nextTempName();
                                                                        Relation rel = evaluateTLExpression(right, tmpname, htable, env);
                                                                        //Domain[]doms = AttribsRel2DomList(rel,left.opcode);
                                                                        Domain [] doms = null;
                                                                        if(left.opcode == OP_ATTRIBUTE)  // October 18, 2004
                                                                                doms = AttribsRel2DomList(rel);
                                                                        domList.addElement(doms);
                                                                        deleteFile(tmpname);                       
                                                                        break;                                     
                                        default:
                                                break;
                                }
                        } // else throw new InterpretError(" to be implemented.");
                        break;

                /******* end of Fan's addition *******/

			case OP_IDENTIFIER:	
                                /******* start adding by Fan  *******/
// Commented off by Zongyan 2005 
// Conflits with ~zwang26/test3/tao/t2.a
                    //          nd = CheckRecursiveNode(nd,env); // Fan August 19, 2004
                     //         nd.jjtSetParent(node);
                      //        node.jjtReplaceChild(i,nd);
                                /******* end of Fan's addition *******/
                                        String s = nd.name;
/*					Relation rd = env.lookupRel(s, true);
   					if ((rd != null) && (rd.rvc == RELATION)) //modified by Yi Zheng May 31, 2002
					{ //Convert relation of type "attribute" to domain list
						Domain doms[] = AttribsRel2DomList(rd);
						domList.addElement(doms);
					}
					else
					{
						Domain d = env.lookupDom(s, true);
						if (d == null)
							throw new InterpretError("domain or relation: " + s + " does not exist");
						else
						{
							Domain[] doms = new Domain[1];
							doms[0] = d;
							domList.addElement(doms);
	 					}
					}
					break;
*/
// added by zongyan wang April,2004
					Domain d = env.lookupDom(s, true);

					if (d != null)

					{

						Domain[] doms = new Domain[1];

						doms[0] = d;

						domList.addElement(doms);

 					}

                                        else{
						Relation rd = env.lookupRel(s, true);

   						if ((rd != null) && (rd.rvc == RELATION)) //modified by Yi Zheng May 31, 2002

						{ //Convert relation of type "attribute" to domain list

							Domain doms[] = AttribsRel2DomList(rd);

							domList.addElement(doms);

						}

						else

							throw new InterpretError("domain or relation: " + s + " does not exist");


					}

					break;
// end:added by zongyan wang April,2004

			case OP_JOINOPERATOR:	String tmpname = nextTempName();
									Relation rel = evaluateTLExpression(nd, tmpname, htable, env);
									Domain doms[] = AttribsRel2DomList(rel);
									domList.addElement(doms);
									deleteFile(tmpname);
									break;

			case OP_EQUIV:
			case OP_PAR:
			case OP_RED: //Yi Zheng May 7, 2002
			case OP_FUN: //Yi Zheng May 7, 2002
	//				throw new InterpretError("Reduction and functional mapping in projection list: to be implemented");
			break; //Yi Zheng June 24, 2002, note this is temporay, I don't know how to cope with the following 2 cases yet:
				// case 1: let a be equiv + of 1 by (equiv ujoin of nestedAttrRel);
				//		if nestedAttrRel contains only one domain of type attr, then
				//		this let statement is legal, and we should allow it in actualization
				//	However, the challenge here is that we don't know what the type of the temp
				//	domain is (in this case equiv ujoin of nestedAttrRel), and we certainly
				//	don't know if this temp domain will be IDLIST typed with one attribute of type attribute;
				//	secondly, it is not a good idea to try to actualize the temp domain here
				//	because this will not allow domain alias, for example,
				//	let b be a; in such case, actulizing on b will be problematic, as the actualizer will not know how to convert relation containing attr to attributes!
				//	this is just another case of the shortcomings of early conversion in actualizer constructor by Andrey
				//case 2: let a be equiv + of 1 by (equiv + of c by d);
				//	This is also a legitimate case, however, we don't know the type of the temp
				//	domain equiv + of c by d, it could have been an IDLIST typed domain and which
				// 	doesn't have attr as its first attribute.
				//	As a temporary compromise, I just let these cases pass with a break. But
				//	the danger is that case 1 will cause errors later on, if no code is installed for handling AttribsRel in actualization code other than the constructor
			case OP_PROJECT: //added by Yi Zheng June 24, 2002
			case OP_SELECT:
			case OP_QSELECT:
			case OP_TSELECT:
			case OP_PICK:
			case OP_ATTRIBSOF:
			case OP_ARRAY:
							tmpname = nextTempName();

							 rel = evaluateTLExpression(nd, tmpname, htable, env);

						     doms = AttribsRel2DomList(rel);
							domList.addElement(doms);

							deleteFile(tmpname);
							break;
							//end of addition June 24, 2002
			default:
				/*			 tmpname = nextTempName();

							 rel = evaluateTLExpression(nd, tmpname, htable, env);
						     doms = AttribsRel2DomList(rel);
							domList.addElement(doms);
							deleteFile(tmpname);
							*/ //commented out by Yi Zheng June 24, 2002
							break;

	/*		throw new
			InterpretError("only IDList and JoinExpression are implemented for ExpressionList");
	*/
		}
	 }
	 int domSize =0;
	 Domain[] doms=null;
	 for(int i=0;i<domList.size();i++)
	 {
		doms=(Domain[])domList.elementAt(i);
		domSize=domSize+doms.length;
	 }

	 Domain[] domsArray=new Domain[domSize];
	 int k=0;

	 for(int i=0;i<domList.size();i++)
	 {
		doms=(Domain[])domList.elementAt(i);
		for(int j=0;j<doms.length;j++)
		{
			domsArray[k]=doms[j];

			k++;
		}
	 }

	 return domsArray;
   }

   /**
    * generate a temporary domain for red node that is intended for level lifting
    * @author Yi Zheng June 28, 2002
    */
   private Domain[] handleRedDomain(Domain[] domArray, SimpleNode node, Relation rel, Environment env)
	throws InterpretError
   {
 	 int n = node.jjtGetNumChildren();
 	 if(n != 1)
 	 	return domArray;

	SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	Domain newDom = null;
	switch(nd.type)
	{
		case OP_RED:


		//starting zhan's adding to handle recursive relation Level-lifting

		SimpleNode pNode = (SimpleNode)node.jjtGetParent();
		SimpleNode ppNode = (SimpleNode)pNode.jjtGetParent();
		SimpleNode cNode = (SimpleNode)ppNode.jjtGetChild(1);

		RecursiveLifting(node,0,cNode.name); //zhan to handle recursive domain lifting

		//end zhan's addding July 2003

                /******* start adding by Fan  *******/
                CheckRecursiveNodeInRedOp(nd,env);  // nd: the vertical Node  Sep. 1 2004
                /******* end of Fan's addition *******/

			SimpleNode ndCopy = (SimpleNode)nd.jjtDeepDuplicate();

			SimpleNode letNode = SimpleNode.jjtCreateNode("Declaration", OP_DECLARATION,
				OP_LET, null);
			letNode.jjtAddChild(ndCopy);
			ndCopy.jjtSetParent(letNode);
			String domname = "reddom" + Surrogate.next();
			newDom = newVirDomain(domname, letNode, env);
	   		Relation rel_saved = env.lookupRel("."+domname,true);
	   		env.put(newDom);
	   		newDom.isRedTempTop = 1;
			if (rel_saved != null)
	     		env.put(rel_saved);

			break;
		default:
			return domArray;
	}
	 Domain[] doms = new Domain[1];
	 doms[0] = newDom;

	 return doms;



   }

   //zhan' addition
   //to handle [red ujoin of [red ujoin of N] in N] in N
   private void RecursiveLifting(SimpleNode node,int L,String recurName)
   {
      //System.out.println("RecursiveLifting");
	  int j = L;
      String recursiveDomain = recurName;

      if(node.jjtGetNumChildren()!=0)
      {
		SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
		if(nd.type == OP_RED && nd.opcode== OP_UJOIN)
        {
	        //System.out.println("type = OP_RED,opcode = OP_UJOIN");
		    SimpleNode projectNode = (SimpleNode)nd.jjtGetChild(0);
            if(projectNode.type == OP_PROJECT)
            {
				SimpleNode idNode = (SimpleNode) projectNode.jjtGetChild(1);

				if(recursiveDomain != null)
				{
				  if(recursiveDomain.equals(idNode.name))
				   {
						 idNode.name = recursiveDomain + "_"+ j;
						 j = j +1;
				   }
				  else
				  {
						 recursiveDomain = idNode.name;
				   }
				 }
			    else
				 recursiveDomain = idNode.name;

			  SimpleNode projectorNode = (SimpleNode)projectNode.jjtGetChild(0);

			  RecursiveLifting(projectorNode,j,recursiveDomain);

		     }//end type = OP_PROJECT

		    //else //node is not project, do nothing


          }//end if nd.type = OP_RED, op.opcode = OP_UJOIN
        else
           RecursiveLifting(nd,j,recursiveDomain);
       }//end if
   }

  //end of zhan'additon



  private void getDomsInCompList(SimpleNode node, Hashtable htable,
				 Environment env)
       throws InterpretError
  {
    if(node.type == OP_IDENTIFIER){
      Domain d = env.lookupDom(node.name, true);
      if(d != null){   // Should never happen.
	htable.put(node.name, d);
	return;
      }else
	throw new InterpretError("domain '"+node.name+"' not found.");
    }else if(node.jjtGetNumChildren() == 2){
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);

      switch(node.opcode){
      case OP_AND:
      case OP_OR:

	getDomsInCompList(left, htable, env);
	getDomsInCompList(right, htable, env);
	break;
      case OP_EQ:
      case OP_NEQ:
      case OP_GT:
      case OP_LT:
      case OP_GE:
      case OP_LE:
	Domain d;
	if(left.type == OP_IDENTIFIER){
	  d = env.lookupDom(left.name, true);
	  if(d != null) //Yi Zheng May 2, 2002
	  	htable.put(left.name, d);
	  else throw new InterpretError("domain " + left.name + " not found");
	}
	if(right.type == OP_IDENTIFIER){
	  d = env.lookupDom(right.name, true);
	  if(d != null) //Yi Zheng May 2, 2002
	  	htable.put(right.name, d);
	  else throw new InterpretError("domain " + right.name + " not found");
	}
	break;
      default:
	// START Since jRelixNS
//	if (env.isSigmaJoinImplemented != true) {
	  throw new InterpretError("getDomsInCompList: to be implemented.");
//	} else {
	  // This is added for the boolean-IDLIST compatibility.
	  // We will try to see if it is an IDLIST if all else
	  // fails for a boolean.
	  // when processing domains in a 'comparions' list
	  // for a OP_SELECT or OP_TSELECT.

//	  if (getIDList(node,env) == null)
//	    throw new InterpretError("getDomsInCompList: BI-OP " +
//				     node.opcode +" not supported; to be implemented.");
//	}
	// END Since jRelixNS
      }
      return;
    }else{
      // START Since jRelixNS
 //     if (env.isSigmaJoinImplemented != true) {
	throw new InterpretError("getDomsInCompList: to be implemented.");
  //    } else {
	// This is added for the boolean-IDLIST compatibility.
	// We will try to see if it is an IDLIST if all else
	// fails for a boolean.
	// when processing domains in a 'comparions' list
	// for a OP_SELECT or OP_TSELECT.
//	if (getIDList(node,env) == null)
//	  throw new InterpretError("getDomsInCompList: Node type "
//+node.type+
//"non-IDLIST/BIOP not supported; to be implemented.");
//      }
      // END Since jRelixNS
    }
  }

  /******* start adding by Fan  *******/
  // wrapper for original getIDList July 4, 2004 Fan Guo
  public Relation getIDList(SimpleNode node, Environment env)
  throws InterpretError{
  	switch(node.type){
  		case OP_RELATION: 
  	    	SimpleNode rel_exprlist = (SimpleNode)node.jjtGetChild(0);  	   		
  	    	if (rel_exprlist.jjtGetNumChildren() == 0)  throw new InterpretError("no domain is defined.");
  	    	
  	    	// get domains defined to a domain array
  	    	String rel_dname = null;
  	    	Domain rel_d = null;
  	    	Vector rel_domvec = new Vector();
  	    	rel_domvec.addElement(env.lookupDom(".id", true));	// add domain .id to domain vector
  	    	for(int i = 0; i < rel_exprlist.jjtGetNumChildren(); i++){
  	    		rel_dname = ((SimpleNode)rel_exprlist.jjtGetChild(i)).name;   //get domain name defined 
  	    		rel_d = env.lookupDom(rel_dname, true);	//check the validation of defined domain name 
  	    		if(rel_d == null)	  throw new InterpretError("domain "+rel_dname+" not found.");
  	    		rel_domvec.addElement(rel_d);					// add the valid domain to domain vector
  	    	}  	
  	    	Domain[] rel_doms = new Domain[rel_domvec.size()];
  	    	rel_domvec.copyInto(rel_doms);
  	    	
  	    	// construct a new relation and return it to the caller
  	    	return new Relation("", rel_doms, 0, rel_doms.length, 0, env);	// constructer a relation without name
  	    case OP_TRANSPOSENEW:  // case: let cn be transpose(attr,tp,val);
  	    		SimpleNode trans_exprlist = (SimpleNode)node.jjtGetChild(0);
  	    		if (trans_exprlist.jjtGetNumChildren() == 0)  throw new InterpretError("no domain is defined.");
  	    		
  	    		//get domains defined to a domain vector
  	    		String trans_dname = null;
  	    		Domain trans_d = null;  	    		
  	    		int total=0, ta=0, tt=0, tu=0;
  	    		Vector trans_domvec = new Vector();
  	    		trans_domvec.addElement(env.lookupDom(".id", true));	// add domain .id to domain vector  	    		
  	    		for(int i = 0; i < trans_exprlist.jjtGetNumChildren(); i++){
  	    			trans_dname = ((SimpleNode)trans_exprlist.jjtGetChild(i)).name;   //get domain name defined 
  	    			trans_d = env.lookupDom(trans_dname, true);	//check the validation of defined domain name 
  	    			if(trans_d == null)	  throw new InterpretError("domain "+trans_dname+" is not found.");  	    			
  	    			switch(trans_d.type){
  	    	  			case ATTRIBUTE: ta++; total++; break;
  	    	  			case TYPE:      tt++; total++; break;
  	    	  			case UNIVERSAL: tu++; total++; break;
  	    	  			default:throw new InterpretError("domain "+ trans_dname+" is not valid.");
  	    	  		}
  	    			trans_domvec.addElement(trans_d);
  	    		}  
  	    		
  	    		// check the validation of attributes defined
  	    		if (total >3) throw new InterpretError("number of domains is not valid.");
  	    		if (ta >1) throw new InterpretError("more than one domain is of type attribute");
  	    		if (tt >1) throw new InterpretError("more than one domain is of type type");
  	    		if (tu >1) throw new InterpretError("more than one domain is of type universal");
  	    		// construct a new relation and return it to the caller 
  	    		Domain[] trans_dom = new Domain[trans_domvec.size()];
  	    		trans_domvec.copyInto(trans_dom);
  	    		return new Relation("", trans_dom, 0, trans_dom.length, 0, env);	// constructer a relation without name
  	    case OP_IFTHENELSE:  // September 15, 2004 
  	    		SimpleNode trueValNode = (SimpleNode) node.jjtGetChild(1);
  	    		SimpleNode falseValNode = (SimpleNode) node.jjtGetChild(2);
  	    		Relation r1, r2;
  	    		r1 = getIDList(trueValNode, env);
  	   		 	r2 = getIDList(falseValNode, env);
  	   		 	if(r1 == null && r2 != null) return r2;
  	   		 	if(r1 != null && r2 == null) return r1;
  	   		 	if(r1!=null && r2!=null){
  	   		 		if(r1.isDomainsEquals(r2.domains)) 
  	   		 			return r1;
  	   		 		else throw new InterpretError("attributes in two relations for true value and false value are not equal");
  	   		 	}
  	   		 	if(r1 == null && r2 == null) 
  	   		 		throw new InterpretError("no attributes in result relations.");
  	   	case OP_LITERAL: 
  	   		return null;
  	    default:
  	    	return getIDList_old(node, env);
  	    }
  }
  /******* end of Fan's addition  *******/


  /**
   * Return a virtual relation which contains the domain array as the
   * result of join operations etc. occurred in 'node'.
   * Added for jRelixNS:
   *    The relation could be real if a top-level relation is
   *    used in the domain algebra.
   *
   * JRelixNS modified/enhanced this method:
   *   1) Handle Sigma-Joins in addition to mu-joins.
   *   2) Handle empty exprlist as it might not exist for an empty projector
   *      for nullary tuple results, more frequently seen for sigma-joins.
   */

   private Relation getIDList_old(SimpleNode node, Environment env) // renamed on July 6, 2004  by Fan Guo
                                                                   // using wrapper

  // private Relation getIDList(SimpleNode node, Environment env)
       throws InterpretError
  {
 //   System.out.println("getIDlistnode type "+node.type);
    switch(node.type){
    case OP_IDENTIFIER:
      String rname = node.name;
      Relation r = env.lookupRel("."+rname, true);
      if(r == null){

	// REVISIT: should be looking into NREnvironment for top-level relation.
	// e.g. let x be S ijoin R; where 'R' is top-level rel.
	if((r = env.lookupRel(rname, true)) == null)
	  // should never happen..
	  throw new InterpretError("join relation not found.");
      }
      return r;
    case OP_JOINOPERATOR:
    if (Relation.isSigmaJoinImplemented != true) { //Andy May 12, 2002, added this if case
	   Relation r1, r2;
	   Domain[] doms;

	   if(node.jjtGetNumChildren() == 2){
		 // No join attributes presented.
		 r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
		 r2 = getIDList((SimpleNode)node.jjtGetChild(1), env);
		 doms = r1.mujoinType(node.opcode, null, r2, null);
	   }else{
		 // Join attributes are given.
		 r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
		 r2 = getIDList((SimpleNode)node.jjtGetChild(3), env);
		 Domain[] doms1, doms2;
		 Vector domsv1 = new Vector();
		 Vector domsv2 = new Vector();

		 // Add .id to vectors.
		 domsv1.addElement(env.lookupDom(".id", true));
		 domsv2.addElement(env.lookupDom(".id", true));

		 SimpleNode expr1, expr2;
		 expr1 = (SimpleNode)node.jjtGetChild(1);
		 expr2 = (SimpleNode)node.jjtGetChild(2);

		 if(expr1.jjtGetNumChildren() != expr2.jjtGetNumChildren())
			// Should never happen...
			throw new InterpretError("different number of join attributes.");

		 for(int i = 0; i < expr1.jjtGetNumChildren(); i++){
			String name = ((SimpleNode)expr1.jjtGetChild(i)).name;
			Domain d = env.lookupDom(name, true);
			if(d != null) domsv1.addElement(d);
			else throw new InterpretError("domain '"+d+"' not found.");

			name = ((SimpleNode)expr2.jjtGetChild(i)).name;
			d = env.lookupDom(name, true);
			if(d != null) domsv2.addElement(d);
			else throw new InterpretError("domain '"+d+"' not found.");
		 }
		 doms1 = new Domain[domsv1.size()];
		 doms2 = new Domain[domsv2.size()];
		 domsv1.copyInto(doms1);
		 domsv2.copyInto(doms2);

		 doms = r1.mujoinType(node.opcode, doms1, r2, doms2);
	   }
	   return new Relation("", doms, 0, doms.length, 0, env);
	 }else{

      Relation r1, r2;
      boolean isSigma;
      Domain[] doms;

      switch(node.opcode)
	{
	case OP_ICOMP :
	case OP_EQJOIN :
	case OP_GTJOIN :
	case OP_GEJOIN :
	case OP_LTJOIN :
	case OP_LEJOIN :
	case OP_IEJOIN :
	case OP_NOTICOMP :
	case OP_NOTEQJOIN :
	case OP_NOTGTJOIN :
	case OP_NOTGEJOIN :
	case OP_NOTLTJOIN :
	case OP_NOTLEJOIN :
	case OP_NOTIEJOIN :
	  isSigma = true;
	  break;

	case OP_IJOIN :
	case OP_UJOIN :
	case OP_SJOIN :
	case OP_LJOIN :
	case OP_DLJOIN :

	case OP_RJOIN :
	case OP_DRJOIN :
	default:
	  isSigma = false;
	  break;
	}
      if(node.jjtGetNumChildren() == 2){
	// No join attributes presented.
	r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
	r2 = getIDList((SimpleNode)node.jjtGetChild(1), env);

	//added by Yi Zheng July 4, 2002 to handle let a be B nop C
	if(node.opcode == OP_NOP)
	{
		if((!r1.isDomainsIn(r2.domains)) || (!r2.isDomainsIn(r1.domains)))
			throw new InterpretError("type mismatch between domains");
		return new Relation("", r1.domains, 0, r1.domains.length, 0, env);

	}


	//end of addition July 4


	if (isSigma)
	  doms = r1.sigmaJoinType(node.opcode, null, r2, null);
	else
	  doms = r1.mujoinType(node.opcode, null, r2, null);

	if (!r1.isDomainIn(env.lookupDom(".id", true),doms))
	  {
	    Domain[] tmpdoms = doms;
	    doms = new Domain[tmpdoms.length+1];
	    doms[0]=env.lookupDom(".id", true);
	    for(int i = 0; i < tmpdoms.length; i++){
	      doms[i+1]=tmpdoms[i];
	    }
	  }
//	else if(doms.length == 1){
//	  doms = new Domain[2];
//	  doms[0]=env.lookupDom(".id", true);
//	  doms[1]=env.lookupDom(".bool", true);
//	}
      }else{
	// Join attributes are given.
	r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
	r2 = getIDList((SimpleNode)node.jjtGetChild(3), env);
	Domain[] doms1, doms2;
	Vector domsv1 = new Vector();
	Vector domsv2 = new Vector();

	// Add .id to vectors.
	domsv1.addElement(env.lookupDom(".id", true));
	domsv2.addElement(env.lookupDom(".id", true));

	SimpleNode expr1, expr2;
	expr1 = (SimpleNode)node.jjtGetChild(1);
	expr2 = (SimpleNode)node.jjtGetChild(2);

	if(expr1.jjtGetNumChildren() != expr2.jjtGetNumChildren())
	  // Should never happen...
	  throw new InterpretError("different number of join attributes.");

	for(int i = 0; i < expr1.jjtGetNumChildren(); i++){
	  String name = ((SimpleNode)expr1.jjtGetChild(i)).name;
	  Domain d = env.lookupDom(name, true);
	  if(d != null) domsv1.addElement(d);
	  else throw new InterpretError("domain '"+d+"' not found.");

	  name = ((SimpleNode)expr2.jjtGetChild(i)).name;
	  d = env.lookupDom(name, true);
	  if(d != null) domsv2.addElement(d);
	  else throw new InterpretError("domain '"+d+"' not found.");
	}
	doms1 = new Domain[domsv1.size()];
	doms2 = new Domain[domsv2.size()];
	domsv1.copyInto(doms1);
	domsv2.copyInto(doms2);

	if (isSigma) {
	  Domain[] sigdoms = r1.sigmaJoinType(node.opcode, doms1, r2, doms2);
	  doms = new Domain[sigdoms.length+1];
	  doms[0]=env.lookupDom(".id", true);
	  for(int i = 0; i < sigdoms.length; i++){
	    doms[i+1]=sigdoms[i];
	  }
	}
	else
	  doms = r1.mujoinType(node.opcode, doms1, r2, doms2);
      }
      return new Relation("", doms, 0, doms.length, 0, env);
    }

    case OP_PROJECT:
    case OP_TSELECT:
      // REVISIT: TSELECT also has a projection list at the first child.
      // Note, we are not checking for validality. Hopefully this will
      // be done at the actualization time.

      // Get the projection domain list.
      // START Since jRelixNS
                 // Support for empty projectors.
      SimpleNode exprlist;
 //     if (env.isSigmaJoinImplemented != true) { //Andy May 12, 2002
   if (Relation.isSigmaJoinImplemented != true) {
	exprlist = (SimpleNode)node.jjtGetChild(0);
      }
      else {
	// exprlist might not exist for an empty projector
	   // for nullary tuple results.
	exprlist = (SimpleNode)node.jjtGetChild(0);
	if(exprlist.jjtGetNumChildren() > 0) // expect 1 or 0 only.
	  exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
      }
      // END Since jRelixNS

      //added by Yi Zheng July 2, 2002
      //for implementation of level lifting
      if((exprlist.jjtGetNumChildren() == 1) &&
      	 ((((SimpleNode) exprlist.jjtGetChild(0)).type) == OP_RED))
      {
      	SimpleNode firstChild = null;
      	//if added by Yi Zheng July 9, 2002
      	if((((SimpleNode) exprlist.jjtGetChild(0)).opcode) != OP_REDFUNCALL)
      		firstChild = (SimpleNode) exprlist.jjtGetChild(0);
      	else
       		firstChild = (SimpleNode) exprlist.jjtGetChild(1);
     	//Aug 13
	//Utility.dump(firstChild, "firstChild ");

      	return getIDList(firstChild, env);

      }

      //end of Yi's addition July 2, 2002
      Vector domvec = new Vector();

      domvec.addElement(env.lookupDom(".id", true));
      for(int i = 0; i < exprlist.jjtGetNumChildren(); i++){
	String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
	Domain d = env.lookupDom(dname, true);
        /*if(d == null)
         throw new InterpretError("domain "+dname+" not found.");
         */
        /******* start modifying by Fan  *******/
        if(d == null) {  // September 4, 2004 for case  such as: "let s be [self] in Q"
          SimpleNode selfNode = ((SimpleNode)exprlist.jjtGetChild(i));
          if(selfNode.type == OP_SELF) {
                Domain selfdom = env.lookupDom("self",true);
                if(selfdom == null){
                        selfdom = new Domain ("self",ATTRIBUTE, 0, null);
                        //selfdom= new Domain();
                        //selfdom.type = ATTRIBUTE;
                        //selfdom.name = "self";
                        env.put(selfdom);
                }
                d = selfdom;
          } else
          throw new InterpretError("domain "+dname+" not found.");
        }
        /******* end of Fan's modification *******/

	domvec.addElement(d);
      }
                 // BEGIN Since jRelixNS BUGFIX 2
	   if (exprlist.jjtGetNumChildren() == 0)
	     domvec.addElement(env.lookupDom(".bool", true));
           // END Since jRelixNS BUGFIX 2

      Domain[] doms = new Domain[domvec.size()];
      domvec.copyInto(doms);
      return new Relation("", doms, 0, doms.length, 0, env);

        case OP_TEXT2ATTR: // Jiantao 2004
        SimpleNode parmlistt = (SimpleNode)node.jjtGetChild(0);
        Vector domv = new Vector(); 
        Domain idDom;
        if((idDom = env.lookupDom(".id", true)) == null) // I dont know why I have to cater the case that .id is undifined here
        {   
        idDom = new Domain(".id", IDLIST);
        env.put(idDom);
         }
        domv.add(idDom);

                for(int i=0;i<parmlistt.jjtGetNumChildren();i++)
                {       String dname = ((SimpleNode)parmlistt.jjtGetChild(i)).name;
                        if(!dname.startsWith("\""))
                        {       Domain d = env.lookupDom(dname, true);
                                if(d == null)
                                        throw new InterpretError("domain "+dname+" not found.");
                                domv.add(d);
                        }
                }

        Domain[] domArrayt = new Domain[domv.size()];
        domv.copyInto(domArrayt);
        return new Relation(".text2attr", domArrayt, 0, domArrayt.length, 0, env);

        case OP_MU2NEST: // Jiantao 2004
                SimpleNode parmlistm = (SimpleNode)node.jjtGetChild(0);
                domv = new Vector();
                domv.add(env.lookupDom(".id", true));

                int parmseq = 0; // the sequence of the parameters in the parameter list
                String dname;

                if(parmlistm.isBitSet(0))
                        parmseq++; // parmseq = 0 is the patern, ignore

                for(int i=1;i<4;i++)
                {       if(parmlistm.isBitSet(i))
                        {       dname = ((SimpleNode)parmlistm.jjtGetChild(parmseq++)).name;
                                Domain d = env.lookupDom(dname, true);
                                if(d == null)
                                        throw new InterpretError("domain "+dname+" not found.");
                                domv.add(d);
                        }
                }

        Domain[] domArraym = new Domain[domv.size()];
        domv.copyInto(domArraym);
        return new Relation(".mu2nest", domArraym, 0, domArraym.length, 0, env);


    case OP_SELECT:
      SimpleNode srcrelnode = (SimpleNode)node.jjtGetChild(1);
      return getIDList(srcrelnode, env);

    case OP_QSELECT: //Added by hongyu Sept 30,01
      SimpleNode srcnode = null;
      Relation srcrel = null;
      if(node.jjtGetNumChildren() == 2)
	srcnode = (SimpleNode)node.jjtGetChild(1);
      else srcnode = (SimpleNode)node.jjtGetChild(2);
      switch(srcnode.type){
      case OP_IDENTIFIER:{
	String srcrelname = srcnode.name;
//	System.out.println("srcrelname "+srcrelname);
	srcrel = env.lookupRel("."+srcrelname, true);
	if(srcrel == null){
	  srcrel = env.lookupRel(srcrelname, true);
	  if(srcrel == null)// Should never happen. Hongyu
	    throw new InterpretError("source relation not found.");

	}
	return srcrel;
      }

      case OP_SELECT:
      case OP_PROJECT:
      case OP_TSELECT:
      case OP_QSELECT:
      case OP_JOINOPERATOR:{
	int t = traverseType(srcnode, env);
	if(t != IDLIST && t != COMP)
	  throw new InterpretError("mismatched type.");
	srcrel = getIDList(srcnode, env);
	return srcrel;
      }
      /******* start adding by Fan  *******/
        // the following part of case OP_RED, OP_EQUIV, OP_FUN, and OP_FUN was copied and added by Fan on September 12, 2004.
            // To support vertical operation as source relation.
            // the motivative case is "let multicite be author quant(#>1)cite in red ujoin of (authors ujoin cite);"
      case OP_RED:
      case OP_EQUIV:
        //Check if OP_FUN and OP_PAR should be included. From sungsoo. Hongyu Feb 2002
        //         Vertical     for "let x be red ujoin of (S ijoin T);"
        //             \
        //             Join
        //             /  \
        //            S    T
        //if added by Yi Zheng July 9, 2002
        if(srcnode.opcode != OP_REDFUNCALL)
        {
//  Aug 13
    //  Utility.dump((SimpleNode)node.jjtGetChild(0), "node.child(0) ");
                return getIDList((SimpleNode)srcnode.jjtGetChild(0), env);
        }
        else
                return getIDList((SimpleNode)srcnode.jjtGetChild(1), env);

      case OP_FUN: //From sungsoo. hongyu added Feb 2002
      case OP_PAR:
      //if added by Yi Zheng July 9, 2002
       if(srcnode.opcode != OP_REDFUNCALL)
        return getIDList((SimpleNode)srcnode.jjtGetChild(0),env);
       else
        return getIDList((SimpleNode)srcnode.jjtGetChild(1),env);

        //System.out.println("getIDList should be taken care of");
        // End of sungsoo's code
       /******* end of Fan's addition *******/

      default:
	throw new InterpretError("Relation: to be implemented.");
      }

      // end of Hongyu Added.
    case OP_RED:
    case OP_EQUIV:
      //Check if OP_FUN and OP_PAR should be included. From sungsoo. Hongyu Feb 2002
      //         Vertical     for "let x be red ujoin of (S ijoin T);"
      //             \
      //             Join
      //             /  \
      //            S    T
      //if added by Yi Zheng July 9, 2002
      if(node.opcode != OP_REDFUNCALL)
	{
//Aug 13
  //  Utility.dump((SimpleNode)node.jjtGetChild(0), "node.child(0) ");
      	return getIDList((SimpleNode)node.jjtGetChild(0), env);
	}
      else
      	return getIDList((SimpleNode)node.jjtGetChild(1), env);

    case OP_FUN: //From sungsoo. hongyu added Feb 2002
    case OP_PAR:
    //if added by Yi Zheng July 9, 2002
     if(node.opcode != OP_REDFUNCALL)
      return getIDList((SimpleNode)node.jjtGetChild(0),env);
     else
      return getIDList((SimpleNode)node.jjtGetChild(1),env);

      //System.out.println("getIDList should be taken care of");
      // End of sungsoo's code
    case OP_ARRAY: {//Yi Zheng June 19, 2002
    	//1, get the identifier
    	SimpleNode id = (SimpleNode) node.jjtGetChild(0);
    	SimpleNode arrayIndex = (SimpleNode) node.jjtGetChild(1);
    	//2. look to see if id is a comp/idlist type domain
    	Domain idd = env.lookupDom(id.name, true);
    	if(idd != null)
    	{
    		if ((idd.type != IDLIST) && (idd.type != COMP))
    			throw new InterpretError(id.name + " is not a valid array identifier");
    		else
    		{
    			Relation rel = env.lookupRel("." + id.name, true);
    			if(rel == null)
    				throw new InterpretError("relation ." + id.name + " not found");
    			Domain dotid = env.lookupDom(".id", true);
    			if(!rel.isDomainIn(dotid))
    				return getProjListRel(rel, arrayIndex, env);
    			else //get rid of .id
    			{
    				Domain[] newD = new Domain[rel.domains.length - 1];
    				int j = 0;
    				for(int i = 0; i < rel.domains.length; i++)
    				{
    					if(rel.domains[i].name.compareTo(".id") != 0)
    						newD[j++] = rel.domains[i];
    				}
    				String tname = nextTempName();
    				Relation result = rel.project(newD, tname);
    				result.name = tname;
    				return getProjListRel(result, arrayIndex, env);
    			}
    		}
    	}
    	else //3. check if id is a top-level RELATION/COMPUTATION
    	{
    		Relation topR = env.lookupRel(id.name, true);
    		if(topR == null)
    			throw new InterpretError(id.name + " is not a valid array identifier");
    		return getProjListRel(topR, arrayIndex, env);
    	}

    }
    default:
      throw new InterpretError ("getIDList: to be implemented.");

    }
  }

  /**
   * method that returns a relation containing the domains to appear in the projection list
   * for array syntax check
   * @author Yi Zheng June 19, 2002
   */
   private Relation getProjListRel(Relation rel, SimpleNode arrayList, Environment env)
 	throws InterpretError
   {

   	Vector domvec = new Vector();
   	for(int i = 0; i < rel.numattrs; i++)
   	{
   		if(!arrayList.isBitSet(i))
   			domvec.addElement(rel.domains[i]);
   	}
   	if(domvec.size() == rel.numattrs)
   		return rel;

   	Domain[] doms = new Domain[domvec.size()];
   	domvec.copyInto(doms);
   	String tname = nextTempName();
   	Relation ret = rel.project(doms, tname);
   	ret.name = tname;
   	return ret;
   }

  private void help(String name)
  //added by Yi Zheng
  throws InterpretError
  {
    //added by Yi Zheng March 18, 2002
//    throw new InterpretError ("help: to be implementd");
	//implemented by Shunhong KANG May 23, 2003
	if (name==null) {
		Global.println("Commands:");
		Global.println("    dd <domlist>;");
		Global.println("    dr <rellist>;");
		Global.println("    help [\"<command> ! <declaration> ! domalg ! Relalg ! qtsel ! compu\"];");
		Global.println("    input \"<unix path>\";");
		Global.println("    quit;");
		Global.println("    pr <rel>;");
		Global.println("    sr [<rel>];");
		Global.println("    sd [<dom>];");
		Global.println("    srd ;");
		Global.println("    time;");
		Global.println("    trace;");
//		Global.println("");
		Global.println("Comment:");
		Global.println("    <<...>>, <<...EOL ");
//		Global.println("");
		Global.println("Declarations :");
		Global.println("    domain <ident> {bool ! intg ! real ! strg};   (help domain)");
		Global.println("    relation <ident> (<dom-list>) [<- {(<values>),(<values>),..} !");
		Global.println("                                   <- \"<unix path>\"];");
		Global.println("Domain algebra :");
		Global.println("   let <ident> be <horiz expr>;");
		Global.println("   let <ident> be red <ass-com-op> of <dom expr>;");
		Global.println("   let <ident> be equiv <ass-com-op> of <dom expr> by <dom-list>;");
		Global.println("   let <ident> be fun <fun-par-op> of <dom expr> order <dom-list>;");
		Global.println("   let <ident> be par <fun-par-op> of <dom expr> order <dom-list> by <dom-list>;");
//		Global.println("");
		Global.println("Relational algebra :");
		Global.println("    <ident>             <-             <rel expr>;");
		Global.println("    <ident>             <+             <rel expr>;");
		Global.println("    <ident> [<dom-list> <- <dom-list>] <rel expr>;");
		Global.println("    <ident> [<dom-list> <+ <dom-list>] <rel expr>;");
		Global.println("    <ident> [initial <rel expr>] is <rel expr> [target {<rel expr> ! empty}];");
		Global.println("    update <ident> <upd-op>;");
//		Global.println("");
		Global.println("QT-selectors:");
		Global.println("     <project-list> quant <quantifier list> where <condition> in <rel expr>;");
//		Global.println("");
		Global.println("Computation :");
		Global.println("     comp < identifier> ([param,...]) is ...");
	}
	else{
		String s = name.substring(1, name.length() - 1);
	  if (s.equals("dd")){
		Global.println("\"dd\" is a command to delete domains.");
		Global.println("Syntax: dd <domlist>;");
		Global.println("Example: dd i,s ;");
	  }
	  else if (s.equals("dr")){
		Global.println("\"dr\" is a command to delete relations.");
		Global.println("Syntax: dr <rellist>;");
		Global.println("Example: dr R,student;");
      }
	  else if (s.equals("input")){
		Global.println("\"input\" is a command to run a batch command file.");
		Global.println("Syntax: input \"<unix path>\";");
		Global.println("Example: input \" course/612/myprog.a\";");
	  }
	  else if (s.equals("quit")){
		Global.println("\"quit\" is a command to exit jRelix system.");
		Global.println("Syntax: quit;");
		Global.println("Example: quit;");
	  }
	  else if (s.equals("pr")){
		Global.println("\"pr\" is a command to print all records in a relation.");
		Global.println("Syntax: pr <rel>;");
		Global.println("Example: pr R;");
	  }
	  else if (s.equals("sr")){
		Global.println("\"sr\" is a command to show the name, degree and other information of a relation.");
		Global.println("Syntax: sr [<rel>];");
		Global.println("Example: sr R;");
	  }
	  else if (s.equals("sd")){
		Global.println("\"sd\" is a command to show the name, type and other in formation of a domain.");
		Global.println("Syntax : sd [<dom>];");
		Global.println("Example: sd i;");
	  }
	  else if (s.equals("srd")){
		Global.println("\"srd\" is a command to show all relations and all domains in the database.");
		Global.println("Syntax: srd;");
		Global.println("Example: srd;");
	  }
	  else if (s.equals("time")){
		Global.println("\"time\" is a command to start (stop) to show interpreting running time of each command. By default, \"time\" is off. ");
		Global.println("Syntax: time;");
		Global.println("Example: time;");
	  }
	  else if (s.equals("trace")){
		Global.println("\"trace\" is a command to start (stop) to store running tracing file ( which is named \".trace\"). By default, \"trace\" is off. If turned on, it will erase old .trace.");
		Global.println("Syntax: trace;");
		Global.println("Example: trace;");
	  }
	  else if (s.equals("domain")){
		Global.println("Declare domains:");
		Global.println("Syntax: domain <domlist> {bool ! intg ! real ! strg};");
		Global.println("    <bool>::= true ! false ! dc ! dk        <<ascii rep. of true: 1; false: 0>>");
		Global.println("    <intg>::= <digit>+ ! dc ! dk            <<ascii rep. of dc: ^E;  dk: ^B>>");
		Global.println("    <strg>::= \"<alphameric>+\" ! dc ! dk     <<ascii rep. of end item: ^F>>");
		Global.println("    <real>::= <digit>* . <digit>* E [+,-] <digit>+ ! dc ! dk ");
		Global.println("Example: domain i,s intg;");
		Global.println("");
		Global.println("Syntax: domain <ident> (<dom-list>);");
		Global.println("Example: domain R(s,t);");
		Global.println("");
		Global.println("Syntax: domain <nest domain>;");
		Global.println("Example: domain N(ii,N);");
	  }
	  else if (s.equals("relation")){
		Global.println("Declare relations:");
		Global.println("Syntax:relation <ident> (<dom-list>) [<- {(<values>),(<values>),..} !<- \"<unix path>\"];");
		Global.println("Example: relation R(b,i,s,r,g)<-{(true,1,-1,3.14,\"hi\")};");
	  }
	  else if (s.equals("domalg")){
		Global.println("\"let\" is a key word in domain algebra.");
		Global.println("Syntax: let <ident> be <horiz expr>;");
		Global.println("        let <ident> be red <ass-com-op> of <dom expr>;");
		Global.println("        let <ident> be equiv <ass-com-op> of <dom expr> by <dom-list>;");
		Global.println("        let <ident> be fun <fun-par-op> of <dom expr> order <dom-list>;");
		Global.println("        let <ident> be par <fun-par-op> of <dom expr> order <dom-list> by <dom-list>;");
		Global.println("Example: let a be b;");
		Global.println("");
		Global.println("Note: <ass-com-op>::= + ! * ! & ! | ! max ! min");
		Global.println("      <fun-par-op>::= <ass-com-op> ! - ! / ! mod ! ** ! cat ! pred ! succ !");
		Global.println("      <horiz op>::= <fun-par-op but not pred, succ> ! < ! <= ! = ! >= ! > ! ~= !");
		Global.println("                  if..then..else.. ! abs ! cos ! sin ! tan ! log10 ! log ! ln !");
		Global.println("                  atan ! acos ! asin ! tanh ! cosh ! sinh ! isknown !");
		Global.println("                  round ! ceil ! floor ! ord ! chr");
		Global.println("                  (function arguments in parentheses, e.g., abs(A))");
	  }
	  else if (s.equals("relalg")){
		Global.println("\"<-\" and \"<+\" are key words to assignment value to a relation. ");
		Global.println("Syntax: <ident>             <-             <rel expr>;");
		Global.println("        <ident> [<dom-list> <- <dom-list>] <rel expr>;");
		Global.println("Example: \"<-\" means copying data.");
		Global.println("        Same domain-list:  R1 <- R2;  ");
		Global.println("        Different domain-list: R1 [i1,s1 <- i2,s2 ] R2;");
		Global.println("");
		Global.println("Syntax: <ident>             <+             <rel expr>;");
		Global.println("        <ident> [<dom-list> <+ <dom-list>] <rel expr>;");
		Global.println("Example: \"<+\" means appending data.");
		Global.println("        Same domain-list:  R1 <+ R2;  ");
		Global.println("        Different domain-list: R1 [i1,s1 <+ i2,s2 ] R2;");
		Global.println("");
		Global.println("Syntax: update <ident> <upd-op>;");
		Global.println("        <upd-op>::= add <rel expr> ! delete <rel expr> !");
		Global.println("                    change <ident> <- <dom-expr> ,.. !");
		Global.println("                    change <ident> <- <dom-expr> ,.. using <join-op> <rel expr>");
		Global.println("Example: update R1 change i1 <- i2 using ijoin R2;");
		Global.println("         update R1 using ijoin R2 change i1 <- i2 ;");
		Global.println("");
		Global.println("Note:<rel expr>::= <ident> !");
		Global.println("     <rel expr> [<dom-list>:<join-op>:<dom-list>] <rel expr> !");
		Global.println("     [# ! . ! [<dom-list>]] [quant (<#.dom-expr>)<dom>,..]");
		Global.println("     [where <dom expr>] in <rel expr> ");
		Global.println("     <join-op> ::= <mu-join-op> ! <sigma-join-op> ! nop");
		Global.println("     <mu-join-op>::= ijoin ! djoin ! ujoin ! sjoin ! ljoin ! rjoin ! drjoin ! natjoin ! dljoin ");
		Global.println("     <sigma-join-op>::= gtjoin !  sup  ! eqjoin  !  sub  !  ltjoin !  sep  !");
		Global.println("                        gejoin !          lejoin !          iejoin !div  !");
		Global.println("                       ~gtjoin ! ~sup  ! ~eqjoin ! ~sub  ! ~ltjoin ! icomp !");
		Global.println("     (<negation>::= not )       natcomp");
		Global.println("     (<negation>::= ! )         natcomp");
	  }
	  else if (s.equals("qtsel")){
		Global.println("QT-selector projects data from a relation.");
		Global.println("Syntax: <project-list> quant <quantifier list> [where <condition>] in <rel expr>;");
		Global.println("Example: [s,r,g] quant (.=1) i where s>20 in R;");
		Global.println("Note: <quantifier list> ::= (<domain-expr with # or .>) <domain-expr> !");
		Global.println("                            (<domain-expr with # or .>) <domain-expr>,<quantifier-list>");
	  }
	  else if (s.equals("compu")){
		Global.println("Computation declarations can declare computation and procedure.");
		Global.println("Syntax: comp <identifier> ([param,...]) is");
		Global.println("          [local <idlist> <type>;]");
		Global.println("          [state <idlist> <type>;]");
		Global.println("	      { <statement> [; <statement>]}");
		Global.println("          [alt {<statement> [; <statement>]}]");
		Global.println("          ;");
		Global.println("Example: comp square( s,a) is { s <- a*a};");
		Global.println("");
		Global.println("Computation is a special relation and may be invocated by any T-selector or join.");
		Global.println("Syntax (invocation) : <identifier> \"[\"<comma-separated value list>\"]\";");
		Global.println("Example: square[,2];");
	  }
	   //Start: 2005, Lili
	  else if (s.equals("display2D"))
	  {
		Global.println("display2D displays a relation representing 2D graphics in Xfig.");
		Global.println("Syntax:  "); 
		Global.println("display2D ( (<vocabulary expression>)? ) <graph expression>; \n");  
		Global.println("Example 1 :   New_Rel    <- display2D ( ) PIC;");
		Global.println("Example 2 :   New_Rel2   <- display2D ( VOCAB ) PIC;\n");
		
		Global.println("In Example 1 only relation PIC is passed by the user. ");
		Global.println("In this case, a system built-in relation .vocabulary will be used.\n"); 
		
		Global.println("In Example 2 relation VOCAB and PIC are passed by the user."); 
		Global.println("VOCAB is a relation specified the attributes of relation PIC and their meanings.");
		Global.println("The 2 attributes of relation VOCAB must be : .attribute and .meaning.");
		Global.println(".attribute is in type attr.\n.meaning is a string.\n");
		
		Global.println("Attribute .meaning may have the following values:");
		Global.println("------------------------------------------------");
		Global.println("cart1\ncart2");
		Global.println("sequence :         The polyline vertex sequence number.  Any int >= 1\n");
		
		Global.println("line_colour :");
		Global.println("fill_colour :"); 
		Global.println("text_colour :"); 
		Global.println("                  -1 = Default");
		Global.println("                   0 = Black");
		Global.println("                   1 = Blue");
		Global.println("                   2 = Green");
		Global.println("                   3 = Cyan");
		Global.println("                   4 = Red");
		Global.println("                   5 = Magenta");
		Global.println("                   6 = Yellow");
		Global.println("                   7 = White");
		Global.println("                   8-11 = four shades of blue (dark to lighter)");
		Global.println("                   12-14 = three shades of green (dark to lighter)");
		Global.println("                   15-17 = three shades of cyan (dark to lighter)");
		Global.println("                   18-20 = three shades of red (dark to lighter)");
		Global.println("                   21-23 = three shades of magenta (dark to lighter)");
		Global.println("                   24-26 = three shades of brown (dark to lighter)");
		Global.println("                   27-30 = four shades of pink (dark to lighter)");
		Global.println("                   31 = Gold\n"); 
		
		Global.println("fill_pattern :");
		Global.println("                   -1 = not filled");
		Global.println("                    0 = black");
		Global.println("                    1-19 = \"shades\" of the color, from darker to lighter");
		Global.println("                    20 = full saturation of the color");
		Global.println("                            A shade is defined as the color mixed with black");
		Global.println("                    21-39 = \"tints\" of the color from the color to white");
		Global.println("                            A tint is defined as the color mixed with white");
		Global.println("                    40 = white");
		Global.println("                    41 = 30 degree left diagonal pattern");
		Global.println("                    42 = 30 degree right diagonal pattern");
		Global.println("                    43 = 30 degree crosshatch");
		Global.println("                    44 = 45 degree left diagonal pattern");
		Global.println("                    45 = 45 degree right diagonal pattern");
		Global.println("                    46 = 45 degree crosshatch");
		Global.println("                    47 = horizontal bricks");
		Global.println("                    48 = vertical bricks");
		Global.println("                    49 = horizontal lines");
		Global.println("                    50 = vertical lines");
		Global.println("                    51 = crosshatch");
		Global.println("                    52 = horizontal \"shingles\" skewed to the right");
		Global.println("                    53 = horizontal \"shingles\" skewed to the left");
		Global.println("                    54 = vertical \"shingles\" skewed one way");
		Global.println("                    55 = vertical \"shingles\"skewed the other way");
		Global.println("                    56 = fish scales");
		Global.println("                    57 = small fish scales");
		Global.println("                    58 = circles");
		Global.println("                    59 = hexagons");
		Global.println("                    60 = octagons");
		Global.println("                    61 = horizontal \"tire treads\"");
		Global.println("                    62 = vertical \"tire treads\"\n");
		
		Global.println("line_style :");
		Global.println("                   -1 = Default");
		Global.println("                    0 = Solid");
		Global.println("                    1 = Dashed");
		Global.println("                    2 = Dotted");
		Global.println("                    3 = Dash-dotted");
		Global.println("                    4 = Dash-double-dotted");
		Global.println("                    5 = Dash-triple-dotted\n");
		
		Global.println("line_thickness: Any int between 0 to 1000, inclusive.\n");
		
		Global.println("depth :         Any int between 0 to 999, inclusive.");
		Global.println("                     where larger value means object is deeper than (under)");
		Global.println("                     objects with smaller depth\n");
		
		Global.println("dash_length :   defined as the length, in 1/80 inches, of the on/off");
		Global.println("                     dashes for dashed lines, and the distance between the dots, ");
		Global.println("                     in 1/80 inches,for dotted lines.\n");
		
		Global.println("font :");
		Global.println("                    -1 = Default font");
		Global.println("                    0  = Times Roman");
		Global.println("                    1  = Times Italic");
		Global.println("                    2  = Times Bold");
		Global.println("                    3  = Times Bold Italic");
		Global.println("                    4  = AvantGarde Book");
		Global.println("                    5  = AvantGarde Book Oblique");
		Global.println("                    6  = AvantGarde Demi");
		Global.println("                    7  = AvantGarde Demi Oblique");
		Global.println("                    8  = Bookman Light");
		Global.println("                    9  = Bookman Light Italic");
		Global.println("                    10 = Bookman Demi");
		Global.println("                    11 = Bookman Demi Italic");
		Global.println("                    12 = Courier");
		Global.println("                    13 = Courier Oblique");
		Global.println("                    14 = Courier Bold");
		Global.println("                    15 = Courier Bold Oblique");
		Global.println("                    16 = Helvetica");
		Global.println("                    17 = Helvetica Oblique");
		Global.println("                    18 = Helvetica Bold");
		Global.println("                    19 = Helvetica Bold Oblique");
		Global.println("                    20 = Helvetica Narrow");
		Global.println("                    21 = Helvetica Narrow Oblique");
		Global.println("                    22 = Helvetica Narrow Bold");
		Global.println("                    23 = Helvetica Narrow Bold Oblique");
		Global.println("                    24 = New Century Schoolbook Roman");
		Global.println("                    25 = New Century Schoolbook Italic\n ");
		
		Global.println("font_size:     Any int between 1 to 500, inclusive.\n");
		
		Global.println("join_style : ");
		Global.println("                       defined for LINES only");
		Global.println("                       0 = Miter");
		Global.println("                       1 = Round");
		Global.println("                       2 = Bevel\n");
		
		Global.println("cap_style :");
		Global.println("                       defined for LINES, OPEN SPLINES and ARCS only");
		Global.println("                       0 = Butt");
		Global.println("                       1 = Round");
		Global.println("                       2 = Projecting\n");
		
		Global.println("forward_arrow  :");
		Global.println("backward_arrow :");
		Global.println("                       defined for LINES, OPEN SPLINES and ARCS only");
		Global.println("                       0 = Stick-type");
		Global.println("                       1 = Closed triangle");
		Global.println("                       2 = Closed with \"indented\" butt");
		Global.println("                       3 = Closed with \"pointed\" butt\n");
		Global.println("Ref Site: http://xfig.org/userman/fig-format.html");
	  }
	  //End: 2005 Lili
	  else Global.println("No such help information. Please type \"help\" to list all help information.");
	} //end if
  } //end help implementation,23 May, 2003

  // convert an IDList node to an array of domains
  private Domain[] IDListToDomains(SimpleNode node, Environment env)
       throws InterpretError
  {
    int n = node.jjtGetNumChildren();
    Domain[] doms = new Domain[n];

    for (int i = 0; i < n; i++) {
      String s = ((SimpleNode) node.jjtGetChild(i)).name;
      Domain d = env.lookupDom(s, true);
      if (d == null)
	throw new InterpretError("domain : " + s + " not exist");
      else if (d.isStateDom == 1) //Yi Zheng May 6, 2002
      throw new InterpretError("domain : " + s + " is a state domain");
      else
	doms[i] = d;
    }

    return doms;
  }
  public long interpret(SimpleNode node, Environment env)
       throws InterpretError
  {
	//zhan


	    /*
	    System.out.println("---- enter interpret --------"); //zhan
		Global global= new Global();
		System.out.println("*** enter interpret ***");
		global.traverse(node);
		System.out.println();
		*/


	//zhan

    //System.out.println("enter interpret");
    if (node == null) return 0;
    long ts = System.currentTimeMillis();
	int type = node.type;

	// start Zongyan's addition, July 22 2002

	if (node.url != null & type != ALDATP_ASSIGNMENT & type != ALDATP_UPDATE & type != ALDATP_DECLARATION) {
		if (Debug) System.out.println("call aldatpClient in interpret");
		aldatpClient.request(node,env);  // need to check response ^_^
		System.out.println(aldatpClient.rspCode);
		return System.currentTimeMillis() - ts;
	}
	// end Zongyan's addition,  July 22 2002

        /******* start adding by Fan  *******/
        relLevel = 0;// June 6, 2004 Fan Guo
        NRelName = new Vector();// June 11, 2004
        /******* end of Fan's addition  *******/

    switch(type)
      {  case OP_COMMAND :
	executeCommand(node, env);
      break;
      case OP_STATEMENT :
	executeStatement(node, env);
	break;
      case OP_DECLARATION :
	executeDeclaration(node, env);
	break;

		//start zhan' addition Nov 2003
		case OP_XML :
		    //Global global= new Global();
			//System.out.println(" ----------------- OP_XML------------");
			//global.traverse(node);

	        //initialization from file
	        //      Declaration
	        //      /         \
	        //    IDList     Init
	        //     /           /
	        //  Identifier   FILEPATH

	        SimpleNode filePathNode = (SimpleNode)node.jjtGetChild(1).jjtGetChild(0);
	        if(filePathNode.opcode == OP_FILEPATH)
	        {
	          SimpleNode relName = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	          String inputFileName = filePathNode.name.substring(1,
	                            filePathNode.name.length()-1);

	          String s1,s2 = new String();

	          try
	          {
	            //get the content of the file
	            DataInputStream inp = new DataInputStream
		          (new BufferedInputStream(new FileInputStream(inputFileName)));
		            while((s1 = inp.readLine())!= null)
		               s2 += s1 + "\n";

	             s2 = s2.trim();
			 }
			 catch(IOException e){System.out.println("IOException in RelationalInitialization");}

	          String input="relation "+relName.name+"<-"+s2+";";

			  byte[] barray=input.getBytes();
			  ByteArrayInputStream binstream=new ByteArrayInputStream(barray);
			  InputStream bin=(InputStream)(binstream);
			  Global.jin.resetIn(bin);
			  parser.ReInit(Global.jin);
			  instack.push(in);
			  in = bin;

			}
	        else  //xml input
	        {
			  node = XMLInitialization(node,env);
			  executeDeclaration(node, env);
		    }

		break;
	//end of zhan's addition Nov2003


		case ALDATP_DECLARATION: //Zongyan Aug 2002
	reverseView(node,env);
	try{
    node = aldatpUtility.getRootNode();
	} catch (IOException e)
	{ throw new InterpretError(" IO error in rebuild tree");
	}
	interpret(node,env);
	break;

      case OP_ASSIGNMENT :
//	executeAssignment(node, env);
	case ALDATP_ASSIGNMENT :  // Zongyan Aug 2002
	executeAssignment(node, env);   // local evaluate, remote assignment
	break;
	   case ALDATP_UPDATE:  //Zongyan Aug 2002
	   lefthandUpdate(node,env);

	break;
      case OP_UPDATE :
	Environment uenv = new Environment(env.reltable, env.domtable);
//	Environment uenv = new Environment(env, env.getTable(), env.reltable, env.domtable);
	// produce trigger tree
	int diftrig = lookUpdate(node, uenv, null, uenv.triggerroot);
	// reverse the result -- undo trigger
	doTrigger(env, uenv.triggerroot, 1);
	if (diftrig == 1)
	  executeUpdate(node, node, uenv);
	RemoveTrigger(uenv.triggerroot);		// clear up
	break;
      default : // impossible to be here
	System.out.println("unknown statement");
	break;
      }
    return System.currentTimeMillis() - ts;
     }

  // Weizhong
  // overload the original interpret to handle the UNDO command in
   // event handlers.
  public long interpret(SimpleNode node, Environment env, String rlname, int opc)
       throws InterpretError
  {
    if (node == null) return 0;
    int type = node.type;

    switch(type)
      {  case OP_COMMAND :
	executeCommand(node, env, rlname, opc);

      break;
      default :
	interpret(node, env);
	break;
      }
    return 1;
  }
    //added by Zongyan Aug,2002
  private void lefthandUpdate(SimpleNode node, Environment env)
	  throws InterpretError
  {
	  String tempname = null;
	  Relation rr = null;
	  String url,name,rname;
	  Domain[] domsr = null ,domsl = null;
	  boolean flag = false;

      switch (node.opcode)
	  {
	  case OP_ADD:

	  case OP_DELETE:
	       tempname = nextTempName();
	       rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), tempname, new Hashtable(), env);
           String lname = ((SimpleNode) node.jjtGetChild(0)).name;
		   tempname = nextTempName();
		   Relation rel = rr.assign(domsr, tempname, domsl);
		   env.put(rel);
		   distributedStmt(rel.name,lname,node.url,node.opcode,env);
		   env.removeRel(rel.name, false);
		   break;

	  case OP_CHANGE:
		   try{
		   File outputFile = new File("_tmp_aldatp_tmpFile.txt");
           FileOutputStream os = new FileOutputStream(outputFile);
           PrintWriter outFile = new PrintWriter(os);
		   if (Global.start_type == 0 )
		    outFile.println( "StatementBlock" + "|" + node.url + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 2);
		   else
		    outFile.println( "StatementBlock" + "|" + node.url + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		   SimpleNode tmpnode = node.updateChangeDump(outFile);

		   if (tmpnode != null){
			 if (tmpnode.opcode==OP_IDENTIFIER){
			   rname = tmpnode.name;
			 } else
			 {
		       tempname = nextTempName();
	           rr = evaluateTLExpression((SimpleNode) tmpnode, tempname, new Hashtable(), env);
               env.put(rr);
			   rname = rr.name;
			   flag = true;
			 }

		  if (Global.start_type == 0 ) { // for stand alone JRelix, normal server can also go this way

			  name = aldatpClient.getRelName(node.url,env);

			 aldatpClient.putRel(rname,name,node.url,env);	// response ^_^
		  }
		  else { // for top-level server, normal server or prtoected server
			name = rname;
			String path = System.getProperty("user.dir");
		    if( path.indexOf("public_aldatp") >= 0 ){
		    if (path.endsWith("public_aldatp"))
	    	  url = "aldatp://" + aldatpUtility.getHostname() +"/"  ;
		    else
			  url = "aldatp://" + aldatpUtility.getHostname() +"/" + path.substring(path.indexOf("public_aldatp") + 14) + System.getProperty("file.separator");
		    }
		    else url = "aldatProtect://" + aldatpUtility.getHostname() +"/" + path;


		    outFile.println( "ALDATP" + "|" + url + "|" + OP_ALDATP + "|" + OP_ALDATP + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		  } //server
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|"
					+ null + "|" + null + "|" + 0);
		   if (Global.start_type == 0 ) {

		   outFile.println( "Command" + "|" + null + "|" + OP_COMMAND + "|" + OP_DELR + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		   outFile.println( "IDList" + "|" + null + "|" + OP_IDLIST + "|" + OP_IDLIST + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|"
					+ null + "|" + null + "|" + 0);
		  }		  }//tmpnode!= null
		  else {
		   if (Global.start_type == 0 ) {		   outFile.println( "Command" + "|" + null + "|" + OP_COMMAND + "|" + OP_PR + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);

		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + ((SimpleNode) node.jjtGetChild(0)).name + "|" + 0 + "|"
					+ null + "|" + null + "|" + 0);
			   }		   }
          outFile.close();
		  node = aldatpUtility.getRootNode();

	     }
	     catch (IOException e){
	  	   Global.processThrowable(e);
		 }

	     interpret(node,env);

	     if(flag) env.removeRel(rr.name, false);

	     break;

	  default: //impossible
	      break;
	  }
  }

  //supose the following declaration is  made on mimi:mydir
  // aldatp://willy/dir1/(myView) is R;
  // reverseView() transfer the left hand remote declaration to the remote statement:
  // aldatp://willy/dir1{ myView is aldatp://mimi/mydir/(R)};
  //
  // Zongyan  Aug 2002
  private void reverseView(SimpleNode node, Environment env)

	  throws InterpretError
  {
	  String URL = node.url;
	   try{
		   File outputFile = new File("_tmp_aldatp_tmpFile.txt");
           FileOutputStream os = new FileOutputStream(outputFile);
           PrintWriter outFile = new PrintWriter(os);
		   outFile.println( "StatemnetBlock" + "|" + URL + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);		 ((SimpleNode) node.jjtGetChild(0)).url = URL;
		 if(Debug) System.out.println("child0.url:"+  ((SimpleNode) node.jjtGetChild(0)).url);			node.dump(outFile,URL,env);
            outFile.close();

	   }
	   catch (IOException e){
		   Global.processThrowable(e);
		}
	}
    public void distributedStmt(String rname, String lname, String URL, int op,Environment env)
	  throws InterpretError
  {
	  SimpleNode node = null;
	  String url = null;
	  String name = null;
	   try{
		   File outputFile = new File("_tmp_aldatp_tmpFile.txt");
           FileOutputStream os = new FileOutputStream(outputFile);
           PrintWriter outFile = new PrintWriter(os);
		   if(Global.start_type == 0 )
		  outFile.println( "StatementBlock" + "|" + URL + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 2);
		   else
		   outFile.println( "StatementBlock" + "|" + URL + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		  switch (op)
		  {
		  case OP_ASSIGN:
		   outFile.println( "Assignment" + "|" + null + "|" + OP_ASSIGNMENT + "|" + OP_ASSIGN + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 2);
		   break;
		  case OP_ASSIGNADD:
		   outFile.println( "Assignment" + "|" + null + "|" + OP_ASSIGNMENT + "|" + OP_ASSIGNADD + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 2);
		   break;
		   case OP_ADD:
		   outFile.println( "Update" + "|" + null + "|" + OP_UPDATE + "|" + OP_ADD + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 2);
		   break;
		   case OP_DELETE:
		   outFile.println( "Update" + "|" + null + "|" + OP_UPDATE + "|" + OP_DELETE + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 2);
		   break;
		  }
		  outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + lname + "|" + 0 + "|"
					+ null + "|" + null + "|" + 0);


		  if (Global.start_type == 0 ) { // for stand alone JRelix, server can also go this way
			 name = aldatpClient.getRelName(URL,env);

			 aldatpClient.putRel(rname,name,URL,env);	// response ^_^
		  }
		  else { // for top-level server, normal server or prtoected server
		  name = rname;
		  String path = System.getProperty("user.dir");
		  if( path.indexOf("public_aldatp") >= 0 ){

		  if (path.endsWith("public_aldatp"))
	    	  url = "aldatp://" + aldatpUtility.getHostname() +"/"  ;
		  else
			 url = "aldatp://" + aldatpUtility.getHostname() +"/" + path.substring(path.indexOf("public_aldatp") + 14) + System.getProperty("file.separator");
		  }
		  else url = "aldatProtect://" + aldatpUtility.getHostname() +"/" + path;

		  outFile.println( "ALDATP" + "|" + url + "|" + OP_ALDATP + "|" + OP_ALDATP + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		  }
		  outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|"
					+ null + "|" + null + "|" + 0);
		  if (Global.start_type == 0 ) {
		   outFile.println( "Command" + "|" + null + "|" + OP_COMMAND + "|" + OP_DELR + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		   outFile.println( "IDList" + "|" + null + "|" + OP_IDLIST + "|" + OP_IDLIST + "|" + null + "|" + 0 + "|"
					+ null + "|" + null + "|" + 1);
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|"
					+ null + "|" + null + "|" + 0);
		  }
          outFile.close();
		  node = aldatpUtility.getRootNode();

	   }
	   catch (IOException e){
		   Global.processThrowable(e);
		}

	  interpret(node,env);

  }
  // check if there is self-reference in a view expression, the semantic
  // check for the existence of operands in the view expression is also
  // done, the hash table contains names already checked to avoid
  // infinite recursive call for mutual recursion
  private boolean isRecursiveView(SimpleNode node, String name,
				  Hashtable htable, Environment env)
       throws InterpretError
  {
    int type = node.type, opcode = node.opcode;
    boolean flag = false; // flag indicating reference to name thus cycle

    switch(type) {
    case OP_IDENTIFIER : {
      if (node.name.equals(name)) { // there is cycle
	flag = true;
      } else { // continue to do semantic check
	// check if this ID is already checked
	String s = (String) htable.get(node.name);
	if (s != null) // already checked
	  return flag;
//	  System.out.println("looking for: " + node.name);

	if (node.url == null){ // added by Zongyan
	// remote view definition will be shipped to remote site.
    // aldatp://mimi/(v) is aldatp://mimi/(V) ijoin R
	// is converted to aldatp://mimi/{ V is V ijoin aldatp://myhost/mydir/(R)}
	// so interpreter will nerver encounter such situation : url!= null and is recursive

	Relation rel = env.lookupRel(node.name, true);
	if (rel == null) {
	  throw new InterpretError("no relation : " + node.name
				   + " in view " + name);
	}
	// REVISIT : computation
	if (rel.rvc == VIEW) {
	  SimpleNode view = rel.tree, nd = null;
	  // get the expression body of the view
	  if (view.isBitSet(0)) // there is initial part
	    nd = (SimpleNode) view.jjtGetChild(1);
	  else
	    nd = (SimpleNode) view.jjtGetChild(0);

	  // add this ID to hash table so no check for it again
	  htable.put(node.name, node.name);

	  // recursively check and set flag
	  flag = flag || isRecursiveView(nd, name, htable, env);
	}
      }
      }
      return flag;
    }
    case OP_JOINOPERATOR : {
      SimpleNode nodel = null, noder = null;


      // get expression tree of left operand
      nodel = (SimpleNode) node.jjtGetChild(0);

      // get expression tree of right operand
      if (node.jjtGetNumChildren() == 2) { // no join attributes
	noder = (SimpleNode) node.jjtGetChild(1);
      } else { // node.jjtGetNumChildren() == 4
	noder = (SimpleNode) node.jjtGetChild(3);
      }

      flag = flag || isRecursiveView(nodel, name, htable, env);
      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_PROJECT : {
      // get expression tree of relational expression

      SimpleNode noder = (SimpleNode) node.jjtGetChild(1);

      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_TSELECT : {
      // get expression tree of the selector
      SimpleNode noder = (SimpleNode) node.jjtGetChild(1);

      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_SELECT : {
      // get relational expression tree of the selector
      SimpleNode noder = (SimpleNode) node.jjtGetChild(1);

      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_QSELECT : {
      //this is modified by hongyu. Sept 30, 01
      //throw new InterpretError("TO BE IMPLEMENTED : QSelector");
      // get relational expression tree of the selector

      SimpleNode noder = null;
      if(node.jjtGetNumChildren() == 2)
	noder= (SimpleNode)node.jjtGetChild(1);
      noder = (SimpleNode)node.jjtGetChild(2);

      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    } // end of hongyu modification

	case OP_ARRAY: //Yi Zheng May 31, 2002
    case OP_PICK : {
      // get relational expression tree
      SimpleNode nd = (SimpleNode) node.jjtGetChild(0);

      flag = flag || isRecursiveView(nd, name, htable, env);
            flag = flag || isRecursiveView(nd, name,  htable, env);
      return flag;
    }
	 case OP_ALDATP : { // Zongyan Aug 2002 ^_^

      flag = flag || false;
      return flag;
    }
    case OP_EDIT :
    case OP_ZORDER :
    case OP_EVAL : // ???
    case OP_RENAME : { // ???
      throw new InterpretError("TO BE IMPLEMENTED");
    }
    // many more other case ...
    default : {
      throw new
	InterpretError("invalid/not implemented relational expression");
    }
    }

  }

     /**
   * method that  traverse and output the dump info to screen
   * for debug purpose
   * @author Zongyan Wang July 19, 2002
   */
    public void traverse(SimpleNode node) {	  
        System.out.println("identifer:" + node.identifier + " name:" + node.name + " opcode:" + node.opcode + " type:" + node.type);
        System.out.println("numChildren:" + node.jjtGetNumChildren() +" info:" + node.info + " url:" + node.url);
        for(int i = 0; i < node.jjtGetNumChildren(); i++) {
             traverse((SimpleNode) node.jjtGetChild(i));
        }  
    }

  /**
   * The main loop for the interpretation.
   * @param args java.lang.String[]
   */
  public void mainLoop(String args[]) {

    in = Global.instream;
    instack = new Stack();
    parser = new JRelixParser(Global.jin);
    Global.print(">"); Global.flush();

    //testing, Sept 29, 2002

    //executeJRelixCommand("pr (where .rel_name = \".rel\" in .rd);");

    //end of testing

    while (true) {
      try {
	Global.beforeParse();
	Global.status = parser.Start();
	if (Global.status == OP_EOF) {


	  if ((in instanceof FileInputStream)||(in
						instanceof ByteArrayInputStream)) {


	    try {
	      Object o = instack.pop();
	      if (o instanceof FileInputStream) {
		in.close();
		in = (FileInputStream) o;
	       } else
	       {
		 in.close();
		 in = (InputStream) o;
	       }
	      Global.jin.resetIn(in);
	      parser.ReInit(Global.jin);
	    }
	     catch (Exception e) {
	      Global.println("Exception during IO redirection: reset to System.in for input");
	      in = Global.instream;
	      Global.jin.resetIn(in);
	      parser.ReInit(Global.jin);
	    }
	  }

	} else {
	  Global.root = (SimpleNode) parser.jjtree.rootNode();
	  Global.afterParse(env);
	  Global.beforeInterpret();
	  Global.interpretTime = interpret(Global.root, env);

	  Global.afterInterpret();
	}
      } catch (ParseError e) {
	Global.processThrowable(e);
	//previous code:	parser.ReInit(new JRelixInputStream(in));

	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } catch (InterpretError e) {
	Global.processThrowable(e);
	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } catch (Exception e) {
	Global.processThrowable(e);
	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } catch (Error e) {
	Global.processThrowable(e);
	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } finally {
		  //Added by KANG, shunhong, Aug,12 2003

		  Global.save();
		  //end add
	parser.jjtree.reset();
       }
    }
  }
  /***************************************************************/
  /*          METHODS USED FOR DECLARING A VIRTUAL DOMAIN        */
  /***************************************************************/

   /**
    * This method is used for virtual domains.

    *   e.g: let X be A+B;
    */
  public Domain newVirDomain(String name, SimpleNode node, Environment env)
       throws InterpretError
  {
    int domtype;
    SimpleNode declNode;

    if(node.jjtGetNumChildren() == 1)
      // if(node.jjtGetNumChildren() == 2)
    {
	declNode = (SimpleNode)node.jjtGetChild(0);
	//		declNode = (SimpleNode)node.jjtGetChild(1);
    }
    else
    {  // Something must be wrong.
	throw new InterpretError("Domain: syntax error: <INITIAL> not supported.");
    }

    // First make sure that all referenced domains in the syntax tree
    // exist, otherwise, throw an exception.

    // May 2, 2002 Yi Zheng: now also checks if any of the reference
    //domains has isStateDom == 1, if so, throw an exception
    traverseNode(declNode, 0, env);

    // If this domain already exists, we just insert/replace the
    // syntax tree to that domain's 'tree' field. Note, the reference
    // counters of those referenced domains are not incremented.
    // However, before we do some actual work, we must make sure
    // that old/new domains' types are identical, even when we want
    // to overwrite a virtual domain since we can't differentiate a
    // true virtual from a 2-meaning domain item (i.e. normal/virtual
    // domain).
    // Domain doment = env.lookupDom(name, false);  //Modified by Zongyan July 28 2005
    Domain doment = env.lookupDom(name, true);

    // Get the actualized type. Note, this type might be changed
    // later, so must be checked again at run-time.
    domtype  = traverseType(declNode, env);
 //   System.out.println(" end of traverseType");
    if(doment != null)
    {  // domain exists.
		if(doment.type != domtype)
	  	{  
                        if(doment.type == UNDEF) // Jiantao 2004
                        {       // update the old domain
                                doment.type = domtype;
                                doment.tree = node;
                        }
                        else
                          throw new InterpretError("error: new type of '" + doment.name
				      + "' is '" + Utility.type2String(domtype)
				      + "', but the old type is '"
						+ Utility.type2String(doment.type) + "'.");
	  	}
		else
	  	{
                        // start adding by Fan on Sep 1, 2004
                //      if(domtype != IDLIST)     // do not insert tree yet if it is the redefining of virtual nested relation
                        //end of Fan's addition
	    	doment.tree = node;
	    // Weizhong
	    //		doment.tree.jjtRemoveChild(0);
	  	}
    }
    else
    {  // this is a new domain
	//	   doment = new Domain(name, domtype, 0, node);
	//	Node removenode = node.jjtGetChild(0);

	//	node.jjtRemoveChild(0);
		doment = new Domain(name, domtype, 0, node);
	//	node.jjtAddChildAtHead(removenode);
    }
    //COMP added by Yi Zheng, March 20, 2002
    if(domtype == COMP) //r is only used here for sd purpuse
    { Domain[] doms = getIDList(declNode, env).domains;
      // Add it to relTable without idlist info.
      Relation r = new Relation("."+name, doms, RELATION, 0,
				doms.length, 0, null, env);
      env.put(r);
    }
    else if(domtype == IDLIST) //Yi Zheng June 17, 2002
    {
	//Aug 13
       //Utility.dump(declNode, "declNode ");

		Relation rdecl = getIDList(declNode, env);
      	Domain[] doms = rdecl.domains;
      	Domain iddom = env.lookupDom(".id", true);
		Domain[] newDoms = null;
		if(!rdecl.isDomainIn(iddom))
		{
	  		newDoms = new Domain[doms.length + 1];
	  		newDoms[0] = iddom;
	  		for(int i = 1; i < newDoms.length; i++)
	  			newDoms[i] = doms[i-1];
		}
		else
		{
			newDoms = new Domain[doms.length];
			for(int j = 0; j < newDoms.length; j++)
				newDoms[j] = doms[j];
		}
/*
                Relation r = new Relation ("." + name, newDoms, RELATION, 0, newDoms.length, 0, null, env);
                env.put(r);
                */  // this is original code. Modified to the following code on July 31, 2004 by Fan Guo
                // these modification are used for renaming of nested virtual domain.
                // start modification: should be used together with modification code in AddVirtDom2DomTable()
                // July 31, 2004
                // name: name of the virtual nested relation declaring;
                // newDoms: the domains that the new virtual nested relaiton defined on
                doment = VirtualNestedRelationCreation(doment,node, newDoms, env);  // August 11, 2004 dealing
                //doment = NestedDomainCreation(doment,newDoms, env);  // August 11, 2004 dealing
    }
    return doment;
  }
  /***************************************************************/
  /*                  OTHER SHORT UTILITY METHODS                */
  /***************************************************************/

  // Return the a temporary name used as the file name for intermediate
  // result in relational operation
  public static String nextTempName() //Yi Zheng June 11, 2002
//public String nextTempName()
  {
    int n = (next_number++)%MAX_TEMP_FILE;
    return temp_name_prefix + Integer.toString(n);
  }
  //REVISIT: It seems this is never used, what is it for?
  private void parse(String sp)
  {
    StringBufferInputStream sin = new StringBufferInputStream(sp);
  }
  /***************************************************************/
  /*            METHODS USED FOR DECLARING A RELATION            */
  /***************************************************************/

  /**
   * Code to handle parsing of a relational declaration with
   * optional initialization
   */
  private void RelationalDeclaration(SimpleNode node, Environment env)
   throws InterpretError
  {

    //start zhan's addition Nov2003
    //To handle relation R(dom list)<-"..\..\"; while file is "{(...)}"
    boolean isFileInput = false;
    if(node.jjtGetNumChildren()==3)
    {
	 SimpleNode filePathNode=(SimpleNode)node.jjtGetChild(2).jjtGetChild(0);
	 if(filePathNode.opcode==OP_FILEPATH)
     {
	  String inputfilename = filePathNode.name.substring(1,filePathNode.name.length()-1);
	  String s1 = new String();
      String s2 = new String();
      String s3= new String();
      try
      {
	  	   //get the content of the file
		   DataInputStream inp = new DataInputStream
		   (new BufferedInputStream(new FileInputStream(inputfilename)));
			while((s1 = inp.readLine())!= null)
			   s2 += s1 + "\n";
 		   s2 = s2.trim();
      }
	  catch(IOException e){System.out.println("IOException in RelationalInitialization");}

	  s3 = s2.substring(0,1);
	  if(s3.equals("{")) //seperate from yuling's filepath
	  {  inputFromFile(node,s2);
	     isFileInput = true;
	  }
	 }

    }    //end of if(node.jjtGetNumChildren()==3)

    if(isFileInput == false)
    {
    //end zhan's addition Nov2003


    // type of the new relation(s)
    SimpleNode type = (SimpleNode) node.jjtGetChild(1);
    // typeArray is the array of domains for the new relation(s)
    Domain[] typeArray = new Domain[type.jjtGetNumChildren()];

    String recursiveDomName = null;//from zhan
    boolean isRecur = false;//from zhan

    for(int i = 0; i < typeArray.length; i++) {
      String dname = ((SimpleNode) type.jjtGetChild(i)).name;
      // check if this domain is already declared
      Domain d = env.lookupDom(dname, true);
      if(d == null) {
		throw new
	  InterpretError("domain "+ dname + " has not been declared");

      }
      else if (d.isStateDom == 1) //Yi Zheng May 9, 2002
      throw new InterpretError("domain " + dname + " is a state domain");

      //***start of zhan's addition
      //record whether the rel contains a recursive domain and what its name
      if(d.type == IDLIST)		
      {
	   Relation Recur = env.lookupRel("."+dname,true);

       //System.out.println("recur dom is "+dname);
 	   for(int j=0;j<Recur.domains.length;j++) //N(A,B,N)
       {
	    if(Recur.domains[j].name.equals(dname))
	    {
		    recursiveDomName = dname;
		    isRecur=true;
		    break;
	    }
	    else if(Recur.domains[j].name.equals(dname+"_0"))    //N(A,B,N_0)
	    {
            //System.out.println("Recur.domains[j].name.equals(dname+_0)");
		recursiveDomName = dname;
		isRecur=true;
		break;
	    }
	    else if (Recur.domains[j].type==UNION)//union type Aug 10,2004 --added by Yu
	    {
           Vector compnames = new Vector(); //stack of domain names to process
	   compnames.addElement(Recur.domains[j].name);
	   while (compnames.size()>0)//hasMoreUnion
	   {
	     String domnm = (String) compnames.lastElement();
	     //System.out.println("union domain:"+domnm);
    	 compnames.removeElementAt(compnames.size() - 1);
 	     Domain domC = env.lookupDom(domnm, true);
	     for (int j1 = 0; j1<domC.union.length; j1++)	//check for each possible type to see if it is a IDList
	     {
		  if (domC.union[j1].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;
		  {
		    String nameD = domC.union[j1].name;
		    //System.out.println("union domain "+j1+" is "+nameD);
		    if (env.lookupDom(nameD,true)!=null)
		    {
			if ((env.lookupDom(nameD,true)).type==IDLIST)
			{//System.out.println("union domain name: "+nameD+" current domain name: "+dname);
		   	  		if(nameD.equals(dname))
		  	  		{//System.out.println("isRecur is true");
		    			 recursiveDomName = dname;
		    			 isRecur=true;
		    			 break;
		  	  		}
		  	  		else if(nameD.equals(dname+"_0"))    //N(A,B,N_0)
		   	  		{
            		//System.out.println("Recur.domains[j].name.equals(dname+_0)");
					 recursiveDomName = dname;
					 isRecur=true;
		    			 break;
	      	  			}
	      	  	} 
	      	  	else if ((env.lookupDom(nameD,true)).type==UNION)
			{
	 	 			compnames.addElement(nameD);
	 		}	      	  	
		    }	//	if (env.lookupDom(nameD,true)!=null)
	   	  }//if (domC.union[j1].opcode==OP_IDENTIFIER)
	     }//	for (int j1 = 0; j1<domC.union.length; j1++)
	   }//while
	  }//else if (Recur.domains[j].type==UNION)
	}//for
      }//end if (d.type == IDLIST)
      //end of zhan's addition
      else if (d.type == UNION)		//need to deal with union type July 08,2004 by Yu
      {
        Vector compnames = new Vector(); //stack of domain names to process
	compnames.addElement(d.name);
	while (compnames.size()>0)//hasMoreUnion
	{	
         String domnm = (String) compnames.lastElement();
         compnames.removeElementAt(compnames.size() - 1);
 	 Domain domC = env.lookupDom(domnm, true);
	 for (int j = 0; j<domC.union.length; j++)	//check for each possible type to see if it is a IDList
	 {
	  if (domC.union[j].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;

	  {
	   String nameD = domC.union[j].name;
	   if (env.lookupDom(nameD,true)!=null)
	   {
	    if ((env.lookupDom(nameD,true)).type==IDLIST)
	    {
	     //System.out.println("here!:union types with idlist:"+domC.union[j].name);
	     Relation Recur = env.lookupRel("."+nameD,true);
             //System.out.println("recur dom is "+nameD);
 	     for(int jj=0;jj<Recur.domains.length;jj++) //N(A,B,N)
             {
	      if(Recur.domains[jj].name.equals(nameD))
	      {
		    	 recursiveDomName = nameD;
		    	 isRecur=true;
		    	 break;
	      }
	      else if(Recur.domains[jj].name.equals(nameD+"_0"))    //N(A,B,N_0)
	      {
            	//System.out.println("Recur.domains[j].name.equals(dname+_0)");
				recursiveDomName = nameD;
				isRecur=true;
		    		break;
	      }
	      else if (Recur.domains[jj].type==UNION)//union type Aug 10,2004 --added by Yu
	      {
        	Vector compnames1 = new Vector(); //stack of domain names to process
	    	compnames1.addElement(Recur.domains[jj].name);
	    	while (compnames1.size()>0)//hasMoreUnion
	    	{
	         String domnm1 = (String) compnames1.lastElement();
    	     	 compnames1.removeElementAt(compnames1.size() - 1);
 	    	 Domain domC1 = env.lookupDom(domnm1, true);
	   	 for (int j1 = 0; j1<domC1.union.length; j1++)	//check for each possible type to see if it is a IDList
	   	 {
		  	  if (domC1.union[j1].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;
		  	  {
		   		String nameD1 = domC1.union[j1].name;
		   		if (env.lookupDom(nameD1,true)!=null)
		   		{
				 if ((env.lookupDom(nameD1,true)).type==IDLIST)
				 {
		   	  		if(nameD1.equals(nameD))
		  	  		{
		    			recursiveDomName = nameD;
		    			isRecur=true;
		    			break;
		  	  		}
		  	  		else if(nameD1.equals(nameD+"_0"))    //N(A,B,N_0)
		   	  		{
            		//System.out.println("Recur.domains[j].name.equals(dname+_0)");
						recursiveDomName = nameD;
						isRecur=true;
		    				break;
	      	  			}
	      	  	 	 } 
	      	  	 	 else if ((env.lookupDom(nameD1,true)).type==UNION)
			     	 { 
	 	 			compnames1.addElement(nameD1);
	 			 }	      	  	
		  	   	}	//	if (env.lookupDom(nameD,true)!=null)
	   	 	  }//if (domC.union[j1].opcode==OP_IDENTIFIER)
	    	 }//	for (int j1 = 0; j1<domC.union.length; j1++)
	       }//while
	      }//else if (Recur.domains[j].type==UNION)

	     }//end for(int jj=0;jj<Recur.domains.length;jj++)
			  //hasMoreUnion=false;//--no!!! idlist can have union domain! deal with it at the following
	    }//end of if ((env.lookupDom(nameD,true)).type==IDLIST)
	    else if ((env.lookupDom(nameD,true)).type==UNION)
				compnames.addElement(nameD);
	   }//if (env.lookupDom(nameD,true)!=null)
	  }//if (domC.union[j].opcode==OP_IDENTIFIER)	   		
	 }//for (int j = 0; j<domC.union.length; j++)
	}//while (compnames.size()>0)
      }	//else if (d.type == UNION)		//July 08,2004 by Yu

      typeArray[i] = d;
    }	//end of for(int i = 0; i < typeArray.length; i++)

    // add the new relations to the reltable
    SimpleNode rels = (SimpleNode) node.jjtGetChild(0);
    for(int i = 0; i < rels.jjtGetNumChildren(); i++) {
      String relname = ((SimpleNode) rels.jjtGetChild(i)).name;
      if(env.lookupText(relname, true) != null)  // Jiantao
                throw new InterpretError("Error: relations are not allowed to have duplicate names as texts.");
      Relation r = new Relation(relname, typeArray, RELATION,
				0, typeArray.length, 0, null, env);
      env.put(r);
    }



    // perform initialization of the relation(s) if necessary
    if (node.jjtGetNumChildren() == 3) {
      // nd is node of tuples		//init->tuplelist --comment by Yu
      SimpleNode nd = (SimpleNode) node.jjtGetChild(2).jjtGetChild(0);

      // initialize the first relation
      String name = ((SimpleNode) rels.jjtGetChild(0)).name;	//relation name
      Hashtable relsHT = CreateDataHashTable(name, env);	//for example: domain c1 string|n1
      														//relation T1(a,c1) will return with T1 and n1
      RelationalInitialization(nd, relsHT, name, 0, 0);

      //zhan starting :handle recursive domains
      if (isRecur == true)
      {
          //System.out.println("isRecur  === true");
          //System.out.println("recursiveDomName = " + recursiveDomName);
          //System.out.println("levels  is "+ levels);
          Domain newRecurDomain = null; //new Domains used to create recursiveRelation
          String newDomainName ;
          String updateDomain;
          Relation recurRelation;
          Relation recursiveRelation = (Relation) relsHT.get("."+recursiveDomName);
          int len = recursiveRelation.domains.length;
          Domain[] recurDoms = new Domain[len];

          Relation firstRecur = (Relation) relsHT.get("."+recursiveDomName);
		  Domain[] Reldom = firstRecur.domains;

          if(levels!=0) //N_0 !=DC
          {//System.out.println("this is it! N_0 !=DC, recursive domain type is idlist "+levels);
	        newRecurDomain = new Domain(recursiveDomName+"_0",IDLIST,0,null);
	      }
	      else //N_0=DC
	        newRecurDomain = new Domain(recursiveDomName+"_0",LONG,0,null);

	      env.putRecurDom(newRecurDomain);

		  for(int j=0;j<Reldom.length;j++)
		  {
		    if(Reldom[j].name.equals(recursiveDomName))
		        Reldom[j] = newRecurDomain;
		  }

	      for (int i=0;i<levels;i++)
          {
		    //N_0[...N_1] N_1(..N_2)...
		   	newDomainName = recursiveDomName + "_"+ i;
			updateDomain = recursiveDomName + "_"+ (i+1); //domain:N[1]
			if((i+1)==levels)
			  newRecurDomain = new Domain(updateDomain,LONG,0,null);
			else

			  newRecurDomain = new Domain(updateDomain,IDLIST,0,null);
			env.putRecurDom(newRecurDomain);

			//change N[0](A,B,N) -> N[0](A,B,N[1])
			recurRelation = (Relation) relsHT.get("."+newDomainName);//relation:N[0]
			recurDoms = recurRelation.domains; //A,B,N  ??????
			String recurN = null;
			for(int j=0;j<recurDoms.length;j++)
			{
			  if(recurDoms[j].name.indexOf('_')!=-1)
			     //in former initialation Domain of N(B,N) was changed to (B,N_0)
                  recurN = recurDoms[j].name.substring(0,recurDoms[j].name.indexOf('_'));
              else
                  recurN = recurDoms[j].name;
			  if(recurN.equals(recursiveDomName))
			  	  recurDoms[j] = newRecurDomain;
			     //System.out.println("j=" + j + " "+ recurDoms[j].name);
			}

	      }

       }     // if (isRecur == true)
     //***ending of zhan's addition


      // append data for nested domains, note that there is no need to
      // check for duplicate in appending data for nested domains
    //start: check if relation has union domain July 24,2004 by Yu
	boolean hasUnion = false;    
    
    Vector unionN = new Vector(); //stack of domain names to process
	unionN.addElement(name);
	while (unionN.size()>0&&!hasUnion)//hasMoreUnion
	{	
     String domnm = (String) unionN.lastElement();
     unionN.removeElementAt(unionN.size() - 1);
     Domain[] domsR = env.lookupRel(domnm, true).domains;
	 for (int i = 0; i < domsR.length; i++) 
	 {
	  String s = domsR[i].name;
      if(domsR[i].type == IDLIST && !s.equals(".id"))//.id of type IDLIST

	  {//System.out.println("domain type is idlist:"+domsR[i].name);
	    	  unionN.addElement("." + domsR[i].name);
	  }
	   //added by Yu, July 24,2004		
	  else if (domsR[i].type == UNION)
	  {
	  	hasUnion = true;
	  	break;
	  }
	 }//for
	} //while (domnames.size()>0)(hasMoreUnion)
	//System.out.println("here");  
    //end: check if relation has union domain July 24,2004 by Yu
	Vector dag;
	if (hasUnion)
	{//System.out.println("have union type domain");
	//System.out.println("recursive domain:"+recursiveDomName+"isRecur:"+isRecur+" and level is :"+ levels);
	  if (isRecur)
      dag = CreateDAGUnionRecursive(name, env,recursiveDomName,levels);
	  else
      dag = CreateDAGUnion(name, env);
    }
	else
	{//System.out.println("do not have union type domain");
      dag = CreateDAG(name, env);
    }
	
	  //recurLevel = 0;//from zhan
	  levels=0; //from zhan
	  level = 0;//from zhan


      for (int i = dag.size() - 1; i >= 0; i--) {
	String s = (String) dag.elementAt(i);
	//System.out.println("relation name in dag:"+s);
	Relation r = (Relation) relsHT.get(s);
//	if (s.indexOf(".")==0)
//	System.out.println(s+" data[2][0]"+((String[])r.data[2])[0]); //why null //casting error 
//	else
//System.out.println(s+" data[1][0]"+((String[])r.data[1])[0]); //why null //casting error 
	int tmp = r.numtuples;
	r.numtuples = r.capacity;
	r.capacity = tmp;
      	//System.out.println("now!! relation:"+r.name+" has "+ r.numtuples+" tuples");
//System.out.println(((int[])r.data[0])[0]+((String[])r.data[1])[0]); //why null
	r.project(null); // project to eliminate duplicate
//System.out.println(((String[])r.data[0])[2]); //why null
	//System.out.println("processing duplicate in " + s);
      }

      for (int i = dag.size() - 1; i > 0; i--) {
	String s = (String) dag.elementAt(i);
	//System.out.println("Today:relation name in dag:"+s);
	Relation r = (Relation) relsHT.get(s);
	int tmp = r.numtuples;
	r.numtuples = r.capacity;
	r.append(r.data, tmp);
	//  r.print();

	r.data = null; r.capacity = 0;
	r.numsortattrs = 0;
//	if (Global.debug) {
//	  Global.println(name);
//	  r.print();
//	}
	//	  Global.println("appending in " + s);
      }

      Relation rel = env.lookupRel(name, true);
      //rel.print();  string is empty?
      Object[] dataArray = rel.data;
      int tuples = rel.numtuples;
      int sort = rel.numsortattrs;
      rel.numtuples = 0;

      // append top level data to each relation, note the need to
      // check for duplicate tuples in initialization
      for (int i = 0; i < rels.jjtGetNumChildren(); i++) {
	name = ((SimpleNode) rels.jjtGetChild(i)).name;
	Relation r = env.lookupRel(name, true);
//System.out.println("append:"+r.name+" tuples:"+tuples);
//System.out.println(((int[])dataArray[0])[0]+((String[])dataArray[1])[0]); //why null
	r.append(dataArray, tuples);
	r.numsortattrs = sort;


	r.data = null; r.capacity = 0;
	  //r.print();
	if (Global.debug) {
	  Global.println(name);
	  r.print();
	}
      }

    }
	 //Global.syncAllToDisk();

   }//zhan's addtion corresponding to "if(isFileInput == false)" Nov 2003

  }

  private void DeclareDeclaration(SimpleNode node, Environment env)	//by Yu
   throws InterpretError
  {
    SimpleNode type = (SimpleNode) node.jjtGetChild(1);           //for example: declare x intg;
    // add the new relations to the reltable
    SimpleNode rels = (SimpleNode) node.jjtGetChild(0);
    for(int i = 0; i < rels.jjtGetNumChildren(); i++)
    {
      String relname = ((SimpleNode) rels.jjtGetChild(i)).name;       //x
	//	  int t = type.opcode-OP_BOOLEAN;
	    //System.out.println("type is "+ t);
/*	  Domain d = new Domain(relname,t,0,null);  
	  Domain [] dom = new Domain[1];
	  dom[0] = d;  */
      Relation r = new Relation(relname, null, SCALAR, 0, 0, 0, null, env);
      env.put(r);
    // perform initialization of the declare if necessary
    //need to give a default value for those without initial value	//Sep 01, 2004
      Relation rel = env.lookupRel(relname,true);

	  //rel.value = child.getInfo();
			//let x be 10;
			//construct vnode
      SimpleNode vNode = SimpleNode.jjtCreateNode("Declaration",OP_DECLARATION,OP_LET,null);
      SimpleNode ideNode = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,relname);
      SimpleNode cNode;
      if (node.jjtGetNumChildren() == 3) 
      {
       SimpleNode child = (SimpleNode) node.jjtGetChild(2);	//value, for example: declare x intg<-10;
       cNode = child.jjtDuplicate();
      }
      else	//for those not initialized
      {
       cNode = SimpleNode.jjtCreateNode("Literal",OP_LITERAL,type.opcode,null);
       //Integer inte = new Integer(0);
       //cNode.setInfo(inte);//node info setto null****?
      }
		  ideNode.jjtSetParent(vNode);
		  (vNode).jjtAddChild(ideNode);
		  cNode.jjtSetParent(vNode);
		  (vNode).jjtAddChild(cNode);

		  AddVirtDom2DomTable(vNode, env);    	
		 //System.out.println("start assignLiteral"+rel.name+" capacity:"+rel.capacity);
/*		 try{
		 //	if (rel.data==null) rel.load();
		      rel.data = rel.createDataArray(rel.domains, rel.numtuples);
		 assignLiteral(rel, 0, 0, child);
		}catch (InterpretError ee)
		{
			throw new InterpretError("Declare error:"+ ee.getMessage());
		} */
    }//for
  }

   /**
    * recursive routine called only by RelationalDeclaration()
    * to initialize a relation using the curly bracket syntax
   * START Since jRelixNS
   * Support empty nested relations.
   * Empty nested relations are all assigned with the Surrogate of value 0.
    */
  private long RelationalInitialization(SimpleNode node, Hashtable relsHT,
		String name, long surr,  int tupleNum) //modified by Yi Zheng Sept 28, 2002
	   	throws InterpretError
  {

   int type = node.type, opcode = node.opcode;

   switch(type) {
	 
   case OP_CONSTANT :
	switch(opcode) {
	 case OP_TUPLE :		
	  //System.out.println("OP_TUPLE starts with relation name:"+name);
	  Relation rel = (Relation) relsHT.get(name);
	     //System.out.println("name:"+name);
	  if ( rel == null) { // should never happen
	       throw new
		 InterpretError("Error: look up of relation " + name + " failed");
	  }
	  Domain[] doms = rel.domains;

	     // do a semantic check to see if user gave right number
	     // of entries for a tuple (careful for ".id" field)
	  int numchld = node.jjtGetNumChildren();
	  if( (!name.startsWith(".") && (doms.length != numchld)) ||
	  	(name.startsWith(".") && (doms.length != numchld+1)) )
	  {
		 throw new InterpretError("Tuple " + tupleNum + " for relation " + name
					  + " contains wrong number of attributes: " + numchld +
					  ". The expected number of attributes is: " +
					  ((!name.startsWith("."))?doms.length:(doms.length-1)));
	  }

	     // find the entries of the tuple
	     // first, see if we have to put a ".id" entry
	     // if there is a ".id" field, then other domains are offset by 1
	  int offset = 0;
	  if(name.startsWith(".")) {
	       ((long[]) rel.data[0])[rel.capacity] = surr;
	       offset = 1;
	  }

         //zhan addition
      String recurRel;
         //int level = 0;
      int tmpCapacity = 0; //zhan
      boolean addCap = false;
	  int star1,star2;
	  Domain[] recurDoms = new Domain[doms.length];
      Domain recursiveDomain;
      Hashtable newrelsHT;
      Relation originalRecurRel; //.N
      String reCurName;
         //end of zhan's addition

	     // now get all other entries of this tuple -- other than .id field //comment by Yu
      for(int i = 0; i < node.jjtGetNumChildren(); i++) {
	   SimpleNode child = (SimpleNode) node.jjtGetChild(i);
	   if(child.type == OP_LITERAL) {
		 //System.out.println("start assignLiteral"+rel.name+" capacity:"+rel.capacity);
		 try{
		 assignLiteral(rel, i, offset, child);
		 //if (i==2&&rel.name.equals("company"))
		 //System.out.println("!!!relation:"+((String[]) rel.data[2])[0] );
		}catch (InterpretError ee)
		{
			throw new InterpretError("Tuple " + tupleNum + " of relation " + rel.name +
			" has invalid value for attribute " + (i+1) + "; message: " + ee.getMessage());
		}
	   }//end of if(child.type == OP_LITERAL) 
                   /******* start adding by Fan *******/
                         else if (child.identifier.equals("TypeIDList") && child.type == OP_IDLIST) { // September 2004
                                try{
                                        //assignIdentifier(rel, i, offset, child.name);
                                assignIdentifier(rel, i, offset, child);
                                }catch (InterpretError ee)
                                {
                                        throw new InterpretError("Tuple " + tupleNum + " of relation " + rel.name +
                                        " has invalid value for attribute " + (i+1) + "; message: " + ee.getMessage());
                                }

                         }
                         /******* end of Fan's addition *******/

	   // Yu. July 07,2004 //haven't been used
	   else if (child.type == OP_UNIONCONSTANT) //can also be unionConstant***** --Yu. July 07, 2004
	   {//OP_IDENTIFIER//OP_TYPE//OP_TUPLELIST
	    SimpleNode childT = (SimpleNode) child.jjtGetChild(0); //identifier or type
	    SimpleNode childL = (SimpleNode) child.jjtGetChild(1); //literal or tuplelist
		String content;
	    switch(child.opcode){
	     case OP_IDENTIFIER :	//a:1
	     			content = childT.name+":"+(childL.getInfo()).toString();
					//SimpleNode cNode = childL.jjtDuplicate();
	       			childL.set(childL.type, OP_STRING);
	       			childL.setInfo((Object)content);
	       			assignLiteral(rel, i, offset, childL);
	       			break;
	     case OP_TYPE :		//string:1
	       			String t = Utility.type2String(Utility.typeConvert(childT.opcode));
	       			//System.out.println("opcode:"+childT.opcode+" and type:"+t);
	       			String l = (childL.getInfo()).toString();
                                        /*** Start adding by Fan ***/
                                        if(childL.opcode == OP_BOOLEAN)  // September 19, 2004
                                                if (l.equals("1")) l ="true";
                                                else if(l.equals("0")) l="false";
                                        /*** enf of Fan's addition ***/


	       			content = t+":"+l;
					//SimpleNode cNode = childL.jjtDuplicate();
	       			childL.set(childL.type, OP_STRING);
	       			childL.setInfo((Object)content);
	       			assignLiteral(rel, i, offset, childL);
	       			break;
	     case OP_TUPLELIST:	//n1:{...}
					String newrel = (childT.name.startsWith(".")) ?
					   childT.name : "." + childT.name;
		   			long result = RelationalInitialization(childL, relsHT, newrel, 0, 0);
		   			String surrogate = (new Long(result)).toString();
		   			((String[]) rel.data[i + offset])[rel.capacity] = childT.name+":"+surrogate;
	       			//System.out.println(childT.name+":"+surrogate);
	    } //switch
	   }//Yu	       	
	   else { // child is a nested domain  tuplelist
 
         //second way:
	     //System.out.println("in nesting");
	    //System.out.println("name:"+name+" in nesting"+" type of "+doms[i+offset].name+" is "+doms[i+offset].type);
		long result=0;
		String newrel = (doms[i+offset].name.startsWith(".")) ?
		   doms[i+offset].name : "." + doms[i+offset].name;

         //start of zhan's addition

		 //recursive case .N(A,B,N)->.N(A,B,N_0) or
		 //.N(A,B,N_0) : N has been declared
		if(newrel.equals(name)||newrel.equals(name+"_0"))
		{
             //recurRel= newrel+"_"+ level; //.N[0]
             recurRel= name+"_"+ level; //Nov 13
			 level = level +1;

			 if(relsHT.containsKey(recurRel)==false)
			 {
			   recurDoms = doms;
               Relation r = new Relation(recurRel, recurDoms, RELATION,
			                    0, recurDoms.length, 0, null, env);
               env.put(r);
               CreateData(relsHT, recurRel, env);
		     }

             tmpCapacity = rel.capacity;
             addCap = true;
			 rel.capacity++;
			 //System.out.println("!!!!!!capacity++");
			 result=RelationalInitialization(child,relsHT,recurRel,0,0);
	         //System.out.println("name = "+rel.name + " tmpCapactiy = "+ tmpCapacity + "  result = "+ result);
			 ((long[]) rel.data[i + offset])[tmpCapacity] = result;
		}
		 //case .N_0(A,B,N_1)
		 //create relaton .N_1(.id,A,B,N_2)
		else if(rel.name.indexOf('_')!=-1)
        {
            star1 = rel.name.indexOf('_');

            star2 = newrel.indexOf('_');
            if(star2!=-1)
                   reCurName = newrel.substring(0,star2);
            else
                   reCurName = newrel;
            if(rel.name.substring(0,star1).equals(reCurName)) //Nov14
            {
			   recurRel= rel.name.substring(0,star1)+"_"+ level; //.N_2
			   level = level+1;
               //System.out.println("level = " + level + " relation r");
               if(relsHT.containsKey(recurRel)==false)
               {
				   originalRecurRel = (Relation)relsHT.get(name.substring(0,star1));

                   recurDoms = originalRecurRel.domains;

			       Relation r = new Relation(recurRel, recurDoms, RELATION,
				                0, recurDoms.length, 0, null, env);
			       env.put(r);

                   CreateData(relsHT, recurRel, env);
			   }

               tmpCapacity = rel.capacity;
			   addCap = true;
			   rel.capacity++;
			 //System.out.println("!!!!!!capacity++");

			   result=RelationalInitialization(child,relsHT,recurRel,0,0);
	           //System.out.println("name = "+rel.name + " tmpCapactiy = "+ tmpCapacity + "  result = "+ result);
			   ((long[]) rel.data[i + offset])[tmpCapacity] = result;
			}
		}
		else //normal nested
        {
         	
            //end of zhan's addition
		/*	if (newrel is union type) 	//July 12,2004 by Yu******
			{extract type from tree;
			 try find match from the definition of this union type with this one;
			 if (no match) throw exception;
			 else
			 {*/
		 if (doms[i+offset].type==UNION)
		 {//System.out.println("hello!!!!!!UNION:"+doms[i+offset].name);
		  boolean hasMatch = false;	 
		  for (int j = 0; j<doms[i+offset].union.length; j++)  //check for each possible type 
		  {												//if it is IDLIST and if it matchs the type from tree
		//		if (doms[i].union[j].opcode==OP_IDLIST||doms[i].union[j].opcode==OP_IDENTIFIER)
		   if (doms[i+offset].union[j].opcode==OP_IDENTIFIER)	//idlist or union or other domain
		   {	
			String cnm = doms[i+offset].union[j].name;
			Domain d = env.lookupDom(cnm, true);
			//System.out.println("here we go!!!!"+cnm+" and its type is:"+d.type);

			if (d.type==IDLIST)
			{	
					String nm = "."+cnm;
					//get its definition then check if it matchs -> type[]<-from tree
					//Domain d = env.lookupDom(nm, true);
					//if (d.type==UNION)****** domain c3 c2|b; domain c2 n1|b
					Relation r = env.lookupRel(nm,true);
			//		System.out.println("@@@here! relation is"+nm);
					if (r!=null)
					{
					  boolean match = match(r,child);
					  if (match)
					  {
					  		hasMatch = true; 
							newrel = (r.name.startsWith(".")) ?
					   		r.name : "." + r.name;
		if(newrel.equals(name)||newrel.equals(name+"_0"))
		{
             //recurRel= newrel+"_"+ level; //.N[0]
             recurRel= name+"_"+ level; //Nov 13
			 level = level +1;

			 if(relsHT.containsKey(recurRel)==false)
			 {
			   recurDoms = doms;
               r = new Relation(recurRel, recurDoms, RELATION,
			                    0, recurDoms.length, 0, null, env);
               env.put(r);
               CreateData(relsHT, recurRel, env);
		     }

             //tmpCapacity = rel.capacity;
             //addCap = true;
			 //rel.capacity++;
			 result=RelationalInitialization(child,relsHT,recurRel,0,0);
	         //System.out.println("name = "+rel.name + " tmpCapactiy = "+ tmpCapacity + "  result = "+ result);
			 //((long[]) rel.data[i + offset])[tmpCapacity] = result;
		}
		 //case .N_0(A,B,N_1)
		 //create relaton .N_1(.id,A,B,N_2)
		else if(rel.name.indexOf('_')!=-1)
        {
            star1 = rel.name.indexOf('_');

            star2 = newrel.indexOf('_');
            if(star2!=-1)
                   reCurName = newrel.substring(0,star2);
            else
                   reCurName = newrel;
            if(rel.name.substring(0,star1).equals(reCurName)) //Nov14
            {
			   recurRel= rel.name.substring(0,star1)+"_"+ level; //.N_2
			   level = level+1;
               //System.out.println("level = " + level + " relation r");
               if(relsHT.containsKey(recurRel)==false)
               {
				   originalRecurRel = (Relation)relsHT.get(name.substring(0,star1));

                   recurDoms = originalRecurRel.domains;

			       r = new Relation(recurRel, recurDoms, RELATION,
				                0, recurDoms.length, 0, null, env);
			       env.put(r);

                   CreateData(relsHT, recurRel, env);
			   }

               //tmpCapacity = rel.capacity;
			   //addCap = true;
			   //rel.capacity++;

			   result=RelationalInitialization(child,relsHT,recurRel,0,0);
	           //System.out.println("name = "+rel.name + " tmpCapactiy = "+ tmpCapacity + "  result = "+ result);
			   //((long[]) rel.data[i + offset])[tmpCapacity] = result;
			}
		}
		else //normal nested
        {



					   		//System.out.println("@@@union type relation name:"+newrel);
				   			result = RelationalInitialization(child, relsHT, newrel, 0, 0);
					  }
               tmpCapacity = rel.capacity;
			   //addCap = true;
			   //rel.capacity++;
//System.out.println("@@@@@@"+rel.name);
					   		String idname = (r.name.startsWith(".")) ?
					   		(r.name).substring(1) : r.name;
				   			String surrogate = (new Long(result)).toString();
				   			//((String[]) rel.data[i + offset])[rel.capacity] = r.name+":"+surrogate;
							((String[]) rel.data[i + offset])[tmpCapacity] = idname+":"+surrogate;
				   			break;
					  
					 }
				    }
					else System.out.println("relation is null "+nm);
//	   				System.out.println("union types are:"+doms[i].union[j].name+" and opcode:"+doms[i].union[j].opcode);
//					relnames.addElement("." + doms[i].union[j].name);
			}	//		 if (d.type==IDLIST)
			else if (d.type==UNION)//****** domain c3 c2|b; domain c2 n1|b
			{
       		 Vector domnames = new Vector(); //stack of domain names to process
	   		 domnames.addElement(doms[i].name);
	   		 while (domnames.size()>0)//hasMoreUnion
	   		 {	
        	  String dnm = (String) domnames.lastElement();
         	  domnames.removeElementAt(domnames.size() - 1);
 	    	  Domain dc = env.lookupDom(dnm, true);
 	    	  if (dc==null) System.out.println("domain is null,"+dnm);
	    	  //System.out.println("union type:"+dc.name);
	    	  for (int j1 = 0; j1<dc.union.length; j1++) //check for each possible type if it is IDLIST
	    	  {
//		 if (doms[i].union[j].opcode==OP_IDLIST||doms[i].union[j].opcode==OP_IDENTIFIER)
		 	   if (dc.union[j1].opcode==OP_IDENTIFIER) //domain c2 n1|b; domain c3 c2|b;

		 	   {
		 		String nameD = dc.union[j1].name;
		 		//System.out.println("union:"+nameD);
				if ((env.lookupDom(nameD,true)).type==IDLIST)
				{
	   		 	 //System.out.println("union types with idlist:"+dc.union[j1].name);
			 	 //relnames.addElement("." + dc.union[j].name);
				 String nm = "."+nameD;
					//get its definition then check if it matchs -> type[]<-from tree
					//Domain d = env.lookupDom(nm, true);
					//if (d.type==UNION)****** domain c3 c2|b; domain c2 n1|b
				 Relation r = env.lookupRel(nm,true);
				 //System.out.println("here!"+nm);
				 if (r!=null)
				 {
						boolean match = match(r,child);
						/*for (int k = 1; k<r.domains.length; k++)
						{
							SimpleNode childIdlist = (SimpleNode) (child.jjtGetChild(0)).jjtGetChild(k-1); 
							System.out.println("domain type:"+r.domains[k].type+" and node type is :"+childIdlist.opcode);
							if (r.domains[k].type!= Utility.typeConvert(childIdlist.opcode))
							{
								System.out.println("not match");
								match = false;
							}
					 }*/
					  if (match)
					  {
					  		hasMatch = true; 
							newrel = (r.name.startsWith(".")) ?

					   		r.name : "." + r.name;
					   		String idname = (r.name.startsWith(".")) ?
					   		(r.name).substring(1) : r.name;
					   		//System.out.println("union type relation name:"+newrel);
				   			result = RelationalInitialization(child, relsHT, newrel, 0, 0);
				   			String surrogate = (new Long(result)).toString();
				   			//((String[]) rel.data[i + offset])[rel.capacity] = r.name+":"+surrogate;
							((String[]) rel.data[i + offset])[rel.capacity] = dc.name+":"+idname+":"+surrogate;
						   			break;
					  }
				 }
				 else System.out.println("relation is null "+nm);
			 //hasMoreUnion=false;//--no!!! idlist can have union domain! deal with it at the following
			    }
				else if ((env.lookupDom(nameD,true)).type==UNION)
				{//System.out.println(nameD+" is union type of union domain :"+dc.name);
					domnames.addElement(nameD);
				}
		 	   }
	    	  }	//for
	   		 } //while (domnames.size()>0)(hasMoreUnion)
			}//else if (d.type==UNION)
		   }//if (doms[i+offset].union[j].opcode==OP_IDENTIFIER)
	   	  }//for
	   	  if (!hasMatch)   
	   	   throw new InterpretError("the input has no match with the union type ");
	 	 }// if (doms[i+offset].type==UNION)

	 	 else

		 {//System.out.println("call RelationalInitialization again with name: "+newrel);
				result = RelationalInitialization(child, relsHT, newrel, 0, 0);
		   		((long[]) rel.data[i + offset])[rel.capacity] = result;
		 }
			//}

	    }//end else //zhan

	   }//end of 	       else { // child is a nested domain  

	 } //end of      for(int i = 0; i < node.jjtGetNumChildren(); i++) {
//System.out.println("here capacity is:"+rel.capacity+((String[])rel.data[1])[0]);

     //zhan's addition
     if(levels < level)
       levels = level;
     if(addCap == false)
        rel.capacity++;

     level = 0;
     //zhan's addition ends

	 //rel.capacity++;//original //modified by zhan
//System.out.println("here capacity is:"+rel.capacity+((String[])rel.data[1])[0]);
	     return 0;

	   case OP_TUPLELIST :
	     //System.out.println("OP_TUPLELIST starts");
	     // get a surrogate number if we are a nested domain
	     long s = 0;
//	     if (name.startsWith("."))
//	       s = Surrogate.next(); //get a surrogate
                 // START Since jRelixNS
                 // Support empty nested relations.
                 // Let surrogate remain 0 so they are always smaller
                 // than any other relation.
         if (Relation.isSigmaJoinImplemented == true) {
		 if (name.startsWith("."))
                     if (node.jjtGetNumChildren() > 0)
		       s = Surrogate.next(); //get a surrogate
                 }
         else
		   if (name.startsWith("."))
		     s = Surrogate.next(); //get a surrogate
                 // END Since jRelixNS //Andy, May 12, 2002

	     //System.out.println("num of children="+node.jjtGetNumChildren());
	     for(int i=0; i < node.jjtGetNumChildren(); i++) {	//num of tuples
	       SimpleNode child = (SimpleNode) node.jjtGetChild(i);
//System.out.println(i+":call RelationalInitialization again with name: "+name+" and s:"+s);
	       RelationalInitialization(child, relsHT, name, s, i + 1);
	     }
//System.out.println("here capacity is:"+rel.capacity+((String[])rel.data[1])[0]);
	     return s;
	   } // end of second switch: switch(opcode)

	   //YuLing added
	 case OP_FILEPATH:
	   String inputfilename=node.name.substring(1,node.name.length()-1);
	   //System.out.println("inputfilename="+inputfilename);
	   FileInputStream fin=null;
	   Relation rel = null;
	   try{
	     rel = (Relation) relsHT.get(name);

	     fin=new FileInputStream(inputfilename);
	     DataInputStream dis=new DataInputStream(fin);

	     char[] ch={06}; //separate character of relation file
	     String s=String.valueOf(ch);

	     String str=null;

	     //create the tuplelist node
	     SimpleNode tplistnd=new SimpleNode("Tuplelist");
	     tplistnd.set(OP_CONSTANT,OP_TUPLELIST,null);
	     tplistnd.jjtSetParent(node);
	     node.jjtAddChild(tplistnd);

	     Domain[] doms = rel.domains;
	     int childnum=0;
	     int i = -1;
	     while((str=dis.readLine())!=null)//read the lines of the relation file
	       {
	       childnum++;
	       try {
		 SimpleNode tpnd=new SimpleNode("Tuple");
		 tpnd.set(OP_CONSTANT,OP_TUPLE,null);
		 tpnd.jjtSetParent(tplistnd);
		 tplistnd.jjtAddChild(tpnd);

	//	 StringTokenizer st= new StringTokenizer(str,s);
	//modified by Yi Zheng as a bug fix: to allow empty string

	//Sept 22, 2002

	     StringTokenizer st = new StringTokenizer(str,s,true);
	     boolean prevIsTok = true;


	 //end of bug fix Sept 22, 2002

		 i=-1;
		 while (st.hasMoreTokens())
		   {
		     String stringval=null;
		     SimpleNode literalnd=new SimpleNode("Literal");
		     String val=st.nextToken();

		     //bug fix Sept 22, 2002
		     if (DEBUG_09_22_2002)
		     	System.out.println("val = " + val);

		     if (val.compareTo(s) == 0)
		     {
		     	if (!prevIsTok)

		     	{
		     		prevIsTok = true;
		     		continue;
		     	}
		     	else //empty string encountered
		     	{
		     		val = "";
		     		i++;
		     		prevIsTok = true;
		     	}
		     }
		     else
		     {
		     	prevIsTok = false;
		     	i++;
		     }
		     //end of bug fix Sept 22, 2002

   		     switch(doms[i].type)
		       {
		       case INTEGER:
			 if(val.charAt(0)=='-')
			   stringval="-"+val.substring(1);
			 else
			   stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			 literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {
			     literalnd.set(OP_LITERAL,OP_INTEGER,null);
			     literalnd.setObject(new Integer(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case  BOOLEAN:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {literalnd.set(OP_LITERAL,OP_BOOLEAN,null);
			   if(stringval.compareTo("1")==0)
			     literalnd.setObject(new Byte((byte)1));
			   else
			     literalnd.setObject(new Byte((byte)0));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case SHORT:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {literalnd.set(OP_LITERAL,OP_SHORT,null);
			   literalnd.setObject(new Short(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case LONG:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {literalnd.set(OP_LITERAL,OP_LONG,null);
			   literalnd.setObject(new Long(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
			  case NUMERIC://from Yiyi
          stringval = val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {literalnd.set(OP_LITERAL,OP_NUMERIC,null);
			   literalnd.setObject(new number(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;

		       case FLOAT:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {literalnd.set(OP_LITERAL,OP_FLOAT,null);
			   literalnd.setObject(new Float(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case DOUBLE:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   { literalnd.set(OP_LITERAL,OP_DOUBLE,null);
			   literalnd.setObject(new Double(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
			 case UNIVERSAL: //from Andrey
			case ATTRIBUTE: //from Andrey
			case TYPE:		//from Yu	July 04, 2004
			case UNION: //from Yu	July 04, 2004
		       case STRING:
		       case TEXT:

		       //stringval = val.substring(0);
		       //bug fix Sept 22, 2002
//System.out.println("here:");
		       if (val.length() > 0)
			 		stringval=val.substring(0);
			 	else
			 		stringval = new String("");
			 		//end of bug fix Sept 22, 2002

			 //debugging Sept 22, 2002:
			 if (DEBUG_09_22_2002)
			 	System.out.println("stringval = " + stringval);


	//		 if(((int)stringval.charAt(0))==05)//dc
	//		   literalnd.set(OP_LITERAL,OP_DC,null);
	//		 else if(((int)stringval.charAt(0))==02)

	//		   literalnd.set(OP_LITERAL,OP_DK,null);
	//bug fix Sept 22, 2002
			   if((stringval.length() > 0) && (((int)stringval.charAt(0))==05))//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if((stringval.length() > 0) && (((int)stringval.charAt(0))==02))
			   literalnd.set(OP_LITERAL,OP_DK,null);
	//end of bug fix Sept 22, 2002
			 else
			   {literalnd.set(OP_LITERAL,OP_STRING,null);
			   literalnd.setObject(new String(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
		        tpnd.jjtAddChild(literalnd);
			break;
		       default:
			 System.out.println("unkown domain type:"+doms[i].type);

			 throw new InterpretError("Error: Unexpected domain type");
		       }//end of switch
		   //  i++; //bug fix Sept 22, 2002

		     //debuggin Sept 22, 2002
		     if (DEBUG_09_22_2002)
		     	System.out.println("tokens = " + i);
		   }
		}
		   catch (ArrayIndexOutOfBoundsException aie)
		   {
		   	 throw new InterpretError("Tuple " + (childnum + 1) + " contains too many attributes;" +
		   	  " offending attribute position: " + (i + 1) + ". Expected number of attributes: " + doms.length);
		   }
	       }

	     RelationalInitialization(tplistnd,relsHT,name,surr,0);
	   }
	   catch(IOException e){System.out.println("IOException in RelationalInitialization");}
	   //added by Yi Zheng Oct 2, 2002
	   if (rel != null)
	   	{
	   		env.put(rel);
	   		rel.save();
	   	}
	   //end of addition Oct 2, 2002

	   return 0;
	   //end of YuLing added
	 case OP_IDENTIFIER://added by Yi Zheng March 23, 2002
	 throw new InterpretError("Initialization by Identifier not implemented yet.");
	 default :
	   //System.out.println("the type is"+type);
	   throw new InterpretError("Error: Unexpected SimpleNode");
	 }//end of 	 switch(type) {

  }
//July 28th,2004
private boolean match(Relation r, SimpleNode child)
       throws InterpretError
{
  if (child.jjtGetNumChildren()==0) 
  {System.out.println("no tuple at all"); 
   return true;
  }
  for (int i = 0; i< child.jjtGetNumChildren(); i++)
  {
	SimpleNode tuple = (SimpleNode)(child.jjtGetChild(i));
	for (int k = 1; k<r.domains.length; k++)
	{
	 SimpleNode domN = (SimpleNode)(tuple.jjtGetChild(k-1)); 
	 //System.out.println("domain type:"+r.domains[k].type+" and node type is :"+domN.opcode);
	 if (domN.type==OP_LITERAL)
	 {
	 	if (r.domains[k].type!=UNION)
	 	{
	 		if (r.domains[k].type!=Utility.typeConvert(domN.opcode))
	 		return false;
	 	}
	 	else
	 	{
		  boolean noMatch = true;
		  String nm = r.domains[k].name;
		  Vector domnames = new Vector(); //stack of domain names to process
	   	  domnames.addElement(nm);
	   	  while (domnames.size()>0)//hasMoreUnion
	   	  {	
        	String dnm = (String) domnames.lastElement();
         	domnames.removeElementAt(domnames.size() - 1);
 	    	Domain dc = env.lookupDom(dnm, true);
 	    	if (dc==null) System.out.println("domain is null,"+dnm);
	    	  //System.out.println("union type:"+dc.name);
	    	for (int j = 0; j<dc.union.length; j++) //check for each possible type if it is IDLIST
	    	{
	    		if (dc.union[j].type!=UNION)
	    		{
			 		if (dc.union[j].type==Utility.typeConvert(domN.opcode))
	 				noMatch = false;
	 			}
	 			else
	 				domnames.addElement(dc.union[j].name);
	    	}
			if (!noMatch) break;
		  }
	 	}
	 }
	 else //tupleList
	 {
	  if (r.domains[k].type!=IDLIST)
	  {
	 	if (r.domains[k].type!=UNION)
	 		return false;
	 	else //check each union type to find idlist
	 	{
		  boolean noMatch = true;
		  String nm = r.domains[k].name;
		  Vector domnames = new Vector(); //stack of domain names to process
	   	  domnames.addElement(nm);
	   	  while (domnames.size()>0)//hasMoreUnion
	   	  {	
        	String dnm = (String) domnames.lastElement();
         	domnames.removeElementAt(domnames.size() - 1);
 	    	Domain dc = env.lookupDom(dnm, true);
 	    	if (dc==null) System.out.println("domain is null,"+dnm);
	    	  //System.out.println("union type:"+dc.name);
	    	for (int j = 0; j<dc.union.length; j++) //check for each possible type if it is IDLIST
	    	{
	    		if (dc.union[j].type!=UNION)
	    		{
			 		if (dc.union[j].type==IDLIST)
			 		{
			 			String nmd = "."+dc.union[j].name;
			 			//Domain domI = env.lookupDom(nmd, true);
			 			Relation R = env.lookupRel(nmd,true);
			 			boolean m = match(R,domN);
			 			if (m) noMatch = false;
			 		}
			 		
	 				//noMatch = false;
	 			}
	 			else
	 				domnames.addElement(dc.union[j].name);
	    	}
			if (!noMatch) break;
		  }//while	 			
	 	}//else
	  }//if
	 }//else
	 
	}
  }
  return true;
}

//starting of zhan's addition Nov 2003
private void inputFromFile(SimpleNode node,String s)
{
   //initialization from file like this {(...)}
         //       Declaration
         //      /      |     \
         //    IDList  IDList Init
         //     /       |      /
         //  Identifier     FILEPATH

      SimpleNode relName = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	  SimpleNode idlistNode = (SimpleNode)node.jjtGetChild(1);

	  String domList = "(";
      for(int i=0;i<idlistNode.jjtGetNumChildren();i++)
      {
		domList = domList +((SimpleNode)idlistNode.jjtGetChild(i)).name+",";
	  }

	  domList = domList.substring(0,domList.length()-1) + ")";

	  String input="relation "+relName.name+ domList +"<-"+s+";";
      //System.out.println("the new input is ==="+input);

	  byte[] barray=input.getBytes();
	  ByteArrayInputStream binstream=new ByteArrayInputStream(barray);
	  InputStream bin=(InputStream)(binstream);
	  Global.jin.resetIn(bin);
	  parser.ReInit(Global.jin);
	  instack.push(in);
	  in = bin;

}


private SimpleNode XMLInitialization(SimpleNode node,Environment env)
         throws InterpretError
{

  SimpleNode tupList = (SimpleNode)node.jjtGetChild(1).jjtGetChild(0);
  String topRel = tupList.name;

  //             Declaration
  //            /   [|]     \
  //       IDList  [IDList] Init
  //         /                |
  //        ID               TupleList (tupList)
  //                        / ... ....\
  //                      Tuple        Tuple
  //                      /.. \         /.. \
  //                  Literal TupleList
  //                             |
  //                           Tuple
  //                           /|...\
  //                         Tuple  TupleList
  //                           ....  ....

  Vector relDom = new Vector();

  //find all domains needed to be declared
  relDom = findAllDomainsInfo(tupList,relDom,env);
//System.out.println("after findallDomainInfo");
/*
  for(int i=0;i<relDom.size();i++)
  {
	 XmlRelInfo xml = (XmlRelInfo)relDom.elementAt(i);
	 System.out.print("name= "+xml.name+" ");
	 System.out.println("isDefined =" + xml.isDefined);
	 for(int j=0;j<xml.doms.size();j++)
	 {
		 XmlDomInfo xmlDom = (XmlDomInfo)xml.doms.elementAt(j);
		 System.out.print(" domName = "+ xmlDom.name + " type = "+ xmlDom.type+" ");
         System.out.println("isDefined="+xmlDom.isDefined);
      }
  }
*/

  //declare the domains
  //String domDeclare = "";


  int opcode;
  for(int i=0;i<relDom.size();i++)
  {
	XmlRelInfo xmlRel = (XmlRelInfo)relDom.elementAt(i);

	//declare non-nested domains
	for(int j=0;j<xmlRel.doms.size();j++)
	{
	  XmlDomInfo xmlDom = (XmlDomInfo)xmlRel.doms.elementAt(j);

	  if(!xmlDom.type.equals("IDList") && xmlDom.isDefined==false)
      {

	   //create the tree consponding to the domain declaration
	   SimpleNode Declaration = SimpleNode.jjtCreateNode("Declaration",140,143,null);
       SimpleNode IDList = SimpleNode.jjtCreateNode("IDList",463,463,null);
       SimpleNode IdNode = SimpleNode.
	            jjtCreateNode("Identifier",230,230,xmlDom.name);

	   IdNode.jjtSetParent(IDList);
	   IDList.jjtAddChild(IdNode);

	   IDList.jjtSetParent(Declaration);
	   Declaration.jjtAddChild(IDList);

       opcode = findOpCode(xmlDom.type);
	   SimpleNode typeNode = SimpleNode.jjtCreateNode("Type",OP_TYPE,opcode,null);
	   typeNode.jjtSetParent(Declaration);
	   Declaration.jjtAddChild(typeNode);

       //declare the domain
	   executeDeclaration(Declaration,env);
      }//end if type !=idlist
	}
  }//end for


  //declare nested domain
  for(int i=relDom.size()-1;i>0; i--)
  {
	  XmlRelInfo xmlRel = (XmlRelInfo)relDom.elementAt(i);

	  if(xmlRel.isDefined==false)
      {
        SimpleNode Declaration = SimpleNode.jjtCreateNode("Declaration",140,143,null);
	    SimpleNode IDList = SimpleNode.jjtCreateNode("IDList",463,463,null);
        SimpleNode IdNode = SimpleNode.jjtCreateNode("Identifier",230,230,xmlRel.name);
        IdNode.jjtSetParent(IDList);
	    IDList.jjtAddChild(IdNode);

	    IDList.jjtSetParent(Declaration);
	    Declaration.jjtAddChild(IDList);

        SimpleNode typeNode = SimpleNode.jjtCreateNode("Type",OP_TYPE,463,null);
	    SimpleNode IDListNode = SimpleNode.jjtCreateNode("IDList",463,463,null);

	    for(int j=0;j<xmlRel.doms.size();j++)
	    {
		  XmlDomInfo xmlDom = (XmlDomInfo)xmlRel.doms.elementAt(j);
		  //domDeclare = domDeclare + xmlDom.name + ",";
		  SimpleNode Identifier = SimpleNode.jjtCreateNode("Identifier",230,230,xmlDom.name);
	      Identifier.jjtSetParent(IDListNode);
	      IDListNode.jjtAddChild(Identifier);
	    }

        IDListNode.jjtSetParent(typeNode);
        typeNode.jjtAddChild(IDListNode);

        typeNode.jjtSetParent(Declaration);
        Declaration.jjtAddChild(typeNode);

        //System.out.println(" nested domain to be returned =============== ");
		//Global global = new Global();
        //global.traverse(Declaration);

        executeDeclaration(Declaration,env);
	}//end if

   }//end of declare nested domains




 //modify the tree
 tupList = createXmlTree(tupList,relDom);

 //modify the IDlist that contain all the top-level domains
 ((SimpleNode)node.jjtGetChild(1)).jjtRemoveChild(0);

  //duplicate Init node for further use
  SimpleNode InitNode =(SimpleNode)((SimpleNode)node.jjtGetChild(1))
                       .jjtDuplicate();

  node.jjtRemoveChild(1);
  SimpleNode IDListNode = SimpleNode.jjtCreateNode("IDList",463,463,null);

  XmlRelInfo xmlRelTop = (XmlRelInfo)relDom.elementAt(0);
  for(int i=0;i<xmlRelTop.doms.size();i++)
  {
	XmlDomInfo xmlDom = (XmlDomInfo)xmlRelTop.doms.elementAt(i);

	//create new "Identifier" node
	SimpleNode idNode = SimpleNode.jjtCreateNode("Identifier",230,230,xmlDom.name);

    idNode.jjtSetParent(IDListNode);
    IDListNode.jjtAddChild(idNode);

  }
  IDListNode.jjtSetParent(node);
  node.jjtAddChild(IDListNode);

  InitNode.jjtSetParent(node);
  node.jjtAddChild(InitNode);

 //connect the new "TupleList" node with "Init" node
  tupList.jjtSetParent((SimpleNode)node.jjtGetChild(2));
  ((SimpleNode)node.jjtGetChild(2)).jjtAddChild(tupList);

 //change the type of declaration node OP_XML to 140: OP_DECLARATION
 node.type = 140;

 //System.out.println("node to be returned =============== ");
 //Global global = new Global();
 //global.traverse(node);

 return node;

}

private int findOpCode(String type)
    throws InterpretError
{
   int opcode = 0;
   if((type.equals("strg")) | type.equals("String")
       | type.equals("string"))
      opcode = OP_STRING;
   else if(type.equals("univ") |type.equals("universal"))
      opcode = OP_UNIVERSAL;
   else if(type.equals("attr") |type.equals("attribute"))
      opcode = OP_ATTRIBUTE;
   else if(type.equals("bool") | type.equals("boolean"))
      opcode = OP_BOOLEAN;
   else if(type.equals("short"))
      opcode = OP_SHORT;
   else if(type.equals("intg") | type.equals("integer"))
      opcode = OP_INTEGER;
   else if(type.equals("long"))
      opcode = OP_LONG;
   else if(type.equals("float") | type.equals("real"))
      opcode = OP_FLOAT;
   else if(type.equals("double"))
      opcode = OP_DOUBLE;
   else if(type.equals("number"))
      opcode = OP_NUMERIC;
   else if(type.equals("text"))
      opcode = OP_TEXT;
   else if(type.equals("stmt") | type.equals("statement"))
      opcode = OP_STATEMENT;
   else if(type.equals("expr") | type.equals("expression"))
      opcode = OP_EXPRESSION;
   //miss OP_COMP
   else
      throw new InterpretError("Error: undefied domain type");
   return opcode;

}



//the recursive function is used to search all domains needed to be declared
private Vector findAllDomainsInfo(SimpleNode node,Vector relDom,Environment env)
       throws InterpretError
{
  boolean hasRel = false;
  int pos=0;
  int pos1,pos2=0;

  //first search whether the dom is already in the vector reldom
  for(int i=0; i<relDom.size();i++)
  {
	XmlRelInfo xmlRel = (XmlRelInfo)relDom.elementAt(i);

	if(xmlRel.name.equals(node.name)) //already defined
	{
	  hasRel = true;
	  pos = i;  //keep the position for further use
	  break;
	}

  }

  //the rel or nested rel haven't been defined
  if(hasRel==false)
  {
	Vector domList = new Vector();
	//check whether the nested doms have been defined
	Domain d = env.lookupDom(node.name.trim(),true);
	boolean isDefined = false;
    if(d!=null && d.numref!=0)      //the domain is not used before
      isDefined = true;

    XmlRelInfo xmlRel = new XmlRelInfo(node.name,domList,isDefined);
	relDom.addElement(xmlRel);
    pos = relDom.size()-1;

  }

  String name,type;
  boolean hasTheDom=false;

  for(int i=0; i<node.jjtGetNumChildren();i++)
  {
	 SimpleNode tuple = (SimpleNode)node.jjtGetChild(i);

     //2-level down to the current level to check if there is nested rel
	 for(int j=0; j<tuple.jjtGetNumChildren();j++)
	 {
		SimpleNode domNode = (SimpleNode)tuple.jjtGetChild(j);
		if(domNode.opcode == OP_TUPLELIST) //nested domain
        {
	      hasTheDom=false;
		  for(int n=0;n<((XmlRelInfo)relDom.elementAt(pos)).doms.size();n++)
		  {
		     XmlDomInfo xmlDom = (XmlDomInfo)((XmlRelInfo)relDom.elementAt(pos))
		                          .doms.elementAt(n);
		     //check if the dom is aleady in the domlist
		     if(xmlDom.name.equals(domNode.name))
		     {
			   hasTheDom=true;
			   break;
			 }

		  }

          //if it is not in the domlist, add it

          if(hasTheDom==false)
          {
			XmlDomInfo xmlDomInfo = new XmlDomInfo(domNode.name,"IDList",false);
		    ((XmlRelInfo)relDom.elementAt(pos)).doms.addElement(xmlDomInfo);
	      }

	      //recursive call
          relDom = findAllDomainsInfo(domNode,relDom,env);
	    }
		else //the dom is not nested dom
 		{

		 //there is a type defined
	     if(domNode.name.indexOf("type")!=-1)
         {
			pos1 = domNode.name.indexOf("type");
            name = domNode.name.substring(0,pos1).trim();
            pos2 = domNode.name.indexOf("=");
            type = domNode.name.substring(pos2+1,domNode.name.length()).trim();

			//check whether the doms have been defined
			Domain d = env.lookupDom(name.trim(),true);
			boolean isDefined = false;

			if(d!=null && d.numref!=0) //the domain is used before
			  isDefined = true;

            //check whether the type is compatible with it defined before
            if(d!=null && d.numref!=0)
            { if(d.type != intType(type))

                throw new InterpretError("Error: wrong domain type of "+ name);
		    }
            XmlDomInfo xmlDomInfo = new XmlDomInfo(name,type,isDefined);
            ((XmlRelInfo)relDom.elementAt(pos)).doms.addElement(xmlDomInfo);
          }
	      else //if(domNode.name.indexOf("type")=-1)
	      {

			hasTheDom = false;
			 for(int n=0;n<((XmlRelInfo)relDom.elementAt(pos)).doms.size();n++)
			 {
			     XmlDomInfo xmlDom = (XmlDomInfo)((XmlRelInfo)relDom.elementAt(pos))
			                          .doms.elementAt(n);

			     //already defined
			     if(xmlDom.name.equals(domNode.name))
			     {

			       hasTheDom=true;
			 	   break;
			 	 }
			 }

			 //in its first occurrence, but it is not defined(doesn't have "type =")
             if(hasTheDom == false)
             {
			    //check whether it's already defined in other definition
                Domain d = env.lookupDom(domNode.name.trim(),true); //Feb,2004

                boolean isDefined = false;

                if(d == null) //the domain is not defined before, Feb,2004
                {
			      //default is strg;
                  isDefined = false;
			    }
			    else
			    {
                  if(d.numref==0) isDefined = false;
                  else isDefined = true;
			   }

			      //default is strg;
			  XmlDomInfo xmlDomInfo = new XmlDomInfo(domNode.name.trim(),"strg",isDefined);
			  ((XmlRelInfo)relDom.elementAt(pos)).doms.addElement(xmlDomInfo);

    	 	 }
		  }

		}

	 }//end for

  }//end of for

  return relDom;
}

private String typeTransfer(int t)
{
  if(t==0) return "boolean";
  else if(t==1) return "short";
  else if(t==2) return "intg";
  else if(t==3) return "long";
  else if(t==4) return "float";
  else if(t==5) return "double";
  else if(t==6) return "numeric";
  else if(t==7) return "strg";
  else if(t==8) return "text";
  else if(t==9) return "stmt";
  else if(t==10) return "expr";
  else if(t==11) return "comp";
  else if(t==12) return "idlist";
  else if(t==18) return "attribute";
  else if(t==25) return "universal";
  else return null;

}

private int intType(String s)
{
  if(s.equals("bool")||s.equals("boolean"))
    return 0;
  else if(s.equals("short"))
    return 1;
  else if(s.equals("intg")||s.equals("integer"))
    return 2;
  else if(s.equals("long"))
    return 3;
  else if(s.equals("float")||s.equals("real"))
    return 4;
  else if(s.equals("double"))
    return 5;
  else if(s.equals("numeric"))
    return 6;
  else if(s.equals("strg")||s.equals("string"))
    return 7;
  else if(s.equals("text"))
    return 8;
  else if(s.equals("stmt")||s.equals("statement"))
    return 9;
  else if(s.equals("expr")||s.equals("expression"))
    return 10;
  else if(s.equals("comp")||s.equals("computation"))
    return 11;
  else if(s.equals("idlist"))
    return 12;
  else if(s.equals("attr")||s.equals("attribute"))
    return 18;
  else if(s.equals("univ")||s.equals("universal"))
    return 25;
  else return -1;

}



private SimpleNode createXmlTree(SimpleNode tupleList,Vector relDom)
             throws InterpretError
{
  //duplicate the tree
  SimpleNode tupList = tupleList.jjtDeepDuplicate();

  //first find the correct DomInfo for the rel with tupList.name
  int pos=0;
  for(int i=0; i<relDom.size();i++)
  {
   	XmlRelInfo xmlRel = (XmlRelInfo)relDom.elementAt(i);

   	if(xmlRel.name.equals(tupList.name))
   	{
   	  pos = i;
   	  break;
   	}
  }

 String domName = null;
 int pos1=0;

 //create new TupleList Node
 SimpleNode newTupleList = SimpleNode.jjtCreateNode("TupleList",210,211,null);


 for(int i=0; i<tupList.jjtGetNumChildren();i++)
  {
   //get the tuple node under tupList
   SimpleNode tuple = (SimpleNode)tupList.jjtGetChild(i);

   int order=0;  //correct order counter
   int j=0;
   int count=0;

   SimpleNode newTupleNode=null;
   while(order<((XmlRelInfo)relDom.elementAt(pos)).doms.size())
   {
    if(count>1024)
      throw new InterpretError("Error: wrong input.");

    count = count+1;

    //create a new "Tuple" node
    if(order==0)
    {
	  newTupleNode = SimpleNode.jjtCreateNode("Tuple",210,212,null);
	  //connect the new tuple node to TupleList
	  newTupleNode.jjtSetParent(newTupleList);
	  newTupleList.jjtAddChild(newTupleNode);
    }


    //get the doms of the nest rel in correct order
    XmlDomInfo xmlDom =(XmlDomInfo)((XmlRelInfo)relDom.elementAt(pos)).doms.elementAt(order);

    //get the current doms under the "Tuple" node
    SimpleNode domNode = (SimpleNode)tuple.jjtGetChild(j);

    //domNode.name = kk type=intg; or kk
    if(domNode.name.indexOf("type")!=-1)
    {
	  pos1 = domNode.name.indexOf("type");
	  domName = domNode.name.substring(0,pos1).trim();
    }
    else
      domName = domNode.name.trim();

    //get a match
    if(domName.equals(xmlDom.name))
    {
	  //this dom is a nested dom
      if(domNode.opcode == OP_TUPLELIST)
	  {
        SimpleNode newDomNode = createXmlTree(domNode,relDom);
        newDomNode.jjtSetParent(newTupleNode);
		newTupleNode.jjtAddChild(newDomNode);
      }
      else //this dom is a flat dom
      {

		//create new dom node and add it to the "Tuple" node
		SimpleNode newDomNode = (SimpleNode)domNode.jjtDeepDuplicate();

        //modify the opcode of the domNode
        //if the domtype != strg, then modify the opcode of newDomNode
		if(!xmlDom.type.equals("strg")|!xmlDom.type.equals("string"))
		{
           if(domNode.info.equals("null"))
           {
              newDomNode.opcode = OP_NULL;
              newDomNode.setInfo(null);
              //System.out.println("info = null");
		   }
           else if(domNode.info.equals("dc"))
           {
			 newDomNode.opcode = OP_DC;
			 newDomNode.setInfo(null);
             //System.out.println("info = DC");
		   }
           else if(domNode.info.equals("dk"))
		   {
		   	 newDomNode.opcode = OP_DK;
		   	 newDomNode.setInfo(null);
		     //System.out.println("info = DK");
		   }

		   //motify info object and opcode
           else
           {
		    String st=(String)(newDomNode.info);
		    if(xmlDom.type.equals("intg")|xmlDom.type.equals("Integer")|
		       xmlDom.type.equals("integer"))
		    {
			 if(st.endsWith("i")|st.endsWith("I"))
     	         st = st.substring(0,(st.length())-1);
     	     try {
			       Integer inte = Integer.valueOf(st);
		           newDomNode.setInfo(inte);
			     } catch (NumberFormatException e)
			      { throw new InterpretError("Invalid integer : "+st);}

             //Integer inte = Integer.valueOf(st);
		     //newDomNode.setInfo(inte);
		     newDomNode.opcode = OP_INTEGER;
		    }
		    else if(xmlDom.type.equals("long")|xmlDom.type.equals("Long"))
		    {
     	      if(st.endsWith("l")|	st.endsWith("L"))
		         st = st.substring(0,(st.length())-1);
		      try {
			        Long l=Long.valueOf(st);
     	            newDomNode.setInfo(l);
			      } catch (NumberFormatException e)
			     {throw new InterpretError("Invalid long : "+st);}

     	     //Long l=Long.valueOf(st);
     	     //newDomNode.setInfo(l);
		     newDomNode.opcode = OP_LONG;
		    }
           else if(xmlDom.type.equals("short")|xmlDom.type.equals("Short"))
           {

     	     if(st.endsWith("s")|st.endsWith("S"))
			 	   st = st.substring
     	                 (0,(st.length())-1);
     	     try {
			        Short s = Short.valueOf(st);
		            newDomNode.setInfo(s);
			      } catch (NumberFormatException e)
			     {throw new InterpretError("Invalid short : "+st);}


     	     //Short s = Short.valueOf(st);
		     //newDomNode.setInfo(s);
		     newDomNode.opcode = OP_SHORT;
		    }
           else if(xmlDom.type.equals("number"))
		    {
			 number n = new number((String)(newDomNode.info));
			 newDomNode.setInfo(n);
			 newDomNode.opcode = OP_NUMERIC;
		    }
           else if(xmlDom.type.equals("bool")|xmlDom.type.equals("boolean"))
           {
			 newDomNode.opcode = OP_BOOLEAN;
			 if(st.equals("true"))
			   newDomNode.setObject(new Byte((byte) 1));
			 else if(st.equals("false"))
			   newDomNode.setObject(new Byte((byte) 0));
			 else
			   throw new InterpretError("Error: invalid boolean literal");

		    }
		   else if(xmlDom.type.equals("double"))
			{
			 if(st.endsWith("d")|st.endsWith("D"))
			 	  st = st.substring(0,(st.length())-1);
			 try {
			        Double d = Double.valueOf(st);
			        newDomNode.setInfo(d);
			      } catch (NumberFormatException e)
			     {throw new InterpretError("Invalid double : "+st);}
			 //Double d = Double.valueOf(st);
			 //newDomNode.setInfo(d);
			 newDomNode.opcode = OP_DOUBLE;
		    }
		    else if(xmlDom.type.equals("float")|xmlDom.type.equals("real"))
			{
			 if(st.endsWith("f")|st.endsWith("F"))
			 	    st = st.substring(0,(st.length())-1);
			  try {
			        Float f = Float.valueOf(st);
			        newDomNode.setInfo(f);
			      } catch (NumberFormatException e)
			     {throw new InterpretError("Invalid float : "+st);}

			 //Float f = Float.valueOf(st);
			 //newDomNode.setInfo(f);
			 newDomNode.opcode = OP_FLOAT;
		    }
		  }


	    }
		newDomNode.jjtSetParent(newTupleNode);
		newTupleNode.jjtAddChild(newDomNode);
      }

     j++;
    }

    else //doesn't match
    {
      //create a new dom node which has dc value
      SimpleNode newDomNode = SimpleNode.jjtCreateNode
               ("Literal",OP_LITERAL,OP_DC,xmlDom.name);
      newDomNode.jjtSetParent(newTupleNode);
	  newTupleNode.jjtAddChild(newDomNode);
	}
	order++;

	if(order == ((XmlRelInfo)relDom.elementAt(pos)).doms.size())
	{
      order =0;
    }

	if(j==tuple.jjtGetNumChildren())
	{
      XmlDomInfo xmlDomLast =(XmlDomInfo)((XmlRelInfo)relDom.elementAt(pos)).doms.lastElement();
      if(domNode.name.indexOf("type")!=-1)
      {
	    pos1 = domNode.name.indexOf("type");
	    domName = domNode.name.substring(0,pos1).trim();
      }
      else
        domName = domNode.name;

      if(!domName.equals(xmlDomLast.name))
      {
        for(int m=order;m<((XmlRelInfo)relDom.elementAt(pos)).doms.size();m++)
        {
        XmlDomInfo xmlDomS =(XmlDomInfo)((XmlRelInfo)relDom.elementAt(pos)).
                            doms.elementAt(m);
        SimpleNode newDomNode = SimpleNode.jjtCreateNode
                    ("Literal",OP_LITERAL,OP_DC,xmlDomS.name);

        newDomNode.jjtSetParent(newTupleNode);
	    newTupleNode.jjtAddChild(newDomNode);
	    }
      }
      break;
	}

   }//end of while


  }//end of For

 //System.out.println(" ---------- newTupleList --------  ");
 //Global global = new Global();
 //global.traverse(newTupleList);

 return newTupleList;
}

/* end zhan's addition Nov 2003





  /**
   * Traverse the syntax tree of a virtual domain.
   * @param node root of the syntax tree
   * @param flag    0: don't change domains' reference counters
   *                1: increment domains' reference counters
   *                2: decrement domains' reference counters
   */
  private void traverseNode(SimpleNode node, int flag, Environment env)
       throws InterpretError
  {
    // REVISIT: have to check other cases e.g. node.name is
    // Rel name etc.

    if(node == null) return;
		//System.out.println("type of node:"+node.type);
    if(node.type == OP_IDENTIFIER){
//System.out.println("node:"+node.name);
      Domain d = env.lookupDom(node.name, true);


      if(d != null){
       	if (d.type==UNION)
	{
	}
	// if it's a declared domain...
	//Yi Zheng May 2, 2002
	if (d.isStateDom == 1)
		throw new InterpretError("virtual domain defined on a state domain:" + d.name);

	if(flag == 1)
	  d.numref++;      // increment domain's numref
	else if(flag == 2)
	  d.numref--;      // decrement domain's numref
      }
      else{
	Relation r = env.lookupRel(node.name, true);
	if(r == null)
/*
  		throw new InterpretError(
				   "domain '"+node.name +"' has not been declared.");
*/
	{   Text t = env.lookupText(node.name, true);  //Jiantao 2004
                if(t == null)
                        throw new InterpretError("domain '"+node.name +"' has not been declared.");
        } 

      }
    }

    // continue traversing...
    int numchildren = node.jjtGetNumChildren();
    if(numchildren != 0){
      // the source relation in projection and selection should not
      // be counted as domain.
      //-- this is no longer true once we allow
      // state and local vars to be of type IDLIST, so the following 2
      //lines are commented out by Yi Zheng May 2, 2002
//      if(node.type == OP_PROJECT)
//	numchildren--;

      // Traverse into the children.
      for(int i = 0; i < numchildren; i++){
				SimpleNode n = (SimpleNode)node.jjtGetChild(i);

	// For join operations, the operand might not be a domain.
	// For example, let x be S ijoin R; where R is a top-level
	// relation. Hence, we postpone the check alittle later.
	//-- for the same reason as above, the following if switch is commented
	//out by Yi Zheng, May 2, 2002
//	if(!(node.type == OP_JOINOPERATOR && n.type == OP_IDENTIFIER))
				
				traverseNode(n, flag, env);
      }
    }
  }


  // Get the actualized type. Note, this type might be changed
  // later, so must be checked again at run-time.
     /**
   * JRelixNS modified/enhanced this method:
   *  1) Check for empty exprlist.
   *  2) Support processing for non-scalars in OP_SELECT.
   *  3) Support processing of stand-alone top-level RELATION for sigma-join.
   */
  private int traverseType(SimpleNode node, Environment env)
       throws InterpretError
  {
    int i = 0;
    int[] typeArray;                    // An array of types
    int size = node.jjtGetNumChildren();
    boolean cond = false; //Added by hongyu. Sept 30,01
    SimpleNode  qtCond = null;//Added by hongyu. Sept 30,01

    switch(node.type){
        /******* start adding by Fan *******/
        case OP_RELATION:       // May 2004
                case OP_TRANSPOSENEW:   // June 2004
                        return IDLIST;
                case OP_TYPEOF:                 // July 2004
                        return TYPE;
                case OP_SELF:                   // August 2004
                        return ATTRIBUTE;       //  case: let srel be self;
                case OP_EVAL:
                        return UNIVERSAL;   // September 29, 2004  // let X be eval A;
                /******* end of Fan's addition *******/

   case OP_QUOTE://******************** QUOTE OPERATION ********************** //from Andrey
	return ATTRIBUTE;		 //????????????????????????
    case OP_RED:
      // There is only only child for Red operations.
    case OP_EQUIV:
      // For Equiv operations, we only need to take care of the first
      // child for type information.
     //Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
        /*
        if(node.opcode != OP_REDFUNCALL)
                return(traverseType((SimpleNode)node.jjtGetChild(0), env));
        else
                return(traverseType((SimpleNode)node.jjtGetChild(1), env));
        */// commented out on Oct. 11 2004 use following modification

        /******* start modifying by Fan  *******/
        SimpleNode tpnode = null;
        int childnum=0;
        if(node.opcode != OP_REDFUNCALL){
                tpnode = (SimpleNode)node.jjtGetChild(0);
                childnum = 0;
        }
        else{
                tpnode = (SimpleNode)node.jjtGetChild(1);
                childnum = 1;
        }
        int tp = 0;
        tp = traverseType(tpnode, env);
        if(isJoinOperator(node))
                tp  = ScalarDom2Relation(node, childnum, tp);
        return tp;
    /******* end of Fan's modification  *******/

    case OP_FUN://From sungsoo. hongyu added Feb 2002
      //System.out.println("OP_FUN: traverseType1");
	//Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
        /*
        if(node.opcode != OP_REDFUNCALL)
              return (traverseType((SimpleNode)node.jjtGetChild(0),env));
        else
                return(traverseType((SimpleNode)node.jjtGetChild(1), env));
        */ // commented out on Oct. 11 2004 use following modification
/******* start modifying by Fan  *******/
        if(node.opcode != OP_REDFUNCALL){
                tpnode = (SimpleNode)node.jjtGetChild(0);
                childnum = 0;
        }
        else{
                tpnode = (SimpleNode)node.jjtGetChild(1);
                childnum = 1;
        }
        tp = traverseType(tpnode, env);
        if(isJoinOperator(node))
                tp  = ScalarDom2Relation(node, childnum, tp);
        return tp;
/******* end of Fan's modification  *******/

    case OP_PAR:
      //System.out.println("OP_PAR: traverseType1");
	//Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
        /*
        if(node.opcode != OP_REDFUNCALL)
              return (traverseType((SimpleNode)node.jjtGetChild(0),env));
        else
                return (traverseType((SimpleNode)node.jjtGetChild(1), env));
      //End of sungsoo's code. hongyu added Feb 2002
        */  // commented out on Oct. 11 2004 use following modification
/******* start modifying by Fan  *******/
        if(node.opcode != OP_REDFUNCALL){
                tpnode = (SimpleNode)node.jjtGetChild(0);
                childnum = 0;
        }
        else{
                tpnode = (SimpleNode)node.jjtGetChild(1);
                childnum = 1;
        }
        tp = traverseType(tpnode, env);
        if(isJoinOperator(node))
                tp  = ScalarDom2Relation(node, childnum, tp);
        return tp;
/******* end of Fan's modification  *******/

    case OP_IFTHENELSE:
     // if(BOOLEAN == traverseType((SimpleNode)node.jjtGetChild(0),env)){
	int type = traverseType((SimpleNode)node.jjtGetChild(1), env);
	if(type == traverseType((SimpleNode)node.jjtGetChild(2), env))
	  return type;
	else
	  throw new InterpretError("type mismatch between 'then' clause"+
				   " and 'else' clause.");
     // }else
     //	throw new InterpretError("type of 'if' clause is not boolean.");
    case OP_SELECT:
    case OP_PROJECT:
    case OP_TSELECT:
    case OP_QSELECT: // this is added by hongyu a <- quant (#>2) COMP in SUPPLY;
      // This branch was for OP_SELECT & OP_PROJECT only. To fit for
      // OP_TSELECT, both SELECT & PROJECT are used. As for TSELECT,

      // the following tree structure is presented for
      //        let x be [A,B] where A=3 or B=2 in (S ijoin T);
      //                  TSelect
      //                  /     \
      //            Projector   Selector
      //                /        /    \
      //           ExprList    Or      Join
      //             /  \     /  \     /  \
      //            A    B  Comp Comp  S  T
      //                    /  \ /  \
      //                    A  3 B  2
      // First, get the source relation.
      Domain[] darray = null;
      Relation srcrel = null;
      //SimpleNode srcnode = (SimpleNode)node.jjtGetChild(1);
      SimpleNode srcnode = null; //revised by hongyu
      if((node.type == OP_PROJECT) ||(node.type == OP_SELECT)){
	srcnode = (SimpleNode)node.jjtGetChild(1);
      }
      else if(node.type == OP_TSELECT)// here is case of OP_TSELECT
	srcnode = (SimpleNode)node.jjtGetChild(1);

      else{//OP_QSELECT  a <- quant (#>2) COMP in SUPPLY;
	if(node.jjtGetNumChildren()==2){
	  cond = true;
	  srcnode = (SimpleNode)node.jjtGetChild(1);
	}
	else{
	  qtCond = (SimpleNode)node.jjtGetChild(1);

	  srcnode = (SimpleNode)node.jjtGetChild(2);
	}
      }
   //   System.out.println(" srcnode.type "  + srcnode.type);
   //   System.out.println(" srcnode.name "  + srcnode.name); // End of Hongyu modification

      switch(srcnode.type){
        /******* start adding by Fan  *******/
        case OP_RELATION:      // May 2004
            case OP_TRANSPOSENEW:  // June 2004
                return IDLIST;
            case OP_QTOPERATOR:  // added on August 8, 2004 Interpreter "." in domain algebra as any level relation here.
                                                // as the wildcard "." is recognized as OP_QTOPERATOR, so far.
                if (srcnode.opcode == OP_DOT)
                        return IDLIST;
                else throw new InterpretError("traverseType1: to be implemented.");

            // the following part of case OP_RED, OP_EQUIV, OP_FUN, and OP_FUN was copied (see the author inside the code)and added by Fan on September 12, 2004.
            // To support vertical operation as source relation.
            // the motivative case is "let multicite be author quant(#>1)cite in red ujoin of (authors ujoin cite);"
            case OP_RED:// There is only only child for Red operations.
            case OP_EQUIV:
                // For Equiv operations, we only need to take care of the first
                // child for type information.
               //Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
                if(srcnode.opcode != OP_REDFUNCALL)
                                return(traverseType((SimpleNode)srcnode.jjtGetChild(0), env));
                  else
                        return(traverseType((SimpleNode)srcnode.jjtGetChild(1), env));
              case OP_FUN://From sungsoo. hongyu added Feb 2002
                //System.out.println("OP_FUN: traverseType1");
                //Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
                if(srcnode.opcode != OP_REDFUNCALL)
                      return (traverseType((SimpleNode)srcnode.jjtGetChild(0),env));
                else
                        return(traverseType((SimpleNode)srcnode.jjtGetChild(1), env));
              case OP_PAR:
                //System.out.println("OP_PAR: traverseType1");
                //Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
                if(srcnode.opcode != OP_REDFUNCALL)
                      return (traverseType((SimpleNode)srcnode.jjtGetChild(0),env));
                else
                        return (traverseType((SimpleNode)srcnode.jjtGetChild(1), env));
                //End of sungsoo's code. hongyu added Feb 2002
        /******* end of Fan's addition *******/

      case OP_IDENTIFIER:
	String srcrelname = "."+srcnode.name;
	srcrel = env.lookupRel(srcrelname, true);

	// REVISIT: use NREnvironment to find the top-level rel.
	if(srcrel == null){
	  srcrel = env.lookupRel(srcnode.name, true);
	  if(srcrel == null)

	    throw new InterpretError("source relation" + srcnode.name + " not found.");
	}
	break;
      case OP_SELECT:
      case OP_PROJECT:
      case OP_TSELECT:
      case OP_JOINOPERATOR:
      case OP_QSELECT:
      case OP_TEXT2ATTR: // Jiantao 2004
      case OP_MU2NEST: // Jiantao 2004

	// Firstly we go through the source relation tree (might be
	// a tree of join operations.
	int t = traverseType(srcnode, env);
	if(t != IDLIST && t != COMP)
	  // should never happen..
	  throw new InterpretError("mismatched type.");
	srcrel = getIDList(srcnode, env);
//	System.out.println("end of getIDList");
	break;
      case OP_TRANSPOSE: //?????????????????????
		 return IDLIST; //from Andrey
		 //added by Yi Zheng July 7, 2002
      case OP_ARRAY:
	//	return IDLIST;
            break; //modified August 13, 2002
		//end of Yi's addition July 7, 2002
      default:
	throw new InterpretError("traverseType1: to be implemented.");
      }
   //   System.out.println("other.type "  + node.type);
      if(node.type == OP_PROJECT || node.type == OP_TSELECT)//Revised by hongyu
	{

	  // Get the projection domain list.
	  // START Since jRelixNS
	                   //  o Check for empty exprlist.
	  SimpleNode exprlist;
//	  if (env.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
           if (Relation.isSigmaJoinImplemented != true) {
	    exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	  }
	  else {
                    // This is an enhancement to jRelix76.
	    // The exprlist might not exist for an empty projector, [],
	    // used for nullary tuple results.
	    exprlist = (SimpleNode)node.jjtGetChild(0);
	    if(exprlist.jjtGetNumChildren() > 0) // expect 1 or 0 only.
	      exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	  }
	  // END Since jRelixNS
	  Vector domvec = new Vector();
	  for(i = 0; i < exprlist.jjtGetNumChildren(); i++){
//In order to to allow for the following:

//let domB be [red + of a] in A;
//  we have to generate a temporary domain for red + of a
//  set domB's type to the temporary domain's type
//	this temporary domain is generated when actualized
// -- Yi Zheng July 1, 2002

		if((i == 0) && (exprlist.jjtGetNumChildren() == 1)
			&& (((SimpleNode) exprlist.jjtGetChild(0)).type == OP_RED))
		{
			SimpleNode firstChild = null;
			if(((SimpleNode)exprlist.jjtGetChild(0)).opcode != OP_REDFUNCALL)
				firstChild = (SimpleNode) exprlist.jjtGetChild(0);
			else
				firstChild = (SimpleNode) exprlist.jjtGetChild(1);

			int redType = traverseType(firstChild, env);
			return redType;
		}
		else
		{
			String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
		    Domain d = env.lookupDom(dname, true);
                    /*
                    if(d == null)
                      throw new InterpretError("domain "+dname+" not found.");
                      */  // original code, commented out September, 2004

                    /******* Start modifying by Fan *******/
                    if(d == null){  // Sept. 4, 2004, for case "let s be [self] in Q;"
                        SimpleNode selfNode = ((SimpleNode)exprlist.jjtGetChild(i));
                                if(selfNode.type == OP_SELF) {
                                        Domain selfdom = env.lookupDom("self",true);
                                        if(selfdom== null){
                                                selfdom= new Domain("self",ATTRIBUTE,0,null);  // gives name and type
                                                env.put(selfdom);
                                        }
                                        d = selfdom;
                          } else
                              throw new InterpretError("domain "+dname+" not found.");
                    }
                    /******* end of Fan's modification *******/


		    domvec.addElement(d);

		}
		/* original code //commented out July 1, 2002

	    String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
	    Domain d = env.lookupDom(dname, true);
	    if(d == null)
	      throw new InterpretError("domain "+dname+" not found.");

	    	domvec.addElement(d);
	    	*/

	  }
	  darray = new Domain[domvec.size()];
	  domvec.copyInto(darray);

	  //Yi Zheng May 2, 2002
	  // "Newing" an actualizer will report any errors involved..
	  new Actualizer(darray, srcrel, env);

	  return IDLIST;
	}
      else if(node.type == OP_SELECT)// revised by hongyu
	{

	  // node.type == OP_SELECT
	                // START Since jRelixNS
              // o Support processing for non-scalars in OP_SELECT.
              if (Relation.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
	  SimpleNode complist = (SimpleNode)node.jjtGetChild(0);
	  Hashtable htbl = new Hashtable();
	  getDomsInCompList(complist, htbl, env);

	  darray = new Domain[htbl.size()];
	  int count = 0;
	  for(Enumeration e = htbl.keys(); e.hasMoreElements();)
	    darray[count++] = (Domain) htbl.get(e.nextElement());
	  		  // "Newing" an actualizer will report any errors involved..
	  new Actualizer(darray, srcrel, env);

	  // Always return IDLIST type at this stage.
	  return IDLIST;

	                }else{
	         SimpleNode sn = new SimpleNode("virtmp");
	         sn.jjtAddChild((SimpleNode)node.jjtGetChild(0));
	         Domain d = new Domain("0select0", BOOLEAN, 0, sn);
	         Domain[] doms = new Domain[1];
	         doms[0] = d;
	         // "Newing" an actualizer will report any errors involved..
	         new Actualizer(doms, srcrel, env);
	         // Always return IDLIST type at this stage.
	         return IDLIST;
              }
              // END Since jRelixNS
	}


      else if(node.type == OP_QSELECT){// revised by hongyu
	return IDLIST;
      }
      else
	throw new InterpretError("node type is not decalred ");
    case OP_JOINOPERATOR:
      typeArray = new int[2];
      typeArray[0] = traverseType((SimpleNode)node.jjtGetChild(0), env);
      typeArray[0] = ScalarDom2Relation(node, 0, typeArray[0]);   // Fan Guo August 7, 2004
      if(node.jjtGetNumChildren() == 2){
	typeArray[1] = traverseType((SimpleNode)node.jjtGetChild(1), env);
        typeArray[1] = ScalarDom2Relation(node, 1, typeArray[1]);   // Fan Guo August 7, 2004
      }else{
	typeArray[1] = traverseType((SimpleNode)node.jjtGetChild(3), env);
        typeArray[1] = ScalarDom2Relation(node, 3, typeArray[1]);   // Fan Guo August 7, 2004
      }
      return(Utility.typeCheck(typeArray, node.opcode));

        case OP_TEXT2ATTR: // Jiantao 2004
        case OP_MU2NEST:
                return IDLIST;

	case OP_TRANSPOSE: //?????????????????????
	 return IDLIST; //from Andrey

        case OP_FUNCTION: //Yi Zheng May 3, 2002
	 SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
	 int argType = traverseType(argument, env);
	 if(argType == NUMERIC) return argType;//from Yiyi

     switch (node.opcode)
     {
		case OP_ABS:
			if(!Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return argType;


		case OP_SQRT:
		case OP_SIN:
		case OP_ASIN:
		case OP_COS:
		case OP_ACOS:
		case OP_TAN:
		case OP_ATAN:
		case OP_LOG:
		case OP_LOG10:
			if(!Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return DOUBLE;

		case OP_ROUND:
		case OP_CEIL:
		case OP_FLOOR:
			if(!Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return INTEGER;
		case OP_ORD:

			if (!Utility.isText(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return INTEGER; //elevate to highest level of precision
		case OP_CHR:
			if (! Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return STRING;

               /*******  Rong  Li, *********************/
                case OP_ISNULL:
                case OP_ISNULLDC:
                case OP_ISNULLDK:
                        return BOOLEAN;
              /********  Rong Li,  *********************/   

		case OP_ISKNOWN:
		case OP_SINH:
		case OP_COSH:

		case OP_TANH:
			throw new InterpretError("Not implemented");
		default:
			throw new InterpretError("Unknown function operator");
     }
    case OP_ARRAY:
    	return IDLIST; //Yi Zheng June 19, 2002
    default:
      // START Since jRelixNS
  //    if (env.isSigmaJoinImplemented == true) {
	// Allow processing for sigma-join.
       if (Relation.isSigmaJoinImplemented == true) { //Andy, May 12, 2002

               // START Since jRelCaNu a.
             // OP_CAST is a OP_BIOPERATOR, but
             // the 2 children nodes should be explicitely
             // checked for cast compatibility.
             // Also store the CastFromType in bits of the OP_TYPE node.
             if(node.opcode == OP_CAST) {// also (node.type == OP_BIOPERATOR)
                int CastToType = Utility.typeConvert(((SimpleNode)node.jjtGetChild(0)).opcode);
                int CastFromType = traverseType((SimpleNode)node.jjtGetChild(1), env);
                ((SimpleNode)node.jjtGetChild(0)).bits = CastFromType;
                if (!Utility.domsCastCompatible(CastToType, CastFromType))
                {
                   Utility.dump(node,"traverseType CAST Error ->");
                   throw new InterpretError("traverseType: Casting incompatible types:"
                   + " CastToType: " + Utility.type2String(CastToType)
                   + " CastFromType: " + Utility.type2String(CastFromType));
                }
                return CastToType;
             }
             /******* start adding by Fan  *******/
             else if(node.opcode == OP_CAT) {  // on August 13, 2004
                SimpleNode left = ((SimpleNode)node.jjtGetChild(0));
                SimpleNode right = ((SimpleNode)node.jjtGetChild(1));
                int lType = traverseType(left,env);
                int rType = traverseType(right,env);
                if(lType == rType && rType == STRING) return STRING;
                else {
                        node.dump("InterpretError traverseType==>");
                  throw new InterpretError("TraverseType: cat types not compatible.");
                }
             }
             /******* end of Fan's addition *******/

             // END Since jRelCaNu a.

       // Support processing of stand-alone top-level RELATION for sigma-join
	Domain IdNodeDom;
	Relation IdNodeRel;

//	if(size != 0){
	//added by Yu, Oct 15, 2004
	if(size != 0 ){
	 if (node.opcode==OP_SUBSTR)
	 {
	 	return 7;
	 }
	 else
	 {
	  typeArray = new int[size];
	  for (java.util.Enumeration e = node.children.elements();
	       e.hasMoreElements();) {
	    SimpleNode n = (SimpleNode)e.nextElement();
//System.out.println("!!!Interpreter:n.type:"+n.type+"!!!e:"+node.opcode);
	    switch(n.type){
	    case OP_LITERAL:
	      typeArray[i++] = Utility.typeConvert(n.opcode);
	      break;
	    case OP_IDENTIFIER:
	      // Handle the relation case:
	      // REVISIT: need use 'true' for lookup?
	      //          and how about relation case?
	      // Old code below will cause a java null pointer exception
	      // in case we have allowd a top level relation node.
	      //typeArray[i++] = env.lookupDom(n.name, true).type;
	      if ((IdNodeDom = env.lookupDom(n.name, true)) != null)
		typeArray[i++] = IdNodeDom.type;
	      //typeArray[i++] = env.lookupDom(n.name, true).type;
	      else if ((IdNodeRel = env.lookupRel(n.name, true)) != null)
		typeArray[i++] = RELATION;
	      break;
	    case OP_BIOPERATOR:

	    case OP_UOPERATOR:
	    case OP_JOINOPERATOR:
	                      //START Since JRelCaNu b.
                  // - to allow project and select in
                  //   nested relational expressions;
                  //   for example, in a OP_LET expression.
                  // Should have been in jRelixNS.
                  case OP_SELECT:
                  case OP_PROJECT:
                  case OP_TSELECT:
                  //End Since JRelCaNu b.

	      typeArray[i++] = traverseType(n, env);
	      break;
	    case OP_RED:
	    case OP_EQUIV:
		//Yi Zheng July 8, 2002, added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0), env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1), env);

	      break;
	    case OP_FUN: //sungsoo's code hongyu added Feb 2002
	    case OP_PAR:
	      //System.out.println("traverseType2");
		//Yi Zheng July 8, 2002, added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0),env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1),env);

	      break;

	    case OP_IFTHENELSE:
	      typeArray[i++] = traverseType(n, env);
	      break;
	    case OP_FUNCTION: //Yi Zheng May 6, 2002
	     typeArray[i++] = traverseType(n, env);
	      break;
	    default:
	      throw new InterpretError("traverseType2 4991: to be implemented.");
	    }
	  }
	  return(Utility.typeCheck(typeArray, node.opcode));
	 }
	}
	else{
	  if(node.type == OP_LITERAL)
	    // Node with children e.g. let b be 1;
	    return(Utility.typeConvert(node.opcode));
	  else
	    // Node without children e.g. let b be a;
	    // REVISIT: need use 'true' for lookup?
	    // Added to handle the relation case:
	    {//System.out.println("node.name:"+node.name);
	      if ((IdNodeDom = env.lookupDom(node.name, true)) != null)
		return(IdNodeDom.type);
	      else if ((IdNodeRel = env.lookupRel(node.name, true)) != null)
	      { if (IdNodeRel.rvc == RELATION || IdNodeRel.rvc == VIEW)
				return(IDLIST);
			else return (COMP);//Yi Zheng May 2, 2002
		}
	      else
		throw new InterpretError("traverseType: OP_IDENTIFIER not LITERAL, DOM, or REL/COMPUTATION.");
	    }
	}
      }
      else {
	if(size != 0){
	  typeArray = new int[size];
	  for (java.util.Enumeration e = node.children.elements();
	       e.hasMoreElements();) {
	    SimpleNode n = (SimpleNode)e.nextElement();
	    switch(n.type){
	    case OP_LITERAL:
	      typeArray[i++] = Utility.typeConvert(n.opcode);
	      break;
	    case OP_IDENTIFIER:
	      // REVISIT: need use 'true' for lookup?
	      //          and how about relation case?
	      typeArray[i++] = env.lookupDom(n.name, true).type;
	      break;
	    case OP_BIOPERATOR:
	    case OP_UOPERATOR:
	    case OP_JOINOPERATOR:
	      typeArray[i++] = traverseType(n, env);
	      break;
	    case OP_RED:
	    case OP_EQUIV:
		//Yi Zheng July 8, 2002 added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0), env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1), env);

	      break;

	    case OP_FUN://sungsoo's code. hongyu added Feb 2002
	    case OP_PAR:
	      //System.out.println("traverseType2");
		//Yi Zheng July 8, 2002 added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0),env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1),env);

	      break;
	    case OP_IFTHENELSE:
	      typeArray[i++] = traverseType(n, env);
	      break;
	    default:
	      throw new InterpretError("traverseType2: to be implemented.");
	    }
	  }
	  return(Utility.typeCheck(typeArray, node.opcode));
	}else{
	  if(node.type == OP_LITERAL)
	    // Node with children e.g. let b be 1;
	    return(Utility.typeConvert(node.opcode));
	  else
	    // Node without children e.g. let b be a;
	    // REVISIT: need use 'true' for lookup?
	    //System.out.println(node.name + " of type " + Utility.type2String(env.lookupDom(node.name, true).type) + " about to be returned from traverseType()");
	    return(env.lookupDom(node.name, true).type);
	}
      }
    }
  }

  // implemented by Weizhong Sun, August 1999
  private int executeEventH(String prefixname, SimpleNode enode, Environment env)
       throws InterpretError
  {
    String relname = new String("");
    String actionname = new String("");
    String attrlist = new String("");

    RelTable reltable = env.reltable;
    Hashtable reltb = reltable.getRelTable();
    //	int resetnested = 0;
    int ret = 0;

    // get the event query
    int opcode = enode.opcode;
    actionname = env.opcodeToString(opcode);
    relname = ((SimpleNode)enode.jjtGetChild(0)).name;

    SimpleNode stmtList = (SimpleNode) enode.jjtGetChild(1);
    int numStmts = stmtList.jjtGetNumChildren();
    // for Add/Delete, no attrlist
    if (opcode == OP_CHANGE) {
      for(int i = 0; i < numStmts; i++)
	{
	  SimpleNode stmt = (SimpleNode) stmtList.jjtGetChild(i);
	  String attr = ((SimpleNode) stmt.jjtGetChild(0)).name;
	  if(attrlist.compareTo("") == 0)
	    attrlist = attr;
	  else
	    attrlist = attrlist+","+attr;
	}
    }

    // and select the right event handler to execute
    for(Enumeration e = reltb.keys(); e.hasMoreElements();)
      {
	Relation rel = (Relation)reltb.get(e.nextElement());
	String name = rel.name;

	if (rel.tree != null && rel.rvc == COMPUTATION && rel.active == 1) {
	  SimpleNode root = (SimpleNode) rel.tree;

	  if (root.jjtGetNumChildren() != 0) {
	    // check the first child's opcode
	    SimpleNode firstchild = (SimpleNode) root.jjtGetChild(0);
	    if (firstchild.opcode == OP_EVENTNAME) {
	      // cut name into ename and eattr
	      int bindex = name.indexOf('[');
	      int eindex;
	      String ename = new String("");
	      String eattr = new String("");
	      if (bindex != -1) {
		eindex = name.indexOf(']');
		ename = name.substring(0, bindex);
		eattr = name.substring(bindex+1, eindex);
	      } else
		ename = name;

	      // compare event name with prefixname, actionname,
	      // relname, attrlist
	      String iname = prefixname+":"+actionname+":"+relname;
	      if (ename.compareTo(iname) == 0) {

		// check semantic of add/delete event handler
		if (((actionname.compareTo("add") == 0) ||
		     (actionname.compareTo("delete") == 0)) &&
		    (eattr.compareTo("") != 0))

		  throw new InterpretError("add/delete event handler semantic error.");

		// check the attribute list
		if (IsEmbody(eattr, attrlist)) {

		  ret=1;
		  SimpleNode secondchild = (SimpleNode)root.jjtGetChild(1);
		  // execute Declaration body whose opcode is OP_COMPVARDELA
		  SimpleNode declnode = (SimpleNode) secondchild.jjtGetChild(0);
		  int declnum = declnode.jjtGetNumChildren();
		  for (int i = 0; i < declnum; i++) {
		    SimpleNode declstmt = (SimpleNode) declnode.jjtGetChild(i);
		    executeDeclaration(declstmt, env);
		  }

		  // execute Computation body whose opcode is OP_COMPBLOCK

		  SimpleNode blocknode = (SimpleNode) secondchild.jjtGetChild(1);
		  int blocknum = blocknode.jjtGetNumChildren();

		  for (int i = 0; i < blocknum; i++) {
		    SimpleNode blockstmt = (SimpleNode) blocknode.jjtGetChild(i);

		    interpret(blockstmt, env, relname, opcode);
		  }

		  /*
		    if (resetnested != 0) {
		    nested = resetnested;
		    resetnested = 0;
		    }
		    */
		}
	      }
	    } // else skip
	  } else {
	    // never should be here
	    throw new InterpretError("computation root error.");
	  }
	}

      }
    return ret;
   }


  //written by YuLing Chen, Oct 28, 2000
  public boolean executeEventH( String actionname, Environment env)
       throws InterpretError
  {

    RelTable reltable = env.reltable;
    Hashtable reltb = reltable.getRelTable();
    boolean exist = false;//whether the event handler exists or not
    int opcode=env.stringToOpcode(actionname);

    // and select the right event handler to execute
    for(Enumeration e = reltb.keys(); e.hasMoreElements();)
      {
	Relation rel = (Relation)reltb.get(e.nextElement());
	String name = rel.name;


	if (rel.tree != null && rel.rvc == COMPUTATION && rel.active == 1) {
	  SimpleNode root = (SimpleNode) rel.tree;

	  if (root.jjtGetNumChildren() != 0) {
	    // check the first child's opcode
	    SimpleNode firstchild = (SimpleNode) root.jjtGetChild(0);
	    if (firstchild.opcode == OP_EVENTNAME) {
	      // cut name into ename and eattr
	      int bindex = name.indexOf('[');
	      int eindex;
	      String ename = new String("");
	      String eattr = new String("");
	      if (bindex != -1) {
		eindex = name.indexOf(']');
		ename = name.substring(0, bindex);
		eattr = name.substring(bindex+1, eindex);
	      } else
		ename = name;

	      // compare event name with  actionname,
	      String iname = actionname+":";
	      if (ename.compareTo(iname) == 0) {

		// check semantic of add/delete event handler
		if (((actionname.compareTo("add") == 0) ||
		     (actionname.compareTo("delete") == 0)) &&
		    (eattr.compareTo("") != 0))
		  throw new InterpretError("add/delete event handler semantic error.");

		// check the attribute list
			SimpleNode secondchild = (SimpleNode)root.jjtGetChild(1);
			// execute Declaration body whose opcode is OP_COMPVARDELA
			SimpleNode declnode = (SimpleNode) secondchild.jjtGetChild(0);
			int declnum = declnode.jjtGetNumChildren();
			for (int i = 0; i < declnum; i++) {
			  SimpleNode declstmt = (SimpleNode) declnode.jjtGetChild(i);
			  executeDeclaration(declstmt, env);
			}

			// execute Computation body whose opcode is OP_COMPBLOCK
			SimpleNode blocknode = (SimpleNode) secondchild.jjtGetChild(1);
			int blocknum = blocknode.jjtGetNumChildren();

			for (int i = 0; i < blocknum; i++) {
			  SimpleNode blockstmt = (SimpleNode) blocknode.jjtGetChild(i);
		//	  System.out.println("execute the computation");

			  interpret(blockstmt, env );
			  exist=true;
			}

			/*
			  if (resetnested != 0) {
			  nested = resetnested;

			  resetnested = 0;
			  }
			  */
	      }
	    } // else skip
	  } else {
	    // never should be here
	    throw new InterpretError("computation root error.");
	  }
	}

      }

    return exist;
  }
  // Weizhong, Aug. 1999
  // check if handAtt <= queAtt, for event handler's attribute list checking
  // 	handAtt: the attrlist of event handler
  //	queAtt: the attrlist of event query
  // example 1:
  // attrlist in event handler: [A,B]
  // attrlist in event query: [B]
  // then return false, so the event handler won't be active.
  // example 2:
  // attrlist in event handler: []
  // attrlist in event query: [B]
  // then return true, so the event handler will be active.
  public boolean IsEmbody(String handAtt, String queAtt)
  {

    if (handAtt.compareTo("") == 0)
      return true;

    int bindex = handAtt.indexOf(',');
    if (bindex == -1) {
      // just one attribute in handAtt list
      if (queAtt.indexOf(handAtt) != -1 )
	return true;
      else
	return false;
    } else {
      // several attribute in handAtt list
      // get first attribute in handAtt list
      String att = handAtt.substring(0, bindex);
      if (queAtt.indexOf(att) == -1)
	return false;
      bindex++;
      int nindex = handAtt.indexOf(',', bindex);
      while(nindex != -1) {
	// get more attributes
	att = handAtt.substring(bindex, nindex);
	if (queAtt.indexOf(att) == -1)
	  return false;
	bindex = nindex + 1;
	nindex = handAtt.indexOf(',', bindex);
      }
      // get the last one
      att = handAtt.substring(bindex);
      if (queAtt.indexOf(att) == -1)
	return false;
    }

    return true;
  }

  // Weizhong, to clear up trigger
  public void RemoveTrigger(TriggerNode triggernode)
  {

    // remove the current level's trigger
    if (triggernode.trigger.relname.compareTo("") != 0) {
      triggernode.trigger.relname = new String("");
      if (triggernode.trigger.oldp.name.compareTo("") != 0) {
	deleteFile(triggernode.trigger.oldp.name);
	triggernode.trigger.oldp = null;
      }
      if (triggernode.trigger.newp.name.compareTo("") != 0) {
	deleteFile(triggernode.trigger.newp.name);
	triggernode.trigger.newp = null;
      }
      if (triggernode.trigger.restp.name.compareTo("") != 0) {
	deleteFile(triggernode.trigger.restp.name);
	triggernode.trigger.restp = null;
      }
    }

    // remove children's trigger
    for (int i = 0; i < triggernode.jjtGetNumChildren(); i++)
      RemoveTrigger((TriggerNode)triggernode.jjtGetChild(i));

    triggernode = null;
  }

  public void printNode(SimpleNode Node)
  {
    System.out.println("Node.type="+Node.type);
    System.out.println("Node.name="+Node.name);
    System.out.println("Node.identifier="+Node.identifier);
    System.out.println("Node.opcode="+Node.opcode);
  }

  public  int executeRelixCommand(String commd)

  {
    byte[] barray=commd.getBytes();
    ByteArrayInputStream binstream=new ByteArrayInputStream(barray);
    InputStream bin=(InputStream)(binstream);
    Global.jin.resetIn(bin);
    parser.ReInit(Global.jin);
    if(in instanceof BufferedInputStream)instack.push(in);
    in=bin;
    try {
      Global.beforeParse();
      Global.status = parser.Start();
      if (Global.status == OP_EOF) {
	System.out.println("OP_EOF in executeRelix");
	if (in instanceof FileInputStream) {
	  try {
	    Object o = instack.pop();
	    if (o instanceof FileInputStream) {
	      in.close();
				in = (FileInputStream) o;
	    } else {
	      in.close();
	      in = (InputStream) o;
	    }
	    Global.jin.resetIn(in);
	    parser.ReInit(Global.jin);
	  } catch (Exception e) {
	    Global.println("Exception during IO redirection: reset to System.in for input");
	    in = Global.instream;
	    Global.jin.resetIn(in);
	    parser.ReInit(Global.jin);
	  }
	}
      } else {
	Global.root = (SimpleNode) parser.jjtree.rootNode();
	Global.afterParse(env);
	Global.beforeInterpret(); //nothing
	Global.interpretTime = interpret(Global.root, env);

	Global.afterInterpret();
      }
    } catch (ParseError e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-1);
    } catch (InterpretError e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-2);
    } catch (Exception e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-3);
    } catch (Error e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-4);
    } finally {
      parser.jjtree.reset();
    }
    return(0);
  }//end of executeRelixCommand

  //method added by Yi Zheng Sept 29, 2002 for srd <rel>
   public  void executeJRelixCommand(String commd)
  {
    byte[] barray=commd.getBytes();
    ByteArrayInputStream binstream=new ByteArrayInputStream(barray);
    InputStream bin=(InputStream)(binstream);
    Global.jin.resetIn(bin);
    parser.ReInit(Global.jin);
    instack.push(in);
    in=bin;
    try {
      Global.beforeParse();
      Global.status = parser.Start();
      if (Global.status == OP_EOF) {
	System.out.println("OP_EOF in executeJRelixCommand");

      } else {
	Global.root = (SimpleNode) parser.jjtree.rootNode();
	Global.afterParse(env);
	Global.beforeInterpret(); //nothing
	Global.interpretTime = interpret(Global.root, env);

	Global.afterSRDInterpret();
      }
    } catch (ParseError e) {
      Global.processThrowable(e);
    } catch (InterpretError e) {
      Global.processThrowable(e);
    } catch (Exception e) {
      Global.processThrowable(e);
    } catch (Error e) {
      Global.processThrowable(e);
    } finally {
    InputStream origIn = (InputStream) instack.pop();
    Global.jin.resetIn(origIn);
    parser.ReInit(Global.jin);
    in = origIn;
      parser.jjtree.reset();
    }
  }//end of executeJRelixCommand

  public Relation getRelbyName(String name) throws InterpretError
  {
    Relation rel = null;
    rel = env.lookupRel(name, true);
    rel.load();
    return rel;
  }

  
  /******************** start of Yu's new supporting methods ***********************/
  
  //for grep
  public Grep[] positionValue(String line, String pattern)                  //June 25,2004 by Yu
  throws InterpretError
  {
   if (pattern.equals(""))
   {
    Grep g = new Grep("",0);
    Grep[] posV = new Grep[1];
    posV[0] = g;
    return posV;
   }
    //int dotstar = pattern.indexOf(".*");
    //int dotplus = pattern.indexOf(".+");
    //int dotqmark = pattern.indexOf(".?");
    /*if (dotstar>=0)//******check \.* later
    {
    	if (dotstart>0)
    	{
    		String x = pattern.substr(dotstart-1,dotstar);
    		if (x=="\\")
    		
    	}
    }*/
  	
//    int star = pattern.indexOf("*");
  //  int plus = pattern.indexOf("+");
    //int qmark = pattern.indexOf("?");
    int star = pattern.indexOf("*");
    int plus = pattern.indexOf("+");
    int qmark = pattern.indexOf("?");
    if (star>=0||plus>=0||qmark>=0)
    {
     Grep[] posV = positionValueS(line,pattern);
     return posV;
    }
    //else if (dotstar<0&&dotplus>=0&&dotqmark<0)
    Vector grepPV= new Vector();
    int start = 0;
    int length = 1;
    int p;
    Grep g = null;
    String val="";
    if (pattern.indexOf("$")>=0)
    {
       	p = backwards(line,pattern);		//position is s
	if (p>=0){
		val = line.substring(p);
        	g = new Grep(val,p);
		grepPV.addElement(g);
	}
	else	System.out.println("not possible!!!");
    }
    else
    {
     while (start+length <= line.length())
     {
		String s = line.substring(start,start+length);	//take the substring
      		String f="";
	    	try {
			       f = saveToAFile(s);		//save s to file f
            	}
      		catch (IOException e) {
                System.err.println(e);
            	}
      		//String cmd="egrep "+"\""+pattern+"\""+" "+f;
//      		String cmd="egrep "+pattern+" "+f;
	   	String[] cmd={"egrep",pattern,f};
	    //System.out.println(cmd+" and string is:"+s);
      		try {
			ExecComm command=new ExecComm();
			String outlist[] = command.runCommand(cmd);
		  	if (outlist.length>0)                         //catch the result of egrep
      			{
       			 p = backwards(s,pattern);		//position is s
			 if (p>=0){
			 	val = s.substring(p);
				//System.out.println("value:"+val);
        			g = new Grep(val,start+p);
				grepPV.addElement(g);
        			start = start+length;
        			length = 1;
			 }
	  		 else	System.out.println("not possible!!!");
      			}
      			else
 		     	length++;
           	}//try
      		catch (IOException e) {
                System.err.println(e);
            	}
      }  //while
     }//else
     Grep [] posV= new Grep[grepPV.size()];
     grepPV.copyInto(posV);   
     //System.out.println("before return from positonValue:"+posV.length);
     return posV;		
   }

  public Grep[] positionValueN(String line, String pattern)                  //June 25,2004 by Yu
  throws InterpretError
  {
    Vector grepPV= new Vector();
    int start = 0;
    int length = 1;
    int p;
    Grep g = null;
    String val="";
    if (pattern.indexOf("$")>=0)
    {
       	p = backwards(line,pattern);		//position is s
	if (p>=0){
		val = line.substring(p);
        	g = new Grep(val,p);
		grepPV.addElement(g);
	}
	else	System.out.println("not possible!!!");
    }
    else
    {
     while (start+length <= line.length())
     {
		String s = line.substring(start,start+length);	//take the substring
      		String f="";
	    	try {
			       f = saveToAFile(s);		//save s to file f
            	}
      		catch (IOException e) {
                System.err.println(e);
            	}
      		//String cmd="egrep "+"\""+pattern+"\""+" "+f;
//      		String cmd="egrep "+pattern+" "+f;
	   	String[] cmd={"egrep",pattern,f};
	    //System.out.println(cmd+" and string is:"+s);
      		try {
			ExecComm command=new ExecComm();
			String outlist[] = command.runCommand(cmd);
		  	if (outlist.length>0)                         //catch the result of egrep
      			{
       			 p = backwards(s,pattern);		//position is s
			 if (p>=0){
			 	val = s.substring(p);
				//System.out.println("value:"+val);
        			g = new Grep(val,start+p);
				grepPV.addElement(g);
        			start = start+length;
        			length = 1;
			 }
	  		 else	System.out.println("not possible!!!");
      			}
      			else
 		     	length++;
           	}//try
      		catch (IOException e) {
                System.err.println(e);
            	}
      }  //while
     }//else
     Grep [] posV= new Grep[grepPV.size()];
     grepPV.copyInto(posV);   
     return posV;		
   }
   
/**
@author Yu Gu Sep 23, 2004
@param Sting line: search line
@param String pattern: search pattern in line
@Domain[] param: parameter domain array
@boolean hasBegin: 
@boolean hasEnd
@String beginP,
@String[] domN name of parameter inside the pattern
@String endP
@return Grep[]
@exception InterpretError
Grep a paticular pattern from a relation, return a relation
*/
  public Grep[] positionValueP(String line, String pattern,String patternA,String[] domN,Domain[] param,
  		boolean hasBegin,boolean hasEnd,String beginP,String endP)                  //Sep 21,2004 by Yu
		throws InterpretError
  {
   //System.out.println("pattern:"+pattern);
   //System.out.println("hasBegin:"+hasBegin+"hasEnd:"+hasEnd+"beginP:"+beginP+"endP:"+endP);
   Grep[] posV=null;
//   System.out.println("!!!!!!!!!"+domN.length);
   if (domN!=null&&domN.length!=0)
    posV = positionValuePS(line,patternA,domN,param);	//take care of the inner parameter Sep 23,2004 by Yu
   else
    posV = positionValue(line,pattern);	//no inner parameter Sep 23,2004 by Yu
   for (int i=0;i<posV.length;i++)
   {
//    System.out.println("here!!!"+i+":"+posV[i].position+","+posV[i].value);
   }
   Grep[] posVP = new Grep[posV.length];
   String[] paraV = new String[param.length];
   for (int i=0;i<posV.length;i++)
   {
    //int pos = posV[i].position;
    //String val = posV[i].value;
    String beforeV,endV;
    if (hasBegin) 
    {
     beforeV = line.substring(0,posV[i].position);
     posV[i].position = 0;
     posV[i].value = beforeV+posV[i].value;
     
     for (int j=0;j<param.length;j++)
     {
      if (beginP.equals(param[j].name))
      {
//       System.out.println(j+":"+param[j].name);
       paraV[j] = new String(beforeV);
      }
     }
    }
    
    if (hasEnd)
    {
//     System.out.println("posV[i].position:"+posV[i].position+"posV[i].value:"+posV[i].value);
     endV = line.substring(posV[i].position+posV[i].value.length());
//     posV[i].position = 0;
     posV[i].value = posV[i].value+endV;
     
     for (int j=0;j<param.length;j++)
     {
      if (endP.equals(param[j].name))
      {
//       System.out.println(j+":"+param[j].name);
  //     System.out.println("end value:"+endV);
       paraV[j] = new String(endV);
      }
     }
    }
    if (posV[i].paraValue!=null)
    for (int j=0;j<posV[i].paraValue.length;j++)
    {
     if (posV[i].paraValue[j]!=null)
     paraV[j]=posV[i].paraValue[j];
    }
    for (int j=0;j<param.length;j++)
//     System.out.println("parameter value:"+j+" is "+paraV[j]);
    if (paraV!=null)
    {Grep g = new Grep(posV[i].value,posV[i].position,paraV);
     posVP[i] = g;
    }
   }//   for (int i=0;i<posV.length;i++)

   return posVP;		
  }
  
  public Grep positionValueSingle(String line, String pattern)                  //Sep 15,2004 by Yu
  {
    int start = 0;
    int length = 1;
    int p;
    Grep g = null;
    String val="";
    if (pattern.indexOf("$")>=0)
    {
       	p = backwards(line,pattern);		//position is s
	if (p>=0){
		val = line.substring(p);
        	g = new Grep(val,p);
		return g;
		//grepPV.addElement(g);
	}
	else	System.out.println("not possible!!!");
    }
    else
    {
     while (start+length <= line.length())
     {
		String s = line.substring(start,start+length);	//take the substring
      		String f="";
	    	try {
			       f = saveToAFile(s);		//save s to file f
            	}
      		catch (IOException e) {
                System.err.println(e);
            	}
//      		String cmd="egrep "+pattern+" "+f;
	   	String[] cmd={"egrep",pattern,f};
	    //System.out.println(cmd+" and string is:"+s);
      		try {
			ExecComm command=new ExecComm();
			String outlist[] = command.runCommand(cmd);
		  	if (outlist.length>0)                         //catch the result of egrep
      			{
       			 p = backwards(s,pattern);		//position is s
			 if (p>=0){
			 	val = s.substring(p);
        			g = new Grep(val,start+p);
				return g;
			 }
	  		 else	System.out.println("not possible!!!");
      			}
      			else
 		     	length++;
           	}//try
      		catch (IOException e) {
                System.err.println(e);
            	}
      }  //while
     }//else
  return g;
  }
  
/**
@author Yu Gu June 25, 2004	modified Dec 01,2004
@param Sting line: search line
@param String pattern: search pattern in line
@return Grep[]
Grep a paticular pattern from a string, return array of Grep.
Deal with .* or .? or .+ ;*,?,+
*/
  public Grep[] positionValueS(String line, String pattern)                  //June 25,2004 by Yu
  throws InterpretError 
  {
    int asci = 1;
    String aChar = new Character((char)asci).toString();
    Vector grepPV= new Vector();
    int dotStar = pattern.indexOf(".*");
    int lastDotStar = pattern.lastIndexOf(".*");
    //System.out.println("line is "+line+" dotStart at:"+dotStar);
    String patternBefore = "";
    String patternAfter = "";
    if (dotStar==0) //at the beginning of the pattern, for example: .*l
    {
     //System.out.println("deal with it later"); Sep 15, 2004
     patternAfter = pattern.substring(2);
     Grep[] posV = positionValue(line,patternAfter);
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      posV[i].value = line.substring(0,pos+valL);
      posV[i].position = 0;
     }
     return posV;     
    }
    if (lastDotStar==pattern.length()-2) //at the end of the pattern, like l.*
    {
     patternBefore = pattern.substring(0,pattern.length()-2);
     Grep[] posV = positionValue(line,patternBefore);
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      posV[i].value = line.substring(pos);
     }
     return posV;          
    }
    //.+
    int dotPlus = pattern.indexOf(".+");
    int lastDotPlus = pattern.lastIndexOf(".+");
    //System.out.println("line is "+line+" dotStart at:"+dotStar);
    if (dotPlus==0) //at the beginning of the pattern, for example: .+l
    {
     //System.out.println("deal with it later"); Sep 15, 2004
     patternAfter = pattern.substring(2);
     Grep[] posV = positionValue(line,patternAfter);
     Vector grepPVp= new Vector();
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      if (pos>0)	//something before l
      {
       int valL = posV[i].value.length();
       posV[i].value = line.substring(0,pos+valL);
       posV[i].position = 0;
       grepPVp.addElement(posV[i]);
      }
     }
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;     
    }
    if (lastDotPlus==pattern.length()-2) //at the end of the pattern, like l.+
    {
     Vector grepPVp= new Vector();
     patternBefore = pattern.substring(0,pattern.length()-2);
     Grep[] posV = positionValue(line,patternBefore);
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      if (pos+valL<line.length()) //something after l
      {
       grepPVp.addElement(posV[i]);
      }
     }
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;          
    }
    //.?
    int dotQ = pattern.indexOf(".?");
    int lastDotQ = pattern.lastIndexOf(".?");
    //System.out.println("line is "+line+" dotStart at:"+dotStar);
    if (dotQ==0) //at the beginning of the pattern, for example: .?l
    {
     //System.out.println("deal with it later"); Sep 15, 2004
     patternAfter = pattern.substring(2);
     Grep[] posV = positionValue(line,patternAfter);
     Vector grepPVp= new Vector();
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      if (pos==0||pos==1)	//0 or 1 before l
      {
       int valL = posV[i].value.length();
       posV[i].value = line.substring(0,pos+valL);
       posV[i].position = 0;
       grepPVp.addElement(posV[i]);
      }
     }
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;     
    }
    if (lastDotQ==pattern.length()-2) //at the end of the pattern, like l.?
    {
     Vector grepPVp= new Vector();
     patternBefore = pattern.substring(0,pattern.length()-2);
     Grep[] posV = positionValue(line,patternBefore);
//     System.out.println("Nov 02:"+posV.length);
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
//      System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      if (pos+valL==line.length()) //0 after l
      {
   	       grepPVp.addElement(posV[i]);
      }
      else if (pos+valL<line.length())
      {
       posV[i].value = line.substring(pos,pos+valL+1);	
       grepPVp.addElement(posV[i]);
      }
     }
//     System.out.println("Nov o2:"+grepPVp.size());
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;          
    }
    //*
    
//    int dotMin = Math.min(Math.min(dotStar,dotPlus),dotQ);
  //  System.out.println("!!!!!!!!!!!dotMin:"+dotMin);
    int dotMin;
    if (dotStar!=-1 && dotPlus!=-1 && dotQ!=-1)
     dotMin = Math.min(Math.min(dotStar,dotPlus),dotQ);
    else if (dotStar==-1 && dotPlus!=-1 && dotQ!=-1)
     dotMin = Math.min(dotPlus,dotQ);
    else if (dotStar!=-1 && dotPlus==-1 && dotQ!=-1)
     dotMin = Math.min(dotStar,dotQ);
    else if  (dotStar!=-1 && dotPlus!=-1 && dotQ==-1)
     dotMin = Math.min(dotPlus,dotStar);
    else if (dotStar==-1 && dotPlus==-1 && dotQ!=-1)
     dotMin = dotQ;
    else if (dotStar==-1 && dotPlus!=-1 && dotQ==-1)
     dotMin = dotPlus;
    else if (dotStar!=-1 && dotPlus==-1 && dotQ==-1)
     dotMin = dotStar;
    else dotMin=-1;
    if (dotMin!=-1) //not .*,.+,.?
    {
    //}
//    else
  //  {
     String dotWCard = pattern.substring(dotMin,dotMin+2);
     //System.out.println("dotWCard:"+dotWCard);
     if (dotWCard.equals(".*"))
     {//System.out.println("^^^^^I am here");
	patternBefore = pattern.substring(0,dotStar);
    	patternAfter = pattern.substring(dotStar+2);
    	Grep[] posVB = positionValue(line,patternBefore);
    	Grep[] posVA = positionValue(line,patternAfter);
    	//Grep[] posProd = new Grep[posVB.length*posVA.length];
    	int k=0;
    	for (int i = 0; i<posVB.length; i++)
     	 for (int j = 0; j<posVA.length; j++)
     	 {
//      int begin = posVB[i].position;
//      int end = posVA[j].position + (posVA[j].value).length();
      	  int begin = posVB[i].position+posVB[i].value.length();
      	  int end = posVA[j].position;
      	  //System.out.println("line is "+line+"posVB[i].value"+posVB[i].value+" begin:"+begin+" end"+end);
      	  if (begin<=end)
//	  if (posVB[i].position<posVA[j].position)
      	  {
//      Grep g = new Grep(line.substring(begin,end),begin);
      		Grep g = new Grep(line.substring(posVB[i].position,posVA[j].position + (posVA[j].value).length()),posVB[i].position);
      		grepPV.addElement(g);
      	  }
     	 }
     	Grep [] posV= new Grep[grepPV.size()];
     	grepPV.copyInto(posV);                     

     	return posV;		
     }
     else if (dotWCard.equals(".+"))
     {
	patternBefore = pattern.substring(0,dotPlus);
    	patternAfter = pattern.substring(dotPlus+2);
    	Grep[] posVB = positionValue(line,patternBefore);
    	Grep[] posVA = positionValue(line,patternAfter);
    	//Grep[] posProd = new Grep[posVB.length*posVA.length];
    	int k=0;
    	for (int i = 0; i<posVB.length; i++)
     	 for (int j = 0; j<posVA.length; j++)
     	 {
//      int begin = posVB[i].position;
//      int end = posVA[j].position + (posVA[j].value).length();
      	  int begin = posVB[i].position+posVB[i].value.length();
      	  int end = posVA[j].position;
      	  //System.out.println("line is "+line+"posVB[i].value"+posVB[i].value+" begin:"+begin+" end"+end);
      	  if (begin<end)//at least one
//	  if (posVB[i].position<posVA[j].position)
      	  {
//      Grep g = new Grep(line.substring(begin,end),begin);
      		Grep g = new Grep(line.substring(posVB[i].position,posVA[j].position + (posVA[j].value).length()),posVB[i].position);
      		grepPV.addElement(g);
      	  }
     	 }
     	Grep [] posV= new Grep[grepPV.size()];
     	grepPV.copyInto(posV);                     

     	return posV;		
     }
     else//.?
     {
	patternBefore = pattern.substring(0,dotQ);
    	patternAfter = pattern.substring(dotQ+2);
    	Grep[] posVB = positionValue(line,patternBefore);
    	Grep[] posVA = positionValue(line,patternAfter);
    	//Grep[] posProd = new Grep[posVB.length*posVA.length];
    	int k=0;
    	for (int i = 0; i<posVB.length; i++)
     	 for (int j = 0; j<posVA.length; j++)
     	 {
//      int begin = posVB[i].position;
//      int end = posVA[j].position + (posVA[j].value).length();
      	  int begin = posVB[i].position+posVB[i].value.length();
      	  int end = posVA[j].position;
      	  //System.out.println("line is "+line+"posVB[i].value"+posVB[i].value+" begin:"+begin+" end"+end);
      	  if (patternAfter.equals("$"))
	  {
	   if (begin==end||begin==end+1)//0 or 1
//	  if (posVB[i].position<posVA[j].position)
      	   {
//      Grep g = new Grep(line.substring(begin,end),begin);
      		Grep g = new Grep(line.substring(posVB[i].position,posVA[j].position + (posVA[j].value).length()),posVB[i].position);
      		grepPV.addElement(g);
      	   }	
	  }
	  else if (begin==end||begin==end-1)//0 or 1
//	  if (posVB[i].position<posVA[j].position)
      	  {
//      Grep g = new Grep(line.substring(begin,end),begin);
      		Grep g = new Grep(line.substring(posVB[i].position,posVA[j].position + (posVA[j].value).length()),posVB[i].position);
      		grepPV.addElement(g);
      	  }
     	 }
     	Grep [] posV= new Grep[grepPV.size()];
     	grepPV.copyInto(posV);                     

     	return posV;		
     }
    }//else -- no .x, .+, .?
    int star = pattern.indexOf("*");
    int lastStar = pattern.lastIndexOf("*");
    if (star==0) //at the beginning of the pattern, for example: *l
	  throw new InterpretError("invalid syntax of pattern:* cannot be at the beginning of a pattern");
    //System.out.println("**********"+lastStar);
    if (lastStar==pattern.length()-1) //at the end of the pattern, like l*
    {//System.out.println("**********"+lastStar);
     String beforeStar = pattern.substring(pattern.length()-2,pattern.length()-1);
     String beforeChars = beforeStar;
     String beforeCharsP = beforeStar;
     //System.out.println("**beforeStar********"+beforeStar);
     int p = pattern.length()-2;
     if (beforeStar.equals("]"))	//for example: [ho]*
     {
      //System.out.println("**beforeStar!here"+beforeStar);
      p = pattern.length()-3;
      while ((p>0)&&!pattern.substring(p,p+1).equals("["))
      { p--;}
      if (pattern.substring(p,p+1).equals("[")) 
      {
      	beforeChars = pattern.substring(p+1,pattern.length()-2);//ho
      	beforeCharsP = pattern.substring(p,pattern.length()-1);//[ho]
      }
      else throw new InterpretError("invalid syntax of pattern: no match for ]");
     }
     else if (beforeStar.equals(")"))	//for example: (he)*
     {
      p = pattern.length()-3;
      while ((p>0)&&!pattern.substring(p,p+1).equals("("))
      { p--;}
      if (pattern.substring(p,p+1).equals("(")) 
      {
      	beforeChars = pattern.substring(p+1,pattern.length()-2);//ho
      	beforeCharsP = pattern.substring(p,pattern.length()-1);//(ho)
      }
      else throw new InterpretError("invalid syntax of pattern: no match for ]");
     }
     
     Vector grepPVp= new Vector();
     patternBefore = pattern.substring(0,p);
     //System.out.println("***********************!!!!patternBefore"+patternBefore+";line:"+line);
     if (patternBefore.length()!=0)
     {
//System.out.println("!!!in here!!!");         
     Grep[] posV = positionValue(line,patternBefore);
//     switch (beforeStar)
//     {
      //case "]":
     if (beforeStar.equals("]"))
     {
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
//      System.out.println("************pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value+"beforeChars:"+beforeChars);
      String x = line.substring(pos+valL,pos+valL+1);
      int px = pos+valL;
      String nx = x;
      int indxx = beforeChars.indexOf(x);	//to see if x is in [ho]	Yu --Oct 29,2004      
      if (indxx>=0)
      {
      	while (px<line.length() && nx.equals(x))
	{
	 px++;
	 nx = line.substring(px,px+1);
	}
	posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }
      }
//      break;
     }
     else  if (beforeStar.equals(")"))
     {
      //case ")":
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      int px = pos+valL+beforeChars.length();
      String x = line.substring(pos+valL,px);
      if (x.equals(beforeChars))
      {
      	while (px<line.length() && x.equals(beforeChars))
	{//System.out.println("px,line"+px+","+line);
	 if (px+beforeChars.length()<=line.length())
	  x = line.substring(px,px+beforeChars.length());
	 px=px+beforeChars.length(); 
	}
	if (x.equals(beforeChars)&&px+beforeChars.length()<=line.length())
	 posV[i].value = line.substring(pos,px);
	else 
	 posV[i].value = line.substring(pos,px-beforeChars.length());
	grepPVp.addElement(posV[i]);
      }
      else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }
      }
     }
     else
     {//System.out.println("!!!here!!!!!!beforestar:"+beforeStar+";"+posV.length);
//      break;
      //default://no [,( just one char
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
//      System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      if (pos+valL==line.length())
       grepPVp.addElement(posV[i]);
      else
      {
      String x = line.substring(pos+valL,pos+valL+1);
      int px = pos+valL;
      String nx = x;
      int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
      if (indxx>=0)
      {//System.out.println("px,x"+px+";"+x);
      	while (px<line.length() && nx.equals(x))
	{
	 px++;
	 if (px<line.length())
	  nx = line.substring(px,px+1);
//	 System.out.println("nx:"+nx+";px:"+px);
	}
//	System.out.println("afterpx,x"+px+";"+x);
	posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }
      }
      }
     } 
//      break;
//     }//switch
     }//if (patternBefore.length()!=0)
     else
     {
//     Grep[] posV = positionValueN(line,pattern);
  //   return posV;
//System.out.println("patternBefore:"+patternBefore);
     //Grep[] posV = positionValueN(line,pattern);
      Grep[] posV = positionValue(line,patternBefore);//modified by Yu Nov 9, 2004 !!!! example: grep "[lm]?" in R;
  //    System.out.println("posV:"+posV.length+"beforeChars:"+beforeCharsP);
      Grep[] posVb = positionValue(line,beforeCharsP);
    //  System.out.println("posV:"+posV.length+"beforeChars:"+beforeCharsP+"posVb:"+posVb.length);
      if (posVb.length==0)
      {
       for (int i=0; i<posV.length;i++)
       	grepPVp.addElement(posV[i]);
      }
      else
      {
       if (beforeStar.equals("]"))
       {
       for (int i=0; i<posVb.length;i++)
       {
        int px = posVb[i].position+posVb[i].value.length();
        if (px==line.length())
         grepPVp.addElement(posVb[i]);
        else
        {
        String x = line.substring(px,px+1);
        String nx = x;
        int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
        if (indxx>=0)
        {
         while (px<line.length() && nx.equals(x))
	 {
	  px++;
	  nx = line.substring(px,px+1);
	 }
	  posVb[i].value = line.substring(posVb[i].position,px);
        }
      	grepPVp.addElement(posVb[i]);
	}	
       }
       }
       else if (beforeStar.equals(")"))
       {
       for (int i=0; i<posVb.length;i++)
       {
        int pos = posVb[i].position;
        int valL = posVb[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
        if (pos+valL==line.length())
         grepPVp.addElement(posVb[i]);
        else
        {
        int px = pos+valL+beforeChars.length();
        String x = line.substring(pos+valL,px);
        if (x.equals(beforeChars))
        {
      	 while (px<line.length() && x.equals(beforeChars))
	 {//System.out.println("px,line"+px+","+line);
	  if (px+beforeChars.length()<=line.length())
	   x = line.substring(px,px+beforeChars.length());
	  px=px+beforeChars.length(); 
	 }
	 if (x.equals(beforeChars)&&px+beforeChars.length()<=line.length())
	  posVb[i].value = line.substring(pos,px);
	 else 
	  posVb[i].value = line.substring(pos,px-beforeChars.length());
        }
	grepPVp.addElement(posVb[i]);
	}
       }
       }
       else
       {
       for (int i=0; i<posVb.length;i++)
       {
        int pos = posVb[i].position;
        int valL = posVb[i].value.length();
      //posV[i].value = line.substring(pos);
        //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
        if (pos+valL==line.length())
         grepPVp.addElement(posVb[i]);
        else
        {
         String x = line.substring(pos+valL,pos+valL+1);
         int px = pos+valL;
         String nx = x;
         int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
         if (indxx>=0)
         {//System.out.println("px,x"+px+";"+x);
      	  while (px<line.length() && nx.equals(x))
	  {
	    px++;
	    if (px<line.length())
	    nx = line.substring(px,px+1);
	 //System.out.println("nx:"+nx+";px:"+px);
	  }
	  //System.out.println("afterpx,x"+px+";"+x);
	  posVb[i].value = line.substring(pos,px);
         }
         grepPVp.addElement(posVb[i]);
        }//else
	}
        }
      }
     }
     //}
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;          
    }
//+
    int plus = pattern.indexOf("+");
    int lastPlus = pattern.lastIndexOf("+");
    if (plus==0) //at the beginning of the pattern, for example: *l
	  throw new InterpretError("invalid syntax of pattern:+ cannot be at the beginning of a pattern");
    if (lastPlus==pattern.length()-1) //at the end of the pattern, like l+
    {
     String beforePlus = pattern.substring(pattern.length()-2,pattern.length()-1);
     String beforeChars = beforePlus;
     int p = pattern.length()-2;
     if (beforePlus.equals("]"))	//for example: [ho]*
     {
      p = pattern.length()-3;
      while ((p>0)&&!pattern.substring(p,p+1).equals("["))
      { p--;}
      if (pattern.substring(p,p+1).equals("[")) 
      {
      	beforeChars = pattern.substring(p+1,pattern.length()-2);//ho
      }
      else throw new InterpretError("invalid syntax of pattern: no match for ]");
     }
     else if (beforePlus.equals(")"))	//for example: (he)*
     {
      p = pattern.length()-3;
      while ((p>0)&&!pattern.substring(p,p+1).equals("("))
      { p--;}
      if (pattern.substring(p,p+1).equals("(")) 
      {
      	beforeChars = pattern.substring(p+1,pattern.length()-2);//ho
      }
      else throw new InterpretError("invalid syntax of pattern: no match for ]");
     }
     
     Vector grepPVp= new Vector();
     patternBefore = pattern.substring(0,p);
     if (patternBefore.length()!=0)
     {
     Grep[] posV = positionValue(line,patternBefore);
//     switch (beforeStar)
//     {
      //case "]":
     if (beforePlus.equals("]"))
     {
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      String x = line.substring(pos+valL,pos+valL+1);
      int px = pos+valL;
      String nx = x;
      int indxx = beforeChars.indexOf(x);	//to see if x is in [ho]	Yu --Oct 29,2004      
      if (indxx>=0)
      {
      	while (px<line.length() && nx.equals(x))
	{
	 px++;
	 nx = line.substring(px,px+1);
	}
	posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      /*else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }*/
      }
//      break;
     }
     else  if (beforePlus.equals(")"))
     {
      //case ")":
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      int px = pos+valL+beforeChars.length();
      String x = line.substring(pos+valL,px);
      if (x.equals(beforeChars))
      {
      	while (px+beforeChars.length()<line.length() && x.equals(beforeChars))
	{System.out.println(px+";"+beforeChars+";"+line);
	 x = line.substring(px,px+beforeChars.length());
	 px=px+beforeChars.length();
	}
	if (px+beforeChars.length()<line.length())
	 posV[i].value = line.substring(pos,px-beforeChars.length());
	else 
	 posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      /*else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }*/
      }
     }
     else
     {
//      break;
      //default://no [,( just one char
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
//      System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      if (pos+valL==line.length())
       grepPVp.addElement(posV[i]);
      else
      {
      String x = line.substring(pos+valL,pos+valL+1);
      int px = pos+valL;
      String nx = x;
      int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
      if (indxx>=0)
      {
      	while (px<line.length() && nx.equals(x))
	{
	 px++;
	 if (px<line.length())
	  nx = line.substring(px,px+1);
	}
	posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      /*else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }*/
      }
      }
     } 
//      break;
//     }//switch
     }
     else
     {
      Grep [] posV = positionValueN(line,pattern);
       if (beforePlus.equals("]"))
       {
       for (int i=0; i<posV.length;i++)
       {
        int px = posV[i].position+posV[i].value.length();
        if (px==line.length())
         grepPVp.addElement(posV[i]);
        else
        {
        String x = line.substring(px,px+1);
        String nx = x;
        int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
        if (indxx>=0)
        {
         while (px<line.length() && nx.equals(x))
	 {
	  px++;
	  nx = line.substring(px,px+1);
	 }
	  posV[i].value = line.substring(posV[i].position,px);
        }
      	grepPVp.addElement(posV[i]);
	}	
       }
       }
       else if (beforePlus.equals(")"))
       {
       for (int i=0; i<posV.length;i++)
       {
        int pos = posV[i].position;
        int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
        if (pos+valL==line.length())
         grepPVp.addElement(posV[i]);
        else
        {
        int px = pos+valL+beforeChars.length();
        String x = line.substring(pos+valL,px);
        if (x.equals(beforeChars))
        {
      	 while (px<line.length() && x.equals(beforeChars))
	 {//System.out.println("px,line"+px+","+line);
	  if (px+beforeChars.length()<=line.length())
	   x = line.substring(px,px+beforeChars.length());
	  px=px+beforeChars.length(); 
	 }
	 if (x.equals(beforeChars)&&px+beforeChars.length()<=line.length())
	  posV[i].value = line.substring(pos,px);
	 else 
	  posV[i].value = line.substring(pos,px-beforeChars.length());
        }
	grepPVp.addElement(posV[i]);
	}
       }
       }
       else
       {
       for (int i=0; i<posV.length;i++)
       {
        int pos = posV[i].position;
        int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
        //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
        if (pos+valL==line.length())
         grepPVp.addElement(posV[i]);
        else
        {
         String x = line.substring(pos+valL,pos+valL+1);
         int px = pos+valL;
         String nx = x;
         int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
         if (indxx>=0)
         {//System.out.println("px,x"+px+";"+x);
      	  while (px<line.length() && nx.equals(x))
	  {
	    px++;
	    if (px<line.length())
	    nx = line.substring(px,px+1);
	 //System.out.println("nx:"+nx+";px:"+px);
	  }
	  //System.out.println("afterpx,x"+px+";"+x);
	  posV[i].value = line.substring(pos,px);
         }
         grepPVp.addElement(posV[i]);
        }//else
	}
        }
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     
     return posVn;
     }
     
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;          
    }

//?
    int qMark = pattern.indexOf("?");
    int lastQ = pattern.lastIndexOf("?");
    if (qMark==0) //at the beginning of the pattern, for example: *l
	  throw new InterpretError("invalid syntax of pattern:? cannot be at the beginning of a pattern");
    //System.out.println("**********"+lastStar);
    if (lastQ==pattern.length()-1) //at the end of the pattern, like l*
    {
     String beforeQ = pattern.substring(pattern.length()-2,pattern.length()-1);
     String beforeChars = beforeQ;
     String beforeCharsP = beforeQ;
     //System.out.println("**beforeStar********"+beforeQ);
     int p = pattern.length()-2;
     if (beforeQ.equals("]"))	//for example: [ho]*
     {
      //System.out.println("**beforeStar!here"+beforeQ);
      p = pattern.length()-3;
      while ((p>0)&&!pattern.substring(p,p+1).equals("["))
      { p--;}
      if (pattern.substring(p,p+1).equals("[")) 
      {
      	beforeChars = pattern.substring(p+1,pattern.length()-2);//ho
	beforeCharsP = pattern.substring(p,pattern.length()-1);//[ho]
      }
      else throw new InterpretError("invalid syntax of pattern: no match for ]");
     }
     else if (beforeQ.equals(")"))	//for example: (he)*
     {
      p = pattern.length()-3;
      while ((p>0)&&!pattern.substring(p,p+1).equals("("))
      { p--;}
      if (pattern.substring(p,p+1).equals("(")) 
      {
      	beforeChars = pattern.substring(p+1,pattern.length()-2);//ho
	beforeCharsP = pattern.substring(p,pattern.length()-1);//(ho)
      }
      else throw new InterpretError("invalid syntax of pattern: no match for ]");
     }
     
     Vector grepPVp= new Vector();
     patternBefore = pattern.substring(0,p);
     // System.out.println("patternBefore:"+patternBefore+";"+patternBefore.length());
     if (patternBefore.length()!=0)
     {
     Grep[] posV = positionValue(line,patternBefore);
//      System.out.println("************posV:"+posV.length);
//     switch (beforeStar)
//     {
      //case "]":
     if (beforeQ.equals("]"))
     {
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      String x = line.substring(pos+valL,pos+valL+1);
//      System.out.println("*pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value+"beforeChars:"+beforeChars+"x:"+x);
      int px = pos+valL;
      String nx = x;
      int indxx = beforeChars.indexOf(x);	//to see if x is in [ho]	Yu --Oct 29,2004      
      if (indxx>=0)
      {
      	//while (px<line.length() && nx.equals(x))
	//{
	 px++;
	 //nx = line.substring(px,px+1);
	//}
	posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }	//comment out by Yu Nov 9,2004 because 
      }
//      break;
     }
     else  if (beforeQ.equals(")"))
     {
      //case ")":
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      int px = pos+valL+beforeChars.length();
      String x = line.substring(pos+valL,px);
      if (x.equals(beforeChars))
      {
      	//while (px<line.length() && x.equals(beforeChars))
//	System.out.println("px,line"+px+","+line);
	 //if (px+beforeChars.length()<=line.length())
	  //x = line.substring(px,px+beforeChars.length());
	 //px=px+beforeChars.length(); 
	//}
	//if (x.equals(beforeChars)&&px+beforeChars.length()<=line.length())
	 posV[i].value = line.substring(pos,px);
	//else 
	 //posV[i].value = line.substring(pos,px-beforeChars.length());
	grepPVp.addElement(posV[i]);
      }
      else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }
      }
     }
     else
     {
//      break;
      //default://no [,( just one char
      for (int i = 0;i<posV.length;i++)
      {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      //posV[i].value = line.substring(pos);
      //System.out.println("pos,valL:"+pos+";"+valL+"posV[i].value"+posV[i].value);
      if (pos+valL==line.length())
       grepPVp.addElement(posV[i]);
      else
      {
      String x = line.substring(pos+valL,pos+valL+1);
      int px = pos+valL;
      String nx = x;
      int indxx = beforeChars.indexOf(x);	//to see if x is same as x	Yu --Oct 29,2004      
      if (indxx>=0)
      {//System.out.println("px,x"+px+";"+x);
      	//while (px<line.length() && nx.equals(x))
	//{
	 px++;
	 //if (px<line.length())
	  //nx = line.substring(px,px+1);
	 //System.out.println("nx:"+nx+";px:"+px);
	//}
	//System.out.println("afterpx,x"+px+";"+x);
	posV[i].value = line.substring(pos,px);
	grepPVp.addElement(posV[i]);
      }
      else	//not in [ho]
      {
       grepPVp.addElement(posV[i]);
      }
      }
      }
     }
     }//  if (patternBefore.length()!=0)
//      break;
//     }//switch
     else
     {
//System.out.println("patternBefore:"+patternBefore);
     //Grep[] posV = positionValueN(line,pattern);
      Grep[] posV = positionValue(line,patternBefore);//modified by Yu Nov 9, 2004 !!!! example: grep "[lm]?" in R;
//      System.out.println("posV:"+posV.length+"beforeChars:"+beforeCharsP);
      Grep[] posVb = positionValue(line,beforeCharsP);
//      System.out.println("posV:"+posV.length+"beforeChars:"+beforeCharsP+"posVb:"+posVb.length);
      if (posVb.length==0)
      {
       for (int i=0; i<posV.length;i++)
       	grepPVp.addElement(posV[i]);
      }
      else
      {
       for (int i=0; i<posVb.length;i++)
       	grepPVp.addElement(posVb[i]);
      }

     } //modified by Yu Nov 9, 2004
     
     Grep [] posVn= new Grep[grepPVp.size()];
     grepPVp.copyInto(posVn);                     
     return posVn;          
    }
     return positionValueN(line,pattern);		
   }


/**
@author Yu Gu Oct 25, 2004
@param Sting line: search line
@param String pattern: search pattern in line
@return Grep[]
Grep a paticular pattern from a string, return array of Grep.
Deal with .*
*/
  public Grep[] positionValuePlus(String line, String pattern)                  //Oct 25,2004 by Yu
  throws InterpretError
  {
    int asci = 1;
    String aChar = new Character((char)asci).toString();
    Vector grepPV= new Vector();
    int dotStar = pattern.indexOf(".+");
    int lastDotStar = pattern.lastIndexOf(".+");
    String patternBefore = "";
    String patternAfter = "";
    //System.out.println("line is "+line+" dotStart at:"+dotStar);
    if (dotStar==0) //at the beginning of the pattern, for example: .*l
    {
     //System.out.println("deal with it later"); Sep 15, 2004
     patternAfter = pattern.substring(2);
     Grep[] posV = positionValue(line,patternAfter);
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      posV[i].value = line.substring(0,pos+valL);
      posV[i].position = 0;
     }
     return posV;     
    }
    if (lastDotStar==pattern.length()-2) //at the end of the pattern, like l.*
    {
     patternBefore = pattern.substring(0,pattern.length()-2);
     Grep[] posV = positionValue(line,patternBefore);
     for (int i = 0;i<posV.length;i++)
     {
      int pos = posV[i].position;
      int valL = posV[i].value.length();
      posV[i].value = line.substring(pos);
     }
     return posV;          
    }
    patternBefore = pattern.substring(0,dotStar);
    patternAfter = pattern.substring(dotStar+2);
    Grep[] posVB = positionValue(line,patternBefore);
    Grep[] posVA = positionValue(line,patternAfter);
    Grep[] posProd = new Grep[posVB.length*posVA.length];
    int k=0;
    for (int i = 0; i<posVB.length; i++)
     for (int j = 0; j<posVA.length; j++)
     {
//      int begin = posVB[i].position;
//      int end = posVA[j].position + (posVA[j].value).length();
      //System.out.println("line is "+line+" begin:"+begin+" end"+end);
      int begin = posVB[i].position+posVB[i].value.length();
      int end = posVA[j].position;
      if (begin<=end)
	  if (posVB[i].position<posVA[j].position)
      {
      Grep g = new Grep(line.substring(posVB[i].position,posVA[j].position + (posVA[j].value).length()),posVB[i].position);
      grepPV.addElement(g);
      }
     }
     Grep [] posV= new Grep[grepPV.size()];
     grepPV.copyInto(posV);                     

     return posV;		
   }

   // deal with the situation when there are parameters in between the pattern
  public Grep[] positionValuePS(String line, String pattern,String[] domN, Domain[] param)         //Sep 23,2004 by Yu
  throws InterpretError
  {
    //System.out.println("in positionValuePS:"+line+";"+pattern+";" +domN[0]);
    String patternBefore = "";
    String patternAfter = "";
    String paV[] = new String[param.length];
    int ind=0;  //the position of the parameter being dealt with
    for (int i=0;i<param.length;i++)
    {
     if (param[i].name.equals(domN[0]))
      ind = i;
    }
    Vector grepPV= new Vector();
    if (domN.length==0) return positionValue(line,pattern);
    String para = "\\"+domN[0];
    int dotStar = pattern.indexOf(para);
    //System.out.println("line is "+line+" dotStart at:"+dotStar);
    patternBefore = pattern.substring(0,dotStar);
    patternAfter = pattern.substring(dotStar+para.length());
    if (domN.length==1)
    {
    Grep[] posVB = positionValue(line,patternBefore);
    Grep[] posVA = positionValue(line,patternAfter);
    for (int i = 0; i<posVB.length; i++)
     for (int j = 0; j<posVA.length; j++)
     {
      int begin = posVB[i].position+posVB[i].value.length();
      int end = posVA[j].position;
      //System.out.println("line is "+line+" begin:"+begin+" end"+end);
      if (begin<=end)
//	  if (posVB[i].position<posVA[j].position)
      {
       String paraV = line.substring(begin,end);	//value of the parameter
       paV[ind]= new String(paraV);
       Grep g = new Grep(line.substring(posVB[i].position,end+posVA[j].value.length()),posVB[i].position,paV);
       grepPV.addElement(g);
      }
     }
     Grep [] posV= new Grep[grepPV.size()];
     grepPV.copyInto(posV);                     
     return posV;
    }
    else //domN.length>1
    {
    Grep[] posVB = positionValue(line,patternBefore);
    String paraN[] = new String[domN.length-1];
    for (int i = 1; i<domN.length; i++)
     paraN[i-1]=domN[i];
    Grep[] posVA = positionValuePS(line,patternAfter,paraN,param);
    for (int i = 0; i<posVB.length; i++)
     for (int j = 0; j<posVA.length; j++)
     {
      int begin = posVB[i].position+posVB[i].value.length();
      int end = posVA[j].position;
      //System.out.println("line is "+line+" begin:"+begin+" end"+end);
      if (begin<=end)
      {
       String paraV = line.substring(begin,end);
       for(int k=0;k<posVA[j].paraValue.length;k++)
       if (posVA[j].paraValue[k]!=null)
	paV[k]=posVA[j].paraValue[k];
       paV[ind]= new String(paraV); //set for position ind the value of parameter
       Grep g = new Grep(line.substring(posVB[i].position,end+posVA[j].value.length()),posVB[i].position,paV);
       grepPV.addElement(g);
      }
     }
     Grep [] posV= new Grep[grepPV.size()];
     grepPV.copyInto(posV);                     

     return posV;		
     }
   }
      
 public int backwards(String s, String pattern)    //July 14,2004 by Yu
 {
		int start = s.length()-1;
		int length = 1;
//System.out.println("start at beginning:"+start+" and string is "+s+" and pattern is "+pattern);
    while (start >= 0)
		{
			String ss = s.substring(start,start+length);
      String f="";
	    try {
			       f = saveToAFile(ss);		//save s to file f
            	}
      catch (IOException e) {
                System.err.println(e);
            	}
//      String cmd="egrep "+pattern+" "+f;
	   String[] cmd={"egrep",pattern,f};
	    //System.out.println(cmd+" and in backwards string is:"+ss);
      try {
					ExecComm command=new ExecComm();
					String outlist[] = command.runCommand(cmd);
		  if (outlist.length>0)                         //return the position
			{
        return start;
			}
      else
      {
       	start--;
        length++;
      }
					}//try
   		catch (IOException e) {
                System.err.println(e);
            	}
		}  //while
		return -1;
	 }

   public String saveToAFile(String s)          //July 14,2004 by Yu
   throws IOException
   {
		String fname="_temp_.txt";
		File outputFile = new File(fname);

    FileWriter out = new FileWriter(outputFile);
	  out.write(s);

    out.close();
    return fname;
   }

	 public int getOpcode(int type)	//July 14,2004 by Yu
	 {
					 int opcode;
           switch(type) {
           case SHORT:
             opcode = OP_SHORT;
           break;
           case INTEGER:
             opcode = OP_INTEGER;
           break;
           case LONG:
             opcode = OP_LONG;
           break;
           case FLOAT:
             opcode = OP_FLOAT;
           break;
           case DOUBLE:
             opcode = OP_DOUBLE;
           break;
           case NUMERIC:
             opcode = OP_NUMERIC;
           break;
					 default:
             opcode = OP_STRING;
					 }//switch
					 return opcode;      
   }

            //jjtThis.setObject(new Integer(s));
 public Object getInfoFromString(String s, int type)	//July 14,2004 by Yu
 {
           switch(type) {
           case SHORT:
             return new Short(s);
           case INTEGER:
             return new Integer(s);
           case LONG:
             return new Long(s);
           case FLOAT:
             return new Float(s);
           case DOUBLE:
             return new Double(s);
           case NUMERIC:
						 return new number(s);
					 default:
             return s;
					 }//switch
 }

 public String extractPattern(Relation r)	//July 15,2004 by Yu
 {//pattern = extractPattern(grepR);
  String[] patternArray = new String[r.numattrs*r.numtuples];
  for (int ij = 0; ij < r.numattrs ; ij++)
  {
  	if (r.data==null)
		try {r.load();}
		catch (InterpretError e) {
			Global.processThrowable(e);  }
					 	
        switch(r.domains[ij].type) {              //s.valueOf(((int[])data[j])[i]
           case SHORT:
             short[] pAS = new short[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAS, 0, r.numtuples);
      		 	 for (int i = 0; i<r.numtuples; i++)
             {
              Short aS = new Short(pAS[i]);
			 				patternArray[ij*r.numtuples+i] = aS.toString();
             }
             break;
           case INTEGER:
             int[] pAI = new int[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAI, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Integer aI = new Integer(pAI[i]);
			 				patternArray[ij*r.numtuples+i] = aI.toString();
             }
             break;
           case LONG:
             long[] pAL = new long[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAL, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Long aS = new Long(pAL[i]);
			 				patternArray[ij*r.numtuples+i] = aS.toString();
             }
             break;
           case FLOAT:
             float[] pAF = new float[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAF, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Float aF = new Float(pAF[i]);
			 				patternArray[ij*r.numtuples+i] = aF.toString();
             }
             break;
           case DOUBLE:
             double[] pAD = new double[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAD, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Double aD = new Double(pAD[i]);
			 				patternArray[ij*r.numtuples+i] = aD.toString();
             }
             break;
           case NUMERIC:
             number[] pAN = new number[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAN, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
			 				patternArray[ij*r.numtuples+i] = pAN[i].toString();
             break;
	   default:
             String[] pASt = new String[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pASt, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
			 				patternArray[ij*r.numtuples+i] = pASt[i];
					 }//switch

	  }
    String pattern = patternArray[0];
    for (int i = 1; i<patternArray.length; i++)
               pattern = pattern + "|" + patternArray[i];
    return pattern;
   }

 public String[] extractWhere(Relation r)	//Sep 19,2004 by Yu
 {//pattern = extractPattern(grepR);
  String[] patternArray = new String[r.numattrs*r.numtuples];
  for (int ij = 0; ij < r.numattrs ; ij++)
  {
  	if (r.data==null)
		try {r.load();}
		catch (InterpretError e) {
			Global.processThrowable(e);  }
					 	
        switch(r.domains[ij].type) {              //s.valueOf(((int[])data[j])[i]
           case SHORT:
             short[] pAS = new short[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAS, 0, r.numtuples);
      		 	 for (int i = 0; i<r.numtuples; i++)
             {
              Short aS = new Short(pAS[i]);
			 				patternArray[ij*r.numtuples+i] = aS.toString();
             }
             break;
           case INTEGER:
             int[] pAI = new int[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAI, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Integer aI = new Integer(pAI[i]);
			 				patternArray[ij*r.numtuples+i] = aI.toString();
             }
             break;
           case LONG:
             long[] pAL = new long[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAL, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Long aS = new Long(pAL[i]);
			 				patternArray[ij*r.numtuples+i] = aS.toString();
             }
             break;
           case FLOAT:
             float[] pAF = new float[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAF, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Float aF = new Float(pAF[i]);
			 				patternArray[ij*r.numtuples+i] = aF.toString();
             }
             break;
           case DOUBLE:
             double[] pAD = new double[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAD, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
             {
              Double aD = new Double(pAD[i]);
			 				patternArray[ij*r.numtuples+i] = aD.toString();
             }
             break;
           case NUMERIC:
             number[] pAN = new number[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pAN, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
			 				patternArray[ij*r.numtuples+i] = pAN[i].toString();
             break;
	   default:
             String[] pASt = new String[r.numtuples];
						 System.arraycopy(r.data[ij], 0, pASt, 0, r.numtuples);
      			 for (int i = 0; i<r.numtuples; i++)
			 				patternArray[ij*r.numtuples+i] = pASt[i];
					 }//switch

	  }
    return patternArray;
   }
  
  public String getType(int type)          //by Yu
   {
   	String typeName="";
   	switch (type) 
   	{
   		case INTEGER:
   		typeName="integer";
   		break;
   		case BOOLEAN:
   		typeName="boolean";
   		break;
   		case SHORT:
   		typeName="short";
   		break;
   		case LONG:
   		typeName="long";
   		break;
   		case FLOAT:
   		typeName="float";
   		break;
   		case DOUBLE:
   		typeName="double";
   		break;
   		case NUMERIC:
   		typeName="numeric";
   		break;
   		case STRING:
   		typeName="string";
   		break;
   		case TEXT:
   		typeName="text";
   		break;
   		case ATTRIBUTE:
   		typeName="attribute";
   		break;
   		case UNIVERSAL:
   		typeName="universal";
   		break;
   		case TYPE:
   		typeName="type";
   		break;
   	}
   	 return typeName;	
   }
  /******************** end of Yu's new supporting methods ***********************/

 	
  /******************** start Fan's new supporting methods ***********************/
  /*
   * This method is for new version of transpose operator
   * 
   */ 
  public Relation evaluateTranspose(SimpleNode node, String filename,
  		 Hashtable htable, Environment env)    //July 30, 2004 Fan Guo
  	       throws InterpretError
  { 
  	  int i;
  	  SimpleNode IdlistNode = (SimpleNode) node.jjtGetChild(0);	  	 	  
  	  Relation srcrel = destrel;   	  
  	  Domain[] domsList = ExpressionListToDomains(IdlistNode, srcrel, htable, env); // domain list without .id
  	  Domain[] fdomsList = new Domain [domsList.length+1];  				// domain list with .id
  	  fdomsList[0]=env.lookupDom(".id", true);
  	  for( i = 0; i < domsList.length; i++){
  	  	fdomsList[i+1]=domsList[i];
  	  }
  	  
  	  String tempname=nextTempName();	 
  	    
  	  // prepare the relation that to be returned
  	  Relation cellRel = new Relation(tempname, domsList, 0, domsList.length, 0, env); // without ".id"
  	    
  	  // get a surrogate
  	  long s= Surrogate.next();  
  	  Domain[] ScalarDomL = srcrel.getScalarAttribute();  
  	  
  	  // get the position of each domains that the transpose operator works on in the new actualized relation
  	  int pos[] = new int[ScalarDomL.length];
  	  for (i=0;i<ScalarDomL.length; i++) {
  	  	for (int j=0;j<srcrel.domains.length; j++){
  	  		if (ScalarDomL[i].name.equals(srcrel.domains[j].name)) {
  	  			pos[i] = j;
  	  			break;
  	  		}
  	  	}  	  		
  	  }  	  
  	  
  	  // get the data in current tuple; 
  	  int numoftuple = ScalarDomL.length; 
  	  Object[] CellRelDataArray = cellRel.createDataArray(domsList, numoftuple);
  	  
  	  //Build the "transpose" nested relation 
  	  for(int k=0;k<fdomsList.length;k++){  // for each attribute in dot relation
  	  	for(i=0; i<ScalarDomL.length; i++) {
  	  			switch(fdomsList[k].type){ 
  	  				case ATTRIBUTE:
  	  					((String[]) CellRelDataArray[k-1])[i]=srcrel.domains[pos[i]].name;
  	  					break;
  	  				case TYPE:
  	  					((String[]) CellRelDataArray[k-1])[i]=Utility.type2String(srcrel.domains[pos[i]].type);
  	  					break;
  	  				case UNIVERSAL:
  	  					String str = new String();
  	  					Object b = srcrel.data[pos[i]];                               
  	  					Object a = java.lang.reflect.Array.get(b,currow);              
  	  					str = a.toString(); 
  	  					
  	  					int tp = srcrel.domains[i].type;    
  	  					if (tp == BOOLEAN && str.equals("0")) str = "false";
  	  					if (tp == BOOLEAN && str.equals("1")) str = "true";  
  	  					
  	  					String tpstr = Utility.type2String(srcrel.domains[pos[i]].type);  	    					
  	  					if (!((str.startsWith("universal")||str.startsWith("anytype") ) && 
  	  						  (tpstr.equals("universal") ||tpstr.equals("anytype") )))   // recursive Universal in data is not blocked
  	  						str = tpstr + ':' + str;
  	  					
  	  					((String[]) CellRelDataArray[k-1])[pos[i]]=str;
  	  					break;          	
  	  				default:
  	  					break;
  	  			}
  	  	} 
  	  }  
  	  
  	  cellRel.data = CellRelDataArray;
  	  cellRel.numtuples=numoftuple; 
  	  
  	  deleteFile(tempname);
  	  
  	  return cellRel;
  }
  
  
  /*
   * This method is for relation operator
   * 
   */ 
  private Relation evaluateRelation(SimpleNode node, String filename,
  		 Hashtable htable, Environment env)	// June 29, 2004 Fan Guo
  	       throws InterpretError
  {
  	  Domain[] domsList=null;
  	  int i;
  	  
  	  Relation srcrel = destrel.duplicateLoadAll();  	  
  	  
  	  SimpleNode IdlistNode = (SimpleNode) node.jjtGetChild(0);
  	  domsList = ExpressionListToDomains(IdlistNode, srcrel, htable, env); // domain list without .id
  	  
  	  String tempname = nextTempName();
  	 
  	  // prepare the relation that to be returned
  	  Relation singletonRel = new Relation(tempname, domsList, 0, domsList.length, 0, env); // without ".id"
  	  
  	  //actualization: so that virtual domains (if any) is actualized
  	  Actualizer actualizer = new Actualizer(domsList, srcrel, env);
  	  Relation rtmp = actualizer.actualize();
  	  rtmp.myEnv = env;     
  	  
  	  // get the position of each domains that the relation operator works on in the new actualized relation
  	  int pos[] = new int[domsList.length];
  	  for (i=0;i<domsList.length; i++) {
  	  	for (int j=0;j<rtmp.domains.length; j++){
  	  		if (domsList[i].name.equals(rtmp.domains[j].name)) {
  	  			pos[i] = j;
  	  			break;
  	  		}
  	  	}  	  		
  	  } 
  	  
  	  // get the data in current tuple; 
  	  int numoftuple = 1;  // singleton
  	  Object[] SgtDataArray = singletonRel.createDataArray(domsList, numoftuple);

  	  for(int k=0;k<domsList.length;k++){  // for each attribute in dot relation
           	switch(domsList[k].type){            	
            	case NUMERIC:
            		((number[]) SgtDataArray[k])[0]=((number[])rtmp.data[pos[k]])[currow];                    
                    break;
                case INTEGER:
                	((int[]) SgtDataArray[k])[0]=((int[])rtmp.data[pos[k]])[currow];                    
                    break;
                case BOOLEAN:
                	((byte[]) SgtDataArray[k])[0]=((byte[])rtmp.data[pos[k]])[currow];                    
                    break;
                case IDLIST:
                	((long[]) SgtDataArray[k])[0]=((long[])rtmp.data[pos[k]])[currow];            
                    break;
                case SHORT:
                	((short[]) SgtDataArray[k])[0]=((short[])rtmp.data[pos[k]])[currow];                    
                    break;
                case LONG:
                	((long[]) SgtDataArray[k])[0]=((long[])rtmp.data[pos[k]])[currow];                    
                    break;
                case FLOAT:
                	((float[]) SgtDataArray[k])[0]=((float[])rtmp.data[pos[k]])[currow];                    
                    break;
                case DOUBLE:
                	((double[]) SgtDataArray[k])[0]=((double[])rtmp.data[pos[k]])[currow];                    
                    break;
                case UNION:
                case UNIVERSAL: 
                case ATTRIBUTE: 
                case TYPE: 	
                case STRING:
                case STMT:
                case EXPR:
                case TEXT:
                	((String[]) SgtDataArray[k])[0]=((String[])rtmp.data[pos[k]])[currow];                    
                    break;
            	default:
            		break;
            }
      }            
      
  	  singletonRel.data = SgtDataArray;
  	  singletonRel.numtuples=1; 
  	  actualizer.cleanup();
  	  deleteFile(tempname);

  	  return singletonRel;
  }
  
  /*    
   * this method is for level lifting of intermediate relation with vertical operation,
   * used for evaluateJoin() method
   * e.g. let A be relation(a) ijoin relation(b);
   * let B be relation (c) ijoin [red ujoin of A] in R1 ;
   * R<-[red ujoin of B] in R2;
   * here when actualizing "[red ujoin of A] in R1" part in the second statement, 
   * we need a flat relation tempname1(a,b); but without level lifting, we can only
   * get a nested relation tempname2(reddom); where reddom is defined on(id, a,b);
   * the following code is to extract a flat relation "tempname1" 
   * from the nested relation "tempname2"; 
   * it is the same when the red ujoin part is on the letf side of the join operator, 
   * so both after-process of relation rl, and rr are performed here.
   */
  private Relation IntermediateLevelLift(Relation rel) throws InterpretError
  {	 if (rel == null) return null;
  	 Relation relnew = rel;
  	 if(rel.domains.length == 1){
      	Domain reldom = rel.domains[0];
      	if(reldom.name.startsWith("reddom")){  
      		Relation tmprel = env.lookupRel("."+reldom.name, true);
      	
      		if(tmprel == null)
      			throw new InterpretError("relation ." + reldom.name + " not found");
      		tmprel.load();
      		
      		String str = new String();
            Object b = new Object();
            b = rel.data[0]; 
            Object a = new Object();   
            a= java.lang.reflect.Array.get(b,0);  
            str = a.toString(); 
      		Long srg = Long.valueOf(str);
      		relnew = tmprel.extract(srg.longValue());    
      		//deleteFile("."+reldom.name);
      	} 
      	//env.removeDom(reldom.name, true);  
      	
      }
  	 return relnew;
  } 
  
  /*	check the type of operand and change it
   *	to "relation (A)" if there is a node A that is the name of an attribute;
   *	parameter: node: the join operator node
   *    i   : which child of the node (0,1 in case of no join operands specified, or 0,2, otherwise)
   *    type: the type of the operand 
   */   
  private int ScalarDom2Relation(SimpleNode node, int i, int type) //added by Fan Guo on August 7, 2004
  throws InterpretError{
  	int tp = type;   
    SimpleNode child = null;
    if (type == STRING) {
    	child = (SimpleNode)node.jjtGetChild(i);
    	if(child.type == OP_IDENTIFIER) {
			Domain dj = env.lookupDom(child.name,true);
			if (dj != null && dj.type== STRING) {
				child = createRelNode(child, env);
				child.jjtSetParent(node);
				node.jjtReplaceChild(i,child);
				tp = IDLIST;
			} // else: invalid type, leave the original code to deal with
		}
    }
    return tp;
  }  
  /*
   * Turn a string node to a relation node
   */
  private SimpleNode createRelNode(SimpleNode node, Environment env) // July 12, 2004 Fan Guo
  throws InterpretError{
  	SimpleNode IDListNode = SimpleNode.jjtCreateNode("idlist", OP_IDLIST, OP_IDLIST, "");
  	IDListNode.jjtAddChild(node);
  	node.jjtSetParent(IDListNode);
  	SimpleNode RelNode = SimpleNode.jjtCreateNode("Relation", OP_RELATION, OP_RELATION, "");
  	RelNode.jjtAddChild(IDListNode);
  	IDListNode.jjtSetParent(RelNode);
  	return RelNode;
  } 
  
  /* 
   * Methods maintaining Relation Vector. 
   * add a relation name to the vector
   */
  private void addRelName(String name)// Fan Guo June 6, 2004
  throws InterpretError {
  	if (NRelName == null)  NRelName = new Vector();
  	NRelName.add(name);  			
  }
  
  /* 
   * Methods maintaining Relation Vector. 
   * delete a relation name at posotion relLevel-1 in the vector
   */
  private void delRelName()// Fan Guo June 20, 2004
  throws InterpretError {
  	if (NRelName != null)  {
  		if (NRelName != null) {
  			if ((relLevel -1)>= 0)
  				NRelName.removeElementAt(relLevel-1);
  		}
  	} 		
  }
  
  /* 
   * Methods maintaining Relation Vector. 
   * get and return a relation name at posotion relLevel-1 in the vector 
   */
  public String getRelName()// Fan Guo June 6, 2004
  throws InterpretError {
  	String name = null;
  	int level = 0;
  	if (NRelName != null) {
  		level = relLevel -1;
  		if (level>= 0)
  			name = (String) NRelName.elementAt(relLevel-1);  		
  	}
  	return name;
  }
  
  // Methods used for redefining virtual nested relation
  /*
   * orgNestedDom: the original nested virtual domain virtual nested relation;
   * newDoms: domains of the new defined nested relation;
   * tree: the definition tree of the redefined virtual nested domain
   */
  private Domain VirtualNestedRelationCreation(Domain orgNestedDom, SimpleNode tree, Domain[] newDoms, Environment env)
   throws InterpretError{
  	
 	Relation oldR = env.lookupRel("."+orgNestedDom.name, true);
 	Domain newNestedDom  = orgNestedDom;
	if (oldR==null ) {		   // not declared;
		oldR = new Relation ("." + orgNestedDom.name, newDoms, RELATION, 0, newDoms.length, 0, null, env);
		//orgNestedDom.tree = tree;
		env.put(orgNestedDom);
		env.put(oldR);
	}
	else if (!oldR.isDomainsEquals(newDoms)){  // same name but different domains;				
		Relation newR = LookupDotRelation(orgNestedDom.name, newDoms, env);
		String str = newR.name;
		str = str.substring(1,str.length());
		newNestedDom = env.lookupDom(str, true);
		if(newNestedDom == null) {
			newNestedDom = new Domain(str, orgNestedDom.type, 0, tree);
	
			env.put(newNestedDom);
			env.put(newR);
		}
		return newNestedDom;
	} else { // else same domains; do nothing	

	}
	return newNestedDom;	
  }
  
  public Relation LookupDotRelation(String name, Domain[] newDoms, Environment env) // August 11, 2004 Fan Guo 
  throws InterpretError{             // Look up the next dot relation according to the domain name
  	  			                     // Create a new one if it does not exist
  	int i = 0;
  	String dotname = null;
	if(!name.startsWith("."))
		dotname = "." + name;
	else dotname = name;
	String srname = dotname;
	String sr = new String("");
	//sr = sr.valueOf(i);
	//srname  += "_" + sr;  // name = ".Visit_1"
	Relation tmprel = null;
	Domain tmpdom = null;
	while (true) {
		tmprel = env.lookupRel(srname,true); 
		tmpdom = env.lookupDom(srname.substring(1,srname.length()), true);
		if (tmprel != null) {  // .Visit_i exists
			if(!tmprel.isDomainsEquals(newDoms)){// domains are different, look for next one
				sr = sr.valueOf(i++);
				srname  = dotname + "_" + sr;							
			} else   // domains are the same; do nothing							
				return tmprel;
		} else if (tmprel == null && tmpdom != null) {
			sr = sr.valueOf(i++);
			srname  = dotname + "_" + sr;	
		} else { // .Visit_i does not exist, create using this name			
			Relation newR = new Relation (srname, newDoms, RELATION, 0, newDoms.length, 0, null, env);
			env.put(newR);
			return newR;
		}
	}
  }

  private SimpleNode WildcardInProject(SimpleNode node, Environment env)  // August 19, 2004
	throws InterpretError
	{
		SimpleNode pNode = (SimpleNode)node.jjtGetParent();
		SimpleNode srcnode = (SimpleNode) node.jjtGetChild(1); 
		SimpleNode projectornode = (SimpleNode) node.jjtGetChild(0);
		if(projectornode.children == null) return node; // one case is in TSELECT: [] where a = "Roux" in Author;"
		SimpleNode exprnode = (SimpleNode) node.jjtGetChild(0).jjtGetChild(0);
		if(srcnode.opcode ==OP_DOT){ 
			SimpleNode[] nestednode = WildcardToRelations(exprnode,srcnode,env);
			Vector nds = new Vector();
			if(nestednode.length>0) {
				for (int i=0;i<nestednode.length; i++) {
					SimpleNode nd = node.jjtDeepDuplicate();
					nd.jjtReplaceChild(1,nestednode[i]);
					if (nds.size()!=0) {
						SimpleNode joinNode = SimpleNode.jjtCreateNode("join",OP_JOINOPERATOR, OP_UJOIN, "");				
						joinNode.jjtAddChild((SimpleNode)nds.lastElement());
						joinNode.jjtAddChild(nd);
						nd.jjtSetParent(joinNode);
						((SimpleNode)nds.lastElement()).jjtSetParent(joinNode);
						nds.addElement(joinNode);
					}else 
						nds.addElement(nd);	
				}
				node.jjtReplace((SimpleNode)nds.lastElement());
				((SimpleNode)nds.lastElement()).jjtSetParent(pNode);				
				return node;
			}   else return null;					
		} else 
			return node;
	}
  
	// deal with all 'one level lower nested relations' inside a relation
  	// exprnode: parent node of all domains;
    // node: the source node, e.g., the wildcard node if it is
	private SimpleNode[] WildcardToRelations(SimpleNode exprnode, SimpleNode node, Environment env)  // August 18, 2004
	throws InterpretError
	{   Relation[] dotRels = null;
		Vector RelNodes = new Vector();
		SimpleNode pNode = null;
		SimpleNode srcnode = node;//(SimpleNode) node.jjtGetChild(1); 
		if(srcnode.opcode ==OP_DOT){
			String relName = getRelName();
			if(relName==null){ // "." represents top level relations
				Vector TopRels= new Vector();
				Relation rel = null;  
				int count = 0;
				for(Enumeration e = env.reltable.keys(); e.hasMoreElements();) {
				    rel = (Relation) env.reltable.getRel((String)(e.nextElement()));
				    if(!rel.name.startsWith(".") ) {
				    	Domain[] tempD = ExpressionListToDomains(exprnode, rel, new Hashtable(), env);
				    	if(actualizableInRel(tempD, rel, env))
				    		TopRels.add(rel);
				    }
				}
				dotRels = new Relation[TopRels.size()];
				TopRels.copyInto(dotRels);
			} else {
				Relation rel = env.lookupRel(relName,true);  // relName represents top level relation
				if (rel==null) rel = env.lookupRel("."+relName,true);  // relName represents nested relation
				dotRels = getNestedRels(exprnode, rel, env);
			}
			pNode = (SimpleNode) srcnode.jjtGetParent();
			String str=null;
			SimpleNode relNode;
			for (int i=0;i<dotRels.length; i++) {
					if(dotRels[i].name.startsWith("."))
						str = dotRels[i].name.substring(1,dotRels[i].name.length());  // get rid of "."
					else str = dotRels[i].name;
					relNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, str);
					relNode.jjtSetParent(pNode);							
					//srcnode.jjtReplace(relNode);
					RelNodes.add(relNode);				
			}	
		}
		SimpleNode[] NewNodes = new SimpleNode[RelNodes.size()];
		RelNodes.copyInto(NewNodes);
	   return NewNodes;
	}
	
	private Relation[] getNestedRels(SimpleNode exprnode, Relation rel, Environment env) // Fan Guo August 20, 2004  
	  throws InterpretError
	{   Vector RelVec = new Vector();
		Domain dom;
		Relation r;
		for (int i=0;i<rel.numattrs; i++){
			dom = rel.domains[i];
			r = env.lookupRel("."+ dom.name,true);
			if(r!=null) {
				Domain[] tempD = ExpressionListToDomains(exprnode, rel, new Hashtable(), env);
				//if(actualizableInRel(tempD, rel, env))
				if(actualizableInRel(tempD, r, env))
	    			RelVec.add(r);
			}
		}
		Relation[] rels = new Relation[RelVec.size()];
		RelVec.copyInto(rels);
		return rels;
	}
	
	/*
	private SimpleNode ExpandRecursiveNode (SimpleNode node, Environment env) // Fan Guo August 26, 2004  
	  throws InterpretError
	  { 
		if (node == null) return null;

		if(IsRerursiveVirtualDomain(node)){
			if(relLevel==0) //schema,
				return node;
			else  {// rename to schema_n
				Domain curDom = env.lookupDom(node.name, true);  // name: schema					
				String newname = node.name+"_"+(relLevel-1);		
				SimpleNode right = (SimpleNode) curDom.tree;
				Domain newDom = newVirDomain(newname, right, env); 
				env.put(newDom);				
				SimpleNode left = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, newname);
				return left;
			}		
		} else return node;
	  }
	  */
	
	/*
	 * check if a domain list is actualizable in a relation
	 *
	 */ 
	private boolean actualizableInRel(Domain[] doms, Relation rel, Environment env)   // September 4, 2004
	throws InterpretError 
	{   		
		for (int i=0;i<doms.length; i++) {  // for each domain
			if (!rel.isDomainIn(doms[i])){  // not a actual domain in rel
				if(doms[i].tree !=null) {   // a virtual one
					SimpleNode child = null; 
					int num = doms[i].tree.jjtGetNumChildren();
					if(num!=0) {
						for(int j=0;j<num; j++) {
							child = (SimpleNode)doms[i].tree.jjtGetChild(j);  // each child node of the irtual domain tree
							if(child.type == OP_IDENTIFIER && child.name != null){
								Domain[] d = new Domain[1];
								d[0] = env.lookupDom(child.name, true);
								if(!actualizableInRel(d,rel,env)) 
									return false;									
							}
						}
					}
				} else if(doms[i].name !=null && !doms[i].name.equals("self"))
					return false;  // an actual domain not in rel
			} // else if a actual domain in rel
		}// end of for each domain
		return true;
		
	}
	
	private SimpleNode IsRedefined (SimpleNode node, Environment env)  // Fan Sep. 1 2004
	throws InterpretError  {
		Domain dom = env.lookupDom(node.name,true);
		Relation rel = env.lookupRel("."+node.name, true);
		if(dom == null || dom.tree == null || rel == null) return node;
		SimpleNode declNode = (SimpleNode) dom.tree.jjtGetChild(0);
		Relation rdecl = getIDList(declNode, env);
		if(rdecl.isDomainsEquals(rel.domains)) return node;
		Relation newRel = LookupDotRelation(dom.name,  rdecl.domains, env);
		String newName = newRel.name.substring(1,newRel.name.length());
		SimpleNode newNode = (SimpleNode)node.jjtDeepDuplicate();
		newNode.name = newName;
		node = newNode;
		return node;
		
	}
	
	private SimpleNode CheckRecursiveNode (SimpleNode node, Environment env)  // Fan Sep. 1 2004
	throws InterpretError  
	{
		if(node == null) return null;		
		if(IsRerursiveVirtualNestedDomain(node.name)){
			String name = getRelName();  // current level relation
			node = ExpandRecursiveNode(name, node, env);
			return node;
		} else return IsRedefined(node,env);
		//return node;
	}
	
	// Chck if the dom node of red OPerator is recursive one
	// node: the vertical node
	private SimpleNode CheckRecursiveNodeInRedOp (SimpleNode node, Environment env)// Fan Sep. 1 2004
	throws InterpretError
	{	
		if(node == null) return null;
		SimpleNode exprNode = (SimpleNode) node.jjtGetParent();
		SimpleNode prjctrNode = (SimpleNode) exprNode.jjtGetParent();
		SimpleNode prjctNode = (SimpleNode) prjctrNode.jjtGetParent();
		SimpleNode srcrelNode = (SimpleNode) prjctNode.jjtGetChild(1);
		
		SimpleNode idNode = (SimpleNode)node.jjtGetChild(0);
		if(IsRerursiveVirtualNestedDomain(idNode.name)){						
			String name = srcrelNode.name;			
			idNode = ExpandRecursiveNode(name, idNode, env);
			node.jjtReplaceChild(0,idNode);
		}
		return node;
	}
 
	/*  modify the recursive node and its domain; 
	 *  modify tree of the domain in different nested level
	 *  name: the name current level relation
	 *  node:  recursive virtual nested relation node
	 *  return: modified recursive node
	 */
	private SimpleNode ExpandRecursiveNode (String name, SimpleNode node, Environment env) // Fan Guo October 4, 2004  
	  throws InterpretError
	  { 
		
			recurpNode = null;
			childOrder = 0;
		
			int level = getNestedLevel(0,name,env);  
			Domain curDom = env.lookupDom(node.name, true);  // name: schema
			Domain newDom = null;
			SimpleNode curTree = (SimpleNode) curDom.tree; 
			SimpleNode tree = null; 
			String newname = node.name+"_"+level;// +"level";
			SimpleNode newNode = null;
			SimpleNode lastNode = null;
			
			for (int i=level;i>=0;i--){	
				newname = node.name+"_"+i;	
				tree = curTree.jjtDeepDuplicate();
				getNodePosition(node.name,tree);
					if (i==0){  // the first node: schema
						
						//newname = node.name+"_" + 0 + 0;
						
						//lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, lastname);
						recurpNode.jjtReplaceChild(childOrder, lastNode);
						//newDom = newVirDomain(node.name, tree, env); // schema
						newDom = newVirDomain(newname, tree, env); // schema_00
						newNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, newDom.name);
						//SimpleNode pNode = (SimpleNode)node.jjtGetParent();
						//newNode.jjtSetParent(pNode);
					} else {	// nodes in middle
						if(i == level) 
							recurpNode.jjtRemoveChild(childOrder);
						else {
							//lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, lastname);
							recurpNode.jjtReplaceChild(childOrder, lastNode);
						}		
						newDom = newVirDomain(newname, tree, env);
						
						lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, newDom.name);
					}
					
					Relation testR = env.lookupRel("."+ newDom.name, true);
					Domain testDom = env.lookupDom(newDom.name,true);
					if (newDom.type != IDLIST || (newDom.type == IDLIST && env.lookupDom(newDom.name,true) == null))
		 	    		env.put(newDom);	
					testR = env.lookupRel("."+ newDom.name, true);
					testDom = env.lookupDom(newDom.name,true);
					System.out.println("Redefinign virtual nested domain test..."+ newDom.name);
			//	}
			}			
		//} 
		//MyTraverse(newNode,env);
		return newNode;
	  }
	/*
	private SimpleNode ExpandRecursiveNode (String name, SimpleNode node, Environment env) // Fan Guo August 26, 2004  
	  throws InterpretError
	  { 
		
			recurpNode = null;
			childOrder = 0;
		
			int level = getNestedLevel(0,name,env);  
			Domain curDom = env.lookupDom(node.name, true);  // name: schema
			Domain newDom = null;
			SimpleNode curTree = (SimpleNode) curDom.tree; 
			SimpleNode tree = null; 
			String newname = node.name+"_"+level;// +"level";
			SimpleNode newNode = null;
			SimpleNode lastNode = null;
			
			for (int i=level-1;i>=-1;i--){	
				newname = node.name+"_"+i;	
				tree = curTree.jjtDeepDuplicate();
				getNodePosition(node.name,tree);
					if (i==-1){  // the first node: schema
						
						newname = node.name+"_" + 0 + 0;
						
						//lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, lastname);
						recurpNode.jjtReplaceChild(childOrder, lastNode);
						//newDom = newVirDomain(node.name, tree, env); // schema
						newDom = newVirDomain(newname, tree, env); // schema_00
						newNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, newDom.name);
						//SimpleNode pNode = (SimpleNode)node.jjtGetParent();
						//newNode.jjtSetParent(pNode);
					} else {	// nodes in middle
						if(i == (level-1)) 
							recurpNode.jjtRemoveChild(childOrder);
						else {
							//lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, lastname);
							recurpNode.jjtReplaceChild(childOrder, lastNode);
						}		
						newDom = newVirDomain(newname, tree, env);
						
						lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, newDom.name);
					}
					
					Relation testR = env.lookupRel("."+ newDom.name, true);
					Domain testDom = env.lookupDom(newDom.name,true);
					if (newDom.type != IDLIST || (newDom.type == IDLIST && env.lookupDom(newDom.name,true) == null))
		 	    		env.put(newDom);	
					testR = env.lookupRel("."+ newDom.name, true);
					testDom = env.lookupDom(newDom.name,true);
					System.out.println("Redefinign virtual nested domain test..."+ newDom.name);
			//	}
			}			
		//} 
		//MyTraverse(newNode,env);
		return newNode;
	  }*/ // this version name the top level recursive relation _00, and then _0, _1,... which is not nice for the tname of top level one, use the above version to use _0, and then _1,..
	
	/*
	// expand the recursive node to a complete tree with the number of nested level
	// node: the node of the recursive virtual domain
	private SimpleNode ExpandRecursiveNode (String name, SimpleNode node, Environment env) // Fan Guo August 26, 2004  
	  throws InterpretError
	  { 
		//if (node == null) return null;		
		recurpNode = null;
		childOrder = 0;
		//SimpleNode recurpNode = null;
		//int childOrder = 0;
		//if (recurpNodeV == null) recurpNodeV = new Vector();
		//if (childOrderV == null) childOrderV = new Vector();
		//if(IsRerursiveVirtualDomain(node)){
			int level=0;
			//String name = getRelName();  // current level relation
			level = getNestedLevel(0,name,env);
			Domain curDom = env.lookupDom(node.name, true);  // name: schema
			Domain newDom = null;
			SimpleNode curTree = (SimpleNode) curDom.tree; 
			SimpleNode tree = null; 
			String newname = null;
			String lastname = null;
			SimpleNode newNode = null;
			SimpleNode lastNode = null;
			for (int i=level;i>=-1;i--){	
				lastname = newname;
				newname = node.name+"_"+i;	
				tree = curTree.jjtDeepDuplicate();
				getNodePosition(node.name,tree);
			//	int rcrNum = recurpNodeV.size();
			//	for (int j=0;j<rcrNum; j++) {
					//recurpNode = (SimpleNode) recurpNodeV.elementAt(j);
					//Object a = new Object(); 
					//a = childOrderV.elementAt(j);					
					//String str = new String();
					//str = a.toString();	          
		      	//	childOrder = Integer.parseInt(str);
					if(i==level){	 // the deepest node				
						SimpleNode dc = SimpleNode.jjtCreateNode("ILITERAL",OP_LITERAL,OP_LONG, lastname);	
						dc.setObject(new Long((long) LONG_DC));
						((SimpleNode)tree.jjtGetChild(0)).jjtReplace(dc);
						((SimpleNode)tree.jjtGetChild(0)).children = null;
						
						Domain d = env.lookupDom(newname, true); // create one if not exist; otherwise, no matter it was of type LING or IDLIST, leave it along
						if(d==null)							
							newDom = newVirDomain(newname, tree, env);
						else newDom = env.lookupDom(newname, true);
						//newDom = newVirDomain(newname, tree, env);
						
					}else if(i==-1){  // the first node: schema
						newname = node.name+"_" + 0 + 0;
						lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, lastname);
						recurpNode.jjtReplaceChild(childOrder, lastNode);
						//newDom = newVirDomain(node.name, tree, env); // schema
						newDom = newVirDomain(newname, tree, env); // schema_00
						newNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, newDom.name);
						//SimpleNode pNode = (SimpleNode)node.jjtGetParent();
						//newNode.jjtSetParent(pNode);
					} else {	// nodes in middle
						lastNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, lastname);
						recurpNode.jjtReplaceChild(childOrder, lastNode);
						
						Domain d = env.lookupDom(newname, true);
						if(d!=null && d.type == LONG) {
							//env.domtable.rmDom(d.name);
							env.domtable.rmRecurDom(d.name);
						}
						d = env.lookupDom(newname, true);
						
						newDom = newVirDomain(newname, tree, env);
					}
					Relation testR = env.lookupRel("."+ newDom.name, true);
					Domain testDom = env.lookupDom(newDom.name,true);
					if (newDom.type != IDLIST || (newDom.type == IDLIST && env.lookupDom(newDom.name,true) == null))
		 	    		env.put(newDom);	
					testR = env.lookupRel("."+ newDom.name, true);
					testDom = env.lookupDom(newDom.name,true);
					System.out.println("Redefinign virtual nested domain test..."+ newDom.name);
			//	}
			}			
		//} 
		//MyTraverse(newNode,env);
		return newNode;
	  }
	*/  // this version keep the deepest recursive node, which will cause LONG type recursive node.
	// use the above version, just delete the deepest node.  September 6, 2004
	
	/* get the nested level of current relation
	 * return the deepest level of the current relation
	 * name: the name of the current relaiton; with (nested relation) or without (top level relation) "." 
	 */
	private int getNestedLevel(int level, String name, Environment env)  // Fan Guo August 31, 2004
	throws InterpretError
	{   int dl=0;
		int lastlevel = level;
		Relation rel = null;
		//if (level ==0) 
			rel = env.lookupRel(name, true);
		//else 
		//	rel = env.lookupRel("."+name, true);		
		if (rel == null) return 0;
		
		for(int i=0;i<rel.domains.length;i++) {
			dl = 0; 
	      if(rel.domains[i].type == IDLIST && !rel.domains[i].name.equals(".id")) {
			//level++;
			name = "."+ rel.domains[i].name;
			dl = getNestedLevel(0, name, env) +1;
	      }  
	      if (lastlevel + dl > level) 
	      	level = lastlevel + dl;	 
	    }
		
	return level;
	}
	/*
	private boolean IsRerursiveVirtualDomain(SimpleNode node) // Fan Guo August 8, 2004  
	  throws InterpretError
	{   
		boolean isRcr = false;
		String name = node.name;
		Domain dom = env.lookupDom(name,true);
		if(dom.tree !=null) {
			SimpleNode treenode = (SimpleNode) dom.tree.jjtGetChild(0);		
			isRcr = IsNameIn(name, treenode);
		}
		return isRcr;
	} */  
	private boolean IsRerursiveVirtualNestedDomain(String name) // Fan Guo August 8, 2004  
	  throws InterpretError
	{   
		boolean isRcr = false;
		
		Domain dom = env.lookupDom(name,true);
		Relation rel = env.lookupRel("."+name,true);
		if(rel != null && rel.isDomainIn(dom)) {
			isRcr = true;
			//SimpleNode treenode = (SimpleNode) dom.tree.jjtGetChild(0);
			//getNodePosition(name,treenode);
		}
		
		return isRcr;
	}  
	//recursive routine checking if a given name is in the tree
	public void getNodePosition(String name, SimpleNode node) // Fan Guo August 8, 2004  
	throws InterpretError
	{ 	
		
		SimpleNode child = null; 
		int num = node.jjtGetNumChildren();
		if(num!=0) {
			for(int i=0;i<num; i++) {
				child = (SimpleNode)node.jjtGetChild(i);
				if(child.name != null && child.name.equals(name)){
					recurpNode = node;	
					childOrder = i;
					//recurpNodeV.addElement(node);	
					//childOrderV.addElement(new Integer((int) i));
					//break;
				} else {			
					getNodePosition(name,child);
				}
			}
		}
	}
		
		
	/*
	// recursive routine checking if a given name is in the tree
	public boolean IsNameIn(String name, SimpleNode node) // Fan Guo August 8, 2004  
	throws InterpretError
	{ 	
		boolean isIn = false;
		boolean isIn1 = false;
		SimpleNode child;
		int num,i;
		switch(node.type){				
			case OP_IDENTIFIER:
				if(node.name.equals(name))
					return true;
				else return false;
			case OP_TSELECT:
    		case OP_SELECT :	
			case OP_PROJECT:
				isIn = IsNameIn(name,(SimpleNode)node.jjtGetChild(1));
				if(isIn)
					return true;
				else {
					child = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
					num = child.jjtGetNumChildren();
					for (i=0;i<num;i++) {
						isIn = IsNameIn(name,(SimpleNode)child.jjtGetChild(i));
						if (isIn) 
							return true;
					}
					return false;
				}
			case OP_QSELECT : 
				num = node.jjtGetNumChildren();			
				if (num==2){
					isIn=  IsNameIn(name,(SimpleNode)node.jjtGetChild(1));
				}else {
					isIn=  IsNameIn(name,(SimpleNode)node.jjtGetChild(2));
				}		
				return isIn;
      		case OP_RED:
				isIn = IsNameIn(name,(SimpleNode)node.jjtGetChild(0));
				if(isIn) return true;
				else return false;
			case OP_JOINOPERATOR: 
				num = node.jjtGetNumChildren();
				isIn = IsNameIn(name,(SimpleNode)node.jjtGetChild(0));				
				if (num==2){
					isIn1=  IsNameIn(name,(SimpleNode)node.jjtGetChild(1));
				}else {
					isIn1=  IsNameIn(name,(SimpleNode)node.jjtGetChild(3));
				}
				isIn = isIn || isIn1;			
				return isIn;				
			case OP_ARRAY: 
    		case OP_PICK : 
      			isIn = IsNameIn(name,(SimpleNode)node.jjtGetChild(0));
				if(isIn) return true;
				else return false;
   			case OP_EDIT :
    		case OP_ZORDER :
    		case OP_EVAL : 
    		case OP_RENAME :  
      			throw new InterpretError("TO BE IMPLEMENTED");
        	// many more other case ...	
			default: 
				return false;
	  }
	}
	*/
	
	public void MyTraverse(SimpleNode node, Environment env) { // June 8, 2004 Fan Guo  
		//System.out.println("identifer:"  + " name:" + " opcode:" + " type:" + " numChildren:" + " info:" );
		//System.out.println(node.identifier + node.name + node.opcode + node.type + node.jjtGetNumChildren() + node.info);
		System.out.println("identifer:" + node.identifier + "  name:" + node.name + "  opcode:" + node.opcode + "  type:" + node.type + "  numChildren:" + node.jjtGetNumChildren() +"  info:" + node.info);
		// System.out.println("numChildren:" + node.jjtGetNumChildren() +" info:" + node.info + " url:" + node.url);
		if (node.type == OP_IDENTIFIER){
			Domain d = env.lookupDom(node.name, true);
			
	      	if (d != null) {
	      		if (d.tree != null) { 
	      			for (int j=0; j<printLevel; j++) System.out.print("    ");
	      			System.out.println("tree:");
	      			printLevel ++;
	      			for (int j=0; j<printLevel; j++) System.out.print("    ");	      				      			
	      			MyTraverse(d.tree, env);
	      			printLevel --;
	      		}
	      	}
	      	
		}
		printLevel ++;
		for(int i = 0; i < node.jjtGetNumChildren(); i++) {
			for (int j=0; j<printLevel; j++) System.out.print("    ");
			//if (DEBUG_20040608) if (DEBUG_20040608_1) 
				MyTraverse((SimpleNode) node.jjtGetChild(i), env);
		}  
		printLevel --;
	}
	/*
	 * This method is for relation initialization.
	 * rel: relation which is being initializing
	 * i:
	 * offset:
	 * str: the name of the identifier as the data of the tuple
	 */
	
	private void assignIdentifier(Relation rel, int i, int offset, String str)  // September 2004
	throws InterpretError {
		Domain[] doms = rel.domains;

		switch(doms[i].type) {
		case TYPE:
			if(Utility.IsaType(str) || env.lookupDom(str,true)!=null)
				((String[]) rel.data[i+offset])[rel.capacity]= str;
			else throw new InterpretError("invalid TYPE value '"+ str + "'");
	 		break;
	 	case ATTRIBUTE:
	 		if(env.lookupDom(str,true)!=null)
				((String[]) rel.data[i+offset])[rel.capacity]= str;
			else throw new InterpretError("invalid ATTRIBUTE value '"+ str + "'");
	 		break;
	 	default:
	 		throw new InterpretError("invalid value ");
	 	}
	}
	
	private void assignIdentifier(Relation rel, int i, int offset, SimpleNode typeIDList)  // September 2004
	throws InterpretError {
		int num = typeIDList.jjtGetNumChildren();
		String str = "";
		SimpleNode tp = null;
		Domain[] doms = rel.domains;
		for(int j=0;j<num;j++) {
			tp = (SimpleNode)typeIDList.jjtGetChild(j);
			switch(doms[i].type) {
				case TYPE:			
					if(Utility.IsaType(tp.name) || env.lookupDom(tp.name,true)!=null)
						str += tp.name;
					else throw new InterpretError("invalid TYPE value '"+ tp.name + "'.");
			 		break;
			 	case ATTRIBUTE:
			 		if(env.lookupDom(tp.name,true)!=null)
			 			str += tp.name;
					else throw new InterpretError("invalid ATTRIBUTE value '"+ tp.name + "'.");
			 		break;
			 	default:
			 		throw new InterpretError("invalid value ");
			 	}
			if(j<(num-1)) str += "|";
		}
		
		((String[]) rel.data[i+offset])[rel.capacity]= str;
		
	}
	
	// for case: relation Schema(sattri,schema_0)
	private String HasRecursiveDom(Relation rel, Environment env)  // september 8, 2004
	throws InterpretError {
		String rcrname = null;
		Domain [] doms = rel.domains;
		for (int i=0;i<doms.length; i++) {  // check each domain in rel
			if(doms[i].type == IDLIST && !(doms[i].name).equals(".id")){
				String dname = doms[i].name;				
				if (dname.indexOf("_")!=-1) 					
					rcrname = dname.substring(0,dname.indexOf("_"));
				else rcrname = dname;
				Relation drel = env.lookupRel("."+dname, true);
				for (int j=0;j<drel.domains.length; j++){
					// if(drel.domains[j].name.startsWith(rcrname) )  // Dec. 9 2004 Fan 
					if(drel.domains[j].name.startsWith(rcrname)  && drel.domains[j].name.indexOf("_")!=-1 )
						//return rcrname;
						return dname;  // the actual recursive name in the top level relation. schema_00 or schema_0
				}				
			}				
		}
		return null;
	}
	// rel: Schema(sattri, schema_0);
	private String ModifyIfHasRecursiveDom(Relation rel, Environment env)  // september 8, 2004
	throws InterpretError {
		String rcrdname = HasRecursiveDom(rel,env);
		if(rcrdname!=null) {
			String s =  rcrdname+ "_0";
			return s;
		}
		else return null;
		/*
		String tmpname0 = nextTempName();
		if (rcrdname!=null) {
			Domain dom = env.lookupDom(rcrdname, true);			
			Relation dotr = env.lookupRel("."+rcrdname,true);
			int offset = 0;
			String s = null;
			while(dotr!=null && dotr.numtuples==0) {
				s = rcrdname+ "_" + offset;
				dotr = env.lookupRel("."+s,true);
			}
			if (dotr!=null) {
				dom = env.lookupDom(s,true);
				Relation newrel = changeDom(rel,dom,tmpname0,env);
			}
			*/	
	}
//	private SimpleNode LookupValidRelNameForPR(SimpleNode firstchild, Environment env)  // September 16, 2004
	private SimpleNode LookupValidRelName(int i, SimpleNode child, Environment env)  // Fan January 7, 2005
	throws InterpretError {
		String name;
		Relation r;
		Domain d;
		switch(child.type){
			case OP_IDENTIFIER:
				return child;
			case OP_LITERAL:
				switch (child.opcode){
					case OP_FLOAT:
						float fn = ((Float)child.info).floatValue(); // 0.2
						name = CastToStrn.value(fn);  // "0.2"
						name = name.substring(1); //  ".2"
						//Relation r;
						//Domain d;
						while(true){
							d = env.lookupDom(name.substring(1), true);
							r = env.lookupRel(name,true);
							if(r != null) {  // ".2000"
								SimpleNode newNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, name);
								SimpleNode pNode = (SimpleNode)child.jjtGetParent();
								return newNode;
								//newNode.jjtSetParent(pNode);
								//pNode.jjtReplaceChild(i,newNode);
								
							}
							if(d==null && r==null)	
								name = name+"0";
							else throw new InterpretError("invalid name. ");
							
						}
					case OP_INTEGER:  // Fan January 7, 2005
						Integer intN = (Integer)child.info;
						if (intN.compareTo(new Integer(INT_DC)) ==0) return child;
						if (intN.compareTo(new Integer(INT_DK)) ==0) return child;
						name = intN.toString(); // "2000"
						while(true){
							d = env.lookupDom(name, true);
							r = env.lookupRel(name,true);
							if(r != null || d != null) { // "2000"								
								SimpleNode newNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, name);
								SimpleNode pNode = (SimpleNode)child.jjtGetParent();
								newNode.jjtSetParent(pNode);
								pNode.jjtReplaceChild(i,newNode);
								return newNode;
							}
							else throw new InterpretError(name + " is an invalid name. ");
							
						}	
					default: return child;
						//throw new InterpretError("PR command to be implemented. ");
				}
				/*
			case OP_BIOPERATOR:
				if(firstchild.opcode == OP_MINUS) { // pr .schema_-1;
					SimpleNode node1 = (SimpleNode)firstchild.jjtGetChild(0);
					SimpleNode node2 = (SimpleNode)firstchild.jjtGetChild(1);
					int i = ((Integer)firstchild.info).intValue(); // 1
					name = CastToStrn.value(i);  // "0.2"
					name = node1.name + "-" + name;
					SimpleNode newNode = SimpleNode.jjtCreateNode("IDENTIFIER",OP_IDENTIFIER,OP_IDENTIFIER, name);
					SimpleNode pNode = (SimpleNode)firstchild.jjtGetParent();
					return newNode;
				}*/
			default: 
				return child;
		}
		
	}
	
	
	private void printRecursiveNestedRelations(Relation rel, Environment env)
	throws InterpretError {
		String rcrdname = HasRecursiveDom(rel,env);  // rcrdname = schema_00, schema_0, or schema;
		if (!rel.name.startsWith(".") && rcrdname != null) {  // a top level relation has a recursive domain
			 String tmpname0 = nextTempName();
			 String tmpname1 = nextTempName();
			 String rcrName = null;
			 String strl = null;
			 if(rcrdname.indexOf("_")==-1) {
			 	rcrName = rcrdname;			 	
			 }else	{		 
			 	rcrName = rcrdname.substring(0,rcrdname.indexOf("_")); // rcrName = schema;			 
			    strl = rcrdname.substring(rcrdname.indexOf("_")+1, rcrdname.length()); // "00", or "0" , "-1" possible
			 }
			 int startLevel = 0;
			 String name0 = null;
			 /*
			 if(strl ==null ||strl.equals("00")) {
			 	startLevel = 0;
			 	name0 = "."+ rcrdname;
			 } 
			 //if(strl ==null ||strl.equals("-1")) {
			 //	startLevel = 0;
			 //	name0 = "."+ rcrdname;
			 //} 
			  	else if(strl.equals("0")){
			 	startLevel = 1;
			 	name0 = "."+rcrName+"_"+0;	
			 } else if(strl.indexOf("_")!=-1) {// "00_0"
			 	String inter = strl.substring(0,strl.indexOf("_"));  // "00"
			 	strl = strl.substring(strl.indexOf("_")+1, strl.length());			 		
			 	rcrName = rcrName+"_"+inter;
			 	name0 = "."+ rcrName;
			 	startLevel = 1;
			 } else return;
			 */  // commented out on October 4, 2004 use the following part. 
			 if(strl ==null ||strl.equals("0")) {
			 	startLevel = 0;
			 	name0 = "."+ rcrdname;
			 } 			 
			 else if(strl.equals("1")){
			 	startLevel = 1;
			 	name0 = "."+rcrName+"_"+1;	
			 } else if(strl.indexOf("_")!=-1) {// "0_0"
			 	String inter = strl.substring(0,strl.indexOf("_"));  // "0"
			 	strl = strl.substring(strl.indexOf("_")+1, strl.length());			 		
			 	rcrName = rcrName+"_"+inter;
			 	name0 = "."+ rcrName;
			 	startLevel = 1;
			 } else return;
			 Relation cumulateR = env.lookupRel(name0, true);
			 Relation tempR = null;
			 Relation middleR = null;
			 String domname = null;
			 Domain dom = env.lookupDom(rcrName, true); 
			 int cumCounter = 0;
			 cumulateR = changeDom(cumulateR, dom, name0 + " (Compact Form)",env);
			 int i = startLevel;
			 while(true) {
		  		//String ns = "."+rcrdname + "_" + i;
			 	String ns = "."+rcrName + "_" + i;
		  		middleR = env.lookupRel(ns, true);	  		
		  		if (middleR!=null) {
		  			middleR = changeDom(middleR,dom,tmpname1,env);
		  			tempR = cumulateR.mujoin(OP_UJOIN,null,middleR,null, cumulateR.name);
		  			i++;
		  			cumCounter++;
		  			cumulateR = tempR;
		  		} else break;
		  	}  
			// if(cumCounter>0) {
		  		cumulateR.load();
		  		cumulateR.save();
		  		cumulateR.print(null,true);
		  		env.removeRel(cumulateR.name,true);
		  		deleteFile(cumulateR.name);
		  		deleteFile(tmpname0);
		  		deleteFile(tmpname1);
		  	//}	   
		}
	}
	/*
	private void printRecursiveNestedRelations(Relation rel, Environment env)
	throws InterpretError {
	  String tmpname0 = nextTempName();
	  String tmpname1 = nextTempName();
	  String relname = rel.name;
	  Relation cumulateR = rel;
	  Relation tempR = null;
	  Relation middleR = null;
	  String domname = null;
	  Domain dom = null; 
	  int cumCounter = 0;
	  
	  if(relname.endsWith("_0")) {	 	
	  	relname = relname.substring(0,relname.length()-2); //indexOf("_"));
	  	domname = relname.substring(1,relname.length());
	  	dom = env.lookupDom(domname, true);
	  	cumulateR = changeDom(cumulateR, dom, tmpname0,env);
	  	int i=1;
	  	while(true) {
	  		String ns = relname + "_" + i;
	  		middleR = env.lookupRel(ns, true);	  		
	  		if (middleR!=null) {
	  			middleR = changeDom(middleR,dom,tmpname1,env);
	  			tempR = cumulateR.mujoin(OP_UJOIN,null,middleR,null,relname + "_0" + "_CompactForm");
	  			i++;
	  			cumCounter++;
	  			cumulateR = tempR;
	  		} else break;
	  	}  

	  	if(cumCounter>0) {
	  		cumulateR.load();
	  		cumulateR.save();
	  		cumulateR.print(null,true);
	  		env.removeRel(cumulateR.name,true);
	  		deleteFile(cumulateR.name);
	  		deleteFile(tmpname0);
	  		deleteFile(tmpname1);
	  	}	  
	  }
	}
	*/ // this version will print the compact form for only relation with name end with _0. e.g., schema_0;
       // use the above version to check a top level relation and print a recursive compact for if necessary
	   // September 8, 2004
	
	// change the recursive domain in relation rel to "dom".
	// the new relation is named tmpname.
	private Relation changeDom(Relation rel, Domain dom, String tmpname, Environment env)
	throws InterpretError {
		Domain[] domsL = new Domain[rel.domains.length];
		rel.load();
		int p;
        for(p=0;p<rel.domains.length;p++){
        	if(rel.domains[p].name.startsWith(dom.name))
        		domsL[p] = dom;
        	else
        		domsL[p]=rel.domains[p];
        }
        Relation newR = new Relation(tmpname ,domsL,rel.numtuples,domsL.length,0,env);
        Object[] dataArray = newR.createDataArray(newR.domains,newR.numattrs*newR.numtuples);
        dataArray = rel.data;
        newR.data = dataArray;
        newR.save();
		return newR;
	}
/*
 * This is an overload method 
 */
	/*// the function of this method is merged to the original method AttribsRel2DomList. on October 18, 2004
	public Domain[] AttribsRel2DomList(Relation rel, int type) // September 14, 2004
	   throws InterpretError
	   {
		   Domain[] domList = null;
		   rel = rel.duplicateLoadAll();
	   	   domList = new Domain[rel.numtuples];	// great! this is a unary relation and the duplication has been eliminaed when the relaiton is formed. 
	   	   String domname = null;
	   	   Domain newDom = null;
		   if(rel.domains[0].type != ATTRIBUTE && type == OP_ATTRIBUTE){
		   		String virdName = rel.domains[0].name+"_evaled";
		   		Domain virDom = env.lookupDom(virdName, true);
		   		SimpleNode tree = (SimpleNode)virDom.tree;  // this tree is "let year be titles;" now we changed it to IfThenElse statement
		   		SimpleNode treeChild = (SimpleNode) tree.jjtGetChild(0);
		   		int treeType = traverseType(treeChild,env);
		   		for(int i=0;i<domList.length;i++) {
		   			switch(rel.domains[0].type){
		   				case STRING:
		   					domname = ((String[]) rel.data[0])[i];
		   					break;
		   				case INTEGER: 
		   					int val = ((int[]) rel.data[0])[i];
		   					domname = CastToStrn.value(val);
		   						break;
		   				default: break;	
		   			}
		   			if(type == OP_ATTRIBUTE){
		   				domList[i] = Eval2IFTHENELSE(domname, rel.domains[0].type, virdName, env);
		   			}
		   			
		   		}
		   }else if (rel.domains[0].type == ATTRIBUTE && type == ATTRIBUTE){ 		   				 
		   		for(int i=0;i<domList.length;i++)
		   			domList[i]= env.lookupDom(((String[]) rel.data[0])[i],true);
		   }   
		return domList;
	   }
	   */ 
	  /* declare a new virtual domain. 
	   * virName: name of the virtual attribute to be declared;  e.g. B
	   * domType: original type of the domain to be evaled before casting to ATTRIBUTE. such as 2000, which is the value of domain year, and year is evaled with eval (attr)year
	   * evaledName: name of the virtual domain containing the eval definition. e.g. A_evaled
	   * => let B be if A = B then definition else null;
	   * Note: type of attribute A must be ATTRIBUTE
	   */
	 private Domain Eval2IFTHENELSE(String virName, int domType, String evaledName, Environment env ) 
	 throws InterpretError{   
	 	String orgName = evaledName.substring(0,evaledName.indexOf("_evaled"));
   		Domain evaledDom = env.lookupDom(evaledName, true);
   		SimpleNode tree = (SimpleNode)evaledDom.tree;  // this tree is "let A_evaled be B+C+D;" now we changed it to IfThenElse statement
   		SimpleNode treeChild = (SimpleNode) tree.jjtGetChild(0); // get the B+C+D part;
   		int treeType = traverseType(treeChild,env);  // get the type of B+C+D;   			
   			
   		// construct IfThenElse tree
   		SimpleNode newTree = (SimpleNode)tree.jjtDeepDuplicate();		   			
   		SimpleNode ifthenelse = SimpleNode.jjtCreateNode("IfThenElse",OP_IFTHENELSE,OP_IFTHENELSE,"");
   		newTree.jjtReplaceChild(0,ifthenelse);
   		ifthenelse.jjtSetParent(newTree);
	   	SimpleNode compare = SimpleNode.jjtCreateNode("Compare",OP_BIOPERATOR,OP_EQ,"");
	   	SimpleNode compleft = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,orgName);
	   	SimpleNode compright = SimpleNode.jjtCreateNode("Literal",OP_LITERAL,Utility.typeConvertToSimpleNode(domType),"");
   		//compright.setObject((Object)virName);
		switch(domType){
			case ATTRIBUTE:
	   		case STRING:
	   			compright.setObject((Object)virName);
   				break;
   			case INTEGER:
   				compright.setObject(new Integer(virName));
   				break;
   			case SHORT:
   				compright.setObject(new Short(virName));
   				break;
   	        case LONG:
   	        	compright.setObject(new Long(virName));
   	        	break;
   	        default: 
   	        	compright.setObject((Object)"dc");			   	        	
   			break;
	   	}
	   	compleft.jjtSetParent(compare);
	   	compright.jjtSetParent(compare);
	   	compare.jjtAddChild(compleft);
	   	compare.jjtAddChild(compright);
	   	compare.jjtSetParent(ifthenelse);
	   	ifthenelse.jjtAddChild(compare);
	   		
	   	SimpleNode trueval = treeChild.jjtDeepDuplicate();
	   	trueval.jjtSetParent(ifthenelse);   
	   	ifthenelse.jjtAddChild(trueval);
	   	//SimpleNode falseval = SimpleNode.jjtCreateNode("Literal",OP_LITERAL,OP_DC,"");
	   	
	   	
	   	SimpleNode falseval = SimpleNode.jjtCreateNode("Literal",OP_LITERAL,Utility.typeConvertToSimpleNode(treeType),"");
	   	switch(treeType){
	   		case STRING:
	   			falseval.setObject((Object)"dc");
   				break;
   			case INTEGER:
   				falseval.setObject(new Integer(INT_DC));
   				break;
   			case SHORT:
   				falseval.setObject(new Short(SHORT_DC));
   				break;
   	        case LONG:
   	        	falseval.setObject(new Long(LONG_DC));
   	        	break;
   	       
   	        default: 
   	        	falseval.setObject((Object)"dc");			   	        	
   			break;
	   	}
	   	falseval.jjtSetParent(ifthenelse);
	   	ifthenelse.jjtAddChild(falseval);
	   	// end of constructing IfThenElseTree
   		Domain newDom = newVirDomain(virName, newTree, env);
   		Relation newRel = env.lookupRel("."+newDom.name, true);  
   		env.put(newDom);
   		if(newRel!=null) env.put(newRel);	
   		return newDom;
	 }
	 
	 private boolean isJoinOperator(SimpleNode node) throws InterpretError 
	 {
	 	switch(node.opcode){
	 		case OP_NOP:    // Mu-Join operator
			case OP_IJOIN:
			case OP_UJOIN:
			case OP_SJOIN:
			case OP_LJOIN:
			case OP_RJOIN:
			case OP_DJOIN:// DJOIN is DLJOIN
			case OP_DRJOIN:			   
			case OP_SHIFT://shift for not // Sigma-Join operator
			case OP_ICOMP:
			case OP_EQJOIN:
			case OP_GTJOIN:
			case OP_GEJOIN:
			case OP_LTJOIN:
			case OP_LEJOIN:
			case OP_IEJOIN:			  
			case OP_NOTICOMP: // Not-Sigma-Join operator
			case OP_NOTEQJOIN:
			case OP_NOTGTJOIN:
			case OP_NOTGEJOIN:
			case OP_NOTLTJOIN:
			case OP_NOTLEJOIN:
			case OP_NOTIEJOIN:
				return true;
			default:
				return false;
	 	}
	 }
	/******************** end of Fan's new supporting methods ***********************/

        /********************* start of Jaintao's methods *******************************/ 
	/**
	* Evaluate binary grep expression
	* @param	node	User input tree
	* @param	filename	Returned relation name
	* @param	env	Environment
	*/
	private Relation evaluateBigrep(SimpleNode node, String filename, Environment env)
									throws InterpretError // Jiantao 2004
	{	Relation rl = null, rr = null;
		String lname = nextTempName();
		String rname = nextTempName();
		final String DOM_SEQ_LEFT = "_seql";
		final String DOM_SEQ_RIGHT = "_seqr";
		
		// get text form the environment
		Text tl = env.textTable.getText(((SimpleNode)node.jjtGetChild(0)).name);
		Text tr = env.textTable.getText(((SimpleNode)node.jjtGetChild(2)).name);
		if(tl == null)
			throw new InterpretError("Text "+ ((SimpleNode)node.jjtGetChild(0)).name + " does not exist.");
		else if(tr == null)
			throw new InterpretError("Text "+ ((SimpleNode)node.jjtGetChild(2)).name + " does not exist.");
		
		// get pattern from input
		String tmp = ((SimpleNode)node.jjtGetChild(1).jjtGetChild(0)).name;
		String pattern = tmp.substring(1, tmp.length()-1);
		
		// assign domains from parameters
		String dname;
		int idx = 0, nString = 0, nInt = 0;
		Domain d;
		Domain vall=null, posl=null, valr=null, posr=null;
		Vector parmList = new Vector();
		for(;idx<Math.min(4, node.jjtGetChild(1).jjtGetNumChildren()-1);idx++)
		{	dname = ((SimpleNode)node.jjtGetChild(1).jjtGetChild(idx+1)).name;
			d = env.lookupDom(dname, true);

			if(d == null) 
				throw new InterpretError("domain "+ dname + " has not been declared");
			else if (d.isStateDom == 1) 
				throw new InterpretError("domain " + dname + " is a state domain");
			else
			{	if(d.type == STRING)
				{	if(nString == 0)
						vall = d;
					else
						valr = d;
					nString++;
				}
				else if(d.type == INTEGER)
				{	if(nInt == 0)
						posl = d;
					else
						posr = d;
					nInt++;
				}
				else
					throw new InterpretError("incorrect domain type!");
				if(node.opcode == OP_DGREP && (nString == 2 && d.type == STRING || nInt == 2 && d.type == INTEGER))
				{
//	Global.println("warning: parameter " + d.name + " is redundant in dgrep.");
 				}
 				else
				{	parmList.add(d); }
			}
		}

		// assign the parameters to domain list
		Domain [] dList = new Domain[parmList.size()];
		parmList.copyInto(dList);
	
		Domain seql, seqr;	// domain defined for mujoin
   		if((seql = env.lookupDom(DOM_SEQ_LEFT, true)) == null)
   		{	seql = new Domain(DOM_SEQ_LEFT, INTEGER);
   			env.put(seql);
   		}
   		if((seqr = env.lookupDom(DOM_SEQ_RIGHT, true)) == null)
   		{	seqr = new Domain(DOM_SEQ_RIGHT, INTEGER);
   			env.put(seqr);
   		}
		
		// grep texts into relations rl and rr
		if (vall == null)
			throw new InterpretError("Not enough parameters for text " + tl.name + ".");
		else
			rl = tl.grep(lname, pattern, vall, posl, seql);
		if (valr == null)
			throw new InterpretError("Not enough parameters for text " + tr.name + ".");
		else
			rr = tr.grep(rname, pattern, valr, posr, seqr);
				// prepare domain list for mujoin
		Domain[] domsl = new Domain[1], domsr = new Domain[1];
	    domsl[0] = vall;
	    domsr[0] = valr;	    
	    int opcode;
	    switch(node.opcode)
	    {	case OP_IGREP:
				opcode = OP_IJOIN;
				break;
			case OP_UGREP:
				opcode = OP_UJOIN;
				break;			
			case OP_SGREP:
				opcode = OP_SJOIN;
				break;
			case OP_DGREP:
				opcode = OP_DJOIN;
				break;
			case OP_LGREP:
				opcode = OP_LJOIN;
				break;
			case OP_RGREP:
				opcode = OP_RJOIN;
				break;
			default:
			    throw new InterpretError("invalid/not implemented binary grep operator!");
		}
		
		Relation rel = rl.mujoin(opcode, domsl, rr, domsr, nextTempName());
		deleteFile(lname);
		deleteFile(rname);
		
		if(rel.data == null)
			rel.load();		
		final int pvall = 0, pposl = 1, pvalr = 3, pposr = 4; // positions of val and pos attributes in the relation
		if(node.opcode != OP_DGREP)
		{	for(int i=0;i<rel.numtuples;i++)
			{	if(((int[])rel.data[pposl])[i] == INT_DC)
					((String[])rel.data[pvall])[i] = STRING_DC;				if(((int[])rel.data[pposr])[i] == INT_DC)
					((String[])rel.data[pvalr])[i] = STRING_DC;
			}
		}		
		rel.save();		
		return rel.project(dList, filename);
	}
	
	/**
	* Evaluate text-to-attr  expression
	* @param	node	User input tree
	* @param	filename	Returned relation name
	* @param	env	Environment
	*/
	private Relation evaluateText2Attr(SimpleNode node, String filename, Environment env)
										throws InterpretError
	{	Vector parmTree = new Vector();
		int nParm = node.jjtGetChild(0).jjtGetNumChildren();
		String parm;
		int n = 0, m = 0; // index used in the parameter tree
		Domain d;
		Domain tmp;		Vector projDom = new Vector();		Vector tmpDom = new Vector();

		for(int i=0;i<nParm;i++)	// in subVectors, position 0 - pattern, 1 - val, 2 - pos, 3 - seq
		{	parm = ((SimpleNode)node.jjtGetChild(0).jjtGetChild(i)).name;
			if(parm.startsWith("\""))
			{	parmTree.add(n, new Vector());				((Vector)parmTree.get(n)).add(0, new String(parm.substring(1, parm.length()-1)));
				((Vector)parmTree.get(n)).add(1, null);				((Vector)parmTree.get(n)).add(2, tmp = new Domain(nextTempName(), INTEGER));				((Vector)parmTree.get(n)).add(3, null);
				env.put(tmp);				tmpDom.add(tmp);
				n++;
			}
			else
			{	d = env.lookupDom(parm, true);
				if(d == null) 
					throw new InterpretError("domain "+ parm + " has not been declared");
				else if (d.isStateDom == 1) 
					throw new InterpretError("domain " + parm + " is a state domain");
				else 
				{	if (d.type == STRING)
						((Vector)parmTree.get(n-1)).set(1, d);
					else if(d.type == INTEGER)
						((Vector)parmTree.get(n-1)).set(3, d);
					projDom.add(d);
				}
			}
		}
		
		String textname = ((SimpleNode)node.jjtGetChild(1)).name;
		Text t = env.textTable.getText(textname);
		if(t == null)
		{	
                    	if(((SimpleNode)node.jjtGetChild(1)).info != null)
				t = new Text(nextTempName(), (String)((SimpleNode)node.jjtGetChild(1)).info, "String", env);
			else
				throw new InterpretError("Text "+ textname + " does not exist.");
		}
				// grep for each group
		Relation [] grepRel = new Relation[parmTree.size()];
		for(int i=0;i<parmTree.size();i++)
		{	grepRel[i] = t.grep(nextTempName(), 
								(String)((Vector)parmTree.get(i)).get(0),
								(Domain)((Vector)parmTree.get(i)).get(1),
								(Domain)((Vector)parmTree.get(i)).get(2),
								(Domain)((Vector)parmTree.get(i)).get(3));
		}
				// join the results
		Relation recuRel = grepRel[0];
		Domain [] dl = new Domain[1], dr = new Domain[1];
		for(int i=1;i<parmTree.size();i++)
		{	dl[0] = (Domain)((Vector)parmTree.get(i-1)).get(2);	// position
			dr[0] = (Domain)((Vector)parmTree.get(i)).get(2);
			recuRel = recuRel.rangejoin(OP_LOJOIN, dl, grepRel[i], dr, nextTempName());
		}
		
		Domain [] domArray = new Domain[projDom.size()];
		projDom.copyInto(domArray);
		
		Relation rel = recuRel.project(domArray, filename);
				// garbage collection
		for(int i = 0; i < tmpDom.size(); i++)			env.removeDom(((Domain)tmpDom.elementAt(i)).name, true);
		
		return rel;
	}
	
	/**
	* Evaluate markup-to-nest  expression
	* @param	node	User input tree
	* @param	filename	Returned relation name
	* @param	t	Input text
	* @param	env	Environment
	*/
	private Relation evaluateMu2Nest(SimpleNode node, String filename, 									Text t, Environment env)
									throws InterpretError // Jiantao 2004
	{	// parse paramaters
		String excluded = "<.*?>", tmpstr;
		Domain content, start, length;
		int currparm = 0;
		SimpleNode parmlist = (SimpleNode)node.jjtGetChild(0);
		Vector commattr = new Vector();	// list of common attrs, e.g. content, start, length
		Vector attrlist = new Vector();	// list of result attrs, parent level
		Vector leafchildattr = new Vector(); // list of attrs of leaf nodes
		Domain[] leafChildDomArray;
		Domain[] nonLeafChildDomArray = new Domain[1];		if(parmlist.isBitSet(0)) // excluded pattern
		{	tmpstr = ((SimpleNode)parmlist.jjtGetChild(currparm++)).name;
			excluded = tmpstr.substring(1, tmpstr.length()-1);
		}
		if(parmlist.isBitSet(1)) // content
		{	tmpstr = ((SimpleNode)parmlist.jjtGetChild(currparm++)).name;
			if((content = env.lookupDom(tmpstr, true)) == null)
			{	throw new InterpretError("domain " + tmpstr + " not found.");
                         }
			commattr.add(content);
		}
		else
		{	if((content = env.lookupDom(DOM_UNANALYZED, true)) == null)
			{	content = new Domain(DOM_UNANALYZED, STRING);
   				env.put(content);
   			}
		}
   		nonLeafChildDomArray[0] = content;	
 	 	if(parmlist.isBitSet(2)) // start
		{	tmpstr = ((SimpleNode)parmlist.jjtGetChild(currparm++)).name;
			if((start = env.lookupDom(tmpstr, true)) == null)
				throw new InterpretError("domain " + tmpstr + " not found.");
			else
				commattr.add(start);
		}
		if(parmlist.isBitSet(3)) // length
		{	tmpstr = ((SimpleNode)parmlist.jjtGetChild(currparm++)).name;
			if((length = env.lookupDom(tmpstr, true)) == null)
				throw new InterpretError("domain " + tmpstr + " not found.");
			else
				commattr.add(length);
		}
		attrlist.addAll(commattr);
		leafchildattr.addAll(commattr);
		leafChildDomArray = new Domain[leafchildattr.size()];
		leafchildattr.copyInto(leafChildDomArray);		
		// parse xML into a tree
		xMLParser xparser = new xMLParser(t.toString());
		SimpleNode parent = xparser.parse(null);		
		Domain childDom;
		Relation oldDomRel;
		boolean domsEqual;
		
		int prev = 0; // number of preceding characters		
		if(parent.hasAttributes())
		{	Hashtable attrs = parent.getAttributes();
			String prevchar;
			if((prevchar = (String)attrs.get(ABSOLUTE_POS)) != null)
			{	prev = Integer.parseInt(prevchar); 
			}
		}
		// 1st traversal, build the domain array for the parent node
		for(int i=0;i<parent.jjtGetNumChildren();i++)
		{	SimpleNode child = (SimpleNode)parent.jjtGetChild(i);
			String elemName = child.identifier;
			Domain[] childDomArray = child.jjtGetNumChildren() == 0 ? leafChildDomArray : nonLeafChildDomArray;
			if((childDom = env.lookupDom(elemName, true)) == null || childDom.type == UNDEF) // if childDom does not exist	
			{	childDom = new Domain(elemName, IDLIST, childDomArray, env);
				env.put(childDom);
			}
			else // childDom exists
			{	oldDomRel = env.lookupRel("." + childDom.name, true);
				if(oldDomRel == null)
					throw new InterpretError("relation ." + childDom.name + " does not exist.");
				domsEqual = true;
				if(oldDomRel.domains.length == (childDomArray.length + 1)) // check whether the new/old domains are equal
				{	for(int j=0;j<childDomArray.length;j++)
						if(!oldDomRel.domains[j+1].equals(leafChildDomArray[j])) // assume .id is in the 1st pos
						{	domsEqual = false;
							break;
						}
				}
				else
				{	domsEqual = false;	}

				if(!domsEqual)
				{
					if(childDom.numref == 0)
					{	//env.domtable.rmDom(childDom.name); // not sure
						childDom = new Domain(elemName, IDLIST, childDomArray, env);
						env.put(childDom);
					}
					else
					{	throw new InterpretError("Error: domain '" + childDom.name + "' is being used.");
					}
				}
			}
			if(!attrlist.contains(childDom))
			{	attrlist.add(childDom); }
		} // end of for
		Domain[] prntDomArray = new Domain[attrlist.size()];
		attrlist.copyInto(prntDomArray);
		Relation rel = new Relation(filename, prntDomArray, 0, prntDomArray.length, 0, env);

		Object[] prntData = Relation.createDataArray(prntDomArray, 1);
		String muStr = t.toString().replaceAll("\n", "");
		int prntPos = 0;
		long s;
		Long sur;
		// assign values to top level relation
		if(parmlist.isBitSet(1)) // content
		{	if(parmlist.isBitSet(0)) // given excluded pattern, revise content
				((String[])prntData[prntPos++])[0] = muStr.replaceAll(excluded, "");
			else
				((String[])prntData[prntPos++])[0] = muStr;
		}
		if(parmlist.isBitSet(2)) // start
		{	((int[])prntData[prntPos++])[0] = prev;
		}	
		if(parmlist.isBitSet(3)) // length
		{	((int[])prntData[prntPos++])[0] = muStr.replaceAll(excluded, "").length();
		}
				Hashtable dom2sur = new Hashtable(); // to map domain names into surrogates		// 2nd traversal, to assign values to child relations and the surrogates in the parent relation
		for(int i=0;i<parent.jjtGetNumChildren();i++)
		{	SimpleNode child = (SimpleNode)parent.jjtGetChild(i);
			String elemName = child.identifier;
			String nodeStr = (String)child.info;
			int absPos = prev + ((Integer)child.getAttributes().get(RELATIVE_POS)).intValue();
			if((sur = (Long)dom2sur.get(elemName)) == null)
			{	s = Surrogate.next(); 
				dom2sur.put(elemName, new Long(s));
				((long[])prntData[prntPos++])[0] = s;
			}
			else
			{	s = sur.longValue(); }
			Relation childRel = env.lookupRel("." + elemName, true);
			if(childRel == null)
				throw new InterpretError("relation ." + elemName + " not found.");
			Object[] childData = Relation.createDataArray(childRel.domains, 1);
			int childPos = 0;			
			if(child.jjtGetNumChildren() == 0) // leaf node
			{	String nodeVal = nodeStr.replaceAll("<.*?>", "");
				((long[])childData[childPos++])[0] = s; // surrogate		
				if(parmlist.isBitSet(1)) // content
				{	((String[])childData[childPos++])[0] = nodeVal;
				}
				if(parmlist.isBitSet(2)) // start
				{	((int[])childData[childPos++])[0] = absPos;
				}
				if(parmlist.isBitSet(3)) // length
				{	((int[])childData[childPos++])[0] = nodeVal.length();
				}
			}
			else // not leaf node
			{	String nodeVal = nodeStr.replaceFirst(">", " " + ABSOLUTE_POS + "=" + absPos + ">"); // insert preceding char info
				((long[])childData[childPos++])[0] = s; // surrogate
				((String[])childData[childPos++])[0] = nodeVal; // content or .unanalized
			}
			childRel.append(childData, 1);			
			childRel.load();
		} // end of for		
		rel.append(prntData, 1);
		rel.load();
		return rel;
	}

	// Jiantao 2004
	private void TextDeclaration(SimpleNode node, Environment env) throws InterpretError
	{
		Text t;
		String textName = ((SimpleNode)node.jjtGetChild(0)).name;		if(env.lookupRel(textName, true) != null)			throw new InterpretError("Error: texts are not allowed to have duplicate names as relations.");
		if(node.jjtGetNumChildren()==2)
		{	
			SimpleNode rightNode = (SimpleNode)node.jjtGetChild(1).jjtGetChild(0);
			
			if (rightNode.identifier == "FilePath")
			{	String inputFile = rightNode.name.substring(1, rightNode.name.length()-1);
				t = new Text(textName, inputFile, rightNode.identifier, env);
			}
			else
			{	String inputText = rightNode.name;
				t = new Text(textName, inputText, rightNode.identifier, env);
			}
		}
		else
		{	t = new Text(textName, 0, "", env);	}
		env.put(t);
	}

/*---------------------------------------------------------------------------
Summer, 2005                 Lili ZHU's code starts here :
------------------------------------------------------------------------------*/
public static int done = 0;
private void runFigfile (String xfig_file_name)
{
	try
	{
		done = 0;
		String str;
		String cmd= "xfig "+xfig_file_name;
		//System.out.println(cmd);
		Process p = Runtime.getRuntime().exec(cmd);
		
		DataInputStream in = new DataInputStream(p.getInputStream());	
		try
		{
			while ((str = in.readLine()) != null)
			{
			}
			done = 1;
		}
		catch (IOException e)
		{
			System.err.println(e);
			System.exit(0);
		}
	} 
	catch (IOException e1)
	{
		System.err.println(e1);
		System.exit(1);
	}
}

public static void cpFile(String from_file_name, String to_file_name)
{
	try
	{
		String str;
		//String cmd0 = "cp "+xfig_file_name+" _"+xfig_file_name;
		String cmd0 = "cp "+from_file_name+" "+to_file_name;
		//System.out.println(cmd0);
		
		Process p0 = Runtime.getRuntime().exec(cmd0);
		DataInputStream in0 = new DataInputStream(p0.getInputStream());
		
		try
		{
			while ((str = in0.readLine()) != null)
			{
				System.out.println(str);
			}
		}
		catch (IOException e)
		{
			System.err.println(e);
			System.exit(0);
		}
	}
	catch (IOException e1)
	{
		System.err.println(e1);
		System.exit(1);
	}
}
private boolean isDomainName( String s, Domain [] d )
{
	boolean result = false;
	for (int i = 0; i < d.length; i++)
	{
		if( s.compareTo(d[i].name) == 0 )
		{
			result = true;
			break;
		}
	}
	return result;
} 

private Domain isDomainName2( String s, Domain [] d )
{
	Domain dom = null;
	for (int i = 0; i< d.length; i++)
	{
		if( s.compareTo(d[i].name) == 0 )
		{
			dom = d[i];
			break;
		}
	}
	return dom;
}

private int domIndex( String s, Domain [] d )
{
	int index = -1;
	for ( int i = 0; i < d.length; i++ )
	{
		if( s.compareTo(d[i].name)== 0 )
		{
			index = i;
			break;
		}
	}
	return index;
} 

public boolean isStringin( String s, String [] strs ) throws InterpretError
{  
	boolean result = true;  int i = 0;
	while( i < strs.length )
	{ 
		Relation r = null;
		r = env.lookupRel(("."+s), true); 
		
		if( r != null) break;   //case: Graph(Polygon, Line), Aug, 09, 2005
		
		if( s.compareTo(".id") == 0 ) break; //case:  .rel(.id, ...)
		
		if( s.compareTo(strs[i]) == 0 ) break;
		else
		{
			if( i != strs.length -1 )
				i++;
			else
			{
				result = false;
				break;
			}
		}
	}
	return result;
}
public static Relation root;

private boolean hasIDLISTchild( String s) throws InterpretError
{
	boolean result = false;
	Relation r = null;	
	r = env.lookupRel(s, true); 
	Domain [] rdom = r.domains;
	
	/*  728 DomTable domtable = env.domtable;  
	Domain d = env.domtable.getDom(dn);	
	Vector domlist = env.domtable.getDoms(dn); 728 */

	if( r!= null )
	{
		for(int i = 0; i < r.domains.length; i++)
		{
			//String dname = (String)domlist.elementAt(j); 728
			//Domain dom = env.domtable.getDom(dname); 728
			if( rdom[i].type == IDLIST && rdom[i].name.compareTo(".id")!=0 )
			{
				result = true;
				break;
			}
		}
	}
	return result; 
}

private int rel_Type (Relation r) throws InterpretError
{
	int reltype = 1; // 1: root, 2: intermediate, 3: leaf
	boolean root = true;
	Domain [] rdom = r.domains;
	
	if( r != null )
	{
		for(int i = 0; i < r.domains.length; i++)
		{
			if( rdom[i].name.compareTo(".id") == 0 )
			{
				root = false;
				break;
			}
		}
		if(root == false)
		{
			if(hasIDLISTchild(r.name)) reltype = 2;
			else  reltype = 3;
		}
	}

	return reltype;
}

int count = 0;
public static boolean nested = false;

private void dispNestedRel(String s, Relation v, long id) throws InterpretError
{
	Relation r = null;	
	r = env.lookupRel(s, true);

	if(r!= null)
	{
		if(r.data == null)  r.load();
		Object[] data = r.data;
		int reltype = rel_Type(r);
		Domain [] rdom = r.domains;
		
		if(reltype == 1) //root
		{
			//System.out.println(s+" is ROOT.   \t  id = "+id);
			for(int i = 0; i < r.domains.length; i++)
			{
				if(rdom[i].type == IDLIST)
				{
					for(int j = 0; j < r.numtuples; j++)
					{
						dispNestedRel( ("."+rdom[i].name), v, ((long[]) data[i])[j] );
					}
				}
			}
		}
		
		if(reltype == 2) //intermediate
		{
			for(int j = 0; j < r.numtuples; j++)
			{
				if(((long[]) data[0])[j] == id)
				{
					for(int i = 1; i < r.domains.length; i++)
					{
						if(rdom[i].type == IDLIST && rdom[i].name.compareTo(".id") != 0)
						dispNestedRel( ("."+rdom[i].name), v, ((long[]) data[i])[j] );	
					}
				}
			}
		}
		
		if(reltype == 3) //leaf
		{
			//System.out.println(s+" is LEAF.   \t  id = "+id);
			//draw the leaf
			drawRel(r, v, id);
		}
	}
}

String parent = "";
private boolean hasDomain(String s, Relation r) throws InterpretError
{
	Relation root = null;
	root = env.lookupRel(s, true);

	if(root!= null)
	{
		Domain [] dom = root.domains;
		for(int i = 0; i < root.domains.length; i++)
		{
			//System.out.println("\t==== "+dom[i].name);
			if(("."+dom[i].name).compareTo(r.name) == 0)
			{
				parent = s; 
				return true;
			}
			else if(dom[i].type == IDLIST && dom[i].name.compareTo(".id") != 0)
			{
				if( hasDomain(("."+dom[i].name), r))
					return true;
			}
		}
	}
	return false;
}
public static int compare2Files(String name1, String name2)
{
	String s = "";String s1 = "";
	try
	{
		File f1 = new File(name1);
		FileInputStream fin1 = new FileInputStream(f1);
		DataInputStream in = new DataInputStream(fin1);
		
		//http://www.javacoffeebreak.com/java103/java103.html
		//read file into a string s
		while (in.available() !=0)
		{
			String tmp = in.readLine();
			s += tmp+"\n";
		}
		in.close();
	
		File f_1 = new File(name2);
		FileInputStream fin_1 = new FileInputStream(f_1);
		DataInputStream in1 = new DataInputStream(fin_1);
	
		while (in1.available() !=0)
		{
			String tmp = in1.readLine();
			s1 += tmp+"\n";
		}
		in1.close();
		//System.out.println(s+"\n-----------------\n"+s1);
	}
	catch(Exception e){}
	if(s.compareTo(s1)== 0)
		return 0;   //0: same
	else
		return 1;   //1: diff
}

static int file_num = 0;
String file_name="";
PrintStream MyOutput;
private Relation evaluateDisplay2D (SimpleNode node, String filename, Hashtable htable, Environment env) throws InterpretError
{
	int num = node. jjtGetNumChildren();
	count = 0;   nested = false;
	Relation vocab = null, pic = null;
	/*Global.println("num===="+tt);
	for(int z=0; z<tt; z++)
	{
		String kdd=((SimpleNode) (node.jjtGetChild(z))).name;
		Global.println(kdd+"****");
	} */
	file_num ++;
	file_name = file_num+".fig";
	
	try
	{
		MyOutput = new PrintStream(new FileOutputStream(file_name,true)); //apend
		if (MyOutput != null)
		{
			MyOutput.println("#FIG 3.2\nLandscape\nCenter\nMetric\nLetter  \n100.00\nSingle\n-2\n1200 2");
		}
		else
		{
			System.out.println("Error. No output file written");
			System.exit(0);
		}
	}
	catch (IOException e)
	{
		System.out.println("IOException");
	}
	
	if( num == 2) // NewPic <- display2D(VOCABULARY) PICTURE;
	{
		SimpleNode nl = (SimpleNode) node.jjtGetChild(0);
		SimpleNode nr = (SimpleNode) node.jjtGetChild(1);
		String tmpname = nextTempName();
		pic = evaluateTLExpression(nr, tmpname, htable, env);
		root= evaluateTLExpression(nr, tmpname, htable, env);
		
		tmpname = nextTempName();
		vocab = evaluateTLExpression(nl, tmpname, htable, env);
	}
	else          // NewPic <- display2D() PICTURE;
	{
		SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
		String tmpname = nextTempName();
		pic = evaluateTLExpression( nd, tmpname, htable, env );
		root = evaluateTLExpression(nd, tmpname, htable, env);
		vocab = env.lookupRel(".vocabulary", true);
		if(vocab == null) vocab.load();
	}
	
	if( pic.data == null )  pic.load();
	Object[] pic_data = pic.data;
		
	if( root.data == null )  root.load();
	
	Domain [] pic_dom = pic.domains;
	RelTable reltable = env.reltable;
	
	//Domain [] vocab_dom=vocab.domains;
	//if (env.lookupNestedRel(pic.name, true) != null)
		//System.out.println(env.getSurrogate("Graph"));
		
	//int levels = getNestedLevel(0, pic.name, env);
	//System.out.println("levels= "+levels);
	
	for(int i = 0; i < pic_dom.length; i++)
	{
		if(pic_dom[i].type == IDLIST)
		{
			nested = true;
			break;
		}
	}
	
	if(!nested)
	{
		//System.out.println("flat relation"); //no .relation
		drawRel( pic, vocab, -1); 
	}
	else  // nested relation
	{
		//System.out.println("nested relation");
		dispNestedRel(pic.name, vocab, 0);
	}
	MyOutput.close();
	
	cpFile (file_name, "_"+file_name);
	cpFile (file_name, "_#"+file_name);
	
	detectFileDiffThread t = new detectFileDiffThread(file_name);
	Thread thr = new Thread(t);
	try
	{
		thr.start();
	}
	catch (Exception e){}

	runFigfile (file_name);
	
	//System.out.println("detectFileDiffThread.polyline= "+detectFileDiffThread.polyline);
	if(detectFileDiffThread.polyline == 1 && compare2Files(file_name, "_#"+file_name) == 1)  //POLYLINE updates
	{
		//System.out.println("------------ if ------------");
		deleteFile(detectFileDiffThread.name);
		deleteFile("_#"+file_name);
		deleteFile("_"+file_name);
		return detectFileDiffThread.r;
	}
	else //no updates or updates on non_Polyline/nested relation
	{
		//System.out.println("------------ else ------------");
		if(detectFileDiffThread.polyline == 0 && compare2Files(file_name, "_#"+file_name) == 1)
		{
			// Updating Text or Nested relation : not implemented yet
			// so copy the original xfig file to the changed copy
			cpFile("_"+file_name, file_name);
		}
		deleteFile("_#"+file_name);
		deleteFile("_"+file_name);
		return pic;
	}
}
public static int polyline_flag = 0 ;
public static int text_flag = 0;
public static int up_text_flag = 0;
public static int othershape = 0;
public static boolean dotrel = false; //true : nested relation;
private void drawRel (Relation pic, Relation vocab, long id) throws InterpretError
{
	if(pic.data == null)  pic.load();
	Object[] pic_data = pic.data;
	
	if(vocab.data == null) vocab.load();
	Object[] vocab_data = vocab.data;
	
	Domain [] pic_dom = pic.domains;
	Domain [] vocab_dom = vocab.domains;
	int atr_index = domIndex( ".attribute", vocab_dom );
	text_flag = 0; up_text_flag = 0;
	String a = "", m = "", upper_rel_name = "";
	int v_idx = -1, cart_count = 0, c1 = 0, c2 = 0, x = 0,  y = 0, k = 0, num = 0;
	int max_seq = 0,  num_text = 0, num_up_text = 0, num_polyline = 0;
	int draw = 0, seq_idx = -1, first_tuple = 0, text_count = 0,up_text_count = 0;
	int gx = 0, gy = 0, g_count = 0; //group coordinate
	dotrel = false; polyline_flag = 0; text_flag = 0; up_text_flag = 0; othershape = 0;
	String [] slist = (String[]) vocab_data[atr_index];
	
	//System.out.println("-----------"+pic.name);
	for (int i = 0; i < vocab.numtuples; i++) 
	{
		a = ((String[]) vocab_data[atr_index])[i];
		m = ((String[]) vocab_data[1-atr_index])[i];
	 
		v_idx = domIndex( a, pic_dom );
		if( v_idx != -1)
		{
			if( m.compareTo("cart1") == 0) 
			{
				c1++;
			}
			else if( m.compareTo("cart2") == 0)
			{
				c2++;
			}
			else if( m.compareTo("line_colour") == 0
				|| m.compareTo("line_thickness") == 0
				|| m.compareTo("line_style") == 0
				|| m.compareTo("dash_length") == 0)
				{
					draw = 1;
				}
		}
		else;
	}
	
	if(c1 == c2 && c1 != 0) cart_count = c1;
	else //# of cart1 is not equal to # of cart2.
	{
		if(c1 != c2 && c1 == 1 && c2 > 1)
			throw new InterpretError("Not implemented.( to draw this vertical line.)");
		else if(c1 != c2 && c2 == 1 && c1 > 1)
			throw new InterpretError("Not implemented.( to draw this horizontal line.)");
		else
			throw new InterpretError("Error: Not enough information to draw the relation "+pic.name);
	}//# of cart1 is not equal to # of cart2.
	
	if( cart_count == 1 )
	{
		for (int j = 0; j < vocab.numtuples; j++)
		{
			a = ((String[]) vocab_data[atr_index])[j];
			m = ((String[]) vocab_data[1-atr_index])[j];
			seq_idx = domIndex(a, pic_dom);
			if(seq_idx != -1 && m.compareTo("sequence") == 0)
			{
				polyline_flag = 1; break;
			}
		}
	}
	 
	/*XfigObj te = new XfigObj(4,true);
	for(int i=1 ;  i<=500; i++)
	{
	te.font_size=i;
	int io=te.findLength(te.font_size);
	System.out.println(i+"     "+io);
	}    */  
	 
	for(int i = 0; i < pic_dom.length; i++)
	{
		if(pic_dom[i].name.compareTo(".id") == 0)
		{
			dotrel = true;
			break;
		}
	}

	if(dotrel == true )
	{
		for(int i = 0; i < pic.numtuples; i++)
		{
			if(id == ((long[]) pic_data[0])[i]) 
			{
				num++;
			}
		}
	}
	//ig = (Integer)tjs.elementAt(e);
	//tj= ig.intValue();

	c1 = 0; c2 = 0;
	Relation upper_rel= null;
	if(hasDomain(root.name, pic))
		upper_rel_name = parent;
	upper_rel = env.lookupRel(upper_rel_name, true);
	
	if(dotrel == false)
		num_polyline = pic.numtuples;
	else
		num_polyline = num;
 
	// ----- text detection -----
	/* 
	checking the parent of the current relation, whether there is a msg (,where msg is not in the vocabulary table)
	e.g. MsPoly(msg, Polygon), Polygon(x, y, seq) 
	or:  MsLine(msg, Lines), Lines (x1, y1, x2, y2) 
	*/
	if( dotrel )
	{
		k = 0;
		
		if(hasDomain(root.name, pic))
			upper_rel_name = parent;
		//System.out.println("parent = "+parent);
		//upper_rel_name = hasDomain(root.name, pic);
		//Relation upper_rel= null;
		 
		upper_rel = env.lookupRel(upper_rel_name, true);
		
		while( k < upper_rel.domains.length )
		{
			if(isStringin(upper_rel.domains[k].name, slist))
				k++;
			else 
			{
				up_text_flag = 1; //up_text_idx = k; 
				break;
			}
		}
		if(up_text_flag == 1) num_up_text = 1;
	}
	
	/*
	checking in the current relation, whether there is a msg
	e.g. Lines (x1, y1, x2, y2, msg)
	*/
	k = 0;
	while( k < pic_dom.length )
	{
		//System.out.println(k+" --- "+pic_dom[k].name);
		if(isStringin(pic_dom[k].name, slist)) 
			k++;
		else
		{
			text_flag = 1; //text_idx = k;
			break;
		}
	}
	if(text_flag == 1) num_text = num_polyline;
	
	//System.out.println("\ttext_flag= "+text_flag);
	//System.out.println("\tup_text_flag= "+up_text_flag);
	
	if(text_flag == 1)
	{
		k = 0;
		while( k < pic_dom.length )
		{
			//System.out.println(k+" --- "+pic_dom[k].name);
			if(isStringin(pic_dom[k].name, slist)) ;
			else  text_count++; 
			k++;
		}
		//System.out.println("\ttext_count= "+text_count);
	}
	
	int [] texts_idx = new int[text_count];
	if(text_count >= 1)  //(x, y, s1, s2, s3)
	{
		k = 0;
		//texts_idx = new int[text_count];
		int i = 0;
		while( k < pic_dom.length )
		{
			if(isStringin(pic_dom[k].name, slist));
			else
			{
				texts_idx[i] = k;
				i++;
				//System.out.println("\ttexts_idx= "+k);
			}
			k++;
		}
	}
	
	if(up_text_flag == 1)
	{
		k = 0;
		
		if(hasDomain(root.name, pic))
			upper_rel_name = parent;
		 
		upper_rel = env.lookupRel(upper_rel_name, true);

		
		while( k < upper_rel.domains.length )
		{
			if(isStringin(upper_rel.domains[k].name, slist));
			else
				up_text_count++;
			k++;
		}
		num_up_text = 1;	
		//System.out.println("up_text_count= "+up_text_count);
	}
	
	int [] up_texts_idx = new int[up_text_count];
	if(up_text_count >= 1)  //MsPoly(m1, m2, Polygon), Polygon(x, y, seq)
	{
		k = 0;
		//texts_idx = new int[text_count];
		int i = 0;
		if(hasDomain(root.name, pic))
			upper_rel_name = parent; 
		 
		upper_rel = env.lookupRel(upper_rel_name, true);
		
		while( k < upper_rel.domains.length )
		{
			if(isStringin(upper_rel.domains[k].name, slist));
			else
			{
				up_texts_idx[i] = k;
				i++;
				//System.out.println("\tup_texts_idx= "+k);
			}
			k++;
		}
	}
	
	int up_line_color_flag = 0, up_fill_color_flag = 0, up_fill_pattern_flag = 0, up_line_thickness_flag = 0, up_text_color_flag = 0, up_text_font_flag = 0, up_text_fontsize_flag = 0;
	int up_line_color_idx = 0, up_fill_color_idx = 0, up_fill_pattern_idx = 0, up_line_thickness_idx = 0, up_text_color_idx = 0, up_text_font_idx = 0, up_text_fontsize_idx = 0;
	int up_line_color = 0, up_fill_color = 0, up_fill_pattern = 0, up_line_thickness = 0, up_text_color = 0, up_text_font = 0;
	double up_text_font_size = 0.0;
	 
	Domain [] dom = null;
	Object[] data = null;
	
	if( dotrel )
	{
		for(int i = 0; i < upper_rel.domains.length; i++ )
		{
			if(isStringin(upper_rel.domains[i].name, slist)) 
			{
				for(int j = 0; j < slist.length; j++)
				{
					if(upper_rel.domains[i].name.compareTo(((String[]) vocab_data[atr_index])[j])==0)
					{
						String meaning = ((String[]) vocab_data[1-atr_index])[j];
						//System.out.println("meaning= "+meaning);
						if( meaning. compareTo("line_colour") == 0)
						{
							draw = 1; //for the case: .Polygon(x, y, seq)
								  //.MsPoly(color, Polygon)
							up_line_color_flag = 1;
							up_line_color_idx = i;
						}
						else if( meaning. compareTo("fill_colour") == 0)
						{
							draw = 1; 
							up_fill_color_flag = 1;
							up_fill_color_idx = i;
						}
						else if( meaning. compareTo("fill_pattern") == 0)
						{
							draw = 1; 
							up_fill_pattern_flag = 1;
							up_fill_pattern_idx = i;
						}
						else if( meaning. compareTo("line_thickness") == 0)
						{
							draw = 1; 
							up_line_thickness_flag = 1;
							up_line_thickness_idx = i;
						}
						else if( meaning. compareTo("text_colour") == 0)
						{
							up_text_color_flag = 1;
							up_text_color_idx = i;
						}
						else if( meaning. compareTo("font") == 0)
						{
							up_text_font_flag = 1;
							up_text_font_idx = i;
						}
						else if( meaning. compareTo("font_size") == 0)
						{
							up_text_fontsize_flag = 1;
							up_text_fontsize_idx = i;
						}
						else;
					}
				}
			}
		}
	
		dom = upper_rel.domains;
		if(upper_rel.data == null)  upper_rel.load();
		data = upper_rel.data;
		for(int p = 0; p < upper_rel.domains.length; p++)
		{
			if(("."+dom[p].name).compareTo(pic.name) == 0)
			{
				for(int q = 0; q < upper_rel.numtuples; q++)
				{
					if(id == ((long[]) data[p])[q])
					{
						if(up_line_color_flag == 1)
							up_line_color = ((int[])data[up_line_color_idx])[q];
						if(up_line_thickness_flag == 1)
							up_line_thickness = ((int[])data[up_line_thickness_idx])[q];
						if(up_fill_color_flag == 1)
							up_fill_color = ((int[])data[up_fill_color_idx])[q];
						if(up_fill_pattern_flag == 1)
							up_fill_pattern = ((int[])data[up_fill_pattern_idx])[q];
						if(up_text_color_flag == 1)
							up_text_color = ((int[])data[up_text_color_idx])[q];
						if(up_text_font_flag == 1)
							up_text_font = ((int[])data[up_text_font_idx])[q];
						if(up_text_fontsize_flag == 1)
							up_text_font_size = ((double[])data[up_text_fontsize_idx])[q];
						break;
					}
				}
			}
		}
		//System.out.println("\tup_text_color = "+up_text_color);
		/*System.out.println("\tup_line_color = "+up_line_color);
		System.out.println("\tup_fill_color = "+up_fill_color);
		System.out.println("\tup_fill_pattern = "+up_fill_pattern);*/
	}
	
	//draw adjustment for case: (x,y)
	if( text_flag == 0 && draw == 0 ) draw = 1;
	//num_text = num_text*text_count;
	
	XfigObj [] text = new XfigObj[num_text*text_count];
	XfigObj [] up_text = new XfigObj[num_up_text*up_text_count];
	XfigObj [] polyline = new XfigObj[num_polyline];
	XfigObj polygon = new XfigObj();
	
	if( cart_count == 1 )
	{
		 //POINTS
		if( polyline_flag == 0 )
		{
			int q = 0;
			if(dotrel == true)
			{
				for(int n = 0; n < pic.numtuples && q < num_polyline; n++)
				{
					/*if(q == 0  && newf == true)
						polyline[q] = new XfigObj(2,true);
					else
						polyline[q] = new XfigObj(2,false);*/
					polyline[q] = new XfigObj(2);
					polyline[q].num_points = 1;
					polyline[q].points = new int[1][2];
					
					// if(idMatch(pic, ((long[]) pic_data[0])[n]))    728
					if(id == ((long[]) pic_data[0])[n]) 
					{
						for (int j = 0; j < vocab.numtuples; j++) 
						{
							a = ((String[]) vocab_data[atr_index])[j];
							
							m = ((String[]) vocab_data[1-atr_index])[j];
					
							v_idx = domIndex(a, pic_dom);
							if( v_idx != -1 )
							{
								x = ((int[]) pic_data[v_idx])[n];
								
								if(m.compareTo("cart1") == 0)
									polyline[q].points[0][0] = x;
								else if( m.compareTo("cart2") == 0)
									polyline[q].points[0][1] = x;
								else if( m.compareTo("line_colour") == 0)
									polyline[q].line_color = x;
								else if( m.compareTo("line_thickness") == 0)
									polyline[q].thickness = x;
								else if( m.compareTo("line_style") == 0)
									polyline[q].pen_style = x;
								else if( m.compareTo("fill_colour") == 0)
									polyline[q].fill_color = x;
								else;
							}
						}
						gx = gx + polyline[q].points[0][0];
						gy = gy + polyline[q].points[0][1];
						g_count ++;
						
						if(up_line_color_flag == 1)
							polyline[q].line_color = up_line_color;
						if(up_line_thickness_flag == 1)
							polyline[q].thickness = up_line_thickness;
						q++;
					}
				} 
			}
			else  //dotrel == false
			{
				for (int i = 0; i < num_polyline; i++)
				{
					/*if( i == 0 && newf == true)
						polyline[i] = new XfigObj(2,true);
					else
						polyline[i] = new XfigObj(2,false);  728*/
					polyline[i] = new XfigObj(2);
					polyline[i].num_points = 1;
					polyline[i].points = new int[1][2];
					for (int j = 0; j < vocab.numtuples; j++)
					{
						a = ((String[]) vocab_data[atr_index])[j];
						//a= x, y....
						//System.out.println("--i: "+i+" vocab = "+a);
				
						m = ((String[]) vocab_data[1-atr_index])[j];
						//m= cart1, cart2...
				
						v_idx = domIndex(a, pic_dom);
						if( v_idx != -1 )
						{
							x = ((int[]) pic_data[v_idx])[i];
							//System.out.println("--i: "+i+" pic = "+x);
								
							if(m.compareTo("cart1") == 0)   
								polyline[i].points[0][0] = x;
								//System.out.println("cart1 "+x);
							else if( m.compareTo("cart2") == 0)
								polyline[i].points[0][1] = x;
								//System.out.println("cart2 "+x);
							else if( m.compareTo("line_colour") == 0)
								polyline[i].line_color = x;
							else if( m.compareTo("line_thickness") == 0)
								polyline[i].thickness = x;
							else if( m.compareTo("line_style") == 0)
								polyline[i].pen_style = x;
							else if( m.compareTo("fill_colour") == 0)
								polyline[i].fill_color = x;
							else;
						}
					} 
				} 
			}
			if(draw == 1)
			{
				for (int i = 0; i < num_polyline; i++)
				{
					String str = polyline[i].outFigFile_Objs();
					MyOutput.println(str);
				}
			}
		} //end of Points
		
		//POLYGON: either open or close
		//open: last point != first point
		//close last point == first point
		else
		{
			int line_color_warning = 0, fill_color_warning = 0;
			if( !dotrel )   // polygon not from nested relation
			{
				//System.out.println(" in polygon: flat ");
				max_seq = 0;
				polygon = new XfigObj(2);
				max_seq = pic.numtuples;
				/*for (int i = 0; i < pic.numtuples; i++) 
				{
					if( max_seq < ((int[]) pic_data[seq_idx])[i] )
						max_seq = ((int[]) pic_data[seq_idx])[i];
				}*/
				//System.out.println("\t max_seq= "+max_seq);
				polygon.num_points = max_seq ;//+ 1;
				polygon.points = new int[max_seq][2]; //+1][2];
		
				for (int i = 0; i < max_seq; i++) 
				{
					int tmp = ((int[]) pic_data[seq_idx])[i];
					for (int j = 0; j < vocab.numtuples; j++) 
					{
						a = ((String[]) vocab_data[atr_index])[j];
						m = ((String[]) vocab_data[1-atr_index])[j];
		
						v_idx = domIndex(a, pic_dom);
						if( v_idx != -1)
						{
							x = ((int[]) pic_data[v_idx])[i];
							if(m.compareTo("cart1") == 0) 
							{
								polygon.points[tmp-1][0] = x;
							}
							else if( m.compareTo("cart2") == 0)
							{
								polygon.points[tmp-1][1] = x;
							}
							else if( m.compareTo("line_colour") == 0)
							{
								line_color_warning = 1;
								polygon.line_color = ((int[]) pic_data[v_idx])[0];
							}
							else if( m.compareTo("line_thickness") == 0)
								polygon.thickness = x;
							else if( m.compareTo("line_style") == 0)
								polygon.pen_style = x;
							else if( m.compareTo("fill_colour") == 0)
							{
								fill_color_warning = 1;
								polygon.fill_color = ((int[]) pic_data[v_idx])[0];
								polygon.fill_pattern = 20;  //solid color
							}
							else if( m.compareTo("fill_pattern") == 0)
							{
								polygon.fill_pattern = x;
							}
							else;
						}
					}
					//polygon.points[ max_seq][0] = polygon.points[0][0];
					//polygon.points[ max_seq][1] = polygon.points[0][1];
				}
				if(draw == 1)
				{
					if(line_color_warning == 1)
						Global.println("Warning: only the first color seen will be used as the line color.");
					if(line_color_warning == 1)
						Global.println("Warning: only the first color seen will be used as the fill color.");
					String str = polygon.outFigFile_Objs();
					MyOutput.println(str);
				}
			}
			else  //polygon from nested relation
			{
				//System.out.println( "in polygon: dotrel");	
				max_seq = 0;
				polygon = new XfigObj(2);
				
				for (int i = 0; i < pic.numtuples; i++) 
				{
					if( ((long[]) pic_data[0])[i] == id) 
					{
						first_tuple = i; break;
					}
				}
				//System.out.println("\t 1st tuple= "+first_tuple);
				for (int i = 0; i < pic.numtuples; i++) 
				{
					if( max_seq < ((int[]) pic_data[seq_idx])[i] && ((long[]) pic_data[0])[i] == id)
						max_seq = ((int[]) pic_data[seq_idx])[i];
				}
				
				polygon.num_points = max_seq; //+ 1;
				polygon.points = new int[max_seq][2]; //+ 1][2];
				
				for (int i = first_tuple; i < first_tuple + max_seq; i++)
				{
					int tmp = ((int[]) pic_data[seq_idx])[i];
					for (int j = 0; j < vocab.numtuples; j++)
					{
						a = ((String[]) vocab_data[atr_index])[j];
						m = ((String[]) vocab_data[1-atr_index])[j];
		
						v_idx = domIndex(a, pic_dom);
						if( v_idx != -1)  
						{
							x = ((int[]) pic_data[v_idx])[i]; 
							if(m.compareTo("cart1") == 0) 
							{
								polygon.points[tmp-1][0] = x;
							}
							else if( m.compareTo("cart2") == 0)
							{
								polygon.points[tmp-1][1] = x; 	
							}
							else if( m.compareTo("line_colour") == 0)
							{
								line_color_warning = 1; //first seen color, need warning msg
								polygon.line_color = ((int[]) pic_data[v_idx])[first_tuple]; 
							 
							}
							else if( m.compareTo("line_thickness") == 0)
								polygon.thickness = x;
							else if( m.compareTo("line_style") == 0)
								polygon.pen_style = x;
							else if( m.compareTo("fill_colour") == 0)
							{
								fill_color_warning = 1; //first seen color, need warning msg
								polygon.fill_color = ((int[]) pic_data[v_idx])[first_tuple];
								polygon.fill_pattern = 20;  //solid color
							}
							else if( m.compareTo("fill_pattern") == 0)
							{
								polygon.fill_pattern = x;
							}
							else;
						}
					}
					//polygon.points[ max_seq][0] = polygon.points[0][0];
					//polygon.points[ max_seq][1] = polygon.points[0][1];
				}
				if(up_line_color_flag == 1)
					polygon.line_color = up_line_color;
				if(up_line_thickness_flag == 1)
					polygon.thickness = up_line_thickness;
				if(up_fill_color_flag == 1)
				{
					polygon.fill_color = up_fill_color;
					polygon.fill_pattern = 20;
					if(up_line_color_flag == 0)  //if no line color is given, but fill color
					{
						polygon.line_color = up_fill_color; //then linecolor = fillcolor
					}
				}
				if(up_fill_pattern_flag == 1)
					polygon.fill_pattern = up_fill_pattern;
					
				if(draw == 1)
				{
					if(line_color_warning == 1 && up_line_color_flag == 0)
						Global.println("Warning: only the first color seen will be used as the line color.");
					if(fill_color_warning == 1 && up_fill_color_flag == 0)
						Global.println("Warning: only the first color seen will be used as the fill color.");
					if(line_color_warning == 1 && up_line_color_flag == 1)
						Global.println("Warning: top level line color will be used as the line color.");
					if(fill_color_warning == 1 && up_fill_color_flag == 1)
						Global.println("Warning: top level fill color will be used as the fill color");	
					String str = polygon.outFigFile_Objs(); 
					MyOutput.println(str);
				}
			}
			 if(up_text_flag == 1)
			{
				double area = 0.00;  int tp_x = 0, tp_y = 0;
				if(polygon.points[max_seq-1][0] != polygon.points[0][0] && polygon.points[max_seq-1][1] != polygon.points[0][1])
				{
					tp_x = polygon.points[0][0];
					tp_y = polygon.points[0][1];
				}
				for (int i = 0; i < max_seq; i++)
				{
					if( i == max_seq-1)
					{
						area += polygon.points[i][0] * tp_y;
						area -= polygon.points[i][1] * tp_x;
					}
					else
					{
					area += polygon.points[i][0] * polygon.points[i+1][1];
					area -= polygon.points[i][1] * polygon.points[i+1][0];
					}
				}
			
				area /= 2.0; 
				if(area < 0) area = -area;
				//System.out.println("-- "+area);
				
				for (int i = 0; i < max_seq; i++)
				{
					if( i == max_seq-1)
					{
						long factor = polygon.points[i][0]*tp_y - polygon.points[i][1]* tp_x;
						polygon.centroid_x += factor * (polygon.points[i][0] + tp_x);
						polygon.centroid_y += factor * (polygon.points[i][1] + tp_y);
					}
					else
					{
						long factor = polygon.points[i][0]*polygon.points[i+1][1] - polygon.points[i][1]* polygon.points[i+1][0];
						//System.out.println("fac = "+factor);
						polygon.centroid_x += factor * (polygon.points[i][0] + polygon.points[i+1][0]);
						polygon.centroid_y += factor * (polygon.points[i][1] + polygon.points[i+1][1]);
					}
				}
				
				polygon.centroid_x = Math.abs((long) (polygon.centroid_x/ (6*area)));
				polygon.centroid_y = Math.abs((long) (polygon.centroid_y/ (6*area)));
				//in xfig display, no negative coordinate, therefore,  Math.abs(...)
				
				//System.out.println("--x "+polygon.centroid_x);
				//System.out.println("--y "+polygon.centroid_y);
			}
		} // end of polygon
	} //end of cart_count = 1
	
	//STRAIGHT LINE
	else if( cart_count == 2 )
	{
		int q = 0;
		
		if(dotrel == true ) //nested
		{
			//System.out.println("dotrel = true ");
			for(int n = 0; n < pic.numtuples && q < num_polyline; n++)
			{
				polyline[q] = new XfigObj(2);
				polyline[q].num_points = 2;
				polyline[q].points = new int[2][2];
				
				 if(id == ((long[]) pic_data[0])[n]) 
				{
					for (int j = 0; j < vocab.numtuples; j++) 
					{
						a = ((String[]) vocab_data[atr_index])[j];
						m = ((String[]) vocab_data[1 - atr_index])[j];
					
						v_idx = domIndex( a, pic_dom );
						if( v_idx != -1)
						{
							x= ((int[]) pic_data[v_idx])[n];
							if(m.compareTo("cart1")==0) 
							{
								polyline[q].points[c1][0] = x;  c1 ++;
							}
							else if(m.compareTo("cart2")==0)
							{
								polyline[q].points[c2][1] = x;  c2 ++;
							}
							else if(m.compareTo("line_colour")== 0)
								polyline[q].line_color = x;
							else if(m.compareTo("line_thickness")== 0)
								polyline[q].thickness = x;
							else if( m.compareTo("line_style") == 0)
								polyline[q].pen_style = x; 
							else if( m.compareTo("fill_colour") == 0)
								polyline[q].fill_color = x;
							else;
						}
					}
					gx = gx + (polyline[q].points[0][0]+polyline[q].points[1][0])/2;
					gy = gy + (polyline[q].points[0][1]+polyline[q].points[1][1])/2;
					g_count ++;
					
					if(up_line_color_flag == 1)
						polyline[q].line_color = up_line_color;
					if(up_line_thickness_flag == 1)
						polyline[q].thickness = up_line_thickness;
					c1 = 0; c2 = 0; q++;
				}
			}
		}
		else  //flat
		{
			//System.out.println("dotrel = false ");
			for (int i = 0; i < num_polyline; i++)
			{
				polyline[i] = new XfigObj(2);
				polyline[i].num_points=2;
				polyline[i].points=new int[2][2];
				
				for (int j = 0; j < vocab.numtuples; j++)
				{
					a = ((String[]) vocab_data[atr_index])[j];
					m = ((String[]) vocab_data[1-atr_index])[j];
			
					v_idx=domIndex(a, pic_dom);
					if(v_idx!=-1)
					{
						x= ((int[]) pic_data[v_idx])[i];
						if(m.compareTo("cart1")==0)
						{
							polyline[i].points[c1][0]=x;  c1++;
						}
						else if(m.compareTo("cart2")==0)
						{
							polyline[i].points[c2][1]=x;  c2++;
						}
						else if(m.compareTo("line_colour")==0)
							polyline[i].line_color=x;
						else if(m.compareTo("line_thickness")==0)
							polyline[i].thickness=x;
						else if( m.compareTo("line_style") == 0)
							polyline[i].pen_style = x; 
						else if( m.compareTo("fill_colour") == 0)
							polyline[i].fill_color = x;
						else;
						}
				}
				c1=0; c2=0;
			}
		}
		
		//compute the slope of each line for text ==> angle
		for (int i = 0; i < num_polyline; i++) 
		{
			if( polyline[i].points[1][0]-polyline[i].points[0][0] != 0)
			{
				double slope = (double)(polyline[i].points[1][1]-polyline[i].points[0][1])/
				(polyline[i].points[1][0]-polyline[i].points[0][0]);
				polyline[i].angle = -Math.atan( slope );
				//System.out.println("== "+ Math.toDegrees(-Math.atan(slope))); 
			}
			else
				polyline[i].angle = Math.PI/2;
			
			if(text_flag == 1)
			{
				polyline[i].centroid_x = (polyline[i].points[0][0]+polyline[i].points[1][0])/2;
				polyline[i].centroid_y = (polyline[i].points[0][1]+polyline[i].points[1][1])/2;
			}
		}
		if(draw == 1)
		{
			for (int i = 0; i < num_polyline; i++) 
			{
				String str = polyline[i].outFigFile_Objs();
				MyOutput.println(str);
			}
		}
	} // end of Straight line (cart_count = 2)
 
	//TRIANGLE
	else if( cart_count == 3 )
	{ 
		int q = 0;
		if(dotrel == true)
		{
			for(int n = 0; n < pic.numtuples && q < num_polyline; n++)
			{
				polyline[q] = new XfigObj(2);
				polyline[q].num_points = 4;
				polyline[q].points = new int[4][2];
				
				//if(idMatch(pic, ((long[]) pic_data[0])[n])) 728
				if(id == ((long[]) pic_data[0])[n]) 
				{
					for (int j = 0; j < vocab.numtuples; j++)
					{
						a = ((String[]) vocab_data[atr_index])[j];
						m = ((String[]) vocab_data[1-atr_index])[j];
				
						v_idx = domIndex(a, pic_dom);
						if(v_idx != -1)  
						{
							x = ((int[]) pic_data[v_idx])[n];
							if(m.compareTo("cart1")== 0)
							{
								polyline[q].points[c1][0] = x; c1++;
							}
							else if(m.compareTo("cart2")== 0)
							{
								polyline[q].points[c2][1] = x;c2++;
							}
							else if(m.compareTo("line_colour")== 0)
								polyline[q].line_color = x;
							else if(m.compareTo("line_thickness")== 0)
								polyline[q].thickness = x;
							else if( m.compareTo("line_style") == 0)
								polyline[q].pen_style = x; 
							else if(m.compareTo("fill_colour")== 0)
							{
								polyline[q].fill_color = x; //((int[]) pic_data[v_idx])[0]; Oct 3, 2005
								polyline[q].fill_pattern = 20;  //solid color
							}
							else if(m.compareTo("fill_pattern")== 0)
							{
								polyline[q].fill_pattern = x;
							}
							else;
						}
					}
					
					polyline[q].points[3][0] = polyline[q].points[0][0];
					polyline[q].points[3][1] = polyline[q].points[0][1];
					
					if(up_line_color_flag == 1)
						polyline[q].line_color = up_line_color;
					if(up_line_thickness_flag == 1)
						polyline[q].thickness = up_line_thickness;
					 
					gx = gx + (polyline[q].points[0][0]+polyline[q].points[1][0]+polyline[q].points[2][0])/3;
					gy = gy + (polyline[q].points[0][1]+polyline[q].points[1][1]+polyline[q].points[2][1])/3;
					g_count ++;
					
					c1 = 0; c2 = 0; q++;
				}
			}
		}
		else //flat ==> dotrel == false
		{
			for (int i = 0; i < num_polyline; i++) 
			{
				polyline[i] = new XfigObj(2);
				polyline[i].num_points = 4;
				polyline[i].points = new int[4][2];
				for (int j = 0; j < vocab.numtuples; j++) 
				{
					a = ((String[]) vocab_data[atr_index])[j];
					//a= x, y....
					//System.out.println("--i: "+i+" vocab = "+a);
			
					m = ((String[]) vocab_data[1-atr_index])[j];
					//m= cart1, cart2...
			
					v_idx = domIndex(a, pic_dom);
					if(v_idx != -1)  
					{
						x = ((int[]) pic_data[v_idx])[i];
						//System.out.println("--i: "+i+" pic = "+x);
						//System.out.println("--i: "+i+" meaning = "+m); 
						if(m.compareTo("cart1")== 0) 
						{       //System.out.println("                     c1=  "+c1); 
							polyline[i].points[c1][0] = x; 
							c1++;
							//System.out.println("cart1 "+x); 
						}
						else if(m.compareTo("cart2")== 0)
						{
							polyline[i].points[c2][1] = x;
							c2++;
							//System.out.println("cart2 "+x);
						}
						else if(m.compareTo("line_colour")== 0)
							polyline[i].line_color = x;
						else if(m.compareTo("line_thickness")== 0)
							polyline[i].thickness = x;
						else if( m.compareTo("line_style") == 0)
							polyline[i].pen_style = x; 
						else if(m.compareTo("fill_colour")== 0)
						{
							polyline[i].fill_color = x; //((int[]) pic_data[v_idx])[0]; Oct 3, 2005
							polyline[i].fill_pattern = 20;  //solid color
						}
						else if(m.compareTo("fill_pattern")== 0)
						{
							polyline[i].fill_pattern = x;
						}
						else;
					}
				} 
				c1 = 0; c2 = 0;
				polyline[i].points[3][0] = polyline[i].points[0][0];
				polyline[i].points[3][1] = polyline[i].points[0][1];
			} 
		}
		if( text_flag == 1 )
		{ 
			for (int i = 0; i < num_polyline; i++)
			{
			  polyline[i].centroid_x = (polyline[i].points[0][0]+polyline[i].points[1][0]+polyline[i].points[2][0])/3;
			  polyline[i].centroid_y = (polyline[i].points[0][1]+polyline[i].points[1][1]+polyline[i].points[2][1])/3;
			}
		}
		if(draw == 1)
		{
			for (int i = 0; i < num_polyline; i++)
			{ 
				String str = polyline[i].outFigFile_Objs(); 
				MyOutput.println(str);
			}
		}
	} //end of Triangle (cart_count = 3)
	
	else othershape = 1;  //other shapes: arc, spline...implementatino is coming later
	
	//TEXT
	if( text_flag == 1 )
	{
		if( dotrel ) // .rel ==> .Line (.id, x1, y1, x2, y2, msg)
		{
			int n = 0;
			for (int q = 0; q < num_text && n < pic.numtuples; q++)
			{
			for (int p = 0; p < text_count; p++)
			{
				//System.out.println("n= "+n+" q = "+q+" p= "+ p+" q*text_count + p = "+(q*text_count + p));
				text[q*text_count + p] = new XfigObj(4); 
				text[q*text_count + p].sub_type = 0;
				//text[q] = new XfigObj(4); 
				//text[q].sub_type = 0;
				//text[i].length=text[i].findLength(text[i].font_size)*text[i].s.length();
				//System.out.println( "  ... "+text[i].length);
				if(id == ((long[]) pic_data[0])[n]) 
				{
					for (int j = 0; j < vocab.numtuples; j++) 
					{
						a = ((String[]) vocab_data[atr_index])[j];
						m = ((String[]) vocab_data[1-atr_index])[j];
					
						v_idx = domIndex(a, pic_dom);
						if(v_idx != -1)
						{
							x = ((int[]) pic_data[v_idx])[n]; 	
							//text[q].s = ((String[]) pic_data[text_idx])[n];
							if(m.compareTo("cart1") == 0) 
							{
								if(cart_count== 1)  text[q*text_count + p].pos_x = x + 180;
								//move text to the right of distance 180 (length of 2 chars)
							}
							else if(m.compareTo("cart2") == 0)
							{
								if(cart_count== 1)   text[q*text_count + p].pos_y = x + 180;
							}
							else if(m.compareTo("text_colour") == 0)
							{
								text[q*text_count + p].text_color = x;
							}
							else if(m.compareTo("font") == 0)
							{
								text[q*text_count + p].font = x;
							}
							else if(m.compareTo("font_size") == 0)
							{
								text[q*text_count + p].font_size = (double) x;
							}
							else;
						}
						int types =pic_dom[texts_idx[p]].type;
						switch(types) //Aug 20, 2005, Move this out from the above "if"
						{
							case STRING:
							text[q*text_count + p].s = ((String[]) pic_data[texts_idx[p]])[n];
							break;
							case INTEGER:
							text[q*text_count + p].s =""+((int[]) pic_data[texts_idx[p]])[n];
							break;
							case SHORT:
							text[q*text_count + p].s =""+((short[]) pic_data[texts_idx[p]])[n];
							break;
							case LONG:
							text[q*text_count + p].s =""+((long[]) pic_data[texts_idx[p]])[n];
							break;
							case FLOAT:
							text[q*text_count + p].s =""+((float[]) pic_data[texts_idx[p]])[n];
							break;
							case DOUBLE:
							text[q*text_count + p].s =""+((double[]) pic_data[texts_idx[p]])[n];
							break;
						}
						//text[q*text_count + p].s = ((String[]) pic_data[texts_idx[p]])[n];
						//System.out.println(" - text=  "+text[q*text_count + p].s);
					}
					text[q*text_count + p].length = 
					text[q*text_count + p].findLength(text[q*text_count + p].font_size)*text[q*text_count + p].s.length();
					
					if( cart_count == 2 )
					{
						text[q*text_count + p].angle = polyline[q].angle;
						
						text[q*text_count + p].pos_x = polyline[q].centroid_x-(int)(text[q*text_count + p].findLength(text[q*text_count + p].font_size)*text[q*text_count + p].s.length()/2*Math.cos(text[q*text_count + p].angle));
						
						//Aug 18, 2005: New Adjustent to the pos_x
						if(text[q*text_count + p].angle > 0.0)
							text[q*text_count + p].pos_x = 
							text[q*text_count + p].pos_x - (int)(180* Math.sin( text[q*text_count + p].angle));
						else
							text[q*text_count + p].pos_x = 
							text[q*text_count + p].pos_x + (int)(180* Math.sin(-text[q*text_count + p].angle));
							
						text[q*text_count + p].pos_y = polyline[q].centroid_y+(int)(text[q*text_count + p].findLength(text[q*text_count + p].font_size)*text[q*text_count + p].s.length()/2*Math.sin(text[q*text_count + p].angle));
						
						//Aug 18, 2005: New Adjustent to the pos_y
						text[q*text_count + p].pos_y =  
						text[q*text_count + p].pos_y - (int)(90* Math.cos(text[q*text_count + p].angle)); 
						
					}
					if( cart_count == 3 )
					{
						text[q*text_count + p].pos_x = polyline[q].centroid_x;
						text[q*text_count + p].pos_y = polyline[q].centroid_y;
					}
					
					text[q*text_count + p].depth = 49 - p;
					if(text[q*text_count + p].depth < 0)
						Global.println("warning: Text depth must be >= 0");
						//i.e. the maximum number of strings in a relation as atrributes must be <=50
						// (x, y, s1, s2....s50)
					
					text[q*text_count + p].pos_x = text[q*text_count + p].pos_x + 27*p;
					text[q*text_count + p].pos_y = text[q*text_count + p].pos_y + 27*p;
				
					String str = text[q*text_count + p].outFigFile_Text();
					MyOutput.println(str);
				}
			} n++;
			}
		}
		else //flat rel: Line (x1, y1, x2, y2, s1)
		{
			for (int n = 0; n < num_text ; n++)
			{
			for (int p = 0; p < text_count ; p++)
			{
				text[n*text_count + p] = new XfigObj(4);
				text[n*text_count + p].sub_type = 0;
				//text[i].length=text[i].findLength(text[i].font_size)*text[i].s.length();
				//System.out.println( "  ... "+text[i].length);
				 
				for (int j = 0; j < vocab.numtuples; j++)
				{
					a = ((String[]) vocab_data[atr_index])[j];
					m = ((String[]) vocab_data[1-atr_index])[j];
				
					v_idx = domIndex(a, pic_dom); 
					if(v_idx != -1)
					{
						x = ((int[]) pic_data[v_idx])[n];
						
						if(m.compareTo("cart1") == 0)
						{
							if(cart_count== 1)   text[n*text_count + p].pos_x = x + 180;
							//move text to the right of distance 180 (length of 2 chars)
						}
						else if(m.compareTo("cart2") == 0)
						{
							if(cart_count== 1)   text[n*text_count + p].pos_y = x + 180;
						}
						else if(m.compareTo("text_colour") == 0)
						{
							text[n*text_count + p].text_color = x;
						}
						else if(m.compareTo("font") == 0)
						{
							text[n*text_count + p].font = x;
						}
						else if(m.compareTo("font_size") == 0)
						{
							text[n*text_count + p].font_size = (double)x;
						}
						else;
					}
					
					int types =pic_dom[texts_idx[p]].type;
					switch(types) //Aug 20, 2005, Move this out from the above "if"
					{
						case STRING: text[n*text_count + p].s = ((String[]) pic_data[texts_idx[p]])[n]; break;
						case INTEGER: text[n*text_count + p].s =""+((int[]) pic_data[texts_idx[p]])[n]; break;
						case SHORT: text[n*text_count + p].s =""+((short[]) pic_data[texts_idx[p]])[n]; break;
						case LONG:  text[n*text_count + p].s =""+((long[]) pic_data[texts_idx[p]])[n];  break;
						case FLOAT: text[n*text_count + p].s =""+((float[]) pic_data[texts_idx[p]])[n]; break;
						case DOUBLE:text[n*text_count + p].s =""+((double[])pic_data[texts_idx[p]])[n]; break;
					}
					//System.out.println( " s= ... "+((Object[]) pic_data[texts_idx[p]])[n]);
					//text[n*text_count + p].s = ((String[]) pic_data[texts_idx[p]])[n];
					//text[n].s = ((String[]) pic_data[text_idx])[n];
					//System.out.println( "  ... "+text[n*text_count + p].s);
				}
				text[n*text_count + p].length = 
				text[n*text_count + p].findLength(text[n*text_count + p].font_size)*text[n*text_count + p].s.length();
				
				if( cart_count == 2 )
				{
					text[n*text_count + p].angle = polyline[n].angle;
					
					text[n*text_count + p].pos_x = 
					polyline[n].centroid_x-(int)(text[n*text_count + p].findLength(text[n*text_count + p].font_size)*text[n*text_count + p].s.length()/2*Math.cos(text[n*text_count + p].angle));
					
					//Aug 18, 2005: New Adjustent to the pos_x
					if(text[n*text_count + p].angle > 0.0)   
						text[n*text_count + p].pos_x = 
						text[n*text_count + p].pos_x - (int)(180* Math.sin(text[n*text_count + p].angle));
					else
						text[n*text_count + p].pos_x = 
						text[n*text_count + p].pos_x + (int)(180* Math.sin(-text[n*text_count + p].angle));
						
					text[n*text_count + p].pos_y = polyline[n].centroid_y+(int)(text[n*text_count + p].findLength(text[n*text_count + p].font_size)*text[n*text_count + p].s.length()/2*Math.sin(text[n*text_count + p].angle));
					
					//Aug 18, 2005: New Adjustent to the pos_y
					text[n*text_count + p].pos_y =  
					text[n*text_count + p].pos_y - (int)(180* Math.cos(text[n*text_count + p].angle));
				}
				if( cart_count == 3 )
				{
					text[n*text_count + p].pos_x = polyline[n].centroid_x;
					text[n*text_count + p].pos_y = polyline[n].centroid_y;
				}
				
				//adjust the depth and the position of a stack of texts
				text[n*text_count + p].depth = 49 - p;
				if(text[n*text_count + p].depth < 0)
					Global.println("warning: Text depth must be >= 0.");
					//i.e. the maximum number of strings in a relation as atrributes must be <=50
					// (x, y, s1, s2....s50)
					
				text[n*text_count + p].pos_x = text[n*text_count + p].pos_x + 27*p;
				text[n*text_count + p].pos_y = text[n*text_count + p].pos_y + 27*p;
				
				String str = text[n*text_count + p].outFigFile_Text(); 
				MyOutput.println(str);
			}
			}
		}
	}
	
	//1 level up : TEXT
	if( up_text_flag == 1 )
	{
		//System.out.println("up_text_flag= 1");
		if(hasDomain(root.name, pic))
			upper_rel_name = parent;
		upper_rel = null;
		upper_rel = env.lookupRel(upper_rel_name, true);
		dom = upper_rel.domains;
		if(upper_rel.data == null)  upper_rel.load();
		data = upper_rel.data;
		
		for (int i = 0; i < num_up_text; i++)
		{
		for (int n = 0; n < up_text_count ; n++)
		{
			up_text[i*up_text_count + n] = new XfigObj(4);
			up_text[i*up_text_count + n].sub_type = 0;

			for (int j = 0; j < vocab.numtuples; j++) 
			{
				a = ((String[]) vocab_data[atr_index])[j];
				m = ((String[]) vocab_data[1-atr_index])[j];
				
				v_idx = domIndex(a, dom);
				if(v_idx != -1)
				{
					x = ((int[]) data[v_idx])[i];
					
					if(m.compareTo("text_colour") == 0)
					{
						up_text[i*up_text_count + n].text_color = up_text_color;
					}
					else if(m.compareTo("font") == 0)
					{
						up_text[i*up_text_count + n].font = up_text_font;
					}
					else if(m.compareTo("font_size") == 0)
					{
						up_text[i*up_text_count + n].font_size = up_text_font_size;
					}
					else;
				} 
				for(int p = 0; p < upper_rel.domains.length; p++) //Aug 20, 2005, Move this out from the above "if"
				{
					if(("."+dom[p].name).compareTo(pic.name) == 0)
					{
						for(int q = 0; q < upper_rel.numtuples; q++)
						{
						if(id == ((long[]) data[p])[q])
						{
							int types = dom[up_texts_idx[n]].type;
							switch(types) 
							{
							case STRING: 
							up_text[i*up_text_count + n].s 
							= ((String[])data[up_texts_idx[n]])[q]; break;
							case INTEGER:
							up_text[i*up_text_count + n].s 
							=""+((int[])data[up_texts_idx[n]])[q]; break;
							case SHORT: 
							up_text[i*up_text_count + n].s =""+((short[])data[up_texts_idx[n]])[q]; break;
							case LONG:  
							up_text[i*up_text_count + n].s =""+((long[])data[up_texts_idx[n]])[q]; break;
							case FLOAT: 
							up_text[i*up_text_count + n].s =""+((float[])data[up_texts_idx[n]])[q]; break;
							case DOUBLE: 
							up_text[i*up_text_count + n].s =""+((double[])data[up_texts_idx[n]])[q]; break;
							}
							break;
						}
						}
					}
				}
				
			}   //System.out.println( " 4;;;;;;;;;");System.out.println("======mm "+up_text[i*up_text_count + n].s);
			up_text[i*up_text_count + n].length = 
			up_text[i*up_text_count + n].findLength(up_text[i*up_text_count + n].font_size)*up_text[i*up_text_count + n].s.length();
			
			if( polyline_flag == 1 )
			{
				up_text[i*up_text_count + n].pos_x = polygon.centroid_x;
				up_text[i*up_text_count + n].pos_y = polygon.centroid_y;
			}
			else
			{
				up_text[i*up_text_count + n].pos_x = gx / g_count; 
				up_text[i*up_text_count + n].pos_y = gy / g_count;
				//System.out.println(up_text[i].s+" -- pos_x= "+up_text[i].pos_x+"  pos_x= "+up_text[i].pos_y);
			}
			
			up_text[i*up_text_count + n].depth = 49 - n;
			if(up_text[i*up_text_count + n].depth < 0)
				Global.println("warning: Text depth must be >= 0.");
				//i.e. the maximum number of strings in a relation as atrributes must be <=50
				// (x, y, s1, s2....s50)
				
			up_text[i*up_text_count + n].pos_x = up_text[i*up_text_count + n].pos_x + 27*n;
			up_text[i*up_text_count + n].pos_y = up_text[i*up_text_count + n].pos_y + 27*n;
			String str = up_text[i*up_text_count + n].outFigFile_Text();
			MyOutput.println(str);
		}
		}
	}
}
//End of Lili's Code. 2005
}













